/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "colcode.h"
#include "browse.h"
#include "tags.h"

// If this line doesn't contain want and a nearby one does, go there
// and return its number.  Otherwise return -1.
int adjust_found_bsc_line(char *want)
{
	char pat[FNAMELEN], *s;
	int orig = point;

	while (s = strchr(want, ':'))	// Convert foo::bar() to plain bar pattern.
		want = s + 1;
	sprintf(pat, ".*%s", want);
	if (s = strchr(pat, '('))
		*s = 0;
	if (parse_string(1, pat))		// The line number looks right already.
		return -1;
	move_by_lines(-10);				// Search 10 previous lines.
	save_var narrow_start = point;
	point = orig;
	if (!re_search(-1, pat)) {
		point = orig;
		return -1;
	}
	to_begin_line();
	restore_vars();
	return lines_between(0, point) + 1;
}

// Delete text in range, replace with new value.
replace_line_number(int start, int end, int newval)
{
	save_spot point = start;
	delete(start, end);
	bprintf("%d", newval);
}

// Search for a pattern in range, where #1 is a source file name and #2
// is a line number, then get that source file line and append it.  If
// that line doesn't contain the text in want but a slightly earlier
// line does, adjust.
extract_color_file_lines(int from, int to, char *pat, char *want)
{
	char file[FNAMELEN], linestr[FNAMELEN];
	int i, obuf = bufnum, adj_line = -1, linestart, lineend, olen;

	save_spot point = from;
	save_var narrow_end = size() - to, abort_file_io = ABORT_ERROR;
	while (re_search(1, pat)) {		// Get each file name and line number.
		grab(find_group(1, 1), find_group(1, 0), file);
		grab(linestart = find_group(2, 1),
			 lineend = find_group(2, 0), linestr);
		to_end_line();
		olen = point - give_begin_line();
		absolute(file);
		if (!look_file(file)) {
			if (check_file(file) != CHECK_FILE
				  || (i = std_find_it(file)) && i != EREADABORT) {
				bufnum = obuf;
				continue;
			}
			on_exit {		// Delete this buffer when function exits.
				i = bufnum;
				bufnum = obuf;
				if (window_bufnum == i)
					window_bufnum = obuf;
				buf_delete(i);
			}
			if (i == EREADABORT)
				return;
			save_var bufnum;
		}
		if (!size()) {
			bufnum = obuf;
			continue;
		}
		save_var point;
		go_line(strtoi(linestr, 10) + 1);
		if (browser_options & 256)
			adj_line = adjust_found_bsc_line(want);
		re_search(1, "[ \t]*");		// Move past indentation, don't copy it.
		i = point;
		to_end_line();
		recolor_buffer_range(i, point);
		if (olen > window_width / 2)		// Too wide, put it on a second line.
			buf_printf(obuf, "\n--     ");
		buf_xfer_colors(obuf, i, point);
		bufnum = obuf;
		if (adj_line != -1)
			replace_line_number(linestart, lineend, adj_line - 1);
	}
}

// Fill in sym_name and sym_descr with info on instance inst.
void get_inst_data(int *pbsc, int inst, char *sym_name, char *sym_descr)
{
	char sym_type[512], sym_attr[512];
	byte *undecor_name, typ[1];
	short atr[1];

	BSCIinstInfo(pbsc, inst, &undecor_name, typ, atr);
    bytes_to_chars(sym_name, (byte *) BSCFormatDname(pbsc, undecor_name));

    bytes_to_chars(sym_type, (byte *) BSCSzFrTyp(pbsc, *typ));
    bytes_to_chars(sym_attr, (byte *) BSCSzFrAtr(pbsc, *atr));
	sprintf(sym_descr, "%s%s%s", sym_attr, *sym_attr ? " " : "", sym_type);
}

filter_to_text(char *s, int i)
{
	sprintf(s, "%s%s%s%s%s",
			(i & MBFVARS) ? "Var " : "",
			(i & MBFFUNCS) ? "Func " : "",
			(i & MBFMACROS) ? "Macro " : "",
			(i & MBFTYPES) ? "Type " : "",
			(i & MBFCLASS) ? "Class " : "");
}

// Save the most recent buffer position for when the browser buffer
// displayed info on the given symbol, so we can restore it later on.
record_browser_position(char *dest)
{
	save_var bufnum = create(dest);
	int pos = point;
	save_var point = 0;
	if (search(1, "\nSymbol: ")) {
		char sym[FNAMELEN];
		grab(point, give_end_line(), sym);
		buffer_to_end(BROWSER_POS, 100000);
		buffer_printf(BROWSER_POS, "%d %s\n", pos, sym);
	}
}

// Go to remembered location in listing for this symbol, return nonzero.
int restore_browser_position(char *sym)
{
	save_var bufnum = create(BROWSER_POS);
	char pat[FNAMELEN];
	sprintf(pat, "([0-9]+) %q\n", sym);
	point = size();
	if (!re_search(-1, pat))
		return 0;
	grab(find_group(1, 1), find_group(1, 0), pat);
	int pos = numtoi(pat);
	restore_vars();
	point = pos;
	return 1;
}

// Put Symbol info for this symbol into buffer dest.
int dump_defs(int *pbsc, char *bsc_filename, char *symbol,
			  int *insts, int num_insts, int flags, char *dest)
{
	char sym_name[512], sym_descr[512], other_name[512], other_descr[512];
	int r = 0, i, j, *list, rstart;

	record_browser_position(dest);
	bufnum = zap(dest);
	browse_mode();
	bprintf("%pBrowser File: %p%s%p\n",
			-1, color_class c_function, bsc_filename, -1);
	bprintf("Symbol: %p%s%p\n", color_class c_function, symbol, -1);
	filter_to_text(other_name, browser_filter);
	bprintf("Filter all but: %s\n", other_name);
	filter_to_text(other_name, browser_filter_usedby);
	bprintf("Filter all Uses/UsedBy but: %s\n\n", other_name);

	for (j = 0; j < num_insts; j++){
		int num_defs = 0, num = 0, inst = insts[j];

		if (num_insts > 100 && (j % 10) == 0)
			noteput("Retrieving symbols: %d%% complete...",
					muldiv(j, 100, num_insts));
		get_inst_data(pbsc, inst, sym_name, sym_descr);
		bprintf("%p%s%p (%s) is defined at:\n",
				color_class c_function, sym_name, -1, sym_descr);
		rstart = point;

		if (BSCGetDefArray(pbsc, inst, &list, &num_defs) && num_defs) {
			for (i = 0; i < num_defs; i++){
				char modname[1024];
				byte *temp;
				short line_num[1];
				BSCIdefInfo(pbsc, list[i], &temp, line_num);
				r++;
				bytes_to_chars(modname, temp);

				convert_to_8_3_filename(modname, 1);
				bprintf("    %s(%d): \n", modname, *line_num);
			}
			BSCDisposeArray(pbsc, list);
		}

		if ((flags & 2) && BSCGetRefArray(pbsc, inst, &list, &num) && num) {
			bprintf("%p%p%s%p (%s) is used at:\n",
					-1, color_class c_function, sym_name, -1, sym_descr);
			for (i = 0; i < num; i++){
				char modname[1024];
				byte *temp;
				short line_num[1];
				BSCIrefInfo(pbsc, list[i], &temp, line_num);
				bytes_to_chars(modname, temp);

				convert_to_8_3_filename(modname, 1);
				bprintf("-    %s(%d): \n", modname, *line_num);
			}
			BSCDisposeArray(pbsc, list);
		}
		if (flags & 1)
			extract_color_file_lines(rstart, point, "-?" BROWSER_ENTRY, sym_name);
		if ((flags & 4) && BSCGetUsesArray(pbsc, inst, browser_filter_usedby,
										   &list, &num) && num) {
			bprintf("%p%s%p (%s) uses:\n",
					color_class c_function, sym_name, -1, sym_descr);
			for (i = 0; i < num; i++){
				get_inst_data(pbsc, list[i], other_name, other_descr);
				bprintf("  %p%s%p (%s)\n",
						color_class c_function, other_name, -1, other_descr);
			}
			BSCDisposeArray(pbsc, list);
		}
		if ((flags & 8)
			  && BSCGetUsedByArray(pbsc, inst, browser_filter_usedby,
								   &list, &num) && num) {
			bprintf("%p%s%p (%s) is used by:\n",
					color_class c_function, sym_name, -1, sym_descr);
			for (i = 0; i < num; i++){
				get_inst_data(pbsc, list[i], other_name, other_descr);
				bprintf("- %p%s%p (%s)\n",
						color_class c_function, other_name, -1, other_descr);
			}
			BSCDisposeArray(pbsc, list);
		}
		if (user_abort)
			break;
	}
	if (!restore_browser_position(symbol)) {
		point = 0;
		if (search(1, "\n\n"))
			nl_forward();
		else
			point = 0;
	}
	if (num_insts > 100)
		say("");
	return r;
}

// Open browser database, putting handle in *bsc.  Look up symbol using
// filter, and put overload array into inst and cnt.  Return
// nonzero on error.
int browser_get_overload_array(char *bsc_filename, int **bsc, char *symbol,
							   int filter, int **insts, int *cnt)
{
	byte temp[FNAMELEN];

	browser_check_dll();
	chars_to_bytes(temp, bsc_filename);
	if (BSCOpen(temp, bsc) == 0) {
		say("Can't use %s to read %s.", BSC_DLL, bsc_filename);
		return 1;
	}
	chars_to_bytes(temp, symbol);
	BSCGetOverloadArray(*bsc, temp, filter, insts, cnt);
	return 0;
}

browser_check_dll()
{
	char dllname[FNAMELEN];
	get_executable_file(dllname, BSC_DLL, 0);
	if (check_file(dllname) != CHECK_FILE)
		error("First run Alt-x configure-epsilon to set up source browsing.");
}

/*
	This gathers def or ref data for the indicated symbol, using the
	indicated browser database.  All the data goes into a buffer, and
	the browser database is closed, and the dll is unloaded.
	Subsequently, you can use browser_find_match() to retrieve
	information on the matches.  If flags & 1, don't reposition in
	history.  If flags & 2, not invoked via tagging, but directly from
	symbols buffers or browse-symbol, so use different bits in
	browser_options to customize.  If flags & 4, reuse current buffer if
	it's in browser mode, even if it's not named BROWSER_BUF.
*/
int browser_query(char *bsc_filename, char *symbol, int flags)
{
	int *bsc, *inst, num_insts, r = 0;
	char pat[1024], *buf = BROWSER_BUF;

	if ((flags & 4) && mode_keys == browse_tab)
		buf = bufname;
	save_var bufnum, abort_searching = 0;
	if (browser_get_overload_array(bsc_filename, &bsc, symbol,
								   browser_filter, &inst, &num_insts))
		return -1;
	r = dump_defs(bsc, bsc_filename, symbol, inst, num_insts,
				  browser_options >> ((flags & 2) ? 4 : 0), buf);
	if (num_insts)
		BSCDisposeArray(bsc, inst);
	BSCClose(bsc);
		// We unload the DLL to be sure that all the resources are released.
	call_dll(BSC_DLL);
	if (!(flags & 1))
		buffer_to_end(BROWSER_HIST, 100000);
	save_var bufnum = create(BROWSER_HIST);
	save_spot point = 0;
	sprintf(pat, "^%q$", symbol);
	while (*symbol && re_search(1, pat))
		delete(matchstart, point + 1);
	point = size();
	bprintf("%s\n", symbol);
	return r;
}

// Write list of symbols starting with prefix to buffer dest, one per
// line.  Return nonzero on error.

int browse_list_prefix_matches(char *bsc_filename, char *prefix, char *dest)
{
	int *bsc, *inst, num_insts, i;
	byte *undecor_name, typ[1];
	char symbol[FNAMELEN];
	short atr[1];

	sprintf(symbol, "%s*", prefix);
	save_var bufnum, abort_searching = 0;
	if (browser_get_overload_array(bsc_filename, &bsc, symbol,
								   browser_filter, &inst, &num_insts))
		return -1;
	zap(dest);
	for (i = 0; i < num_insts; i++)
		if (BSCIinstInfo(bsc, inst[i], &undecor_name, typ, atr)) {
			bytes_to_chars(symbol, (byte *) BSCFormatDname(bsc, undecor_name));
			buffer_printf(dest, "%s\n", symbol);
		}
	if (num_insts)
		BSCDisposeArray(bsc, inst);
	BSCClose(bsc);
		// We unload the DLL to be sure that all the resources are released.
	call_dll(BSC_DLL);
	save_var bufname = dest, show_status = 0;
	point = 0;					// Now clean up matches.
	while (simple_re_replace(1, "%([^()\n]*%)", "") > 0)
		;						// Discard parenthesized arg lists.
	while (simple_re_replace(1, "%<[^<>\n]*%>", "") > 0)
		;						// Discard template names.
	if (!strchr(prefix, ':'))				   // Separate at ::'s.
		simple_re_replace(1, "::", "\n");
	sprintf(symbol, "^%q", prefix);			   // Drop entries that don't start
	do_drop_matching_lines(REGEX, symbol, 0);  // with prefix.
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	point = 0;
	return 0;
}

char *browse_matcher(char *s, int flags)
{
	if (flags & STARTMATCH)
		browse_list_prefix_matches(bsc_file, s, _MATCH_BUF);
	return general_matcher(s, flags);
}

/*
	This returns nonzero if there is a remaining match that hasn't
	been reported yet.  The filename and line number of the match
	are returned.  The n_fwd parameter says which match to report.
	n_fwd=1 means the next match, n_fwd=0 means the current match,
	n_fwd=-1 means the previous match, n_fwd=2 means the match after
	next, etc.
 */

int browser_find_match(char *file, int *linenum, int n_fwd)
{
	int buf = name_to_bufnum(BROWSER_BUF);
	int dir = RE_FORWARD;
	char linestr[30];

	save_var bufnum;
	if (!buf)
		return 0;
	else
		bufnum = buf;

	if (n_fwd < 0){
		dir = RE_REVERSE;
		n_fwd = -n_fwd;
		to_begin_line();
	} else if (n_fwd > 0) {
		to_end_line();
	} else {			// n_fwd == 0: return current match again
		to_begin_line();
		n_fwd = 1;
	}
		
	while (n_fwd--){
		if (!re_search(dir, "^" BROWSER_ENTRY))
			return 0;
	}
	grab(find_group(1, 1), find_group(1, 0), file);
	grab(find_group(2, 1),  find_group(2, 0), linestr);
	*linenum = strtoi(linestr, 10) + 1;
	return 1;
}

when_loading()	/* disable normal keys */
{
	fix_key_table(reg_tab, (short) normal_character, browse_tab, -1);
	set_list_keys(browse_tab);
}

// Look up an earlier symbol again.  If 0, do current symbol again, -1
// does previous, 1 asks user.
browse_old_symbol(int dir)
{
	char symbol[FNAMELEN];

	if (!exist(BROWSER_HIST))
		error("No previous symbol to look up.");
	save_var point = 0;
	if (re_search(1, "^Symbol: "))
		grab(matchend, give_end_line(), symbol);
	restore_vars();
	save_var bufname = BROWSER_HIST;
	if (point == size())
		point--;
	if (dir == 1) {
		to_begin_line();
		zap(_MATCH_BUF);
		xfer(_MATCH_BUF, 0, size());
		set_buf_point(name_to_bufnum(_MATCH_BUF), size());
		restore_vars();
		if (!get_choice(name_to_bufnum(_MATCH_BUF), symbol,
						"Previous symbols", "Look up a symbol again",
						"OK", "Cancel", ""))
			return;
	} else {
		if (dir == -1)
			nl_reverse();
		else
			to_end_line();
		if (!parse_string(-1, ".*", symbol))
			error("No previous symbol.");
		delete(point, give_end_line() + 1);
	}
	restore_vars();
	browser_query(bsc_file, symbol, (dir == -1) ? 7 : 6);
}

browse_previous() on browse_tab['l']
{
	iter = 0;
	browse_old_symbol(has_arg ? 1 : -1);
}

browse_pick() on browse_tab[' '], browse_tab['\n'],
				browse_tab['e'], browse_tab['\r'], browse_tab[GREYENTER]
{
	char fname[FNAMELEN], linestr[FNAMELEN], symbol[FNAMELEN];

	tiled_only();
	to_begin_line();
	if (parse_string(1, "--   ")) {	// Continuation line, move back.
		nl_reverse();
		to_begin_line();
	}
	if (parse_string(1, "Browser File: ")) {
		if (tags_from_bsc)
			select_tag_file();
		else
			select_browse_file();
	} else if (parse_string(1, "Symbol: ")) {
		browse_symbol();
	} else if (parse_string(1, "Filter all but: ")) {
		browse_set_filter();
	} else if (parse_string(1, "Filter all Uses/UsedBy but: ")) {
		browse_set_usedby_filter();
	} else if (parse_string(1, "-?" BROWSER_ENTRY)) {	// File/line reference.
		grab(find_group(1, 1), find_group(1, 0), fname);
		grab(find_group(2, 1),  find_group(2, 0), linestr);
		if (!window_number)		// Any window but this one.
			window_number = 1;
		else
			window_number--;
		get_to_tag_match_line(fname, strtoi(linestr, 10) + 1);
	} else if (parse_string(1, "[- ] ([^ (<\n]+)")) {	// Another symbol.
		grab(find_group(1, 1), find_group(1, 0), symbol);
		browser_query(bsc_file, symbol, 6);
	} else
		error("No reference on this line.");
}


// Set display_func_name to what we want displayed as the "function
// name" we're editing, and return 1.

int browse_func_name_finder()
{
	save_var point;
	*display_func_name = 0;
	if (re_search(-1, ".* (is defined at|is used at|uses|is used by):"))
		grab(point, matchstart, display_func_name);
	else {
		point = 0;
		if (re_search(1, "^Symbol: "))
			grab(matchstart, give_end_line(), display_func_name);
	}
	return 1;
}

browse_mode()
{
	mode_keys = browse_tab;		/* Use these keys. */
	major_mode = _browse_mode_name;
	discardable_buffer = 1;
	mouse_dbl_selects = 1;
	try_calling("browse-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

command browse_set_filter() on browse_tab['f']
{
	do_browse_set_filter(&browser_filter, "", "browse-set-filter");
}

command browse_set_usedby_filter() on browse_tab['b']
{
	do_browse_set_filter(&browser_filter_usedby, " uses/usedby",
						 "browse-set-usedby-filter");
}

do_browse_set_filter(char *var, char *type, char *cmd)
{
	char buf[FNAMELEN];

	for (;;) {
		filter_to_text(buf, *var);
		noteput("Only show%s: %s[VFMTC toggle, or Enter] ", type, buf);
		getkey();
		check_abort();
		if (binding_of(key, root_keys) == (short) help
			  || key == 'h' || key == '?')
			help_on_command(cmd);
		else if (IS_ENTER_KEY(key))
			break;
		switch (toupper(key)) {
			case 'V': *var ^= MBFVARS; break;
			case 'F': *var ^= MBFFUNCS; break;
			case 'M': *var ^= MBFMACROS; break;
			case 'T': *var ^= MBFTYPES; break;
			case 'C': *var ^= MBFCLASS; break;
		}
	}
	say("");
	save_var point;
	browse_old_symbol(0);
	to_begin_line();
}

// Set browser-file variable to alternative .bsc file.
command select_browse_file()
{
	char file[FNAMELEN];

	save_var only_file_extensions = "|.bsc|";
	iter = 0;
	get_file(file, "Source browser (.bsc) file", browser_file);
	if (*file)
		strcpy(browser_file, file);
}

browse_get_default_tag(char *def)
{
	save_var point, mark;
	if (is_highlight_on() && point != mark && ABS(point - mark) < 250) {
		fix_region();
		grab(point, mark, def);
	} else
		grab_tag_name(def);
}

do_browse_symbol(int ask)
{
	browser_check_dll();
	tiled_only();
	iter = 0;
	if (!tags_from_bsc) {
		if (!*browser_file || check_file(browser_file) != CHECK_FILE) {
			select_browse_file();
			if (check_file(browser_file) != CHECK_FILE)
				error("To view symbols, you must select a .bsc file.");
		}
		save_var tags_from_bsc = 1;
		strcpy(bsc_file, browser_file);
		absolute(bsc_file);
	}
	char def[FNAMELEN], tag[FNAMELEN];
	browse_get_default_tag(def);
	if (ask || !*def) {
		save_var menu_width = screen_cols - _view_left - _view_right;
		prompt_comp_read(tag, "Find symbol", browse_matcher, CAUTIOUS, def);
	} else
		strcpy(tag, def);
	quiet_set_bookmark();
	if (browser_query(bsc_file, tag, 6) == -1)
		return;
	if (mode_keys != browse_tab)
		to_buffer(BROWSER_BUF);
}

// Prompt for symbol, then look it up and show result in symbol browser.
command browse_symbol() on reg_tab[NUMCTRL(GREYSLASH)]
{
	do_browse_symbol(1);
}

command browse_current_symbol()
{
	do_browse_symbol(0);
}
