/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B15 to 13.00.

#include "eel.h"
#include "dired.h"
#include "diff.h"
#include "proc.h"
#include "vspell.h"

int do_compare_to_prior_version(int edits)
{
	int prev = zap(edits > 0 ? "-prior-version" : "-disk-version"), i, j;
	if (edits > 0) {
		save_var narrow_start = 0, narrow_end = 0;
		int ocnt = undo_count(0);
		for (i = 0; i < edits; i++)
			if ((j = undo_it(1, 1, 0)) == -1)
				break;
			else if (!(j & UNDO_CHANGES))
				i--;
		buf_xfer(prev, 0, size());
		while (undo_count(0) > ocnt)
			if (undo_it(0, 1, 0) == -1)
				break;
	} else {
		int orig = bufnum, trans = translation_type;
		char *fname = filename;
		bufnum = prev;
		if (err_file_read(fname, trans))
			return 2;
		bufnum = orig;
	}
	if (buffers_identical(bufnum, prev))
		return 1;
	safe_copy_buffer_variables(prev, bufnum);
	struct screen_info winsetup;
	winsetup.wins = 0;
	save_screen(&winsetup);
	one_window();
	window_split(HORIZONTAL);
	window_bufnum = prev;
	on_exit restore_screen(&winsetup);
	has_arg = 0;
	switch (compare_to_prior_version_style) {
		case 0:		diff(); break;
		case 2:		merge_diff(); break;
		default:	visual_diff(); break;
	}
	restore_vars();
	locate_window(_diff_buffer, "");
	prev_forget_buf(prev);
	return 0;
}

dired_unix_fix_lines()
{
	save_var modified, indent_with_tabs = 0, case_fold = 1;
	right_align_columns("[0-9]+ *" UNIX_DATE);
	right_align_columns(UNIX_DATE);
	save_var point = 0;
	if (re_search(1, ".*" UNIX_DATE))
		goal_column = current_column();
}

char *man_use_program(int type);
int literal_if_args();
char *man_match(char *s, int flags);

show_man_page(int type)
{
	char topic[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN];
	char *s, *prog = man_use_program(type);

	tiled_only();
	if (is_unix_flavor == IS_UNIX_MACOS && (s = getenv("MANPATH"))
		  && !strcmp(s, "/usr/man:/usr/X11R6/man") && !check_file("/usr/man"))
		putenv("MANPATH=");		// Avoid Mac's bogus MANPATH bug.

	save_var show_status = 0, abort_searching = 0, case_fold = 1;
	grab_man_word_from_buffer(bufnum, point, buf,
							  "[-+_0-9a-z.:]+(%(..?%))?");
	if (type == 1 && !strcmp(buf, "use")) {
		save_var point;
		point += parse_string(1, "<word>*<space>*");
		grab_man_word_from_buffer(bufnum, point, buf,
								  "[-+_0-9a-z.:]+(%(..?%))?");
	}
	man_rewrite(buf);
	save_var completion_literal_char = literal_if_args;
	sprintf(pr, "Show %s entry for", type ? prog : "manual");
	prompt_comp_read(topic, pr, prog ? man_match : 0,
					 COMP_FOLD | SPACE_VALID | CAUTIOUS, buf);
	if (!*topic)
		return;
	make_man_page(topic, type);
}

compile_handler(int activity, int buf, int from, int to)
{
	UNREFERENCED_PARAMETER(to);
	if (activity == NET_DONE) {
		save_var bufnum = buf;
		point = size();
		if (size() && character(point - 1) != '\n')
			bprintf("\n");
		bprintf("Process exited with status %d\n", from);
		update_mode_more_extra(-1);
		if (post_compile_hook_buffer && buf_exist(post_compile_hook_buffer)) {
			save_var bufnum = post_compile_hook_buffer;
			if (post_compile_hook)		// Track this compilation back to orig
				(*post_compile_hook)(6, bufnum); // buffer, then run its hook.
		}
	}
}

do_spell_correct(int mult)
{
	char word[FNAMELEN];
	save_var point--;
	re_search(RE_FORWARD | RE_IGNORE_COLOR, get_spell_word_pattern());
	re_search(RE_REVERSE | RE_IGNORE_COLOR, get_spell_word_pattern());
	int wstart, wend, i = 1, hline = -1, have_file = *filename;
	grab(wstart = matchend, wend = matchstart, word);
	if (!*word)
		return 2;
	int case_repl = (get_num_var(vspell_get_mode_variable()) & 0x2000)
					? 0 : check_for_case_replace(wstart, wend);
	build_window();
	int wcol = window_left + column_in_window;
	int wline = window_top + line_in_window, obuf = bufnum;
	int sugg = spell_find_suggestions(word, 9, case_repl);
	save_var bufnum = sugg;
	remove_line_highlight();
	go_line(10);
	delete(point, size());
	int wcnt = lines_between(0, size());
	point = 0;
	if (point < size()) {
		do {
			bprintf("%d ", i++);
		} while (nl_forward() && point < size());
		stuff("\n");
	} else
		stuff("No suggestions\n\n");
	stuff("O Ignore word once\n");
	if (have_file) {
		stuff("F Ignore in this file\n");
		stuff("D Ignore in this directory\n");
	}
	stuff("M Ignore for this mode only\n");
	stuff("G Ignore globally\n");
	if (mult)
		stuff("Ctrl-R for recursive edit\n");
	point = 0;
	int cnt = lines_between(0, size());
	restore_vars();
	highlight_on();
	save_spot mark = wstart;
	save_spot point = wend;
	on_exit highlight_off();
	save_var want_gui_prompts = 0;
	save_var window_handle;
	int row = 5;
	if (wline + cnt + 4 < screen_lines)
		row = wline + 2;			// Position below text.
	else if (wline > cnt + 2)
		row = wline - cnt - 3;		// Or above.
	else if (row + cnt >= screen_lines - 1)
		if ((row = screen_lines - cnt - 3) + cnt <= wline)
			row = 0;
	int col = wcol - 3;				// Align with text.
	if (col + 35 > screen_cols)		// Too far to right, so move back.
		col = screen_cols - 35;
	if (col < 0)					// Too far to left, move right.
		col = 0;							// If not visible above or below, 
	if (row <= wline && wline < row + cnt)	// move left or right
		if (wcol + 35 > screen_cols)
			col = wcol - 35;
		else
			col = wcol + strlen(word) + 3;
	int win = add_popup(col, row, 30, cnt + 2, _view_border, sugg);
	on_exit remove_window(win);
	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	window_title(win, BTOP, TITLECENTER,  " Suggestions ");

	display_scroll_bar = 0;
	for (;;) {
		if (wcnt > 1)
			noteput("1-%d to replace, ignore \"%s\" Once, or add "
					"to an ignore list [1-%dO%sMG] ", wcnt, word, wcnt,
					have_file ? "FD" : "");
		else if (wcnt == 1)
			noteput("1 to replace, ignore \"%s\" Once, or add "
					"to an ignore list [1O%sMG] ", word, have_file ? "FD" : "");
		else
			noteput("No suggestions.  Ignore \"%s\" Once, or add "
					"to an ignore list [O%sMG] ", word, have_file ? "FD" : "");
		refresh();
		getkey();
		say("");
		check_abort();
		if (binding_of(key, reg_tab) == down_line)
			hline = spell_correct_move_highlight(hline, win, 1, 0);
		else if (binding_of(key, reg_tab) == up_line)
			hline = spell_correct_move_highlight(hline, win, -1, 0);
		else if (IS_ENTER_KEY(key) && hline != -1)
			hline = spell_correct_move_highlight(hline, win, 0, 1);
		if (IS_MOUSE_LEFT(key)
			  && window_at_coords(mouse_y, mouse_x, mouse_screen) == win
			  && (i = mouse_y - window_text_top(win)) >= 0 && i < cnt)
			hline = spell_correct_move_highlight(-1, win, i + 1,
				IS_MOUSE_DOUBLE(key));
		if (tolower(key) == 'g')
			spell_ignore_this_word(SPELLB_GLOBAL, word, obuf);
		else if (tolower(key) == 'd' && have_file) {
			spell_ignore_this_word(SPELLB_PERDIR, word, obuf);
			break;
		} else if (tolower(key) == 'f' && have_file) {
			spell_ignore_this_word(SPELLB_PERFILE, word, obuf);
			break;
		} else if (tolower(key) == 'm')
			spell_ignore_this_word(SPELLB_PERMODE, word, obuf);
		if (key == CTRL('R') && mult)
			return 1;
		if (IS_ASCII_KEY(key) && strchr("sogm", tolower(key))
			  || key == abort_key || key == ESC)
			break;
		if (isdigit(key) && key != '0' && key - '0' <= wcnt) {
			restore_vars();
			spell_get_suggestion_line(sugg, key, word);
			point = wstart;
			delete(wstart, wend);
			stuff(word);
			return 0;
		} else if (IS_ALT_KEY(key) && isdigit(i = KEY_PLAIN(key))
				   && i != '0' && i - '0' <= wcnt) {
			spell_get_suggestion_line(sugg, i, word);
			copy_line_to_clipboard(word, 0);
		}
	}
	return 0;
}

// Called during buffer display.  Mark up misspelled words in the
// current window using a region.
vspell_buffer_when_displaying()
{
	int from, to, cnt = 0;
	if (!modified_buffer_region(&from, &to, "vspell")
		  && vspell_data.window_start == window_start
		  && vspell_data.window_end == window_end
		  && vspell_data.display_column == display_column
		  && !vspell_data.skipped_word)
		return;
	vspell_data.window_start = window_start;
	vspell_data.window_end = window_end;
	reset_modified_buffer_region("vspell");
	if (modify_region(SPELL_REGION_HANDLE, MRTYPE, -1) == REGNORM)
		remove_region(SPELL_REGION_HANDLE);
	if (!vspell_data.want)
		return;
	int opos = point;		// Ignore the word just before point when typing.
	char *pat = get_spell_word_pattern();
	save_var narrow_start = window_start
							- matches_at_length(window_start, -1, pat);
	save_var narrow_end = size() - window_end
						  - matches_at_length(window_end, 1, pat);
	save_var point = 0, abort_searching = ABORT_ERROR;
	int opt = get_num_var(vspell_get_mode_variable());
	vspell_data.skipped_word = 0;
	while (re_search(1, pat) == 1) {
		char word[FNAMELEN];
		if (spell_get_and_check_word(opt, word))
			continue;
		int skey = key;
		skey = (skey == GREYBACK) ? '\b' : IS_ASCII_KEY(skey) ? skey : '/';
		if (matchstart <= opos && opos <= matchend + (strchr("'", skey) != 0)
			  && (isalnum(skey) || (strchr("_'", skey)) || skey == '\b')) {
			vspell_data.skipped_word = 1;	// Catch it next time.
			continue;
		}
		if (cnt++ > 2000)	// More than this many misspellings in the window?
			break;			// Must be something wrong; bail out.
		spot start = alloc_spot(0), end = alloc_spot(1);
		*start = point - strlen(word);
		*end = point;
		add_region(start, end, color_class spell_misspelled, REGNORM,
				   SPELL_REGION_HANDLE);
		if (display_column >= 0)	// Don't check words we can't see.
			if (current_column() - strlen(word) > display_column + window_width)
				nl_forward();
			else if (current_column() < display_column - 50)
				move_to_column(display_column - 50);
	}
	modify_region(SPELL_REGION_HANDLE, MRAUTODEL, 1);
}

set_spell_mode(int per_buf)
{
	int old = get_num_var(vspell_get_mode_variable()), want;

	if (vspell_data.buf_want)
		old = (vspell_data.buf_want == 1);
	want = (old & ~1) | (has_arg ? (iter != 0) : (~old & 1));
	if ((want & 1) && !spell_load_file(SPELLB_ORDERED, 0)) {
		say("No dictionary found, running spell-configure.");
		spell_configure();
		return;
	}
	if ((want & 1) && !spell_mode_ok(bufnum))
		error("Spell mode not available for this system buffer.");
	if (per_buf)
		vspell_data.buf_want = want ? 1 : 2;
	else {
		set_num_var(vspell_get_mode_variable(), want);
		vspell_data.buf_want = 0;
	}
	do_when_make_mode_vspell();
	save_customization(vspell_get_mode_variable());
	if ((want & 1) != vspell_data.want)
		say("Spell mode not available for this type of buffer.");
	else if (per_buf)
		say("Turning Spell minor mode %s for this buffer.",
			  vspell_data.want ? "on" : "off");
	else
		say("Setting %s to turn Spell minor mode %s.",
			  name_name(vspell_get_mode_variable()),
			  vspell_data.want ? "on" : "off");
	make_mode();
}
