/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.00 to 13.01.

#include "eel.h"
#include "dired.h"
#include "diff.h"
#include "proc.h"
#include "kill.h"
#include "menu.h"
#include "bufed.h"
#include "c.h"
#include "dired.h"
#include "colcode.h"
#include "tags.h"
#include "epsnet.h"
#include "lowlevel.h"
#include "html.h"
#include "info.h"
#include "perl.h"
#include "session.h"
#include "tex.h"
#include "vspell.h"

command brief_keyboard()
{
	load_keyboard_file("brief.kbd", "Brief");
	alt_invokes_menu = 0;
	fwd_search_key = -1;
	virtual_space.default = 2;	// Brief always uses virtual space
	display_column.default = 0;	// and horiz scrolling
	virtual_space = 2;		// Also set for current buffer & window
	display_column = 0;
	mark_rectangle_expands = 1;
	resize_rectangle_on_tab = 0;
	mark_unhighlights = 1;
	yank_rectangle_to_corner = 3;
	yank_line_retains_position = 1;
	typing_deletes_highlight = 2;	// Only at prompts.
	typing_hides_highlight = 0;
	search_positions_at_start = 1;
	search_in_region = 1;
	replace_in_region = 2;
	recall_prior_response_options = 3;	// At prompts, display old responses in-place.
	recall_response_selected = 1;	// Highlight such responses.
	search_defaults_from = 2;
	abort_key = ESC;
	strcpy(menu_file, "brief.mnu");
	rebuild_menu();
	say("Brief keyboard loaded.");
}

command epsilon_keyboard()
{
	load_keyboard_file("epsilon.kbd", "Epsilon");
	alt_invokes_menu = 0;
	fwd_search_key = -1;
	virtual_space.default = 0;	// set back to factory defaults
	virtual_space = 0;		// Also set for current buffer & window
	mark_unhighlights = 0;
	mark_rectangle_expands = 0;
	resize_rectangle_on_tab = 0;
	typing_deletes_highlight = 1;
	abort_key = CTRL('G');
	yank_rectangle_to_corner = 1;
	yank_line_retains_position = 0;
	typing_deletes_highlight = 1;
	typing_hides_highlight = 1;
	search_positions_at_start = 0;
	search_in_region = 0;
	replace_in_region = 0;
	recall_prior_response_options = 0;
	recall_response_selected = 0;
	search_defaults_from = 1;
	strcpy(menu_file, "epsilon.mnu");
	strcpy(gui_menu_file, "gui.mnu");
	rebuild_menu();
	say("Epsilon keyboard loaded.");
}

make_buffer_list(int flags, char *inbuffer)
{
	int old = bufnum, opos = 0, start = 0;
	int sort = (bufed_grouping & 1) != 0;	// Sort each group separately.
	int sep_unsaved = (bufed_grouping & 2) != 0; // Put unsaved first.

	if (exist(inbuffer)) {
		bufname = inbuffer;
		set_read_only(0);
		bufnum = old;
	}
	zap(inbuffer);
	if (sep_unsaved || (flags & BUFED_ONLY_UNSAVED))
		start = opos = add_to_buffer_list(inbuffer, opos, old, 0, 0, 1);
	if (!(flags & BUFED_ONLY_UNSAVED)) {
		opos = add_to_buffer_list(inbuffer, opos, old, 1, sep_unsaved, sort);
		opos = add_to_buffer_list(inbuffer, opos, old, 2, sep_unsaved, sort);
		if (flags & BUFED_ALL)
			opos = add_to_buffer_list(inbuffer, opos, old, 3, sep_unsaved, sort);
	}
	if (!(bufed_grouping & 1))
		sort_buffer_list(start, inbuffer);
	bufname = inbuffer;
	point = 0;
	delete_final_newlines();
	set_read_only(1);
	bufnum = old;
}

sort_buffer_list(opos, buf)
char *buf;		/* sort from opos to the end, & return new end */
{
	int col, want_sort = 1, rev = 0;

	save_var case_fold = (preserve_filename_case & 8) != 0;
	switch (tolower(_bufed_sort)) {
		case 'f':	col = BNAME_COL + ACCESS_WIDTH + bufed_column_width; break;
		case 'b':	col = BNAME_COL + ACCESS_WIDTH; break;
		case 'i':	col = ACCESS_WIDTH; break;
		case 'a':	col = 0; rev = 1; break;
		default:	want_sort = 0; break;
	}
	if (isupper(_bufed_sort))
		rev = !rev;
	save_var bufnum;
	bufname = buf;
	if (want_sort)
		do_sort_region(opos, size(), col, rev);
	save_var narrow_start = opos;
	strip_columns(0, ACCESS_WIDTH);		// remove access timestamps
	return size();
}

look_file(fname)	/* is there a buffer with this file? */
char *fname;		/* if so, switch to it */
{
	int i, orig = bufnum;
	int new_file_id[3];

	unique_filename_identifier(fname, new_file_id);
	if (!fnamecmp(filename, fname) && force_save_as != FSA_NEWFILE)
		return 1;	/* already in it */
	if (check_unique_file_id(new_file_id) && force_save_as != FSA_NEWFILE)
		return 1;
	i = buf_list(0, 2);	/* start with most recently created buf */
	do {
		bufnum = i;
		if (!fnamecmp(filename, fname) && force_save_as != FSA_NEWFILE)
			return 1;
		if (check_unique_file_id(new_file_id) && force_save_as != FSA_NEWFILE)
			return 1;
	} while (i = buf_list(-1, 1));
	bufnum = orig;
	return 0;
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN], classname[FNAMELEN], pat[FNAMELEN];
	int start, funcstart, j, got_class = 0;

	start_of_function = -1;
	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	if (c_extra_keywords & OBJC_KEYWORDS)
		got_class |= find_objc_class(type, class);
	to_begin_line();
	start = point;
	point--;		// Begin searching at the last } char on the left margin...
	while (search(-1, "\n}")) {
		point++;		// ...that's not in a comment.
		if (!in_c_comment(c_look_back))
			break;
		point--;
	}

	strcpy(pat, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(");
	if (c_extra_keywords & OBJC_KEYWORDS)
		strcat(pat, "|\n[-+]");
	while (re_search(1, pat)) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*(define|if|pragma).*(\\\n.*)*")) {
			nl_forward();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				j = point;
				if (c_inside_block(point, start)) {
					got_class |= find_c_class_info(start, type, class, func);
					point = j + 1;
					if (got_class)
						continue;
				}
				point = j;
				c_skip_level('{', '}');
				break;
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					nl_forward();	/* c++ comment */
				break;
			case '\n':		// Objective-C method
				point++;
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				if (find_objc_method(func) == 1)
					return CF_INFO_FUNC | got_class;
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return got_class;
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + 1000);
				if (tag_c_func(0) == 2 && !find_c_func_oneliner(start)) {
					point = funcstart;	// Checks out as a function def.
					start_of_function = give_begin_line();
					*classname = 0;
					if (parse_string(-1, (c_extra_keywords == CSHARP_KEYWORDS)
									 ? "[ \t\n]*%.[ \t\n]*"
									 : "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						parse_string(-1, "[a-zA-Z0-9_]+", classname);
					}
					if (func)
						strcpy(func, funcname);
					if (class && *classname)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0)
							| got_class;
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return got_class;
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */
	int in_initializer;

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!(reindent_c_comments & 1) && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if (!reindent_one_line_c_comments && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' && reindent_c_preprocessor_lines == 1
		  || first_char == '\f'
		  || first_char == '@' && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(1, "@(" OBJC_TOPBLOCKS ")[ \t]+<word>|@end"))
		return 0;
	else if (first_char == '}')
		return cind_close();
	else if (have_c_close_equivalent(orig))
		return indent_at_c_close_equivalent(orig);

	to_begin_line();
	if (parse_string(-1, PREPROC_ELSE ".*\n")) {
		point = matchend;
		locate_by_ifdef(-1);
	}
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig) || c_align_break_with_case
		  && matches_at(orig, 1, "[ \t]*break;")) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig) && !((c_extra_keywords & OBJC_KEYWORDS)
									 && matches_at(pstart, 1, "^[-+@]"))) {
		if (c_in_struct())		// Could it be a bitfield, not a label?
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';' || c_semi_equivalent(pend, orig)) {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (c_close_equivalent(pend)) {
		point = pend;
		ind = indent_after_c_close_equivalent(orig);
	} else if (prev_end == '{' || c_open_equivalent(pend, orig)) {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
			  || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_initializer = in_c_initializer(orig)) {
		return ind;
	} else if ((c_extra_keywords & OBJC_KEYWORDS)
			   && objc_next_no_indent(orig, pstart)) {
		return 0;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, in_initializer);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
			   && !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

is_dispatch_table(int p)
{
	save_var point = p, case_fold = 0;
	if (!want_c_macro_rules() || !parse_string(1, "ON_[A-Z0-9_]+%("))
		return 0;
	return re_search(-1, "(BEGIN_MESSAGE_MAP|AFX_MSG_MAP)%(") != 0;
}

int want_c_macro_rules()
{
	if (in_perl_buffer || in_shell_buffer)
		return 0;
	switch (use_c_macro_rules) {
		case 0:
		case 1:		return use_c_macro_rules;
	}
	if (c_extra_keywords & (OBJC_KEYWORDS | JAVA_KEYWORDS | JS_KEYWORDS))
		return 0;
	return is_win32;
}

// Is point after a macro that should act like an opening curly brace?
int c_open_equivalent(int pend, int orig)
{
	UNREFERENCED_PARAMETER(orig);
	if (!want_c_macro_rules() || character(pend - 1) != ')')
		return 0;
	save_var case_fold = 0, point = pend;
	if (c_rev_skip_level('(', ')') && parse_string(-1, c_block_macro_open_pat))
		return 1;
	return 0;
}

// Is point after a macro that should act like a closing curly brace?
int c_close_equivalent(int p)
{
	if (!want_c_macro_rules() || character(p - 1) != ')')
		return 0;
	save_var case_fold = 0, point = p;
	if (c_rev_skip_level('(', ')') && parse_string(-1, c_block_macro_close_pat))
		return 1;
	return 0;
}

// Is there an END_-style macro at p, one that should act like a }
// character?
int have_c_close_equivalent(int p)
{
	save_var case_fold = 0;
	if (want_c_macro_rules() && matches_at(p, 1, c_block_macro_close_pat)
		  && character(matchend) == '(')
		return 1;
	return 0;
}

// Return the indent for a line with an END_-style macro, by finding the
// matching BEGIN_ line and copying its indent.  If we can't find the
// current line's macro or it doesn't start with END_, find any
// BEGIN_-style line and copy that.
int indent_at_c_close_equivalent(int p)
{
	save_var case_fold = 0, point = p;
	char endname[FNAMELEN], beginname[FNAMELEN], *s = c_block_macro_open_pat;
	if (parse_string(1, c_block_macro_close_pat, endname)
		  && !strncmp(endname, "END_", 4))
		sprintf(s = beginname, "BEGIN_%s", endname + 4);
	while (re_search(-1, s) && character(matchstart) != '(')
		;
	return get_indentation(point);
}

// Return the indent for a line following one that acts like a closing
// curly brace, but uses a macro.  Point is at the line with the macro.
int indent_after_c_close_equivalent(int orig)
{
	UNREFERENCED_PARAMETER(orig);
	return get_indentation(point);
}

// Return nonzero if this line has a macro that should act like it ends
// with a semicolon.  We start from just after the last character on
// the line (ignoring comments and spacing).
int c_semi_equivalent(int pend, int orig)
{
	UNREFERENCED_PARAMETER(orig);
	if (!want_c_macro_rules() || character(pend - 1) != ')')
		return 0;
	save_var case_fold = 0, point = pend;
	if (c_rev_skip_level('(', ')')
		  && parse_string(-1, c_block_macro_inner_pat))
		return 1;
	return 0;
}

int after_c_macro_line()
{
	if (!want_c_macro_rules())
		return 0;
	if (parse_string(-1, c_block_macro_open_pat))
		return 1;
	if (parse_string(-1, c_block_macro_close_pat))
		return 1;
	if (parse_string(-1, c_block_macro_inner_pat))
		return 1;
	return 0;
}

indent_after_statement(orig) // orig is place needing indentation, for toplevel
{
	int lastcont, i, c, atparen = -1;

	if ((i = parse_string(1, "[ \t\f]*for[ \t\f]*%([ \t\f]*"))
		  && for_unclosed(point, orig))
		return get_column(point + i); // part of "for" statement
	for (;;) { /* was only the end of a larger statement, find start */
		do {	/* look back for line that ends with ;{}: */
			lastcont = point;
			to_begin_line();
			if (cind_try_space())
				break;
			if ((c = character(point - 1)) == ')') {
				atparen = lastcont;
				if (!c_rev_skip_level('(', ')'))
					break;
				if (after_c_macro_line())
					break;
				to_begin_line();
			} else if (c == ';')
				if (move_before_for_statement(orig))
					c = 'x';	// Keep looking.
		} while (!index(";{}:", c));
		if (c == ':' && parse_string(-1, "%)[ \t]*:")
			  && (c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS | JS_KEYWORDS))
			  && !matches_at(give_begin_line(), 1, "[ \t]*case</word>"))
			continue;	// Not a label but C++ inheritance or similar.
		point = lastcont;	/* use indentation of line */
		to_c_indentation();	/* just after it */
		if (parse_string(RE_FORWARD, "(}[ \t\f]*)?e!lse[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_if())
				break;
		} else if (parse_string(RE_FORWARD,
								"(}[ \t\f]*)?!(catch|finally)[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_try())
				break;
		} else
			break;
	}
	i = current_column();
	if (!i && atparen != -1 && (Topindent || !c_statement_start()))
		i = get_indentation(atparen);	// 2nd func param decl
	if ((atparen = parse_string(1, "for[ \t\f]*%([ \t\f]*"))
		  && for_unclosed(point, orig))
		i = get_column(point + atparen); // part of "for" statement
	if (i < c_top_braces)
		return 0;	// 1st line of new func
	return i;
}

c_in_struct()
{
	char type[FNAMELEN];
	find_c_func_info(type, NULL, NULL, 0);
	return (!strcmp(type, "struct") || !strcmp(type, "union"));
}

parse_label()	/* parse for a case:, default: or label construct at point */
{		/* return nonzero if found, setting matchend to end */
	if (!have_colon())
		return 0;
	if (parse_string(1, "case[^a-zA-Z0-9_\n]([^:\n]*|.*(':'|::).*)"
					 ":[ \t\f]*![^:]"))
		return 1;
	if (parse_string(1, "[a-zA-Z0-9_]+[ \t\f]*:[ \t\f]*![^:]")) {
		save_var matchend;		// A simple "word:" could be a bitfield.
		if (!is_case(point) && c_in_struct())
			return 0;	// A bitfield if inside a struct, not a known keyword.
		return 1;
	}		// Otherwise it's a label.
	return 0;
}

have_label(pt)	/* is there a label here? */
{
	save_var point = pt;
	if (have_colon() && parse_string(1,"[a-zA-Z0-9_]+[ \t\f]*:[^[:]")) {
		if (c_extra_keywords & OBJC_KEYWORDS) {
			save_var narrow_end = size() - point;
			move_by_lines(-10);			// Not a label if in unclosed [] block.
			save_var narrow_start = point;
			point = size();
			return !c_rev_skip_level('[', ']');
		}
		// Not a label but a bitfield if it's not a known keyword and
		// it's in a struct.
		if (!is_case(point) && c_in_struct())
			return 0;
		return 1;
	}
	return 0;
}

is_contin_statement(p, fd, orig) // would line after p be a continuation line?
{			// assumes p is just after last interesting char
	// if fd, pretend func param decls are a "continuation" of func decl
	// orig is later position in buffer, for unclosed "for" scan
	switch (character(p - 1)) {
		case ';':		// check for multi-line "for" statemnt
			if (unmatched_rparen(p, orig))
				return 1;
		case '{':
		case '}':
			return 0;
		case ')':
			save_var point = p;
			save_var narrow_start = p - 500;
			if (!c_rev_skip_level('(', ')'))
				return 1;
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && parse_string(-1, "@(synchronized|catch)"))
				return 0;
			if (sublanguage_code == SUBLANG_PHP)
				return !parse_string(-1, "[^a-z0-9_](if|while|for"
									 "|switch|foreach)[ \t\f]*")
						&& (fd || !Topindent || get_indentation(point));
			if (after_c_macro_line())
				return 0;
			return !parse_string(-1, "[^a-z0-9_](if|while|for"
								 "|switch)[ \t\f]*")
					&& (fd || !Topindent || get_indentation(point))
					|| in_perl_buffer;
		case 'e':
			return !is_else(p - 4);
		case 'o':
			save_var point = p;
			return !parse_string(-1, "[^a-z0-9_]do");
		case 'y':
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && matches_at(p, -1, "@(try|finally)"))
				return 0;
			return 1;
		case ':':
			save_var point = p;
			to_indentation();
			return !is_label(point);
		default:
			return 1;
	}
}

get_clipboard_fix_lines(int buf)
{
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_BINARY)
		return;
	switch (clipboard_convert_mac_lines) {
		case 0:
			return;
		case 1:
			break;
		default:
			if (is_unix_flavor != IS_UNIX_MACOS)
				return;
	}
	save_var bufnum = buf, point = 0;	// If \r's but no \n's, convert.
	if (!nl_forward() && re_search(-1, "\r"))
		simple_re_replace(0, "\r", "\n");
}

char *clipboard_buf_name();

command insert_clipboard()
{
	int buf = zap(clipboard_buf_name());
	if (clipboard_to_buffer(buf,
							clipboard_convert_newlines(), clipboard_format)) {
		get_clipboard_fix_lines(buf);
		mark = point;
		set_region_type(REGNORM);
		grab_buffer(buf);
		build_first = 1;		/* redisplay hint */
	} else
		file_error(errno, "", "Couldn't retrieve clipboard.");
}

get_clipboard()		// push clipboard text on kill ring
{
	int new, old;

	if (!should_use_clipboard() || !clipboard_available())
		return;
	new = zap("-new-clipboard-");
	if (!clipboard_to_buffer(new, clipboard_convert_newlines(),
							 clipboard_format))
		return;
	get_clipboard_fix_lines(new);
	old = create(clipboard_buf_name());
	if (!buffers_identical(old, new)) {
		save_var bufnum = new;
		check_num_kill_buffers();	// copy clipboard to kill ring
		push_kill();
		_kill_width[_num_kill] = REGION;
		xfer(_cur_kill_buf, 0, size());
		buf_zap(old);		// copy new text to old clipboard buf
		buf_xfer(old, 0, size());
	}
	buf_zap(new);
}

set_color_handler(int activity, int buf, int from, int to)
{
	char msg[FNAMELEN], scheme[FNAMELEN];
	int i, nums[4];

	save_var bufnum = buf;
	if (activity != NET_RECV)
		return;
	save_var narrow_start, narrow_end, point = to, abort_searching = 0;
	to_begin_line();
	narrow_end = size() - point;
	point = from;
	to_begin_line();
	do {
		if (parse_string(1, "=")) {
			point++;
			parse_string(1, "[^\n ]+", scheme);
			i = find_index(scheme);
			point++;
			parse_string(1, ".*", msg);
			if (break_into_numbers(msg, nums) == 3 && is_color_scheme(i)) {
				save_var selected_color_scheme = i;
				set_color_pair(nums[0], nums[1], nums[2]);
			}
		} else if (parse_string(1, "@.*", scheme)) {
			i = find_index(scheme + 1);
			if (is_color_scheme(i))
				selected_color_scheme = i;
		} else if (parse_string(1, ".*", msg))
			say("%s", msg);	// Display the error message.
	} while (nl_forward());
	refresh();
}

get_choice(list, resp, title, msg, b1, b2, b3)
char *resp, *title, *msg, *b1, *b2, *b3;
{
	int w = window_handle, win;

	error_if_input(1);	/* error if we're in the middle of input */
	save_var _doing_input = DI_LINEINPUT;
	win = new_menu(list);
	window_title(win, BTOP, TITLECENTER, title);
	window_to_fit(screen_lines - 7);
	window_width_to_fit();
	save_var _window_flags |= HIGHLIGHT_LINE;
	save_var search_in_menu = 0;
	save_var display_column = 0;	// don't wrap
	if (*resp) {			// Otherwise use initial position.
		point = 0;
		col_search(resp, 0);	// Searches from current position.
	}
	if (is_gui && want_gui_prompts) {
		one_window_to_dialog(title, win, b1, b2, b3);
	} else if (msg)
		say(msg);
	select_menu_item(0, win, w, 0);
	grab_full_line(list, resp);
	ungot_key = -1;
	return !user_abort;
}

dired_show_live_link(char *file, char *msg)
{
	struct file_info fi;

	save_var readonly_warning |= ROWARN_GREP, no_popup_errors = 1;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	save_var bell_on_read_error = 0;
	if (!*msg)
		if (check_file(file, &fi) != CHECK_FILE)
			sprintf(msg, "Can't read %s: ", get_tail(file, 1));
		else if (fi.fsize > dired_live_link_limit || fi.fsize_high)
			sprintf(msg, "%s too big: ", get_tail(file, 1));
	set_read_only(0);
	filename = "";
	zap(bufname);
	fundamental_mode();
	if (!*msg) {
		if (do_file_read(file, FILETYPE_AUTO))
			sprintf(msg, "Can't read %s: ", get_tail(file, 1));
		else {
			unique_filename_identifier(file, unique_file_identifier);
			file_io_converter = 0;
			file_convert_read(FILE_CONVERT_QUIET);
			call_mode(filename);
			check_filevars();
		}
	}
	modified = 0;
	set_window_wrap(wrap_dired_live_link);	// Binary files can be slow to
	make_mode();							//  display when wrapped.
	set_read_only(1);
	buffer_not_saveable = 1;
	force_save_as = FSA_NEWFILE;
	if (!*msg)
		sprintf(msg, "%s: ", get_tail(file, 1));
	set_mode_message(msg);
}

dired_search() on dired_tab['f']
{
	if (!is_highlight_on()) {
		save_var _highlight_control;
		highlight_off();
		save_var restrict_search_type = REGRECT;
		set_spot(&restrict_search.start, 0);
		set_spot(&restrict_search.end, size());
		restrict_search.colstart = goal_column;
		restrict_search.colend = 10000;
		int h = add_region(restrict_search.start, restrict_search.end,
						   color_class pull_highlight, REGRECT);
		on_exit remove_region(h);
		modify_region(h, MRSTARTCOL, restrict_search.colstart);
		modify_region(h, MRENDCOL, restrict_search.colend);
	}
	incremental_search();
}

dired_do_delete(name, to, in_dired)
char *name, *to;
{
	int isdir = in_dired ? dired_isdir() : 0;

	to = to;
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER,
					  isdir == 1 ? "RMD %s\n" : "DELE %s\n",
					  get_url_file_part(name, 0));
	} else {
		if (in_dired)
			noteput("Deleting %f.", name);
		if ((isdir && dired_isdir() == 1) ? dired_warn_rmdir(name)
			  : dired_warn_delete(name))
			return 1;
		if (!isdir)
			maybe_delete_spell_ignore_list(name);
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

char *dired_skip_line_pat();

color_dired_range(from, to)
{
	int color, namecolor, old, namestart, nameend, isdir;
	char buf[FNAMELEN];

	if (from >= to)
		return to;
	detect_dired_format();
	save_var point, matchstart, matchend;
	point = from;
	save_var narrow_start = give_begin_line();
	point = to;
	nl_forward();
	save_var narrow_end = size() - point;
	point = 0;
	while (point < size() - narrow_end) {
		switch (curchar()) {
			case 'D': color = color_class dired_deleted; break;
			case 'U': case 'L': case 'R': case 'G':
			case 'M': color = color_class dired_moved; break;
			case 'C': color = color_class dired_copied; break;
			default: color = color_class dired_normal; break;
		}
		old = point;
		if (parse_string(1, dired_skip_line_pat()) != 0) {
			color = color_class dired_header;
			namestart = matchend;
			nameend = give_end_line();
			namecolor = color_class dired_dir_name;
		} else {
			dired_getindex(buf, goal_column);
			namestart = matchstart;
			nameend = matchend;
			namecolor = color_class dired_file_name;
			if (dired_format == DF_VMS)
				nameend = point + parse_string(1, "[^\n;]*;?[0-9]*");
			isdir = dired_isdir();
			if (isdir) {
				namecolor = color_class dired_dir_name;
				if (!dired_format)
					namestart = column_to_pos(ISDIR_START);
				if (isdir == 3 && parse_string(1, ".* -> "))
					nameend = matchend;
			}
			if (!is_remote_file(get_dired_buffer_pattern())
				  && character(namestart - 1) == ' ')
				namestart--;
		}
		if (nl_forward())
			set_character_color(point - 1, point, -1);
		set_character_color(old, point - 1, color);
		set_character_color(namestart, nameend, namecolor);
	}
	return point;
}

abbreviate_file_name(char *p, int room)
{
	char fname[FNAMELEN * 2], *s, *slash = 0;

	if (strlen(p) <= room || strlen(p) * sizeof(char) > sizeof(fname) - 1)
		return;		// No change if it all fits, or it's too big for us.
	if (room < 10)
		room = 10;
	strcpy(fname, p);
	*p = 0;

	for (s = fname; *s; s++)	// Find last path sep.
		if (is_path_separator(*s) && s > fname + 2 && s[1])
			  slash = s + 1;		// Ignore ones at start, or if it ends with one.
	if (!slash) {		// If there's no path sep, just chop end of filename.
		sprintf(p, "%s", fname);
		if (strlen(fname) > room)
			strcpy(p + room - 3, "...");
		return;
	}
	// If there's room for all of last element, chop start of filename.
	if (strlen(slash) + 2 < room) {
		// Since strlen(fname) > room, this pos isn't too low.
		sprintf(p, "...%s", fname + strlen(fname) - room + 3);
		return;
	}
	// Otherwise (not enough room for all of last element), chop at
	// both ends, including only start of last element.
	sprintf(p, "...%.*s...", room - 6, slash);
}

command edit_variables()
{		// list variables for editing; with arg, includes system vars
	int listbuf, orig_buf, orig, part = 1;
	int listwin, helpwin, orig_win, win, listheight = 9;
	int setit, origpos, ind, type, hasdef, i, lasthelp;
	TIMER ts;	// for searching in list
	char var[100], partial[100], *val;
	char tmp[FNAMELEN * 2], curval[FNAMELEN * 2];
	jmp_buf this_level;

	tiled_only();
	if (use_winhelp()) {
		if (has_arg)
			win_help_string("", "all variables, list of");
		else
			win_help_string("", "preference variables, list of");
		return;
	}
	if (screen_lines > 20)
		listheight = screen_lines * 40 / 100;
	orig_buf = bufnum;
	orig_win = window_handle;
	noteput("Listing variables...");
	save_var abort_searching = 0;
	if (has_arg)
		save_var show_all_variables = 1;
	iter = has_arg = lasthelp = 0;
	listbuf = tmp_buf();	// put list of variables here
	bufnum = listbuf;
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	save_var _nt_filter = BIT(NT_VAR) | BIT(NT_BUFVAR) | BIT(NT_WINVAR)
			      | BIT(NT_BUILTVAR);
	list_matches("", var_match, 0, listbuf);	// construct list of vars
	point = 0;
	listwin = add_popup(_view_left, 1, 30, listheight,
						ALL_BORD(BSINGLE), listbuf);
	set_menu_look(listwin);
	window_title(listwin, BTOP, TITLECENTER, " Variables ");
	mode_keys = menu_tab;
	_window_flags |= HIGHLIGHT_LINE;

	get_doc(0);
	helpwin = add_popup(_view_left, listheight + 2,	// put help text here
						MIN(screen_cols - _view_left - _view_right, 80),
						11, _view_border, zap(HELPBUF));

	text_color = (is_gui && want_gui_prompts) ? color_class text : color_class viewed_text;
	display_column = 0;		/* prevent continuation lines */

	restore_vars();		// don't let save_var undo user's changes
	time_begin(&ts, 0);
	while (is_window(listwin) && is_window(helpwin)) {
		setit = 0;		// assume we don't want to set variable
		window_handle = listwin;
		origpos = point;
		to_begin_line();
		grab_full_line(bufnum, var);	// get name of current variable
		ind = find_index(var);
		switch (name_type(ind)) {	// has a default value too?
			case NT_BUFVAR:	hasdef = 1; break;
			case NT_WINVAR: hasdef = 2; break;
			default:	hasdef = 0; break;
		}
		type = vartype(ind);

		if (ind != lasthelp) {	// show info on variable
			window_handle = orig_win; // if buffer/window-specific,
			bufnum = orig_buf;	// get value from before this
			val = get_var_val(ind, curval, 0); // command

			window_handle = helpwin;
			make_var_help(var, orig_buf, orig_win, 1);
			window_to_fit(screen_lines - listheight - 6);
			sprintf(tmp, " %s ", var);
			window_title(helpwin, BTOP, TITLECENTER, tmp);
			display_scroll_bar = !window_extra_lines();
			lasthelp = ind;
		}
		window_handle = helpwin;
		build_window();
		window_title(helpwin, BBOTTOM, TITLERIGHT(2), 
					 all_blanks(window_end, size() - narrow_end)
					 ? "" : " Press <Space> for more ");
		window_handle = listwin;
//		do {			// prompt for a key
			maybe_refresh();
			note("Select a variable & press <Enter> to set, "
				 "<Esc> when done, ? for help.");
			i = permit_window_keys;
			permit_window_keys = PERMIT_SCROLL_KEY;
			getkey();
			permit_window_keys = i;
//		} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		if (key == ESC || user_abort || key == abort_key)
			break;
		else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (is_window(win) != ISPOPUP)
				break;
			window_handle = win;
			get_text_area(win);
			if (in_text_area(mouse_x, mouse_y)) {
				point_to_pos(mouse_x, mouse_y, mouse_screen);
				if (win == listwin && IS_MOUSE_DOUBLE(key))
					setit = 1;
				else if (win == helpwin)
					key = KEYPGDN;
			} else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
		} else if (is_search_dir()) {
			save_var top_level = &this_level;
			if (!setjmp(top_level))	// come here on abort
				do_again();	// let search cmds run normally
			restore_vars();
			continue;
		}
		if (!isdigit(key))	// do arrow keys, C-n, etc.
			switch (get_direction()) {
				case BTOP:
					if (!nl_reverse()) point = size();
					break;
				case BBOTTOM:
					if (!nl_forward()) point = 0;
					break;
			}
		switch (key) {
			case KEYHOME:
			case NUMDIGIT(7): point = 0; break;
			case KEYEND:
			case NUMDIGIT(1): point = size(); break;
			case KEYPGDN:
			case NUMDIGIT(3): next_page(); break;
			case KEYPGUP:
			case NUMDIGIT(9): previous_page(); break;
			case '\n':
			case '\r':
				setit = 1; break;
			case ' ':
				window_handle = helpwin;
				next_page();
				break;
			case '\b':
			case GREYBACK:
				if (!time_done(&ts)) { // backsp in srch
					if (part > 1)
						part--;
					goto search_for;
				}
				window_handle = helpwin; // or scroll help
				previous_page();
				break;
			default:
				if (binding_of(key, root_keys) == (short) help
				    || key == '?') {
					help_on_command("edit-variables");
					lasthelp = 0;
				} else if (key >= ' ' && key <= '~') {
					if (time_done(&ts))
						part = 1; // new search
					partial[part++] = key;
search_for:				orig = point, point = 0;
					partial[0] = '^';
					partial[part] = 0;
					save_var see_delay = new_search_delay;
					say("Searching for %s", partial+1);
					restore_vars();
					if (do_searching(FOLD|REGEX, partial)
					    != 1)
						point = orig;
					time_begin(&ts, new_search_delay);
				} else if (IS_WIN_KEY(key) || key == '\t' || key == BACKTAB) {
					run_topkey();
				} else if (IS_TRUE_KEY(key)
					 && !setit && point == origpos) {
					maybe_ding(bell_on_bad_key);
				}
				break;
		}
		if (setit) {		// user wants to set a variable
			refresh();
			part = 1;	// reset searcher
			ungot_key = CTRL('S');	// bring in default value
			save_var top_level = &this_level;
			i = 0;
			if (setjmp(top_level))	// come here on abort
				i = 1;		// and record it
			else if (is_gui && want_gui_prompts)
				get_strpopup(tmp, var, val, var);
			else
				get_strnone(tmp, "New value: ", val);
			restore_vars();
			if (i)		// aborted
				continue;
			if (type <= TYPE_INT) {
				numtoi(tmp);
				if (got_bad_number)
					continue;
			}
			i = 0;		// assume current val only
			if (is_gui && want_gui_prompts && hasdef)
				i = 1;			// assume set both, for now
			else if (hasdef) {
				sayput("Set D=default value of %s, C=current"
				       " value, or B=both? [B]", var);
				getkey();
				if (user_abort)
					continue;
				if (toupper(key) == 'D')
					i = -1;
				else if (toupper(key) != 'C')
					i = 1;	// set current & default
			}
			window_handle = orig_win;
			bufnum = orig_buf;
			if (i >= 0)
				set_var_val(ind, tmp, 0);
			if (i)
				set_var_val(ind, tmp, 1);
			full_redraw = 1; // might have set a display-type var
			lasthelp = 0;	// refresh help win
		}
	}
	say("");
	if (is_window(orig_win))
		window_handle = orig_win;
	remove_window(listwin);
	remove_window(helpwin);
	buf_delete(listbuf);
	delete_buffer(HELPBUF);
	user_abort = 0;
}

// Handle activity in scp buffer.
scp_handler(int activity, int buf, int from, int to)
{
	struct scp_job *p;

	save_var bufnum = buf, abort_searching = 0;
	for (p = scp_job; p; p = p->next)	// Find the data for this buffer.
		if (bufnum == p->buf)
			break;
	if (!p || !p->destbuf)
		return;
	if (activity == NET_DONE)
		scp_handle_done(p);
	else if (activity == NET_RECV) {
		if (p->op == FTP_LIST || p->op == FTP_MISC || (p->flags & SCP_IS_XFER))
			scp_handle_list(p, from, to);
		else
			scp_handle_xfer(p, from, to);
		if (matches_at(to, -1, "(.* |^)pass(word|phrase).*: ?")) {
			ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
			menu_command = "password-prompt";
		}
	}
	refresh();
}

check_dates(save)
{
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum, pos;

	save_var reject_client_connections |= 2;
	if (num_of_popups() > 0)	// Don't pop up over bufed or similar.
		return 0;
	if (file_date_skip_drives && check_dates_skip_drive())
		return 0;
	ignoring_file_change = 0;	// Resume checking on getting focus.
	if (!want_warn || !filename || then->check_type != CHECK_FILE
			|| check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if ((readonly_warning & ROWARN_REFRESH) && !save
		  && (file_info.attr ^ now.attr) & ATTR_READONLY)
		return 2;

	if (!compare_dates(&now, then)) {
		maybe_add_readonly_once(&now);
		return 0;
	}
	switch_to_buffer(bufnum);
	maybe_add_readonly_once(&now);

	if ((auto_read_changed_file || _force_auto_read_changed_file)
		&& !save && maybe_auto_read(&now))
		return 1;
	if ((readonly_warning & (ROWARN_GREP | ROWARN_REFRESH)) && !save)
		return 2;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 30;
	save_var return_raw_buttons = 1;
	while (!user_abort) {
		show_date_warning(temp, orig, then, &now, save);
		bufnum = orig;
		pos = point;
		if (is_gui && want_gui_prompts) {
			int win, owin = window_handle;

			remove_final_view();
			save_var _doing_input = DI_VIEW;
			win = add_popup(0, 0, 75, 15, _view_border, temp);
			go_to_link = 0;
			display_dialog_box("FileDateBox2", "File has changed on disk",
							   win, 0, 0,
							   "Compare", "Read", save ? "Save" : "Ignore");
			text_color = (is_gui && want_gui_prompts) ? color_class text
						 : color_class viewed_text;
			point = 0;
			view_loop(win);
			switch (key_is_button) {
				case 1:	key = 'C'; break;
				case 2:	key = 'R'; break;
				case 3:	key = save ? 'S' : 'I'; break;
			}
			window_handle = owin;
			remove_window(win);
			if (user_abort)
				break;
		} else {
			view_buf(temp, 0);
			if (key == abort_key)
				user_abort = 1;
			if (user_abort)
				break;
			getkey();
		}
		char junk[FNAMELEN];
		if (key == WIN_DRAG_DROP)	// Ignore dropped files here.
			while(drag_drop_result(junk))
				;
		key = toupper(key);
		if (key == 'C') {
			ignoring_file_change = 2;	// Don't check for a while.
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, get_default_translation_type(filename));
			point = pos;
			to_begin_line();
			set_buffer_filename("");
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, reading_translation_type())) {
				say("%s: Read from disk.", filename);
				decompress_after_reading();
			}
			point = pos;
			to_begin_line();	// Don't check immediately after dialog ends.
			ignoring_file_change = is_gui ? 2 : 0;
			break;
		} else if (key == 'D') {
			save_var compare_to_prior_version_style = 1;
			switch (do_compare_to_prior_version(0)) {
				case 1:
					say("The version on disk is identical.");
					delay(200, 0);
					break;
				case 0:
					say("Displaying changes from disk version to edited version.");
					window_one();
					buf_delete(temp);
					quick_abort();		/* abort out of calling func */
					break;
			}
			break;
		} else if (key == (save ? 'S' : 'I')) {
			if (key == 'I')
				ignoring_file_change = 100;	// Don't check on getting focus.
			else
				ignoring_file_change = 2;	// Don't check immediately after dialog ends.
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		else if (key == 'A')
			ask_read_changed_file();
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

compare_dates(a, b)
struct file_info *a, *b;
{
	int diff;

	// Avoid Windows network bug that returns size of zero just after
	// writing.  Check sizes only if both are nonzero.
	if (b->fsize && a->fsize)
		if (a->fsize != b->fsize || a->fsize_high != b->fsize_high)
			return 1;
	// We test raw times to avoid an NT problem with daylight savings
	// time.  (Files on some file systems change their timestamps when
	// daylight savings time starts or ends, while files on other file
	// systems do not.)  Also we can easily ignore small time differences
	// which can occur over networks or in various other situations.
	diff = b->raw_file_date_low - a->raw_file_date_low;
	if (diff < 0)
		diff = -diff;
	if (a->raw_file_date_high || a->raw_file_date_low)	// Good, we have raw times.
		if (a->raw_file_date_high == b->raw_file_date_high)
			if (diff < file_date_tolerance)		// Raw times are available,
				return 0;					// and close, so they're equal.

	return a->year != b->year || a->month != b->month
			|| a->day != b->day || a->hour != b->hour || a->minute != b->minute
			|| a->second != b->second;
}

do_print_region(b, all, silent)
{
	int buf, err, skipped_lines, cnt, cols, width = 80;
	char to[FNAMELEN], title[FNAMELEN], *dest;

	iter = 0;
	dest = is_unix ? print_destination_unix : print_destination;
	if ((has_feature & FEAT_OS_PRINTING) && want_gui_printing) {
		print_options.flags = all ? 0 : PD_SELECTION;
		if (silent)
			print_options.flags |= PD_RETURNDEFAULT;
		if (!select_printer(&print_options))
			return -1;		// User canceled.
		all = !(print_options.flags & PD_SELECTION);
		width = print_options.width;
	} else if (!silent) {
		get_strdef(to, "Print to device (or !command)", dest);
		strcpy(dest, to);
	}
	if (!*filename)
		strcpy(title, bufname);
	else if (print_heading & PRHEAD_PATH) {
		strcpy(title, filename);
		abbreviate_file_name(title, width);
	} else
		strcpy(title, get_tail(filename, 1));
	save_var bufnum = b;
	buf = tmp_buf();
	save_spot point, mark;
	fix_region();		// handle other types of regions
	recolor_buffer_range(0, size());
	skipped_lines = all ? 0 : lines_between(0, point, 0);
	if (all)
		buf_xfer_colors(buf, 0, size());
	else if (region_type() == REGRECT)
		extract_rectangle_colors(buf, 0, 1);
	else
		buf_xfer_colors(buf, point, mark);
	bufnum = buf;
	safe_copy_buffer_variables(buf, b);

	point = 0;
	if (!print_tabs || print_line_numbers)
		tab_convert(0, size(), 0);
	if (print_line_numbers) {
		cnt = lines_between(0, size(), 0) + 1 + skipped_lines;
		for (cols = 0; cnt > 0; cols++)
			cnt /= 10;		// Compute maximum width of line numbers.
		cnt = 1;
		do {
			if (point >= size())
				break;
			bprintf("%*d  ", cols, skipped_lines + cnt++);
			set_character_color(give_begin_line(), point, -1);
		} while (nl_forward());
		point = 0;
	}
	do_print_adjustments();
	if ((has_feature & FEAT_OS_PRINTING) && want_gui_printing) {
		gui_print(title, title);
		err = 0;
	} else
		err = send_to_file_or_cmd(buf, dest, title);
	bufnum = b;
	buf_delete(buf);
	if (!err && !all)
		highlight_off();
	if (!err && !((has_feature & FEAT_OS_PRINTING) && want_gui_printing))
		say("%s printed.", all ? "Buffer" : "Region");
	return err;
}

do_revert_file(int trans)
{
	char pr[FNAMELEN], rel[FNAMELEN];

	save_var point;
	if (!*filename || force_save_as)
		error("No file for this buffer.");
	iter = 0;
	if (modified) {		/* buffer need to be saved? */
		relative(filename, rel);
		sprintf(pr, "Discard changes and revert to %s? ", rel);
		if (!ask_yn("Discard changes", pr, "Discard", "Cancel", 1))
			return 0;
	}
	read_file(filename, trans);
	build_first = 1;
	fix_window_start();
	return 1;
}

command revert_file()
{
	do_revert_file(reading_translation_type());
}

maybe_delete_autosave()
{
	auto_save_failed = 0;
	if (!(want_auto_save & 4) && strcmp(auto_save_name, "%f")
		  && autosave_ok(bufnum)) {
		char name[FNAMELEN];
		build_filename(name, auto_save_name, filename);
		delete_file(name);
		if (check_file(filename) != CHECK_FILE)
			error("Deleted auto-saved file and now original file is gone: %f",
				  filename);
	}
}

do_when_idle_autosave(int cnt)
{
	if (auto_save_idle_seconds < 1 || !want_auto_save)
		return -1;
	if (!cnt)
		time_begin(&auto_save_idle_timer, auto_save_idle_seconds * 100);
	int left = time_remaining(&auto_save_idle_timer);
	if (left > 0)
		return left;
	_auto_save_counter = 0;
	auto_save_buffers();
	time_begin(&auto_save_idle_timer, auto_save_idle_seconds * 100);
	return -1;
}

auto_save_one()
{
	char name[FNAMELEN], altname[FNAMELEN];
	int err, origfile = !strcmp(auto_save_name, "%f"), from, to;

	if (auto_save_failed)	// It failed last time, so no saving for now.
		return;
	if (!modified_buffer_region(&from, &to, "auto-save"))
		return;			// Buffer hasn't been modified since last auto-save.
	build_filename(name, auto_save_name, filename);
	if (_read_aborted || !autosave_ok(bufnum))		// If we didn't read
		return;						// whole file in, don't save it.
	else if (!origfile) {			// auto-save to a different file
		sprintf(altname, "%s.epstmp", name);
		delete_file(altname);
		rename_file(name, altname);
		err = file_convert_write(name, translation_type, NULL);
		save_var case_fold = (get_foldfile() != 0);
		if (strfcmp(altname, name))
			delete_file(altname);
	} else if (check_dates(1))	// make sure other prog hasn't changed
		return;
	else				// get new time info
		err = file_convert_write(filename, translation_type, &file_info);
	if (err) {
		save_var no_popup_errors = 1; // Don't pop up dialog when autosave fails.
		file_error(err, name, "couldn't autosave");
		maybe_ding(bell_on_autosave_error);
		auto_save_failed = 1;
	} else {
		if (want_auto_save & 2)
			show_text(0, 1, "%f autosaved.", name);
		reset_modified_buffer_region("auto-save");
	}
	if (!origfile)
		modified = 1;
}

check_buffer_filevars(int target)
{
	char buf[FNAMELEN], var[FNAMELEN], pat[2048], *list, *s;
	int i, cur = bufnum;

	save_var bufnum;
	if (look_for_filevar_pattern("tab-(size|width)", "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			tab_size = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(margin-right|fill-column)",
								 "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			margin_right = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(auto-)?fill-mode",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			call_maybe_break_line(i);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(indent-with-tabs|indent-tabs-mode)",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		indent_with_tabs = numtoi(buf);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(write-line-translate)",
								 "([0-9]|dos|windows|binary|unix|mac|auto)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		set_line_translate_from_text(buf);
		bufnum = cur;
	}
	list = "c-indent|soft-tab-size|over-mode|auto-indent"
		   "|goal-column|auto-fill-indents|virtual-space"
		   "|want-backups"
		   "|case-fold|sort-case-fold|delete-hacking-tabs"
		   "|undo-size|indents-separate-paragraphs"
		   "|tex-paragraphs|comment-column|want-warn"
		   "|auto-read-changed-file|perl-indent"
		   "|html-paragraph-is-container|tex-force-latex"
		   "|concurrent-compile|vbasic-indent";
	sprintf(pat, "(%s)", list);
	if (look_for_filevar_pattern(pat, "([0-9]+|nil)", NULL)) {
		for (;;) {
			strncpy(var, list, (sizeof(var) / sizeof(*var)) - 1);
			if (s = strchr(var, '|'))
				*s = 0;
			if (look_for_filevar_pattern(var, "([0-9]+|nil)", NULL)) {
				grab(find_group(1, 1), find_group(1, 0), buf);
				bufnum = target;
				if (i = find_index(var))
					do_setting(i, buf, 0);
				else
					say("Unknown mode variable %s.", var);
				bufnum = cur;
			}
			if (s = strchr(list, '|'))
				list = s + 1;
			else
				break;
		}
	}
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "(sh|csh|bash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

set_line_translate_from_text(char *s)
{
	int trans, enc = GET_ENCODING(translation_type);

	if (isdigit(*s))
		trans = numtoi(s);
	else switch (tolower(*s)) {
		case 'd':
		case 'w':	trans = FILETYPE_MSDOS; break;
		case 'b':	trans = FILETYPE_BINARY; break;
		case 'u':	trans = FILETYPE_UNIX; break;
		case 'm':	trans = FILETYPE_MAC; break;
		case 'a':	trans = FILETYPE_AUTO; break;
	}
	translation_type = MAKE_TRANSLATE(enc, trans, 1);
}

html_help_string(char *topic)
{
	char url[FNAMELEN], prog[FNAMELEN], *p, *q, cmdline[100], dir[FNAMELEN];

	if (epsilon_manual_port > 0)
		sprintf(url, "http://127.0.0.1:%d/man/", epsilon_manual_port);
	else
		strcpy(url, "http://www.lugaru.com/cgi-bin/lhelp/man/");
	for (p = topic, q = url + strlen(url); *p; p++, q++)
		if (*p == ' ') {
			strcpy(q, "%20");
			q += 2;
		} else
			*q = *p;
	*q = 0;
	if (!*topic)
		strcat(url, "0");
	p = lookpath("lhelp/epsilon.lhl");
	if (p) {		// Find name of directory containing our lhelp files.
		strcpy(dir, p);
		*get_tail(dir, 0) = 0;
	} else
		get_executable_directory(dir);
	convert_to_8_3_filename(dir);
	if (opsys == OS_UNIX) {
		sayput("Running a browser...");
		get_executable_file(prog, "lhelpd", 1);
		sprintf(prog + strlen(prog), " -p %d -d \"%s\" &",
				epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			shell("", prog, "");
	} else if (is_gui || is_win32) {
		sayput("Starting a browser, please wait...");
		expire_message = 1;
		get_executable_directory(prog);
		strcat(prog, "lhelp.exe");
		sprintf(cmdline, "lhelp -p %d -d %s", epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			winexec(prog, cmdline, SW_SHOWMINNOACTIVE, 0);
	} else
		error("Don't know how to run a web browser in this environment, sorry.");
	show_url(url);
	note("");
}

show_windows_help2(char *topic)
{
	char prog[FNAMELEN], cmdline[FNAMELEN], *s, col[FNAMELEN];
	int talk = 0, regkey;

	if (!is_win32)
		error("This type of help is only available under Windows.");
	save_var bufnum = create("-windows-help2-helper");
	if (!is_process_buffer(bufnum)) {
		if (size())
			say("Possible problem running mshelp2.vbs helper detected.");
		if (RegOpenKey(HKEY_CLASSES_ROOT, "DExplore.AppObj", &regkey))
			return 1;
		RegCloseKey(regkey);
		sayput("Starting MS-Help2...");
		talk = 1;
		s = lookpath("mshelp2.vbs");
		if (!s) {
			get_executable_directory(prog);
			strcat(prog, "mshelp2.vbs");
		} else
			strcpy(prog, s);
		convert_to_8_3_filename(prog);
		sprintf(cmdline, "cscript %s", prog);
		if (concur_shell("", cmdline, NULL, bufname))
			return 2;
		mshelp2_choose_collection(col);
		point = size();
		if (*col)
			bprintf("collection %s\n", col);
	}
	point = size();
	bprintf("keyword %s\n", topic);
	delay(10, 0);
	if (talk)
		say("");
	if (!is_process_buffer(bufnum))
		return 3;
	return 0;
}

color_html_include(int limit)
{
	int s, t, c, lang, orig = point, stop;
	char *end, *name;

	search(-1, "<");	// Find start of include.
	c = character(point + 1);
	s = point;
	switch (c) {
		case '%':	if (parse_string(1, "%<%%--")) {
						search(1, "--%>");	// Found a JSP "hidden comment".
						set_character_color(s, point, color_class html_comment);
						return 1;
					}
					point += parse_string(1, "%<%%[a-z0-9]*");
					end = "%>";
					lang = get_html_asp_language(sublanguage_code
						== SUBLANG_XML ? xml_asp_coloring : html_asp_coloring);
					break;
		case '?':	if (sublanguage_code == SUBLANG_XML)
						lang = xml_php_coloring;
					else if (parse_string(1, "%<%?php"))
						lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
					else if (parse_string(1, "%<%?python"))
						lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
					else
						lang = html_php_coloring;
					point += parse_string(1, "%<%?[a-z0-9]*");
					end = "?>";
					break;
		default:	point++;
					end = ">";
					lang = get_html_asp_language(html_asp_coloring);
					break;
	}
	if (lang == HT_IGNORE) {
		point = orig;
		return 0;
	}
	t = point;
	// First mark whole section embedded.
	set_character_color(s, point, color_class html_embedded);
	set_tagged_region("color-as-unit", s, point, 99);
	for (;;) {
		stop = search(1, end) ? matchstart : point;	// Find the end marker.
		color_embedded_code(t, point, lang, limit);	// Color it too for now.
		if (c == '%' && lang == HT_VBSCRIPT)
			break;	// Don't do this for comment in ASP vbscript.
		// If it wound up being colored as a comment or string, that's
		// not really the end, so find the next marker.
		name = name_color_class(get_character_color(stop, &t));
		if (point <= stop || !name || !strstr(name, "comment")
			  && !strstr(name, "string") && !strstr(name, "constant"))
			break;
		// Force recoloring of that entire comment/string.
		set_tagged_region("needs-color", t, s = point, -1);
	}
	// Color the marker ending the embedded code.
	set_character_color(stop, point, color_class html_embedded);
	set_tagged_region("color-as-unit", stop, point, 99);
	return 1;
}

int give_html_indent()
{
	if (sublanguage_code == SUBLANG_XML)
		return (xml_indent > 0) ? xml_indent : tab_size;
	return (html_indent > 0) ? html_indent : tab_size;
}

command indent_region()
{
	int temp;
	int (*func)() = (int (*)()) (mode_keys['\t'] ? mode_keys['\t']
		: root_keys['\t']);
	int progress, cnt = 0;

	fix_region();
	highlight_off();
	if (nl_reverse())
		point++;
	temp = point;
	progress = (mark - point > 4000);
	this_cmd = CMD_INDENT_REG;
	while (point < mark) {
		(*func)();
		if (curchar() == '\n')
			delete_horizontal_space();
		nl_forward();
		if (!(cnt % 25) && progress && temp - mark != 0)
			note("Indenting region: %d%%...",
				 muldiv(point - temp, 100, mark - temp));
		cnt++;
		check_abort();
	}
	if (progress)
		say("");
	mark = temp;
	this_cmd = iter = 0;
}

to_info_buffer()
{
	to_buffer(INFO_BUF);
	discardable_buffer = 1;
	force_save_as = FSA_READONLY;
	set_window_wrap(wrap_info_mode);
	tab_size = 8;
	undo_keeps_narrowing = 1;	// Keep it from removing narrowing.
}

ini_func_name_finder()
{
	int from, to;
	save_var point;
	to_end_line();	// Might be starting from a section line.
	*display_func_name = 0;
	if (re_search(-1, "^<hspace>*%[(.*)%]")) {
		from = find_group(1, 1);
		to = find_group(1, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
	}
	return 1;
}

get_text_area(w_handle)
{
	save_var window_handle = w_handle;

	w_left = window_text_left(window_handle);
	w_top = window_text_top(window_handle);
	w_right = w_left + window_content_width() - 1;
	w_bottom = w_top + text_height() - 1;
}

proc_view_error_lines(int buf, int pos)
{
	if (!process_view_error_lines)
		return;
	if (process_view_error_lines < -90
		  || process_view_error_lines > screen_lines - 2)
		process_view_error_lines = screen_lines / 3;
	if (process_view_error_lines < 0)	// It's a percentage.
		save_var process_view_error_lines
				= muldiv(-process_view_error_lines, screen_lines, 100);
	int owin = window_handle;
	int win = buf_in_window(buf);
	if (win >= 0)
		window_handle = win;
	else {
		window_number = 0;
		window_number--;		// Get to last window on screen.
		while (window_height < process_view_error_lines) {
			int oh = window_height++;
			if (oh == window_height)
				if (give_window_space(BTOP))
					break;
		}
		if (window_height > process_view_error_lines) {
			window_split(HORIZONTAL);
			window_height = process_view_error_lines;
		}
		to_buffer_num(buf);
	}
	point = window_start = pos;
	set_window_wrap(wrap_view_error_lines);
	if (is_window(owin))
		window_handle = owin;
	else
		window_number--;
}

color_perl_keyword(int start)
{
	char buf[100];

	if (point - start < sizeof(buf) / sizeof(*buf) - 10
		  && (point - start > 1 || character(start - 1) != '-')) {
		buf[0] = '|';		// get identifier, between | chars
		grab(start, point, buf + 1);
		strcat(buf, "|");
	} else	// Too long to be a keyword, or else it's -s or similar.
		strcpy(buf, "not a keyword");
	if (sublanguage_code == SUBLANG_PHP) {
		if (strstr("|if|else|elseif|while|do|for|foreach|break|continue"
				   "|switch|declare|return|require|require_once|include"
				   "|include_once|endif|endwhile|endfor|endforeach|endswitch"
				   "|function|class|", buf))
			  set_character_color(start, point, color_class perl_keyword);
		else if (isdigit(character(start))) {		// a number
			if (matches_at(start, -1, "-?%.?"))	// number like -.25
				start = matchend;
			point += parse_string(1, "([A-Za-z0-9._]|[Ee]-)*");		// -3.0e-7
			set_character_color(start, point, color_class perl_constant);
		} else
			set_character_color(start, point, color_class perl_function);
		return;
	}
	if (strstr("|package|sub|", buf)) {
		point += parse_string(1, "[ \t]+");
		set_character_color(start, point, color_class perl_keyword);
		if (parse_string(1, "(<word>+('|::))*<word>+"))
			set_character_color(matchstart, point = matchend,
								color_class perl_function);
	} else if (strstr("|import|no|require|use|", buf)) {
		point += parse_string(1, "[ \t]+");
		set_character_color(start, point, color_class perl_keyword);
		if (parse_string(1, "(<word>+('|::))*<word>+"))
			set_character_color(matchstart, point = matchend,
								color_class perl_constant);
	} else if (strstr("|die|do|dump|else|elsif|eval|exec|exit|for|foreach|if|"
			   "return|unless|until|while|BEGIN|END|", buf)) {
		set_character_color(start, point, color_class perl_keyword);
	} else if (strstr("|local|my|", buf)) {		// special color
		set_character_color(start, point, color_class perl_constant);
	} else if (strstr("|continue|goto|last|next|redo|", buf)) {
		point += parse_string(1, "[ \t]+");
		set_character_color(start, point, color_class perl_keyword);
		if (parse_string(1, "[0-9A-Za-z_]+"))	// A label.
			color_perl_label(matchstart, matchend);
	} else if (strstr("|qq|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 2, 0);	// One pattern, always interpolated.
	} else if (strstr("|qx|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 1, 0);	// One pattern, interpolated.
	} else if (strstr("|m|qr|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 1, 2);	// One pattern, interp, w/modifiers.
	} else if (strstr("|q|qw|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 0, 0);	// One pattern, uninterpolated.
	} else if (strstr("|s|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(1, 1, 2);	// Two patterns, interpolated.
	} else if (strstr("|tr|y|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(1, 0, 1);	// Two patterns, uninterpolated.
	} else if (strstr("|format|", buf)) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_format();
	} else if (strstr("|__END__|__DATA__|", buf)) {	// Skip to the end.
		set_character_color(start, point, color_class perl_keyword);
		set_character_color(point, size(), color_class perl_constant);
		point = size();
	} else if (character(point) == ':' && matches_at(point, 1, ":[^:]")
			   && matches_at(start, -1, "^[ \t]*")) {		// a label
		set_character_color(start, point, color_class perl_constant);
	} else
		set_character_color(start, point, color_class perl_function);
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
											   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				if (in_shell_buffer && !parse_string(-1, "([ \n\t|&;()<>]|^)#"))
					break;
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "#?((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_variable);
				else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer ? character(t - 1) != '\\'
									: !strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && character(t - 1) == '\\')
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_><word>*)"
									 : "(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

tag_mode_perl()
{
	char func[TAGLEN], pat[FNAMELEN], class_name[FNAMELEN];
	int start, i, have_class, class_end = -1;

	*class_name = 0;
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	sprintf(pat, "<^word>(%s)[ \t]+([a-zA-Z_]<word>*"
			"((::|')[a-zA-Z_]<word>*)?)",
			(sublanguage_code == SUBLANG_PHP) ? "function|class" : "sub");
	while (re_search(1, pat)) {
		if (get_character_color(find_group(1, 1)) == color_class perl_keyword
			&& get_character_color(find_group(2, 1)) == color_class perl_function) {
			grab(start = find_group(2, 1), find_group(2, 0), func);
			have_class = (sublanguage_code == SUBLANG_PHP)
						 && character(find_group(1, 1)) == 'c';
			if (*class_name && point > class_end)
				*class_name = 0;
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (curchar() != '{')
				have_class = 0;
			if (have_class && (i = perl_move_level(1, 0)) > 0) {
				if (tag_which_items & 16)
					add_tag(func, start);
				strcpy(class_name, func);
				class_end = i;
				continue;
			}
			if (curchar() == '(') {
				i = perl_move_level(1, 0);
				if (i > 0 && i < size() - narrow_end)
					point = i;
			}
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (curchar() == '{' || (tag_declarations & 1))
				add_class_tag(class_name, "::", func, start);
		}
	}
}

prefix_fill_one_paragraph(char *pat, char *skip, int force)
{
	char prefix[FNAMELEN], quoted[FNAMELEN];

	save_spot point--;
	re_search(1, "[^\n\t ]");
	if (force || matches_at(give_begin_line(), 1, pat)) {
		save_spot mark, narrow_start, narrow_end, margin_right;
		to_begin_line();
		*prefix = 0;
		parse_string(1, skip, prefix);
		if (!*prefix) {
			fill_paragraph();
			return;
		}
		margin_right -= current_column();
		if (margin_right < 1)
			margin_right = 1;
		sprintf(quoted, "^%q", prefix);
		prefix_find_fill_region(1, skip, pat);
		// We've now narrowed to one paragraph, and know the quoting
		// prefix shared by all its lines.  So strip the prefix, fill,
		// and put it back.
		if (*quoted)
			simple_re_replace(1, quoted, "");
		fill_paragraph();
		if (*prefix)
			simple_re_replace(1, "^", prefix);
	} else
		fill_paragraph();
}

compile_handler(int activity, int buf, int from, int to)
{
	UNREFERENCED_PARAMETER(to);
	if (activity == NET_DONE) {
		save_var bufnum = buf, abort_searching = 0;
		point = size();
		if (size() && character(point - 1) != '\n')
			bprintf("\n");
		bprintf("Process exited with status %d\n", from);
		update_mode_more_extra(-1);
		if (post_compile_hook_buffer && buf_exist(post_compile_hook_buffer)) {
			save_var bufnum = post_compile_hook_buffer;
			if (post_compile_hook)		// Track this compilation back to orig
				(*post_compile_hook)(6, bufnum); // buffer, then run its hook.
		}
	}
}

// The process is now ready for or has just been sent some input.
do_when_process_input(int buf, int type, int start)
{
	char hist_buf[FNAMELEN];

	save_var abort_searching = 0;
	save_var bufnum = buf;
	save_var point = start;
	if (type == PROCESS_INPUT_LINE) {
		if (opsys == OS_UNIX) {	// Our only opportunity to color.
			if (spot_to_buffer(process_input_starts_at) <= 0)
				process_input_starts_at = alloc_spot();
			*process_input_starts_at = start;
			if (process_coloring_rules & 1)
				set_character_color(give_begin_line(), start,
									color_class process_prompt);
		}
		sprintf(hist_buf, "-process-history-%s", bufname);
		create(hist_buf);
		process_cmds_to_end(hist_buf);
		nl_forward();
		if (point > start + 1)		// don't store empty (just \n) cmds
			xfer(hist_buf, start, point);
		process_cmds_to_end(hist_buf);
	} else if (type == PROCESS_INPUT_CHAR) {
		if (process_coloring_rules & 1)
			set_character_color(start, start + 1, color_class process_input);
	} else if (!type) {		// Got a request for line input (might now wait).
		if (spot_to_buffer(process_input_starts_at) <= 0)
			process_input_starts_at = alloc_spot();
		*process_input_starts_at = start;
		point = start;
		to_begin_line();
		if (process_coloring_rules & 1)
			set_character_color(point, start, color_class process_prompt);
		try_grab_directory_name(point, start);
	}
}

remove_xterm_sequences(int start)
{
	save_var point = start - 25;
	save_var case_fold = 0;
	simple_re_replace(1, "\x1B%][0-9];.*\7", "");
}

do_when_process_output(int buf, int start, int end)
{
	save_var bufnum = buf;
	save_var abort_searching = 0;
	set_character_color(start, end, color_class process_output);
	if (process_coloring_rules & 32)
		remove_xterm_sequences(start);
	if (process_coloring_rules & 2)
		color_ansi_output(start);
	if (process_coloring_rules & 8)
		process_backspace_sequences(start - 100);
	if (process_coloring_rules & 16)
		process_remove_controls(start);
	notice_password_prompt(start);
	save_var position_window_on_screen_line = 100;
	if (buf_in_window(buf) >= 0)
		refresh();
}

command filter_region()
{
	int in_place = has_arg, b, width, orig = bufnum;
	char cmd[FNAMELEN];

	iter = 0;
	get_str_auto_def(cmd, in_place ? "Send region to command" :
			   "Filter region through command");
	b = zap("filter-output");
	if (region_type() == REGRECT) {
		width = extract_rectangle(b, 0);
	} else {
		save_spot point, mark;
		fix_region();		// handle other types of regions
		buf_xfer(b, point, mark);
	}
	if (!buf_pipe_text(b, b, cmd, "", PIPE_CLEAR_BUF | (in_place ? PIPE_SYNCH : 0))) {
		if (bufnum != b)
			buf_delete(b);
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	if (in_place) {
		if (region_type() == REGRECT) {
			save_var bufnum = b;
			xfer_rectangle(orig, width, 1);
		} else {
			delete(point, mark);
			restore_vars();
			grab_buffer(b);
			fix_region();
		}
	} else {
		tiled_only();
		to_buffer_num(b);
		point = 0;
	}
	highlight_off();
}

// Does this region have unmatched (, [, or {?
int python_open_delimiter(int start, int end)
{
	save_var point = start;
	save_var narrow_end = size() - end;
	while (re_search(1, "![[({]") && get_character_syntax_color(point) == -1)
		if (to_default_move_level(1, "[({", "])}", 1) == -1)
			return 1;
	return 0;
}

// Does this region have unmatched ), ] or }?
int python_close_delimiter(int start, int end)
{
	save_var narrow_start = start;
	save_var point = end;
	while (re_search(-1, "[])}]!") && get_character_syntax_color(point - 1) == -1)
		if (to_default_move_level(-1, "])}", "[({", 1) == -1)
			return 1;
	return 0;
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN], search_other_buf_name[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	if (search_in_region && is_highlight_on()) {
		save_var _highlight_control;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	} else if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var _highlight_control;
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	if (!search_other_name) {
		sprintf(search_other_buf_name, "search-%s", bufname);
		save_var search_other_name = search_other_buf_name;
	}
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = full_key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))), "search");
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, comp_tab) == (short) inp_get_other) {
			flags = pick_search_recall(flags);
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		flags = grab_search_string(flags, &trystr);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = editwin;
	if (size()) {
		flags = grab_search_string(flags, &trystr);
		save_search_recall(flags, trystr);
	}
	if (key == abort_key || user_abort) {
		window_handle = origwin;
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			if (_last_was_regex)
				grab_search_string(flags, &_default_regex_search);
			else
				grab_search_string(flags, &_default_search);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

save_search_recall(int flags, char *str)
{
	char *srch;
	int buf = tmp_buf();

	save_var bufnum = buf, abort_searching = 0;
	bprintf("%s%s%s",
		!(flags & MODFOLD) ? "" : (flags & FOLD) ? "Fold " : "Nofold ",
		(flags & WORD) ? "Word " : "",
		(flags & REGEX) ? "Regex " : "");
	if (character(point - 1) == ' ')
		delete(point - 1, point);
	bprintf(": %s", str);
	simple_re_replace(1, "%<NL>", "%<NL>");	// Put a % before actual <NL> strings.
	simple_re_replace(1, "\n", "<NL>");	// Convert newlines to <NL>.
	srch = malloc(size() + 1);
	grab(0, size(), srch);
	restore_vars();
	buf_delete(buf);
	save_recall(srch, "search");
	if (search_other_name)
		save_recall(srch, search_other_name);
	free(srch);
}

int pick_prev_search_from_end(int *flags, int pos, char *tag)
{
	char res[FNAMELEN];
	int ret;

	save_var recall_id = tag;
	ret = get_last_recall_from_end((char *(*)())0, res, pos);
	buf_zap(bufnum);
	stuff(res);
	*flags &= ~(MODFOLD | FOLD);
	if (case_fold)
		*flags |= FOLD;
	*flags = extract_search_flags(*flags);
	if (recall_response_selected) {
		highlight_on();
		shift_selecting = 2;
	}
	return ret;
}

int pick_search_recall(int flags)
{
	if (search_other_name)
		pick_prev_search_from_end(&flags, 1, search_other_name);
	return flags;
}

restore_session(b)	/* if error in session file, return error code # */
{			/* for debugging purposes, else 0 */
	int j, err, readerrs = 0, def_buf = 0;
	char buf[FNAMELEN], file[FNAMELEN];
	char sdir[FNAMELEN];	// dir of session file

	give_session_file(sdir);
	absolute(sdir);
	*get_tail(sdir, 0) = 0;
	bufnum = b;
	point = 0;
	try_calling("restore-session-conversion-hook");
	if (!grab_line(b, buf) || strcmp(buf, SES_HDR))
		return 1;
	while (grab_line(b, buf)) {
		get_file_name_from_session_file(b, file);
		if (session_restore_files)
			bufnum = create(buf);
		if (!def_buf) {
			if (is_gui)
				show_text(0, 1, "Restoring session...");
			else
				show_text(-1, 1, "Restoring session...");
			def_buf = bufnum;
		}
		j = session_restore_one_file(b, file, sdir, &readerrs);
		if (j != -1)
			return j;
	}
	if (session_restore_prev_buffers(b))
		return 12;
	if (restore_bookmark_session(b))
		return 13;
	err = session_restore_windows(b, def_buf);
	if (err < 0)
		return 5;
	if (session_restore_misc(b))
		err = 11;
	return err ? err : readerrs ? 100 : 0;
}

jmp_buf *jbuf;

parse_cmds()	/* load cmds in current buf, return 1 if error occurred */
{
	point = 0;
	save_var _errors_in_cmd_file = 0;
	save_var now_parsing_cmds = 1;
	jmp_buf my_jbuf;
	save_var jbuf = &my_jbuf;
	if (setjmp(jbuf))			/* errors come here */
		if (_abort_load_on_error)
			return 1;
	for (;;) {
		re_search(1, "[ \n\t]*");	/* skip whitespace */
		if (point >= size() - narrow_end)
			break;
		if (_abort_load_on_error != 2)
			noteput("Loading: %d%%...", muldiv(point, 100, size()));
		switch (curchar()) {
			case ';':		/* found comment */
			case '#': nl_forward(); break;
			case '(': paren_cmds(); break;
			case '+': load_var(); break;
			case '~': load_binding(); break;
			case '&': load_color(); break;
			default:  load_other(); break;
		}
	}
	copy_bindings();
	int i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	if (!_abort_load_on_error
		  || !_errors_in_cmd_file && _abort_load_on_error != 2)
		say("%d error%s detected.", _errors_in_cmd_file,
			  _errors_in_cmd_file == 1 ? "" : "s");
	else if (_abort_load_on_error != 2)
		say("%d error%s detected, see buffer %s.", _errors_in_cmd_file,
			  _errors_in_cmd_file == 1 ? "" : "s", MSG_BUFFER);
	return _errors_in_cmd_file != 0;
}

init_error(msg, s)		/* display error and abort loading */
char *msg, *s;
{
	init_warning(msg, s);
	longjmp(jbuf, 1);
}

do_go_tag(s, n_fwd)
char *s;
{
	char pat[2048], tdir[FNAMELEN], *file, *p;
	int pos, oldbuf = bufnum, i;

	i = get_mode_based_index("tag_to_next_%s");
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	*pat = 0;
	if (*s) {
		if (p = index(s, '\t')) {
			strcpy(_last_tag_file, p + 1);
			if (strchr(p + 1, '\t'))	// Have full info already.
				sprintf(pat, "^%q", s);
			*p = 0;		/* remove file name from s */
			if (p = index(_last_tag_file, '\t'))
				*p = 0;
		} else
			*_last_tag_file = 0;
		strcpy(_last_tag, s);
		if (!*_last_tag)	/* no last tag */
			return 0;
		if (!*pat)
			if (*_last_tag_file && p && !strchr(p + 1, '\t'))
				sprintf(pat, "^%q\t%q\t%q", _last_tag, _last_tag_file, p + 1);
			else if (*_last_tag_file)
				sprintf(pat, "^%q\t%q\t", _last_tag, _last_tag_file);
			else
				sprintf(pat, "^%q\t", _last_tag);
		point = 0;
		if (!re_search(1, pat))		/* look for 1st */
			return 1;
	} else {		/* find next matching tag */
		pos = point;
		sprintf(pat, "^%q\t", _last_tag);
		switch (name_type(i)) {		// Call mode-specific version.
			case NT_SUBR: case NT_COMMAND:
			case NT_AUTOSUBR: case NT_AUTOLOAD:
				if ((*(int (*)())i)(pat, n_fwd))
					return 1;
				break;
			default:
				if (!tag_search_by_offset(pat, n_fwd))
					return 1;
		}
	}
	grab(give_begin_line(), give_end_line(), pat);	/* tag, file, pos */
	file = index(pat, '\t');
	*file++ = 0;

	p = index(file, '\t');		/* find start of pos */
	pos = strtoi(p + 1, 10);
	*p = 0;
	p = index(p + 1, '\t');		/* if p, p+1 has line's text */
	strcpy(_last_tag_file, file);
	put_directory(tdir);
	bufnum = oldbuf;
	absolute(_last_tag_file, tdir);
	tag_find_file();
	if (p)
		sprintf(tdir, "%q[^a-zA-Z0-9_]", _last_tag);
	else			// Might not include class info.
		tag_make_loose_pattern(_last_tag, tdir);
	point = pos;
	if (!parse_string(1, tdir, NULL))
		if (!p || !prox_line_search(p + 1))	/* search by text */
			return 2;
	to_begin_line();
	window_start = prev_screen_line(show_tag_line);
	return 0;
}

command tex_close_environment()
{
	char env[FNAMELEN];
	int cnt = 0 , level = 0, pos = point, col = 0;

	while (re_search(-1, "\\(begin|end){")) {
		if (want_code_coloring
			  ? (get_character_syntax_color(point) == color_class tex_comment)
			  : parse_string(-1, "%%.*"))
			continue;
		cnt++;
		if (character(point + 1) == 'e')
			level++;
		else if (--level < 0) {
			if (!parse_string(1, "\\begin{([^\n}]+)}"))
				error("Can't interpret \\begin statement");
			grab(find_group(1, 1), find_group(1, 0), env);
			if (parse_string(-1, "^[ \t]*"))
				col = current_column();
			point = pos;
			delete_horizontal_space();
			tex_open_new_line();
			to_column(col);
			bprintf("\\end{%s}", env);
			return;
		}
	}
	if (cnt)
		error("Mismatched \\begin & \\end pairs detected.");
	point = pos;
	error("No \\begin block before point");
}

do_uniq(incl_uniq, incl_dups, talk)
{
	int old = 0, ocnt, ncnt;

	if (talk)
		ocnt = count_lines_in_buf(bufnum, 0);
	save_spot point;
	for (;;) {
		point = old;
		if (!nl_forward())
			break;
		if (uniq_count_dups(old, point) ? incl_dups : incl_uniq)
			old = point;
		else
			delete(old, point);
		if (talk && size())
			noteput("%d%% complete...", muldiv(point, 100, size()));
		check_abort();
	}
	if (!incl_uniq)
		delete(old, point);
	if (talk) {
		ncnt = count_lines_in_buf(bufnum, 0);
		say("Deleted %d lines leaving %d.", ocnt - ncnt,
			  ncnt + (size() && character(size() - 1) != '\n'));
	}
}

int spell_download_one(char *dir, char *arch, int sz, int ask)
{
	char archpath[FNAMELEN], url[FNAMELEN], cmdline[FNAMELEN], *s;
	struct file_info fi;
	sprintf(archpath, "%s/%s", dir, arch);
	sprintf(url, "http://www.lugaru.com/%s", arch);
	s = lookpath(arch);
	if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9) {
		if (s && check_file(s, &fi) == CHECK_FILE && fi.fsize * 10 >= sz * 9)
			strcpy(archpath, s);
		else if (!ask)
			return 0;
		else if (!ask_yn("Download and install dictionary file?", 
						 "Download and install a dictionary file "
						 "(about 700 KB)? ", "Yes", "No", 1))
			return 1;
		else {
			say("Please wait...");
			if (download_file_to_disk(url, archpath, sz))
				error("Couldn't download dictionary file from %s", url);
		}
	}
	if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9)
		error("Bad dictionary archive %f, please retrieve manually.",
			  archpath);
	convert_to_8_3_filename(archpath);
	sprintf(cmdline, is_win32 ? "%s" : "tar xzf %s", archpath);
	pipe_text(NULL, "-spell-config", cmdline,
			  dir, PIPE_SYNCH | PIPE_NOREFRESH, 0);
	return 2;
}

spell_download(char *dir)
{
	if (!spell_download_one(dir, is_win32 ? "spell.exe" : "spell.tgz",
							is_win32 ? 2_398_133 : 2_258_676, 0))
		spell_download_one(dir, is_win32 ? "spell2.exe" : "spell2.tgz",
						   is_win32 ? 771_401 : 692_231, 1);
}

command spell_configure()
{
	char dir[FNAMELEN], file[FNAMELEN], other[FNAMELEN];
	if (exist(WORD_LIST_BUFFER))
		reload_spell_word_list();
	save_var want_gui_prompts = 0;
	get_customization_directory(dir);
	absolute(dir);
	strcat(dir, "/spell");
	mkdir(dir);
	if (check_file(dir) != CHECK_DIR)
		error("Couldn't create working directory %f", dir);
	sprintf(file, "%s/final/english-words.10", dir);
	if (check_file(file) != CHECK_FILE)		// Download and extract.
		spell_download(dir);
	if (check_file(file) != CHECK_FILE)
		error("Couldn't extract files from dictionary archive.");
	
	int region = get_key_response("Use American, British, British with -ize, "
								  "or Canadian spellings [ABZC]? ",
								  "ABZC", 0, "spell-configure");

	int help = tmp_buf(), w = add_popup(10, 10, 60, 10, _view_border, help);
	on_exit buf_delete(help), remove_window(w);
	window_title(w, BTOP, TITLECENTER,  " Spell Configuration Help ");

	show_popup_help(w, "Select a dictionary size.  Bigger dictionaries "
					" recognize rarer words, but won't catch misspellings "
					"that result in rare words.");
	int dictsz = get_key_response("Dictionary size (small, medium, large) "
								  "[SML, default L]? ", "SML", 'L',
								  "spell-configure");

	save_var insert_default_response = 1;
	show_popup_help(w, "Should the dictionary include abbreviations, "
					"contractions, uppercase words, or proper names?  Enter "
					"A, C, U, and P, in any combination, to select these.\n\nThe "
					"Abbreviations category includes abbreviations and acronyms "
					"like \"TV\", \"VIP\", and \"rpm\".\n\nThe Contractions "
					"category includes \"doesn't\" and \"I'm\".\n\nThe Uppercase "
					"category includes words that must start with an uppercase "
					"letter, like \"Doctor\", \"December\", or \"Europe\".\n\n"
					"The Proper Names category includes less common names like "
					"\"Brian\" or \"Altman\" that are not included in the "
					"Uppercase category.\n\nSome words appear in more than one "
					"category.");
	get_strdef(other, "Include Abbrevs, Contractions, "
			   "Uppercase, Proper names in dictionary", "ACUP");

	show_popup_help(w, "Should the speller permit variant spellings, instead "
					"of permitting only a single preselected spelling of each "
					"word?\n\nThis permits words like \"fiascos\" in addition to "
					"\"fiascoes\" or \"Hanukah\" in addition to \"Hanukkah\".");
	sprintf(other + strlen(other), "%c",
			get_key_response("Permit variant spellings (None, Some, Many, Lots) "
							 "[Some]? ", "NSML", 'S', "spell-configure"));
	spell_configure_helper(w);
	save_var bufnum = zap("-spell-config");
	switch (dictsz) {
		case 'L':
			grab_spell_dict_size("70", region, other);
			grab_spell_dict_size("60", region, other);
			grab_spell_dict_size("55", region, other);
		case 'M':
			grab_spell_dict_size("50", region, other);
			grab_spell_dict_size("40", region, other);
		case 'S':
			grab_spell_dict_size("35", region, other);
			grab_spell_dict_size("20", region, other);
			grab_spell_dict_size("10", region, other);
	}
	sprintf(file, "%s/../%s", dir, WORD_ORDER_FILE);
	int err = file_write(file, FILETYPE_AUTO);
	if (err)
		file_error(err, file, "write error");
	case_fold = 0;
	do_sort_region(0, size(), 0, 0);
	sprintf(file, "%s/../%s", dir, WORD_FILE_NAME);
	err = file_write(file, FILETYPE_AUTO);
	if (err)
		file_error(err, file, "write error");
	restore_vars();
	reload_spell_word_list();
	say("Dictionary configured, run spell-mode or spell-buffer.");
}

char *get_spell_language_prefix()
{
	return spell_language_prefix ? spell_language_prefix : "";
}

int spell_load_file(int flavor, int action)
{
	char fname[FNAMELEN], bname[FNAMELEN], mode[FNAMELEN], *s = 0, *t;
	char lookfor[FNAMELEN];
	int b, bit = (1 << flavor);
	switch (flavor) {
		case SPELLB_ORDERED:
			strcpy(bname, WORD_ORDER_BUFFER);
			break;
		case SPELLB_SORTED:
			strcpy(bname, WORD_LIST_BUFFER);
			break;
		case SPELLB_GLOBAL:
			strcpy(bname, "-spell-ignore-global");
			break;
		case SPELLB_PERDIR:
			if (!*filename)
				return 0;
			sprintf(bname, "-spell-ignore-%s", filename);
			*get_tail(bname, 0) = 0;
			break;
		case SPELLB_PERFILE:
			if (!*filename)
				return 0;
			sprintf(bname, "-spell-ignore-%s", filename);
			break;
		case SPELLB_PERMODE:
			clean_mode(mode);
			sprintf(bname, "-spell-ignore-mode-%s", mode);
			break;
		case SPELLB_PEREXTEN:
			if (!*filename)
				return 0;
			sprintf(bname, "-spell-ignore-extension-%s",
					get_extension(filename));
			break;
	}
	strcat(bname, get_spell_language_prefix());
	if ((b = name_to_bufnum(bname)) && !action)
		return buf_size(b) || (action == 2) ? b : 0;
	switch (flavor) {
		case SPELLB_ORDERED:
			s = WORD_ORDER_FILE;
			break;
		case SPELLB_SORTED:
			s = WORD_FILE_NAME;
			break;
		case SPELLB_GLOBAL:
			s = "ignore.lst";
			break;
		case SPELLB_PERDIR:
			strcpy(fname, filename);
			strcpy(get_tail(fname, 0), ".directory.espell");
			strcat(fname, get_spell_language_prefix());
			break;
		case SPELLB_PERFILE:
			strcpy(fname, filename);
			sprintf(get_tail(fname, 0), ".%s.espell", get_tail(filename, 0));
			strcat(fname, get_spell_language_prefix());
			break;
		case SPELLB_PERMODE:
			sprintf(fname, "ignore.%s.mode.lst", mode);
			strcpy(s = mode, fname);
			break;
		case SPELLB_PEREXTEN:
			sprintf(fname, "ignore.%s.extension.lst", get_extension(filename));
			strcpy(s = mode, fname);
			break;
	}
	if (s) {	// Search for these.
		create(bname);
		sprintf(lookfor, "%s%s", s, get_spell_language_prefix());
		if (t = lookpath(lookfor))
			strcpy(fname, t);
		else if (action == 2 && (flavor == SPELLB_GLOBAL
								 || flavor == SPELLB_PERMODE
								 || flavor == SPELLB_PEREXTEN)) {
			get_customization_directory(fname);
			strcat(fname, lookfor);
		} else
			return 0;
	} else {	// Look just once from each buffer.
		if (action)		// Clear the missing bit.
			vspell_data.missing_file &= ~bit;
		if (!(vspell_data.missing_file & bit)
			  && (check_file(fname) != CHECK_FILE))
			vspell_data.missing_file |= bit;	// Remember it's missing.
		if ((vspell_data.missing_file & bit) && action != 2)
			return 0;
	}
	save_var bufnum = create(bname);
	filename = fname;
	case_fold = 0;
	if (modified && action != 2)
		return 0;		// Don't mess with buffer while user modifies it.
	if (action == 2)
		file_read(fname, FILETYPE_AUTO);
	else if (err_file_read(fname, FILETYPE_AUTO) && action != 2)
		return 0;
	switch (flavor) {	// Sort only user's ignore buffers.
		case SPELLB_GLOBAL:
		case SPELLB_PERDIR:
		case SPELLB_PERFILE:
		case SPELLB_PERMODE:
		case SPELLB_PEREXTEN:
			save_var abort_searching = 0, show_status = 0, modified;
			buf_sort_and_uniq(bufnum);
			break;
	}
	return bufnum;
}

// Load the two dictionary files.  Sort the user one.  Return 1 if ok.
int load_spell_word_list(int force_reload)
{
	if (!spell_load_file(SPELLB_ORDERED, force_reload)) {
		if (!vspell_data.told_err) {
			if (*get_spell_language_prefix())
				say("Can't find speller word list %f%s.",
					  WORD_ORDER_FILE, get_spell_language_prefix());
			else
				say("Can't find speller word list %f, run spell-configure.",
					  WORD_ORDER_FILE);
			vspell_data.told_err = 1;
		}
		return 0;
	}
	spell_load_file(SPELLB_SORTED, force_reload);
	spell_load_file(SPELLB_GLOBAL, force_reload);
	spell_load_file(SPELLB_PERDIR, force_reload);
	spell_load_file(SPELLB_PERFILE, force_reload);
	spell_load_file(SPELLB_PERMODE, force_reload);
	spell_load_file(SPELLB_PEREXTEN, force_reload);
	return 1;
}

// Return 1 if this word is spelled correctly.  Dictionary buffers
// should already be loaded; create empty ones if not.
int good_word_spelling(char *word)
{
	char listbuf[FNAMELEN];
	sprintf(listbuf, "%s%s", WORD_LIST_BUFFER, get_spell_language_prefix());
	return word_in_dictionary(create(listbuf), word)
			|| word_in_dictionary(spell_load_file(SPELLB_GLOBAL, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PERDIR, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PERFILE, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PERMODE, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PEREXTEN, 0), word);
}

// Get word that matched pattern we just searched for.  Return 0 if it's
// misspelled, nonzero otherwise.
int spell_get_and_check_word(int opt, char *word)
{
	int start = matchstart, end = matchend;

	save_var matchstart, matchend;
	if (matchend - matchstart > FNAMELEN - 1)
		return 1;
	grab(matchstart, matchend, word);
	if (spell_ignore_word(opt, matchstart, matchend, word))
		return 2;
	if (good_word_spelling(word))
		return 3;
	char *s = get_mode_string_variable("%s-spell-ignore-pattern-prefix");
	if (s && *s && matches_at(start, -1, s))
		return 4;
	s = get_mode_string_variable("%s-spell-ignore-pattern-suffix");
	if (s && *s && matches_at(end, 1, s))
		return 4;
	return 0;
}

// Add this word to the specified ignore list and refresh.
spell_ignore_this_word(int flavor, char *word, int doc)
{
	save_var bufnum = doc;
	char fname[FNAMELEN];
	int ignore = spell_load_file(flavor, 2);
	if (!ignore)
		error("No file associated with this buffer.");
	bufnum = ignore;
	point = size();
	bprintf("%s%s\n", size() && !parse_string(-1, "\n") ? "\n" : "", word);
	if (global_spell_options & 1)
		do_save_file(0, 0, 0);
	bufnum = doc;
	switch (flavor) {
		case SPELLB_PERDIR:
			strcpy(fname, filename);
			*get_tail(fname, 0) = 0;
			absolute(fname);
			say("\"%s\" added to ignore list for %f.", word, fname);
			break;
		case SPELLB_PERFILE:
			say("\"%s\" added to ignore list for %f.", word, filename);
			break;
		case SPELLB_PERMODE:
			say("\"%s\" added to ignore list for %s mode.", word, major_mode);
			break;
		case SPELLB_PEREXTEN:
			say("\"%s\" added to ignore list for %s extension.",
				  word, get_extension(filename));
			break;
		default:
			say("\"%s\" added to global ignore list.", word);
			break;
	}
	spell_add_to_helper_ignore_list(flavor, word);
	reload_spell_word_list();
}

maybe_delete_spell_ignore_list(char *file)
{
	char fname[FNAMELEN];

	if (!(global_spell_options & 2) || !*file)
		return;
	strcpy(fname, file);
	sprintf(get_tail(fname, 0), ".%s.espell", get_tail(file, 0));
	delete_file(fname);
}

spell_insert_runtogethers(char *word)
{
	char w1[FNAMELEN], w2[FNAMELEN], pat[FNAMELEN];
	int len = strlen(word), i, p1, p2;

	sprintf(pat, "%s%s", WORD_LIST_BUFFER, get_spell_language_prefix());
	int buf = create(pat);
	for (i = 2; i < len - 1; i++) {
		strncpy(w1, word, i);
		strcpy(w2, word + i);
		if (!word_in_dictionary(buf, w1) || !word_in_dictionary(buf, w2))
			continue;		// Fast binary search rejects most combinations.
		sprintf(pat, "^%s$", w1);
		point = 0;
		if (!re_search(1, pat))
			continue;
		p1 = give_begin_line();
		sprintf(pat, "^%s$", w2);
		point = 0;
		if (!re_search(1, pat))
			continue;
		p2 = give_begin_line();
		point = MAX(p1, p2);
		bprintf("%s %s\n", w1, w2);
	}
}

char *get_misspell_pattern(char *word, int exact);

int spell_find_suggestions(char *goal, int max, int case_repl)
{
	int poss = tmp_buf(), sugg = zap("-spell-suggestions");
	char sound[FNAMELEN], *wild, ord[FNAMELEN], pat[FNAMELEN];

	sprintf(ord, "%s%s", WORD_ORDER_BUFFER, get_spell_language_prefix());
	on_exit buf_delete(poss);
	if (*spell_helper_program) {
		save_var bufnum = sugg;
		spell_get_helper_suggestions(goal);
		return sugg;
	}
	int fast = get_num_var(vspell_get_mode_variable()) & 0x1000;
	if (fast || !exist(ord) || !buffer_size(ord))
		sprintf(ord, "%s%s", WORD_LIST_BUFFER, get_spell_language_prefix());
	save_var bufnum = poss;
	if (fast) {
		bprintf("%s\n", goal);
		phoneticize_lines(sugg, 5);	// Find the sound code for the goal word.
		get_line_from_buffer(sugg, 0, sound);	// Record it here.
		buf_zap(sugg);
		buf_zap(poss);
		if (!exist(SOUNDS_BUFFER)) {	// Make sound codes for
			load_spell_word_list(0);	// all dictionary words.
			bufname = ord;	
			phoneticize_lines(zap(SOUNDS_BUFFER), 5);
		}
		bufname = SOUNDS_BUFFER;
		// Copy words with the same sound to poss.  If there aren't many
		// and the sound is more than one character long, try copying
		// variations too.
		if (spell_copy_matching(sound, 0, ord, poss) < max
			  && strlen(sound) > 1)
			spell_copy_matching(sound, 1, ord, poss);
	} else {		// Make a copy of buffer into poss, insert runtogethers.
		grab_buffer(name_to_bufnum(ord));
		spell_insert_runtogethers(goal);
	}
	bufname = bufnum_to_name(poss);
	wild = get_misspell_pattern(goal, 0);
	point = 0;
	while (re_search(1, wild))
		buf_xfer(sugg, give_begin_line(), give_end_line() + 1);
	free(wild);
	bufnum = sugg;
	if (case_repl)
		spell_fix_suggestion_case(case_repl);
	save_var case_fold = 0;
	point = 0;
	sprintf(pat, "^%s$", goal);
	if (re_search(1, pat))
		bprintf(" (OK)");
	return sugg;
}

spell_ignore_word_this_run(char *word)
{
	say("Will resume checking \"%s\" when this command ends.", word);
	save_var bufnum = create(SPELL_THIS_RUN);
	case_fold = 0;
	point = size();
	bprintf("%s\n", word);
	buf_sort_and_uniq(bufnum);
}

do_spell_correct(int mult)
{
	char word[FNAMELEN];
	save_var point--;
	re_search(RE_FORWARD | RE_IGNORE_COLOR, get_spell_word_pattern());
	re_search(RE_REVERSE | RE_IGNORE_COLOR, get_spell_word_pattern());
	int wstart, wend, i = 1, hline = -1, have_file = *filename;
	grab(wstart = matchend, wend = matchstart, word);
	if (!*word)
		return 2;
	int case_repl = (get_num_var(vspell_get_mode_variable()) & 0x2000)
					? 0 : check_for_case_replace(wstart, wend);
	build_window();
	int wcol = window_left + column_in_window;
	int wline = window_top + line_in_window, obuf = bufnum;
	int sugg = spell_find_suggestions(word, 9, case_repl);
	save_var bufnum = sugg;
	remove_line_highlight();
	go_line(10);
	delete(point, size());
	int wcnt = lines_between(0, size());
	point = 0;
	if (point < size()) {
		do {
			bprintf("%d ", i++);
		} while (nl_forward() && point < size());
		stuff("\n");
	} else
		stuff("No suggestions\n\n");
	stuff("O Ignore word once\n");
	if (mult)
		stuff("R Ignore on this run only\n");
	if (have_file) {
		stuff("F Ignore in this file\n");
		stuff("D Ignore in this directory\n");
	}
	stuff("M Ignore for this mode only\n");
	int by_exten = have_file && (global_spell_options & 0x10);
	if (by_exten)
		stuff("E Ignore for this extension only\n");
	stuff("G Ignore globally\n");
	if (mult)
		stuff("Ctrl-R for recursive edit\n");
	point = 0;
	int cnt = lines_between(0, size());
	restore_vars();
	highlight_on();
	save_spot mark = wstart;
	save_spot point = wend;
	on_exit highlight_off();
	save_var want_gui_prompts = 0;
	save_var window_handle;
	int row = 5;
	if (wline + cnt + 4 < screen_lines)
		row = wline + 2;			// Position below text.
	else if (wline > cnt + 2)
		row = wline - cnt - 3;		// Or above.
	else if (row + cnt >= screen_lines - 1)
		if ((row = screen_lines - cnt - 3) + cnt <= wline)
			row = 0;
	int col = wcol - 3;				// Align with text.
	if (col + 35 > screen_cols)		// Too far to right, so move back.
		col = screen_cols - 35;
	if (col < 0)					// Too far to left, move right.
		col = 0;							// If not visible above or below, 
	if (row <= wline && wline < row + cnt)	// move left or right
		if (wcol + 35 > screen_cols)
			col = wcol - 35;
		else
			col = wcol + strlen(word) + 3;
	int win = add_popup(col, row, 30, cnt + 2, _view_border, sugg);
	on_exit remove_window(win);
	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	window_title(win, BTOP, TITLECENTER,  " Suggestions ");

	display_scroll_bar = 0;
	for (;;) {
		char keys[20];
		sprintf(keys, "O%s%sM%sG", mult ? "R" : "", have_file ? "FD" : "",
				by_exten ? "E" : "");
		if (wcnt > 1)
			noteput("1-%d to replace, ignore \"%s\" Once, or add "
					"to an ignore list [1-%d%s] ", wcnt, word, wcnt, keys);
		else if (wcnt == 1)
			noteput("1 to replace, ignore \"%s\" Once, or add "
					"to an ignore list [1%s] ", word, keys);
		else
			noteput("No suggestions.  Ignore \"%s\" Once, or add "
					"to an ignore list [%s] ", word, keys);
		refresh();
		getkey();
		say("");
		check_abort();
		if (binding_of(key, reg_tab) == down_line)
			hline = spell_correct_move_highlight(hline, win, 1, 0);
		else if (binding_of(key, reg_tab) == up_line)
			hline = spell_correct_move_highlight(hline, win, -1, 0);
		else if (IS_ENTER_KEY(key) && hline != -1)
			hline = spell_correct_move_highlight(hline, win, 0, 1);
		if (IS_MOUSE_LEFT(key)
			  && window_at_coords(mouse_y, mouse_x, mouse_screen) == win
			  && (i = mouse_y - window_text_top(win)) >= 0 && i < cnt)
			hline = spell_correct_move_highlight(-1, win, i + 1,
				IS_MOUSE_DOUBLE(key));
		if (tolower(key) == 'g')
			spell_ignore_this_word(SPELLB_GLOBAL, word, obuf);
		else if (tolower(key) == 'r' && mult) {
			spell_ignore_word_this_run(word);
			break;
		} else if (tolower(key) == 'd' && have_file) {
			spell_ignore_this_word(SPELLB_PERDIR, word, obuf);
			break;
		} else if (tolower(key) == 'f' && have_file) {
			spell_ignore_this_word(SPELLB_PERFILE, word, obuf);
			break;
		} else if (tolower(key) == 'm')
			spell_ignore_this_word(SPELLB_PERMODE, word, obuf);
		else if (tolower(key) == 'e' && by_exten)
			spell_ignore_this_word(SPELLB_PEREXTEN, word, obuf);
		if (key == CTRL('R') && mult)
			return 1;
		if (IS_ASCII_KEY(key) && strchr("sogm", tolower(key))
			  || key == abort_key || key == ESC)
			break;
		if (isdigit(key) && key != '0' && key - '0' <= wcnt) {
			restore_vars();
			spell_get_suggestion_line(sugg, key, word);
			point = wstart;
			delete(wstart, wend);
			stuff(word);
			return 0;
		} else if (IS_ALT_KEY(key) && isdigit(i = KEY_PLAIN(key))
				   && i != '0' && i - '0' <= wcnt) {
			spell_get_suggestion_line(sugg, i, word);
			copy_line_to_clipboard(word, 0);
		}
	}
	this_cmd = SPELL_CORRECT;
	return 0;
}

// Called during buffer display.  Mark up misspelled words in the
// current window using a region.
vspell_buffer_when_displaying()
{
	int from, to, cnt = 0;
	if (!modified_buffer_region(&from, &to, "vspell")
		  && vspell_data.window_start == window_start
		  && vspell_data.window_end == window_end
		  && vspell_data.display_column == display_column
		  && !vspell_data.skipped_word)
		return;
	vspell_data.window_start = window_start;
	vspell_data.window_end = window_end;
	reset_modified_buffer_region("vspell");
	if (modify_region(SPELL_REGION_HANDLE, MRTYPE, -1) == REGNORM)
		remove_region(SPELL_REGION_HANDLE);
	if (!vspell_data.want)
		return;
	int opos = point;		// Ignore the word just before point when typing.
	char *pat = get_spell_word_pattern();
	save_var narrow_start = window_start
							- matches_at_length(window_start, -1, pat);
	save_var narrow_end = size() - window_end
						  - matches_at_length(window_end, 1, pat);
	save_var point = 0, abort_searching = ABORT_ERROR;
	int opt = get_num_var(vspell_get_mode_variable());
	vspell_data.skipped_word = 0;
	while (re_search(1, pat) == 1) {
		char word[FNAMELEN];
		if (spell_get_and_check_word(opt, word))
			continue;
		int skey = key;
		skey = (skey == GREYBACK) ? '\b' : IS_ASCII_KEY(skey) ? skey : '/';
		if (is_current_window && prev_cmd != SPELL_CORRECT && matchstart <= opos
			  && opos <= matchend + (strchr("'", skey) != 0)
			  && (isalnum(skey) || (strchr("_'", skey)) || skey == '\b')) {
			vspell_data.skipped_word = 1;	// Catch it next time.
			continue;
		}
		if (cnt++ > 2000)	// More than this many misspellings in the window?
			break;			// Must be something wrong; bail out.
		spot start = alloc_spot(0), end = alloc_spot(1);
		*start = point - strlen(word);
		*end = point;
		add_region(start, end, color_class spell_misspelled, REGNORM,
				   SPELL_REGION_HANDLE);
		if (display_column >= 0)	// Don't check words we can't see.
			if (current_column() - strlen(word) > display_column + window_width)
				nl_forward();
			else if (current_column() < display_column - 50)
				move_to_column(display_column - 50);
	}
	modify_region(SPELL_REGION_HANDLE, MRAUTODEL, 1);
}

set_spell_mode(int per_buf)
{
	int old = get_num_var(vspell_get_mode_variable()), want;

	if (vspell_data.buf_want)
		old = (vspell_data.buf_want == 1);
	want = (old & ~1) | (has_arg ? (iter != 0) : (~old & 1));
	if ((want & 1) && !spell_load_file(SPELLB_ORDERED, 0)) {
		if (*get_spell_language_prefix()) {
			say("No dictionary found for \"%s\" language, install one.",
				  get_spell_language_prefix());
			return;
		}
		say("No dictionary found, running spell-configure.");
		spell_configure();
		return;
	}
	if ((want & 1) && !spell_mode_ok(bufnum))
		error("Spell mode not available for this system buffer.");
	if (per_buf)
		vspell_data.buf_want = want ? 1 : 2;
	else {
		set_num_var(vspell_get_mode_variable(), want);
		vspell_data.buf_want = 0;
	}
	do_when_make_mode_vspell();
	save_customization(vspell_get_mode_variable());
	if ((want & 1) != vspell_data.want)
		say("Spell mode not available for this type of buffer.");
	else if (per_buf)
		say("Turning Spell minor mode %s for this buffer.",
			  vspell_data.want ? "on" : "off");
	else
		say("Setting %s to turn Spell minor mode %s.",
			  name_name(vspell_get_mode_variable()),
			  vspell_data.want ? "on" : "off");
	make_mode();
}

command spell_buffer_or_region()
{
	if (!load_spell_word_list(0))
		return;
	save_spot point, mark;
	spot end = alloc_spot(1);
	on_exit if (spot_to_buffer(end) != -1) free_spot(end);
	on_exit highlight_off();
	if (is_highlight_on()) {
		fix_region();
		*end = mark;
	} else {
		point = 0;
		*end = size();
	}
	on_exit delete_buffer(SPELL_THIS_RUN);
	int opt = get_num_var(vspell_get_mode_variable());
	char *pat = get_spell_word_pattern();
	while (point < *end && re_search(1, pat) == 1) {
		char word[FNAMELEN];
		mark = matchstart;
		if (spell_get_and_check_word(opt, word))
			continue;
		int this_run = name_to_bufnum(SPELL_THIS_RUN);
		if (this_run && word_in_dictionary(this_run, word))
			continue;
		highlight_on();
		if (do_spell_correct(1) == 1) {
			highlight_on();
			show_binding("Continue spell checking by typing %s", "exit-level");
			if (!recursive_edit_preserve())
				break;
		}
	}
	say("Spell checking completed.");
}

move_by_pages(dir)	// page up or down by one page
{
	int col, wide = text_width() - 1, n;

	if (prev_cmd != SAMECOL)
		_preserve_column = virtual_space ? virtual_column()
					: current_column();
	if (window_overlap < 0)
		n = ((100 - (-window_overlap)) * text_height()) / 100;
	else
		n = text_height() - window_overlap;
	if (n < 1)
		n = 1;
	n *= dir;		// move by n lines
	window_start = move_by_screen_lines(window_start, n);
	if (paging_centers_window)	// go halfway down window
		point = move_by_screen_lines(window_start, text_height() / 2);
	else				// go one screenful from orig place
		point = move_by_screen_lines(point, n);
	if (is_highlight_on())
		save_var goal_column = -1;
	col = (goal_column >= 0) ? goal_column : _preserve_column;
	if (virtual_space)
		to_virtual_column(col);
	else {
		if (display_column < 0)	/* move to same screen column */
			col = (col % wide) + (current_column() / wide) * wide;
		move_to_column(col);
	}
}

set_window_wrap(int wrap)
{
	if (wrap == WRAP_ALWAYS)
		display_column = -1;
	else if (wrap == WRAP_NEVER)
		if (display_column < 0)
			display_column = 0;
}

int window_wrap_width()
{
	int res = text_width();
	if (display_column < 0)
		res--;
	return res;
}

command to_right_edge()
{
	to_virtual_column(MAX(display_column, 0) + window_wrap_width() - 1);
}

command split_window_vertically()
{
	iter = 0;
	tiled_only();
	set_window_wrap(wrap_split_vertically);	/* turn on horiz scrolling in both */
	window_split(VERTICAL);
	set_window_wrap(wrap_split_vertically);
}

when_loading()
{
	print_heading |= PRHEAD_PATH;
}
