/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2007 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.01 to 13.02.

#include "eel.h"
#include "proc.h"
#include "c.h"
#include "colcode.h"
#include "epsnet.h"
#include "hex.h"
#include "html.h"
#include "python.h"
#include "info.h"
#include "perl.h"
#include "kill.h"
#include "session.h"

charname_lister(char *s, char *(*finder)(), int flags, int mbuf)
{
	int i, len = strlen(s);
	char buf[FNAMELEN];

	UNREFERENCED_PARAMETER(flags);
	UNREFERENCED_PARAMETER(finder);
	if (strchr(s, '>'))
		len = strchr(s, '>') - s + 1;
	buf_zap(mbuf);
	save_var case_fold = 1;
	if (!*s || *s == '<') {
		note("Listing matches...");
		for (i = 0; i < MAX_CHAR; i++) {
			*buf = 0;
			show_char(buf, i, 3);
			if (buf[0] == '<' && buf[1] && buf[1] != '#'
				  && !strnfcmp(s, buf, len))
				buf_printf(mbuf, "%s  U+%04x (%c)\n", buf, i, i);
		}
		note("");
	}
	point = 0;
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */
	int in_initializer;

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!(reindent_c_comments & 1) && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if ((!reindent_one_line_c_comments || 
		 (indenter_action == IACT_REINDENT_PREV) && !(reindent_c_comments & 8))
		  && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' && reindent_c_preprocessor_lines == 1
		  || first_char == '\f'
		  || first_char == '@' && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(1, "@(" OBJC_TOPBLOCKS ")[ \t]+<word>|@end"))
		return 0;
	else if (first_char == '}')
		return cind_close();
	else if (have_c_close_equivalent(orig))
		return indent_at_c_close_equivalent(orig);

	to_begin_line();
	if (parse_string(-1, PREPROC_ELSE ".*\n")) {
		point = matchend;
		locate_by_ifdef(-1);
	}
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig) || c_align_break_with_case
		  && matches_at(orig, 1, "[ \t]*break;")) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig) && !((c_extra_keywords & OBJC_KEYWORDS)
									 && matches_at(pstart, 1, "^[-+@]"))) {
		if (c_in_struct())		// Could it be a bitfield, not a label?
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';' || c_semi_equivalent(pend, orig)) {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (c_close_equivalent(pend)) {
		point = pend;
		ind = indent_after_c_close_equivalent(orig);
	} else if (prev_end == '{' || c_open_equivalent(pend, orig)) {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_initializer = in_c_initializer(orig)) {
		return ind;
	} else if (c_next_no_indent(orig, pstart)) {
		return 0;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, in_initializer);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

command do_c_indent()
{
	int orig = point, ind = get_c_indent();
	int orig_column = current_column();

	if (maybe_indent_rigidly(0))
		return;
	to_indentation();
	if (!c_tab_always_indents && orig_column > current_column() || in_shell_buffer) {
					/* if not in indentation */
		point = orig;
		int soft = get_soft_tab_size();
		if (this_cmd != CMD_INDENT_REG)
			to_column(orig_column + soft - orig_column % soft);
	} else if (prev_cmd == C_INDENT)	/* repeated, make bigger */
		to_column(orig_column + ind);
	else if ((c_extra_keywords & IDL_KEYWORDS))
		prev_indenter();
	else {
		c_indenter();
		if (in_perl_buffer && !get_indentation(point) && !orig_column
			  && this_cmd != CMD_INDENT_REG)
			// Nothing to indent from, but indent anyway.
			if (sublanguage_code != SUBLANG_PHP)
				to_column(ind);
			else
				to_column(php_top_level_indent);
	}
	if (this_cmd != CMD_INDENT_REG)
		this_cmd = C_INDENT;
}

// Force this line to the left edge?
int c_next_no_indent(int orig, int pstart)
{
	if (c_extra_keywords & OBJC_KEYWORDS)
		return objc_next_no_indent(orig, pstart);
	if ((c_extra_keywords & (JAVA_KEYWORDS)) && matches_at(pstart, 1, "@[A-Z0-9]"))
		return 1;	// Java annotation on prev line of this definition.
	return 0;
}

when_restoring()
{
	int tflag = 0, i, splitmode = HORIZONTAL, have_files = 0;
	char *flag, *arg, init_dir[FNAMELEN], *dir = init_dir;

	getcd(init_dir);
	early_restoring();
	early_init();
	set_toolbar();

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			have_files = 1;
			continue;
		}
		if (!strcmp(argv[i], "--"))
			break;
		if (!stricmp(argv[i], "-noinit")) {
			noinit_flag = 1;
			continue;
		} else if (!strcmp(argv[i], "-dir")) {
			if (++i < argc)
				dir = argv[i];
			continue;
		}
		flag = argv[i];
		switch (flag[1]) {
		case 'f':
			arg = (flag[2] && flag[3] || i + 1 >= argc) ? (flag + 3) : argv[++i];
			switch (flag[2]) {
			case 's':
				set_swapname(arg);
				break;
			case 'd':
				strcpy(_docname, arg);
				break;
			default:
				bad_flag(flag);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'd':
		case 'r':
		case 'l': if (!flag[2] && i + 1 < argc)
					  i++;
				  break;
		case 'p': break;
		case 'h': case '?': usage(); break;
		case 'v':
			switch (flag[2]) {
				case 'v':
					splitmode = VERTICAL;
					break;
				default:
					bad_flag(flag);
			}
			break;
		case '-':
			if (!strcmp(flag + 2, "help"))
				usage();
			else
				bad_flag(flag);
			break;
		default:
			bad_flag(flag);
		}			/* i is 1st file argument */
	}
	if (i < argc)
		have_files = 1;
	term_init();
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	delete_buffer("-customizations");	// Don't record any settings before now.
	parse_command_line_flags(1);
	if (!noinit_flag && load_from_state)
		init_from_file();
	parse_command_line_flags(2);
	if (tflag) {
		preserve_session_once = 1;
		tutorial();
	} else {
		if (load_from_state && (!have_files || session_always_restore)) {
			if (session_restore_files == 2 && have_files)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (have_files)
			load_cmd_line_files(1, splitmode, dir);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	restore_check_version();
	show_text(0, 1, "Epsilon Programmer's Editor %s, EEL version %s",
			  version, eel_version);
	try_calling("start-up");
	parse_command_line_flags(3);
	if (!noinit_flag && load_from_state)
		load_version_patch_file();
}

// See if there's an EEL patch file for this version available, and load it.
load_version_patch_file()
{
	char file[FNAMELEN], ver[FNAMELEN], *s;
	strcpy(ver, eel_version);
	while (s = strchr(ver, '.'))
		strcpy(s, s + 1);
	sprintf(file, "v%spatch.e", ver);
	show_text(-2, 0, "Looking for patch file %s on EPSPATH", file);
	load_eel_from_path(file, 1);
}

color_c_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	char pat[200];
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
		       "|-?%.?[0-9]([A-Za-z0-9._]|[Ee]-)*");
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
		case '/':			// found // one-line comment
			while (nl_forward() && character(point - 2) == '\\')
				;		// Continued // comment.
			set_character_color(t, point, color_class c_comment);
			break;
		case '*':			// found /* starting comment
			search(1, "*/");
			set_character_color(t, point, color_class c_comment);
			if (get_character_color(point, (int *) 0, &s)
				  == color_class c_comment)
				to = MAX(s, to);	// Comment extended further before; rescan it.
			break;
		case '#':			// found preproc line
			c_preproc_color();	// move past correct part
			set_character_color(t, point,
					    color_class c_preprocessor);
			break;
		case '"':		// found a string literal
			point = t;
			if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
				t--;
			if ((c_extra_keywords & CSHARP_KEYWORDS)
				  && character(t - 1) == '@')
				t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
			else
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
			set_character_color(t, point, color_class c_string);
			if (get_character_color(point, (int *) 0, &s) == 
			    color_class c_string && s > to)  // fix up after
				c_init_color(point, to = s); // quoted "'s
			break;
		case '\'':		// found a char const
			point = t;
			re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
			set_character_color(t, point, color_class c_charconst);
			break;
		default:		// found identifier, kywd, or number
			if ((c_extra_keywords & (OBJC_KEYWORDS | JAVA_KEYWORDS))
				  && character(t - 1) == '@')
				t--;
			set_character_color(t, point, c_keyword_color(t));
			if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
				&& (point - t == 4)			// uuid's in IDL files.
				&& parse_string(-1, "uuid")
				&& parse_string(1, "%([0-9A-Fa-f]"))
				maybe_color_uuid();
			break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

zeroed TIMER idle_display_time;
int auto_show_delimiter_delay;	// time before next check (.01 sec units)
#define MAX_BUFFER_WHEN_IDLE	5
buffer int (*buffer_when_idle[MAX_BUFFER_WHEN_IDLE])();

// Called when there's nothing to do, while waiting for key.
when_idle(cnt)
{
	int delay, disp, i, j;

	if (leave_blank)		// Exiting or similar--don't touch screen.
		return -1;
	if (!cnt || time_done(&idle_display_time)) {
		if (unseen_msgs_time() <= 0)
			show_idle_text();
		time_begin(&idle_display_time, 10);	// wait .1 seconds next time
	}
	if ((auto_show_matching_characters && *auto_show_matching_characters
		 || mode_auto_show_delimiter) && cnt == 1) {
		auto_show_delimiter_delay = 5;
		if (maybe_auto_show_delimiter() == 2)	// No match, wait .5 sec
			auto_show_delimiter_delay = 50;	 // before looking on next key.
		refresh();
	}
	if (is_unix == IS_UNIX_XWIN && is_key_repeating())
		return -1;
	if (cnt == 2 && set_display_func_name())
		maybe_refresh();
	delay = when_idle_coloring(cnt);
	if (*show_when_idle) {		// Wait for .1 seconds.
		disp = time_remaining(&idle_display_time);
		if (delay < 0 || disp < delay)
			delay = disp;
	}
	if (cnt < 2 && (delay < 0 || auto_show_delimiter_delay < delay))
		delay = auto_show_delimiter_delay;
	for (i = 0; i < MAX_BUFFER_WHEN_IDLE; i++)
		if (buffer_when_idle[i]) {
			j = (*buffer_when_idle[i])(cnt);
			if (delay < 0 || j >= 0 && j < delay)
				delay = j;
		}
	if (!_doing_input
		  && (text_color == color_class echo_area || system_window)) {
		// Perhaps this should be in when_ function called after eel error.
		say("Recovering from previous error...");
		for (i = 0; i < 5; i++) {
			window_number = 0;
			window_number--;		// to topmost popup
			if (text_color == color_class echo_area || system_window)
				window_kill();
		}
		in_echo_area = 0;
	}
	i = name_match("do_when_idle_", 1);
	for (; i; i = name_match("do_when_idle_", 0))
		if (name_type(i) == NT_SUBR) {
			j = (*(int (*)())i)(cnt);
			if (delay < 0 || j >= 0 && j < delay)
				delay = j;
		}
	return (cnt < 2) ? 0 : delay;
}

command edit_variables()
{		// list variables for editing; with arg, includes system vars
	int listbuf, orig_buf, orig, part = 1;
	int listwin, helpwin, orig_win, win, listheight = 9;
	int setit, origpos, ind, type, hasdef, i, lasthelp;
	TIMER ts;	// for searching in list
	char var[100], partial[100], *val;
	char tmp[FNAMELEN * 2], curval[FNAMELEN * 2];
	jmp_buf this_level;

	tiled_only();
	if ((has_feature & FEAT_WINHELP)
		  && get_epsilon_help_format() != HELP_FORMAT_POPUP) {
		if (has_arg)
			win_help_string("", "all variables, list of");
		else
			win_help_string("", "preference variables, list of");
		return;
	}
	if (screen_lines > 20)
		listheight = screen_lines * 40 / 100;
	orig_buf = bufnum;
	orig_win = window_handle;
	noteput("Listing variables...");
	save_var abort_searching = 0;
	if (has_arg)
		save_var show_all_variables = 1;
	iter = has_arg = lasthelp = 0;
	listbuf = tmp_buf();	// put list of variables here
	bufnum = listbuf;
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	save_var _nt_filter = BIT(NT_VAR) | BIT(NT_BUFVAR) | BIT(NT_WINVAR)
						  | BIT(NT_BUILTVAR);
	list_matches("", var_match, 0, listbuf);	// construct list of vars
	point = 0;
	listwin = add_popup(_view_left, 1, 30, listheight,
						ALL_BORD(BSINGLE), listbuf);
	set_menu_look(listwin);
	window_title(listwin, BTOP, TITLECENTER, " Variables ");
	mode_keys = menu_tab;
	_window_flags |= HIGHLIGHT_LINE;

	get_doc(0);
	helpwin = add_popup(_view_left, listheight + 2,	// put help text here
						MIN(screen_cols - _view_left - _view_right, 80),
						11, _view_border, zap(HELPBUF));

	text_color = (is_gui && want_gui_prompts) ? color_class text : color_class viewed_text;
	display_column = 0;		/* prevent continuation lines */

	restore_vars();		// don't let save_var undo user's changes
	time_begin(&ts, 0);
	while (is_window(listwin) && is_window(helpwin)) {
		setit = 0;		// assume we don't want to set variable
		window_handle = listwin;
		origpos = point;
		to_begin_line();
		grab_full_line(bufnum, var);	// get name of current variable
		ind = find_index(var);
		switch (name_type(ind)) {	// has a default value too?
			case NT_BUFVAR:	hasdef = 1; break;
			case NT_WINVAR: hasdef = 2; break;
			default:	hasdef = 0; break;
		}
		type = vartype(ind);

		if (ind != lasthelp) {	// show info on variable
			window_handle = orig_win; // if buffer/window-specific,
			bufnum = orig_buf;	// get value from before this
			val = get_var_val(ind, curval, 0); // command

			window_handle = helpwin;
			make_var_help(var, orig_buf, orig_win, 1);
			window_to_fit(screen_lines - listheight - 6);
			sprintf(tmp, " %s ", var);
			window_title(helpwin, BTOP, TITLECENTER, tmp);
			display_scroll_bar = !window_extra_lines();
			lasthelp = ind;
		}
		window_handle = helpwin;
		build_window();
		window_title(helpwin, BBOTTOM, TITLERIGHT(2), 
					 all_blanks(window_end, size() - narrow_end)
					 ? "" : " Press <Space> for more ");
		window_handle = listwin;
//		do {			// prompt for a key
		maybe_refresh();
		note("Select a variable & press <Enter> to set, "
			 "<Esc> when done, ? for help.");
		i = permit_window_keys;
		permit_window_keys = PERMIT_SCROLL_KEY;
		getkey();
		permit_window_keys = i;
//		} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		if (key == ESC || user_abort || key == abort_key)
			break;
		else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (is_window(win) != ISPOPUP)
				break;
			window_handle = win;
			get_text_area(win);
			if (in_text_area(mouse_x, mouse_y)) {
				point_to_pos(mouse_x, mouse_y, mouse_screen);
				if (win == listwin && IS_MOUSE_DOUBLE(key))
					setit = 1;
				else if (win == helpwin)
					key = KEYPGDN;
			} else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
		} else if (is_search_dir()) {
			save_var top_level = &this_level;
			if (!setjmp(top_level))	// come here on abort
				do_again();	// let search cmds run normally
			restore_vars();
			continue;
		}
		if (!isdigit(key))	// do arrow keys, C-n, etc.
			switch (get_direction()) {
				case BTOP:
					if (!nl_reverse()) point = size();
					break;
				case BBOTTOM:
					if (!nl_forward()) point = 0;
					break;
			}
		switch (key) {
			case KEYHOME:
			case NUMDIGIT(7): point = 0; break;
			case KEYEND:
			case NUMDIGIT(1): point = size(); break;
			case KEYPGDN:
			case NUMDIGIT(3): next_page(); break;
			case KEYPGUP:
			case NUMDIGIT(9): previous_page(); break;
			case '\n':
			case '\r':
				setit = 1; break;
			case ' ':
				window_handle = helpwin;
				next_page();
				break;
			case '\b':
			case GREYBACK:
				if (!time_done(&ts)) { // backsp in srch
					if (part > 1)
						part--;
					goto search_for;
				}
				window_handle = helpwin; // or scroll help
				previous_page();
				break;
			default:
				if (binding_of(key, root_keys) == (short) help
					  || key == '?') {
					help_on_command("edit-variables");
					lasthelp = 0;
				} else if (key >= ' ' && key <= '~') {
					if (time_done(&ts))
						part = 1; // new search
					partial[part++] = key;
search_for:				orig = point, point = 0;
						partial[0] = '^';
						partial[part] = 0;
						save_var see_delay = new_search_delay;
						say("Searching for %s", partial+1);
						restore_vars();
						if (do_searching(FOLD|REGEX, partial)
							  != 1)
							point = orig;
						time_begin(&ts, new_search_delay);
				} else if (IS_WIN_KEY(key) || key == '\t' || key == BACKTAB) {
					run_topkey();
				} else if (IS_TRUE_KEY(key)
						   && !setit && point == origpos) {
					maybe_ding(bell_on_bad_key);
				}
				break;
		}
		if (setit) {		// user wants to set a variable
			refresh();
			part = 1;	// reset searcher
			ungot_key = CTRL('S');	// bring in default value
			save_var top_level = &this_level;
			i = 0;
			if (setjmp(top_level))	// come here on abort
				i = 1;		// and record it
			else if (is_gui && want_gui_prompts)
				get_strpopup(tmp, var, val, var);
			else
				get_strnone(tmp, "New value: ", val);
			restore_vars();
			if (i)		// aborted
				continue;
			if (type <= TYPE_INT) {
				numtoi(tmp);
				if (got_bad_number)
					continue;
			}
			i = 0;		// assume current val only
			if (is_gui && want_gui_prompts && hasdef)
				i = 1;			// assume set both, for now
			else if (hasdef) {
				sayput("Set D=default value of %s, C=current"
					   " value, or B=both? [B]", var);
				getkey();
				if (user_abort)
					continue;
				if (toupper(key) == 'D')
					i = -1;
				else if (toupper(key) != 'C')
					i = 1;	// set current & default
			}
			window_handle = orig_win;
			bufnum = orig_buf;
			if (i >= 0)
				set_var_val(ind, tmp, 0);
			if (i)
				set_var_val(ind, tmp, 1);
			full_redraw = 1; // might have set a display-type var
			lasthelp = 0;	// refresh help win
		}
	}
	say("");
	if (is_window(orig_win))
		window_handle = orig_win;
	remove_window(listwin);
	remove_window(helpwin);
	buf_delete(listbuf);
	delete_buffer(HELPBUF);
	user_abort = 0;
}

is_remote_buffer(int buf)
{
	save_var bufnum = buf;
	return (mode_keys == ssh_tab) || telnet_id != 0
			|| *filename && is_remote_file(filename);
}

telnet_mode()
{
	mode_default_settings();
	mode_keys = telnet_tab;
	fix_key_table(reg_tab, find_index("cc-tab"),
				  telnet_tab, find_index("telnet-ctrlc-tab"));
	telnet_tab[ALT('p')] = find_index("process-previous-cmd");
	telnet_tab[ALT('n')] = find_index("process-next-cmd");
	major_mode = _telnet_mode_name;
	tab_size = 8;
	try_calling("telnet-mode-hook");
	make_mode();
}

when_loading()
{
	copy_key_table(process_tab, ssh_tab);
	ssh_tab['\t'] = 0;
	ssh_tab[NUMSHIFT(GREYTAB)] = 0;
	fix_key_table(reg_tab, find_index("cc-tab"),
				  ssh_tab, find_index("telnet-ctrlc-tab"));
}

show_date_warning(temp, orig, then, now, save)
struct file_info *now, *then;
{
	char msg[FNAMELEN], *file;
	int mod, warn, auto;

	save_var bufnum = orig;
	file = filename;
	mod = modified;
	warn = want_warn;
	auto = auto_read_changed_file;
	bufnum = temp;
	buf_zap(temp);
	bprintf("%s\nhas changed since it was read from disk", file);
	bprintf(mod ? "!\n" : ".\nFortunately, you haven't edited it.\n");

	stuff_file_date("\nThe version being edited was dated %s on disk.\n",
					then);
	stuff_file_date("         The copy on disk is dated %s.\n\n", now);
	current_time_and_date(msg);
	bprintf("                         It is now %s.\n\n", msg);

	bprintf("Options:\n");
	bprintf("D) Display a diff listing comparing the two versions.\n");
	bprintf("C) Compare the two versions by reading the disk version ");
	bprintf("in its own buffer.\n");
	bprintf("   You may then use compare-windows or diff to ");
	bprintf("see the changes.\n");
	bprintf("R) Read in the new version now, replacing the ");
	bprintf("version already read in.\n");
	if (save)
		bprintf("S) Save your changes, replacing ");
	else
		bprintf("I) Ignore ");
	bprintf("the version on disk.\n\n");
	bprintf("!) %s this warning for this buffer.  ",
			warn ? "Disable" : "Re-enable");
	bprintf("(You may then choose\n");
	bprintf("   another option.)\n");

	bprintf("A) %s auto-read.  Read changed files without prompting, "
			"if it's safe.\n", auto ? "Disable" : "Enable");
	bprintf("   (You may then choose another option.)\n");

	bprintf("\nYour choice? ");
	point = 0;
}

stuff_file_date(fmt, p)	/* bprintf the date in p using fmt's %s */
char *fmt;
struct file_info *p;
{
	char msg[FNAMELEN];

	format_date(msg, p->year, p->month, p->day,
				p->hour, p->minute, p->second);
	bprintf(fmt, msg);
}

format_date(msg, year, month, day, hour, minute, second)
char *msg;
{
	char ampm = 'a', args[24][30], *vals[24], *s, *t;
	char *months = "January February March April May June July August "
				   "September October November December ";
	char *weekdays = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday ";
	int i, dow;

	sprintf(args[0], "%d", hour);
	sprintf(args[1], "%02d", hour);
	if (hour >= 12)
		ampm = 'p', hour -= 12;
	sprintf(args[2], "%d", hour ? hour : 12);
	sprintf(args[3], "%02d", hour ? hour : 12);
	sprintf(args[4], "%d", minute);
	sprintf(args[5], "%02d", minute);
	sprintf(args[6], "%d", second);
	sprintf(args[7], "%02d", second);
	sprintf(args[8], "%c", ampm);
	sprintf(args[9], "%c", toupper(ampm));
	sprintf(args[10], "%d", day);
	sprintf(args[11], "%02d", day);
	sprintf(args[12], "%d", month);
	sprintf(args[13], "%02d", month);
	sprintf(args[14], "%d", year);
	sprintf(args[15], "%02d", year % 100);
	for (i = month - 1, s = months; i > 0 && s; i--) {
		s = strchr(s, ' ');
		if (s)
			s++;
	}
	if (month >= 1 && month <= 12 && s && (t = strchr(s, ' ')))
		sprintf(args[16], "%.*s", t - s, s);
	else
		sprintf(args[16], "???");
	strncpy(args[18], args[16], 3);
	for (i = 0; i < strlen(args[16]) + 1; i++)
		args[17][i] = toupper(args[16][i]);
	for (i = 0; i < strlen(args[18]) + 1; i++)
		args[19][i] = toupper(args[18][i]);
	dow = day_of_week(day, month, year);
	for (i = dow, s = weekdays; i > 0 && s; i--) {
		s = strchr(s, ' ');
		if (s)
			s++;
	}
	if (dow >= 0 && dow <= 6 && s && (t = strchr(s, ' ')))
		sprintf(args[20], "%.*s", t - s, s);
	else
		sprintf(args[20], "???");
	strncpy(args[22], args[20], 3);
	for (i = 0; i < strlen(args[20]) + 1; i++)
		args[21][i] = toupper(args[20][i]);
	for (i = 0; i < strlen(args[22]) + 1; i++)
		args[23][i] = toupper(args[22][i]);
	for (i = 0; i < 24; i++)
		vals[i] = args[i];
	if (expand_string_template(msg, date_format, "iIhHnNeEaAdDmMyYfFtTwWsS", vals))
		error("Invalid date-format variable.");
}

general_help()
{
	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_contents("");		// Show Epsilon's contents topic.
				return;
			case HELP_FORMAT_HTML:
				if (!try_html_help_string(""))
					return;
		}
	prompt_for_help();
}

int get_epsilon_help_format()
{
	char varname[FNAMELEN];
	sprintf(varname, "epsilon-help-format-%s-%s",
			is_unix ? "unix" : is_win32 ? "win" : "other",
			(is_gui || is_unix == IS_UNIX_XWIN) ? "gui" : "console");
	int code = get_num_var(find_index(varname));
	if (code == HELP_FORMAT_WINHELP && !(has_feature & FEAT_WINHELP))
		code = 0;		// Ignore setting, it's only available in Windows.
	if (!code)	// Pick a default format.
		if (has_feature & FEAT_WINHELP_NATIVE)	// WinHelp is always installed.
			code = HELP_FORMAT_WINHELP;
		else if (is_win32)
			code = HELP_FORMAT_HTML;	// Try using a browser.
		else			// Fall back to internal help system.
			code = HELP_FORMAT_POPUP;
	return code;
}

command epsilon_manual()
{
	if (!(has_feature & FEAT_WINHELP_NATIVE))
		epsilon_manual_html();
	else
		win_help_string("", "Epsilon Manual Contents");
}

// Open this URL in a browser and return 1.  Return 0 if we can't.
int try_show_url(char *url)
{
	char cmd[300], prog[300];

	if (is_win32) {
		get_executable_directory(prog);
		strcat(prog, "lhelp.exe");
		sprintf(cmd, "lhelp -u \"%s\"", url);
		winexec(prog, cmd, SW_SHOWMINNOACTIVE, 0);
		return 1;
	} else if (is_unix == IS_UNIX_XWIN) {
		if (run_helper("goto_url", url, ""))
			return 1;
	}
	return 0;
}

// Cause this URL to be displayed by a browser.
show_url(char *url)
{
	if (!try_show_url(url))
		error("Can't start a browser, please access %s manually.", url);
}

// Display topic in a browser and return 1.  Return 0 if running a web
// browser is unsupported, or -1 if we couldn't for some other reason.
int try_html_help_string(char *topic)
{
	char url[FNAMELEN], prog[FNAMELEN], *p, *q, cmdline[100], dir[FNAMELEN];

	if (epsilon_manual_port > 0)
		sprintf(url, "http://127.0.0.1:%d/man/", epsilon_manual_port);
	else
		strcpy(url, "http://www.lugaru.com/cgi-bin/lhelp/man/");
	for (p = topic, q = url + strlen(url); *p; p++, q++)
		if (*p == ' ') {
			strcpy(q, "%20");
			q += 2;
		} else
			*q = *p;
	*q = 0;
	if (!*topic)
		strcat(url, "0");
	p = lookpath("lhelp/epsilon.lhl");
	if (p) {		// Find name of directory containing our lhelp files.
		strcpy(dir, p);
		*get_tail(dir, 0) = 0;
	} else
		get_executable_directory(dir);
	convert_to_8_3_filename(dir);
	if (opsys == OS_UNIX) {
		sayput("Running a browser...");
		get_executable_file(prog, "lhelpd", 1);
		sprintf(prog + strlen(prog), " -p %d -d \"%s\" &",
				epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			shell("", prog, "");
	} else if (is_gui || is_win32) {
		sayput("Starting a browser, please wait...");
		expire_message = 1;
		get_executable_directory(prog);
		strcat(prog, "lhelp.exe");
		sprintf(cmdline, "lhelp -p %d -d %s", epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			winexec(prog, cmdline, SW_SHOWMINNOACTIVE, 0);
	} else
		return 0;
	if (!try_show_url(url))
		return -1;
	note("");
	return 1;
}

html_help_string(char *topic)
{
	switch (try_html_help_string(topic)) {
		case 0:
			error("Don't know how to run a web browser in this environment, sorry.");
		case -1:
			error("Error starting a browser.");
	}
}

tell_key(full, pr)
char *pr;
{
	char msg[300], tmp[300], *flavor;
	int index;

	index = *ask_key(pr, tmp, 0);
	user_abort = 0;
	say("");
	if (key != full_key)
		sprintf(tmp + strlen(tmp), " (becomes %k)", key);
	if (index <= 0)
		sprintf(msg, "The key %s is not bound to any command.", tmp);
	else {
		switch (name_type(index)) {
			case NT_MACRO:	flavor = "macro"; break;
			case NT_SUBR:	flavor = "subroutine"; break;
			default:	flavor = "command"; break;
		}
		sprintf(msg, "%s runs the %s %s.",
				tmp, flavor, name_name(index));
		if (full == 2)
			info_to_epsilon_node(name_name(index), 1);
		else if (full) {
			switch (get_epsilon_help_format()) {
				case HELP_FORMAT_WINHELP:
					show_help_for(name_name(index));
					break;
				case HELP_FORMAT_HTML:
					if (try_html_help_string(name_name(index)) == 1)
						break;
				default:	// Fall through.
					dodesc(name_name(index), msg, 5);
			}
			return;
		}
	}
	say("%s", msg);
}

command describe_command()
{
	char cmd[80];

	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_string("", "commands, list of");
				return;
			case HELP_FORMAT_HTML:
				if (try_html_help_string("command reference") == 1)
					return;
		}
	iter = 0;
	save_var show_mouse_choices = 1;
	get_cmd(cmd, "Describe command", "");
	show_cmd_help(cmd);
}

show_cmd_help(char *cmd)
{
	char msg[300];

	switch (get_epsilon_help_format()) {
		case HELP_FORMAT_WINHELP:
			sprintf(msg, "%s command", cmd);
			win_help_string("", msg);
			break;
		case HELP_FORMAT_HTML:
			if (try_html_help_string(cmd) == 1)
				break;
		default:		// Fall through.
			descr_bindings(cmd, msg);
			dodesc(cmd, msg, 5);
			break;
	}
}

command describe_variable()
{
	char var[80];

	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_string("", "all variables, list of");
				return;
			case HELP_FORMAT_HTML:
				if (try_html_help_string("variable reference") == 1)
					return;
		}
	iter = 0;
	save_var show_mouse_choices = 1, show_all_variables = 1;
	get_var(var, "Describe variable", "", MUST_MATCH);
	restore_vars();
	show_var_help(var);
}

show_var_help(var)
char *var;
{
	char msg[300];

	switch (get_epsilon_help_format()) {
		case HELP_FORMAT_WINHELP:
			sprintf(msg, "%s variable", var);
			win_help_string("", msg);
			break;
		case HELP_FORMAT_HTML:
			if (try_html_help_string(var) == 1)
				break;
		default:		// Fall through.
			make_var_help(var, bufnum, window_handle, 0);
			sprintf(msg, " Help on %s ", var);
			view_help(msg, 1);
			break;
	}
}

give_binding(s, cmd) // put text telling how to invoke cmd in s
char *s, *cmd;
{
	int ind;

	ind = find_index(cmd);
	*s = 0;
	if (ind <= 0 || !find_bindings(s, ind, mode_keys, root_keys, "", 0))
		sprintf(s, "Alt-x %s", cmd);
}

// Display edoc-style help for str.  Insert "also" text at the start of
// the topic.  Flags: 1 means this is the last action of the command, so
// set up the window and return.  Without it, we don't return until the
// user presses some key, then return that key.  The 2 flag prevents
// displaying the h option.  The 4 flag forces using the internal help
// format.

dodesc(char *str, char *also, int flags)
{
	char *old = bufname, msg[80];
	int start, i, browser_ok = !(flags & 2);

	if (!(flags & 4))
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
			case HELP_FORMAT_HTML:
				show_help_for(str);
				return;
		}
	if (!get_doc(0))
		return;
	i = find_index(str);
	if (!i || name_help(i) < 0) {
		gripe("No information on %s", str, flags & 1);
		bufname = old;
		return;
	}
	point = name_help(i);
	while (isspace(curchar()))
		point++;
	start = point;
	re_search(1, "^!~");
	xfer(HELPBUF, start, point);
	bufname = HELPBUF;
	if (also && *also)
		point = 0, stuff(also);
	insert('\n');
	point = size();
	if (is_unix != IS_UNIX_XWIN && !is_gui && !is_win32)
		browser_ok = 0;
	if (!(flags & 1) && browser_ok)
		bprintf("Press H to see this topic in a web browser, with\n"
				"links to the rest of Epsilon's online manual.\n");
	point = 0;
	bufname = old;
	sprintf(msg, " Help on %s ", str);
	view_help(msg, flags & 1);
	if (browser_ok && tolower(key) == 'h') {
		ungot_key = -1;
		html_help_string(str);
		drop_pending_says();
	}
}

char *man_use_program(int type);

make_man_page(char *topic, int type)
{
	char cmdline[FNAMELEN], buf[FNAMELEN];
	int color;

	save_var case_fold = 1;
	sprintf(cmdline, "%s %s", man_use_program(type), topic);
	sprintf(buf, "man-%s", topic);
	quiet_set_bookmark();
	to_buffer(buf);
	man_mode();
	if (!size() || type != _man_program) {
		set_read_only(0);
		zap(buf);
		translation_type = FILETYPE_AUTO;
		pipe_text(NULL, buf, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		point = 0;
		if (parse_string(1, "Reformatting.*please wait.*\n"))
			delete(0, matchend);
		simple_re_replace(1, "\n\n\n+", "\n\n");
		set_character_color(0, size(), -1);
		point = 0;
		while (re_search(1, ".\b.")) {
			if (character(point - 3) == '_')
				color = color_class man_underlined;
			else
				color = color_class man_bold;
			delete(point - 3, point - 1);
			set_character_color(point - 1, point, color);
			point--;
		}
		color_ansi_buffer();
		filename = "fake-man-file.man";	// So find-linked-files works.
		force_save_as = FSA_NEWFILE;
		discardable_buffer = buffer_not_saveable = 1;
	}
	modified = point = 0;
	_man_program = type;
	set_read_only(3);
}

command show_last_keys()
{
	char line[300];
	int i, c, len;

	len = MIN(screen_cols - _view_left - _view_right - 15,
			  sizeof(line) / sizeof(*line));
	if (is_gui && want_gui_prompts)
		len = 48;
	bufnum = zap(HELPBUF);
	bprintf("The last %d keys pressed were:\n", MAXRECENT);
	*line = 0;
	for (i = 0; i < MAXRECENT; i++) {
		c = _recent_keys[(i + _recent_place) % MAXRECENT];
		if (c == ' ')
			strcat(line, "<Sp>");
		else
			show_char(line, c, 1);
		if (strlen(line) >= len)
			bprintf("%s\n", line), *line = 0;
		else
			strcat(line, " ");
	}
	bprintf("%s", line);
	view_help(" Last Keys Pressed ", 1);
}

hex_reload_file()
{
	if (!*filename || force_save_as)
		return;
	if (!ask_yn("Reread in binary mode?", "Warning: File was not "
				"read in binary mode.  Reload in binary mode?",
				"Yes", "No", 1))
		return;
	if (!do_revert_file(MAKE_TRANSLATE(GET_ENCODING(translation_type),
									   FILETYPE_BINARY, 1)))
		quick_abort();
}

int hex_old_on_modify();
int hex_on_modify();
int color_hex_range();
char *hex_prefix();

command hex_mode()
{
	char bname[FNAMELEN];
	int pos = point, trans = GET_LINE_TRANSLATE(translation_type);
	struct file_info saved = file_info;

	if (is_hex_buffer(bufnum)) {	// "Turn off mode" -- save back to real buf.
		hex_save_key();
		return;
	}
	if (!size())
		error("Buffer is empty.");
	sprintf(bname, "%s%s", HEX_BUFNAME_PREFIX, bufname);
	if (exist(bname)) {
		say("This buffer is in hex mode.  Press %s? for help or %sq to quit.",
			  hex_prefix(), hex_prefix());
		to_buffer(bname);
		return;
	}
	if (trans == FILETYPE_MSDOS || trans == FILETYPE_MAC)
		hex_reload_file();
	buffer_to_hex(bname);
	if (user_abort) {
		delete_buffer(bname);
		error("Canceled.");
	}
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_old_on_modify;	// special funcs can modify it.
	file_info.check_type = 0;	// Disable file date checking during hex mode.
	to_buffer(bname);
	file_info = saved;		// Remember old file date checking info.
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_on_modify;	// special funcs can modify it.
	mode_keys = hex_tab;		/* Use these keys. */
	major_mode = hex_overtype_mode ? _hex_overtype_mode_name : _hex_mode_name;
	recolor_range = color_hex_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_delay = -1; // No idle coloring, it slows down apply-changes.
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("hex-mode-hook");
	drop_all_colored_regions();
	make_mode();
	say("Entering hex mode, press %s? for help or %sq to quit.",
		  hex_prefix(), hex_prefix());
	hex_goto_place(pos, hex_overtype_mode);
}

html_tag_region(int type, int from, int to)
{
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	point = 0;
	switch (type) {
		case HT_JSCRIPT:
			tag_mode_jscript();
			break;
		case HT_VBSCRIPT:
			tag_mode_vbasic();
			break;
		case HT_PHP:
			save_var sublanguage_code = SUBLANG_PHP;
			tag_mode_perl();
			break;
		case HT_PYTHON:
			tag_mode_python();
			break;
	}
}

do_delete_blank_lines(int force, int cnt)
{
	int num = 0, c;
	int start;

	check_modify(bufnum);
	re_search(-1, "[ \t\n]*");	/* go to start of this whitespace */
	if (point)			/* skip prev line */
		point += parse_string(1, "[ \t]*\n", NULL);
	start = point;
	while (point < size()) {	/* count lines in whitespace */
		if ((c = curchar()) == '\n')
			num++;
		else if (c != ' ' && c != '\t')
			break;
		point++;
	}
	re_search(-1, "[ \t]*");	/* skip indentation of last line */
	if (num)
		delete(start, point);
	if (!force)			/* with no argument, */
		cnt = (num > 1);   /* 1 or 0 blank lines => 0, more => 1 */
	while (cnt-- > 0)		/* insert this many newlines */
		insert('\n');
	point = start;
}

command delete_blank_lines()
{
	if (is_highlight_on()) {
		save_spot point, mark, narrow_start, narrow_end;
		fix_region();
		narrow_start = point;
		narrow_end = size() - mark;
		if (!parse_string(1, "<wspace>*\n"))
			if (!re_search(1, "\n<wspace>*\n"))	// Find first blank line.
				error("No blank line in region.");
		do {
			do_delete_blank_lines(has_arg, iter);
			re_search(1, "\n*");	// Pass this batch of blank lines.
			re_search(1, "\n<wspace>*\n");	// To next batch.
			check_abort();
		} while (point < size() - narrow_end);
	} else
		do_delete_blank_lines(has_arg, iter);
	iter = 0;
}

get_text_area(w_handle)
{
	save_var window_handle = w_handle;

	w_left = window_text_left(window_handle);
	w_top = window_text_top(window_handle);
	// For broken updates: Avoid calling new prim if running old executable.
	w_right = w_left + (strcmp(versioned_file_string, "13.00")
						? window_content_width() : text_width()) - 1;
	w_bottom = w_top + text_height() - 1;
}

process_find_directory(char *file, int buf, int pos)
{
	char dir[FNAMELEN], url[FNAMELEN];
	int level = 0, got = 0, c;

	save_var bufnum = buf, point = pos;
	*dir = 0;
	if (is_process_buffer(bufnum) && *process_current_directory
		  && !is_remote_buffer(bufnum))
		strcpy(dir, process_current_directory);
	while (!got && re_search(-1, "(Entering|Leaving|Running `.*' in)"
							 " directory `(.*)'$")) {
		switch (c = character(find_group(1, 1))) {
			case 'L':	level++; break;
			case 'E':	if (--level >= 0)
							break;
			case 'R':	got = c;
						grab(find_group(2, 1), find_group(2, 0), dir);
						break;
		}
	}
	if (telnet_error_converter(file, *dir ? dir : NULL, url)) {
		strcpy(file, url);
		return;
	}
	if (got != 'R')	// "Running" message is from us, so don't translate it.
		convert_cygwin_filename(dir, 0);
	if (opsys == OS_DOS && strchr(file, '/'))
		convert_cygwin_filename(file, 0);
	if (*dir)
		absolute(file, dir);
	else
		absolute(file);
}

// First, define some regex pieces that could be helpful in multiple
// definitions.

// These file name matchers say, optionally a drive spec c:, then a
// basename, a dot, and an extension.  The OK_SPACE version permits
// spaces in the basename, starting with the 2nd character.  Both permit
// tags like <3> at the end of the extension.
#define FNAME_NO_SPACE	"(([a-z]:)?[+-9A-z~]+<.>[+-9A-z]+[0-9<>]*)"
#define FNAME_OK_SPACE	"(([a-z]:)?[+-9A-z~][+-9A-z~ ]*<.>[+-9A-z]+[0-9<>]*)"

// Some things we want to ignore in an error message.

// 3> or E " or W " or ***, for example.
#define START_STUFF1	"[1-9]>|e \"|w \"|<*>*"
// One of these three keywords, or [someother].
#define START_STUFF2	"error|warning|fatal| *%[[a-z]+%]"
// Optionally, a code like E123 with a space before it, then perhaps
// some punc chars * : after.  Or just punc chars " or (.
#define START_STUFF3	"( [a-z][0-9]+)?[*:]*|[\"(]?"

// Between a file name and line number, we must have space, open paren,
// comma or colon, then perhaps L or the word "line", then perhaps
// space.  Also skip over any " at the start, which may have surrounded
// file name.
#define BEFORE_NUM		"\"?([ (:,]( *line|L)?) *"
// The line number.
#define LINE_NUM		"([0-9]+)"
// Stuff to require after line number.  Either ) or colon, or else some
// stuff with colon or semicolon after it.
#define END_STUFF		"(<rparen|:>|[^0-9\\\n].*[:;])"

/* If you change ERROR_PATTERN, you may have to change the next 3 macros
 * as well. */
#define ERROR_PATTERN	"^(" START_STUFF1 \
						"(" START_STUFF2 "|[a-z0-9]+ )" \
						START_STUFF3 "|[ \t]*from) *" \
						FNAME_NO_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN	4	/* filename is in 4th () pair above */
#define LINE_PAREN	8	/* line# is in 8th pair of parens above */

#define SIMP_PATTERN	"[:;)]| line "	/* search with simple pattern first */

// This version permits spaces in file names, but with more
// restrictions on ancillary text before it.
#define ERROR_PATTERN2	"^(" START_STUFF1 \
						"(" START_STUFF2 ")" \
						"( [a-z][0-9]+)?[*:]*|[\"(]?) *" \
						FNAME_OK_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN2	4	/* filename is in 4th () pair above */
#define LINE_PAREN2	8	/* line# is in 8th pair of parens above */

#define ERROR_PATTERN3	".* at (.+) line ([0-9]+)[.,].*"
#define FILE_PAREN3	1	/* filename in () pair above */
#define LINE_PAREN3	2	/* line# in pair of parens above */

// An error pattern that permits file names with no . character, but
// with more restrictions on the format.
#define ERROR_PATTERN4	"^([+-9A-z~:]+[0-9<>]*)" BEFORE_NUM \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN4	1	/* filename is in 1st () pair above */
#define LINE_PAREN4	2	/* line# is in 2nd pair of parens above */

// An error pattern for Java runtime errors.

#define ERROR_PATTERN5	"^.*%(([+-9A-z~:]+):([0-9]+)%)"
#define FILE_PAREN5	1	/* filename is in 1st () pair above */
#define LINE_PAREN5	2	/* line# is in 2nd pair of parens above */

// An error pattern that permits file names with no . character,
// allowing spaces, but with even more restrictions on the rest of the
// message.
#define ERROR_PATTERN6	"^([a-z]:\\[ +-9A-z~]+)[(]" \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN6	1	/* filename is in 1st () pair above */
#define LINE_PAREN6	2	/* line# is in 2nd pair of parens above */

find_error(int dir, int *file_paren, int *line_paren)
{
	int old = point;

	if (point > 0 && dir > 0)
		to_end_line();
	else
		to_begin_line();
	while (re_search(dir, SIMP_PATTERN)) {
		to_begin_line();
		if ((!*ignore_error || !parse_string(1, ignore_error, NULL))
			  && !parse_string(1, "In file included from ", NULL)) {
			if (parse_string(1, ERROR_PATTERN, (char *)0)) {
				*file_paren = FILE_PAREN;
				*line_paren = LINE_PAREN;
				return 1;
			}
#ifdef ERROR_PATTERN2
			if (parse_string(1, ERROR_PATTERN2, (char *)0)) {
				*file_paren = FILE_PAREN2;
				*line_paren = LINE_PAREN2;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN3
			if (parse_string(1, ERROR_PATTERN3, (char *)0)) {
				*file_paren = FILE_PAREN3;
				*line_paren = LINE_PAREN3;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN4
			if (parse_string(1, ERROR_PATTERN4, (char *)0)) {
				*file_paren = FILE_PAREN4;
				*line_paren = LINE_PAREN4;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN5
			if (parse_string(1, ERROR_PATTERN5, (char *)0)) {
				*file_paren = FILE_PAREN5;
				*line_paren = LINE_PAREN5;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN6
			if (parse_string(1, ERROR_PATTERN6, (char *)0)) {
				*file_paren = FILE_PAREN6;
				*line_paren = LINE_PAREN6;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN7
			if (parse_string(1, ERROR_PATTERN7, (char *)0)) {
				*file_paren = FILE_PAREN7;
				*line_paren = LINE_PAREN7;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN8
			if (parse_string(1, ERROR_PATTERN8, (char *)0)) {
				*file_paren = FILE_PAREN8;
				*line_paren = LINE_PAREN8;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN9
			if (parse_string(1, ERROR_PATTERN9, (char *)0)) {
				*file_paren = FILE_PAREN9;
				*line_paren = LINE_PAREN9;
				return 1;
			}
#endif
		}
		if (dir > 0)
			nl_forward();
	}
	*error_spot = old;
	return 0;
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
			   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				if (in_shell_buffer && !parse_string(-1, "([ \n\t|&;()<>]|^)#"))
					break;
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "#?((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_variable);
				else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer) {
					if (character(t - 1) != '\\') {
						search(1, "'");	// No \-quoting in '' for shell mode.
						set_character_color(t, point, color_class perl_string);
					}
				} else if (!strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && character(t - 1) == '\\')
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_><word>*)"
									 : "(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

int prefill_line_matches(char *pat, int one_parag, char *skip, char *sing_pat)
{
	int i;

	save_var point;
	to_begin_line();
	if (!*pat && one_parag)
		return !parse_string(1, skip) && !matches_at(point, 1, "[ \t]*$");
	i = parse_string(1, pat);
	if (!one_parag)
		return (i || prefill_have_single_line(sing_pat));
	return i && i == parse_string(1, skip)
			&& (!one_parag || matches_at(point + i, 1, "[^ \t\n]"));
}

prefix_fill_one_paragraph(char *pat, char *skip, int force)
{
	char prefix[FNAMELEN], quoted[FNAMELEN];

	save_spot point--;
	re_search(1, "[^\n\t ]");
	if (force || matches_at(give_begin_line(), 1, pat)) {
		save_spot mark, narrow_start, narrow_end;
		save_var margin_right;
		to_begin_line();
		*prefix = 0;
		parse_string(1, skip, prefix);
		if (!*prefix && !force) {
			fill_paragraph();
			return;
		}
		margin_right -= current_column();
		if (margin_right < 1)
			margin_right = 1;
		sprintf(quoted, "^%q", prefix);
		prefix_find_fill_region(1, skip, pat);
		// We've now narrowed to one paragraph, and know the quoting
		// prefix shared by all its lines.  So strip the prefix, fill,
		// and put it back.
		if (*prefix)
			simple_re_replace(1, quoted, "");
		highlight_off();
		fill_paragraph();
		if (*prefix)
			simple_re_replace(1, "^", prefix);
	} else
		fill_paragraph();
}

// Fill a region or a paragraph, using the specified rules for
// determining a possible fill prefix.  If !force, do a plain
// fill-paragraph when the line doesn't match the prefix pattern.
prefix_fill_region(char *pat, char *skip, int force)
{
	int last = 0;
	iter = 0;
	if (has_arg)
		save_var margin_right = current_column() + 1;
	if (!is_highlight_on())
		prefix_fill_one_paragraph(pat, skip, force);
	else {
		char prefix[FNAMELEN], quoted[FNAMELEN];
		save_spot point, mark, narrow_start, narrow_end;
		fix_region();
		do {
			last = point;
			prefix_fill_one_paragraph(pat, skip, 1);
			to_begin_line();
			*prefix = 0;
			parse_string(1, skip, prefix);
			sprintf(quoted, "%q", prefix);
			while (prefill_line_matches(quoted, 1, skip, pat) && nl_forward())
				check_abort();
			while (character(parse_string(1, skip) ? matchend : point) == '\n'
				   && nl_forward())	// Move past all-prefix lines to next paragraph.
				check_abort();
		} while (point < mark && point > last);
	}
}

color_ansi_buffer()
{
	save_var current_output_color;
	color_ansi_output(0);
}

command process_yank()
{
	int buf, lines;
	char msg[FNAMELEN];

	get_clipboard();
	buf = name_to_bufnum(_cur_kill_buf);
	lines = buf_exist(buf) ? count_lines_in_buf(buf, 0) : 0;
	sprintf(msg, "Are you sure you want to yank %d line%s "
			"for immediate execution? ", lines, (lines == 1) ? "" : "s");
	if (!in_macro() && _len_def_mac <= 0 && point >= type_point
		  && (process_yank_confirm > 0 && lines >= process_yank_confirm)
		  && is_process_buffer(bufnum))
		if (!ask_yn("Confirm yank", msg, "Yes", "No", 0))
			return;
	delete_if_highlighted();
	retrieve(0);
}

suffix_jy()
{
	python_mode();
}

tag_mode_python()
{
	char func[TAGLEN];
	int start, color;

	python_mode();
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	while (re_search(1, "^[ \t]*(def|class)[ \t]+([a-zA-Z_]<word>*)[(:]")) {
		color = get_character_color(matchstart);
		if (color == color_class python_comment
			  || color == color_class python_string)
			continue;
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN], search_other_buf_name[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	if (search_in_region && is_highlight_on()) {
		save_var _highlight_control;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	} else if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var _highlight_control;
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	if (!search_other_name) {
		sprintf(search_other_buf_name, "search-%s", bufname);
		save_var search_other_name = search_other_buf_name;
	}
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = full_key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps & 1) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))), "search");
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, comp_tab) == (short) inp_get_other) {
			flags = pick_search_recall(flags);
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		flags = grab_search_string(flags, &trystr);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = editwin;
	if (size()) {
		flags = grab_search_string(flags, &trystr);
		save_search_recall(flags, trystr);
	}
	if (key == abort_key || user_abort) {
		window_handle = origwin;
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			if (_last_was_regex)
				grab_search_string(flags, &_default_regex_search);
			else
				grab_search_string(flags, &_default_search);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

ask_wrap(int flags)
{
	char msg[FNAMELEN];
	sprintf(msg, "Not found.  Wrap to %s of buffer and continue searching?",
			(flags & SEARCH_REVERSE) ? "end" : "start");
	return ask_yn("Wrap and continue searching?", msg, "Yes", "No", 1);
}

repeat_search(cnt, flags)	 // search & report errors
{
	int orig, ret = 0, wrap = (search_wraps & 2) ? 1 : 0;
	char *s;

	s = default_search_string(flags);
	while (cnt-- > 0) {
        if (wrap == 2)      // Just wrapped, no more wrapping.
            wrap = 0;
        else
            orig = point;
		ret = do_color_searching(flags, s);
		if (ret == 1)
			continue;
		if (ret == DSABORT)
			say("Search canceled");
		else if (wrap == 1 && ask_wrap(flags)) {
            wrap = 2;
            point = (flags & REVERSE) ? size() : 0;
            say("Search wrapped.");
            cnt++;
            continue;
        } else {
			say("%s: not found", s);
			point = orig;
			if (in_macro(1))
				quick_abort();
		}
		break;
	}
	if (search_positions_at_start && ret == 1)
		point = matchstart;
}

struct find {
	char *find_text;
	char *replace_text;
	int flags;
} find_data;		// The find dialog looks at this variable directly.

int our_matchstart;	// make sure matchstart no longer changes unexpectedly

int search_and_show(flags, find_text, start_here, on_failure)
char *find_text;
{
	char msg[256];
	int r = find_from(flags, find_text, start_here, on_failure);
	switch (r){
		case 0:
			sprintf(msg, "Not found: %s", find_data.find_text);
			redisplay_in(window_lines_visible(window_handle));
			find_dialog_say(msg);
			if (!(search_wraps & 2) || !ask_wrap(flags)
				  || !find_from(flags, find_text,
								(flags & SEARCH_REVERSE) ? size() : 0, on_failure))
				break;
			// Fall through.
		case 1:
			mark = matchstart;
			highlight_on();
			redisplay_in(window_lines_visible(window_handle));
			our_matchstart = matchstart;
			find_dialog_say("");
			break;
		case 3:
			sprintf(msg, "Bad regex pattern: %s", find_data.find_text);
			find_dialog_say(msg);
			break;
	}
	return r;
}

do_sort_buffer(rev)	/* ask user where to sort, then do it */
{
	char buf[FNAMELEN];

	save_var show_status = 1;
	save_var inp_get_other_name = bufname;
	get_buf(buf, "Sort current buffer into buffer", "sorted");
	if (name_to_bufnum(buf) != bufnum)
		zap(buf);
	do_buffer_sort(buf, has_arg ? iter : 0, rev);
	to_buffer(buf);
	iter = 0;
}

do_buffer_sort(buf, col, rev)
char *buf;
{
	if (sort_case_fold == 0 || sort_case_fold == 1)
		save_var case_fold = sort_case_fold;
	if (name_to_bufnum(buf) == bufnum) {
		char *tmp = temp_buf();
		on_exit delete_buffer(tmp);
		int res = buffer_sort(tmp, col, rev != 0);
		delete(0, size());
		grab_buffer(name_to_bufnum(tmp));
		return res;
	}
	return buffer_sort(buf, col, rev != 0);
}

when_loading()
{
	drop_name("tag-suffix-py");
	drop_name("tag-suffix-python");
	cc_tab[ALT('d')] = find_index("insert-date");
	int i = find_index("want-gui-help");
	if (i && !get_num_var(i) && !epsilon_help_format_win_gui)
		epsilon_help_format_win_gui = HELP_FORMAT_POPUP;
	drop_name("want-gui-help");
	i = find_index("want-gui-help-console");
	if (i && !get_num_var(i) && !epsilon_help_format_win_console)
		epsilon_help_format_win_console = HELP_FORMAT_POPUP;
	drop_name("want-gui-help-console");
	drop_name("use-winhelp");
}
