/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2007 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.02 to 13.03.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "misclang.h"
#include "perl.h"

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	strcpy(last, "");	// See if "class foo" is before this.
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (c_ignore_declspec())
			continue;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords & (CSHARP_KEYWORDS
				| JAVA_KEYWORDS | JS_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum") || !strcmp(this, "__interface")
				  || !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, (c_extra_keywords == CSHARP_KEYWORDS) ? "." : "::");
						class += strlen(class);
					}
					strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */
	int in_initializer;

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!(reindent_c_comments & 1) && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if ((!reindent_one_line_c_comments || 
		 (indenter_action == IACT_REINDENT_PREV) && !(reindent_c_comments & 8))
		  && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' && reindent_c_preprocessor_lines == 1
		  || first_char == '\f'
		  || first_char == '@' && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(1, "@(" OBJC_TOPBLOCKS ")[ \t]+<word>|@end"))
		return 0;
	else if (first_char == '}')
		return cind_close();
	else if (have_c_close_equivalent(orig))
		return indent_at_c_close_equivalent(orig);

	to_begin_line();
	if (parse_string(-1, PREPROC_ELSE ".*\n")) {
		point = matchend;
		locate_by_ifdef(-1);
	}
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig) || c_align_break_with_case
		  && matches_at(orig, 1, "[ \t]*break;")) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig) && !((c_extra_keywords & OBJC_KEYWORDS)
									 && matches_at(pstart, 1, "^[-+@]"))) {
		if (c_in_struct())		// Could it be a bitfield, not a label?
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';' || c_semi_equivalent(pend, orig)) {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (c_close_equivalent(pend)) {
		point = pend;
		ind = indent_after_c_close_equivalent(orig);
	} else if (prev_end == '{' || c_open_equivalent(pend, orig)) {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_initializer = in_c_initializer(orig)) {
		return ind;
	} else if (c_next_no_indent(orig, pstart)) {
		return 0;
	} else if (c_next_same_indent(orig, pstart)) {
		return ind;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, in_initializer);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

// Force this line to the left edge?
int c_next_no_indent(int orig, int pstart)
{
	if (c_extra_keywords & OBJC_KEYWORDS)
		return objc_next_no_indent(orig, pstart);
	return 0;
}

// Force this line to match the last line?
int c_next_same_indent(int orig, int pstart)
{
	UNREFERENCED_PARAMETER(orig);
	if ((c_extra_keywords & (JAVA_KEYWORDS)) && matches_at(pstart, 1, "@[A-Z0-9]"))
		return 1;	// Java annotation on prev line of this definition.
	return 0;
}

// See if there's an EEL patch file for this version available, and load it.
load_version_patch_file()
{
	char file[FNAMELEN], ver[FNAMELEN], *s;
	strcpy(ver, eel_version);
	while (s = strchr(ver, '.'))
		strcpy(s, s + 1);
	sprintf(file, "v%spatch.e", ver);
	show_text(-2, 0, "Looking for patch file %s on EPSPATH", file);
	if (!load_eel_from_path(file, 1))
		show_text(-2, 0, "Loaded %s", (s = lookpath(file)) ? s : file);
}

do_insert_file(file, strip)	/* insert file into current buffer at pt */
char *file;
{
	int orig = bufnum, buf, err;

	bufnum = buf = tmp_buf();
	err = err_file_read(file, strip);
	if (!err)
		buf_xfer(orig, 0, size());
	bufnum = orig;
	buf_delete(buf);
	return err;
}

err_file_read(file, strip)	/* like file_read, but show error */
char *file;
{
	int err;
	if (is_remote_file(file)) {
		save_var synchronous_file_io = 1;
		err = find_remote_file(file, strip);
	} else
		err = file_read(file, strip);
	if (window_bufnum == bufnum)
		switching_to_buffer();
	_read_aborted = (err == EREADABORT);
	if (err && !_read_aborted) {
		file_error(err, file, "read error");
		maybe_ding(bell_on_read_error);
	}
	return err;
}

int narrow_to_new_comment()
{
	char pat[FNAMELEN];

	if (strcmp(major_mode, "C"))
		sprintf(pat, "[ \t]*%s", get_comment_start());
	else
		strcpy(pat, "[ \t]*//");
	save_var point;
	to_begin_line();
	if (!parse_string(1, pat)) {
		narrow_start = point;	// This line doesn't start with one.
		restore_vars();
		to_end_line();
		if (!re_search(-1, pat))
			return 0;		// Can't find comment start.
		point = 0;		// Found on this line, but now find 1st instance.
		if (!re_search(1, pat))
			return 0;		// Shouldn't happen: found searching back, not fwd.
		narrow_start = point = matchstart;
	}
	while (nl_forward())		// Find the last such line.
		if (!parse_string(1, pat))
			break;
	narrow_end = size() - point;
	do {						// Find the first such line.
		point--;
		to_begin_line();
		if (!parse_string(1, pat)) {
			nl_forward();
			break;
		}
	} while (point > narrow_start);
	narrow_start = point;
	return 1;
}

int color_conf_range();
int color_batch_range();

command conf_mode()
{
	mode_default_settings();
	major_mode = _conf_mode_name;
	mode_keys = conf_tab;		/* Use these keys. */
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_conf_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 2) != 0;
	if (auto_show_conf_delimiters)
		auto_show_matching_characters = conf_auto_show_delim_chars;
	try_calling("conf-mode-hook");
	make_mode();
}

command batch_mode()
{
	mode_default_settings();
	major_mode = _batch_mode_name;
	mode_keys = batch_tab;		/* Use these keys. */
	strcpy(comment_start, "rem[ \t]*");
	strcpy(comment_pattern, "rem.*$");
	strcpy(comment_begin, "rem ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_batch_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 4) != 0;
	if (auto_show_batch_delimiters)
		auto_show_matching_characters = batch_auto_show_delim_chars;
	try_calling("batch-mode-hook");
	make_mode();
}

// Color simple variable references within this string-like thing (already
// colored itself).
color_perl_interpolated(int from, int to, int comments)
{
	int p;

	save_var point = from;
	save_var narrow_end = size() - to;
	while (re_search(1, (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					 ? "[\\$]" : "[\\$@#]")) {
		p = point;
		if (character(point - 1) == '\\')
			point++;
		else if (character(point - 1) == '#') {
			if (!comments)
				continue;
			from = point - 1;
			to_end_line();
			if (point > to)
				point = to;
			set_character_color(from, point, color_class perl_comment);
		} else if (sublanguage_code == SUBLANG_PHP) {
			if (parse_string(-1, "[^\\]{%$")) {
				point -= 2;
				to_default_move_level(1, "{", "}", 1);
				set_character_color(p - 2, point, color_class perl_variable);
			} else if (parse_string(1, "<word>+(-%><word>+)*|{<word>+}")) {
				point = matchend;
				if (curchar() == '[')
					to_default_move_level(1, "[", "]", 0);
				set_character_color(p - 1, point, color_class perl_variable);
			}
		} else if (in_shell_buffer) {
			if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
							 "|[^ \t\n`]"))
				set_character_color(matchstart - 1, point = matchend,
									color_class perl_variable);
		} else if (curchar() == '{') {
			to_default_move_level(1, "{", "}", 1);
			set_character_color(p - 1, point, color_class perl_variable);
		} else if (parse_string(1, "#?([0-9]+|[A-Za-z_][0-9A-Za-z_]*)|%^[A-Z]"
								"|[]-[&`'+*|%=~^_./,\\\";:#?!@$<>()]") && point < to) {
			point = matchend;
			for (;;) {
				if (curchar() == '-' && character(point + 1) == '>'
					  && (isalnum(character(point + 2))
						  || strchr("[{", character(point + 2))))
					point += 2;
				else if (curchar() == '[' || curchar() == '{') {
					if (to_default_move_level(1, "[{", "]}", 1) < 0)
						break;
				} else
					break;
			}
			set_character_color(p - 1, point, color_class perl_variable);
		}
	}
}

color_perl_delimiters(int two, int interpolated, int modifiers)
{
	int start = point, c, end, level = 0, rstart, rend;
	char *openers = "([{<", *closers = ")]}>", *p;
	char pat[10];

	point += parse_string(1, "[ \t\n]+");
	set_character_color(start, point, color_class perl_comment);
	rstart = start = point;
	p = strchr(openers, c = end = curchar());
	if (isalnum(c) || isspace(c))
		return;			// No valid delimiter, so stop now.
	if (c == '\'' && interpolated != 2)
		interpolated = 0;
	if (p) {			// A paired delimiter.
		end = closers[p - openers];
		sprintf(pat, "[%c%c]", end, c);
		while (re_search(1, pat))
			if (character(point - 1) == c)
				level++;
			else if (--level <= 0)
				break;
		set_character_color(start, point, color_class perl_string);
		if (interpolated)
			color_perl_interpolated(start, point, 0);
	} else {
		point++;
		color_perl_fwd_quote(c, start, interpolated);
	}
	rend = point;
	if (two)		// A command like s///, so look for 2nd set.
		if (p) {	// After a paired delimiter, whitespace and pick new char.
			start = point;
			point += parse_string(1, "([ \t\n]|#.*)+");
			set_character_color(start, point, color_class perl_comment);
			color_perl_delimiters(0, interpolated, 0);	// One more pair.
		} else		// Find 3rd instance of char.
			color_perl_fwd_quote(c, point - 1, interpolated);
	if (modifiers) {
		start = point;
		if (modifiers == 2 && parse_string(1, "[a-z]*x"))
			color_perl_interpolated(rstart + 1, rend - 1, 1);
		point += parse_string(1, "[a-z]+");
		set_character_color(start, point, color_class perl_keyword);
	}
}

remove_xterm_sequences(int start)
{
	save_var point = start - 25, narrow_start = point;
	save_var case_fold = 0;
	simple_re_replace(1, "\x1B%][0-9];.*\7", "");
}

process_remove_controls(int start)
{
	save_var point = start, narrow_start = start;
	simple_re_replace(1, "\x1B\x5B[&=<>?]?[-0-9;]*[a-zA-Z@]", "");
	simple_re_replace(1, "[\x01-\x08\x0B-\x1A\x1C-\x1F]", "");
}

sort_region_with_settings(from, to, col, rev, buf)
{	/* sort a region in place, using case-folding from buf */
	char *temp1 = temp_buf(), *temp2 = temp_buf();
	int this_tab_size, this_case_fold, this_sort_case_fold;
	int res = ABORT_JUMP, orig = bufnum, sz, fixed_end = 0;
	jmp_buf this_level;

	bufnum = buf;
	this_case_fold = case_fold;
	this_sort_case_fold = sort_case_fold;
	this_tab_size = tab_size;

	bufnum = orig;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		point = from;
		if (curchar() == '\n' && character(point - 1) != '\n')
			point++;
		else
			to_begin_line();
		from = point;
		point = to;
		if (curchar() == '\n' && character(point - 1) != '\n')
			point++;
		else if (character(point - 1) != '\n' && point == size() - narrow_end) {
			insert('\n');	// Unterminated last line, nothing after.
			fixed_end = 1;
		} else
			to_begin_line();
		to = point;
		xfer(temp1, from, to);
		bufname = temp1;
		case_fold = this_case_fold;
		sort_case_fold = this_sort_case_fold;
		tab_size = this_tab_size;
		res = do_buffer_sort(temp2, col, rev);
		if (!res) {
			bufnum = orig;
			point = MIN(from, to);
			sz = buffer_size(temp2);
			grab_buffer(name_to_bufnum(temp2));
			delete(fix_narrowing(from) + sz, fix_narrowing(to) + sz);
			if (fixed_end && character(point - 1) == '\n')
				delete(point - 1, point);
		}
	}
	bufnum = orig;
	delete_buffer(temp1);
	delete_buffer(temp2);
	return res;
}
