/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2007 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.03 to 13.04.

when_loading()
{
	drop_name("strkeep-data");	// Changed type of variable.
}

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "misclang.h"
#include "perl.h"
#include "bufed.h"
#include "tags.h"
#include "kill.h"
#include "epsnet.h"
#include "proc.h"
#include "grep.h"
#include "html.h"
#include "vbasic.h"
#include "python.h"

do_switch_buffers(int orig_key)
{
	char bname[FNAMELEN];
	int other_key = -1;

	// If the key that invoked us is a function key, numpad or cursor
	// pad, or "gray key", find its shifted equivalent and look for that
	// also.  Usually Shift+orig key will give us the original key code,
	// if the original included Ctrl, but sometimes it might give us a
	// different one.
	if (!IS_EXT_ASCII_KEY(orig_key) && IS_TRUE_KEY(orig_key))
		other_key = NUMSHIFT(NUMPLAIN(orig_key));
	error_if_input(1);
	tiled_only();
	if (!(switch_buffers_options & 1))
		save_var _bufed_sort = 'a';
	if (!(switch_buffers_options & 2))
		save_var bufed_grouping = 0;
	do_bufed(BUFED_SWITCH | (has_arg ? BUFED_ALL : 0));
	move_in_switch_buffers(shift_pressed() & KB_SHIFT_DN);
	while ((shift_pressed() & KB_CTRL_DN)) {
		maybe_refresh();
		if (char_avail()) {
			getkey();
			if (key == GETFOCUS || key == LOSEFOCUS) {
				do_topkey();
				continue;
			} else if (key == orig_key || key == other_key)
				move_in_switch_buffers(shift_pressed() & KB_SHIFT_DN);
			else if (binding_of(key, root_keys) == (short) down_line)
				move_in_switch_buffers(0);
			else if (binding_of(key, root_keys) == (short) up_line)
				move_in_switch_buffers(1);
		} else
			delay(10, COND_KEY | COND_RETURN_ABORT);
		if (user_abort || key == abort_key || key == ESC
			  || key == NUMCTRL(GREYESC)) {
			while (char_avail())	// Discard keys so we don't get a delayed
				getkey();			// Canceled msg.
			user_abort = 0;
			remove_bufed();
			return;
		}
	}
	if (!bufed_getname(bname))
		*bname = 0;
	remove_bufed();
	if (*bname && exist(bname))
		to_buffer(bname);
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN], classname[FNAMELEN], pat[FNAMELEN];
	int start, funcstart, j, got_class = 0;

	start_of_function = -1;
	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	if (c_extra_keywords & OBJC_KEYWORDS)
		got_class |= find_objc_class(type, class);
	to_begin_line();
	start = point;
	point--;		// Begin searching at the last } char on the left margin...
	while (search(-1, "\n}")) {
		point++;		// ...that's not in a comment.
		if (!in_c_comment(c_look_back))
			break;
		point--;
	}

	strcpy(pat, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(");
	if (c_extra_keywords & OBJC_KEYWORDS)
		strcat(pat, "|\n[-+]");
	while (re_search(1, pat)) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*(define|if|pragma).*(\\\n.*)*")) {
			to_end_line();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				j = point;
				if (c_inside_block(point, start)) {
					got_class |= find_c_class_info(start, type, class, func);
					point = j + 1;
					if (got_class)
						continue;
				}
				point = j;
				c_skip_level('{', '}');
				break;
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					to_end_line();	/* c++ comment */
				break;
			case '\n':		// Objective-C method
				point++;
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				if (find_objc_method(func) == 1)
					return CF_INFO_FUNC | got_class;
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return got_class;
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + 1000);
				if (tag_c_func(0) == 2 && !find_c_func_oneliner(start)) {
					point = funcstart;	// Checks out as a function def.
					start_of_function = give_begin_line();
					*classname = 0;
					if (parse_string(-1, (c_extra_keywords == CSHARP_KEYWORDS)
									 ? "[ \t\n]*%.[ \t\n]*"
									 : "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						parse_string(-1, "[a-zA-Z0-9_]+", classname);
					}
					if (func)
						strcpy(func, funcname);
					if (class && *classname)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0)
							| got_class;
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return got_class;
}

indent_after_statement(orig) // orig is place needing indentation, for toplevel
{
	int lastcont, i, c, atparen = -1;

	if ((i = parse_string(1, "[ \t\f]*for[ \t\f]*%([ \t\f]*"))
		  && for_unclosed(point, orig))
		return get_column(point + i); // part of "for" statement
	for (;;) { /* was only the end of a larger statement, find start */
		do {	/* look back for line that ends with ;{}: */
			lastcont = point;
			to_begin_line();
			if (cind_try_space())
				break;
			if ((c = character(point - 1)) == ')') {
				atparen = lastcont;
				if (!c_rev_skip_level('(', ')'))
					break;
				if (after_c_macro_line())
					break;
				to_begin_line();
			} else if (c == ';')
				if (move_before_for_statement(orig))
					c = 'x';	// Keep looking.
		} while (!index(";{}:", c));
		if (c == ':' && parse_string(-1, "%)[ \t]*:")
			  && (c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS | JS_KEYWORDS))
			  && !matches_at(give_begin_line(), 1, "[ \t]*case</word>")
			  && !is_php_alt_open(point))
			continue;	// Not a label but C++ inheritance or similar.
		point = lastcont;	/* use indentation of line */
		to_c_indentation();	/* just after it */
		if (parse_string(RE_FORWARD, "(}[ \t\f]*)?e!lse[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_if())
				break;
		} else if (parse_string(RE_FORWARD,
								"(}[ \t\f]*)?!(catch|finally)[^a-zA-Z0-9_]")) {
				point = matchend;
				if (!to_matching_try())
					break;
			} else
			break;
	}
	i = current_column();
	if (!i && atparen != -1 && (Topindent || !c_statement_start()))
		i = get_indentation(atparen);	// 2nd func param decl
	if ((atparen = parse_string(1, "for[ \t\f]*%([ \t\f]*"))
	    && for_unclosed(point, orig))
		i = get_column(point + atparen); // part of "for" statement
	if (i < c_top_braces)
		return 0;	// 1st line of new func
	return i;
}

is_contin_statement(p, fd, orig) // would line after p be a continuation line?
{			// assumes p is just after last interesting char
	// if fd, pretend func param decls are a "continuation" of func decl
	// orig is later position in buffer, for unclosed "for" scan
	switch (character(p - 1)) {
		case ';':		// check for multi-line "for" statemnt
			if (unmatched_rparen(p, orig))
				return 1;
		case '{':
		case '}':
			return 0;
		case ')':
			save_var point = p;
			save_var narrow_start = p - 500;
			if (!c_rev_skip_level('(', ')'))
				return 1;
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && parse_string(-1, "@(synchronized|catch)"))
				return 0;
			if (sublanguage_code == SUBLANG_PHP)
				return !parse_string(-1, "[^a-z0-9_](if|while|for"
									 "|switch|foreach)[ \t\f]*")
						&& (fd || !Topindent || get_indentation(point));
			if (after_c_macro_line())
				return 0;
			return !parse_string(-1, "[^a-z0-9_](if|while|for"
								 "|switch)[ \t\f]*")
					&& (fd || !Topindent || get_indentation(point))
					|| in_perl_buffer;
		case 'e':
			return !is_else(p - 4);
		case 'o':
			save_var point = p;
			return !parse_string(-1, "[^a-z0-9_]do");
		case 'y':
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && matches_at(p, -1, "@(try|finally)"))
				return 0;
			return 1;
		case ':':
			if (is_php_alt_open(p))
				return 0;
			save_var point = p;
			to_indentation();
			return !is_label(point);
		default:
			return 1;
	}
}

char *clipboard_buf_name();

command insert_clipboard()
{
	if (buffer_flags(0) & 1)	// Turn off converting if orig buffer's
		save_var clipboard_convert_unicode = 0;	// already Unicode.
	int buf = zap(clipboard_buf_name());
	if (clipboard_to_buffer(buf,
							clipboard_convert_newlines(), clipboard_format)) {
		get_clipboard_fix_lines(buf);
		mark = point;
		set_region_type(REGNORM);
		grab_buffer(buf);
		build_first = 1;		/* redisplay hint */
	} else
		file_error(errno, "", "Couldn't retrieve clipboard.");
}

get_clipboard()		// push clipboard text on kill ring
{
	int new, old;

	if (!should_use_clipboard() || !clipboard_available())
		return;
	if (buffer_flags(0) & 1)	// Turn off converting if orig buffer's
		save_var clipboard_convert_unicode = 0;	// already Unicode.
	new = zap("-new-clipboard-");
	if (!clipboard_to_buffer(new, clipboard_convert_newlines(),
							 clipboard_format))
		return;
	get_clipboard_fix_lines(new);
	old = create(clipboard_buf_name());
	if (!buffers_identical(old, new)) {
		save_var bufnum = new;
		check_num_kill_buffers();	// copy clipboard to kill ring
		push_kill();
		_kill_width[_num_kill] = REGION;
		xfer(_cur_kill_buf, 0, size());
		buf_zap(old);		// copy new text to old clipboard buf
		buf_xfer(old, 0, size());
	}
	buf_zap(new);
}

c_keyword_color(from)
{
	char buf[500];

	save_var point;
	if (point - from > sizeof(buf) / sizeof(char) - 10)
		point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-.", buf[1])) {
		if ((c_extra_keywords & CSHARP_KEYWORDS)) {
			restore_vars();		// Rules for C# numbers are slightly different.
			point = from;
			point += parse_string(1, "-?%.?[0-9]([A-Za-z0-9_]|[Ee]-|.[0-9])*");
			grab(from, point, buf + 1);
		}
		return c_number_color(buf + 1);
	}
	strcpy(buf + point - from + 1, "|");
	if (is_c_keyword(buf) && (!(c_extra_keywords & CSHARP_KEYWORDS) ||
							  character(from - 1) != '@'))
		return color_class c_keyword;
	if (color_class c_function != color_class c_identifier
		  && paren_follows())
		return color_class c_function;
	return color_class c_identifier;
}

restore_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes)
{
	int i, j, k, classes = number_of_color_classes();
	int schemes = ptrlen(orig_schemes) / sizeof(int);

	for (i = 0; i < schemes; i++) {
		selected_color_scheme = orig_schemes[i];
		for (j = 0; j < classes; j++) {
			k = i * classes + j;
			set_color_pair(j, orig_fg[k], orig_bg[k]);
		}
	}
}

preserve_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes,
						int orig_scheme)
{
	int i, j, k, n, classes = number_of_color_classes();
	int schemes = ptrlen(orig_schemes) / sizeof(int), fg_color, bg_color;

	*get_color_scheme_variable() = selected_color_scheme;
	if (want_record_customizations()) {
		k = prep_customizations();
		if (selected_color_scheme != orig_scheme)
			list_one(k, color_scheme_variable_index(), 1);
		for (i = 0; i < schemes; i++) {
			selected_color_scheme = orig_schemes[i];
			for (j = 0; j < classes; j++) {
				fg_color = get_foreground_color(j, 1);
				bg_color = get_background_color(j, 1);
				n = i * classes+ j;
				if (fg_color != orig_fg[n] || bg_color != orig_bg[n])
					list_one_color_setting(k, j, 1);
			}
		}
		selected_color_scheme = *get_color_scheme_variable();
		say("");
	} else
		show_binding("To remember these colors for "
					 "future sessions, use %s.", "list-customizations");
}

// Make a copy of null-terminated string s that will be valid until the
// end of this session.  A pointer to static data won't be valid if the
// defining EEL code is changed.  Calling this makes mode names safe.
char *strkeep(char *s)
{
	if (!strkeep_data)
		strkeep_data = (char **) malloc(0);
	int sz = ptrlen(strkeep_data, 1) / sizeof(*strkeep_data), i;
	for (i = 0; i < sz; i++)
		if (!strcmp(s, strkeep_data[i]))	// Already here.
			return strkeep_data[i];
	strkeep_data = (char **) realloc((char *) strkeep_data, (sz + 1)
									 * sizeof(*strkeep_data) / sizeof(char));
	return strkeep_data[sz] = strsave(s);
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 0x20) {
				point = from;
				simple_re_replace(0, "<#0-#1>", "");
			}
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

void do_telnet(char *host, int port, char *bname)
{
	to_buffer(bname);
	type_point = 0;
	telnet_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_net_activity = telnet_activity;
	telnet_id = telnet_host(host, port, bname);
	char url[FNAMELEN];
	sprintf(url, "telnet://%s", host);
	buffer_url = strkeep(url);
}

telnet_send_line()
{
	char telnet_cbuf[200], hist_buf[FNAMELEN];

	if (point < type_point || !telnet_id) {
		enter_key();
		return;
	}
	sprintf(hist_buf, "-process-history-%s", bufname);
	create(hist_buf);
	process_cmds_to_end(hist_buf);
	while (point > type_point) {
		int i = MIN(point - type_point,
					sizeof(telnet_cbuf) / sizeof(*telnet_cbuf) - 1);
		grab(type_point, type_point + i, telnet_cbuf);
		if (telnet_server_echoes(telnet_id))
			delete(type_point, type_point + i);
		else
			type_point += i;
		telnet_send(telnet_id, telnet_cbuf);
		if (i > 0)
			buffer_printf(hist_buf, "%s", telnet_cbuf);
	}
	if (!has_arg) {
		if (!telnet_server_echoes(telnet_id))
			insert('\n'), type_point = point++;
		telnet_send(telnet_id, "\r\n");
	}
	buffer_printf(hist_buf, "\n");
}

maybe_restart_remote_process()
{
	if (buffer_url && !telnet_id && !is_process_buffer(bufnum)
		  && is_remote_command_line(buffer_url))
		url_operation(buffer_url, FTP_RECV);
}

int is_remote_command_line(char *cmd)
{
	return (!strncmp(cmd, "telnet://", 9) || !strncmp(cmd, "ssh://", 6));
}

// Fill in the hostname for active ssh buffers and return nonzero.
ssh_get_hostname(char *host)
{
	char *p;
	struct url_parts parts;

	if (mode_keys != ssh_tab)
		return 0;
	if (p = strchr(bufname, '@'))
		p++;
	else if (!strncmp(bufname, "ssh-", 4))
		p = bufname + 4;
	else if (*filename && is_remote_file(filename)
			 && parse_url(filename, &parts) && !strcmp(parts.service, "ssh")) {
		p = parts.host;
	} else
		return 0;
	strcpy(host, p);	// Make a copy of host name.
	if (p = strchr(host, ' '))
		*p = 0;
	return 1;
}

int ssh_activity(int activity, int buf, int from, int to);
char *scp_variable(struct url_parts *parts, char *varname);

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s");
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
	char url[FNAMELEN];
	sprintf(url, "ssh://%s", host);
	if (parts.have_usr)
		sprintf(url, "ssh://%s@%s", username, host);
	buffer_url = strkeep(url);
}

command ssh()
{
	char hostname[FNAMELEN], dir[FNAMELEN];
	struct url_parts parts;

	inet_only();
	if (!(has_feature & FEAT_MULT_CONCUR))
		error("SCP functions require Unix or Windows NT/2K/XP");
	tiled_only();
	*hostname = 0;
	if (get_buffer_directory(dir) && is_remote_file(dir)
		  && parse_url(dir, &parts) && strstr(" ssh scp ", parts.service)) {
		if (parts.have_usr)
			sprintf(dir, "%s@%s", parts.usr, parts.host);
		else
			sprintf(dir, "%s", parts.host);
		get_strdef(hostname, "Ssh to (hostname or user@hostname)", dir);
	} else
		get_str_auto_def(hostname, "Ssh to (hostname or user@hostname)");
	if (!*hostname)
		return;
	do_ssh_pick_buffer(hostname, "");
}

do_ssh_pick_buffer(char *hostname, char *usr)
{
	char buf[FNAMELEN], *p;
	int i = 1;

	sprintf(buf, "ssh-%s", hostname);
	p = buf + strlen(buf);
	for (;;) {		// Find an unused buffer.
		if (!exist(buf))
			break;
		bufname = buf;
		if (!is_process_buffer(bufnum))
			break;
		sprintf(p, " %d", i++);
	}
	do_ssh(hostname, buf, usr);
}

int divide_url(char *url, struct url_parts *p)
{
	memset(p, 0, sizeof(*p));
	strcpy(p->service, url);			// Parse service:// part.
	if (!split_string(p->service, ":", p->host) || !*p->service
		  || is_path_separator(p->host[0]) != 1
		  || is_path_separator(p->host[1]) != 1)	// It's not valid.
		return 0;
	strcpy(p->host, p->host + 2);		// Split off the service part.
	split_string(p->host, "/\\", p->fname);		// Remove the file name.
	return 1;
}

find_it(fname, strip)	/* find named file (must be in absolute form) */
char *fname;
{
	if (!*fname)
		return 2;
	if (look_file(fname)) {
		save_var _force_auto_read_changed_file = 1;
		to_buffer(bufname);
		if (!size() && !file_info.check_type)
			find_in_this_buf(fname, strip);
		else
			maybe_restart_remote_process();
		return 0;
	}
	return find_in_other_buf(fname, strip);
}

do_fill_region()
{
	int orig = point, start, atend = 0;
	spot end = alloc_spot();

	if (point > (*end = mark))
		*end = point, orig = point = mark;
	while (point < *end && !atend) {
		start = point;
		if (re_search(1, "\n[ \t]*\n"))
			point = matchstart + 1;
		else
			atend = 1;
		if (point > *end + 1)
			point = *end + 1;
		region_fill(start, point - 1);
		re_search(1, "[ \t\n]*");	/* skip paragraph indent */
	}
	mark = orig;
	point = *end;
	free_spot(end);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();
}

prepare_grep_buffer(char *buf, int flags, char *pat, char *filepat,
					char *forcedir)
{
	char search_type[FNAMELEN], dir[FNAMELEN];
	int before_head;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (grep_empties_buffer)
		modified = 0;
	save_var modified;
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	if (forcedir && *forcedir)
		strcpy(dir, forcedir);
	else
		getcd(dir);
	bprintf("Found 0 lines with %s\"%s\" in %s%s (cur dir \"%s\")\n",
			search_type, pat, (flags & BUFGREP) ? "(buffers) " : "",
			filepat, dir);
	set_character_color(before_head, point, color_class grep_header);

	if (wrap_grep == WRAP_ALWAYS)
		display_column = -1;
	else if (wrap_grep == WRAP_NEVER) {
		if (display_column < 0)
			display_column = 0;
	} else if (wrap_grep == WRAP_SPECIAL && display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	refresh();
	return before_head;
}

// If there's a collection with a date, use it.  Prefer later dates to
// earlier ones.
mshelp2_choose_collection(char *col)
{
	if (*mshelp2_collection) {		// Override default selection.
		strcpy(col, mshelp2_collection);
		return;
	}
	*col = 0;
	if (!mshelp2_wait_until_pat("Using default collection.*\n"))
		return;
	point = matchend;
	bprintf("listcoll\n");
	int before = size();
	if (!mshelp2_wait_until_pat("^--End\n"))
		return;
	int tmp = tmp_buf();
	buf_xfer(tmp, before, matchstart);
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	case_fold = 1;
	if (mshelp2_check_std_name("MS.NETFramework.v30", col)
		  || mshelp2_check_std_name("MS.VSCC.v90", col)
		  || mshelp2_check_std_name("MS.NETFramework.v20", col)
		  || mshelp2_check_std_name("MS.VSCC.v80", col)
		  || mshelp2_check_std_name("MS.VSCC.2003", col)
		  || mshelp2_check_std_name("MS.NETFramework", col)
		  || mshelp2_check_std_name("MS.VSCC", col))
		return;
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ", pat[100];
	int year, month;
	for (year = 2020; year > 2000; year--) {
		for (month = 12; month > 0; ) {
			sprintf(pat, "qtr.%d%.3s", year, months + --month * 4);
			point = 0;
			if (search(1, pat)) {
				grab_full_line(bufnum, col);
				return;
			}
		}
		sprintf(pat, "%d", year);
		point = 0;
		if (search(1, pat)) {
			grab_full_line(bufnum, col);
			return;
		}
	}
}

do_html_indent(int manual)
{
	int region, start, i;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 1);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	i = (1 << (region - 1 + (IS_ENTER_KEY(key) ? 8 : 0)));
	if (region > 0 && region <= HT_CSS && !(html_indenting_rules & i))
		region = 0;
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
		case HT_PHP:
			html_exclude_script_head();
			do_php_indent();
			return;
		case HT_PYTHON:
			html_exclude_script_head();
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			html_exclude_script_head();
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

int css_indent_and_show_matching_delimiter();

html_redirect_active_key()
{
	int region, start, end, f = 0, i;
	short *tab = 0;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	i = (1 << (region - 1 + 16));
	if (region > 0 && region <= HT_CSS && !(html_indenting_rules & i))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			save_var auto_show_matching_characters = c_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			tab = c_tab;
			break;
		case HT_VBSCRIPT:
			save_var auto_show_matching_characters = vbasic_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = vbasic_tab;
			break;
		case HT_PHP:
			save_var auto_show_matching_characters = perl_auto_show_delim_chars;
			save_var mode_move_level = perl_move_level;
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			tab = c_tab;
			break;
		case HT_PYTHON:
			save_var auto_show_matching_characters = python_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = python_tab;
			break;
		case HT_CSS:
			save_var auto_show_matching_characters = css_auto_show_delim_chars;
			save_var mode_move_level = 0;
			f = css_indent_and_show_matching_delimiter;
			break;
	}
	if (tab)
		f = binding_of(key, tab);
	if (!f && Matchdelim && strchr(")]}", key)) {
		save_var inside_show_matching_delimiter = 1;
		f = show_matching_delimiter;
	}
	if (f > 0) {
		save_var narrow_start = start, narrow_end = size() - end;
		html_exclude_script_head();
		do_command(f);
		return;
	}
	normal_character();
}

/* yank the current kill buffer into the current buffer */
command yank()
{
	int pos = point, oldsize = size();

	get_clipboard();
	if (typing_deletes_highlight && is_highlight_on() == REGRECT
		  && (typing_deletes_highlight != 2 || _doing_input)) {
		check_num_kill_buffers();
		if (_kill_width[_num_kill] > 0)
			fix_region();
	}
	delete_if_highlighted();
	retrieve(0);
	if (_doing_input == DI_LINEINPUT && insert_default_response == 3 
		  && pos == oldsize) { 
		undo_mainloop();        // Let undo add stuff back. 
		if ((yank_options & 1) && pos > 2 && is_path_separator(character(pos))
			  && is_path_separator(character(pos - 1))) 
			delete(pos - 1, pos);
		restart_file_paths(); 
	} 
}

int is_php_alt_open(int p)
{
	if (sublanguage_code != SUBLANG_PHP)
		return 0;
	save_var point = p;
	if (parse_string(-1, "</word>else[ \t]*:"))
		return 1;
	if (!parse_string(-1, "%)[ \t]*:"))
		return 0;
	return c_rev_skip_level('(', ')')
			&& parse_string(-1, "(if|elseif|for|foreach|while|switch)[ \t]+");
		
	return 1;
}

int is_php_alt_outdent(int p)
{
	if (sublanguage_code != SUBLANG_PHP)
		return 0;
	save_var point = p;
	return parse_string(1, "(else(if)?|end(if|while|for|foreach|switch))</word>");
}

perl_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point, i, j;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	save_var Closeback = perl_closeback;
	save_var Topindent = perl_topindent;
	save_var c_indent = perl_indent;
	save_var c_label_indent = perl_label_indent;
	save_var c_contin_offset = perl_contin_offset;
	save_var c_align_contin_lines = perl_align_contin_lines;
	save_var c_brace_offset = perl_brace_offset;
	save_var c_top_braces = perl_top_braces;
	save_var c_top_contin = perl_top_contin;
	save_var c_top_struct = perl_top_struct;

	if (sublanguage_code != SUBLANG_PHP)
		save_var narrow_start = 0;
	recolor_buffer_range(orig, orig + 1);
	if (get_tagged_region("color-perl-as-unit", orig, &i) != -1
		  || get_character_color(orig, &i) == color_class perl_constant
		  && matches_at(i, 1, "%<%<"))		// Keep original's indentation
		return get_indentation(orig);		// if in here doc.
	if (get_character_color(orig - 1, &i) == color_class perl_string)
		return get_indentation(orig);		// or in string.
	if (get_character_color(orig - 1, &i) == color_class perl_comment
		  && matches_at(i, 1, "="))
		return get_indentation(orig);		// or in POD.

	norm_narrow = fix_narrowing(point - c_look_back);
	narrow_start = norm_narrow;
	if (first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (sublanguage_code == SUBLANG_PHP		// Line following here doc.
		  && (get_tagged_region("color-perl-as-unit", give_end_line(), &i, &j) != -1
			  || get_character_color(orig, &i) == color_class perl_constant
			  && matches_at(i, 1, "%<%<"))) {
		point = j;		// We might have "EOF;" after the here doc.
		point += parse_string(1, "[ \t]*");
		prev_end = character((pend = point) - 1);
	}

	if (is_php_alt_outdent(orig))
		ind -= get_c_indent();
	else if (sublanguage_code == SUBLANG_PHP && is_case(orig))
		ind = indent_case(orig);
	else if (have_label(orig)) {
		ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{' || is_php_alt_open(pend)) {
		if (ind <= c_top_braces	// special case if no previous indent
			|| get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig))
		return indent_just_open(ind);
	else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, 0);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
			   && !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

command search_region()
{
	save_var search_in_region = 1;
	on_exit highlight_off();
	highlight_on();
	incremental_search();
}

when_loading()
{
	if (color_html_look_back < 500_000)
		color_html_look_back = 500_000;
	clipboard_convert_unicode |= 2;
	telnet_interpret_output |= 0x20;
	html_indenting_rules |= 0xff0000;
}
