/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2007 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.04 to 13.05.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "misclang.h"
#include "perl.h"
#include "bufed.h"
#include "tags.h"
#include "python.h"
#include "epsnet.h"
#include "proc.h"
#include "grep.h"
#include "html.h"
#include "vbasic.h"
#include "tex.h"
#include "gams.h"

scan_for_string()	// check current line to see if we're in
{			// string or new-style comment
	int c, old;

	save_var point;
	save_var narrow_end = size() - point;
	to_begin_line();		// scan this line before point
	while (parse_string(-1, ".*\\\n"))
		point = matchend;	// Go back past this continuation line.
	while (re_search(1, "<squote|dquote>|//|/<*>")) {
		old = point = matchstart;
		c = curchar();
		if (c == '/' && character(point + 1) == '/')  // found "//"
			return IN_NEW_COMMENT;
		else if (!c_skip_item(c, RE_FORWARD)) {
			if (c == '"' && (c_extra_keywords & CSHARP_KEYWORDS))
				return csharp_in_string();
			return c == '/' ? IN_OLD_COMMENT
							: c == '"' ? IN_STR_CONST : IN_CHAR_CONST;
		}
		if (point < old)		// Something's wrong, so give up.
			return 0;
	}
	return 0;
}

init_from_file()
{
	char *p, file[FNAMELEN];
	int res;
	jmp_buf this_level;

	save_var top_level = &this_level;	// Don't let errors abort,
	if (setjmp(top_level)) {		// so session file can still load.
		show_text(0, -1, "Error loading %f, see %s buffer.",
				  einit_file_name, MSG_BUFFER);
		return;
	}
	save_var _abort_load_on_error = 2;
	if ((load_customizations & 1) && (p = lookpath(einit_file_name))) {
		strcpy(file, p);
		show_text(-2, 0, "Loaded init file %s", file);
		save_var bufnum = zap("-customizations");
		do_file_read(file, FILETYPE_AUTO);
		ini_mode();
		res = parse_cmds();
		tab_size = tab_size.default;
		restore_vars();
		if (res)
			show_text(0, -1, "Error loading %f, see %s buffer.", file, MSG_BUFFER);
	}
}

// Return 1 if ok, 0 if error.
int dired_do_set_attributes(char *fname, char *dir)
{
	char flags[FNAMELEN], cmd[FNAMELEN], buf[FNAMELEN], msg[FNAMELEN];
	char abbrev[FNAMELEN];

	save_var _show_help_on = "dired-mode";
	strcpy(abbrev, fname);
	abbreviate_file_name(abbrev, screen_cols - 50);
	if (is_unix || is_remote_file(fname)) {
		sprintf(msg, is_remote_file(fname) ? "Octal mode for chmod command on %s"
										   : "Flags for chmod command on %s",
				abbrev);
		get_str_auto_def(flags, msg);
		if (!*flags)
			return 0;
		if (is_remote_file(fname)) {
			sprintf(cmd, "chmod %s %s", flags, get_url_file_part(fname, 0));
			ftp_misc_operation(dir, cmd);
			say("Operation queued.");
			return 0;	// Suppress refreshing listing.

		} else {
			sprintf(cmd, "chmod %s ", flags);
			quote_for_shell(cmd + strlen(cmd), fname);
			strcpy(buf, "-dired-set-attributes-result");
			pipe_text(NULL, buf, cmd, dir,
					  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
			if (buffer_size(buf)) {
				to_buffer(buf);
				discardable_buffer = 1;
				point = 0;
				return 0;
			}
		}
	} else {
		struct file_info fi;
		int attrs;

		if (!check_file(fname, &fi))
			error("Can't get attributes for %f", fname);
		attrs = fi.opsysattr;
		sprintf(cmd, "%s%s%s%s",
				(attrs & DOS_ATTR_RDONLY) ? "R" : "",
				(attrs & DOS_ATTR_SYSTEM) ? "S" : "",
				(attrs & DOS_ATTR_HIDDEN) ? "H" : "",
				(attrs & DOS_ATTR_ARCHIVE) ? "A" : "");
		sprintf(msg, "Set attributes (RO, Sys, Hid, Arch) for %s to: ", abbrev);
		get_strnone(flags, msg, cmd);
		save_var case_fold = 1;
		attrs = (strstr(flags, "R") ?  DOS_ATTR_RDONLY : 0)
				| (strstr(flags, "S") ? DOS_ATTR_SYSTEM : 0)
				| (strstr(flags, "H") ? DOS_ATTR_HIDDEN : 0)
				| (strstr(flags, "A") ? DOS_ATTR_ARCHIVE : 0);
		set_file_opsys_attribute(fname, attrs);
	}
	return 1;
}

net_during_wait(int cond)
{
	delay(1, cond);
	if (ungot_key == WIN_HELP_REQUEST && menu_command
		  && !strcmp(menu_command, "password-prompt")) {
		ungot_key = -1;
		password_prompt();
	}
}	

struct scp_job *do_scp(struct url_parts *parts, int op);

scp_get_response(int code, char *url, struct url_parts *parts, int b)
{
	char cmd[FNAMELEN], filter[FNAMELEN], *s, *val;
	struct scp_job *p;

	save_var bufnum = b;
	buf_zap(bufnum);
	p = do_scp(parts, FTP_MISC);
	strcpy(p->file, url);
	*get_url_file_part(p->file, 0) = 0;
	if (!(p->flags & SCP_CLIENT_MODE)) {
		switch (code) {
			case 1:
				s = "";
				break;
			case 2:
				p->flags |= SCP_CHECK_IS_DIR;
				s = "cd \"%s\"\n\n";
				break;
			default:		// Cases 0 and 3: list matches.
				p->flags |= SCP_CHECK_MATCHES;
				val = get_str_var(find_variable_by_host("scp-list-flags",
					parts->host));
				sprintf(s = cmd, "ls %s", (p->flags & SCP_OLD_SFTP) ? ""
					: val ? val : "-l");
				if (s[strlen(s) - 1] != ' ')
					strcat(s, " ");
				strcat(s, "%s");
				strcpy(filter, get_tail(parts->fname, 1));
				*get_tail(parts->fname, 1) = 0;	// Must list all files.
				break;
		}
		scp_send_some_command(parts, p, s);
	} else {
		sprintf(cmd, (code == 1 || code == 2) ? "prdir %s\n" : "comp %s\n",
				parts->fname);
		p->cmd = p->cmdlist = strsave(cmd);
		if (p->busy != 2)
			scp_next_cmd(p);
	}
	delayed_say(0, 100, 30, "Working...");
	while (p->busy && !user_abort)
		net_during_wait(COND_KEY);
	if ((p->flags & SCP_RESTART) && code == 2 && !(p->flags & SCP_CLIENT_MODE)) {
		p->flags &= ~SCP_RESTART;
		scp_restart_process(p);
		while (!user_abort && p->cmd && *p->cmd)
			net_during_wait(COND_KEY);
	}
	set_buf_mode_message(bufnum, NULL);
	user_abort = 0;
	delayed_say(0, 0, 0, "");
	say("");
	point = 0;
	if (!(scp_flags(parts) & SCP_CLIENT_MODE)) {
		switch (code) {
			case 0:
			case 3:		// Discard the extra info and filter nonmatches.
				scp_filter_sftp_matches(filter, code != 3, url);
				break;
			case 1:
				buf_printf(b, "%s%s", p->file, p->lastdir);
				set_buf_point(b, 0);
				break;
		}
	}
}

char *remote_file_match(char *s, int flags)
{
	struct url_parts parts;
	int b;

	if (parse_url(s, &parts) && (!strcmp(parts.service, "scp")
								 || !strcmp(parts.service, "ftp"))) {
		b = create(_MATCH_BUF);
		if (flags & STARTMATCH) {
			if (!strcmp(parts.service, "scp"))
				scp_get_response((flags & EXACTONLY) ? 3 : 0, s, &parts, b);
			else
				ftp_file_match(s, flags);
			if (flags & FM_NO_DIRS) {
				save_var bufnum = b;
				simple_re_replace(0, ".*[/\\]\n", "");
				restore_vars();
			}
		}
		if (!grab_line(b, _this_match))
			return 0;
		return _this_match;
	}
	return NULL;
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			point = type_point;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 0x20) {
				point = from;
				simple_re_replace(0, "<#0-#1>", "");
			}
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

command insert_date()
{
	char time[FNAMELEN];

	current_time_and_date(time);
	stuff(time);
}

check_some_vi_filevars()
{
	int res, had_set;
	save_var point = 0;
	while (re_search(1, "([ \t\n](vi|vim)|[ \t]ex):[ \t]*(set? )?")) {
		had_set = parse_string(-1, "set? ");
		do {		// Found a settings line, apply all settings.
			if (check_one_vi_filevar("(textwidth|tw)=([0-9]+)", 2, &res))
				margin_right = res;
			if (check_one_vi_filevar("(tabstop|ts)=([0-9]+)", 2, &res))
				tab_size = res;
			if (check_one_vi_filevar("(shiftwidth|sw|softttabstop|sts)"
									 "=([0-9]+)", 2, &res))
				soft_tab_size = res;
			if (check_one_vi_filevar("no(expandtab|et)", -1, &res)
				  || check_one_vi_filevar("(expandtab|et)", 0, &res))
				indent_with_tabs = res;	// Set to 1 or 0.
			point += res = parse_string(1, "[^ \n:]*[ :]");
			if (had_set && character(point - 1) == ':')
				break;		// A set command ends with next :.
		} while (res > 0);	// To next setting on this line.
	}
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(php)")) {
		sprintf(res, pat, "php");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "(sh|csh|bash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

int follow_when_idle(int cnt)
{
	int did_any = 0, i;
	if (follow_mode_on == 1 && !cnt) {
		follow_mode_on++;		// Lock out if error.
		int buf = bufnum, win = window_number, start = window_start;
		while (window_number > 0 && --window_number < win
			   && bufnum == buf && is_window(window_handle) == ISTILED) {
			did_any = 1;
			int old = point;
			point = start;
			i = MAX(text_height() - follow_mode_overlap, 0);
			point = window_start = prev_screen_line(i);
			if (point < old && old < start)
				point = old;
			start = window_start;
		}
		window_number = win;
		start = window_end + 1;
		if (follow_mode_overlap) {
			save_var point = start;
			start = prev_screen_line(MIN(text_height() - 1, follow_mode_overlap));
		}
		while (++window_number > win && is_window(window_handle) == ISTILED
			   && bufnum == buf) {
			did_any = 1;
			int old = point;
			point = window_start = start;
			i = MAX(text_height()- follow_mode_overlap, 0);
			start = next_screen_line(i);
			if (window_start < old && old < start)
				point = old;
		}
		window_number = win;
		modify_region(SHOW_MATCHING_REGION, MRCONTROL, 0);
		follow_mode_on--;
	}
	restore_vars();
	if (did_any)
		refresh();
	return -1;
}

#define TEX_PARAG_SEP	"^\\(begin|end)|^[ \t]*(%%|%$%$|\\[][])"

command forward_paragraph()
{
	char pat[100];

	do_shift_selects();
	re_search(1, ".");
	sprintf(pat, "\n%s|\f",
			indents_separate_paragraphs ? "[ \t\n]" : "[ \t]*\n");
	if (tex_paragraphs) {	/* skip formatter lines */
		to_begin_line();
		re_search(1, "(([@.]|" TEX_PARAG_SEP ").*\n)*");
		re_search(1, "[^ \t\n@.\\]");
		strcat(pat, "|" TEX_PARAG_SEP "|\\\\\n");
	}
	if (re_search(1, pat)) {
		point--;
		if (!strchr("\f\n", curchar())) {
			if (!indents_separate_paragraphs)
				re_search(-1, "[ \t]*");
			if (tex_paragraphs)
				re_search(-1, "^(([@.]|" TEX_PARAG_SEP ").*\n)*");
		}
	} else
		point = size();
}

command backward_paragraph()
{
	do_shift_selects();
	if (tex_paragraphs) {
		if (character(point - 1) == '\n')
			re_search(-1, "^(([@.]|" TEX_PARAG_SEP ").*\n)*");
		re_search(-1, "[^ \t\n]");
		if (matches_at(give_begin_line(), 1, TEX_PARAG_SEP))
			to_begin_line();
		if (indents_separate_paragraphs) {
			if (re_search(-1, "\n[ \t\n]|\f|" TEX_PARAG_SEP "|\\\\\n"))
				re_search(1, "(\\\\)?\n*");
		} else
			re_search(-1, "(\n[ \t]*\n|\f|" TEX_PARAG_SEP "|\\\\\n)!");
		if (curchar() != '\f')
			re_search(1, "^(([@.\f]|" TEX_PARAG_SEP ").*\n)*");
	} else {
		re_search(-1, "[^ \t\n]");
		if (indents_separate_paragraphs) {
			if (re_search(-1, "\n[ \t\n]|\f|\n(" TEX_PARAG_SEP ")"))
				re_search(1, "\n*");
		} else
			re_search(-1, "((\n|<bob>)[ \t]*\n|\f)!");
	}
}

suffix_inc()
{
	char ext[FNAMELEN];
	if (gams_files)
		gams_mode();
	else if (!guess_mode_without_extension(ext, "suffix-%s") || !try_calling(ext))
		asm_mode();
}

color_embedded_code(int from, int to, int lang, int limit)
{
	save_var point;
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	set_tagged_region("color-as-unit", from, to, lang);
	if (limit > 0 && to > limit + 1000 && lang == HT_PHP
		  && count_windows_with_buf(bufnum, 0) <= 1) {
		set_tagged_region("html-coloring-postponed", limit, to, 1);
		to = limit;		// Note that we only colored part of this.
	}
	if (lang == HT_JSCRIPT) {
		save_var c_extra_keywords |= JS_KEYWORDS;
		save_var recolor_range = color_c_range;
		save_var recolor_from_here = color_c_from_here;
		coloring_flags |= COLOR_INVALIDATE_FORWARD;
		// Invalidating flag fixes issue where deleting a </script>
		// doesn't recolor rest of block as scripting.
	} else if (lang == HT_VBSCRIPT) {
		save_var recolor_range = color_vbasic_range;
		save_var recolor_from_here = recolor_by_lines;
		save_var sublanguage_code = SUBLANG_VBSCRIPT;
	} else if (lang == HT_PHP) {
		save_var recolor_range = color_full_perl_range;
		save_var recolor_from_here = recolor_from_top;
		save_var sublanguage_code = SUBLANG_PHP;
		// Any buffer with embedded PHP must invalidate coloring after
		// edits, as in Perl mode.
		coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_RESETS;
	} else if (lang == HT_PYTHON) {
		save_var recolor_range = color_python_range;
		save_var recolor_from_here = recolor_from_top;
	} else if (lang == HT_CSS) {
		save_var recolor_range = color_css_range;
		save_var recolor_from_here = recolor_from_top;
	} else {
		set_character_color(from, to, color_class html_attribute);
		return;
	}
	color_for_narrowed_area(from, to);
}

html_find_linked_file(char *file)
{
	char buf[FNAMELEN], *p;
	int orig;

	if (!re_search(1, "((href|src)[ \t\n]*=|@import)[ \t\n]*")) {
		to_begin_line();
		if (!re_search(1, "((href|src)[ \t\n]*=|@import)[ \t\n]*"))
			error("Can't find any link here.");
	}
	orig = point;
	if (curchar() == '"') {
		point++, orig++;
		search(1, "\"");
	} else if (curchar() == '\'') {
		point++, orig++;
		search(1, "'");
	} else	// No quotes, assume it's whitespace-delimited.
		re_search(1, "[^ \t\n<>]*"), point++;
	grab(orig, point - 1, buf);		// Get link from buffer.
	p = strchr(buf, '#');		// Remove #fragment if any.
	if (p)
		*p = 0;
	if (!*buf)
		error("Can't follow fragment links.");
	point = orig;
	if (parse_string(1, "[A-Za-z]+:"))
		strcpy(file, buf);					// An absolute URL.
	else if (parse_string(1, "/")) {		// An absolute URL path.
		strcpy(file, filename);
		strcpy(get_url_file_part(file, 1), buf);
	} else {								// A relative URL path.
		strcpy(file, filename);
		strcpy(get_tail(file, 1), buf);
	}
}

do_html_indent(int manual)
{
	int region, start;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 1);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	if (region > 0 && region <= HT_CSS && !IS_ENTER_KEY(key)
		  && !(html_indenting_rules & (1 << (region - 1)))) {
		prev_indenter();	// Can move back to prev line's indent.
		return;
	}
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
		case HT_PHP:
			html_exclude_script_head();
			do_php_indent();
			return;
		case HT_PYTHON:
			html_exclude_script_head();
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			html_exclude_script_head();
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

int css_indent_and_show_matching_delimiter();

html_redirect_active_key()
{
	int region, start, end, f = 0, i;
	short *tab = 0;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	i = (1 << (region - 1 + 16));
	if (region > 0 && region <= HT_CSS && !(html_indenting_rules & i))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			save_var auto_show_matching_characters = c_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			tab = c_tab;
			break;
		case HT_VBSCRIPT:
			save_var auto_show_matching_characters = vbasic_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = vbasic_tab;
			break;
		case HT_PHP:
			save_var auto_show_matching_characters = perl_auto_show_delim_chars;
			save_var mode_move_level = perl_move_level;
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			save_var perl_indent = php_indent;
			save_var perl_closeback = php_closeback;
			save_var perl_topindent = php_topindent;
			save_var perl_label_indent = php_label_indent;
			save_var perl_contin_offset = php_contin_offset;
			save_var perl_align_contin_lines = php_align_contin_lines;
			save_var perl_brace_offset = php_brace_offset;
			save_var perl_top_braces = php_top_braces;
			save_var perl_top_contin = php_top_contin;
			save_var perl_top_struct = php_top_struct;
			tab = c_tab;
			break;
		case HT_PYTHON:
			save_var auto_show_matching_characters = python_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = python_tab;
			break;
		case HT_CSS:
			save_var auto_show_matching_characters = css_auto_show_delim_chars;
			save_var mode_move_level = 0;
			f = css_indent_and_show_matching_delimiter;
			break;
	}
	if (tab)
		f = binding_of(key, tab);
	if (!f && Matchdelim && strchr(")]}", key)) {
		save_var inside_show_matching_delimiter = 1;
		f = show_matching_delimiter;
	}
	if (f > 0) {
		save_var narrow_start = start, narrow_end = size() - end;
		html_exclude_script_head();
		do_command(f);
		return;
	}
	normal_character();
}

html_func_name_finder()
{
	int start, end, region, res = 1;
	*display_func_name = 0;
	if (sublanguage_code == SUBLANG_CSS)
		return 1;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	save_var narrow_start = start, narrow_end = size() - end, bufnum;
	if (region <= 0 && html_display_nesting_width >= 5
		  && (html_display_definition
			  & ((sublanguage_code == SUBLANG_XML) ? 4 : 8))) {
		int b = tmp_buf();
		narrow_start = narrow_end = 0;
		save_var point = give_begin_line();
		do_list_element_nesting(b, must_find_func_name ? 0 : 2);
		html_trim_element_nesting(b);
		restore_vars();
		buf_delete(b);
		return 1;
	}
	if (region > 0 && (html_display_definition & 3)) {
		html_exclude_script_head();
		switch (region) {
			case HT_JSCRIPT:
				save_var mode_move_level = c_move_level;
				if (html_display_definition & 1)
					res = javascript_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "JavaScript");
				return res;
			case HT_VBSCRIPT:
				save_var mode_move_level = 0;
				if (html_display_definition & 1)
					res = vbasic_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "VBScript");
				return res;
			case HT_PHP:
				save_var mode_move_level = perl_move_level;
				if (html_display_definition & 1)
					res = inner_php_func_name_finder();
				if ((html_display_definition & 2) && res
					  && !*display_func_name && strcmp(major_mode, "PHP"))
					strcpy(display_func_name, "PHP");
				return res;
			case HT_PYTHON:
				save_var mode_move_level = 0;
				if (html_display_definition & 1)
					res = python_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "Python");
				return res;
			case HT_CSS:
				if (html_display_definition & 2)
					strcpy(display_func_name, "CSS");
				return res;
		}
	}
	return 1;
}

// First, define some regex pieces that could be helpful in multiple
// definitions.

// These file name matchers say, optionally a drive spec c:, then a
// basename, a dot, and an extension.  The OK_SPACE version permits
// spaces in the basename, starting with the 2nd character.  Both permit
// tags like <3> at the end of the extension.
#define FNAME_NO_SPACE	"(([a-z]:)?[+-9A-z~]+<.>[+-9A-z]+[0-9<>]*)"
#define FNAME_OK_SPACE	"(([a-z]:)?[+-9A-z~][+-9A-z~ ]*<.>[+-9A-z]+[0-9<>]*)"

// Some things we want to ignore in an error message.

// 3> or E " or W " or ***, for example.
#define START_STUFF1	"[1-9]>|e \"|w \"|<*>*"
// One of these three keywords, or [someother].
#define START_STUFF2	"error|warning|fatal| *%[[a-z]+%]"
// Optionally, a code like E123 with a space before it, then perhaps
// some punc chars * : after.  Or just punc chars " or (.
#define START_STUFF3	"( [a-z][0-9]+)?[*:]*|[\"(]?"

// Between a file name and line number, we must have space, open paren,
// comma or colon, then perhaps L or the word "line" or a [, then
// perhaps space.  Also skip over any " at the start, which may have
// surrounded file name.
#define BEFORE_NUM		"\"?([ (:,]( *line|L|%[)?) *"
// The line number.
#define LINE_NUM		"([0-9]+)"
// Stuff to require after line number.  Either ) or colon, or else some
// stuff with colon or semicolon after it, or else a column number and ].
#define END_STUFF		"(<rparen|:>|[^0-9\\\n].*[:;]|,[0-9]+%][ \n]*)"

/* If you change ERROR_PATTERN, you may have to change the next 3 macros
 * as well. */
#define ERROR_PATTERN	"^(" START_STUFF1 \
						"(" START_STUFF2 "|[a-z0-9]+ )" \
						START_STUFF3 "|[ \t]*from) *" \
						FNAME_NO_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN	4	/* filename is in 4th () pair above */
#define LINE_PAREN	8	/* line# is in 8th pair of parens above */

#define SIMP_PATTERN	"[:;)]| line "	/* search with simple pattern first */

// This version permits spaces in file names, but with more
// restrictions on ancillary text before it.
#define ERROR_PATTERN2	"^(" START_STUFF1 \
						"(" START_STUFF2 ")" \
						"( [a-z][0-9]+)?[*:]*|[\"(]?) *" \
						FNAME_OK_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN2	4	/* filename is in 4th () pair above */
#define LINE_PAREN2	8	/* line# is in 8th pair of parens above */

#define ERROR_PATTERN3	".* at (.+) line ([0-9]+)[.,].*"
#define FILE_PAREN3	1	/* filename in () pair above */
#define LINE_PAREN3	2	/* line# in pair of parens above */

// An error pattern that permits file names with no . character, but
// with more restrictions on the format.
#define ERROR_PATTERN4	"^([+-9A-z~:]+[0-9<>]*)" BEFORE_NUM \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN4	1	/* filename is in 1st () pair above */
#define LINE_PAREN4	2	/* line# is in 2nd pair of parens above */

// An error pattern for Java runtime errors.

#define ERROR_PATTERN5	"^.*%(([+-9A-z~:]+):([0-9]+)%)"
#define FILE_PAREN5	1	/* filename is in 1st () pair above */
#define LINE_PAREN5	2	/* line# is in 2nd pair of parens above */

// An error pattern that permits file names with no . character,
// allowing spaces, but with even more restrictions on the rest of the
// message.
#define ERROR_PATTERN6	"^([a-z]:\\[ +-9A-z~]+)[(]" \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN6	1	/* filename is in 1st () pair above */
#define LINE_PAREN6	2	/* line# is in 2nd pair of parens above */

// You can define ERROR_PATTERN7 with another alternate pattern.
// Epsilon will look for either one, on lines that contain SIMP_PATTERN.
// Define FILE_PAREN7 and LINE_PAREN7 to match.  (Also ERROR_PATTERN8
// etc.)  You can also define an error pattern where the PAREN codes
// refer to an empty set of parens, for error messages that don't
// include a file name; Epsilon will jump to the process buffer line
// reporting the error for these.

do_next_error(int count, char *cmd)	// return 0 if no errors, else 1
{
	char file[FNAMELEN], number[30], line[FNAMELEN], *p;
	int lineno, origpt, dir, from, to, is_bufname, file_paren, line_paren;
	int orig = bufnum, procbuf, start;

	tiled_only();
	_last_was_grep = 0;
	if (spot_to_buffer(error_spot) <= 0 && !exist(PROCBUF))
		error("No process buffer to read errors from.");
	tiled_only();
	if (spot_to_buffer(error_spot) > 0)
		bufnum = spot_to_buffer(error_spot);
	else
		bufname = PROCBUF;
	telnet_error_set_to_telnet();
	procbuf = bufnum;
	origpt = point;
	case_fold = 1;
	if (spot_to_buffer(error_spot) > 0)
		point = *error_spot;
	else {
		error_spot = alloc_spot();
		point = 0;
	}
	if (!count && procbuf == orig)
		point = origpt;

	noteput("Searching...");
	if (!count)
		find_error(-1, &file_paren, &line_paren), count = 1;
	dir = (count > 0) ? 1 : -1;
	for (; count; count -= dir) {
		if (!find_error(dir, &file_paren, &line_paren)) {
			point = origpt;
			bufnum = orig;
			if (_had_errors == 2)
				return 0;
			else if (_had_errors)
				say("No more errors.");
			else if (cmd)
				say("No errors running \"%s\".", cmd);
			else
				say("No errors.");
			return 0;
		}
	}
	_had_errors = 1;
	grab(find_group(file_paren, 1), find_group(file_paren, 0), file);
	grab(find_group(line_paren, 1), find_group(line_paren, 0), number);
	start = find_group(0, 1);
	from = find_group(line_paren, 0);
	to = find_group(0, 0);
	if (to - from < 20) {	// error message at end is too short
		from = start;	// so try including whole line
		if (to - from > screen_cols - 5) // Only rightmost if it's too big.
			from = to - screen_cols + 5;
	}
	if (to > from + sizeof(line) / sizeof(*line) - 1)
		to = from + sizeof(line) / sizeof(*line) - 1;
	grab(from, to, line);
	is_bufname = (parse_string(1, "Buffer ") != 0);
	if (dir > 0)
		to_end_line();
	*error_spot = point;
	point = origpt;
	lineno = strtoi(number, 10);
	bufnum = orig;
	if (!*file || !*number) {
		locate_window(bufnum_to_name(procbuf), "");
		point = *error_spot;
		to_begin_line();
		say("=> %s", line);
		return 1;
	}
	if (is_bufname)
		locate_window(file, "");
	else {
		process_find_directory(file, procbuf, from);
		if (!check_file(file) && !strcmp(get_extension(file), ".java"))
			proc_find_java_source(file, procbuf, start);
		locate_window("", file);
		if (process_view_error_lines
			  && window_number == number_of_windows() - 1)
			window_number = 0;
		std_find_it(file);
	}
	go_line(lineno);
	if (is_remote_file(file))
		ftp_go_to_line = lineno;
	for (p = line; *p && index("()[] :;,\t\n0123456789", *p); p++)
		;			// skip over these chars at start of msg
	say("=> %s", p);
	restore_vars();
	proc_view_error_lines(procbuf, start);
	return 1;
}

/* Find next error in direction, return 1 and set vars to find_group() paren count */
find_error(int dir, int *file_paren, int *line_paren)
{
	int old = point;

	if (point > 0 && dir > 0)
		to_end_line();
	else
		to_begin_line();
	while (re_search(dir, SIMP_PATTERN)) {
		to_begin_line();
		if ((!*ignore_error || !parse_string(1, ignore_error, NULL))
			  && !parse_string(1, "In file included from ", NULL)) {
			if (parse_string(1, ERROR_PATTERN, (char *)0)) {
				*file_paren = FILE_PAREN;
				*line_paren = LINE_PAREN;
				return 1;
			}
#ifdef ERROR_PATTERN2
			if (parse_string(1, ERROR_PATTERN2, (char *)0)) {
				*file_paren = FILE_PAREN2;
				*line_paren = LINE_PAREN2;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN3
			if (parse_string(1, ERROR_PATTERN3, (char *)0)) {
				*file_paren = FILE_PAREN3;
				*line_paren = LINE_PAREN3;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN4
			if (parse_string(1, ERROR_PATTERN4, (char *)0)) {
				*file_paren = FILE_PAREN4;
				*line_paren = LINE_PAREN4;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN5
			if (parse_string(1, ERROR_PATTERN5, (char *)0)) {
				*file_paren = FILE_PAREN5;
				*line_paren = LINE_PAREN5;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN6
			if (parse_string(1, ERROR_PATTERN6, (char *)0)) {
				*file_paren = FILE_PAREN6;
				*line_paren = LINE_PAREN6;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN7
			if (parse_string(1, ERROR_PATTERN7, (char *)0)) {
				*file_paren = FILE_PAREN7;
				*line_paren = LINE_PAREN7;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN8
			if (parse_string(1, ERROR_PATTERN8, (char *)0)) {
				*file_paren = FILE_PAREN8;
				*line_paren = LINE_PAREN8;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN9
			if (parse_string(1, ERROR_PATTERN9, (char *)0)) {
				*file_paren = FILE_PAREN9;
				*line_paren = LINE_PAREN9;
				return 1;
			}
#endif
		}
		if (dir > 0)
			nl_forward();
	}
	*error_spot = old;
	return 0;
}

int perl_skip_level(int dir, char left_d, char right_d)
{
	char pattern[40], ch;
	int level = 0, pos;

	sprintf(pattern, "[%c%c]", right_d, left_d);
	do {
		if (!re_search(dir, pattern))
			return 0;
		pos = point - (dir > 0);
		if (in_perl_comment(pos))
			continue;
		if (character(pos - 1) == '$' && get_character_syntax_color(pos - 1)
			  == color_class perl_variable && !strchr("{}", character(pos)))
			continue;		// Actually a $[ or similar.
		ch = character(pos);
		if (ch == right_d)
			level++;
		else if (ch == left_d)
			level--;
	} while ((dir > 0) ? (level < 0) : (level > 0));
	return 1;
}

update_mode_more_extra(int offset)
{
	char *s;
	int cnt = 0;

	for (s = buf_match("compilation-output", STARTMATCH); s
			 ; s = buf_match("compilation-output", 0)) {
		if (is_process_buffer(name_to_bufnum(s)))
			  cnt++;
	}
	cnt += offset;
	if (cnt < 0)
		cnt = 0;
	switch (cnt) {
		case 0:		*mode_more_extra = 0; break;
		case 1:		strcpy(mode_more_extra, "[Compiling] "); break;
		default:	sprintf(mode_more_extra, "[Compiling %d] ", cnt); break;
	}
	all_must_build_mode = 1;
	maybe_refresh();
}

color_tex_dollar(int s)
{
	char pat[30];
	int orig, color = color_class tex_math;

	strcpy(pat, "\n[ \t]*\n|\\par![^a-z]|[\\%]|%$");
	if (curchar() == '$') {	// A $$ that signals displayed math?
		point++;			// Get to after $$.
		strcat(pat, "%$");
		color = color_class tex_display_math;
	}
	orig = point;
	while (re_search(1, pat)) {
		if (character(point - 1) == '\\') {	// skip quoted ones
			if (strchr("$%", curchar()))
				point++;
		} else if (character(point - 1) == '$') {
			color_tex_math(s, point, color);
			break;
		} else if (character(point - 1) == '%') {
			color_tex_math(s, point - 1, color);
			set_character_color(point, give_end_line(), color_class tex_comment);
			to_end_line();
			s = point;
		} else {
			point = orig;
			break;
		}
	}
}

is_vbasic_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	save_var case_fold = 1;

	if (sublanguage_code == SUBLANG_VBNET) {
		if (strstr("|#const|#else|#elseif|#externalsource|#if|#then|addhandler"
				   "|addressof|alias|and|andalso|as|boolean|byref|byte|byval"
				   "|call|case|catch|cbool|cbyte|cchar|cdate|cdbl|cdec|char"
				   "|cint|class|clng|cobj|const|cshort|csng|cstr|ctype|date"
				   "|decimal|declare|default|delegate|dim|directcast|do|double"
				   "|each|else|elseif|end|enum|erase|error|event|exit|false"
				   "|finally|for|friend|function|get|gettype|global|goto"
				   "|handles|if|implements|imports|in"
				   "|inherits|integer|interface|is|lib|like|long|loop|me|mod"
				   "|module|mustinherit|mustoverride|mybase|myclass|namespace|new"
				   "|next|not|nothing|notinheritable|notoverridable|object|on"
				   "|option|optional|or|orelse|overloads|overridable|overrides"
				   "|paramarray|private|property|protected|public"
				   "|raiseevent|readonly|redim|rem|removehandler|resume|return"
				   "|select|set|shadows|shared|short|single|static|step|stop"
				   "|string|structure|sub|synclock|then|throw|to|true|try|typeof"
				   "|when|while|with|withevents|writeonly"
				   "|xor|widening|narrowing|operator|datetime|int16|int32|int64"
				   "|sbyte|uinteger|uint32|ulong|uint64|ushort|uint16", p))
			return 1;
		if ((vbasic_language_level & 0x40)	// Unreserved VB.NET keywords.
			  && strstr("|#region|#externalsource|ansi|assembly|auto|binary"
						"|compare|custom|explicit|isfalse|istrue|mid|off"
						"|preserve|strict|text|unicode|until|", p))
			return 1;

		if ((vbasic_language_level & 0x80)	// Obsolete VB.NET keywords.
			  && strstr("|endif|gosub|wend|let|variant|", p))
			return 1;

		if ((vbasic_language_level & 2)		// Added in VB 2005.
			  && strstr("|continue|csbyte|cuint|culng|cushort|isnot|of|partial"
						"|trycast|using|", p))
			return 1;
		if ((vbasic_language_level & 4)		// Added in VB 2008.
			  && strstr("|aggregate|by|distinct|from|group|join|let|order"
						"|skip|take|where|", p))
			return 1;
		return 0;
	}
	if (strstr("|abs|array|asc|ascb|ascw|atn|cbool|cbyte|ccur|cdate|cdbl"
			   "|cdec|choose|chr|chr$|chrb|chrb$|chrw|cint|clng|command"
			   "|command$|cos|createobject|csng|cstr|curdir|curdir$|cvar"
			   "|cvdate|cverr|date|date$|dateadd|datediff|datepart"
			   "|dateserial|datevalue|day|ddb|dir|dir$|doevents|environ"
			   "|environ$|eof|error|error$|exp|fileattr|filedatetime|filelen"
			   "|fix|format|format$|freefile|fv|getallsettings|getattr"
			   "|getobject|getsetting|hex|hex$|hour|iif|imestatus|input"
			   "|input$|inputb|inputb$|inputbox|instr|instrb|int|ipmt|irr"
			   "|isarray|isdate|isempty|iserror|ismissing|isnull|isnumeric"
			   "|isobject|lbound|lcase|lcase$|left|left$|leftb|leftb$|len"
			   "|lenb|loc|lof|log|long|ltrim|ltrim$|mid|mid$|midb|midb$|minute"
			   "|mirr|month|msgbox|now|nper|npv|oct|oct$|partition|pmt|ppmt"
			   "|pv|qbcolor|rate|rgb|right|right$|rightb|rightb$|rnd|rtrim"
			   "|rtrim$|second|seek|sgn|shell|sin|sln|space|space$|spc|sqr"
			   "|str|str$|strcomp|strconv|string|string$|switch|syd|tab|tan"
			   "|time|time$|timer|timeserial|timevalue|trim|trim$|typename"
			   "|ubound|ucase|ucase$|val|vartype|weekday|year|", p))
			return 2;		// Function
	if (strstr("|#const|#else|#elseif|#end|if|#if|access|alias|any"
			   "|appactivate|append|as|base|beep|binary|byref|byval|call"
			   "|case|cdecl|chdir|chdrive|close|compare|const|date|declare"
			   "|defbool|defbyte|defcur|defdate|defdbl|defdec|defint|deflng"
			   "|defobj|defsng|defstr|defvar|deletesetting|dim|do|each|else"
			   "|elseif|empty|end|endif|erase|error|exit|explicit|false"
			   "|filecopy|for|function|get|global|gosub|goto|if|in|input|is"
			   "|kill|len|let|lib|line|load|local|lock|loop|lset|me|mid|midb"
			   "|mkdir|module|name|new|next|nothing|null|on|open|option"
			   "|optional|output|paramarray|preserve|print|private|property"
			   "|public|put|random|randomize|read|redim|rem|reset|resume"
			   "|return|rmdir|rset|savesetting|seek|select|sendkeys|set"
			   "|setattr|shared|static|step|stop|string|sub|text|then|time"
			   "|to|true|type|typeof|unload|unlock|until|wend|while|width"
			   "|with|withevents|write|", p))
			return 1;	// Statement or keyword.
	if (sublanguage_code == SUBLANG_VBSCRIPT &&
		  strstr("|class|eval|execute|scriptengine|", p))
		return 3;		// VBScript-specific
	return 0;
}
