/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2007 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.05 to 13.06.

#include "eel.h"
#include "c.h"
#include "session.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "proc.h"

int exptoi(char *s)
{
	save_var use_default = 0;	// Don't run compile with use_default 1.
	while (isspace(*s))
		s++;
	if (!strchr("<", *s) && (strpbrk(s, "+/*^|&<>%=")
							  || *s && strpbrk(s + 1, "-") || isalpha(*s))
		  && (has_feature & FEAT_EEL_COMPILE))
		return evaluate_numeric_expression(s);
	return numtoi(s);
}

/* Indent continuation line. */
indent_c_contin(ind, orig, lastl, in_initializer)
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0, slevel = 0;
	int objc_sel = 0, objc_indent = -1;

 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	if (!c_align_contin_lines)	// never line up
		if (!in_initializer)
			return c_default_indent(ind, orig, 1);
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	if ((c_extra_keywords & OBJC_KEYWORDS) && c_align_selectors
		  && matches_at(give_begin_line(), 1, "^[ \t]*(<word>+):"))
		objc_sel = find_group(1, 0) - find_group(1, 1);
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[][()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0 || objc_indent >= 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (objc_indent >= 0)
					break;
				if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind, orig);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				if (objc_sel && !plevel && !slevel
					  && parse_string(-1, "(^|<word>|[]])[ \t]*<word>+")
					  && current_column() <= c_align_selectors)
					objc_indent = current_column() - objc_sel;
				else
					qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				if (objc_indent >= 0)
					return objc_indent;
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '[':
				if (slevel-- > 0 || objc_indent >= 0)
					break;
				if (findc != ']')
					point++;
				save_var c_align_extra_space = 0;
				return limit_c_contin_indent(ind, orig);
			case ']':
				slevel++;
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)) || objc_indent >= 0)
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind, orig);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (objc_indent >= 0)
		return objc_indent;
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		if (curchar() == '#')
			past_initial_preprocessor();
		return limit_c_contin_indent(ind, orig);
	}
	return c_default_indent(ind, orig, 2);
}

limit_c_contin_indent(ind, orig)	// if lining up under something would put us
{				// too far to right, just use std indent
	int std, i = current_column();	// get lining-up indent

	if (c_align_contin_max_width == -1)
		save_var c_align_contin_max_width = text_width();
	if (i > c_align_contin_lines) {
		std = ind + get_c_indent() + c_contin_offset;
		i = MIN(std, i);
	} else if (c_align_contin_max_width > 0
			   && i + text_width_after_indent(orig) > c_align_contin_max_width) {
		std = ind + c_align_contin_max_offset;
		if (c_align_contin_max_offset <= 0)
			std = c_align_contin_max_width - text_width_after_indent(orig)
				  + c_align_contin_max_offset;
		i = MIN(std, i);
	} else if (i == ind + get_c_indent())	// The same as a plain indent
		i += c_align_extra_space;			// so fiddle with it slightly.
	return i;
}

fix_drive_only_file_names(int i)
{
	char fname[FNAMELEN];
	int all_files = 0;
	for (; i < argc; i++) {
		if (argv[i][0] == '+') {
		} else if (argv[i][0] == '-' && !all_files) {
			if (argv[i][1] == '-')
				all_files = 1;
			else if (!strcmp(argv[i], "-dir")
					 || strchr("ldr", argv[i][1]) && !argv[i][2]
					 || strchr("f", argv[i][1]) && !argv[i][3])
				i++;
		} else if (isalpha(argv[i][0]) && argv[i][1] == ':' && argv[i][2]
				   && !is_path_separator(argv[i][2])) {
			strcpy(fname, argv[i]);
			convert_cygwin_filename(fname, 0);
			absolute(fname);
			argv[i] = strsave(fname);
		}
	}
}

when_restoring()
{
	int tflag = 0, i, splitmode = HORIZONTAL, have_files = 0;
	char *flag, *arg, init_dir[FNAMELEN], *dir = init_dir;

	getcd(init_dir);
	early_restoring();
	early_init();
	set_toolbar();

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			have_files = 1;
			continue;
		}
		if (!strcmp(argv[i], "--"))
			break;
		if (!stricmp(argv[i], "-noinit")) {
			noinit_flag = 1;
			continue;
		} else if (!strcmp(argv[i], "-dir")) {
			if (++i < argc)
				dir = argv[i];
			continue;
		}
		flag = argv[i];
		switch (flag[1]) {
		case 'f':
			arg = (flag[2] && flag[3] || i + 1 >= argc) ? (flag + 3) : argv[++i];
			switch (flag[2]) {
			case 's':
				set_swapname(arg);
				break;
			case 'd':
				strcpy(_docname, arg);
				break;
			default:
				bad_flag(flag);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'd':
		case 'r':
		case 'l': if (!flag[2] && i + 1 < argc)
					  i++;
				  break;
		case 'p': break;
		case 'h': case '?': usage(); break;
		case 'v':
			switch (flag[2]) {
				case 'v':
					splitmode = VERTICAL;
					break;
				default:
					bad_flag(flag);
			}
			break;
		case '-':
			if (!strcmp(flag + 2, "help"))
				usage();
			else
				bad_flag(flag);
			break;
		default:
			bad_flag(flag);
		}			/* i is 1st file argument */
	}
	if (i < argc)
		have_files = 1;
	term_init();
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	delete_buffer("-customizations");	// Don't record any settings before now.
	parse_command_line_flags(1);
	fix_drive_only_file_names(1);
	if (!noinit_flag && load_from_state)
		init_from_file();
	parse_command_line_flags(2);
	if (tflag) {
		preserve_session_once = 1;
		tutorial();
	} else {
		if (load_from_state && (!have_files || session_always_restore)) {
			if (session_restore_files == 2 && have_files)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (have_files)
			load_cmd_line_files(1, splitmode, dir);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	restore_check_version();
	show_text(0, 1, "Epsilon Programmer's Editor %s, EEL version %s",
			  version, eel_version);
	try_calling("start-up");
	parse_command_line_flags(3);
	if (!noinit_flag && load_from_state)
		load_version_patch_file();
}

dired_refresh() on dired_tab['r']
{
	char msg[FNAMELEN];
	int yes, remote;

	if (modified) {
		sprintf(msg, "Relist %s, discarding markings? ", bufname);
		yes = ask_yn("Discard markings", msg, "Yes", "No", 1);
		check_abort();
		if (!yes)
			return;
	}
	remote = is_remote_file(get_dired_buffer_pattern());
	if (remote)
		point = 0;
	else
		save_var point, window_start;
	save_var file_pattern_wildcards = dired_buffer_pattern_wildcards;
	create_dired_listing(get_dired_buffer_pattern(), 1);
	restore_vars();
	if (!remote)
		move_to_column(goal_column);
	dired_orig_position = point;
	fix_window_start();
	if (point >= size())
		window_start = prev_screen_line(text_height() / 2);
}

void ssh_activity(int activity, int buf, int from, int to);
char *scp_variable(struct url_parts *parts, char *varname);

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = is_unix ? 6 : 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s");
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
	char url[FNAMELEN];
	sprintf(url, "ssh://%s", host);
	if (parts.have_usr)
		sprintf(url, "ssh://%s@%s", username, host);
	buffer_url = strkeep(url);
}

text_width_after_indent(int p)
{
	save_var point = p;
	to_end_line();
	re_search(-1, "[ \t]*");
	return current_column()- get_indentation(p);
}

// If the file doesn't exist, substitute a file with the same basename.
proc_fix_filename(char *fname)
{
	if (!(process_next_error_options & 1) || is_remote_file(fname)
		  || check_file(fname))
		return;
	int orig = bufnum;
	int i = buf_list(0, 2);	/* start with most recently created buf */
	do {
		bufnum = i;
		if (*filename
			  && !fnamecmp(get_tail(filename, 0), get_tail(fname, 0))
			  && force_save_as != FSA_NEWFILE
			  && check_file(filename) == CHECK_FILE) {
			show_text((process_next_error_options & 2) ? 0 : -2, see_delay,
					  "No file %f found, picking another %s",
					  fname, get_tail(fname, 0));
			strcpy(fname, filename);
			return;
		}
	} while (i = buf_list(-1, 1));
	bufnum = orig;
}

do_next_error(int count, char *cmd)	// return 0 if no errors, else 1
{
	char file[FNAMELEN], number[30], line[FNAMELEN], *p;
	int lineno, origpt, dir, from, to, is_bufname, file_paren, line_paren;
	int orig = bufnum, procbuf, start, have_file = 1;

	tiled_only();
	_last_was_grep = 0;
	if (spot_to_buffer(error_spot) <= 0 && !exist(PROCBUF))
		error("No process buffer to read errors from.");
	tiled_only();
	if (spot_to_buffer(error_spot) > 0)
		bufnum = spot_to_buffer(error_spot);
	else
		bufname = PROCBUF;
	telnet_error_set_to_telnet();
	procbuf = bufnum;
	origpt = point;
	case_fold = 1;
	if (spot_to_buffer(error_spot) > 0)
		point = *error_spot;
	else {
		error_spot = alloc_spot();
		point = 0;
	}
	if (!count && procbuf == orig)
		point = origpt;

	noteput("Searching...");
	if (!count)
		find_error(-1, &file_paren, &line_paren), count = 1;
	dir = (count > 0) ? 1 : -1;
	for (; count; count -= dir) {
		if (!find_error(dir, &file_paren, &line_paren)) {
			point = origpt;
			bufnum = orig;
			if (_had_errors == 2)
				return 0;
			else if (_had_errors)
				say("No more errors.");
			else if (cmd)
				say("No errors running \"%s\".", cmd);
			else
				say("No errors.");
			return 0;
		}
	}
	_had_errors = 1;
	grab(find_group(file_paren, 1), find_group(file_paren, 0), file);
	grab(find_group(line_paren, 1), find_group(line_paren, 0), number);
	start = find_group(0, 1);
	from = find_group(line_paren, 0);
	to = find_group(0, 0);
	if (to - from < 20) {	// error message at end is too short
		from = start;	// so try including whole line
		if (to - from > screen_cols - 5) // Only rightmost if it's too big.
			from = to - screen_cols + 5;
	}
	if (to > from + sizeof(line) / sizeof(*line) - 1)
		to = from + sizeof(line) / sizeof(*line) - 1;
	grab(from, to, line);
	is_bufname = (parse_string(1, "Buffer ") != 0);
	if (dir > 0)
		to_end_line();
	*error_spot = point;
	point = origpt;
	lineno = strtoi(number, 10);
	bufnum = orig;
	if (!*file || !*number) {
		locate_window(bufnum_to_name(procbuf), "");
		point = *error_spot;
		to_begin_line();
		say("=> %s", line);
		return 1;
	}
	if (is_bufname)
		locate_window(file, "");
	else {
		process_find_directory(file, procbuf, from);
		if (!check_file(file) && !strcmp(get_extension(file), ".java"))
			proc_find_java_source(file, procbuf, start);
		proc_fix_filename(file);
		locate_window("", file);
		if (process_view_error_lines
			  && window_number == number_of_windows() - 1)
			window_number = 0;
		if ((process_next_error_options & 4) && !is_remote_file(file)
			  && check_file(file) != CHECK_FILE)
			have_file = 0;
		else
			std_find_it(file);
	}
	if (have_file) {
		go_line(lineno);
		if (is_remote_file(file))
			ftp_go_to_line = lineno;
	} else
		say("%f not found", file);
	for (p = line; *p && index("()[] :;,\t\n0123456789", *p); p++)
		;			// skip over these chars at start of msg
	say("=> %s", p);
	restore_vars();
	proc_view_error_lines(procbuf, start);
	return 1;
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN], search_other_buf_name[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	if (search_in_region && is_highlight_on()) {
		save_var _highlight_control;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	} else if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var _highlight_control;
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	if (!search_other_name) {
		sprintf(search_other_buf_name, "search-%s", bufname);
		save_var search_other_name = search_other_buf_name;
	}
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = full_key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps & 1) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))), "search");
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, comp_tab) == (short) inp_get_other) {
			flags = pick_search_recall(flags);
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (binding_of(full_key, reg_tab) == (short) delete_character
				   && (search_delete_match & 1) && (flags & INCR)) {
			window_handle = origwin;
			if (!(oldflags & FAILING))
				delete(point, mark);
			full_key = -1;
			break;
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		flags = grab_search_string(flags, &trystr);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = editwin;
	if (size()) {
		flags = grab_search_string(flags, &trystr);
		save_search_recall(flags, trystr);
	}
	if (key == abort_key || user_abort) {
		window_handle = origwin;
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			if (_last_was_regex)
				grab_search_string(flags, &_default_regex_search);
			else
				grab_search_string(flags, &_default_search);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}
