/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2009 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.06 to 13.07.

#include "eel.h"
#include "c.h"
#include "session.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "proc.h"
#include "bufed.h"
#include "grep.h"
#include "html.h"
#include "perl.h"
#include "kill.h"
#include "misclang.h"
#include "tags.h"
#include "vbasic.h"
#include "vhdl.h"
#include "inimode.h"
#include "makemode.h"
#include "tcl.h"
#include "tex.h"
#include "python.h"

command down_line()
{
	int true;

	down_or_up_lines(1);
	if (in_macro() && window_extra_lines())		// Don't scroll unnecessarily.
		return;
	true = point;
	while (scroll_at_end > 0 && true > window_end) {
		window_scroll(scroll_at_end);
		if (window_end + 1 >= size() - narrow_end || in_macro())
			break;		/* make sure we stop at end */
		point = window_end + 1;	/* compute window_end explicitely */
		window_end = next_screen_line(scroll_at_end) - 1;
		check_abort();
	}
	point = true;
}

bufed_getname(s)
char *s;
{
	int start, oline;

	*s = 0;
	save_var abort_searching = 0;
	move_to_column(BNAME_COL - 1);
	start = point;
	if (character(point) == '"') {
		oline = lines_between(0, point, 0);
		start = ++point;
		if (!search(1, "\"") || oline != lines_between(0, point, 0))
			return 0;
		point--;
	} else {
		if (character(start) == ' ')
			start++;
		move_to_column(BNAME_COL + bufed_column_width);
		if (character(point - 1) != ' ')
			if (!search(1, " "))
				return 0;
		if (!re_search(-1, " *"))
			return 0;
	}
	grab(start, point, s);
	to_bufed_column();
	return 1;
}

to_bufed_column()
{
	move_to_column(BNAME_COL);
}

char _bufed_mode_name[] = "Bufed";

bufed_mode()
{
	mode_default_settings();
	mode_keys = bufed_tab;
	major_mode = _bufed_mode_name;
	mouse_dbl_selects = 1;
	goal_column = BNAME_COL;
	make_mode();
}

list_line(inbuffer)	/* insert a line into the buffer list */
char *inbuffer;
{
	char file[FNAMELEN * 2], msg[FNAMELEN * 2], bname[FNAMELEN * 2], *s;
	int width;

	strcpy(bname, bufname);
	quote_bufname(bname);
	if (force_save_as == FSA_NEWFILE)
		*file = 0;
	else if (bufed_show_absolute_path == 3) {
		strcpy(file, filename);
		if (*filename && !discardable_buffer) {
			sprintf(file, "%s [", get_tail(filename, 0));
			s = file + strlen(file);
			strcpy(s, filename);
			*get_tail(s, 1) = 0;
			drop_final_slash(s);
			strcat(s, "]");
		}
	} else if (bufed_show_absolute_path == 4) {
		char dir[FNAMELEN];
		relative(filename, file);
		getcd(dir);
		if (*filename && is_relative(file) && !discardable_buffer)
			sprintf(file + strlen(file), " [%s]", dir);
	} else if (bufed_show_absolute_path) {
		strcpy(file, filename);
		if (bufed_show_absolute_path == 2)
			convert_to_8_3_filename(file);
	} else
		relative(filename, file);
	width = sprintf(msg, "%10d%10d %c%c%-*s %s\n",
					buf_accessed, size(), modified ? '*' : ' ', 
					_buf_readonly ? 'R' : ' ', bufed_column_width,
					bname, file);
	bufname = inbuffer;
	stuff(msg);
	width = width - ACCESS_WIDTH + 2;
	if (width > bufed_width)
		bufed_width = width;
}

sort_buffer_list(opos, buf)
char *buf;		/* sort from opos to the end, & return new end */
{
	int col, want_sort = 1, rev = 0;

	save_var case_fold = (preserve_filename_case & 8) != 0;
	switch (tolower(_bufed_sort)) {
		case 'f':	col = BNAME_COL + ACCESS_WIDTH + bufed_column_width; break;
		case 'b':	col = BNAME_COL + ACCESS_WIDTH; break;
		case 'i':	col = ACCESS_WIDTH; break;
		case 'a':	col = 0; rev = 1; break;
		default:	want_sort = 0; break;
	}
	if (isupper(_bufed_sort))
		rev = !rev;
	save_var bufnum;
	bufname = buf;
	if (want_sort)
		do_sort_region(opos, size(), col, rev);
	save_var narrow_start = opos;
	strip_columns(0, ACCESS_WIDTH);		// remove access timestamps
	return size();
}

bufed_when_displaying(from, to);

do_bufed(flags)
{
	char find[FNAMELEN * 2], msg[FNAMELEN], title[FNAMELEN];
	int count;

	strcpy(title, " Size      Buffer");
	bufed_column_width = MAX(MIN(bufed_column_width, 200), 10);
	for (count = 6; count < bufed_column_width; count++)
		strcat(title, " ");
	strcat(title, "File ");
	if (in_bufed()) {
		bufnum = window_bufnum;
		set_read_only(0);
		if (bufed_getname(find)) {
			quote_bufname(find);
			strcat(find, " ");
		} else
			*find = 0;
		bufed_restore_orig();
		make_buffer_list(flags, BUFED_BUF);
		window_handle = _bufed_window;
		window_bufnum = name_to_bufnum(BUFED_BUF);
		point = 0;
	} else {
		_bufed_orig_window = window_handle;
		_bufed_orig_buffer = window_bufnum;
		_bufed_orig_highlight =_highlight_control;
		strcpy(find, bufname);
		quote_bufname(find);
		strcat(find, " ");
		make_buffer_list(flags, BUFED_BUF);
		build_first = 1;
		_bufed_window = add_popup(2, 2, bufed_width, screen_lines - 6,
								  ALL_BORD(BSINGLE), name_to_bufnum(BUFED_BUF));
		if (is_gui && want_gui_prompts) {
			count = lines_between(0, size(), 0) + 1;
			mouse_cursor = &left_pointer;
			_bufed_caption_window = add_popup(0, 0, bufed_width, 1, 0,
											  create(BUFED_CAPTION_BUF));
			buf_zap(bufnum);
			bprintf("   %s", title);
			point = 0;
			system_window = 1;
			undo_size = 0;
			mouse_cursor = &left_pointer;
			window_handle = _bufed_window;

			if (flags & BUFED_EXIT) {
				sprintf(msg, "%d buffer%s not saved.  Exit?", count,
						(count == 1) ? "" : "s");
				display_dialog_box("AskExitBox", msg,
								   _bufed_caption_window, _bufed_window, 0,
								   "E&xit", "&Cancel", "");
			} else if (flags & BUFED_SAVE) {
				sprintf(msg, "Save %d unsaved buffers first? "
						"(! to never ask)", count);
				display_dialog_box("AskSaveBox", msg,
								   _bufed_caption_window, _bufed_window, 0,
								   "&Save", "&No Save", "&Cancel");
			} else if (flags & BUFED_SWITCH)
				display_dialog_box("SwitchBufBox", "Select a Buffer",
								   _bufed_caption_window, _bufed_window, 0,
								   "OK", "Cancel", "Help");
			else
				display_dialog_box("CaptionBox", "Select a Buffer",
								   _bufed_caption_window, _bufed_window, 0,
								   "OK", "Cancel", "Help");
		}
		window_handle = _bufed_window;
	}
	set_menu_look(_bufed_window);
	if (flags & (BUFED_EXIT | BUFED_SAVE))
		remove_line_highlight();
	else
		make_line_highlight();
	window_title(_bufed_window, BTOP, TITLELEFT(4), title);
	undo_size = 0;
	display_column = 0;	/* horizontally scroll this window */
	popup_near_window(_bufed_window, _bufed_orig_window);
	bufed_mode();
	if (*find)
		col_search(find, BNAME_COL - 1);
	to_bufed_column();
	modified = 0;
	goal_column = BNAME_COL;
	when_displaying = bufed_when_displaying;
}

do_make_bname(char *fname, char *bname, char *dup_ok)
{
	int uniq = 1, ok;
	char *to, *oldbuf;

	strcpy(bname, get_tail(fname, 0));
	if (!(preserve_filename_case & 2)
		  && (filename_rules(fname) & FSYS_CASE_MASK) != FSYS_CASE_SENSITIVE)
		for (to = bname; *to; to++)
			*to = tolower(*to);
	for (to = bname; *to; to++)
		if (*to < ' ' || *to == '"')
			*to = '!';
	to = bname + strlen(bname);
	while (exist(bname) && (!dup_ok || strcmp(bname, dup_ok))) {
		oldbuf = bufname;
		bufname = bname;
		ok = (size() == 0);
		bufname = oldbuf;
		if (ok) return;
		sprintf(to, "<%d>", uniq++);
	}
}

/* Act like C-x b using supplied prompt */
ask_which_buffer(pr)
char *pr;
{
	char bfname[FNAMELEN];
	int prev = give_prev_buf();

	get_buf(bfname, pr, prev ? bufnum_to_name(prev) : "");
	if (!*bfname)
		return;
	int newbuf = !exist(bfname);
	to_buffer(bfname);
	if (newbuf) {
		char fname[FNAMELEN];
		int unique = 1;
		do {
			sprintf(fname, "Scratch %d%s", unique++, new_file_ext);
			absolute(fname);
			if (unique > 99999) {
				say("Couldn't find unused scratch file name.");
				return;
			}
		} while (do_file_match(fname, STARTMATCH | EXACTONLY));
		force_save_as = FSA_NEWFILE;
		filename = fname;
	}
}

move_in_switch_buffers(int up)
{
	if (up) {
		if (!nl_reverse())
			point = size();
	} else {
		if (!nl_forward() || point == size())
			point = 0;
	}
	move_to_column(BNAME_COL);
}

in_c_comment(lookback)		// is point inside a c comment, string, etc?
{
	int i = -1, orig = point;

	if (in_perl_buffer)
		return in_perl_comment(point);
	if (lookback) {			// If !lookback, check only this line.
		save_var narrow_start = point - lookback;
		save_var point;
		for (; re_search(RE_REVERSE | RE_FASTEST, "/<*>|<*>/"); point++) {
			i = scan_for_string();	// Parse just this line.  A string?
			if (curchar() == '/')
				switch (character(point - 1)) {
					case '/':		// Found //*, so no block comment
						if (!(in_c_comment_at(point - 1, lookback)
							  & IN_OLD_COMMENT))
							continue;
						else	// unless it's /*****//****
							break;
					case '*':		// Found */*, end of comment, not start.
						point--;
						break;
				}
			if (!i || i == IN_OLD_COMMENT)
				break;		// The delimiter wasn't quoted.
			if (curchar() == '*' && c_in_old_comment()) {
				save_var narrow_start = point;
				break;	// Found */ sequence, ends a comment?
			}	// If it looks like it was quoted, double-check.
		}
		if (i != -1 && curchar() == '/' && character(point + 1) == '*')
			return IN_OLD_COMMENT;	// Found /* sequence.
	}
	point = orig;	// Not in a block comment, so use a local check.
	return scan_for_string();
}

int in_c_comment_at(int pos, int lookback)
{
	save_var point = pos;
	return in_c_comment(lookback);
}

early_restoring()
{
	char fname[FNAMELEN];
	int i;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_DOS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
	if (load_from_state)
		show_text(-2, 0, "Loaded from state file %s", state_file);
	else if (*state_file)
		show_text(-2, 0, "Loaded from state file %s (quickup)", state_file);
	else
		show_text(-2, 0, "Loaded from bytecode files");
	for (i = 0; original_argv(i); i++)
		show_text(-2, 0, "Argv[%d]=%s", i, original_argv(i));
	get_customization_directory(fname);
	show_text(-2, 0, "Customization directory is %s", fname);
}

set_color_handler(int activity, int buf, int from, int to);

set_color_xwin()
{
	int i, j, classes = number_of_color_classes();
	int *orig_fg, *orig_bg, *orig_schemes;
	int orig_scheme = selected_color_scheme, this_scheme = orig_scheme;
	char *s, cmd[FNAMELEN];
	int orig_class = get_character_syntax_color(point);

	if (no_mac_dialog_helper()) {
		mac_gtk_message(0);
		return;
	}
	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	on_exit free(orig_schemes), free(orig_fg), free(orig_bg);
	save_var bufnum;
	get_executable_file(cmd, "epsilon-dialogs", 1);
	sprintf(cmd + strlen(cmd), " --parent %x --color", screen_to_window_id(0));

	for (;;) {
		bufnum = zap("-epsilon-dialog-data");
		bprintf("%d\n%s\n", font_styles, font_fixed);
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i))
				bprintf("%s\n", name_name(i));
		bprintf("\n");
		for (i = 0; i < classes; i++)
			bprintf("%s\n", name_color_class(i));
		bprintf("\n");
		bprintf("%s\n", (s = lookpath("colclass.txt")) ? s : "");
		bprintf("%s\n", name_name(this_scheme));

		selected_color_scheme = find_index("color-defaults");
		for (j = 0; j < classes; j++)
			bprintf("default %x %x\n", get_foreground_color(j, 1),
					get_background_color(j, 1));
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i)) {
				selected_color_scheme = i;
				for (j = 0; j < classes; j++)
					bprintf("%d %d %x %x\n", i, j, get_foreground_color(j, 1),
							get_background_color(j, 1));
			}
		make_color_class_list(1);
		point = size();
		s = name_color_class(orig_class);
		bprintf("\n%s\n", s ? s : "");
		selected_color_scheme = this_scheme;
		if (!pipe_text("-epsilon-dialog-data", "-epsilon-dialog-out", cmd,
					   NULL, PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH,
					   set_color_handler))
			file_error(errno, "Couldn't run dialog helper", "Error");
		this_scheme = selected_color_scheme;
		bufname = "-epsilon-dialog-out";
		i = process_exit_status;
		if (matches_at(0, 1, ".*(ibrary|shared)")) {
			mac_gtk_message(0);
			return;
		}
		if (i < 0 || i > 2)
			i = !ask_yn("Keep changes", "The dialog process exited "
						"unexpectedly.  Keep these settings? ",
						"Yes", "No", 0);
		if (i == 0) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			return;
		} else if (i == 1) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			return;
		} else {
			show_binding("Continue changing colors "
						 "by typing %s", "exit-level");
			recursive_edit();
		}
	}
}

save_recall(text, key)	// save response to prompt using this key
char *text, *key;
{
	char resp[FNAMELEN * 2];

	save_var matchstart, matchend, user_abort = 0, abort_searching = 0;
	if (strlen(text) > recall_longest_response) {
		save_var see_delay = 500;
		sayput("Response to prompt was too long to save, not saved.");
		sayput("Recall-longest-response is %'d, this response was %'d characters.",
			   recall_longest_response, strlen(text));
		return;
	}
	if (*text) {
		save_var bufnum = create(OLDRESP);
		if (size() > recall_maximum_size) {
			point = size() - recall_maximum_size;
			to_begin_line();
			delete(0, point);
		}
		if (strlen(key) + strlen(text) < sizeof(resp) / sizeof(*resp) / 2) {
			point = size();		// delete old identical resps
			sprintf(resp, "^%q: %q\n", key, text);
			if (re_search(-1, resp))
				delete(matchstart, matchend);
		}
		point = size();
		bprintf("%s: %s\n", key, text);
	}
}

get_choice(list, resp, title, msg, b1, b2, b3)
char *resp, *title, *msg, *b1, *b2, *b3;
{
	int w = window_handle, win;

	error_if_input(1);	/* error if we're in the middle of input */
	save_var _doing_input = DI_LINEINPUT;
	win = new_menu(list);
	window_title(win, BTOP, TITLECENTER, title);
	window_to_fit(screen_lines - 7);
	window_width_to_fit();
	save_var _window_flags |= HIGHLIGHT_LINE;
	if (search_in_menu != 2)	// Special code for single-char options.
		save_var search_in_menu = 0;
	save_var display_column = 0;	// don't wrap
	if (*resp) {			// Otherwise use initial position.
		point = 0;
		col_search(resp, 0);	// Searches from current position.
	}
	if (is_gui && want_gui_prompts) {
		one_window_to_dialog(title, win, b1, b2, b3);
	} else if (msg)
		say(msg);
	select_menu_item(0, win, w, 0);
	grab_full_line(list, resp);
	ungot_key = -1;
	return !user_abort;
}

select_menu_item(resbuf, menuwin, owin, dir)
{	// let user select line from menu window, maybe copy line to resbuf
	int res; // dir tells how to srch on norm keys, or 0 if they're cmds
	char str[FNAMELEN], pat[FNAMELEN * 2];

	save_var mode_keys;
	cursor_to_column = -1;
	do {
		mode_keys = menu_tab;
		res = do_recursion();
		if (!is_window(menuwin))
			res = INP_ABORT;
		else if (res == INP_ADD && search_in_menu == 2) {
			sprintf(pat, "^%c", key);
			window_handle = menuwin;
			to_end_line();
			if (!re_search(1, pat)) {
				point = 0;
				re_search(1, pat);
			}
			to_begin_line();
			res = 0;
		} else if (res == INP_ADD && search_in_menu && dir) {
			window_handle = owin;
			run_topkey();
			window_handle = owin;
			grab(0, FNAMELEN - 1, str);
			sprintf(pat, "^%q", str);
			window_handle = menuwin;
			point = (dir == 1) ? 0 : size();
			re_search(dir, pat);
			to_begin_line();
			res = 0;
		} else if (res == INP_ADD && !dir)
			switch (isalpha(key) ? toupper(key) : 0) {
				case 'N':
				case 'P':	ungot_key = CTRL(key);
				default:	res = 0;	// fall through
				case 'Q':	break;
			}
	} while (!res);
	switch (res) {		// try all other chars
		case INP_ADD:	ungot_key = full_key;
				break;
		case INP_GET_EXIT:
		case INP_GET:	if (resbuf)
					get_menu_line(resbuf);
				break;
	}
	if (res == INP_ABORT && !user_abort)	// user deleted edit window?
		user_abort = 1;
	remove_line_highlight();
	window_handle = owin;
	remove_window(menuwin);
	return res;
}

int get_key_choice(int list, char *resp, char *title, char *msg,
				   char *b1, char *b2, char *b3)
{
	save_var search_in_menu = 2;
	return get_choice(list, resp, title, msg, b1, b2, b3);
}

bind_universally(int k, int f)
{
	int i, max = final_index();
	for (i = 1; i <= max; i++)
		if (name_type(i) == NT_TABLE && index_table(i)[generic_key(k)] > 0)
			index_table(i)[k] = f;
}

dired_refresh() on dired_tab['r']
{
	char msg[FNAMELEN];
	int yes, remote;

	if (modified) {
		sprintf(msg, "Relist %s, discarding markings? ", bufname);
		yes = ask_yn("Discard markings", msg, "Yes", "No", 1);
		check_abort();
		if (!yes)
			return;
	}
	remote = is_remote_file(get_dired_buffer_pattern());
	if (remote && (ftp_job || is_busy_doing_scp(bufnum) != -1))
		error("Still retrieving directory listing in this buffer.");
	if (remote)
		point = 0;
	else
		save_var point, window_start;
	save_var file_pattern_wildcards = dired_buffer_pattern_wildcards;
	create_dired_listing(get_dired_buffer_pattern(), 1);
	restore_vars();
	if (!remote)
		move_to_column(goal_column);
	dired_orig_position = point;
	fix_window_start();
	if (window_start >= size())
		window_start = prev_screen_line(text_height() / 2);
}

#include "dispchar.e"
#include "editvars.e"

#define HTTP_HEADERS	"HTTP Headers"

http_activity(int activity, int buf, int from, int to)
{
	int headers, pos, orig;
	char err[FNAMELEN];

	UNREFERENCED_PARAMETER(to);
	if (!buf_exist(buf))
		return;
	save_var bufnum = buf;
	if (activity == NET_DONE) {
		set_mode_message(NULL);
		refresh();
		if (when_net_activity == http_activity)
			when_net_activity = 0;
		return;
	}
	if (!from)
		position_in_all_windows(buf, 0);
	bufnum = buf;
	if (mode_extra && !strstr(mode_extra, "Connecting")) {
		if (from < 1000 && mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
		return;
	}
	save_spot point = 0;
	if (re_search(1, "\r?\n\r?\n")) {
		set_mode_message("[HTTP: Retrieving] ");
		pos = point;
		headers = create(HTTP_HEADERS);	// Got end of headers yet?
		set_buf_point(headers, orig = buf_size(headers));
		buffer_printf(HTTP_HEADERS, "Headers for %s\n", filename);
		xfer(HTTP_HEADERS, 0, pos);
		point = 0;
		if (parse_string(1, "HTTP/[.0-9]+ [1345]")) {
			int m = matchend - 1;
			grab(m, MIN(parse_string(1, "[^\n\r]*"), 200), err);
			say("Response: %s, see buffer \"HTTP Headers\" for details.", err);
		}
		save_var modified = 0;	// Ignore following change to buffer.
		delete(0, pos);
		if (mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
		bufname = HTTP_HEADERS;
		discardable_buffer = 1;	// Deleting the headers is ok.
		save_var narrow_start = orig;
		simple_re_replace(1, "\r\n", "\n");
		translation_type = FILETYPE_MSDOS;
	}
}

scp_restart_process(struct scp_job *p)
{
	char dir[FNAMELEN];

	strcpy(dir, p->file);
	if (is_remote_file(dir))
		*dir = 0;
	else
		*get_tail(dir, 1) = 0;
	buffer_printf("-scp-trace", "Run %s in %s\n", p->cmdline, dir);
	if (concur_shell("", p->cmdline, *dir ? dir : NULL,
					 bufnum_to_name(p->buf)))
		error("Couldn't start scp process");
	save_var bufnum = p->buf;
	if (is_unix && strstr(p->cmdline, "sftp"))
		process_echo &= 3;	// Turn off our echoing, since sftp does it too.
}

scp_send_some_command(struct url_parts *parts, struct scp_job *p, char *fmt)
{
	int len = strlen(parts->fname) * 2 + strlen(fmt) + 40;
	if (!(p->flags & SCP_CLIENT_MODE) && p->homedir)
		len += strlen(p->homedir);
	p->cmd = p->cmdlist = malloc(len);
	if (!(p->flags & SCP_CLIENT_MODE)) {
		char dirpart[FNAMELEN];
		if (p->busy == 2)
			sprintf(p->cmd, "pwd\n");
		else if (p->homedir)
			sprintf(p->cmd, "cd \"%s\"\n", p->homedir);
		else
			error("Can't determine home directory.");
		strcpy(dirpart, parts->fname);
		*get_tail(dirpart, 1) = 0;
		if (*dirpart)
			sprintf(p->cmd + strlen(p->cmd), "cd \"%s\"\n", dirpart);
		if (*fmt)
			sprintf(p->cmd + strlen(p->cmd), fmt, get_tail(parts->fname, 1));
	} else
		sprintf(p->cmdlist, "list \"%s\"\n", parts->fname);
	if (p->busy != 2)
		scp_next_cmd(p);
}

wait_for_remote_file(int b)
{
	save_var bufnum = b;
	for (;;) {
		if (!ftp_job && is_busy_doing_scp(b) == -1
			  && when_net_activity != http_activity)
			return;
		net_during_wait(0);
		check_abort();
	}
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			point = type_point;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 0x20) {
				point = from;
				simple_re_replace(0, "<#0-#1>", "");
			}
			if (telnet_interpret_output & 0x40)
				remove_xterm_sequences(from);
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			if (telnet_interpret_output & 0x80)
				process_backspace_sequences(from - 100);
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

// For sending chars to ssh job, act like a process, but for getting
// chars, act like telnet.
ssh_activity(int activity, int buf, int from, int to)
{
	save_var telnet_interpret_output = ssh_interpret_output, abort_searching = 0;
	if (activity == NET_SEND) {
		save_var bufnum = buf;
		color_process_range(from, to);
		do_when_process_input(buf, from, to);
	} else
		telnet_activity(activity, buf, from, to);
}

maybe_add_readonly(struct file_info *now)
{
	if (!(now->attr & ATTR_READONLY) || (file_info.attr & ATTR_READONLY))
		return;
	if (modified) {
		char msg[FNAMELEN];
		save_var window_bufnum = bufnum;
		refresh();
		sprintf(msg, "%s is now read-only.  Set buffer read-only too? ",
				get_tail(filename, 0));
		if (!ask_yn("File is read-only", msg, "Yes", "No", 1))
			return;
	}
	do_readonly_warning();
}

command set_file_name()
{
	char tmp[FNAMELEN];

	get_file_dir(tmp, "Enter new output file name: ");
	if (!*tmp || warn_existing_file(tmp) > 0)
		return;
	set_buffer_filename(tmp);
	modified = 1;
}

call_mode(char *f)
{
	if (!retain_forced_mode && !check_filevar_mode(bufnum)) {
		if (!strfcmp(get_extension(f), ".in"))	// Remove .in extension.
			*get_extension(f = strkeep(f)) = 0;
		call_by_suffix(f, "suffix-%s");
	}
}

#define C_COMMENT_FILLER		"[^0-9a-zA-Z\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

do_fill_comment(int generic, int margin)
{
	int pbuf, width = 1000, i, old_lines, new_lines, orig_lines;
	int col, col2, orig = bufnum, type;
	char last_prefix[FNAMELEN], orig_prefix[FNAMELEN], pat[FNAMELEN], comm[30];

	save_var margin_right = margin;
	save_spot point;
	type = locate_c_comment();	// 1 if block comment, 2 if one-line
	if (!type)					// comment, or 0 if not a comment.
		error("Can't fill comment here, not in a comment.");
	point += parse_string(1, C_COMMENT_FILLER);	// Past 1st line's prefix.
	sprintf(pat, "[ \t]*%s[^0-9a-z_\n]*", generic ? get_comment_start() : "/[*/]");
	grab(point, point - parse_string(-1, pat), orig_prefix);
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (!generic)
		point -= parse_string(-1, "<*>/");
	save_var narrow_end = size() - point + 1;

	if (type == 1 && !generic)		// Combine adjacent block comments.
		simple_re_replace(1, "[ \t]*<*>/[ \t]*\n?[ \t]*/<*>[ \t]*", " ");
	point = 0;
	while (nl_forward()) {	// Find width of narrowest prefix.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		width = MIN(width, col);
	}
	pbuf = tmp_buf();
	point = 0;
	get_comment_marker(comm, generic);
	sprintf(pat, "[ \t]*%q", comm);
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		if (col > width)	// Make sure it's not too wide.
			col = width;
		if (type == 2) {	// But if it's made of one-line comments,
			i = parse_string(1, pat);
			col2 = get_column(point + i);
			if (col < col2)		// each prefix must include a "//".
				col = col2;
		}
		i = column_to_pos(col);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	if (width == 1000)	// No prefix, so use default.
		width = 3;
	orig_lines = old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 20),
						   MAX(margin_right - width, 20));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;
	*last_prefix = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		grab(point, give_end_line(), last_prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	if (orig_lines < 2) {		// Default for new lines if 1 line originally.
		strcpy(last_prefix, orig_prefix);
		if (type == 2 && !strstr(last_prefix, comm))
			sprintf(last_prefix, "%s ", comm);
		else if (type == 1 && !generic)
			if (!strstr(last_prefix, "/*"))
				strcpy(last_prefix, (auto_fill_comment_rules & 4) ? " " : " * ");
			else {
				replace_in_string(last_prefix, '/', ' ');
				if (fill_c_comment_plain)
					replace_in_string(last_prefix, '*', ' ');
			}
	}
	bufnum = orig;
	narrow_start = 0;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(last_prefix);			// Add some prefixes.
		if (generic) {
			save_var indenter_action = IACT_FILL_COMMENT;
			to_begin_line();
			if (indenter)
				(*indenter)();
		}
		nl_forward();
	}
	buf_delete(pbuf);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

int do_maybe_break_line(int c_like)
{
	int comm, res = 1, starts_with_comm, starts_in_comm, after_prefix;
	int comm_start = -1, break_block_comment = 0;
	int generic = strcmp(major_mode, "C");
	char fill[FNAMELEN], pat[FNAMELEN], raw_comm[FNAMELEN];
	spot after, next_ends = 0;

	delete_if_highlighted();

	// Various reasons we might want not want to fill right now.
	if (give_end_line() - give_begin_line() > FNAMELEN - 10)
		return 1;			// The line's too long for us.
	if (IS_ENTER_KEY(key)
		  && (c_like ? in_c_comment(c_look_back) : in_comment()) == IN_OLD_COMMENT
		  && (auto_fill_comment_rules & 1))
		break_block_comment = 1, res = 0;
	else if (current_column() + 1 <= margin_right)
		return 1;		// We're not past the margin.
	comm = c_like ? in_c_comment(c_look_back) : in_comment();
	if (!(comm & IN_COMMENT))
		return 1;			// We're not in a comment.

	if (c_like)
		generic = 0;
	// Collect info on the situation.
	if (key == ' ') {		// Insert <Space> now, then remember not
		insert(key);		// to later, so we can break here.
		res = 0;
	}
	after = alloc_spot(1);	// Remember where we should end up.
	to_begin_line();	// See if a comment's the first thing on this line.
	get_comment_marker(raw_comm, generic);
	sprintf(pat, "[ \t]*%s", generic ? raw_comm : "/[*/]");
	starts_with_comm = parse_string(1, pat);	// Yes, comment's first on line.
	starts_in_comm = c_like ? in_c_comment(c_look_back) : in_comment();
	// Does this line start with a comment?
	sprintf(pat, "[ \t%s]*", generic ? raw_comm : "*/");
	parse_string(1, pat, fill);	// Repeat this prefix on new lines.
	after_prefix = point;
	if (!starts_in_comm) {		// Find where this line's comment starts.
		to_begin_line();
		sprintf(pat, "<min>.*%s",
				generic ? raw_comm : comm == IN_NEW_COMMENT ? "//" : "/%*");
		point += parse_string(1, pat);
		comm_start = point;
	}
	if (nl_forward()) {		// Remember the end of the next line, unless
		to_end_line();		// there's a comment there.
		if (c_like && !in_c_comment(c_look_back))
			next_ends = alloc_spot(0);
	}

	// Look for a breaking point.
	point = *after;
	if ((c_like && c_auto_fill_mode != 1 || starts_with_comm || starts_in_comm)
		  && (auto_fill_comment_rules & 2) && !break_block_comment
		  && pick_break(margin_right) && point > after_prefix
		  && point > comm_start)
		break_block_comment = 1;

	if (break_block_comment)
		comm_break_line_here(c_like, comm, starts_with_comm, starts_in_comm,
							 comm_start, next_ends, after, fill);

	point = *after;
	free_spot(after);
	if (next_ends)
		free_spot(next_ends);
	return res;
}

int maybe_break_this_line()
{
	int h, found = 1;
	spot old;

	h = current_column();
	if (h >= margin_right) {
		old = alloc_spot(1);
		on_exit free_spot(old);
		if (h == margin_right && (key == '\n' || (key == ' '
			&& (!mode_restrict_break || mode_restrict_break(point))))) {
			break_line_here();	/* just change the ' ' to a '\n' */
			return 0;
		}
		found = pick_break(margin_right);
		if (found) {
			break_line_here();
			if (IS_ENTER_KEY(key))
				full_key = key = ' ';
			point = *old;
		} else if (key == ' ')
			if (!mode_restrict_break || mode_restrict_break(point))
				break_line_here();
			else
				return 1;
		if (!found && key == ' ')
			return 0;
	}
	return 1;
}

do_fill_indented_paragraph()
{
	int pbuf, width, i, old_lines, new_lines, orig = bufnum;
	char prefix[FNAMELEN];

	save_spot point;
	point--;
	forward_paragraph();
	mark = point - (point < size() - narrow_end);
	backward_paragraph(); 
	point += parse_string(1, "([ \t]*\n)+");
	parse_string(1, "[ \t]*", prefix);
	width = current_column();
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	point -= parse_string(-1, "\n[ \t]*");
	save_var narrow_end = size() - point;

	point = 0;
	while (nl_forward())	// Find width of narrowest indentation.
		width = MIN(width, get_indentation(point));
	pbuf = tmp_buf();
	point = 0;
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = column_to_pos(width);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 2),
						   MAX(margin_right - width, 2));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		if (point < give_end_line())
			grab(point, give_end_line(), prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	bufnum = orig;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(prefix);			// Add some prefixes.
		nl_forward();
	}
	buf_delete(pbuf);
}

grep_prompt(flags, flavor)
{
	char filepat[FNAMELEN], old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;
	int inbufs = (flavor & 1), nextwin = (flavor & 2);

	_last_was_grep = 1;
	tiled_only();
	iter = 0;
	error_if_input(1);	/* error if we're in the middle of input */
	flags = get_search_string("List lines with: ",
				  flags | (inbufs ? BUFGREP : FILEGREP));
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	if (!inbufs)
		save_var full_path_on_mode_line = 1;
	all_must_build_mode = 1;
	if (!get_grep_file(inbufs, filepat))
		return;
	restore_vars();
	all_must_build_mode = 1;
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	if (inbufs && (search_in_region & 2) && !strcmp(filepat, bufname)
		  && is_highlight_on()) {
		save_var point, mark;
		fix_region();
		save_var narrow_start = point, narrow_end = size() - mark;
	}
	if (nextwin)
		next_window();
	if (!setjmp(top_level))
		do_grep("grep", default_search_string(flags), flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];
	int show_rel = 1;		// Show the relative path of the default.

	if (!inbufs) {
		if (is_dired_buf() && !inp_get_other_name)
			save_var inp_get_other_name = get_dired_buffer_pattern();
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory
			  || grep_default_directory == 3)
			show_rel = 0;
		if (grep_prompt_with_buffer_directory == 2 && !grep_default_directory)
			show_rel = 0;	// Ensure we show which files we'll search by default.
		build_prompt(pr, "In file(s)", _grep_files, 0, show_rel);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory)
			strcpy(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else {
		sprintf(pr, "In buffers matching pattern (default: this buffer): ");
		comp_read(filepat, pr, b_match,
				  ((preserve_filename_case & 4) ? FM_FOLD : 0), bufname);
		if (!*filepat)
			strcpy(filepat, bufname);
	}
	return 1;
}

grep_make_names(rel, abs, s, inbufs)	// put name to use in grep results
char *rel, *abs, *s;				// for any matches in rel
{
	if (inbufs) {
		save_var bufname = s;
		if (*filename && force_save_as != FSA_NEWFILE)
			relative(filename, rel);
		else
			sprintf(rel, BUFPREFIX "%s", s);
	} else
		relative(s, rel);
	strcpy(abs, rel);
	if (!inbufs)
		absolute(abs);
}

count_instances(s)	/* how many times does s occur before pt in this */
char *s;		/* search? */
{
	int count = 0;

	save_var point, narrow_end = size() - point;
	search(-1, "\nFound ");
	save_var narrow_start = point;
	if (strlen(s) < 2)		// Guard against null inserted into grep buffer.
		return 0;
	for (; search(1, s); point--)
		count++;
	return count;
}

do_file_replace(str, with, flags, filepat, inbufs)
char *str, *with, *filepat;	// (query-) replace across files/buffers
{
	int res, oldbuf = 1, prevbuf = 0, total = 0, cnt = 0, skipped = 0;
	char rel[FNAMELEN], abs[FNAMELEN], *s, *(*finder)(), msg[FNAMELEN];
	int total_found = 0, total_changed = 0, total_readonly = 0, old_readonly;

	save_var synchronous_file_io = 1;
	old_readonly = readonly_warning;
	save_var readonly_warning;			// Don't show msg or ring bell
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL); // unless matches.
	save_var abort_file_matching = ABORT_ERROR;
	save_var bufnum, window_bufnum;
	if (file_read_kibitz & 4)
		save_var file_read_kibitz = 0;
	finder = give_grep_finder(inbufs, filepat, &s);	// Get 1st match
	for (; s; s = (*finder)(filepat, 0)) {
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("Searching %s...", rel);
		if (!inbufs) {
			oldbuf = look_file(abs);
			if (std_find_it(abs) == EREADABORT)
				user_abort = 1;
		} else
			to_buffer(s);
		undo_mainloop();
		if (!user_abort) {
			res = one_file_replace(str, with, flags, oldbuf, old_readonly);
			total_found += replace_num_found;
			total_changed += replace_num_changed;
			total++;
			if (res == -2)
				total_readonly++;
			else if (res >= 0)
				cnt++;
		} else {
			res = DSABORT;
			user_abort = 0;
		}
		if (prevbuf && buf_in_window(prevbuf) == -1) {
			buf_delete(prevbuf);	// delete untouched new bufs
			prev_forget(_prev_num - 1); // fix default
		}
		prevbuf = (oldbuf || modified) ? 0 : bufnum;
		if (res == 0 && (flags & QUERY) && key == '!'
			&& ask_continue_replace("Replace without asking"))
			flags &= ~QUERY;	// no more querying
		else if (res == DSABORT || res == 1 && ungot_key != -1) {
			if (!ask_continue_replace("Continue replacing"))
				break;
		} else if (res > 0) {	// User said to stop, or bad pattern.
			if (res == DSBAD)
				show_replace(str, with, flags | BAD);
			user_abort = 0;
			return;
		}
		ungot_key = -1;
		if (user_abort)
			break;
	}
	ungot_key = -1;
	user_abort = 0;
	sprintf(msg, "Replaced %d of %d matches in %d of %d %s",
			total_changed, total_found, cnt, total,
			inbufs ? "buffers" : "files");
	if (skipped) {
		show_text(-2, 0, "Skipped %d %s%s due to grep-ignore-file- variables.",
				  skipped, inbufs ? "buffers" : "files", (skipped == 1) ? "" : "s");
		sprintf(msg + strlen(msg), ", %d skipped", skipped);
	}
	if (total_readonly)
		sprintf(msg + strlen(msg), ", %d read-only", total_readonly);
	if (skipped || total_readonly)
		sprintf(msg + strlen(msg), ", see %s buffer", MSG_BUFFER);
	say("%s.", msg);
	restore_vars();
	if (prevbuf && buf_in_window(prevbuf) == -1) {
		buf_delete(prevbuf);	// delete untouched new bufs
		prev_forget(_prev_num - 1); // fix default
	}
}

general_help()
{
	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_contents("");		// Show Epsilon's contents topic.
				return;
			case HELP_FORMAT_HTML:
				if (try_html_help_string("") == 1)
					return;
		}
	prompt_for_help();
}

make_var_help(var, buf, win, strip_links)	// make help on var with
char *var;			// index i using values in buf, win
{					// maybe remove links to other topics from help text
	char *quotes, *val;
	int i, start, is_number;
	char tmp[FNAMELEN * 2];	// temp loc for values

	i = find_index(var);
	is_number = vartype_class(i) == 1;
	save_var window_handle = win;	// if buffer/window-specific,
	save_var bufnum = buf;		// get value from before this
	val = get_var_val(i, tmp, 0); // command
	restore_vars();

	save_var bufnum = zap(HELPBUF);
	if (!i || !vartype(i)) {
		buffer_printf(HELPBUF, "%s is not a variable.", var);
		return;
	} else if (!get_doc(1)) {
		say("");
		buffer_printf(HELPBUF, "\n\nCan't find edoc file.");
	} else if (name_help(i) < 0) {
		buffer_printf(HELPBUF, "\n\nNo information on %s.", var);
	} else {
		point = name_help(i);
		start = point;
		save_var abort_searching = 0;
		do_searching(REGEX, "^!~");
		xfer(HELPBUF, start, point);
	}
	bufname = HELPBUF;
	point = 0;
	if (strip_links)
		remove_links();
	to_end_line();	// add to end of 1st help line
	bprintf("%s%s variable %s.\n", current_column() ? " " : "",
		variable_type_name(vartype(i)), var);
	replace(0, toupper(character(0)));
	quotes = is_number ? "" : "\"";
	bprintf("Current value: %s%s%s",
		quotes, val ? val : "<Not available>", quotes);
	switch (name_type(i)) {	// has a default value too?
		case NT_BUFVAR:
		case NT_WINVAR:	// show default value
			val = get_var_val(i, tmp, 1);
			bprintf("%sefault value: %s%s%s",
				is_number ? ", d" : "\nD", quotes,
				val ? val : "<Not available>", quotes);
	}
	point = 0;
	delete_final_newlines();
}

color_html_attributes(int script, int orig, int limit)
{
	int len, start, lang;

	if (script == 2)		// A <jsp:> tag.
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x10, HT_JSCRIPT));
	else if (script == 3)
		lang = HT_CSS;
	else if (script == 4)
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x20, HT_JSCRIPT));
	else if (script)
		lang = get_html_asp_language(html_other_coloring);
	for (;;) {
		if (curchar() == '<' && parse_string(1, "%<%!---")) {
			point = matchend;
			color_html_comment();
			re_search(1, "[ \t\n]*");	// Parse whitespace.
		}
		if (!(len = parse_string(1, "[-_./a-z0-9]*")))	// An attribute?
			break;
		if (script == 1 && parse_string(1, "(language|type)[ \t\n]*=[ \t\n]*"
										"[\"']*(text/)?")) {
			start = matchend;
			if (matches_at(start, 1, "(javascript|jscript|ecmascript)"))
				lang = HTML_SPECIFIC_COLORING(1, HT_JSCRIPT);
			else if (matches_at(start, 1, "(vb)"))
				lang = HTML_SPECIFIC_COLORING(2, HT_VBSCRIPT);
			else if (matches_at(start, 1, "(php)"))
				lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
			else if (matches_at(start, 1, "(python)"))
				lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
		}
		if (character(point + len) == ':'
			  && matches_at(point + len, 1, ":[-_./a-z0-9]*")) {
			set_character_color(point, point + len + 1, color_class html_namespace);
			point += len + 1;
			len = matchend - point;
		}
		set_character_color(point, point + len, color_class html_attribute);
		point += len;			// Followed by a value?
		if (len = parse_string(1, "[ \t\n]*=[ \t\n]*")) {
			set_character_color(point, point + len, color_class html_punctuation);
							// Color the = and whitespace.
			point += len;		// Value might be quoted.
			start = point;
			if (curchar() == '"') {
				point++;
				html_find_matching_quote('"');
			} else if (curchar() == '\'') {
				point++;
				html_find_matching_quote('\'');
			} else	// No quotes, assume it's whitespace-delimited.
				re_search(1, "[^ \t\n<>]*");
			html_color_value(start, point);
		}
		re_search(1, "[ \t\n]*");	// Parse whitespace.
		set_character_color(matchstart, matchend, color_class html_punctuation);
	}
	if (curchar() == '>') {
		set_character_color(point, point + 1, color_class html_punctuation);
		point++;
	}
	start = point;
	if (script && lang != HT_IGNORE && !parse_string(-1, "/%>")) {
		if (re_search(1, script == 3 ? "%</style%>"
					  : script == 2 ? "%</jsp:.*%>" : "%</(cf)?script(text)?%>")) {
			set_tagged_region("color-as-unit", orig, point, 99);
			point = matchstart;
		} else
			set_tagged_region("color-as-unit", orig, point, 99);
		color_embedded_code(start, point, lang, limit);
	}
}

color_html_comment()
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (!sublanguage_code && html_recognize_coldfusion_comments
			   && parse_string(1, "%<%!---") && !parse_string(1, "%<%!----%>")) {
		int level = 0;	// A legal ColdFusion comment.
		while (re_search(1, "%<%!---|---%>"))
			if (character(matchstart) == '<')
				level++;
			else if (--level <= 0)
				break;
		set_character_color(s, point, color_class html_altcomment);
		set_tagged_region("color-as-unit", s, point, 99);
		return;
	} else if (parse_string(1, "%<[ \t\n]*%![ \t\n]*--")) {
		point = matchend;		// A legal comment.
		re_search(1, "--[ \t\n]*>");	// Search for end of legal comment.
	} else if (parse_string(1, "%<%!DOCTYPE[^][<>]*%[")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		return;
	} else if (parse_string(1, "%<%!(ENTITY|ELEMENT|ATTLIST)[ \t\n]*")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		s = point;
		search(1, ">");
		set_character_color(s, point, color_class html_value);
		if (character(point - 1) == '>')
			set_character_color(point - 1, point, color_class html_punctuation);
		return;
	} else if (parse_string(1, "%<%!DOCTYPE")) {
		if (re_search(1, "[[>]") && character(matchstart) == '[') {
			search(1, "]");
			search(1, ">");
		}
		set_character_color(s, point, color_class html_value);
		return;
	} else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

color_html_loop(int end)
{
	int s, t, i, script;

	for (s = point; point < end; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							 ? "([#a-z0-9]+);?"
							 : "([a-z0-9]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()) && color_html_include(end))
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = (s == 6 && parse_string(1, "script")
					  || s == 10 && sublanguage_code == SUBLANG_XML
					  && parse_string(1, "scripttext"));
			if (s == 3 && parse_string(1, "jsp:(declaration|expression"
									   "|scriptlet)</word>"))
				script = 2;
			if (s == 5 && parse_string(1, "style"))
				script = 3;
			if (s == 8 && parse_string(1, "cfscript"))
				script = 4;
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			for (i = point; strchr(" \t\n", curchar()); )
				point++;	// Parse whitespace.
			set_character_color(i, point, color_class html_punctuation);
			if (parse_string(1, "/[ \t\n]*>")) {
				set_character_color(point, point + 1, color_class html_tag);
				set_character_color(point + 1, matchend,
									color_class html_punctuation);
				point = matchend;
			} else
				color_html_attributes(script, t, end);
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
}

html_find_linked_file(char *file)
{
	char buf[FNAMELEN], *p;
	int orig;

	if (!re_search(1, "((href|src)[ \t\n]*=|@import)[ \t\n]*")) {
		to_begin_line();
		if (!re_search(1, "((href|src)[ \t\n]*=|@import)[ \t\n]*"))
			error("Can't find any link here.");
	}
	orig = point;
	if (curchar() == '"') {
		point++, orig++;
		search(1, "\"");
	} else if (curchar() == '\'') {
		point++, orig++;
		search(1, "'");
	} else	// No quotes, assume it's whitespace-delimited.
		re_search(1, "[^ \t\n<>]*"), point++;
	grab(orig, point - 1, buf);		// Get link from buffer.
	p = strchr(buf, '#');		// Remove #fragment if any.
	if (p)
		*p = 0;
	if (!*buf)
		error("Can't follow fragment links.");
	point = orig;
	if (parse_string(1, "[A-Za-z]+:"))
		strcpy(file, buf);					// An absolute URL.
	else if (parse_string(1, "/")) {		// An absolute URL path.
		strcpy(file, filename);
		strcpy(get_url_file_part(file, 1), buf);
	} else {								// A relative URL path.
		strcpy(file, filename);
		if (string_matches_regex(file, "[a-z]+:[/\\][/\\][^\n /\\]+<eof>", 1))
			sprintf(file + strlen(file), "/%s", buf);
		else
			strcpy(get_tail(file, 1), buf);
	}
}

xml_find_linked_file(char *file)
{
	html_find_linked_file(file);
}

html_move_level(int dir, int stop_on_key)
{
	char *findch = (dir > 0) ? "<([" : ">)]";
	char *otherch = (dir < 0) ? "<([" : ">)]";

	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point - (dir < 0))) {
		case HT_JSCRIPT:
			return c_move_level(dir, stop_on_key);
		case HT_VBSCRIPT:
		case HT_PYTHON:
		case HT_CSS:
			findch = (dir > 0) ? "[({" : "])}";
			otherch = (dir < 0) ? "[({" : "])}";
			break;
		case HT_PHP:
			save_var sublanguage_code = SUBLANG_PHP;
			return perl_move_level(dir, stop_on_key);
	}
	return default_move_level(dir, findch, otherch);
}

html_indent_in_tag(int pos, int cur)
{
	int attrlen = 0, tagind;
	if (!(html_indenting_rules & 0x100))
		return 0;
	point += parse_string(1, "[ \t]*");
	// Get width of "attname=" on current line.
	if ((html_indenting_rules & 0x8000) && parse_string(1, HTML_NAME "[ \t]*="))
		attrlen = get_column(matchend) - current_column();
	save_var point = pos;
	if (parse_string(-1, "(^|\n)[ \t]*")) {
		// Get indentation of "<tagname".
		tagind = get_indentation(point);
		point += parse_string(1, "[<]" HTML_NAME "[ \t]*");
		if (attrlen && parse_string(1, "[ \t\n]*" HTML_NAME "[ \t]*=")) {
			int i = get_column(matchend) - attrlen;	// Indent to align ='s.
			if (i >= 0)		// Always indent at least to level of tag name.
				return MAX(i, tagind + 1);
		}		// Otherwise, don't align ='s.
		return current_column();
	}
	restore_vars();
	save_var point;
	if (!re_search(-1, "[^ \t\n]"))
		return cur;
	return get_indentation(point);
}

// Search forward counting start and end tags, and return difference in
// counts.
int html_tag_level()
{
	int level = 0;

	if (parse_string(1, "[ \t]*[<]/") && html_is_start_tag(matchend - 2, 1))
		level++;
	while (do_color_searching(REGEX, "<c:html-punctuation>[<]") == 1)
		switch (html_is_start_tag(point - 1, 1)) {
			case -1:	level--; break;
			case 1:		level++; break;
		}
	if (html_indenting_rules & 0x80)
		level = (level > 0) ? 1 : (level < 0) ? -1 : 0;
	return level;
}

// Return indent of this HTML line by examining start and end tags.
int html_indent_level(int manual)
{
	char tag[FNAMELEN];
	save_var case_fold = (sublanguage_code != SUBLANG_XML);
	int ind = get_indentation(point), i, level = 0;
	i = html_in_tag();
	if (i >= 0)
		return html_indent_in_tag(i, ind);
	if (!(html_indenting_rules & 0x200) && !parse_string(1, "[ \t]*[<]"))
		return 0;
	// Indent end tag if due to pressing tab, or to reindent previous
	// line option, but not when auto-fill moves an end tag down.
	if ((manual || indenter_action != IACT_AUTO_FILL)
		  && parse_string(1, "[ \t]*[<]/"))	// Indenting an end tag.
		return html_indent_of_matching_tag(matchend - 2, ind);
	if (sublanguage_code != SUBLANG_XML	// Special rules for some tags.
		  && parse_string(1, "[ \t]*%<(" HTML_NAME ")</" HTML_NAME_CHAR ">")) {
		grab(find_group(1, 1), find_group(1, 0), tag);
		if (!strfcmp(tag, "option") && (i = html_open_option(point)) >= 0)
			return get_indentation(i);
		if (!strfcmp(tag, "li") && (i = html_open_li(point)) >= 0)
			return get_indentation(i);
		if ((!strfcmp(tag, "dd") || !strfcmp(tag, "dt"))
			  && (i = html_open_dd_dt(point)) >= 0)
			return get_indentation(i);
		if (!strfcmp(tag, "colgroup"))
			level += html_colgroup_adjust();
	}
	save_var point;
	if (do_color_searching(REGEX | REVERSE, "<^c:html-comment>[^ \t\n]") != 1)
		return 0;
	save_var narrow_end = size() - give_end_line();
	to_indentation();
	while ((i = html_in_tag()) >= 0 && i < give_begin_line()) {
		point = i;	// Go to start of tag.
		to_indentation();
	}

	ind = current_column();
	level += html_tag_level();
	return ind + level * give_html_indent();
}

do_html_indent(int manual)
{
	int region, start;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 1);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	if (region > 0 && region <= HT_CSS && !IS_ENTER_KEY(key)
		  && !(html_indenting_rules & (1 << (region - 1)))) {
		prev_indenter();	// Can move back to prev line's indent.
		return;
	}
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	if (region >= HT_JSCRIPT && region < HT_IGNORE) {
		if (all_blanks(narrow_start, point)) {	// Start of script block.
			narrow_start = 0;	// 0x400 indent; 0x800 indent even if <% has none.
			int i = get_indentation(start);
			if ((html_indenting_rules & 0x400)
				  && ((html_indenting_rules & 0x800) || i))
				indent_to_column(i + give_html_indent());
			return;
		}
		html_exclude_script_head();
	}
	switch (region) {
		case HT_JSCRIPT:
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			do_vbasic_indent();
			return;
		case HT_PHP:
			do_php_indent();
			return;
		case HT_PYTHON:
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

char *html_comment_begin_override()
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			int start;
			if (get_character_color(point, &start) == color_class perl_comment)
				if (matches_at(start, 1, "//"))
					return "// ";
				else if (character(start) == '#')
					return "# ";
			switch (php_comment_style) {
				case 1: return "# ";
				case 2: return "// ";
			}
			return "/* ";
		case HT_JSCRIPT:
			return new_c_comments ? "// " : "/* ";
		case HT_VBSCRIPT:
			return "' ";
		case HT_PYTHON:
			return "# ";
		case HT_CSS:
			return "/* ";
	}
	return 0;
}

int html_restrict_break(int pos)
{
	if (get_character_syntax_color(pos) == color_class html_value) {
		int start;
		get_character_color(pos, &start);
		return !strchr("\"'", character(start));
	}
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_fill_combine
												: html_auto_fill_combine;
	if (var && parse_string(-1, "^[ \t]*%<[ \t]*(" HTML_NAME ")[ \t]*")
		  && find_group(1, 0) - find_group(1, 1) <= var)
		return 0;
	return 1;
}

int html_maybe_break_line(int mode)
{
	char *var = (sublanguage_code == SUBLANG_XML) ? &xml_auto_fill_mode
				: &html_auto_fill_mode;
	switch (mode) {
		case 0:
			*var &= ~1;
			return 0;
		case 1:
			*var |= 1;
			return 0;
		case 2:
			return *var & 1;
	}
	if (!(*var & 1))		// We're not in auto-fill mode.
		return 1;
	recolor_buffer_range(point, point + 1);
	int have = 16;	// Assume we're in scripting.
	if (html_region_type(point) == -1)	// We're not. Comment, tag, or text?
		have = (get_character_color(point - 1) == color_class html_comment)
			   ? 8 : (html_in_tag() == -1) ? 2 : 4;
	if (!(have & *var))
		return 1;	// Only do one sort of filling?
	switch (html_region_type(point)) {
		case HT_PHP:			 // Break only if in a comment.
		case HT_VBSCRIPT:
		case HT_PYTHON:
		case HT_CSS:
			return do_maybe_break_line(0);
		case HT_JSCRIPT:
			return do_maybe_break_line(1);
	}
	save_var mode_restrict_break = html_restrict_break;
	return maybe_break_this_line();	// Break always as text.
}

tag_mode_jscript()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (re_search(1, "(^|;)[ \t]*(var|function)[ \t]+(<word>+)")) {
		int var = tolower(character(find_group(2, 1))) == 'v';
		grab(start = find_group(3, 1), find_group(3, 0), func);
		if (!isdigit(*func))
			add_tag(func, start);
		if (var)
			for (;;) {
				point += parse_string(1, "([ \t\n]|//.*)*");
				if (curchar() == '=')
					re_search(1, "![\n,;]");
				if (!parse_string(1, ",[ \t\n]*(<word>+)"))
					break;
				grab(start = find_group(1, 1), find_group(1, 0), func);
				point = matchend;
				if (!isdigit(*func))
					add_tag(func, start);
			}
		else
			c_skip_level('{', '}');
	}
}

suffix_ascx()
{
	html_mode();
}

command kill_rectangle()
{				/* save, maybe replace w/ spaces */
	pick_rectangle(1, (kill_rectangle_removes != 0) != has_arg ? 1 : 2);
}

int set_font_for_unix()
{
	char buf[FNAMELEN];

	if (no_mac_dialog_helper()) {
		mac_gtk_message(1);
		return simple_set_font_for_unix();
	}
	get_executable_file(buf, "epsilon-dialogs", 1);
	sprintf(buf + strlen(buf), " --parent %x --font \"%s\"",
			screen_to_window_id(0), font_fixed);
	if (!pipe_text("-epsilon-dialog-out", "-epsilon-dialog-out", buf, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF, 0))
		file_error(errno, "Couldn't run dialog helper", "Error");
	save_var bufname = "-epsilon-dialog-out";
	point = 0;
	if (parse_string(1, ".*(ibrary|shared)")) {
		mac_gtk_message(1);
		return simple_set_font_for_unix();
	} else if (re_search(1, "^=") && parse_string(1, ".*", buf))
		font_fixed = buf;
	return 1;			// User picked one, even if it was bad.
}

// Display a message explaining how to install needed library, trying
// to set font or colors.
mac_gtk_message(int is_font)
{
	int b = zap("-epsilon-dialog-error");
	buf_printf(b, "This command requires GTK+ 1.4 shared libraries, which ");
	buf_printf(b, "are not\ninstalled.  Under OS X, install the fink application from ");
	buf_printf(b, "\nfink.sourceforge.net, then run\n\n");
	buf_printf(b, "   fink install gtk+-shlibs\n\n");
	buf_printf(b, "On other platforms, locate and install the above libraries\n");
	buf_printf(b, "to enable this dialog.\n\n");
	if (is_font)
		buf_printf(b, "Epsilon will revert to a simpler font setting method.\n");
	else {
		buf_printf(b, "For now, you can set colors by running the Alt-x ");
		buf_printf(b, "export-colors\ncommand, editing the source file it ");
		buf_printf(b, "generates, and loading it\nusing the compile-buffer ");
		buf_printf(b, "command on Alt-F3.\n");
	}
	save_var _view_title = " Can't run dialog helper ";
	view_buf(b, 0);
}

color_postscript_range(from, to)
{
	int t = -1, talk, r, talk_at = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();	// Move back over continued lines.
	point -= parse_string(-1, "(.*\\\n)+");
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &r) != -1)
		point = r;		// Move before binary blocks.
	talk = (to - point > 1000000);	// Show status during long delays.
	while (point < to) {
		if (!re_search(1, "[%(]|/[-_a-z0-9]+|</word>(doclutimage|doNimage)</word>")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(t)) {		// check first char
			case '%':
				nl_forward();
				set_character_color(t, point, color_class postscript_comment);
				if (parse_string(-1, "^%%%%Begin(Binary|Data).*\n")) {
					r = point;
					if (re_search(1, "\n%%%%End(Binary|Data)"))
						point = matchstart;
					set_character_color(r, point, color_class postscript_text);
					set_tagged_region("color-as-unit", t, point, 1);
				}
				break;
			case '(':		// A text string.
				point--;
				fwd_postscript_string();
				set_character_color(t, point, color_class postscript_text);
				break;
			case '/':		// Found /literalname.
				set_character_color(t, point, color_class postscript_literal);
				break;
			default:		// Found start of Windows PS driver binary code.
				re_search(1, "~>[ \n]*(Z|endimage)");	// Find end.
				set_character_color(t, point, color_class postscript_text);
				set_tagged_region("color-as-unit", t, point, 1);
				break;
		}
		if (talk && point > talk_at + 20000) {
			note("Coloring PostScript: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

lose_focus() on reg_tab[LOSEFOCUS], cx_tab[LOSEFOCUS], cc_tab[LOSEFOCUS]
{
	invisible_cmd();
}

win_resize() on reg_tab[WIN_RESIZE], cx_tab[WIN_RESIZE], cc_tab[WIN_RESIZE]
{
	invisible_cmd();
}

lost_selection() on reg_tab[LOST_SELECTION], cx_tab[LOST_SELECTION],
					cc_tab[LOST_SELECTION]
{
	invisible_cmd();
}

mouse_leaving_window() on reg_tab[WIN_MOUSE_LEAVES], cx_tab[WIN_MOUSE_LEAVES],
						  reg_tab[WIN_MOUSE_ENTERS], cx_tab[WIN_MOUSE_ENTERS],
						  cc_tab[WIN_MOUSE_ENTERS], cc_tab[WIN_MOUSE_LEAVES]
{
	invisible_cmd();
}

mouse_get_another() on cx_tab[MOUSE_MOVE], cc_tab[MOUSE_MOVE]
{
	invisible_cmd();
	while (key == MOUSE_MOVE) {
		table_prompt();
		getkey();
	}
	say("");
	do_again();
}

when_loading()
{
	cc_tab[GETFOCUS] = cx_tab[GETFOCUS];
	cc_tab[WIN_MOUSE_ENTERS] = cx_tab[WIN_MOUSE_ENTERS];
}

shell_fwd_quote(int c)
{
	char *pat;
	int t;

	switch (c) {
		case '\'':	pat = "['\\]"; break;
		case '"':	pat = "[`\"\\]|%$%("; break;
		case '`':	pat = "[`\"'\\]|%$%("; break;
		case '(':	pat = "[)\"'`]|%$%(|\\![\"'`]";
					c = ')'; break;
		default:	say("Internal error: shell_fwd_quote('%c')", c);
					return 0;
	}
	while (re_search(1, pat)) {
		t = character(point - 1);
		if (t == '\\')
			point++;
		else if (t == c)
			return 1;
		else if (!shell_fwd_quote(t))
			return 0;
	}
	return 0;
}

// At a $(, starting at p, so find matching ), and color through it,
// moving past.
shell_parse_substitution(int p)
{
	shell_fwd_quote('(');
	set_character_color(p, point, color_class perl_constant);
}

// Color simple variable references within this string-like thing (already
// colored itself).
color_perl_interpolated(int from, int to, int comments)
{
	int p;

	save_var point = from;
	save_var narrow_end = size() - to;
	while (re_search(1, (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					 ? "[\\$]" : "[\\$@#]")) {
		p = point;
		if (character(point - 1) == '\\')
			point++;
		else if (character(point - 1) == '#') {
			if (!comments)
				continue;
			from = point - 1;
			to_end_line();
			if (point > to)
				point = to;
			set_character_color(from, point, color_class perl_comment);
		} else if (sublanguage_code == SUBLANG_PHP) {
			if (parse_string(-1, "[^\\]{%$")) {
				point -= 2;
				to_default_move_level(1, "{", "}", 1);
				set_character_color(p - 2, point, color_class perl_variable);
			} else if (parse_string(1, "<word>+(-%><word>+)*|{<word>+}")) {
				point = matchend;
				if (curchar() == '[')
					to_default_move_level(1, "[", "]", 0);
				set_character_color(p - 1, point, color_class perl_variable);
			}
		} else if (in_shell_buffer) {
			if (curchar() == '(') {
				shell_parse_substitution(point - 1);
			} else if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									"|[^ \t\n`]"))
				set_character_color(matchstart - 1, point = matchend,
									color_class perl_variable);
		} else if (curchar() == '{') {
			to_default_move_level(1, "{", "}", 1);
			set_character_color(p - 1, point, color_class perl_variable);
		} else if (parse_string(1, "#?([0-9]+|[A-Za-z_][0-9A-Za-z_]*)|%^[A-Z]"
								"|[]-[&`'+*|%=~^_./,\\\";:#?!@$<>()]") && point < to) {
			point = matchend;
			for (;;) {
				if (curchar() == '-' && character(point + 1) == '>'
					  && (isalnum(character(point + 2))
						  || strchr("[{", character(point + 2))))
					point += 2;
				else if (curchar() == '[' || curchar() == '{') {
					if (to_default_move_level(1, "[{", "]}", 1) < 0)
						break;
				} else
					break;
			}
			set_character_color(p - 1, point, color_class perl_variable);
		}
	}
}

color_perl_fwd_quote(int c, int start, int interpolated)
{
	if (in_shell_buffer)
		shell_fwd_quote(c);
	else
		perl_fwd_quote(c);
	set_character_color(start, point, color_class perl_string);
	if (interpolated)
		color_perl_interpolated(start + 1, point - 1, 0);
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
			   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				if (in_shell_buffer && !parse_string(-1, "([ \n\t|&;()<>]|^)#"))
					break;
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (curchar() == '(') {
						shell_parse_substitution(point - 1);
						break;
					}
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "[#$@]*((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_variable);
				else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer) {
					if (!(matches_at_length(t, -1, "\\+") & 1)) {
						search(1, "'");	// No \-quoting in '' for shell mode.
						set_character_color(t, point, color_class perl_string);
					}
				} else if (!strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && (matches_at_length(t, -1, "\\+") & 1))
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_><word>*)"
									 : "(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

command process_complete()
{
	char pat[FNAMELEN], pref[FNAMELEN];
	int cnt, start, end, buf = create(_MATCH_BUF);
	int offset = IS_SHIFT_KEY(full_key) ? -1 : 1;

	start = end = point;
	if (start < type_point) {
		if (!maybe_indent_rigidly(0))
			indent_like_tab();
		return;
	}
	*pat = 0;
	if (process_line_inside_quoting() || parse_string(-1, "[^\n\t <>=\"]+", pat)) {
		start = point;
		point = end;
		if (process_line_inside_quoting())
			grab(start = matchend + 1, end, pat);
		if (start < type_point && type_point < end) {
			start = type_point;		// exclude prompt
			grab(start, end, pat);
		}
	}
	if (!*pat)
		strcpy(pat, "*");
	if (!(process_completion_style & 1) || prev_cmd != CMD_PROC_COMPL) {
		int dircmdlen = matches_at_length(start, -1, process_completion_dircmds);
		int dirsonly = dircmdlen && ((dircmdlen == start - type_point)
									 || matches_at(start - dircmdlen, -1, "[][#&;()] *"));
		int initial = all_blanks(type_point, start)
					  || matches_at(start, -1, "([&;|]|</word>(sudo)) +");
		cnt = proc_complete_do_list_matches(pat, initial,
											process_completion_style & 2,
											dirsonly ? FM_ONLY_DIRS : 0);
		check_abort();
		offset = 0;
		if (!cnt)
			error("No matches for %s.", pat);
		set_buf_point(buf, 0);
	}
	if (!(process_completion_style & 1))
		find_buffer_prefix_folding(name_to_bufnum(_MATCH_BUF), pref, !is_unix);
	else {
		buf_delete_final_newlines(buf);
		grab_line_offset(buf, pref, offset, 1);
	}
	if ((process_completion_style & 1) || strlen(pref) > end - start) {
		delete(start, end);
		if (strchr(pref, ' ') && !process_line_inside_quoting())
			stuff("\"");
		stuff(pref);
		if ((process_completion_style & 1)) {
			if (is_path_separator(character(point - 1)))
				delete(point - 1, point);
			this_cmd = CMD_PROC_COMPL;
			return;
		}
		if (cnt == 1 && !is_path_separator(character(point - 1))) {
			if (strchr(pref, ' ') && process_line_inside_quoting())
				stuff("\"");
			stuff(" ");
		}
	} else if (prev_cmd == CMD_PROC_COMPL) {
		proc_complete_show_matches();
		this_cmd = CMD_PROC_COMPL;
	} else if (cnt == 1 && !is_path_separator(character(point - 1))) {
		if (strchr(pref, ' ') && process_line_inside_quoting())
			stuff("\"");
		stuff(" ");
	} else if (!(process_completion_style & 4)
			   && buffer_size(_MATCH_BUF) < screen_cols - 20 - strlen(pat)) {
		save_var bufname = _MATCH_BUF;
		simple_re_replace(1, "\n", " ");
		grab(0, size(), pref);
		say("%d matches for %s: %s", cnt, pat, pref);
		expire_message = 1;
	} else {
		say("%d matches for %s: press again to see matches.", cnt, pat);
		expire_message = 1;
		this_cmd = CMD_PROC_COMPL;
	}
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end)) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		} else if (IS_MOUSE_SINGLE(key)) {
			ret = 1;
			ungot_key = full_key;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
							show = 1;
							goto ask;
						}
						if (!IS_ENTER_KEY(key))
							ungot_key = full_key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(str, with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
				case ',': case '^':	skip = 1; goto ask;
				case '.':		ret = 1; goto done;
				case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
				  replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

get_replace_key(flags, str, skip)	// show match, then get a key
char *str;
{
	int oreg_type = region_type();

	if (!skip) {	// make sure we're in the right place given current flags
		save_var abort_searching = 0;
		point = matchstart;
		if (!do_color_searching(flags, str))
			return 0;
	}
	save_var mark = matchstart, _highlight_control;
	set_region_type(REGNORM);
	highlight_on();
	save_var matchstart, matchend;	// refresh could clobber this
	if (display_column > 0
		  && get_column((flags & REVERSE) ? mark : point) < text_width())
		display_column = 0;	// Position to left edge if whole match fits.
	do {
		maybe_refresh();		/* show the selection */
		getkey();
	} while (!IS_TRUE_KEY(key) && !IS_MOUSE_SINGLE(key));
	if (key == CTRL('^'))
		full_getkey("", CTRLIFY_KEY);
	user_abort = 0;
	restore_vars();
	set_region_type(oreg_type);
	if (!skip) {	// refresh could've clobbered regex state info
		save_var abort_searching = 0;	// so repeat srch
		point = matchstart;
		if (!do_color_searching(flags, str))
			return 0;
	}
	return 1;
}

char *get_var_val(i, tmp, def)	/* return value of variable with index i */
char *tmp;			/* (maybe in tmp), or 0 if bad type */
{
	char *p = tmp;

	save_var use_default = def;
	switch (vartype(i)) {
		case TYPE_BYTE:
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT:
			sprintf(tmp, "%d", get_num_var(i));
			break;
		case TYPE_CPTR:
			p = get_str_var(i);
			return p ? p : "<Null Pointer>";
		case TYPE_CARRAY:
			strncpy(tmp, varptr(i), ptrlen(tmp) - 1);
			break;
		default:
			return 0;
	}
	return p;
}

do_setting(int i, char *tmp, int def)
{
	save_var use_default = def;
	switch (vartype(i)) {
		case TYPE_BYTE:
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT:
			if (*tmp)
				set_num_var(i, exptoi(tmp));
			break;
		case TYPE_CARRAY:
			strncpy(varptr(i), tmp, ptrlen(varptr(i)) - 1);
			int j = ptrlen(varptr(i)) - strlen(tmp);
			if (j > 0)
				set_chars(varptr(i) + strlen(tmp), 0, j);
			else
				say("%s truncated: %d characters, room for %d.",
					  name_name(i), strlen(tmp), ptrlen(varptr(i)) - 1);
			break;
		case TYPE_CPTR:
			set_str_var(i, strsave(tmp)); break;
	}
}

char *variable_type_name(i)	// translate type code into text
{
	switch (i) {
		case TYPE_BYTE:		return "byte";
		case TYPE_CHAR:		return "character";
		case TYPE_SHORT:	return "short integer";
		case TYPE_INT:		return "integer";
		case TYPE_CARRAY:
		case TYPE_CPTR:		return "string";
	}
	return "other";
}

var_settable(char *name)	/* is variable settable? 1=number, 2=string, else 0 */
{
	return vartype_class(find_index(name));
}

vartype_class(int i)
{
	switch (vartype(i)) {
		case TYPE_BYTE: case TYPE_CHAR: case TYPE_SHORT: case TYPE_INT:
			return 1;
		case TYPE_CARRAY: case TYPE_CPTR:
			return 2;
	}
	return 0;
}

command create_variable()
{
	char res[80];
	int type = -1, ind, len = 1;

	iter = 0;
	save_var show_all_variables = 1;
	get_var(res, "Create variable", "", 0);
	if (find_index(res)
		  && !ask_yn(" Replace ", "Replace existing item with that name? ",
					 "Yes", "No", 1))
		error("No changes made.");
	do {
		sayput("Character, Short, Integer, or Array of characters"
		       " [C, S, I, A] ");
		getkey();
		check_abort();
		switch (toupper(key)) {
			case 'C': type = TYPE_CHAR; break;
			case 'S': type = TYPE_SHORT; break;
			case 'I': type = TYPE_INT; break;
			case 'A': type = TYPE_CARRAY;
				  len = get_number("Length of new variable: ");
				  break;
			default: if (IS_TRUE_KEY(key))
					 maybe_ding(bell_on_bad_key);
				 break;
		}
	} while (type < 0);
	ind = new_variable(res, NT_VAR, type, len);
	set_name_user(ind, 1);		/* make it a user variable */
	do_set_variable(ind);
}

// Append to buffer b a description of name table entry i.  Flags:
// 1=ignore non-user variables, 2=include color schemes, 4=use final
// colors in schemes, ignoring flags to build colors from others.
list_one(int b, int i, int flags)
{
	int j, last, val, type;
	char *p, *nm, *tname;

	save_var bufnum = b;
	save_var point = size();
	save_var use_default;
	nm = name_name(i);
	switch (name_type(i)) {
		case NT_COMMAND:
		case NT_AUTOLOAD:
			bprintf("%s: command\n", nm); break;
		case NT_SUBR:
		case NT_AUTOSUBR:
			bprintf("%s: subroutine\n", nm); break;
		case NT_MACRO:
			bprintf("%s: macro ", nm);
			stuff_macro(get_macro(i), 1);
			insert('\n');
			break;
		case NT_TABLE:
			list_table(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR: case NT_VAR:
			use_default = 1;
			p = varptr(i);
			type = 0;
			if ((flags & 1) && !name_user(i) && (strncmp(nm, "-our-", 5)
				|| !strstr(nm, "-scheme")))
				break;
			switch (vartype(i)) {
				case TYPE_BYTE:	val = *(byte *) p;
								tname = "byte ";
								break;
				case TYPE_CHAR:	val = *(char *) p;
								tname = "char ";
								break;
				case TYPE_SHORT: val = *(short *) p;
								 tname = "short ";
								 break;
				case TYPE_INT:	val = *(int *) p;
								tname = "int ";
								break;
				case TYPE_CPTR:	tname = "char ptr ";
								type = 2; break;
				case TYPE_CARRAY: tname = "char array ";
								  type = 1; break;
				default:	tname = ""; type = 2; break;
			}
			bprintf("+%s %s%svariable", nm,
					(name_type(i) == NT_WINVAR)
					? "window " :
					(name_type(i) == NT_BUFVAR)
					? "buffer " : "", tname);
			if (type == 0) {
				if (val && !strncmp(nm, "-our-", 5)
					  && parse_string(-1, "^%+-our-(color|mono|gui"
									  "|unixconsole)-scheme .*"))
					bprintf(": %s\n", name_name(val));
				else
					bprintf(": %d\n", val);
			} else if (type == 2)
				stuff(": <complex>\n");
			else {
				last = ptrlen(p);
				bprintf(": %d \"", last);
				while (last > 0 && !p[last - 1])
					last--;
				for (j = 0; j < last; j++)
					switch (p[j]) {
						case '\t':	stuff("\\t"); break;
						case '\r':	stuff("\\r"); break;
						case '\n':	stuff("\\n"); break;
						case 0:		stuff("\\z"); break;
						case '\\':
						case '"':	insert('\\');
						default:	insert(p[j]); break;
					}
				stuff("\"\n");
			}
			use_default = 0;
			break;
		case NT_COLSCHEME:
			if (flags & 2)
				list_color_scheme(i, !(flags & 4));
			break;
	}
}

mark_tags(sorted)	/* update header line (adding if needed) */
{
	save_var bufnum = _tags_buffer;
	if (are_tags_sorted() != sorted) {
		delete(0, tags_header());
		bprintf("\tTAGS\t%sSORTED\n", sorted ? "" : "UN");
	}
}

command tag_files() on cx_tab[ALT('.')]
{
	char pat[FNAMELEN], *s, file[FNAMELEN], def[FNAMELEN];
	int aborted = 0;

	init_tags();
	save_var synchronous_file_io = 1;
	if (tags_from_bsc)
		error("Must tag files using the BSCMAKE program.");
	if (has_arg)
		save_var tag_declarations = 3;
	save_var recall_id = "f-match tag-files";
	get_last_recall((char *(*)()) f_match, def);
	get_file(pat, "Add/update tags for files matching", def);
	iter = 0;
	if (!(s = do_file_match(pat, STARTMATCH | EXACTONLY | FM_NO_DIRS)))
		error("No matches");
	to_another_buffer("-tag-errors");
	delete_buffer("-tag-errors");
	save_var tag_batch_mode = 1;
	if (tag_options & 1)
		clear_tags();
	for (; s; s = do_file_match(pat, EXACTONLY | FM_NO_DIRS)) {
		strcpy(file, s);
		absolute(file);
		if (aborted = do_retag_file(file))
			break;
	}
	if (!aborted) {
		do_save_tags();
		say("%s tagged.", pat);
	}
	if (exist("-tag-errors")) {
		to_buffer("-tag-errors");
		discardable_buffer = 1;
	}
}

int find_tagger(int run)
{
	char subr[FNAMELEN], mode[FNAMELEN], *s;

	if (tags_from_bsc)
		return 0;
	*subr = 0;
	s = get_extension(filename);
	if (*s == '.')
		sprintf(subr, "tag-suffix-%s", s + 1);
	clean_mode(mode);
	if (!*subr || !is_runnable(subr))
		if (*mode)
			sprintf(subr, "tag-mode-%s", mode);
	if (!*subr || !is_runnable(subr))
		guess_mode_without_extension(subr, "tag-mode-%s");
	if (!*subr || !is_runnable(subr))
		strcpy(subr, "tag-suffix-default");
	time_begin(&tag_show_msg, 200);
	return run ? try_calling(subr) : is_runnable(subr);
}

skip_c_delims(int start, int end, int restore)
{
	int level = 0, c, orig = point, iflevel = 0, skipping = 0;
	char buf[4], pat[80];

	strcpy(buf, "X|\\");
	sprintf(pat, "[%c%c\"']|/<*|/>|(" PREPROC_LINE ")", end, start);
	while (re_search(1, pat)) {
		buf[0] = c = character(point - 1);
		if (c == '\"' || c == '\'')
			while (re_search(1, buf)
				   && character(point - 1) == '\\')
				point++;
		else if (c == '*')				/* Found block comment */
			search(1, "*/");
		else if (c == '/')		/* Found c++ one line comment */
			nl_forward();
		else if (c == start) {			/* Found char like (. */
			if (!skipping)
				level++;
		} else if (c == end) {			/* Found char like ). */
			if (!skipping) {
				if (!--level)			/* Found proper end. */
					return 0;
				else if (level < 0) {	/* Found extra closing delim, give up. */
					if (restore)
						point = orig;
					return 2;
				}
			}
		} else if (parse_string(-1, PREPROC_ELSE)) {
			skipping |= (1 << iflevel);	// Found #else or #elif line, skip through #endif.
		} else if (parse_string(-1, PREPROC_ENDIF)) {
			skipping &= ~(1 << iflevel);	// found #endif line, maybe stop skipping.
			if (iflevel > 0)			/* Ignore extra #endif's. */
				iflevel--;
		} else if (parse_string(-1, PREPROC_IF)) {
			iflevel++;				// found #if/#ifdef line
		}
	}
	if (restore)						/* Couldn't find match, give up. */
		point = orig;
	return 1;
}

add_tag(func, pos)
char *func;
{
	char rel[FNAMELEN], *file = filename, tdir[FNAMELEN];

	if (tag_relative) {		/* use relative version instead */
		put_tag_directory(tdir);
		relative(file, rel, tdir);
		file = rel;
	}
	if (tag_show_percent == 2)
		noteput("%s: adding %s...", file, func);
	else if (tag_show_percent == 1 && time_done(&tag_show_msg)) {
		noteput("Working %2d%%...", muldiv(point, 100, size()));
		time_begin(&tag_show_msg, 50);
	}
	if (tag_by_text) {
		int start;

		save_var point;
		point = pos;
		to_begin_line();
		start = point;
		to_end_line();
		buf_printf(_tags_buffer, "%s\t%s\t%d\t",
				   func, file, pos);
		buf_xfer(_tags_buffer, start, point);
		buf_printf(_tags_buffer, "\n");
	} else
		buf_printf(_tags_buffer, "%s\t%s\t%d\n", func, file, pos);
}

int tex_start_construct()
{
	int s;

	if (get_tagged_region("color-tex-as-unit", point, &s) == 1)
		return s;
	save_var point, narrow_end = size() - point;
	re_search(-1, "\n([ \t]*\n|\\def[^a-z0-9])");
	for (; re_search(1, "[%$\\]"); ) {	// Look for %, $, or \\ characters
		s = matchstart;
		switch (character(point - 1)) {
			case '\\':
				if (curchar() == '[') {
					if (!search(1, "\\]"))
						return s;
				} else
					point++;
				break;
			case '$':
				return s;
			case '%':
				to_end_line();
				break;
		}
	}
	return point;
}

// Put info for current buffer in data.
create_undo_movements_data(int data)
{
	int ocnt = undo_count(0), cnt = 0, i, flags, tmp = tmp_buf();
	char num[100];
	spot tspot;

	buf_zap(data);
	spot last = alloc_spot();
	save_var narrow_start = 0, narrow_end = 0;
	while ((flags = undo_it(1, 1, 0)) != -1) {
		if ((flags & UNDO_FLAG) && undo_flag == UNDO_READ_FILE)
			break;
		if (!undo_nearby(*last, point)) {
			buf_printf(tmp, "%d, %d from %d\n",
					   (int) alloc_spot(), point, *last);
			cnt++;
		}
		*last = point;
	}
	while (undo_count(0) > ocnt)
		if (undo_it(0, 1, 0) == -1)
			break;
	for (i = 1; i <= cnt; i++) {
		get_line_from_buffer(tmp, i, num);
		tspot = (spot) numtoi(num);
		if (!undo_nearby(*last, *tspot))
			buf_printf(data, "%d\n", *tspot);
		*last = *tspot;
		free_spot(tspot);
	}
	free_spot(last);
	buf_delete(tmp);
	set_buf_point(data, 0);
	buf_printf(data, "%d\n", point);
	set_buf_point(data, 0);
}

do_undo_movements(int dir)
{
	int data = create("-undo-movements-data"), i, lines, p;
	char num[100];

	if (prev_cmd != UNDO_MOVEMENTS)
		create_undo_movements_data(data);
	this_cmd = UNDO_MOVEMENTS;
	lines = count_lines_in_buf(data, 0);
	i = buf_position_to_line_number(data, get_buf_point(data)) + dir;
	if (i < 0 || i > lines)
		error("No more edit locations.");
	get_line_from_buffer(data, i, num);
	buf_go_line(data, i);
	p = numtoi(num);
	if ((p < narrow_start || p > size() - narrow_end && !undo_keeps_narrowing)) {
		say("Undoing outside narrowed region, removing narrowing.");
		narrow_start = narrow_end = 0;
	}
	point = p;
	say("At edit location %d of %d.", lines - i, lines);
}

int vbasic_indent_more()
{
	char kwd[FNAMELEN], pat[FNAMELEN];

	if (vbasic_indent_subroutines
		  && parse_string(1, "[ \t]*(%<[^\n>]+%>)?[ \t]*"
						  "(</word>(Private|Public|Default|Friend"
						  "|Static|Protected|Shadows|MustInherit"
						  "|NotInheritable|Overloads|Overrides|Overridable"
						  "|NotOverridable|MustOverride|Shared|ReadOnly"
						  "|WriteOnly|Widening|Narrowing)</word>[ \t]*)*"
						  "</word>(Sub|Function|Property|Class|Operator"
						  "|Module|Interface|Namespace|Structure)</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(If)</word>.*"
					 "</word><^c:*-comment|*-string>Then</word>[ \t]*[^ \t\n':]"))
		return 0;		// Old one-line form of If statement.
	if (parse_string(1, "[ \t]*</word>((private|public)[ \t]+)?"
					 "(Type|Enum|Else|With|Select|Case|Try|Catch|Finally"
					 "|SyncLock|If|Elseif)</word>")) {
		grab(find_group(3, 1), find_group(3, 0), kwd);
		sprintf(pat, ".*:[ \t]*</word><^c:*-comment|*-string>End[ \t]+%s</word>", kwd);
		if (!parse_string(1, pat)) {	// Multiple statements: ignore block.
			if (vbasic_indent_case
				  && parse_string(1, "[ \t]*</word>Select</word>"))
				return 2;		// Force extra indent for select case/case.
			return 1;
		}
	}
	if (parse_string(1, "[ \t]*</word>(For)</word>")
		  && !parse_string(1, ".*:[ \t]*<^c:*-comment|*-string></word>Next</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(Do)</word>")
		  && !parse_string(1, ".*:[ \t]*<^c:*-comment|*-string></word>Loop</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(While)</word>")
		  && !parse_string(1, ".*:[ \t]*<^c:*-comment|*-string></word>Wend</word>"))
		return 1;
	if (parse_string(1, "[ \t]*(Get|Set)[ \t]*(%([ \t]*%))?[ \t]*[\n(':]"))
		return 1;	// Get/set keyword with nothing after it, or (, starts block.
	return 0;
}

// Indent the current line less?
int vbasic_indent_less()
{
	if (vbasic_indent_case
		  && parse_string(1, "[ \t]*</word>End[ \t]+Select</word>"))
		return 2;
	if (parse_string(1, "[ \t]*</word>(End|Wend|Else|Elseif"
					 "|Loop|Next|Case|Catch|Finally)</word>"))
		return 1;
	return 0;
}

vbasic_indenter()
{
	int orig = point, ind;

	if (matches_at(give_begin_line(), 1, "[ \t]*#")
		  || point <= narrow_start) {
		indent_to_column(0);
		return;
	}
	if (vbasic_is_continuation()) {
		vbasic_indent_continuation(orig);
		return;
	}
	do {
		to_begin_line();
		if (!re_search(-1, "[^ \t\n]"))	/* Find previous non-blank line */
			break;
		to_indentation();
	} while (parse_string(1, "['#]"));	// that's not a comment or preproc.
	vbasic_before_continuation();
	ind = get_indentation(point);
	ind += vbasic_indent_more() * vbasic_step();	// Based on previous line.
	point = orig;
	ind -= vbasic_indent_less() * vbasic_step();	// Based on current line.
	to_indentation();		/* go to current line's indent */
	indent_to_column(ind);
}

int spell_download_one(char *dir, char *arch, int sz, int ask)
{
	char archpath[FNAMELEN], url[FNAMELEN], cmdline[FNAMELEN], *s;
	struct file_info fi;
	sprintf(archpath, "%s/%s", dir, arch);
	sprintf(url, "http://www.lugaru.com/%s", arch);
	s = lookpath(arch);
	if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9) {
		if (s && check_file(s, &fi) == CHECK_FILE && fi.fsize * 10 >= sz * 9)
			strcpy(archpath, s);
		else if (!ask)
			return 0;
		else if (!ask_yn("Download and install dictionary file?", 
						 "Download and install a dictionary file "
						 "(about 700 KB)? ", "Yes", "No", 1))
			return 1;
		else {
			say("Please wait...");
			if (download_file_to_disk(url, archpath, sz))
				error("Couldn't download dictionary file from %s", url);
		}
	}
	if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9)
		error("Bad dictionary archive %f, please retrieve manually.",
			  archpath);
	convert_to_8_3_filename(archpath);
	sprintf(cmdline, is_win32 ? "%s" : "\"tar\" xzf %s", archpath);
	pipe_text(NULL, "-spell-config", cmdline,
			  dir, PIPE_SYNCH | PIPE_NOREFRESH, 0);
	return 2;
}

when_loading()
{
	vhdl_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	inifile_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	makefile_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	pscript_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	conf_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	batch_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	tcl_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	vbasic_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	vhdl_tab[ALT('Q')] = (short) prefix_fill_paragraph;
	c_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	html_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	tex_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	python_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	if (!strcmp(print_destination_unix, "!lpr %f"))		// New default values.
		strcpy(print_destination_unix, "!lpr \"%f\"");
	if (recall_maximum_size < 500_000)
		recall_maximum_size = 500_000;
	if (recall_maximum_session < 500_000)
		recall_maximum_session = 500_000;
	process_next_error_options |= 7;
	mouse_selection_copies = 0;		// New default.
	drop_name("ssh-coloring-rules");	// Replaced by ssh-interpret-output.
	switch (html_auto_fill_mode) {	// Convert to new bit assignments
		case 0: html_auto_fill_mode = 30; break;	// All on but disabled.
		case 1: html_auto_fill_mode = 31; break;	// All on.
		case 2: html_auto_fill_mode = 25; break;	// Only in scripting & comments.
		case 3: html_auto_fill_mode = 15; break;	// All but scripting.
	}
	switch (xml_auto_fill_mode) {	// Convert to new bit assignments
		case 0: xml_auto_fill_mode = 30; break;	// All on but disabled.
		case 1: xml_auto_fill_mode = 31; break;	// All on.
		case 2: xml_auto_fill_mode = 25; break;	// Only in scripting & comments.
		case 3: xml_auto_fill_mode = 15; break;	// All but scripting.
	}
	drop_name("ssh-coloring-rules");
}
