/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2009 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.07 to 13.08.

#include "eel.h"
#include "c.h"
#include "session.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "proc.h"
#include "bufed.h"
#include "grep.h"
#include "html.h"
#include "perl.h"
#include "kill.h"
#include "misclang.h"
#include "tags.h"
#include "vbasic.h"
#include "vhdl.h"
#include "inimode.h"
#include "makemode.h"
#include "tcl.h"
#include "tex.h"
#include "python.h"

// Mark comments and strings in region from...to in the current C mode
// buffer.
c_lex_region(int from, int to)
{
	// If from is the in middle of an already-lexed region, go to its start.
	get_tagged_region("c-lex", from, &from);
	save_var point = from, narrow_start = 0, narrow_end = 0;
	int start, c, val, init = from, p;
	while (re_search(1, "<squote|dquote>|//|/<*>")) {
		start = matchstart;
		c = character(start);
		p = point;	// Mark that we're not in comment/string before the march.
		set_tagged_region("c-lex", init, p, 0);
		if (c == '/' && character(start + 1) == '/') {
			val = IN_NEW_COMMENT;	// Handle // comment (c_skip_item() won't).
			nl_forward();
		} else {
			point = start;
			c_skip_item(c, 0);
			switch (c) {
				case '/':	val = IN_OLD_COMMENT; break;
				case '"':	val = IN_STR_CONST; break;
				case '\'':	val = IN_CHAR_CONST; break;
				default:	say("Internal error in c_lex_region in %s at %d (%d)",
								  bufname, point, c);
							val = 0;
							break;
			}
		}
		set_tagged_region("c-lex", p, point, val);
		if (point > to)
			return;
		init = point;
	}
	set_tagged_region("c-lex", init, point, 0);	// Mark final region.
}

// Is point in a comment (C++ style or block) or a string or character
// constant?  Return codes for each, or 0.  We cache this info in a
// tagged region, and scan for it from the start of the buffer when
// needed.

in_c_comment(int lookback)
{
	UNREFERENCED_PARAMETER(lookback);
	if (in_perl_buffer)
		return in_perl_comment(point);
	int pos = point, from, to, res, f, t;
	if (modified_buffer_region(&from, &to, "c-lex")) {
		// First, discard region modified since last scan.
		if (get_tagged_region("c-lex", from, &f, &t) != -1)
			from = f;		// Include full regions at edges.
		if (get_tagged_region("c-lex", to, &f, &t) != -1)
			to = t;
		set_tagged_region("c-lex", from, to, -1);	// Set to unknown.
	}
	reset_modified_buffer_region("c-lex");

	res = get_tagged_region("c-lex", pos, &from, &to);
	if (res == -1)		// No cached data, so lex it.
		c_lex_region(from, pos);
	res = get_tagged_region("c-lex", pos, &from, &to);
	if (res == -1) {
		say("Internal error in in_c_comment: %s at %d", bufname, pos);
		return 0;
	}
	return res;
}

void set_color_handler(int activity, int buf, int from, int to);

set_color_xwin()
{
	int i, j, classes = number_of_color_classes();
	int *orig_fg, *orig_bg, *orig_schemes;
	int orig_scheme = selected_color_scheme, this_scheme = orig_scheme;
	char *s, cmd[FNAMELEN];
	int orig_class = get_character_syntax_color(point);

	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	on_exit free(orig_schemes), free(orig_fg), free(orig_bg);
	save_var bufnum;
	get_executable_file(cmd, "epsilon-dialogs", 1);
	sprintf(cmd + strlen(cmd), " --parent %x --color", screen_to_window_id(0));

	for (;;) {
		bufnum = zap("-epsilon-dialog-data");
		bprintf("%d\n%s\n", font_styles, font_fixed);
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i))
				bprintf("%s\n", name_name(i));
		bprintf("\n");
		for (i = 0; i < classes; i++)
			bprintf("%s\n", name_color_class(i));
		bprintf("\n");
		bprintf("%s\n", (s = lookpath("colclass.txt")) ? s : "");
		bprintf("%s\n", name_name(this_scheme));

		selected_color_scheme = find_index("color-defaults");
		for (j = 0; j < classes; j++)
			bprintf("default %x %x\n", get_foreground_color(j, 1),
					get_background_color(j, 1));
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i)) {
				selected_color_scheme = i;
				for (j = 0; j < classes; j++)
					bprintf("%d %d %x %x\n", i, j, get_foreground_color(j, 1),
							get_background_color(j, 1));
			}
		make_color_class_list(1);
		point = size();
		s = name_color_class(orig_class);
		bprintf("\n%s\n", s ? s : "");
		selected_color_scheme = this_scheme;
		if (!pipe_text("-epsilon-dialog-data", "-epsilon-dialog-out", cmd,
					   NULL, PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH,
					   set_color_handler))
			file_error(errno, "Couldn't run dialog helper", "Error");
		this_scheme = selected_color_scheme;
		bufname = "-epsilon-dialog-out";
		i = process_exit_status;
		if (matches_at(0, 1, ".*(ibrary|shared)")) {
			dialog_helper_failed(0);
			return;
		}
		if (i < 0 || i > 2)
			i = !ask_yn("Keep changes", "The dialog process exited "
						"unexpectedly.  Keep these settings? ",
						"Yes", "No", 0);
		if (i == 0) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			return;
		} else if (i == 1) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			return;
		} else {
			show_binding("Continue changing colors "
						 "by typing %s", "exit-level");
			recursive_edit();
		}
	}
}

dired_mark_for_copy() on dired_tab['c']
{
	if (no_remote_copy(get_dired_buffer_pattern()))
		error("Copy operation not implemented for remote files.");
	int isdir = dired_isdir();
	if (isdir == 1 || isdir == 2)
		error("Can't copy directories.");
	dired_repl('C');
}

char *dired_skip_line_pat();

int dired_getname(char *fullname, int flags)
{
	char *p, *q;

	dired_try_getdir(fullname, 1);
	p = fullname + strlen(fullname);
	dired_getindex(p, goal_column);		/* the file name */
	to_begin_line();
	if (parse_string(1, dired_skip_line_pat()))  /* on dir line, make name illegal */
		*p = 0;
	move_to_column(goal_column);
	if (!*p || isspace(*p))
		if (flags & 1)
			error("Can't read file name");
		else
			return 0;
	q = strstr(p, " -> ");			// replace file name with link dest
	if (q)
		if (flags & 2)
			*q = 0;
		else
			strcpy(p, q + 4);
	if (is_path_separator(*p) == 1	// If link dest starts with /,
		  || isalpha(p[0]) && p[1] == ':' && is_path_separator(p[2]) == 1)
		strcpy(get_url_file_part(fullname, 0), p);	// just use that.
	if (dired_format == DF_SIMPLE) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
	} else if (dired_format == DF_VMS) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
		q = strstr(fullname, ".DIR;");	// Refer to directories without
		if (q)							// .DIR extension to see contents.
			*q = 0;
		q = strstr(fullname, ";");		// Remove version number.
		if (q && isdigit(q[1]))
			*q = 0;
	}
	p += strlen(p) - 1;		// Unix file name might use "ls -F" format
	if (*p == '/' || *p == '*')
		*p = 0;
	return 1;
}

dired_isdir()
{
	int i = get_dired_item("dired-isdir-", 1);
	if (i)
		return (*(int (*)())i)();
	i = get_dired_item("dired-dirsearch-", 0);
	if (i) {
		save_var point;
		to_begin_line();
		return parse_string(1, get_str_var(i)) != 0;
	}
	if (matches_at(column_to_pos(ISDIR_START), 1, "  %<SYMLINK%>"))
		return 3;
	else if (matches_at(column_to_pos(ISDIR_START), 1,
						"  %<(JUNCTION|SYMLINKD)%>"))
		return 2;
	return matches_at(column_to_pos(ISDIR_START), 1, DIRSEARCH) != 0;
}

dired_sorter(start, end)
{
	int tmp;
	char *pat = DIRSEARCH;
	int i = get_dired_item("dired-dirsearch-", 0);

	if (i)
		pat = get_str_var(i);
	save_var case_fold = 1, sort_case_fold = (dired_format != DF_UNIX);
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	if (dired_groups_dirs) {
		save_var point = 0;
		tmp = tmp_buf();
		int dirs = tmp_buf();	// Put ., then .., in here, don't sort.
		while (column_color_searching(REGEX, "<.>[/\\]?$", goal_column, -1)
			   && !user_abort)
			move_line_to_buffer(dirs);
		point = 0;
		while (column_color_searching(REGEX, "<.><.>[/\\]?$", goal_column, -1)
			   && !user_abort)
			move_line_to_buffer(dirs);
		point = 0;
		while (re_search(1, pat) && !user_abort)
			move_line_to_buffer(tmp);
		dired_group_sorter(start, point);
		point = 0;
		grab_buffer(dirs);
		buf_delete(dirs);
		grab_buffer(tmp);
		buf_delete(tmp);
		dired_group_sorter(mark, point);
	} else
		dired_group_sorter(start, end);
}

command configure_epsilon()
{
	char prog[FNAMELEN];
	if (!is_win32)
		error("This command is only available in Epsilon for Windows.");
	get_executable_file(prog, "econfig.exe", 0);
	do_shell_execute(prog, "", "run setup");
}

// Return 0 if we can't because program's missing.
int set_font_for_unix()
{
	char buf[FNAMELEN];

	get_executable_file(buf, "epsilon-dialogs", 1);
	sprintf(buf + strlen(buf), " --parent %x --font \"%s\"",
			screen_to_window_id(0), font_fixed);
	if (!pipe_text("-epsilon-dialog-out", "-epsilon-dialog-out", buf, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF, 0))
		file_error(errno, "Couldn't run dialog helper", "Error");
	save_var bufname = "-epsilon-dialog-out";
	point = 0;
	if (parse_string(1, ".*(ibrary|shared)")) {
		dialog_helper_failed(1);
		return simple_set_font_for_unix();
	} else if (re_search(1, "^=") && parse_string(1, ".*", buf))
		font_fixed = buf;
	return 1;			// User picked one, even if it was bad.
}

// Display a message explaining how to install needed library, trying
// to set font or colors.
dialog_helper_failed(int is_font)
{
	int b = zap("-epsilon-dialog-error");
	buf_printf(b, "Epsilon was unable to run its external helper program.\n\n");
	if (is_font)
		buf_printf(b, "Epsilon will revert to a simpler font setting method.\n");
	else {
		buf_printf(b, "For now, you can set colors by running the Alt-x ");
		buf_printf(b, "export-colors\ncommand, editing the source file it ");
		buf_printf(b, "generates, and loading it\nusing the compile-buffer ");
		buf_printf(b, "command on Alt-F3.\n");
	}
	save_var _view_title = " Can't run dialog helper ";
	view_buf(b, 0);
}

when_loading()
{
	drop_name("c_in_old_comment");
	drop_name("csharp_in_string");
	drop_name("in_c_comment_at");
	drop_name("scan_for_string");
	drop_name("no_mac_dialog_helper");
	drop_name("mac_gtk_message");
}
