/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2009 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.08 to 13.09.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "tags.h"
#include "vbasic.h"

int c_lex_region(int from, int to)
{
	// If from is the in middle of an already-lexed region, go to its start.
	get_tagged_region("c-lex", from, &from);
	save_var point = from, narrow_start = 0, narrow_end = 0;
	int start, c, val, init = from, p, found_end = 1;
	while (re_search(1, "<squote|dquote>|//|/<*>")) {
		start = matchstart;
		c = character(start);
		p = point;	// Mark that we're not in comment/string before the march.
		set_tagged_region("c-lex", init, p, 0);
		if (c == '/' && character(start + 1) == '/') {
			val = IN_NEW_COMMENT;	// Handle // comment (c_skip_item() won't).
			found_end = nl_forward();
		} else {
			point = start;
			found_end = c_skip_item(c, 0);
			switch (c) {
				case '/':	val = IN_OLD_COMMENT; break;
				case '"':	val = IN_STR_CONST; break;
				case '\'':	val = IN_CHAR_CONST; break;
				default:	say("Internal error in c_lex_region in %s at %d (%d)",
								  bufname, point, c);
							val = 0;
							break;
			}
		}
		set_tagged_region("c-lex", p, point, val);
		if (point > to)
			return found_end;
		init = point;
	}
	set_tagged_region("c-lex", init, point, 0);	// Mark final region.
	return found_end;
}

c_lex_safe_start(int from)
{
	int i, f;
	do {
		i = get_tagged_region("c-lex", from - 1, &f);
		if (f >= from)	// Give up if we don't progress.
			break;
		from = f;
	} while (i);
	return from;
}

in_c_comment(int lookback)
{
	UNREFERENCED_PARAMETER(lookback);
	if (in_perl_buffer)
		return in_perl_comment(point);
	int pos = point, from, to, res;
	if (modified_buffer_region(&from, &to, "c-lex")) {
		// First, discard region modified since last scan.
		from = c_lex_safe_start(from);
		set_tagged_region("c-lex", from, size(), -1);	// Set to unknown.
	}
	reset_modified_buffer_region("c-lex");

	res = get_tagged_region("c-lex", pos, &from, &to);
	if (res == -1)		// No cached data, so lex it.
		c_lex_region(from, pos);
	res = get_tagged_region("c-lex", pos, &from, &to);
	if (pos >= size() && (res = get_tagged_region("c-lex", pos, &from, &to))
		  && c_lex_region(c_lex_safe_start(from), to))
		res = 0;	// If region ends right at EOF, don't count it.
	if (res == -1) {
		say("Internal error in in_c_comment: %s at %d", bufname, pos);
		return 0;
	}
	return res;
}

dired_do_delete(name, to, in_dired)
char *name, *to;
{
	int isdir = in_dired ? dired_isdir() : 0;

	to = to;
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER,
					  isdir == 1 ? "RMD %s\n" : "DELE %s\n",
					  get_url_file_part(name, 0));
	} else {
		if (in_dired)
			noteput("Deleting %f.", name);
		if (isdir == 1) {
			if (dired_warn_rmdir(name))
				return 1;
		} else if (dired_warn_delete(name)
				   && (isdir <= 1 || !is_win32 || rmdir(name)))
			return 1;	// If NTFS junction, try rmdir too.
		if (!isdir)
			maybe_delete_spell_ignore_list(name);
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class") || !strcmp(func, "__interface"))
		return tag_c_struct(1);		// or similar class or managed c++ def
	else if (!strcmp(func, "interface"))
		return tag_c_struct(1);		// for java or managed c++
	else if ((c_extra_keywords & OBJC_KEYWORDS) && character(start - 1) == '@'
			 && word_in_list(func, "|interface|implementation|protocol|", 0))
		return tag_objc_block();
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if ((c_extra_keywords & CPP_KEYWORDS) && !strcmp(func, "using")) {
		tag_modifier_kwd |= 4;	// just a declaration 
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (!strcmp(func, "__attribute__") && parse_string(1, "<wspace>*%(%(")
			   && !skip_c_delims('(', ')', 1)) {
		return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	if (parse_string(1, "(_ANSI_ARGS_|__P|PROTO)</word>")) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template();
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
												"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (tag_c_variable()
			   && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
				   || (tag_modifier_kwd & 2))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}

int vbasic_indent_more()
{
	char kwd[FNAMELEN], pat[FNAMELEN];

	if (vbasic_indent_subroutines
		  && parse_string(1, "[ \t]*(%<[^\n>]+%>)?[ \t]*"
						  "(</word>(Private|Public|Default|Friend"
						  "|Static|Protected|Shadows|MustInherit"
						  "|NotInheritable|Overloads|Overrides|Overridable"
						  "|NotOverridable|MustOverride|Shared|ReadOnly"
						  "|WriteOnly|Widening|Narrowing)</word>[ \t]*)*"
						  "</word>(Sub|Function|Property|Class|Operator"
						  "|Module|Interface|Namespace|Structure)</word>")) {
		if (parse_string(1, "[ \t]*(%<[^\n>]+%>)?[ \t]*"
						 "(</word>(Private|Public|Default|Friend"
						 "|Static|Protected|Shadows|MustInherit"
						 "|NotInheritable|Overloads|Overrides|Overridable"
						 "|NotOverridable|MustOverride|Shared|ReadOnly"
						 "|WriteOnly|Widening|Narrowing)</word>[ \t]*)*"
						 "</word>MustOverride</word>"))
			return 0;	// No body for MustOverride
		save_var point;
		if (re_search(-1, "</word><^c:*-comment|*-string>interface</word>")
			  && !parse_string(-1, "</word>End[ \t]+"))
			return 0;	// No body in Interface block
		return 1;
	}
	if (parse_string(1, "[ \t]*</word>(If)</word>.*"
					 "</word><^c:*-comment|*-string>Then</word>[ \t]*[^ \t\n':]"))
		return 0;		// Old one-line form of If statement.
	if (parse_string(1, "[ \t]*</word>((private|public)[ \t]+)?"
					 "(Type|Enum|Else|With|Select|Case|Try|Catch|Finally"
					 "|SyncLock|If|Elseif)</word>")) {
		grab(find_group(3, 1), find_group(3, 0), kwd);
		sprintf(pat, ".*:[ \t]*</word><^c:*-comment|*-string>End[ \t]+%s</word>", kwd);
		if (!parse_string(1, pat)) {	// Multiple statements: ignore block.
			if (vbasic_indent_case
				  && parse_string(1, "[ \t]*</word>Select</word>"))
				return 2;		// Force extra indent for select case/case.
			return 1;
		}
	}
	if (parse_string(1, "[ \t]*</word>(For)</word>")
		  && !parse_string(1, ".*:[ \t]*<^c:*-comment|*-string></word>Next</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(Do)</word>")
		  && !parse_string(1, ".*:[ \t]*<^c:*-comment|*-string></word>Loop</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(While)</word>")
		  && !parse_string(1, ".*:[ \t]*<^c:*-comment|*-string></word>Wend</word>"))
		return 1;
	if (parse_string(1, "[ \t]*(Get|Set)[ \t]*(%([ \t]*%))?[ \t]*[\n(':]"))
		return 1;	// Get/set keyword with nothing after it, or (, starts block.
	return 0;
}
