/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2011 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.09 to 13.10.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "html.h"
#include "indent.h"
#include "gams.h"
#include "perl.h"
#include "proc.h"
#include "python.h"
#include "session.h"
#include "tags.h"
#include "vbasic.h"

int quoted_insert();

short *ask_key(char *pr, char *keyname, int flags)
{
	int index, k, fk, ask_if_ambiguous = flags & 1, warn_if_std = flags & 2;
	short *tab, *alttab;
	char msg[FNAMELEN], *p;

	tab = mode_keys;
	alttab = root_keys;
	*keyname = 0;			// put readable form of key here
	key_binding[0] = 1;		// also save sequence in this global
	for (;;) {
		sprintf(msg, "%s%s", pr, keyname);
		full_getkey(msg, 0);
		if (root_keys[key] == (short) quoted_insert) {
			mention("%s<Quote> ", msg);
			getkey();
			user_abort = 0;
		}
		check_abort();
		p = keyname + strlen(keyname);
		k = full_key;
		index = (tab[k] == 0) ? alttab[k] : tab[k];
		if (index <= 0) {
			k = key;
			index = (tab[k] == 0) ? alttab[k] : tab[k];
		}
		show_char(keyname, ask_if_ambiguous ? k : full_key, 1);
		key_binding[key_binding[0]] = k; // add to constructed key
		key_binding[0]++;
		if (index <= 0 || name_type(index) != NT_TABLE)
			break;
		alttab = index_table((tab[k] && !alttab[k]) ?
							 tab[k] : alttab[k]);
		tab = index_table(index);
		if (!alttab)
			alttab = tab;
		strcat(keyname, " ");
	}
	say("");
	if (warn_if_std && (key >= ' ' && key <= 65535 || key == '\r' || key == '\n')) {
		sprintf(msg, "Are you sure you want to rebind the standard key %k? ", key);
		if (!ask_yn("Confirm rebinding a standard key", msg, "Yes", "No", 0))
			quick_abort();
	}
	if (ask_if_ambiguous && full_key != key) {
		k = key;
		fk = full_key;
		sprintf(msg, "Bind only specific key %k (not general key %k)? ",
				fk, k);
		if (ask_yn("Specific key or general?", msg, "Yes", "No", 2)) {
			key = k = fk;
			*p = 0;
			show_char(keyname, k, 1);
		}
	} else if (full_key != key && k == full_key)
		key = k;
	return ((tab[k] == 0) ? alttab : tab) + k;
}

command brief_delete_window()
{
	int dir, wantbuf = bufnum;

	tiled_only();
	while ((dir = get_direction()) == -1) {
		say("Select window edge to delete (use cursor keys)");
		getkey();
		check_abort();
	}
	if (give_window_space(dir))
		error("Can't delete that edge.");
	to_buffer_num(wantbuf);
	say("");
}

c_indenter()		/* replace surrounding indentation with new */
{
	if (c_extra_keywords & IDL_KEYWORDS)
		return;
	if (in_shell_buffer)
		prev_indenter();
	else if (!(in_c_comment(c_look_back) & IN_COMMENT))
		indent_to_column(c_compute_indent());
	else if ((in_perl_buffer ? reindent_perl_comments : reindent_c_comments)
			 & (indenter_action == IACT_REINDENT_PREV ? 4 : 1))
		indent_c_comment();
}

int c_lex_eof_rules(int res, int from, int to)
{
	if (res && c_lex_region(c_lex_safe_start(from), to))
		res = 0;	// If region ends right at EOF, don't count it.
	else if (!res)
		if (!get_tagged_region("c-lex", size() - 1) && strchr("\"'", character(size() - 1)))
			return character(size() - 1) == '"' ? IN_STR_CONST : IN_CHAR_CONST;
		else if (!get_tagged_region("c-lex", size() - 2) && character(size() - 2) == '/')
			if (character(size() - 1) == '/')
				return IN_NEW_COMMENT;
			else if (character(size() - 1) == '*')
				return IN_OLD_COMMENT;
	return res;
}

in_c_comment(int lookback)
{
	UNREFERENCED_PARAMETER(lookback);
	if (in_perl_buffer)
		return in_perl_comment(point);
	int pos = point, from, to, res;
	if (modified_buffer_region(&from, &to, "c-lex")) {
		// First, discard region modified since last scan.
		from = c_lex_safe_start(from);
		set_tagged_region("c-lex", from, size(), -1);	// Set to unknown.
	}
	reset_modified_buffer_region("c-lex");

	res = get_tagged_region("c-lex", pos, &from, &to);
	if (res == -1)		// No cached data, so lex it.
		c_lex_region(from, pos);
	res = get_tagged_region("c-lex", pos, &from, &to);
	if (pos >= size())		// Special rules for EOF, as we don't cache that.
		res = c_lex_eof_rules(res, from, to);
	if (res == -1) {
		say("Internal error in in_c_comment: %s at %d", bufname, pos);
		return 0;
	}
	return res;
}

have_label(pt)	/* is there a label here? */
{
	save_var point = pt;
	if (c_extra_keywords & (JS_KEYWORDS))	// Not a label in JavaScript.
		return 0;
	if (have_colon() && parse_string(1,"[a-zA-Z0-9_]+[ \t\f]*:[^[:]")) {
		if (c_extra_keywords & OBJC_KEYWORDS) {
			save_var narrow_end = size() - point;
			move_by_lines(-10);			// Not a label if in unclosed [] block.
			save_var narrow_start = point;
			point = size();
			return !c_rev_skip_level('[', ']');
		}
		// Not a label but a bitfield if it's not a known keyword and
		// it's in a struct.
		if (!is_case(point) && c_in_struct())
			return 0;
		return 1;
	}
	return 0;
}

c_keyword_color(from)
{
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(char) - 10)
		return color_class c_identifier;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-.", buf[1])) {
		if ((c_extra_keywords & CSHARP_KEYWORDS)) {
			point = from;		// Rules for C# numbers are slightly different.
			point += parse_string(1, "-?%.?[0-9]([A-Za-z0-9_]|[Ee]-|.[0-9])*");
			grab(from, point, buf + 1);
		}
		return c_number_color(buf + 1);
	}
	strcpy(buf + point - from + 1, "|");
	if (is_c_keyword(buf) && (!(c_extra_keywords & CSHARP_KEYWORDS) ||
							  character(from - 1) != '@'))
		return color_class c_keyword;
	if (color_class c_function != color_class c_identifier
		  && paren_follows())
		return color_class c_function;
	return color_class c_identifier;
}

suffix_gz()
{
	// Calling gzip not gunzip works in more cases with Cygwin utils.
	// Use .exe suffix to avoid Take Command's internal incompatible gzip.
	if (uncompress_files & 1)
		do_decompression("gzip", is_win32 ? "gzip.exe --decompress --stdout"
										  : "gzip --decompress --stdout");
}

// Name the new macro and bind it too.
command bind_last_macro()
{
	char name[FNAMELEN], msg[FNAMELEN], keyname[FNAMELEN], def[FNAMELEN];
	int i, j;
	short *p;

	i = find_index("last-kbd-macro");
	if (!i || name_type(i) != NT_MACRO)
		error("No keyboard macro.");
	for (j = 1; ; j++) {
		sprintf(def, "anonymous-macro-%d", j);
		if (!find_index(def))
			break;
	}
	get_cmd(name, "Name the macro", def);
	if (find_index(name)) {
		sprintf(msg, "Replace existing %s? ", name);
		if (!ask_yn("Replace with macro", msg, "Yes", "No", 1))
			return;
	}
	name_macro(name, get_macro(i));
	save_customization(find_index(name));
	sprintf(msg, "Bind new macro %s to key (%k to skip binding): ", name, abort_key);
	p = ask_key(msg, keyname, 3);
	*p = find_index(name);
	save_binding(p);
	rebuild_menu();
	say("");
}

command bind_to_key()
{
	char msg[300];
	int name_index;
	short *p;

	if (name_index = get_command_index("Bind command")) {
		p = ask_key("To key: ", msg, 3);
		*p = name_index;
		save_binding(p);
	}
	rebuild_menu();
}

command unbind_key()
{
	char keyname[300];
	short *binding, val = -1;

	binding = ask_key("Unbind key: ", keyname, 3);
	sayput("Make \"%s\" Unbound, or make it Self-insert [U, S]? ", keyname);
	while (!user_abort && val == -1) {
		getkey();
		switch (toupper(key)) {
			case 'U':	val = 0; break;
			case 'S':	val = normal_character;
						if (!strcmp(keyname, "<EnterKey>") || !strcmp(keyname, "Ctrl-M"))
							val = find_index("enter_key");
						break;
		}
	}
	say("");
	check_abort();
	if (val != -1) {
		*binding = val;
		save_binding(binding);
	}
	rebuild_menu();
}

command create_prefix_command()
{
	char keyname[300];
	short *loc;

	loc = ask_key("New prefix key: ", keyname, 3);
	*loc = make_anon_keytable();
	set_case_indirect(index_table(*loc));
	set_mouse_keys(index_table(*loc));
	save_binding(loc);
}

load_eel_from_path(char *f, int flags)
{
	char *p = lookpath(f, flags & 4), orig_dir[FNAMELEN], new_dir[FNAMELEN];
	int res, tmp = tmp_buf();

	on_exit buf_delete(tmp);
	if (!p) {
		if (flags & 2)
			say("Can't find EEL source file %f.", f);
		return 1;
	}
	getcd(orig_dir);
	strcpy(new_dir, p);
	*get_tail(new_dir, 0) = 0;
	on_exit chdir(orig_dir);
	chdir(new_dir);
	res = eel_compile(p, 1, "-n -q", bufnum_to_name(tmp));
	if (buf_size(tmp)) {
		save_var bufnum = create("eel-errors");
		point = size();
		bprintf("\nLoading %s:\n", p);
		grab_buffer(tmp);
	}
	if (res == 1) {
		if (flags & 3)
			show_text(0, 300, "Error loading %f, see %s buffer.",
					  p, "eel-errors");
		return 2;
	}
	return 0;
}

dired_examine_deleting()
{
	int buf = bufnum;
	dired_examine();
	if (is_dired_buf() && buf && buf != bufnum)
		buf_delete(buf);
}

dired_unix_fix_lines()
{
	save_var modified, indent_with_tabs = 0, case_fold = 1;
	right_align_columns("[0-9]+ *" UNIX_DATE);
	right_align_columns(UNIX_DATE);
	save_var point = 0;
	if (re_search(1, ".*" UNIX_DATE))
		goal_column = current_column();
}

char *dired_skip_line_pat();

detect_dired_format()
{
	int old = dired_format, i;

	goal_column = NAME_START;
	save_var point = 0, case_fold = 1;
	dired_format = DF_VMS;		// Skip VMS lines too for now.
	while (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	dired_format = 0;
	if (point >= size())
		return;
	if (parse_string(1, ".*" UNIX_DATE)) {
		dired_format = DF_UNIX;
		goal_column = get_column(matchend);
	} else if (parse_string(1, ".*" OLDNT_DATE)) {
		dired_format = DF_OLDNT;
		goal_column = 41;
	} else if (parse_string(1, ".*" VMS_DATE)) {
		dired_format = DF_VMS;
		goal_column = 2;
	} else if (!parse_string(1, ". [ 0-9][0-9]-[a-z?][a-z?][a-z?]-[0-9][0-9]")
			   && !parse_string(1, ". [0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]")) {
		dired_format = DF_SIMPLE;
		goal_column = 2;
	} else {
		for (i = 1; i <= 8; i <<= 1)
			if (dired_layout & i)
				goal_column++;
	}
	if (dired_format != old)
		drop_all_colored_regions();
}

dired_sort_by_size_unix()
{
	char line[FNAMELEN * 2], *p;

	point = 0;
	while (re_search(1, UNIX_DATE)) {
		point = matchstart;
		if (!parse_string(-1, "[0-9]+[MK]? *", line)) {
			note("Unrecognized format, can't sort.");
			strcpy(line, "0");
		}
		if (p = strchr(line, 'M'))
			strcpy(p, "000000");
		if (p = strchr(line, 'K'))
			strcpy(p, "000");
		move_to_column(goal_column);
		bprintf("%20s", line);
		to_end_line();
	}
	return 20;
}

dired_sort_by_date_unix()
{
	char line[FNAMELEN * 2], *p, *months, *time;
	int month, year;
	struct time_info now;

	time_and_day(&now);
	months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	point = 0;
	while (re_search(1, UNIX_DATE)) {
		grab(matchstart + 1, point, line);
		if (line[3] != ' ' || line[6] != ' ') {
			say("Unrecognized date format, can't sort.");
			strcpy(line, " Jan  1  1970 ");	// Fake default date.
		}
		line[3] = line[6] = 0;	// break into pieces
		p = strstr(months, line);
		if (!p) {
			say("Unrecognized month %s, can't sort.", line);
			p = months;
		}
		month = 1 + (p - months) / 4;	// Compute month 1..12.

		if (line[9] == ':') {	// Either a time (in the last 6 months)...
			time = line + 7;	// Year is this year or late last year.
			year = now.year - (month > now.month + 1);
		} else {			// Or a prior year, so make up a fake time.
			time = "00:00 ";
			year = numtoi(line + 7);
		}
		move_to_column(goal_column);
		bprintf("%5d%02d%2s%5s", year, month, line + 4, time);
		to_end_line();
	}
	return 15;
}

command edit_variables()
{		// list variables for editing; with arg, includes system vars
	int listbuf, orig_buf, orig, part = 1;
	int listwin, helpwin, orig_win, win, listheight = 9;
	int setit, origpos, ind, type, hasdef, i, lasthelp;
	TIMER ts;	// for searching in list
	char var[100], partial[100], *val;
	char tmp[FNAMELEN * 2], curval[FNAMELEN * 2];
	jmp_buf this_level;

	tiled_only();
	if ((has_feature & FEAT_WINHELP)
		  && get_epsilon_help_format() == HELP_FORMAT_WINHELP) {
		if (has_arg)
			win_help_string("", "all variables, list of");
		else
			win_help_string("", "preference variables, list of");
		return;
	}
	if (screen_lines > 20)
		listheight = screen_lines * 40 / 100;
	orig_buf = bufnum;
	orig_win = window_handle;
	noteput("Listing variables...");
	save_var abort_searching = 0;
	if (has_arg)
		save_var show_all_variables = 1;
	iter = has_arg = lasthelp = 0;
	listbuf = tmp_buf();	// put list of variables here
	bufnum = listbuf;
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	save_var _nt_filter = BIT(NT_VAR) | BIT(NT_BUFVAR) | BIT(NT_WINVAR)
			      | BIT(NT_BUILTVAR);
	list_matches("", var_match, 0, listbuf);	// construct list of vars
	point = 0;
	listwin = add_popup(_view_left, 1, 30, listheight,
						ALL_BORD(BSINGLE), listbuf);
	set_menu_look(listwin);
	window_title(listwin, BTOP, TITLECENTER, " Variables ");
	mode_keys = menu_tab;
	_window_flags |= HIGHLIGHT_LINE;

	get_doc(0);
	helpwin = add_popup(_view_left, listheight + 2,	// put help text here
						MIN(screen_cols - _view_left - _view_right, 80),
						11, _view_border, zap(HELPBUF));

	text_color = (is_gui && want_gui_prompts) ? color_class text : color_class viewed_text;
	display_column = 0;		/* prevent continuation lines */
	char dclass[256];
	memcpy(dclass, _std_disp_class, 256);
	dclass['\t'] = BC;
	_display_class = dclass;

	restore_vars();		// don't let save_var undo user's changes
	time_begin(&ts, 0);
	while (is_window(listwin) && is_window(helpwin)) {
		setit = 0;		// assume we don't want to set variable
		window_handle = listwin;
		origpos = point;
		to_begin_line();
		grab_full_line(bufnum, var);	// get name of current variable
		ind = find_index(var);
		switch (name_type(ind)) {	// has a default value too?
			case NT_BUFVAR:	hasdef = 1; break;
			case NT_WINVAR: hasdef = 2; break;
			default:	hasdef = 0; break;
		}
		type = vartype(ind);

		if (ind != lasthelp) {	// show info on variable
			window_handle = orig_win; // if buffer/window-specific,
			bufnum = orig_buf;	// get value from before this
			val = get_var_val(ind, curval, 0); // command

			window_handle = helpwin;
			make_var_help(var, orig_buf, orig_win, 1);
			window_to_fit(screen_lines - listheight - 6);
			sprintf(tmp, " %s ", var);
			window_title(helpwin, BTOP, TITLECENTER, tmp);
			display_scroll_bar = !window_extra_lines();
			lasthelp = ind;
		}
		window_handle = helpwin;
		build_window();
		window_title(helpwin, BBOTTOM, TITLERIGHT(2), 
					 all_blanks(window_end, size() - narrow_end)
					 ? "" : " Press <Space> for more ");
		window_handle = listwin;
//		do {			// prompt for a key
			maybe_refresh();
			note("Select a variable & press <Enter> to set, "
				 "<Esc> when done, ? for help.");
			i = permit_window_keys;
			permit_window_keys = PERMIT_SCROLL_KEY;
			getkey();
			permit_window_keys = i;
//		} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		if (key == ESC || user_abort || key == abort_key)
			break;
		else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (is_window(win) != ISPOPUP)
				break;
			window_handle = win;
			get_text_area(win);
			if (in_text_area(mouse_x, mouse_y)) {
				point_to_pos(mouse_x, mouse_y, mouse_screen);
				if (win == listwin && IS_MOUSE_DOUBLE(key))
					setit = 1;
				else if (win == helpwin)
					key = KEYPGDN;
			} else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
		} else if (is_search_dir()) {
			save_var top_level = &this_level;
			if (!setjmp(top_level))	// come here on abort
				do_again();	// let search cmds run normally
			restore_vars();
			continue;
		}
		if (!isdigit(key))	// do arrow keys, C-n, etc.
			switch (get_direction()) {
				case BTOP:
					if (!nl_reverse()) point = size();
					break;
				case BBOTTOM:
					if (!nl_forward()) point = 0;
					break;
			}
		switch (key) {
			case KEYHOME:
			case NUMDIGIT(7): point = 0; break;
			case KEYEND:
			case NUMDIGIT(1): point = size(); break;
			case KEYPGDN:
			case NUMDIGIT(3): next_page(); break;
			case KEYPGUP:
			case NUMDIGIT(9): previous_page(); break;
			case '\n':
			case '\r':
				setit = 1; break;
			case ' ':
				window_handle = helpwin;
				next_page();
				break;
			case '\b':
			case GREYBACK:
				if (!time_done(&ts)) { // backsp in srch
					if (part > 1)
						part--;
					goto search_for;
				}
				window_handle = helpwin; // or scroll help
				previous_page();
				break;
			default:
				if (binding_of(key, root_keys) == (short) help
				    || key == '?') {
					help_on_command("edit-variables");
					lasthelp = 0;
				} else if (key >= ' ' && key <= '~') {
					if (time_done(&ts))
						part = 1; // new search
					partial[part++] = key;
search_for:				orig = point, point = 0;
					partial[0] = '^';
					partial[part] = 0;
					save_var see_delay = new_search_delay;
					say("Searching for %s", partial+1);
					restore_vars();
					if (do_searching(FOLD|REGEX, partial)
					    != 1)
						point = orig;
					time_begin(&ts, new_search_delay);
				} else if (IS_WIN_KEY(key) || key == '\t' || key == BACKTAB) {
					run_topkey();
				} else if (IS_TRUE_KEY(key)
					 && !setit && point == origpos) {
					maybe_ding(bell_on_bad_key);
				}
				break;
		}
		if (setit) {		// user wants to set a variable
			refresh();
			part = 1;	// reset searcher
			ungot_key = CTRL('S');	// bring in default value
			save_var top_level = &this_level;
			i = 0;
			if (setjmp(top_level))	// come here on abort
				i = 1;		// and record it
			else if (is_gui && want_gui_prompts)
				get_strpopup(tmp, var, val, var);
			else
				get_strnone(tmp, "New value: ", val);
			restore_vars();
			if (i)		// aborted
				continue;
			if (type <= TYPE_INT) {
				numtoi(tmp);
				if (got_bad_number)
					continue;
			}
			i = 0;		// assume current val only
			if (is_gui && want_gui_prompts && hasdef)
				i = 1;			// assume set both, for now
			else if (hasdef) {
				sayput("Set D=default value of %s, C=current"
				       " value, or B=both? [B]", var);
				getkey();
				if (user_abort)
					continue;
				if (toupper(key) == 'D')
					i = -1;
				else if (toupper(key) != 'C')
					i = 1;	// set current & default
			}
			window_handle = orig_win;
			bufnum = orig_buf;
			if (i >= 0)
				set_var_val(ind, tmp, 0);
			if (i)
				set_var_val(ind, tmp, 1);
			full_redraw = 1; // might have set a display-type var
			lasthelp = 0;	// refresh help win
		}
	}
	say("");
	if (is_window(orig_win))
		window_handle = orig_win;
	remove_window(listwin);
	remove_window(helpwin);
	buf_delete(listbuf);
	delete_buffer(HELPBUF);
	user_abort = 0;
}

int scp_echoed_cmdline(char *cmd, char *list)
{
	char *end;
	if (!cmd || !list)
		return 0;
	while (end = strchr(list, '\n')) {
		if (!strncmp(list, cmd, end - list))
			return 1;
		list = end + 1;
	}
	return !strcmp(cmd, list);
}

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1,
			".* (not a directory|No such|not found)"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & (SCP_CHECK_IS_DIR
											  | SCP_CHECK_MATCHES)))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is| is now) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connect(ing|ed) to "
								 "|Connection closed|Listing directory "
								 "|Using username \"|sftp%>"
								 "|(.* |^)pass(word|phrase).*:)")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg && !scp_echoed_cmdline(msg, p->cmdlist))
				say("%s", msg);		// Any other line is an error msg.
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

scp_filter_sftp_matches(char *filter, int prefix, char *url)
{
	char fname[FNAMELEN];
	delete_final_newlines();
	point = 0;
	for (;;) {	// Result is in ls -l format.
		int start = point, isdir = curchar() == 'd';
		int len = parse_string(1, "^.*" UNIX_DATE " *");
		if (!len)
			break;
		delete(start, start + len);	// Want just the name.
		to_end_line();
		grab(start, point, fname);
		if (!fpatmatch(fname, filter, prefix, 0))
			delete(start, point + 1);	// Discard this line.
		else {
			if (isdir)
				stuff("/");
			if (!nl_forward())
				break;
		}
	}
	delete_final_newlines();
	point = 0;
	strcpy(fname, url);	// Now insert rest of URL before each match.
	*get_tail(get_url_file_part(fname, 0), 1) = 0;
	if (*fname && !is_path_separator(fname[strlen(fname) - 1]))
		strcat(fname, "/");
	if (size())
		simple_re_replace(1, "^", fname);
}

char *scp_variable(struct url_parts *parts, char *varname);

scp_run_helper(struct url_parts *parts, struct scp_job *p)
{
	char cmd[FNAMELEN], *args[3], cmdname[FNAMELEN];

	args[0] = parts->usr;
	args[1] = parts->host;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command");
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(parts, "run-helper"),
							   "uhc", args))
		error("Invalid run helper template.");
	if (!is_process_buffer(p->buf)) {
		if (is_win32) {
			char *oldval = getenv("CYGWIN");
			if (oldval) {
				char *newval = malloc(strlen(oldval) + 50);
				sprintf(newval, "CYGWIN=%s nodosfilewarning", oldval);
				putenv(strkeep(newval));
				free(newval);
			} else
				putenv("CYGWIN=nodosfilewarning");
		}
		scp_start_process(p, cmd);
		if (p->homedir)
			free(p->homedir);
		p->homedir = 0;
		p->busy = 2;
	}
}

ftp_file_match(char *pat, int flags)
{
	char mypat[FNAMELEN * 2], name[FNAMELEN * 2], *p;

	save_var new_buffer_translation_type = FILETYPE_AUTO;
	save_var bufnum = zap("-ftp-file-match-listing");
	strcpy(mypat, pat);
	if (!(flags & EXACTONLY))
		strcat(mypat, "*");
	strcpy(name, mypat);
	p = get_tail(name, 1);	// If name contains wildcards, strip them.
	if (index(p, '*') || index(p, '?')) {
		*p = 0;								// Make FTP listing
		save_var ftp_compatible_dirs |= 1;	// show just the basename.
	}
	url_operation(name, FTP_LIST | FTP_WAIT);
	point = 0;
	bprintf("     Directory of %s\n", name);
	while (re_search(1, "^.")) {
		point--;
		stuff("  ");
		point++;
	}
	modified = 0;
	dired_orig_position = point = 0;
	dired_ftp_activity(NET_DONE, bufnum, 0, 0);
	color_dired_range(0, size());
	int b = zap(_MATCH_BUF);
	point = 0;
	for (point = 0; point < size() && nl_forward(); )
		if (dired_getname(name, 2) && (dired_isdir() ? !(flags & FM_NO_DIRS)
			: !(flags & FM_ONLY_DIRS)))
			if (fpatmatch(name, mypat, 0, 0))
				buf_printf(b, "%s%s\n", name, dired_isdir() == 1 ? "/" : "");
	set_buf_point(b, 0);
}

int http_activity();

is_retrieving_remote_file(int b)
{
	save_var bufnum = b;
	return ftp_job && (ftp_job->operation & FTP_OP_MASK) == FTP_RECV
			|| (is_busy_doing_scp(b) & FTP_OP_MASK) == FTP_RECV
			|| when_net_activity == http_activity;
}

int is_ftp_url_a_dir(char *s)
{
	char cmd[FNAMELEN];
	struct url_parts parts;
	int port, res;

	prepare_url_operation(s, FTP_MISC | FTP_WAIT, &parts);
	if (!strcmp(parts.service, "scp"))
		return scp_is_url_a_dir(s, &parts);
	if (ftp_compatible_dirs & 4)
		return 0;
	if (*parts.fname)
		sprintf(cmd, "CWD %s", parts.fname);
	else
		return 1;
	port = numtoi(parts.port);
	if (got_bad_number)
		port = 21;
	char *t1 = temp_buf(), *t2 = temp_buf();
	set_mode_message("[FTP: Resolving link] ");
	ftp_op(t1, t2, parts.host, port, parts.usr, parts.pwd, cmd,
		   FTP_MISC | FTP_WAIT);
	set_mode_message(NULL);
	save_var bufname = t2;
	point = 0;
	res = !re_search(1, "^(550|530) ");
	restore_vars();
	delete_buffer(t1);
	delete_buffer(t2);
	return res;
}

int reading_translation_type()
{
	if (translation_type & FORCED_TRANS)
		return translation_type;
	return SET_TRANSLATE(translation_type, FILETYPE_AUTO);
}

do_save_all_buffers(int newfile_bufs, int *count, int *queued, int *failed)
{
	int i;

	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		if (!*filename || !modified || buffer_not_saveable
			  || is_retrieving_remote_file(i))
			continue;
		if (newfile_bufs ? (force_save_as != FSA_NEWFILE)
						 : (force_save_as == FSA_NEWFILE))
			continue;	// Only do one group.
		if (newfile_bufs) {
			if (!size())		// Nothing to save, so don't bother to prompt.
				continue;
			if (save_all_without_asking) {
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not"
							  " save buffer %s, no file name\n", bufname);
				(*failed)++;
				continue;
			}
			to_buffer_num(i);	// Show the buffer we'll be prompting for.
		}
		switch (do_save_file(1, 1, 1)) {
			case 1:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not save %s\n",
							  filename);
				(*failed)++;
				break;
			case 0:
				show_text(-2, 1, "save-all-buffers: %s %s", 
						  is_remote_file(filename) ? "Queued" : "Saved",
						  filename);
				if (!save_without_prompt)
					if (is_remote_file(filename))
						(*queued)++;
					else
						(*count)++;
				break;
		}
	} while (i = buf_list(1, 1));
}

// Get user to select an encoding for writing this file.
pick_write_encoding()
{
	int tmp = tmp_buf();
	int t = GET_LINE_TRANSLATE(translation_type);
	int e = GET_ENCODING(translation_type);
	int def = encoding_from_name(default_write_encoding);

	if (is_window(window_handle) == ISPOPUP)
		error("Choose an encoding for this buffer first.");
	save_var window_bufnum;
	to_buffer_num(bufnum);
	save_var _view_title = "Choose an encoding";
	buf_printf(tmp, "%s\ncontains characters that cannot be represented\n",
			   bufname);
	buf_printf(tmp, "in a disk file without using an encoding.  Please choose an\n");
	buf_printf(tmp, "encoding for this file.  The default encoding shown or any\n");
	buf_printf(tmp, "of the UTF- encodings will preserve all the characters.\n");
	buf_printf(tmp, "See \"Using National Characters\" in the online manual.\n\n");
	buf_printf(tmp, "Here are some of the characters that can't be represented\n");
	buf_printf(tmp, "without an encoding:\n\n");
	grab_regex_match_lines(tmp, "<#256-#65535>", " ", " ", 10000);
	view_buf(tmp, 0);
	if (def < 0)
		def = 2;		// Default to UTF-8-No-BOM
	save_var translation_type = MAKE_TRANSLATE(def, t, 1);
	e = ask_encoding("Set buffer encoding to", 1);
	iter = 1;
	buf_delete(tmp);
	restore_vars();
	if (e < 1)
		e = def;
	translation_type = MAKE_TRANSLATE(e, t, (e || t != FILETYPE_AUTO));
	make_mode();
}

#define C_COMMENT_FILLER		"[^0-9a-zA-Z{}()\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

do_fill_comment(int generic, int margin)
{
	int pbuf, width = 1000, i, old_lines, new_lines, orig_lines;
	int col, col2, orig = bufnum, type;
	char last_prefix[FNAMELEN], orig_prefix[FNAMELEN], pat[FNAMELEN], comm[30];

	save_var margin_right = margin;
	save_spot point;
	type = locate_c_comment();	// 1 if block comment, 2 if one-line
	if (!type)					// comment, or 0 if not a comment.
		error("Can't fill comment here, not in a comment.");
	point += parse_string(1, C_COMMENT_FILLER);	// Past 1st line's prefix.
	sprintf(pat, "[ \t]*%s[^0-9a-z_\n]*", generic ? get_comment_start() : "/[*/]");
	grab(point, point - parse_string(-1, pat), orig_prefix);
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (!generic)
		point -= parse_string(-1, "<*>/");
	save_var narrow_end = size() - point + 1;

	if (type == 1 && !generic)		// Combine adjacent block comments.
		simple_re_replace(1, "[ \t]*<*>/[ \t]*\n?[ \t]*/<*>[ \t]*", " ");
	point = 0;
	while (nl_forward()) {	// Find width of narrowest prefix.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		width = MIN(width, col);
	}
	pbuf = tmp_buf();
	point = 0;
	get_comment_marker(comm, generic);
	sprintf(pat, "[ \t]*%q", comm);
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		if (col > width)	// Make sure it's not too wide.
			col = width;
		if (type == 2) {	// But if it's made of one-line comments,
			i = parse_string(1, pat);
			col2 = get_column(point + i);
			if (col < col2)		// each prefix must include a "//".
				col = col2;
		}
		i = column_to_pos(col);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	if (width == 1000)	// No prefix, so use default.
		width = 3;
	orig_lines = old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 20),
						   MAX(margin_right - width, 20));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;
	*last_prefix = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		grab(point, give_end_line(), last_prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	if (orig_lines < 2) {		// Default for new lines if 1 line originally.
		strcpy(last_prefix, orig_prefix);
		if (type == 2 && !strstr(last_prefix, comm))
			sprintf(last_prefix, "%s%s", comm,
					(*comm && comm[strlen(comm) - 1] == ' ') ? "" : " ");
		else if (type == 1 && !generic)
			if (!strstr(last_prefix, "/*"))
				strcpy(last_prefix, (auto_fill_comment_rules & 4) ? " " : " * ");
			else {
				replace_in_string(last_prefix, '/', ' ');
				if (fill_c_comment_plain)
					replace_in_string(last_prefix, '*', ' ');
			}
	}
	bufnum = orig;
	narrow_start = 0;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(last_prefix);			// Add some prefixes.
		if (generic) {
			save_var indenter_action = IACT_FILL_COMMENT;
			to_begin_line();
			if (indenter)
				(*indenter)();
		}
		nl_forward();
	}
	buf_delete(pbuf);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

int locate_c_comment()
{
	int com, type = 1, generic = strcmp(major_mode, "C");

	save_spot narrow_start, narrow_end;
	com = in_c_comment(c_look_back);	// Comment?  If so, what type?
	if (com == IN_OLD_COMMENT || !com && !generic && parse_string(1, "[ \t]*/<*>"))
		narrow_to_block_comment();
	else if (narrow_to_new_comment())
		type = 2;
	else
		return 0;
	to_begin_line();
	if (parse_string(1, C_COMMENT_PREFIX "$"))
		point--;		// If on line right after a parag, go back.
	to_begin_line();
	if (parse_string(1, C_COMMENT_PREFIX "$")) // If no text on line,
		if (!re_search(-1, "[^ \t\n]"))	// Find a line with text.
			re_search(1, "[^ \t\n]");

	re_search(1, C_COMMENT_PREFIX "$");	// Find "blank" line w/only filler.
	if (parse_string(-1, C_COMMENT_PREFIX "$"))
		to_begin_line();	// And move to its start.
	re_search(-1, "[^ \t\n]");	// Find text of some parag before it.
	nl_forward();
	mark = point;
	point--;
	re_search(-1, C_COMMENT_PREFIX "$");	// Find "blank" line before.
	if (parse_string(1, C_COMMENT_PREFIX "\n"))
		nl_forward();
	return type;
}

when_setting_c_fill_column()
{
	if (now_parsing_cmds
		  || !ask_yn("Set for existing too?",
					 "Set right margin for all existing C mode buffers also?",
					 "Yes", "No", 1))
		return;
	save_var bufnum;
	int buf = buf_list(0, 2);
	do {
		bufnum = buf;
		if (!strcmp(major_mode, "C"))
			margin_right = c_fill_column;
	} while (buf = buf_list(-1, 1));
}

command fill_paragraph()
{
	int end;

	if (has_arg)
		save_var margin_right = (iter > 5) ? iter : 32000;
	if (is_highlight_on()) {
		fill_region();
		return;
	}
	save_var point;
	iter = 0;
	point--;
	forward_paragraph();
	re_search(-1, "[ \t]*");
	end = point - 1;
	backward_paragraph(); 
	re_search(1, "[ \t\n]*");	/* leave leading whitespace intact */
	if (point < end)
		region_fill(point, end);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

// Fill paragraph between a and b.  Make the first line up to
// first_width long.  Make remaining lines up to rest_width.
region_fill_first_line(int a, int b, int first_width, int rest_width)
{
	int start, atend, len;

	if (a > b)			/* switch so a < b */
		start = a, a = b, b = start;
	save_spot point = a;
	if (character(b - 1) == '\n')	// Don't merge into line following this region.
		b--;
	save_var narrow_start = a, narrow_end = size() - b;
	save_var abort_searching = 0;
	point = 0;		/* remove previous line breaks */
	while (re_search(1, "( [ \t\n]+)|([\t\n][ \t\n]*)")) {
		len = point - matchstart;
		atend = character(matchstart) == '\n';
		delete(matchstart, point);
		if (sentence_end_double_space && (len >= 2 || atend)
			  && parse_string(-1, "[.?!][])'\"]*", NULL))
			stuff("  ");
		else
			insert(' ');
	}
	point = 0;
	widened_move_to_column(first_width);
	for (;;) {		/* put in new line breaks */
		if (point >= size() - narrow_end)
			break;
		while (curchar() == ' ')
			point++;
		while (re_search(-1, " +"))
			if (!mode_restrict_break || mode_restrict_break(point))
				break;
		if (!re_search(1, " +"))
			break;
		delete(matchstart, point);
		insert('\n');
		narrow_start = point;	/* never look before here */
		widened_move_to_column(rest_width);
	}
	build_first = 1;		/* redisplay hint */
}

#define GAMS_DATA_KEYWORDS "set|sets|scalar|scalars|parameter|parameters|table" \
						   "|variable|variables|equation|equations|model|models" \
						   "|file|files|acronym|acronyms"

// These keywords (plus the ones above) start units.  Semicolon is
// optional before them.
#define GAMS_UNIT_STARTERS "option|options|loop|while|alias|abort|solve|display"

// Other keywords.
#define GAMS_KEYWORDS "card|ord|inf|na|eps|yes|no|smax|smin|not|and|or|sum|prod" \
					  "|exp|all|assign|diag|else|eq|for|ge|gt|if|le|lt|maximizing" \
					  "|minimizing|ne|putpage|puttl|repeat|sameas|system|then" \
					  "|until|using|xor"

// These can come before a VARIABLE declaration.
#define GAMS_TYPE_KEYWORDS "binary|negative|positive|integer|free|semicont" \
						   "|semiint|sos1|sos2|nonnegative|nonpositive"

// Identifiers that aren't really GAMS keywords but should be
// highlighted anyway.
#define GAMS_SPECIALS	"abs|arctan|ceil|cos|errorf|exp|floor|include|log" \
						"|log10|mapval|max|min|mod|normal|power|put|putclose" \
						"|round|sign|sin|sqr|sqrt|trunc|undf|uniform"

// A GAMS statement ends with a semicolon, or with a line break
// followed by a keyword.
#define GAMS_STATEMENT_END	"(;|[ \t]*\n([ \t\n]|[*$].*\n)*(" \
							GAMS_DATA_KEYWORDS "|" GAMS_UNIT_STARTERS "|" \
							GAMS_TYPE_KEYWORDS ")</word|lparen>)"

// A list of possible suffixes.  For speed we color any word
// using the syntax of a suffix, even if it's not in this list.
#define GAMS_SUFFIXES	"date|ifile|ofile|page|rdate|rfile|rtime|sfile|time" \
						"|title|ts|tl|te|l|lo|m|prior|scale|up|lj|nj|sj|tj" \
						"|lw|nw|sw|tw|nd|nr|nz|cc|hdcc|tlcc|cr|hdcr|tlcr|ll" \
						"|hdll|tlll|lp|ws|pc|ps|pw|tm|bm|case|bratio|domlim" \
						"|domusd|holdfixed|iterlim|iterusd|limcol|limrow" \
						"|modelstat|numequ|numinfes|numnopt|numnz|numunbnd" \
						"|numvar|optca|optcr|optfile|reslim|resusd|scaleopt" \
						"|solprint|solveopt|solvestat|sysout|workspace"


color_gams_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk = 0, c, talk_at = from + 40000;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	while (re_search(-1, GAMS_STATEMENT_END))	// Move back to start of a statement.
		if (get_character_color(point) != color_class gams_comment
			  && get_tagged_region("color-gams-as-unit", matchstart) == -1)
			break;
	from = point;
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.]+|^[*$]|[!#?\"'{]|/![*/]|%<%<|_!%$")) {
			t = size();
			break;
		}
		t = matchstart;
		c = character(point - 1);
		gams_color_common(c, t);
		if (isalnum(c) || strchr("_.", c))	// found identifier, kywd, or number
			gams_keyword_color(t, point);
		if (point > talk_at) {
			note("Coloring GAMS file: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point + 2000;
			talk = 1;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk)
		note("");
	return point;
}

gams_color_descriptions()
{
	int descr, done, table, c, p;

	if (parse_string(-1, GAMS_TYPE_KEYWORDS)) {
		gams_move_past_spaces_and_comments();
		if (parse_string(1, GAMS_DATA_KEYWORDS)) {
			set_character_color(matchstart, matchend, color_class gams_keyword);
			point = matchend;
		}
	}
	table = parse_string(-1, "table");	// Special rules for this keyword.
	gams_move_past_spaces_and_comments();
	move_past_gams_identifier();
	do {
		if (parse_string(1, GAMS_STATEMENT_END))
			return;
		point += parse_string(1, "[ \t]*");
		descr = point;
		done = 0;
		c = curchar();
		if (strchr("\"'", c)) {
			gams_parse_and_color_string(p = point);
		} else {
			if (re_search(1, "[\n,/]|" GAMS_STATEMENT_END)) {
				c = character(p = point - 1);
				if (isalpha(c)) {		// relaxed syntax rule: \n plus keyword means ;
					c = ';';
					p = matchstart;
					point = p + 1;
				}
			} else {
				done = 1;
				p = point;
				c = ';';
			}
		}
		set_character_color(descr, p, color_class gams_string);
		switch (c) {
			case ';':
				return;
			case '/':
				gams_color_data_list();
				if (parse_string(1, GAMS_STATEMENT_END)) {
					point = matchstart;
					return;
				}
				point += parse_string(1, "[ \t\n,]*");
				gams_move_past_spaces_and_comments();
				move_past_gams_identifier();
				break;
			case '\n':
			case ',':
			case '"':
			case '\'':
				if (parse_string(1, GAMS_STATEMENT_END)) {
					point = matchstart;
					return;
				}
				point += parse_string(1, "[ \t\n]*");
				if (table) {
					if (re_search(1, GAMS_STATEMENT_END))
						point = matchstart;
					return;
				}
				gams_move_past_spaces_and_comments();
				move_past_gams_identifier();
				break;
		}
	} while (!done);
}

is_gams_data(p)
char *p;
{
	if (strstr("|" GAMS_DATA_KEYWORDS "|" GAMS_TYPE_KEYWORDS "|", p))
		return 1;
	return 0;
}

int get_epsilon_help_format()
{
	char varname[FNAMELEN];
	sprintf(varname, "epsilon-help-format-%s-%s",
			is_unix ? "unix" : is_win32 ? "win" : "other",
			(is_gui || is_unix == IS_UNIX_XWIN) ? "gui" : "console");
	int code = get_num_var(find_index(varname));
	if (code == HELP_FORMAT_WINHELP && !(has_feature & FEAT_WINHELP))
		code = 0;		// Ignore setting, it's only available in Windows.
	if (!code)	// Pick a default format.
		if (has_feature & FEAT_WINHELP_NATIVE)	// WinHelp is always installed.
			code = HELP_FORMAT_WINHELP;
		else if (is_unix == IS_UNIX_XWIN || is_gui || is_win32)
			code = HELP_FORMAT_HTML;	// Try using a browser.
		else			// Fall back to internal help system.
			code = HELP_FORMAT_POPUP;
	return code;
}

int html_recolor_from_here_helper()
{
	int start = point;

	// Look for start or end of XML's CDATA block, but ignore the
	// marker if it might be part of a string or comment.
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>!")
		  && !parse_string(-1, "(['\"#]|//||\\\n).*"))
		return point;
	point = start;
	while (re_search(-1, sublanguage_code == SUBLANG_XML
					 ? "%<%!--|--[ \t\n]*>|%<%?php"
					 : "%<%!--|-[ \t\n]*>|%<[%?]|[%?]%>") && curchar() == '<')
		if (html_region_type(point) == -1)
			return point;
	return -1;
}

// Same as previous function, but keep track of regions we've already
// searched to improve speed on huge buffers.
int html_recolor_from_here(int safe)
{
	int start = point, from, to, p, q;
	if (sublanguage_code == SUBLANG_PHP)	// Can have <?'s in strings.
		return safe;		// Can't parse it backwards, so start from top.

	if (modified_buffer_region(&from, &to, "html-color")) {
		// Buffer's changed, so forget saved info on this range.
		if (get_tagged_region("html-color", from, &p, &q) == 1)
			from = p;
		if (get_tagged_region("html-color", to, &p, &q) == 1)
			to = q;
		set_tagged_region("html-color", from, to, -1);
	}
	reset_modified_buffer_region("html-color");

	if (get_tagged_region("html-color", point, &p, &q) == 1)
		point = p;		// Skip back over section previously scanned.
	p = html_recolor_from_here_helper();
	if (p == -1) {		// Remember we found nothing before starting pos.
		set_tagged_region("html-color", narrow_start, start, 1);
		point = start;
		return start;
	}
	// Remember which region had no matches.
	set_tagged_region("html-color", point, p, 1);
	return p;
}

color_html_attributes(int script, int orig, int limit)
{
	int len, start, lang;

	if (script == 2)		// A <jsp:> tag.
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x10, HT_JSCRIPT));
	else if (script == 3)
		lang = HT_CSS;
	else if (script == 4)
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x20, HT_JSCRIPT));
	else if (script)
		lang = get_html_asp_language(html_other_coloring);
	for (;;) {
		if (curchar() == '<' && parse_string(1, "%<%!---")) {
			point = matchend;
			color_html_comment();
			re_search(1, "[ \t\n]*");	// Parse whitespace.
		}
		if (!(len = parse_string(1, "[-_./a-z0-9]*")))	// An attribute?
			break;
		if (script == 1 && parse_string(1, "(language|type)[ \t\n]*=[ \t\n]*"
										"[\"']*(text/)?")) {
			start = matchend;
			if (matches_at(start, 1, "(javascript|jscript|ecmascript)"))
				lang = HTML_SPECIFIC_COLORING(1, HT_JSCRIPT);
			else if (matches_at(start, 1, "(vb)"))
				lang = HTML_SPECIFIC_COLORING(2, HT_VBSCRIPT);
			else if (matches_at(start, 1, "(php)"))
				lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
			else if (matches_at(start, 1, "(python)"))
				lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
		}
		if (character(point + len) == ':'
			  && matches_at(point + len, 1, ":[-_./a-z0-9]*")) {
			set_character_color(point, point + len + 1, color_class html_namespace);
			point += len + 1;
			len = matchend - point;
		}
		set_character_color(point, point + len, color_class html_attribute);
		point += len;			// Followed by a value?
		if (len = parse_string(1, "[ \t\n]*=[ \t\n]*")) {
			set_character_color(point, point + len, color_class html_punctuation);
							// Color the = and whitespace.
			point += len;		// Value might be quoted.
			start = point;
			if (curchar() == '"') {
				point++;
				html_find_matching_quote('"');
			} else if (curchar() == '\'') {
				point++;
				html_find_matching_quote('\'');
			} else	// No quotes, assume it's whitespace-delimited.
				re_search(1, "[^ \t\n<>]*");
			html_color_value(start, point);
		}
		re_search(1, "[ \t\n]*");	// Parse whitespace.
		set_character_color(matchstart, matchend, color_class html_punctuation);
	}
	if (curchar() == '>') {
		if (character(point - 1) == '/')
			set_character_color(point - 1, point, color_class html_tag);
		set_character_color(point, point + 1, color_class html_punctuation);
		point++;
	}
	start = point;
	if (script && lang != HT_IGNORE && !parse_string(-1, "/%>")) {
		if (re_search(1, script == 3 ? "%</style%>"
					  : script == 2 ? "%</jsp:.*%>" : "%</(cf)?script(text)?%>")) {
			set_tagged_region("color-as-unit", orig, point, 99);
			point = matchstart;
		} else
			set_tagged_region("color-as-unit", orig, point, 99);
		color_embedded_code(start, point, lang, limit);
	}
}

color_html_comment()
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (!sublanguage_code && html_recognize_coldfusion_comments
			   && parse_string(1, "%<%!---") && !parse_string(1, "%<%!----%>")) {
		int level = 0;	// A legal ColdFusion comment.
		while (re_search(1, "%<%!---|---%>"))
			if (character(matchstart) == '<')
				level++;
			else if (--level <= 0)
				break;
		set_character_color(s, point, color_class html_altcomment);
		set_tagged_region("color-as-unit", s, point, 99);
		return;
	} else if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							? "%<%!--" : "%<[ \t\n]*%![ \t\n]*--")) {
		point = matchend;		// A legal comment.
		re_search(1, "--[ \t\n]*>");	// Search for end of legal comment.
	} else if (parse_string(1, "%<%!DOCTYPE[^][<>]*%[")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		return;
	} else if (parse_string(1, "%<%!(ENTITY|ELEMENT|ATTLIST)[ \t\n]*")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		s = point;
		search(1, ">");
		set_character_color(s, point, color_class html_value);
		if (character(point - 1) == '>')
			set_character_color(point - 1, point, color_class html_punctuation);
		return;
	} else if (parse_string(1, "%<%!DOCTYPE")) {
		if (re_search(1, "[[>]") && character(matchstart) == '[') {
			search(1, "]");
			search(1, ">");
		}
		set_character_color(s, point, color_class html_value);
		return;
	} else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

color_html_move_back()
{
	int res = color_html_move_back_helper(), p;
	if (get_character_color(point, &p, NULL) != color_class html_comment)
		return res;
	point = p;		// Anything we found inside a comment doesn't count.
	return 4;		// Go to top of comment instead.
}

color_html_move_back_helper()
{
	int start = point, p, q, from, to, res;

	save_var narrow_start = point - color_html_look_back;

	if (modified_buffer_region(&from, &to, "html-move")) {
		if (get_tagged_region("html-move", from, &p, &q) == 1)
			from = p;
		if (get_tagged_region("html-move", to, &p, &q) == 1)
			to = q;
		set_tagged_region("html-move", from, to, -1);
	}
	reset_modified_buffer_region("html-move");
	if (get_tagged_region("html-move", point, &p, &q) == 1)
		point = p;

	res = re_search(-1, "%<%!%[CDATA%[|%]%]%>");
	p = MIN(point + 10, start);
	if (res)
		p = matchstart;
	set_tagged_region("html-move", p, start, 1);
	if (res && !parse_string(-1, "%<script[^<]*%><wspace>*"))
		if (curchar() == '<')
			return 1;
		else {
			point += parse_string(1, "%]%]%>");
			return -1;
		}
	point = start;
	if (sublanguage_code != SUBLANG_XML
		  && re_search(-1, "<Langle>[ \t\n]*/?((cf)?script|style)")
		  && parse_string(1, "<Langle>[ \t\n]*((cf)?script|style)"))
		return 2;				// Found a <script>, not a </script>.
	point = start;
	if (sublanguage_code != SUBLANG_XML
		  && re_search(-1, "<Langle>[%?]|[%?]<Rangle>") && curchar() == '<'
		  && get_tagged_region("color-as-unit", point) == -1)
		return 3;				// Found a <% or <?, not a %> or ?>.
	point = start;
	do {			// start coloring from the previous tag
		if (!search(-1, "<"))
			break;	// but ignore <'s before " or ' characters
		if (get_character_color(point, &p, NULL) == color_class html_comment) {
			point = p + 2;	// If we're in a comment,
			search(-1, "<!");	// go back to its start.
		}
	} while (html_in_quoted_attribute());
	return 0;
}

color_html_loop(int end)
{
	int s, t, i, script;

	for (s = point; point < end; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							 ? "(<:|_|p:ID_Start><:|hyphen|.|p:ID_Continue>*|#[0-9]+|#x[0-9a-f]+);?"
							 : "([a-z0-9]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()) && color_html_include(end))
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = (s == 6 && parse_string(1, "script")
					  || s == 10 && sublanguage_code == SUBLANG_XML
					  && parse_string(1, "scripttext"));
			if (sublanguage_code != SUBLANG_XML) {
				if (s == 3 && parse_string(1, "jsp:(declaration|expression"
										   "|scriptlet)</word>"))
					script = 2;
				if (s == 5 && parse_string(1, "style"))
					script = 3;
				if (s == 8 && parse_string(1, "cfscript"))
					script = 4;
			}
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			for (i = point; strchr(" \t\n", curchar()); )
				point++;	// Parse whitespace.
			set_character_color(i, point, color_class html_punctuation);
			if (parse_string(1, "/[ \t\n]*>")) {
				set_character_color(point, point + 1, color_class html_tag);
				set_character_color(point + 1, matchend,
									color_class html_punctuation);
				point = matchend;
			} else
				color_html_attributes(script, t, end);
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
}

do_html_indent(int manual)
{
	int region, start;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 200);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	if (region > 0 && region <= HT_CSS && !IS_ENTER_KEY(key)
		  && !(html_indenting_rules & (1 << (region - 1)))) {
		prev_indenter();	// Can move back to prev line's indent.
		return;
	}
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	if (region >= HT_JSCRIPT && region < HT_IGNORE) {
		if (all_blanks(narrow_start, point)) {	// Start of script block.
			narrow_start = 0;	// 0x400 indent; 0x800 indent even if <% has none.
			int i = get_indentation(start);
			if ((html_indenting_rules & 0x400)
				  && ((html_indenting_rules & 0x800) || i))
				indent_to_column(i + give_html_indent());
			return;
		}
		html_exclude_script_head();
	}
	switch (region) {
		case HT_JSCRIPT:
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			do_vbasic_indent();
			return;
		case HT_PHP:
			do_php_indent();
			return;
		case HT_PYTHON:
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

int do_html_find_matching_tag(int pos, char *tag)
{
	return try_html_find_matching_tag(pos, tag, 0);
}

// Like above but accept flags: 1=on keypress immediately return -4.
int try_html_find_matching_tag(int pos, char *tag, int flags)
{
	char pat[FNAMELEN], *endcolor = "html-punctuation";
	int level, dir = 0, i;

	*tag = 0;
	save_var point = pos, case_fold = (sublanguage_code != SUBLANG_XML);
	i = parse_string(1, "%<[ \t\n]*");
	if (!i)
		return -2;
	point += i;
	if (curchar() == '/') {
		dir = REVERSE;
		level = 0;
		point++;
	} else
		level = 1;
	point += parse_string(1, "[ \t\n]");
	int color = get_character_syntax_color(point);
	if (color == color_class html_value)
		endcolor = name_color_class(color);
	if (!parse_string(1, "[-_./A-Za-z0-9:]+", tag))
		return -2;
	if (html_is_self_delimiting(point, endcolor))
		return -3;
	sprintf(pat, "%%<(/?<wspace>*<c:%s>%q)</word>",
			(color >= 0) ? name_color_class(color) : "-1", tag);
	while (do_color_searching(REGEX | dir, pat) == 1) {
		if ((flags & 1) && (char_avail() || is_key_repeating()))
			return -4;
		i = (character((dir ? matchend : matchstart) + 1) == '/') ? -1 : 1;
		if (!html_is_self_delimiting(point, endcolor))
			if (!(level += i))
				break;
	}
	if (dir == REVERSE)
		point++;
	if (level)
		return -1;
	return point;
}

// List element context at point into buffer b.  Flags: 1=multi-line
// format, 2=give up on keypress and return 0.
int do_list_element_nesting(int b, int flags)
{
	char tag[FNAMELEN], depth[FNAMELEN];
	save_var point, case_fold = (sublanguage_code != SUBLANG_XML);
	int open = html_in_tag() != -1;
	while (do_color_searching(REVERSE | REGEX, "<c:html-punctuation>[<]") == 1) {
		if ((flags & 2) && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		set_buf_point(b, 0);
		int i = html_is_start_tag(point, 0);
		if (i == -1) {
			i = try_html_find_matching_tag(point, tag, (flags & 2) ? 1 : 0);
			if (i >= 0) {
				point = i;
				search(-1, "<");	// Get from element name to before its <.
			} else if (i == -4)
				return 0;
			else if (!(flags & 1))
				buf_printf(b, "%s? ", tag);
			else if (i == -1)
				buf_printf(b, "</%s> unmatched at line %d, position %d.\n",
						   tag, lines_between(0, point) + 1, point);
			else
				buf_printf(b, "Can't recognize tag at line %d, position %d.\n",
						   lines_between(0, point) + 1, point);
		} else if ((i == 1 || !i && open) && html_get_tag_name(point, tag) == 1) {
			char *bad = html_invalid_tag(point + 1) ? "?" : "";
			if ((flags & 1) && *bad)
				buf_printf(b, "Unclosed <%s at line %d, position %d.\n",
						   tag, lines_between(0, point) + 1, point);
			else
				buf_printf(b, !(flags & 1) ? (open ? "<%s%s " : "%s%s ")
										   : (open ? "<%s%s\n" : "<%s%s>\n"),
						   tag, bad);
		}
		open = 0;
	}
	if (flags & 1) {
		save_var bufnum = b, point = 0;
		*depth = 0;
		while (re_search(1, "^[<]" HTML_NAME)) {
			set_character_color(matchstart + 1, matchend, color_class html_tag);
			point = matchstart;
			bprintf("%s", depth);
			strcat(depth, "  ");
			to_end_line();
		}
	}
	return 1;
}

html_func_name_finder()
{
	int start, end, region, res = 1;
	*display_func_name = 0;
	if (sublanguage_code == SUBLANG_CSS)
		return 1;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	save_var narrow_start = start, narrow_end = size() - end, bufnum;
	if (region <= 0 && html_display_nesting_width >= 5
		  && (html_display_definition
			  & ((sublanguage_code == SUBLANG_XML) ? 4 : 8))) {
		int b = tmp_buf();
		narrow_start = narrow_end = 0;
		save_var point = give_begin_line();
		if (!do_list_element_nesting(b, must_find_func_name ? 0 : 2))
			return 0;
		html_trim_element_nesting(b);
		restore_vars();
		buf_delete(b);
		return 1;
	}
	if (region > 0 && (html_display_definition & 3)) {
		html_exclude_script_head();
		switch (region) {
			case HT_JSCRIPT:
				save_var mode_move_level = c_move_level;
				if (html_display_definition & 1)
					res = javascript_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "JavaScript");
				return res;
			case HT_VBSCRIPT:
				save_var mode_move_level = 0;
				if (html_display_definition & 1)
					res = vbasic_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "VBScript");
				return res;
			case HT_PHP:
				save_var mode_move_level = perl_move_level;
				if (html_display_definition & 1)
					res = inner_php_func_name_finder();
				if ((html_display_definition & 2) && res
					  && !*display_func_name && strcmp(major_mode, "PHP"))
					strcpy(display_func_name, "PHP");
				return res;
			case HT_PYTHON:
				save_var mode_move_level = 0;
				if (html_display_definition & 1)
					res = python_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "Python");
				return res;
			case HT_CSS:
				if (html_display_definition & 2)
					strcpy(display_func_name, "CSS");
				return res;
		}
	}
	return 1;
}

int html_restrict_break(int pos)
{
	if (get_character_syntax_color(pos) == color_class html_value) {
		int start;
		get_character_color(pos, &start);
		return !strchr("\"'", character(start));
	}
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_fill_combine
												: html_auto_fill_combine;
	if (var && parse_string(-1, "%<[ \t]*(" HTML_NAME ")[ \t]*")
		  && find_group(1, 0) - find_group(1, 1) <= var)
		return 0;
	return 1;
}

command html_fill_paragraph()
{
	int region, start, end;

	recolor_buffer_range(point, point + 1);
	region = html_region_type(point);
	switch (region) {
		case HT_JSCRIPT:
			save_var major_mode = _c_mode_name;
			fill_comment();
			return;
		case HT_VBSCRIPT:
		case HT_PHP:
		case HT_PYTHON:
		case HT_CSS:
			fill_comment();
			return;
	}
	if (get_character_color(point, &start, &end) == color_class html_comment)
		save_var narrow_start = start, narrow_end = size() - end;
	save_var mode_restrict_break = html_restrict_break;
	fill_paragraph();
}

suffix_ui()			// Qt user interface file
{
	xml_mode();
}

suffix_xaml()
{
	xml_mode();
}

int find_indent_under(int rev)
{
	int indent = current_column();
	int cnt = (iter > 0) ? iter : -iter;
	iter = 0;
	save_var point;
	to_begin_line();
	re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
	move_to_column(indent);		/* go to same column on prev line */
	if (current_column() > indent)
		point--;

	if (rev) {		// Look backward.
		while (cnt-- > 0)
			if (!re_search(-1, "[ \t\n].") || character(point) == '\n')
				return -1;
		return current_column() + 1;
	} else {		// Look forward.
		while (cnt-- > 0)
			if (!re_search(1, "[ \t\n]") || character(point - 1) == '\n'
				  || !re_search(1, "[^ \t]") || character(point - 1) == '\n')
				return -1;
		return current_column() - 1;
	}
}

command indent_under()
{
	int ind, rev;

	if (is_highlight_on()) {
		save_spot mark, point;
		fix_region();
		to_indentation();
		rev = IS_SHIFT_KEY(full_key) != 0 || iter < 0;
		ind = find_indent_under(rev);
		iter = ind - current_column();	// Use this indent instead.
		has_arg = 1;
	}
	if (maybe_indent_rigidly(0))
		return;
	rev = iter < 0;
	if ((ind = find_indent_under(rev)) != -1)
		indent_to_column(ind);
	else if (rev)
		indent_to_column(0);
	else
		indent_like_tab();
}

command indent_to_tab_stop()
{
	if (maybe_indent_rigidly(0))
		return;
	indent_like_tab();
}

// Alignment rules for C mode.
c_mode_aligner()
{
	// Align variable names being defined.
	ALPAT(4, "((" C_IDENT "[ \t]+)+)([*(]*" C_IDENT "[ \t]*[,;=].*)", 2);
	ALPAT(1, "((" C_IDENT "|[][ \t])+)(=.*)", 1); // Align = in assignments.
	ALPAT(8, "(#define[ \t]+" C_IDENT "(%((" C_IDENT
		  "|[, \t])*%))?[ \t]+)([^ \t\n]+)", 2);	// Align definitions.
	// Align comments.
	ALPAT(2, "(<^c:c-comment>.*)(<c:c-comment>/[/*].*)", 2);
	ALPAT(16, "(.)(\\)$", 2);
}

scroll_by_wheel(clicks, per_click)
{
	if (!mouse_screen) {
		clip_mouse();
		int win = window_at_coords(mouse_y, mouse_x, mouse_screen);
		if (win != -1)
			save_var window_handle = win;
	}
	if (wheel_click_lines >= 0)		// Override control panel setting?
		per_click = wheel_click_lines;
	if (per_click < 1 || per_click >= text_height())
		per_click = text_height() - 1;	// scroll by pages
	window_scroll(clicks * per_click);
	set_display_func_name();
	restore_vars();
	redisplay();
}

// Color simple variable references within this string-like thing (already
// colored itself).
color_perl_interpolated(int from, int to, int comments)
{
	int p;

	save_var point = from;
	save_var narrow_end = size() - to;
	while (re_search(1, (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					 ? "[\\$]" : "[\\$@#]")) {
		p = point;
		if (character(point - 1) == '\\')
			point++;
		else if (character(point - 1) == '#') {
			if (!comments)
				continue;
			from = point - 1;
			to_end_line();
			if (point > to)
				point = to;
			set_character_color(from, point, color_class perl_comment);
		} else if (sublanguage_code == SUBLANG_PHP) {
			if (parse_string(-1, "[^\\]{%$")) {
				point -= 2;
				to_default_move_level(1, "{", "}", 1);
				set_character_color(p - 2, point, color_class perl_variable);
			} else if (parse_string(1, "<word>+(-%><word>+)*|{<word>+}")) {
				point = matchend;
				if (curchar() == '[')
					to_default_move_level(1, "[", "]", 0);
				set_character_color(p - 1, point, color_class perl_variable);
			}
		} else if (in_shell_buffer) {
			if (curchar() == '(') {
				shell_parse_substitution(point - 1);
			} else if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									"|[^ \t\n`]"))
				set_character_color(matchstart - 1, point = matchend,
									color_class perl_variable);
		} else if (curchar() == '{') {
			to_default_move_level(1, "{", "}", 1);
			set_character_color(p - 1, point, color_class perl_variable);
		} else if (parse_string(1, "#?([0-9]+|[A-Za-z_][0-9A-Za-z_]*"
								"[:0-9A-Za-z_]*)|%^[A-Z]"
								"|[]-[&`'+*|%=~^_./,\\\";:#?!@$<>()]") && point < to) {
			point = matchend;
			for (;;) {
				if (curchar() == '-' && character(point + 1) == '>'
					  && (isalnum(character(point + 2))
						  || strchr("[{", character(point + 2))))
					point += 2;
				else if (curchar() == '[' || curchar() == '{') {
					if (to_default_move_level(1, "[{", "]}", 1) < 0)
						break;
				} else
					break;
			}
			set_character_color(p - 1, point, color_class perl_variable);
		}
	}
}

tag_mode_perl()
{
	char func[TAGLEN], pat[FNAMELEN], class_name[FNAMELEN];
	int start, i, have_class, class_end = -1, have_const, have_php_const;

	*class_name = 0;
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	sprintf(pat, "<^word>(%s)([a-zA-Z_]<word>*"
			"((::|')[a-zA-Z_]<word>*)?)",
			(sublanguage_code == SUBLANG_PHP)
			? "(function|class)[ \t]+|define[ \t\n]*%([ \t\n]*\""
			: "(sub|use[ \t\n]+constant)[ \t]+");
	while (re_search(1, pat)) {
		have_const = character(find_group(1, 1))
					 == ((sublanguage_code == SUBLANG_PHP) ? 'd' : 'u');
		have_php_const = have_const && (sublanguage_code == SUBLANG_PHP);
		if (get_character_color(find_group(1, 1))
			  == (have_php_const ? color_class perl_function : color_class perl_keyword)
			  && get_character_color(find_group(3, 1))
			  == (have_php_const ? color_class perl_string : color_class perl_function)) {
			grab(start = find_group(3, 1), find_group(3, 0), func);
			have_class = (sublanguage_code == SUBLANG_PHP)
						 && character(find_group(1, 1)) == 'c';
			if (*class_name && point > class_end)
				*class_name = 0;
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (curchar() != '{')
				have_class = 0;
			if (have_class && (i = perl_move_level(1, 0)) > 0) {
				if (tag_which_items & 16)
					add_tag(func, start);
				strcpy(class_name, func);
				class_end = i;
				continue;
			}
			if (!have_const && curchar() == '(') {
				i = perl_move_level(1, 0);
				if (i > 0 && i < size() - narrow_end)
					point = i;
			}
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (have_const) {
				if (tag_which_items & 4)
					add_tag(func, start);
			} else if (curchar() == '{' || (tag_declarations & 1))
				add_class_tag(class_name, "::", func, start);
		}
	}
}

command push()
{
	char cmd[FNAMELEN * 2];

	_last_was_grep = 0;
	save_var clear_process_buffer = (clear_process_buffer & 3) != 0;
	do_push(get_cmdline("Push With Command", cmd), capture_output, 1);
}

command start_process()
{
	char cmd[FNAMELEN * 2], *cmdline, *sh, fixup[FNAMELEN], *p;
	char cust1[FNAMELEN], cust2[FNAMELEN], buf[FNAMELEN];

	save_var clear_process_buffer = (clear_process_buffer & 5) != 0;
	_last_was_grep = 0;
	tiled_only();
	error_if_input(1);
	strcpy(buf, PROCBUF);
	if (has_arg && (has_feature & FEAT_MULT_CONCUR))
		pick_other_proc_buffer(buf);
	else if (no_running(1))
		return;
	iter = has_arg = 0;
	cmdline = get_cmdline("Start Process With Command", cmd);
	create_concur(cmdline, buf);
	locate_window(buf, "");
	sh = getenv("EPSSHELL");
	if (!sh)
		sh = getenv("SHELL");
	if (opsys == OS_UNIX && sh) {
		*cust1 = *cust2 = *fixup = 0;
		if (p = lookpath("epsilon.sh", 0))
			strcpy(cust1, p);
		if (p = lookpath("epsilon.csh", 0))
			strcpy(cust2, p);
		if (!*cust1 || !*cust2)
			sprintf(fixup, "echo \"Customization file not found, "
					"no customizations performed.\">/dev/null");
		else if (strstr(sh, "/sh") || strstr(sh, "/bash") || strstr(sh, "/ksh")
				 || strstr(sh, "/zsh"))
			sprintf(fixup, ". \"%s\"; set +o emacs", cust1);
		else if (strstr(sh, "/csh") || strstr(sh, "/tcsh"))
			sprintf(fixup, "source \"%s\"", cust2);
		else
			sprintf(fixup, "echo \"Unrecognized shell, no customizations "
					"performed.\">/dev/null");
		delay(40, 0, bufnum);	// Let's first get a prompt if we can.
		point = type_point;
		bprintf("%s\n", fixup);
	}
	point = size();
	process_mode();
}

command make()
{
	char cmd[FNAMELEN * 2];

	save_var clear_process_buffer = (clear_process_buffer & 17) != 0;
	tiled_only();
	get_cmdline("Make using command", cmd);
	save_var start_process_in_buffer_directory
			= start_make_in_buffer_directory;
	do_compile(cmd);
}

int proc_complete_list_matches(char *pat, int initial,
							   int skip_dots, int flags)
{
	int proc_tab_size = tab_size;
	char *pat2 = 0, file[FNAMELEN], *s, *t, cmdnames = 0;

	if (initial && (s = getenv("PATH")) && !strchr(pat, '/') &&
		  (opsys == OS_UNIX || !strchr(pat, '\\') && !strchr(pat, ':'))) {
		pat2 = malloc(strlen(s) + strlen(pat) + 100);
		if (opsys == OS_UNIX)
			sprintf(pat2, "{%s}/%s*", s, pat);
		else
			sprintf(pat2, "{%s,.}/%s*.{%s}", s, pat,
					process_completion_windows_programs);
		while (s = strchr(pat2, path_list_char))
			*s = ',';
		flags |= EXACTONLY;
		pat = pat2;
		cmdnames = 1;
	}
	zap(_MATCH_BUF);
	char last[FNAMELEN], *orig;
	*last = 0;
	if (s = do_file_match(pat, STARTMATCH | flags))
		do {
			if (opsys == OS_UNIX && initial && lowaccess(s, LOWACC_X))
				continue;
			orig = s;
			if (cmdnames) {
				strcpy(file, s);
				s = get_tail(file, 0);
				if (opsys != OS_UNIX)
					*get_extension(s) = 0;
			}
			if (skip_dots && *get_tail(s, 0) == '.'
				  && string_matches_pattern(get_tail(s, 0), "%.%.?[/\\]?<eof>"))
				continue;
			if (cygwin_filenames & 64)
				while (t = strchr(s, '\\'))
					*t = '/';
			buffer_printf(_MATCH_BUF, "%s\n", s);
			strcpy(last, orig);
		} while ((s = do_file_match(pat, 0)) && !user_abort);
	free(pat2);	// Free if we allocated it.
	save_var bufname = _MATCH_BUF;
	save_var abort_searching = 0;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	tab_size = proc_tab_size;
	int cnt = lines_between(0, size(), 0);
	if (cnt == 1 && (process_completion_style & 8) && *last) {
		absolute(last);
		say("Completed: %s", last);
	}
	return cnt;
}

command process_complete()
{
	char pat[FNAMELEN], pref[FNAMELEN];
	int cnt = 0, start, end, buf = create(_MATCH_BUF);
	int offset = IS_SHIFT_KEY(full_key) ? -1 : 1;

	start = end = point;
	if (start < type_point) {
		if (!maybe_indent_rigidly(0))
			indent_like_tab();
		return;
	}
	*pat = 0;
	if (process_line_inside_quoting() || parse_string(-1, "[^\n\t <>=\"]+", pat)) {
		start = point;
		point = end;
		if (process_line_inside_quoting())
			grab(start = matchend + 1, end, pat);
		if (start < type_point && type_point < end) {
			start = type_point;		// exclude prompt
			grab(start, end, pat);
		}
	}
	if (!*pat)
		strcpy(pat, "*");
	if (!(process_completion_style & 1) || prev_cmd != CMD_PROC_COMPL) {
		int dircmdlen = matches_at_length(start, -1, process_completion_dircmds);
		int dirsonly = dircmdlen && ((dircmdlen == start - type_point)
					 || matches_at(start - dircmdlen, -1, "[][#&;()] *"));
		int initial = all_blanks(type_point, start)
					  || matches_at(start, -1, "([&;|]|</word>(sudo)) +");
		cnt = proc_complete_do_list_matches(pat, initial,
											process_completion_style & 2,
											dirsonly ? FM_ONLY_DIRS : 0);
		check_abort();
		offset = 0;
		if (!cnt)
			error("No matches for %s.", pat);
		set_buf_point(buf, 0);
	}
	if (!(process_completion_style & 1))
		find_buffer_prefix_folding(name_to_bufnum(_MATCH_BUF), pref, !is_unix);
	else {
		buf_delete_final_newlines(buf);
		grab_line_offset(buf, pref, offset, 1);
	}
	if ((process_completion_style & 1) || strlen(pref) > end - start) {
		delete(start, end);
		if (strchr(pref, ' ') && !process_line_inside_quoting())
			stuff("\"");
		stuff(pref);
		if ((process_completion_style & 1)) {
			if (is_path_separator(character(point - 1)))
				delete(point - 1, point);
			this_cmd = CMD_PROC_COMPL;
			return;
		}
		if (cnt == 1 && !is_path_separator(character(point - 1))) {
			if (strchr(pref, ' ') && process_line_inside_quoting())
				stuff("\"");
			stuff(" ");
		}
	} else if (prev_cmd == CMD_PROC_COMPL) {
		proc_complete_show_matches();
		this_cmd = CMD_PROC_COMPL;
	} else if (cnt == 1 && !is_path_separator(character(point - 1))) {
		if (strchr(pref, ' ') && process_line_inside_quoting())
			stuff("\"");
		stuff(" ");
	} else if (!(process_completion_style & 4)
			   && buffer_size(_MATCH_BUF) < screen_cols - 20 - strlen(pat)) {
		save_var bufname = _MATCH_BUF;
		simple_re_replace(1, "\n", " ");
		grab(0, size(), pref);
		say("%d matches for %s: %s", cnt, pat, pref);
		expire_message = 1;
	} else {
		say("%d matches for %s: press again to see matches.", cnt, pat);
		expire_message = 1;
		this_cmd = CMD_PROC_COMPL;
	}
}

is_python_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	save_var case_fold = 0;
	if (strstr("|and|del|for|is|raise|assert|elif|from|lambda|return|"
			   "break|else|global|not|try|class|except|if|or|while|"
			   "continue|import|pass|def|finally|in|yield|", p))
		return 1;
	if (strstr((python_language_level >= 3) ? "|False|None|True|as|nonlocal|with|"
									 : "|print|exec|", p))
		return 1;
	return 0;
}

save_session()
{
	char file[FNAMELEN];

	if (preserve_session_once)
		save_var preserve_session = preserve_session_once - 1;
	if (!(preserve_session & (SAVE_RESTORE_SESSION | SAVE_SESSION)))
		return 0;
	if (!preserve_session_flags && preserve_session != SAVE_SESSION)
		return 0;	// Never read a session, don't write (unless flags say to).
	give_session_file(file);
	int err = do_save_session(file);
	if (err && session_warn_when_saving) {
		save_var no_popup_errors = 1;
		file_error(err, file, "session file write error");
		if (ask_yn("Can't save session, exit anyway?",
				   "Error writing session file, exit anyway? ",
				   "Yes", "No", 1) != 1)
			quick_abort();
	}
	return err;
}

session_save_buffer(int b, char *sdir)
{
	char file[FNAMELEN];
	int transl = GET_LINE_TRANSLATE(translation_type);
	int uni_compat = 0;

	if (has_feature & FEAT_UNICODE) {
		char *s = encoding_to_name(GET_ENCODING(translation_type));
		if (s)
			uni_compat = strstr(s, "UTF-16") != 0;
		if (uni_compat)
			transl = FILETYPE_BINARY;
	}
	buf_printf(b, "%s\n", bufname);
	if (is_dired_buf()) {
		insert_char_in_buf(b, 0);	// Hide file pattern from old versions.
		buf_printf(b, "%s\n", get_dired_buffer_pattern());
	} else {
		relative(filename, file, sdir);
		buf_printf(b, "%s\n", file);
	}
	buf_printf(b, "%d %d %d %d %d %d %d",
			   point, mark, transl,
			   is_dired_buf(), dired_buffer_pattern_wildcards,
			   dired_sorts_files, force_save_as);
	if ((has_feature & FEAT_UNICODE) && (translation_type & FORCED_TRANS))
		buf_printf(b, " %d @%s", GET_LINE_TRANSLATE(translation_type),
				   encoding_to_name(GET_ENCODING(translation_type)));
	buf_printf(b, "\n");
	buf_printf(b, "%d %d %d %d\n", over_mode, fill_mode,
			   indent_with_tabs, margin_right);
	transl = FILETYPE_AUTO;
	if (uni_compat || (translation_type & FORCED_TRANS))
		transl = GET_LINE_TRANSLATE(translation_type);
	// This flavor of the translation type is used by non-Unicode/
	// pre-v12 versions only.
	buf_printf(b, "%d %d %d %d %d %d %d %d %d %d\n", tab_size,
			   case_fold, 0, want_warn,
			   _buf_readonly, buf_accessed,
			   (file_io_converter == oem_file_converter),
			   uni_compat,
			   transl, soft_tab_size);
}

int session_restore_one_file(int b, char *file, char *sdir, int *readerrs)
{
	int i, trans = FILETYPE_AUTO;
	int num1[10], num1cnt, num2[10], num2cnt, num3[10], num3cnt;
	char firstline[150];
	jmp_buf this_level;

	grab_line(b, firstline);
	if ((num1cnt = break_into_numbers(firstline, num1)) < 3)
		return 2;
	if ((num2cnt = grab_numbers(b, num2)) < 4)
		return 3;
	if ((num3cnt = grab_numbers(b, num3)) < 5)
		return 4;
	if (session_restore_files)
		filename = file;
	save_var top_level = &this_level;
	if (setjmp(top_level))
		return -1;
	if (num1cnt >= 4 && num1[3] && session_restore_files) {	// restore dired buf
		filename = "";
		if (num1cnt >= 5)
			save_var file_pattern_wildcards = num1[4];
		save_var show_status = 0;
		if (num1cnt >= 6)
			dired_sorts_files = num1[5];
		create_dired_listing(file, 0);
		if (is_remote_file(get_dired_buffer_pattern()))
			dired_orig_position = -1;
	} else if (*file && session_restore_files) {
		if (!is_remote_file(file))
			absolute(file, sdir);
		if (has_feature & FEAT_UNICODE) {
			if (num1cnt >= 7) {		 // If explicit trans, force.
				char *p;
				if (p = strchr(firstline, '@')) {
					i = encoding_from_name(p + 1);
					if (i)
						trans = MAKE_TRANSLATE(i, num1[7], 1);
				}
			}
		} else {
			if (num3cnt > 8 && num3[7])
				trans = num1[2];	// If explicit trans, or common unicode, force.
		}
		i = do_file_read(file, trans);
		force_save_as = 0;
		if (num1cnt >= 7)
			force_save_as = num1[6];
		if (user_abort || i == EREADABORT)
			return 0;
		if (!i || i == 2) {
			if (!i)
				set_readonly_from_file();
			window_bufnum = bufnum;		// make suffix- kbd macros work
			call_mode(filename);
			make_mode();
		} else if (!*readerrs++) {
			file_error(i, file, "read error");
			maybe_ding(bell_on_read_error);
		} else
			show_text(0, -1, "%d files could not be read", *readerrs);
	}
	if (session_restore_files) {
		if (bufnum != b)	/* same tmp name as session buffer */
			point = num1[0];
		mark = num1[1];
		if (is_dired_buf())
			move_to_column(goal_column);
	}
	if (session_restore_files) {
		session_restore_one_file_1(num2, num2cnt);
		session_restore_one_file_2(num3, num3cnt);
	}
	return -1;
}

ev_header(char *buf)
{
	save_var bufname = buf;
	if (lookpath("include/eel.h", 0))
		bprintf("#include <eel.h>\n");
	else
		bprintf("#include \"eel.h\"\n");
}

void eval_template(char *expr, char *src, int in_buf)
{
	char quoted[300], *p, *q, *format = 0, copy[300];

	strcpy(copy, expr);
	expr = copy;
	for (p = expr; p = strchr(p, ';'); p++) {
		if (strchr(p, '"'))
			continue;
		format = p + 1;
		*p = 0;
		break;
	}
	save_var bufnum = zap(src);
	for (p = expr, q = quoted; *q = *p; p++, q++)
		if (*p == '"') {
			*q++ = '\\';
			*q = '"';
		}
	ev_header(src);
	bprintf("ev_template(){\n");
	bprintf("char *strval;\n");
	bprintf("  if (sizeof(%s) <= sizeof(int))\n", expr);
	char prefix[20];
	*prefix = 0;
	if (format)
		if (p = strpbrk(format, "xo"))
			sprintf(prefix, "0%c", *p);
	if (in_buf)
		bprintf("    bprintf(\"%s%%%s\", (%s));\n",
				prefix, format ? format : "d", expr);
	else
		bprintf("    say(\"%%s == %s%%%s\", \"%s\", (%s));\n",
				prefix, format ? format : "'d", quoted, expr);
	bprintf("  else {\n");
	bprintf("    strval = (char *)(%s);\n", expr);
	bprintf("    if (!strval) strval = \"<null>\";\n");
	if (in_buf)
		bprintf("    bprintf(\"%%%s\", strval);\n", format ? format : "s");
	else
		bprintf("    say(\"%%s == %%%s\", \"%s\", strval);\n",
				format ? format : "s", quoted);
	bprintf("}\n");
	bprintf("}\n");
}

// Take an expression like 1+2 and evaluate it.  If it's invalid,
// return 0.  Set got_bad_number to indicate if it's valid.
int evaluate_numeric_expression(char *expr)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	save_var bufnum = zap(src);
	got_bad_number = 0;
	ev_header(src);
	bprintf("ev_numexpr(){ return %s; }\n", expr);
	if (!eel_compile(src, 0, "-s -b", "eval-comp"))
		return ev_numexpr();
	got_bad_number = 1;
	return 0;
}

void execute_eel_template(char *expr, char *src)
{
	save_var bufnum = zap(src);
	ev_header(src);
	bprintf("ev_template(){ %s; }\n", expr);
}

inline_eel(char *code)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	ev_header(src);
	buffer_printf(src, "\n%s\n", code);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		error("Could not compile EEL code.");
	}
}

get_tag(res, pr)	/* do completion on tags */
char *res, *pr;		/* returns either tag, or tag <Tab> file */
{			/* or tag <Tab> file <Tab> pos */
	save_var menu_width = screen_cols - _view_left - _view_right;
	if (tag_display_width)
		menu_width = MIN(menu_width, tag_display_width * 2);
	if (!tags_from_bsc)
		save_var completion_lister = tag_lister;
	comp_read(res, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
			  | MUST_MATCH | NONE_OK | CAUTIOUS, "");
}

tag_searcher_c(char *tag)
{
	char classname[FNAMELEN], fulltag[FNAMELEN], pr[FNAMELEN];
	int err, cnt = 0, selected = 0, try = 1;

	if (parse_string(-1, "-%>|%.[ \t\n]"))	// Ignore func's class.
		*classname = 0;
	else
		find_c_func_info(NULL, classname, NULL, 0);
again:
	strcpy(fulltag, tag);
	if (!strchr(tag, ':')) {		// Class not known yet.
withclass:
		if (*classname && strcmp(classname, tag)) {		// See if current class
			sprintf(fulltag, "%s::%s", classname, tag);	// has such a member.
			if (!(err = go_tag(fulltag, 1)))
				return err;
			bufnum = window_bufnum;
		}			// Is there more than one function by that name in any class?
		*fulltag = 0;
		bufnum = window_bufnum;
		if (tag_match(tag, STARTMATCH | EXACTONLY)) {
			do {
				cnt++;
			} while (tag_match(tag, EXACTONLY));
			if (cnt != 1) {
				sprintf(pr, "%d matches for %s: ", cnt, tag);
				save_var menu_width = screen_cols - _view_left - _view_right;
				if (tag_display_width)
					menu_width = MIN(menu_width, tag_display_width * 2);
				if (!tags_from_bsc)
					save_var completion_lister = tag_lister;
				save_var insert_default_response = 2, tag_list_exact_only = 1;
				ungot_key = '?';
				comp_read(fulltag, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
						  | MUST_MATCH | NONE_OK | CAUTIOUS, tag);
				selected = 1;
			} else
				strcpy(fulltag, tag_match(tag, STARTMATCH | EXACTONLY));
		} else
			strcpy(fulltag, tag);
		err = go_tag(fulltag, 1);
	} else {
		err = go_tag(fulltag, 1);
		if (err && *classname)
			goto withclass;
	}
	strcpy(_last_tag, tag);
	bufnum = window_bufnum;
	if (err && !selected && find_tagger(0)) {
		// Try retagging current file before giving up.
		sprintf(pr, "Tag %s has moved, retag file %s? ",
				fulltag, filename);
		if (!tag_ask_before_retagging || try > 1
			  || ask_yn("Retag file", pr, "Yes", "No", 1)) {
			if (try < 2)
				do_retag_file(filename);
			err = go_tag(fulltag, 1);
			if (try < 2)
				do_save_tags();
			if (err && try++ < 2)
				goto again;
			if (!err)
				say("Found %s in this file.", *tag ? tag : "tag");
		}
	}
	if (!err && !selected && strcmp(tag, fulltag))
		say("Searched for %s, found %s", tag, fulltag);
	report_tag_error(err);
}

tag_c_func(show)
{
	int ret = 2;

	if (skip_c_delims('(', ')', 1)) {
		if (show && !in_c_comment(0))
			buffer_printf("-tag-errors", "Warning: couldn't parse %s at line %d\n",
						  filename, lines_between(0, point, 0) + 1);
		return 0;	/* after identifier, skip over arg list */
	}
	c_fwd_space();
	while (tag_c_skip_before_body(show))
		c_fwd_space();
	if (parse_string(1, "on[ \t]")) {	// eel command
		search(1, "\n{");
		point--;
		ret = 2;
	} else if (parse_string(1, "[A-Za-z{]")) // func definition
		ret = skip_param_decls();
	else if (curchar() == ':')		// C++ initializer
		ret = skip_cpp_init();
	else if (curchar() == '=')		// C++ function using =0.
		return 3;
	else if (!parse_string(1, "[,;]"))	// not even a declaration
		ret = (point >= size());	// Skip parens if it looks like a macro.
	else if (!(tag_declarations & 1))	// a declaration, don't tag it?
		ret = 1;
	if (ret)
		tag_modifier_kwd = 0;
	return ret;
}

add_tag(func, pos)
char *func;
{
	char rel[FNAMELEN], *file = filename, tdir[FNAMELEN];

	if (tag_relative) {		/* use relative version instead */
		put_tag_directory(tdir);
		relative(file, rel, tdir);
		file = rel;
	}
	if (tag_show_percent == 2)
		noteput("%r: adding %s...", file, func);
	else if (tag_show_percent == 1 && time_done(&tag_show_msg)) {
		noteput("Working %2d%%...", muldiv(point, 100, size()));
		time_begin(&tag_show_msg, 50);
	}
	if (tag_by_text) {
		int start;

		save_var point;
		point = pos;
		to_begin_line();
		start = point;
		to_end_line();
		buf_printf(_tags_buffer, "%s\t%s\t%d\t",
			func, file, pos);
		buf_xfer(_tags_buffer, start, point);
		buf_printf(_tags_buffer, "\n");
	} else
		buf_printf(_tags_buffer, "%s\t%s\t%d\n", func, file, pos);
}

// Does this region have unmatched { with text after it?  Return pos if
// so.
tcl_unclosed(int start, int end)
{
	save_var narrow_start = start;
	save_var point = end;
	int level = 0;
	while (re_search(-1, "[{}]"))
		if ((level += (curchar() == '{') ? -1 : 1) < 0)
			return parse_string(1, "{[ \t]*(#|$)") ? -1 : point;
	return -1;
}

when_loading()
{
	process_completion_style |= 2;	// Apply new defaults.
	session_restore_max_files = MAX(session_restore_max_files, 25);
	session_restore_biggest_file = MAX(session_restore_biggest_file, 20_000_000);
	int tmp = tmp_buf();		// Try to modify part of recognize_password_pattern
	on_exit buf_delete(tmp);	// (it may be customized).
	save_var bufnum = tmp;
	stuff(recognize_password_pattern);
	simple_re_replace(1, "%.%*%@%.%* ", ".*@.* |%[.*%]");
	grab(0, size(), recognize_password_pattern);
}
