/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2012 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.12 to 13.13.

#include "eel.h"
#include "colcode.h"
#include "c.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "info.h"
#include "kill.h"
#include "search.h"
#include "session.h"
#include "proc.h"

// At point there's a < character that may mark the start of a template.
// Move to just after the > that ends it and return 1.  Return 0 if it
// doesn't seem to be a template, without moving.
int find_template_end()
{
	int opos = point, level = 1, parens = 0, bad = 0;

	if (curchar() != '<')
		return 0;
	point++;
	for (;; point++) {
		if (parse_string(1, "[ \n\t,a-zA-Z0-9_:*=]+|//.*"))
			point = matchend - 1;
		else
			switch (curchar()) {
				case '/':
					if (character(point + 1) != '*')
						bad = 1;
					point += 2;
					if (!search(1, "*/"))
						bad = 1;
					break;
				case '<':
					if (parens == 0)
						level++;
					break;
				case '>':
					if (--level <= 0) {
						point++;
						return 1;
					}
					break;
				case '(':
					parens++;
					break;
				case ')':
					if (--parens <= 0)
						bad = 1;
					break;
				default:
					bad = 1;
					break;
			}
		if (bad) {
			point = opos;
			return 0;
		}
	}
}

// Are we in a struct or union definition?
c_in_struct()
{
	char type[FNAMELEN];
	find_c_func_info(type, NULL, NULL, 0);
	return (!strcmp(type, "struct") || !strcmp(type, "union"));
}

// Return nonzero at the definition of a eember of a JavaScript object
// literal.  It is if there's a label-like thing next, and it follows a
// comma or ={ sequence.
is_js_object_literal(int p)
{
	if (!(c_extra_keywords & (JS_KEYWORDS)) || !have_colon())
		return 0;
	save_var point = p;
	if (!parse_string(1, "[a-zA-Z0-9_]+[ \t\f]*:[ \t\f]*![^:]"))
		return 0;
	cind_try_space();
	switch (character(point - 1)) {
		case ',':
			return 1;
		case '{':
			point--;
			cind_try_space();
			if (character(point - 1) == '=')
				return 1;
			break;
	}
	return 0;
}

parse_label()	/* parse for a case:, default: or label construct at point */
{		/* return nonzero if found, setting matchend to end */
	if (!have_colon())
		return 0;
	if (parse_string(1, "case[^a-zA-Z0-9_\n]([^:\n]*|.*(':'|::).*)"
					 ":[ \t\f]*![^:]"))
		return 1;
	if (parse_string(1, "default:"))
		return 1;
	if (!(c_extra_keywords & (JS_KEYWORDS))
		  && parse_string(1, "[a-zA-Z0-9_]+[ \t\f]*:[ \t\f]*![^:]")) {
		save_var matchend;		// A simple "word:" could be a bitfield.
		if (!is_case(point) && c_in_struct())
			return 0;	// A bitfield if inside a struct, not a known keyword.
		return 1;
	}		// Otherwise it's a label.
	return 0;
}

to_statement_start()
{
	int lastcont, count = 100;	/* give up after this many lines */
	int orig = point;

	to_begin_line();
	if (parse_string(1, "[ \t\f]*}")) {
		search(1, "}");
		c_rev_skip_level('{', '}');
	}
	if (!is_js_object_literal(point)) {
		do {
			to_begin_line();
			lastcont = point;
			if (parse_string(1, "[ \t\f]*},")) {	// initializer
				search(1, "}");
				c_rev_skip_level('{', '}');
			}
		} while (!cind_try_space() && is_contin_statement(point, 0, orig)
				 && count-- > 0);
		point = lastcont;
	}
	to_indentation();
	if (curchar() == '#')
		past_initial_preprocessor();
}

char *clipboard_buf_name();

save_to_clipboard()	// try to save current kill buffer on clipboard
{
	int strip = clipboard_convert_newlines();

	clipboard_translation_type = translation_type;
	if (should_use_clipboard() && clipboard_available()) {
		buffer_to_clipboard(name_to_bufnum(_cur_kill_buf), strip,
							clipboard_format);
		clipboard_to_buffer(zap(clipboard_buf_name()), strip,
							clipboard_format);
	}
}

get_clipboard()		// push clipboard text on kill ring
{
	int new, old;

	if (!should_use_clipboard() || !clipboard_available())
		return;
	if (buffer_flags(0) & 1)	// Turn off converting if orig buffer's
		save_var clipboard_convert_unicode = 0;	// already Unicode.
	new = zap("-new-clipboard-");
	int trans = (GET_LINE_TRANSLATE(clipboard_translation_type) != FILETYPE_BINARY
				 && opsys != OS_UNIX) ? CLIP_CONVERT_NEWLINES : 0;
	if (!clipboard_to_buffer(new, trans, clipboard_format))
		return;
	get_clipboard_fix_lines(new);
	old = create(clipboard_buf_name());
	if (!buffers_identical(old, new)) {
		save_var bufnum = new;
		check_num_kill_buffers();	// copy clipboard to kill ring
		push_kill();
		_kill_width[_num_kill] = REGION;
		xfer(_cur_kill_buf, 0, size());
		buf_zap(old);		// copy new text to old clipboard buf
		buf_xfer(old, 0, size());
	}
	buf_zap(new);
}

paren_follows() // is there whitespace followed by a "(" at point in buffer?
{		// if so, we assume it's a function name
	save_var point;
	for (;; point++) {
		switch (curchar()) {
			case ' ': case '\t': case '\n':
				continue;
			case '(':		return 1;
			case '<':		if (find_template_end()) {
								point--;
								break;
							}
			default:		return 0;
		}
	}
}

// List matching buffer names, showing file names too.
buf_lister(char *s, char *(*finder)(), int flags, int mbuf)
{
	char *res;
	int f = STARTMATCH | LISTMATCH, len = 0;

	UNREFERENCED_PARAMETER(flags);
	if (preserve_filename_case & 4)
		f |= FM_FOLD;
	save_var bufnum;
	buf_zap(mbuf);
	for (; res = finder(s, f); f &= ~STARTMATCH) {
		bufname = res;
		buf_printf(mbuf, "%s\t%s\n", res, filename);
		if (strlen(res) > len)
			len = strlen(res);
	}
	bufnum = mbuf;
	tab_size = MAX(len + 2, 15);
	do_sort_region(0, size(), 0, 0);
	point = 0;
	delete_final_newlines();
}

do_shell_command(char *file, char *dir)
{
	char cmd[FNAMELEN], cmdline[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN], *s;
	int sync = 1;

	sprintf(pr, "! on %s: ", get_tail(file, 0));
	get_string(cmd, pr);
	if (strlen(cmd) > 1 && cmd[strlen(cmd) - 1] == '&') {
		cmd[strlen(cmd) - 1] = 0;
		sync = 0;
	}
	sprintf(buf, "shell-output-%s-%s", cmd, get_tail(file, 0));
	if (!strchr(cmd, '*'))
		strcat(cmd, " *");
	strcpy(cmdline, cmd);
	strcpy(strchr(cmdline, '*'), " ");
	quote_for_shell(cmdline + strlen(cmdline), file);
	strcat(cmdline, strchr(cmd, '*') + 1);
	while (s = strchr(buf, '"'))	// No " in buffer names.
		*s = '\'';
	if (opsys == OS_UNIX || !sync) {
		pipe_text(NULL, buf, cmdline, dir,
				  (sync ? PIPE_SYNCH : 0) | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			point = 0;
		}
	} else
		do_push(cmdline, 1, 1);
}

// Return nonzero if rel (a junction destination name in a dired
// listing) is valid. Windows sometimes makes reparse points that use
// fake names.
int dired_is_valid_link(char *rel)
{
	char name[FNAMELEN], *p;
	dired_try_getdir(name, 1);
	p = name + strlen(name);
	strcpy(p, rel);
	if (is_path_separator(*p) == 1	// If link dest starts with /,
		  || isalpha(p[0]) && p[1] == ':' && is_path_separator(p[2]) == 1)
		strcpy(get_url_file_part(name, 0), p);	// just use that.
	return check_file(name);
}

char *dired_skip_line_pat();

int dired_getname(char *fullname, int flags)
{
	char *p, *q;

	dired_try_getdir(fullname, 1);
	p = fullname + strlen(fullname);
	dired_getindex(p, goal_column);		/* the file name */
	to_begin_line();
	if (parse_string(1, dired_skip_line_pat()))  /* on dir line, make name illegal */
		*p = 0;
	move_to_column(goal_column);
	if (!*p || isspace(*p))
		if (flags & 1)
			error("Can't read file name");
		else
			return 0;
	q = strstr(p, " -> ");			// replace file name with link dest
	if (q)
		if ((flags & 2) || !dired_is_valid_link(q + 4))
			*q = 0;
		else
			strcpy(p, q + 4);
	if (is_path_separator(*p) == 1	// If link dest starts with /,
		  || isalpha(p[0]) && p[1] == ':' && is_path_separator(p[2]) == 1)
		strcpy(get_url_file_part(fullname, 0), p);	// just use that.
	if (dired_format == DF_SIMPLE) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
	} else if (dired_format == DF_VMS) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
		q = strstr(fullname, ".DIR;");	// Refer to directories without
		if (q)							// .DIR extension to see contents.
			*q = 0;
		q = strstr(fullname, ";");		// Remove version number.
		if (q && isdigit(q[1]))
			*q = 0;
	}
	p += strlen(p) - 1;		// Unix file name might use "ls -F" format
	if (*p == '/' || *p == '*')
		*p = 0;
	return 1;
}

int ssh_activity(int activity, int buf, int from, int to);


char *ssh_command_variable(char *dest, char *def, char *pat, struct url_parts *parts)
{
	char s[FNAMELEN];
	int i;

	strcpy(dest, def);
	sprintf(s, pat, opsys == OS_UNIX ? "unix" : "windows");
	i = find_variable_by_host(s, parts->host);
	if (i)
		expand_string_template(dest, get_str_var(i), "", NULL);
}

do_scp_xfer(struct url_parts *parts, struct scp_job *p, char *template,
			int mode, int use_tmp)
{
	char tmpfile[FNAMELEN], cmd[FNAMELEN], *args[5], *s, cmdname[FNAMELEN];

	if (use_tmp) {
		new_file_write("", translation_type, NULL, FILE_IO_TEMPFILE, mode);
		strcpy(p->file, get_tempfile_name());
		p->flags |= SCP_TMP_FILE;
	}
	if (!(p->flags & (SCP_SCP_XFER | SCP_CLIENT_MODE))
		  && !strchr(p->file, '%')) {
		scp_run_helper(parts, p);
		p->flags |= SCP_IS_XFER;
		sprintf(cmd, (p->op == FTP_SEND) ? "put \"%s\" \"%%s\""
										 : "get \"%%s\" \"%s\"", p->file);
		if (is_win32)
			for (s = cmd; s = strchr(s, '\\'); )
				*s = '/';	// Windows sftp gets confused by \'s sometimes.
		scp_send_some_command(parts, p, cmd);
		return;
	}
	strcpy(tmpfile, get_tail(p->file, 1));
	args[0] = parts->usr;
	args[1] = parts->host;
	args[2] = parts->fname;
	args[3] = tmpfile;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command", parts);
	args[4] = cmdname;
	if (expand_string_template(cmd, template, "uhflc", args))
		error("Invalid scp template.");
	scp_start_process(p, cmd);
}

char *scp_variable(struct url_parts *parts, char *varname);

scp_run_helper(struct url_parts *parts, struct scp_job *p)
{
	char cmd[FNAMELEN], *args[3], cmdname[FNAMELEN];

	args[0] = parts->usr;
	args[1] = parts->host;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command", parts);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(parts, "run-helper"),
							   "uhc", args))
		error("Invalid run helper template.");
	if (!is_process_buffer(p->buf)) {
		if (is_win32) {
			char *oldval = getenv("CYGWIN");
			if (oldval) {
				char *newval = malloc(strlen(oldval) + 50);
				sprintf(newval, "CYGWIN=%s nodosfilewarning", oldval);
				putenv(strkeep(newval));
				free(newval);
			} else
				putenv("CYGWIN=nodosfilewarning");
		}
		scp_start_process(p, cmd);
		if (p->homedir)
			free(p->homedir);
		p->homedir = 0;
		p->busy = 2;
	}
}

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = is_unix ? 6 : 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s", &parts);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
	char url[FNAMELEN];
	sprintf(url, "ssh://%s", host);
	if (parts.have_usr)
		sprintf(url, "ssh://%s@%s", username, host);
	buffer_url = strkeep(url);
}

// Check for file vars in the current buffer, then apply to target
// buffer.
check_buffer_filevars(int target)
{
	char buf[FNAMELEN], var[FNAMELEN], pat[2048], *list, *s;
	int i, cur = bufnum;

	save_var bufnum;
	if (look_for_filevar_pattern("tab-(size|width)", "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			tab_size = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(margin-right|fill-column)",
								 "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			margin_right = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(auto-)?fill-mode",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		call_maybe_break_line(i);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(indent-with-tabs|indent-tabs-mode)",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		indent_with_tabs = numtoi(buf);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(write-line-translate)",
								 "([0-9]|dos|windows|binary|unix|mac|auto)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		set_line_translate_from_text(buf);
		bufnum = cur;
	}
	list = "c-indent|soft-tab-size|over-mode|auto-indent"
		   "|goal-column|auto-fill-indents|virtual-space"
		   "|want-backups"
		   "|case-fold|sort-case-fold|delete-hacking-tabs"
		   "|undo-size|indents-separate-paragraphs"
		   "|tex-paragraphs|comment-column|want-warn"
		   "|auto-read-changed-file|perl-indent"
		   "|html-paragraph-is-container|tex-force-latex"
		   "|concurrent-compile|vbasic-indent";
	sprintf(pat, "(%s)", list);
	if (look_for_filevar_pattern(pat, "([0-9]+|nil)", NULL)) {
		for (;;) {
			strncpy(var, list, (sizeof(var) / sizeof(*var)) - 1);
			if (s = strchr(var, '|'))
				*s = 0;
			if (look_for_filevar_pattern(var, "([0-9]+|nil)", NULL)) {
				grab(find_group(1, 1), find_group(1, 0), buf);
				bufnum = target;
				if (i = find_index(var))
					do_setting(i, buf, 0);
				else
					say("Unknown mode variable %s.", var);
				bufnum = cur;
			}
			if (s = strchr(list, '|'))
				list = s + 1;
			else
				break;
		}
	}
}

// Look for an .epsilon_vars file with settings for many files and apply
// its settings to the current buffer.
apply_vars_file()
{
	char varsfile[FNAMELEN], bname[FNAMELEN], pat[FNAMELEN];
	char *file = filename, *mode = major_mode, *s;
	int orig = bufnum;

	if (!stricmp(get_tail(filename, 1), ".epsilon_vars")
		  || !get_buffer_directory(varsfile) || is_remote_file(filename))
		return;
	strcat(varsfile, ".epsilon_vars");
	if (check_file(varsfile) != CHECK_FILE)
		return;
	save_var bufnum;
	if (!look_file(varsfile)) {	// Load the .epsilon_vars file we found.
		make_bname(varsfile, bname);
		bufnum = create(bname);
		find_in_this_buf(varsfile, get_default_translation_type(varsfile));
	}
	int tmp = tmp_buf();	// Make a buffer with just the matching settings.
	buf_printf(tmp, "Local Variables:\n");
	on_exit buf_delete(tmp);
	save_var bufnum;
	save_var point = 0, case_fold = 0, abort_searching = ABORT_ERROR;
	while (re_search(1, "^(Extensions|Modes|Filenames): *([^ \n]*)") == 1) {
		grab(find_group(2, 1), find_group(2, 0), pat);
		switch (toupper(character(matchstart))) {
			case 'E':	s = get_extension(file); break;
			case 'M':	s = mode; break;
			default:	s = get_tail(file, 1); break;
		}
		if (fpatmatch(s, pat, 0, 0)) {
			int start = give_end_line();
			if (re_search(1, "^(Extensions|Modes|Filenames): *"))
				to_begin_line();
			buf_xfer(tmp, start, point);
			buf_printf(tmp, "\n");
		}
	}
	bufnum = tmp;	// Now apply settings from tmp to orig buffer.
	point = 0;
	simple_re_replace(1, "^#.*", "");
	check_buffer_filevars(orig);
	if (!retain_forced_mode)
		check_filevar_mode(orig);
}

get_foldfile()
{
	if (opsys == OS_UNIX && is_unix_flavor != IS_UNIX_MACOS && is_unix_flavor != IS_UNIX_CYGWIN)
		return 0;
	return COMP_FOLD;
}

dodesc(char *str, char *also, int flags)
{
	char *old = bufname, msg[80];
	int start, i, browser_ok = !(flags & 2);

	if (!(flags & 4))
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
			case HELP_FORMAT_HTML:
				show_help_for(str);
				return;
		}
	if (!get_doc(0))
		return;
	i = find_index(str);
	if (!i || name_help(i) < 0) {
		gripe("No information on %s", str, flags & 1);
		bufname = old;
		return;
	}
	point = name_help(i);
	while (isspace(curchar()))
		point++;
	start = point;
	re_search(1, "^!~");
	xfer(HELPBUF, start, point);
	bufname = HELPBUF;
	if (also && *also)
		point = 0, stuff(also);
	insert('\n');
	point = size();
	if (is_unix != IS_UNIX_XWIN && !is_gui && !is_win32 || (flags & 1))
		browser_ok = 0;	// No h key unless we keep control after showing.
	if (browser_ok)
		bprintf("Press H to see this topic in a web browser, with\n"
				"links to the rest of Epsilon's online manual.\n");
	point = 0;
	bufname = old;
	sprintf(msg, " Help on %s ", str);
	view_help(msg, flags & 1);
	if (browser_ok && tolower(key) == 'h') {
		ungot_key = -1;
		html_help_string(str);
		drop_pending_says();
	}
}

color_html_move_back_helper()
{
	int start = point, p, q, from, to, res;

	save_var narrow_start = point - color_html_look_back;

	if (modified_buffer_region(&from, &to, "html-move")) {
		if (get_tagged_region("html-move", from, &p, &q) == 1)
			from = p;
		if (get_tagged_region("html-move", to, &p, &q) == 1)
			to = q;
		set_tagged_region("html-move", from, to, -1);
	}
	reset_modified_buffer_region("html-move");
	if (get_tagged_region("html-move", point, &p, &q) == 1)
		point = p;

	res = re_search(-1, "%<%!%[CDATA%[|%]%]%>");
	p = MIN(point + 10, start);
	if (res)
		p = matchstart;
	set_tagged_region("html-move", p, start, 1);
	if (res && sublanguage_code != SUBLANG_XML
		  && !parse_string(-1, "%<script[^<]*%><wspace>*"))
		if (curchar() == '<')
			return 1;
		else {
			point += parse_string(1, "%]%]%>");
			return -1;
		}
	point = start;
	if (sublanguage_code != SUBLANG_XML
		  && re_search(-1, "<Langle>[ \t\n]*/?((cf)?script|style)")
		  && parse_string(1, "<Langle>[ \t\n]*((cf)?script|style)"))
		return 2;				// Found a <script>, not a </script>.
	point = start;
	if (sublanguage_code != SUBLANG_XML
		  && re_search(-1, "<Langle>[%?]|[%?]<Rangle>") && curchar() == '<'
		  && get_tagged_region("color-as-unit", point) == -1)
		return 3;				// Found a <% or <?, not a %> or ?>.
	point = start;
	do {			// start coloring from the previous tag
		if (!search(-1, "<"))
			break;	// but ignore <'s before " or ' characters
		if (get_character_color(point, &p, NULL) == color_class html_comment) {
			point = p + 2;	// If we're in a comment,
			search(-1, "<!");	// go back to its start.
		}
	} while (html_in_quoted_attribute());
	return 0;
}

color_html_loop(int end)
{
	int s, t, i, script;

	for (s = point; point < end; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							 ? "(<:|_|p:ID_Start><:|hyphen|.|p:ID_Continue>*|#[0-9]+|#x[0-9a-f]+);?"
							 : "([a-z0-9]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()) && color_html_include(end))
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = 0;
			if (sublanguage_code != SUBLANG_XML) {
				if (s == 6 && parse_string(1, "script"))
					script = 1;
				if (s == 3 && parse_string(1, "jsp:(declaration|expression"
										   "|scriptlet)</word>"))
					script = 2;
				if (s == 5 && parse_string(1, "style"))
					script = 3;
				if (s == 8 && parse_string(1, "cfscript"))
					script = 4;
			}
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			for (i = point; strchr(" \t\n", curchar()); )
				point++;	// Parse whitespace.
			set_character_color(i, point, color_class html_punctuation);
			if (parse_string(1, "/[ \t\n]*>")) {
				set_character_color(point, point + 1, color_class html_tag);
				set_character_color(point + 1, matchend,
									color_class html_punctuation);
				point = matchend;
			} else
				color_html_attributes(script, t, end);
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
}

// Parse an info node name and return a pointer to the ')' at the end of
// the filename portion, or null if none.  If it doesn't start with '('
// assume we're just past it. File names may contain balanced parens.
char *info_filename_end(char *s)
{
	char *t;
	int cnt = 1;
	if (*s == '(')
		s++;
	while (t = strpbrk(s + 1, "()")) {
		if (*t == '(')
			cnt++;
		else if (--cnt <= 0)
			return t;
		s = t;
	}
	return 0;
}

// Get the file name for the current node.
info_get_file(char *file)
{
	char *p;

	if (*filename == '(') {
		strcpy(file, filename + 1);
		if (p = info_filename_end(file)) {
			*p = 0;
			return;
		}
	}
	*file = 0;
}

int info_color_node(int start, int end);

info_go_node(char *node, int flags)
{
	char file[FNAMELEN], *p, nodename[FNAMELEN], thisfile[FNAMELEN];
	int ret = 1;

	if (flags & 1)
		info_record();
	save_var abort_searching = 0;
	info_standardize(node);
	*file = 0;
	if (*node == '(' && (p = info_filename_end(node))) {
		strcpy(file, node + 1);		// Node includes a file name.
		file[p - node - 1] = 0;
		node = p + 1;
	} else
		p = node;
	if (!*file)		// Assume it's to the same file as the current one.
		info_get_file(file);
	if (!*file)		// No current file, so use this default one.
		strcpy(file, "dir");
	if (!*node)		// No node within this file, use default.
		node = "Top";
	to_info_buffer();
	info_get_file(thisfile);	// Get file name for current node.
	if (strcmp(file, thisfile))	// Doesn't match, get to right file.
		if (!load_info_file(file, (flags & 2) ? 1 : 0) && (flags & 2))
			return 0;
	if (!info_to_local_node(node)) {	// Get to right node.
		say("Info file %s contains no node %s.", file, node);
		ret = 0;
		node = "Top";
		if (!info_to_local_node("Top")) {
			strcpy(file, "dir");
			load_info_file(file, 0);
			info_to_local_node("Top");
		}
	}
	sprintf(nodename, "(%s)%s", file, node);
	filename = nodename;		// Record this node name.
	make_mode();
	recolor_range = info_color_node;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	coloring_flags |= COLOR_RETAIN_NARROWING;
	when_setting_want_code_coloring();
	idle_coloring_delay = -1;		// No idle coloring.
	return ret;
}

command info_up() on info_tab['u']
{
	char orignode[FNAMELEN], pat[FNAMELEN * 3];

	to_info_buffer();
	char *s = info_filename_end(filename);
	strcpy(orignode, s ? s + 1 : "");
	if (!info_get_link("Up"))
		return;
	if (*orignode) {
		point = 0;
		if (re_search(1, "^%* Menu:.*\n")) {
			sprintf(pat, "^%%* ((%%([^\t\n,()]+%%))?%q+::"
					"|.*: *(%%([^\t\n,()]+%%))?%q[\t\n,.])",
					orignode, orignode);
			if (re_search(1, pat)) {
				to_begin_line();
				return;
			}
		}
	}
	info_restore_point();	// If we've been to this node before, to same place within.
}

int info_search_continuation(int code, int flags, char *str)
{
	char disp[FNAMELEN], *s;
	int i, res = 0, orig = narrow_start;

	save_var abort_searching = 0;
	switch (code) {
		case SCON_RECORD:				/* remember this message number */
			sprintf(str, "%d", narrow_start);
			return 0;
		case SCON_RESTORE:				/* return to remembered message */
			info_node_by_start(numtoi(str));
			return 0;
		case SCON_COMPARE:				/* is remembered node < or == or >? */
			i = narrow_start - numtoi(str);
			return (i < 0) ? -1 : (i > 0) ? 1 : 0;
	}
	narrow_start = narrow_end = 0;
	point += (flags & REVERSE) ? 1 : -1;
	if (re_search((flags & REVERSE) ? -1 : 1, "\n" INFO_CSTR "\f?\n"))
		res = 1;
	if (!info_node_by_start(point)) {
		res = 0;
		info_node_by_start(orig);
	}
	point = (flags & REVERSE) ? size() : 0;
	if (code % 40 == 0) {	// Display progress.
		make_search_prompt(disp, flags);
		s = info_filename_end(filename);
		note("%sSearch in node \"%s\" for %s...",
			 disp, s ? (s + 1) : filename, str);
	}
	return res;
}

int do_info_index(char *item, int record)
{
	char node[FNAMELEN], pat[FNAMELEN];
	char title[FNAMELEN], file[FNAMELEN];
	int cnt = 0;

	info_to_index();
	if (!*item)		// Just go there.
		return -1;
	save_var abort_searching = ABORT_ERROR;
	zap(INFO_INDEX);	// Look for menu items containing this pat.
	sprintf(pat, "^%%* ([^\n:]*%q[^\n:]*):([^.\n]+)%%.[ \t]*[0-9]*$", item);
	while (info_is_index() && !user_abort) {
		if (re_search(1, "\n%* Menu:")) {
			file[0] = '(';
			info_get_file(file + 1);	// Get file name for current node.
			strcat(file, ")");
			while (re_search(1, pat) == 1) {
				grab(find_group(1, 1), find_group(1, 0), title);
									// Get the separate node name.
				grab(find_group(2, 1), find_group(2, 0), node);
				info_standardize(node);	// Make list of choices.
				cnt++;
				buffer_printf(INFO_INDEX, "%s\t%s%s\t%s\t%s\n",
							  title, strchr(node, '(') ? "" : file,
							  node, info_filename_end(filename) + 1, item);
			}
		}
		info_give_linked_node("Next", node);
		if (!*node)
			break;
		info_go_node(node, 0);
	}
	info_return(0);
	if (cnt) {
		info_prefer_matches(item, "^%s.*\n", 0);
		info_prefer_matches(item, "^%s[^-_a-z0-9].*\n", 0);
		info_prefer_matches(item, "^%s[ \t].*\n", 0);
		info_prefer_matches(item, "^%s[ \t].*Concept Index.*\n", 1);
		do_info_index_next(record);
	}
	return cnt;
}

// Look in node for the first of several patterns that indicate s
// is being defined there.  Return 1 if we found one, else 0.
info_locate_match(char *s)
{
	char pat[FNAMELEN], pat2[FNAMELEN], *p;

	point = 0;
	sprintf(pat, "[a-z]+: %q!( |$)", s);
	if (re_search(1, pat))		// sometext: ourmatch
		return 1;
	point = 0;
	sprintf(pat, "`%q'", s);	// `ourmatch'
	if (re_search(1, pat))
		return 1;
	point = 0;
	sprintf(pat, "^%q$", s);	// ourmatch on its own line
	if (re_search(1, pat))
		return 1;
	point = 0;
	strcpy(pat, s);			// If we're searching for ourmatch (sometext),
	p = strstr(pat, " (");		// try looking for ourmatch.
	if (p && info_filename_end(p)) {
		*p = 0;
		sprintf(pat2, "`%q'", s);	// `ourmatch'
		if (re_search(1, pat2))
			return 1;
		point = 0;
		sprintf(pat2, "^%q$", pat);	// ourmatch on its own line
		if (re_search(1, pat2))
			return 1;
		point = 0;
		if (search(1, pat))
			return 1;
	}
	point = 0;
	if (search(1, s))		// Look for the plain text.
		return 1;
	point = 0;
	return 0;
}

// Return 0 if we can't because program's missing.
int set_font_for_unix()
{
	char buf[FNAMELEN];

	get_executable_file(buf, "epsilon-dialogs", 1);
	sprintf(buf + strlen(buf), " --parent %x --font \"%s\"",
			screen_to_window_id(0), font_fixed);
	if (!pipe_text("-epsilon-dialog-out", "-epsilon-dialog-out", buf, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF, 0))
		file_error(errno, "Couldn't run dialog helper", "Error");
	save_var bufname = "-epsilon-dialog-out";
	point = 0;
	if (parse_string(1, ".*(ibrary|shared|X Error)")) {
		dialog_helper_failed(1);
		return simple_set_font_for_unix();
	} else if (re_search(1, "^=") && parse_string(1, ".*", buf))
		font_fixed = buf;
	return 1;			// User picked one, even if it was bad.
}

int do_run_viewer(char *fname, char *op, char *dirname)
{
	char orig_dir[FNAMELEN];
	int ret;

	if (opsys != OS_UNIX || is_unix_flavor == IS_UNIX_CYGWIN)
		return run_viewer(fname, op, dirname);
	getcd(orig_dir);
	if (dirname && *dirname)
		chdir(dirname);
	if (!stricmp(op, "open"))
		op = "";	// Use the default operation for this.
	ret = run_helper("epsilon-viewer", fname, op);
	chdir(orig_dir);
	return ret;
}

int grab_search_word(flags, origwin, line)
{
	char text[FNAMELEN];
	int orig, len = size(), ret = 0;

	save_var window_handle = origwin;
			// First time, if we're inside a word, go to its start.
	if (!len && !line && !(flags & FAILING) && parse_string(1, word_pattern)
		&& parse_string(-1, word_pattern)) {
		point -= parse_string(-1, word_pattern);
		ret = 1;
	}
	if (!(flags & REVERSE))
		save_var point;
	else if (len && !(flags & FAILING))
		point = matchstart;
	orig = point;
	if (!(flags & FAILING))
		if (line) {
			if (curchar() == '\n')
				point++;
			to_end_line();
		} else
			re_search(1, word_pattern);		// Get the next word.
	int max = orig + sizeof(text) / sizeof(*text) - 1;
	if (point >= max)
		point = max;
	grab(orig, point, text);
	restore_vars();
	if (search_defaults_from && shift_selecting == 2 && is_highlight_on()) {
		delete(0, size());
		shift_selecting = 0;
	}
	point = size();
	bprintf((flags & REGEX) ? "%q" : "%s", text);
	return ret;
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end)) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command(help_on_search(flags));
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		} else if (IS_MOUSE_SINGLE(key)) {
			ret = 1;
			ungot_key = full_key;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
							show = 1;
							goto ask;
						}
						if (!IS_ENTER_KEY(key))
							ungot_key = full_key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(str, with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
				case ',': case '^':	skip = 1; goto ask;
				case '.':		ret = 1; goto done;
				case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
				  replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

int session_restore_one_file(int b, char *file, char *sdir, int *readerrs)
{
	int i, trans = FILETYPE_AUTO;
	int num1[10], num1cnt, num2[10], num2cnt, num3[10], num3cnt;
	char firstline[150];
	jmp_buf this_level;

	grab_line(b, firstline);
	if ((num1cnt = break_into_numbers(firstline, num1)) < 3)
		return 2;
	if ((num2cnt = grab_numbers(b, num2)) < 4)
		return 3;
	if ((num3cnt = grab_numbers(b, num3)) < 5)
		return 4;
	if (session_restore_files)
		filename = file;
	save_var top_level = &this_level;
	if (setjmp(top_level))
		return -1;
	if (num1cnt >= 4 && num1[3] && session_restore_files) {	// restore dired buf
		filename = "";
		if (num1cnt >= 5)
			save_var file_pattern_wildcards = num1[4];
		save_var show_status = 0;
		if (num1cnt >= 6)
			dired_sorts_files = num1[5];
		create_dired_listing(file, 0);
		if (is_remote_file(get_dired_buffer_pattern()))
			dired_orig_position = -1;
	} else if (*file && session_restore_files) {
		if (!is_remote_file(file))
			absolute(file, sdir);
		if (has_feature & FEAT_UNICODE) {
			if (num1cnt >= 7) {		 // If explicit trans, force.
				char *p;
				if (p = strchr(firstline, '@')) {
					i = encoding_from_name(p + 1);
					if (i)
						trans = MAKE_TRANSLATE(i, num1[7], 1);
				}
			}
		} else {
			if (num3cnt > 8 && num3[7])
				trans = num1[2];	// If explicit trans, or common unicode, force.
		}
		i = do_file_read(file, trans);
		force_save_as = 0;
		if (num1cnt >= 7)
			force_save_as = num1[6];
		if (user_abort || i == EREADABORT)
			return 0;
		if (!i || i == 2) {
			if (!i)
				set_readonly_from_file();
			window_bufnum = bufnum;		// make suffix- kbd macros work
			call_mode(filename);
			make_mode();
			check_filevars();
		} else if (!*readerrs++) {
			file_error(i, file, "read error");
			maybe_ding(bell_on_read_error);
		} else
			show_text(0, -1, "%d files could not be read", *readerrs);
	}
	if (session_restore_files) {
		if (bufnum != b)	/* same tmp name as session buffer */
			point = num1[0];
		mark = num1[1];
		if (is_dired_buf())
			move_to_column(goal_column);
	}
	if (session_restore_files) {
		session_restore_one_file_1(num2, num2cnt);
		session_restore_one_file_2(num3, num3cnt);
	}
	return -1;
}

do_uniq(incl_uniq, incl_dups, talk)
{
	int old = narrow_start, ocnt, ncnt;

	if (talk)
		ocnt = count_lines_in_buf(bufnum, 0);
	save_spot point;
	for (;;) {
		point = old;
		if (!nl_forward())
			break;
		if (uniq_count_dups(old, point) ? incl_dups : incl_uniq)
			old = point;
		else
			delete(old, point);
		if (talk && size())
			noteput("%d%% complete...", muldiv(point - narrow_start, 100,
											   size() - narrow_start - narrow_end));
		check_abort();
	}
	if (!incl_uniq)
		delete(old, point);
	if (talk) {
		ncnt = count_lines_in_buf(bufnum, 0);
		int sz = size() - narrow_end;
		say("Deleted %d lines leaving %d.", ocnt - ncnt,
			  ncnt + (sz && character(sz - 1) != '\n'));
	}
}

when_loading()
{
	if (!strstr(grep_ignore_file_extensions, "|.ses|"))	// Default changed.
		strcat(grep_ignore_file_extensions, ".ses|");
}
