/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2017 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.13 to 13.14.

#include "eel.h"
#include "colcode.h"
#include "c.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "info.h"
#include "kill.h"
#include "search.h"
#include "session.h"
#include "proc.h"
#include "diff.h"
#include "tags.h"
#include "gams.h"
#include "hex.h"
#include "indent.h"
#include "perl.h"
#include "python.h"
#include "tcl.h"
#include "tex.h"

/* Delete trailing whitespace, then break line here and indent it. */
break_and_indent_line()
{
	delete(point, point - parse_string(-1, "[ \t]*"));
	insert('\n');
	do_reindent_previous_line();
	save_var indenter_action = IACT_AUTO_INDENT;
	point += parse_string(1, "[ \t]*");
	(*indenter)();
}

void charname_lister(char *s, char *(*finder)(), int flags, int mbuf);
char *charname_matcher(char *s, int flags);

command insert_ascii()
{
	int res;
	char buf[FNAMELEN];

	if (run_by_mouse)
		has_arg = 0;
	save_var completion_lister = charname_lister;
	save_var menu_width = MAX(menu_width, 70);
	comp_read(buf, "Insert character with code: ",
			  charname_matcher, COMP_FOLD | SPACE_VALID, "");
	res = exptoi(buf);
	if (got_bad_number || res < 0 || res > MAX_CHAR)
		error("Invalid code %s.", buf);
	insert(res);
}

list_a_bookmark(list, i)	/* maybe add a line to list for mark i */
{				/* return 1 if we did, else 0 */
	int b, opt;
	spot m = _bookmark_spots[i];

	if (!m || (b = spot_to_buffer(m)) <= 0)
		return 0;
	bufnum = b;
	buf_printf(list, "%c %s ", bookmark_to_key(i), bufname);
	opt = point;
	if (_bookmark_saved_pos[i] && !*m) {
		*m = _bookmark_saved_pos[i];
		_bookmark_saved_pos[i] = 0;
	}
	point = *m;
	to_begin_line();
	point += parse_string(1, "<hspace>*");
	buf_printf(list, "%d: ", buf_position_to_line_number(b, point));
	recolor_buffer_range(point, give_end_line());
	buf_xfer_colors(list, point, give_end_line());
	point = opt;
	buf_printf(list, "%p\n", -1);
	return 1;
}

c_indenter()		/* replace surrounding indentation with new */
{
	if (c_extra_keywords & IDL_KEYWORDS)
		return;
	if (in_shell_buffer)
		prev_indenter();
	else if (!(in_c_comment(c_look_back) & IN_COMMENT))
		indent_to_column(c_compute_indent());
	else if ((in_perl_buffer ? reindent_perl_comments : reindent_c_comments)
			 & (indenter_action == IACT_REINDENT_PREV ? 4 :
				(this_cmd == CMD_INDENT_REG) ? 16 : 1))
		indent_c_comment();
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN * 4], classname[FNAMELEN * 4], pat[FNAMELEN];
	int start, funcstart, j, got_class = 0;

	start_of_function = -1;
	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	if (c_extra_keywords & OBJC_KEYWORDS)
		got_class |= find_objc_class(type, class);
	to_begin_line();
	start = point;
	point--;		// Begin searching at the last } char on the left margin...
	while (search(-1, "\n}")) {
		point++;		// ...that's not in a comment.
		if (!in_c_comment(c_look_back))
			break;
		point--;
	}

	strcpy(pat, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(");
	if (c_extra_keywords & OBJC_KEYWORDS)
		strcat(pat, "|\n[-+]");
	while (re_search(1, pat)) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*(define|if|pragma).*(\\\n.*)*")) {
			to_end_line();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				j = point;
				if (c_inside_block(point, start)) {
					got_class |= find_c_class_info(start, type, class, func);
					point = j + 1;
					if (got_class)
						continue;
				}
				point = j;
				c_skip_level('{', '}');
				break;
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					to_end_line();	/* c++ comment */
				break;
			case '\n':		// Objective-C method
				point++;
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				if (find_objc_method(func) == 1)
					return CF_INFO_FUNC | got_class;
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return got_class;
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + c_biggest_declarator);
				if (tag_c_func(0) == 2 && !find_c_func_oneliner(start)) {
					point = funcstart;	// Checks out as a function def.
					start_of_function = give_begin_line();
					*classname = 0;
					if (parse_string(-1, (c_extra_keywords == CSHARP_KEYWORDS)
									 ? "[ \t\n]*%.[ \t\n]*"
									 : "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						parse_string(-1, "[a-zA-Z0-9_]+", classname);
					}
					if (func)
						strcpy(func, funcname);
					if (class && *classname)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0)
							| got_class;
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return got_class;
}

suffix_as()
{
	suffix_js();
}

suffix_ts()
{
	suffix_js();
	compile_buffer_cmd = compile_typescript_cmd;	// can compile this
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */
	int in_initializer;

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!(reindent_c_comments & ((this_cmd == CMD_INDENT_REG) ? 16 : 1))
		  && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if ((!reindent_one_line_c_comments || 
		 (indenter_action == IACT_REINDENT_PREV) && !(reindent_c_comments & 8))
		  && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' && reindent_c_preprocessor_lines == 1
		  || first_char == '\f'
		  || first_char == '@' && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(1, "@(" OBJC_TOPBLOCKS ")[ \t]+<word>|@end"))
		return 0;
	else if (first_char == '}')
		return cind_close();
	else if (have_c_close_equivalent(orig))
		return indent_at_c_close_equivalent(orig);

	to_begin_line();
	if (parse_string(-1, PREPROC_ELSE ".*\n")) {
		point = matchend;
		locate_by_ifdef(-1);
	}
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig) || c_align_break_with_case
		  && matches_at(orig, 1, "[ \t]*break;")) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig) && !((c_extra_keywords & OBJC_KEYWORDS)
									 && matches_at(pstart, 1, "^[-+@]"))) {
		if (c_in_struct())		// Could it be a bitfield, not a label?
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';' || c_semi_equivalent(pend, orig)) {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (c_close_equivalent(pend)) {
		point = pend;
		ind = indent_after_c_close_equivalent(orig);
	} else if (prev_end == '{' || c_open_equivalent(pend, orig)) {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_initializer = in_c_initializer(orig)) {
		return ind;
	} else if (c_next_no_indent(orig, pstart)) {
		return 0;
	} else if (c_next_same_indent(orig, pstart)) {
		return ind;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, in_initializer);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

in_c_comment(int lookback)
{
	UNREFERENCED_PARAMETER(lookback);
	if (in_perl_buffer)
		return in_perl_comment(point);
	int pos = point, from, to, res;
	save_var narrow_start = 0, narrow_end = 0;
	if (modified_buffer_region(&from, &to, "c-lex")) {
		// First, discard region modified since last scan.
		from = c_lex_safe_start(from);
		set_tagged_region("c-lex", from, size(), -1);	// Set to unknown.
	}
	reset_modified_buffer_region("c-lex");

	res = get_tagged_region("c-lex", pos, &from, &to);
	if (res == -1)		// No cached data, so lex it.
		c_lex_region(from, pos);
	res = get_tagged_region("c-lex", pos, &from, &to);
	if (pos >= size())		// Special rules for EOF, as we don't cache that.
		res = c_lex_eof_rules(res, from, to);
	if (res == -1) {
		say("Internal error in in_c_comment: %s at %d", bufname, pos);
		return 0;
	}
	return res;
}

indent_c_comment()		/* indenter when inside a comment */
{
	int prev_indent;
	int orig, has_star, no_star = 0;

	has_star = parse_string(1, "[ \t\f]*<*>");
	orig = point;
	re_search(-1, "[^ \t\f\n]");	/* find previous non-blank line */
	to_indentation();
	if (has_star && parse_string(1, "[/ \t\f]*<*>"))
		point = matchend - 1;	// line up star under prev star
	else {
		no_star = 1;
		point += parse_string(1, "[ \t\f]*/<*>[ \t\f]*!<^wspace>");
	}
	prev_indent = current_column();
	point = orig;
	if (no_star && parse_string(1, "<*>/"))
		prev_indent = 0;
	to_column(prev_indent);		/* indentation as previous */
}

/* Indent continuation line. */
indent_c_contin(ind, orig, lastl, in_initializer)
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0, slevel = 0;
	int objc_sel = 0, objc_indent = -1;

 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	if (!c_align_contin_lines)	// never line up
		if (!in_initializer)
			return c_default_indent(ind, orig, 1);
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	if ((c_extra_keywords & OBJC_KEYWORDS) && c_align_selectors
		  && matches_at(give_begin_line(), 1, "^[ \t]*(<word>+):"))
		objc_sel = find_group(1, 0) - find_group(1, 1);
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[][()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0 || objc_indent >= 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (objc_indent >= 0)
					break;
				if (findc != ')')
					point++;
				else if (Closeback & 2)
					return get_indentation(point);
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind, orig);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				if (objc_sel && !plevel && !slevel
					  && parse_string(-1, "(^|<word>|[]])[ \t]*<word>+")
					  && current_column() <= c_align_selectors)
					objc_indent = current_column() - objc_sel;
				else
					qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				if (objc_indent >= 0)
					return objc_indent;
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '[':
				if (slevel-- > 0 || objc_indent >= 0)
					break;
				if (findc != ']')
					point++;
				save_var c_align_extra_space = 0;
				return limit_c_contin_indent(ind, orig);
			case ']':
				slevel++;
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)) || objc_indent >= 0)
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind, orig);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (objc_indent >= 0)
		return objc_indent;
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		if (curchar() == '#')
			past_initial_preprocessor();
		return limit_c_contin_indent(ind, orig);
	}
	return c_default_indent(ind, orig, 2);
}

char *clipboard_buf_name();

save_to_clipboard()	// try to save current kill buffer on clipboard
{
	int strip = clipboard_convert_newlines();

	clipboard_translation_type = GET_LINE_TRANSLATE(translation_type) | FORCED_TRANS;
	if (should_use_clipboard() && clipboard_available()) {
		buffer_to_clipboard(name_to_bufnum(_cur_kill_buf), strip,
							clipboard_format);
		clipboard_to_buffer(zap(clipboard_buf_name()), strip,
							clipboard_format);
	}
}

get_clipboard()		// push clipboard text on kill ring
{
	int new, old;

	if (!should_use_clipboard() || !clipboard_available())
		return;
	if (buffer_flags(0) & 1)	// Turn off converting if orig buffer's
		save_var clipboard_convert_unicode = 0;	// already Unicode.
	new = zap("-new-clipboard-");
	int trans = GET_LINE_TRANSLATE(translation_type);
	if (clipboard_translation_type & FORCED_TRANS)
		trans = GET_LINE_TRANSLATE(clipboard_translation_type);
	int conv = (trans != FILETYPE_BINARY && opsys != OS_UNIX) ? CLIP_CONVERT_NEWLINES : 0;
	if (!clipboard_to_buffer(new, conv, clipboard_format))
		return;
	get_clipboard_fix_lines(new);
	old = create(clipboard_buf_name());
	if (!buffers_identical(old, new)) {
		save_var bufnum = new;
		check_num_kill_buffers();	// copy clipboard to kill ring
		push_kill();
		_kill_width[_num_kill] = REGION;
		xfer(_cur_kill_buf, 0, size());
		buf_zap(old);		// copy new text to old clipboard buf
		buf_xfer(old, 0, size());
	}
	buf_zap(new);
}

early_restoring()
{
	char fname[FNAMELEN];
	int i;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_DOS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
	if (load_from_state)
		show_text(-2, 0, "Loaded from state file %s", state_file);
	else if (*state_file)
		show_text(-2, 0, "Loaded from state file %s (quickup)", state_file);
	else
		show_text(-2, 0, "Loaded from bytecode files");
	for (i = 0; original_argv(i); i++)
		show_text(-2, 0, "Argv[%d]=%s", i, original_argv(i));
	get_customization_directory(fname);
	show_text(-2, 0, "Customization directory is %s", fname);
	char *path = verenv("EPSPATH");
	show_text(-2, 0, "EPSPATH is %s", path ? path : "undefined");
}

usage()
{
	char name[FNAMELEN], *p;

	bufnum = zap("usage");
	bprintf("\n");
	strcpy(name, get_tail(argv[0], 0));
	*get_extension(name) = 0;
	if (opsys != OS_UNIX)
		for (p = name; *p; p++)
			*p = tolower(*p);
	bprintf("Usage: %s [flags...] [files...]\n", name);
	bprintf("A file name may have +linenum or +linenum:column before it.\n\n");

	bprintf("Flags:\n");
	bprintf("-sFILE   Load state from FILE.sta (default: epsilon.sta).\n");
	bprintf("-bFILE   Build new system from FILE.b, not .sta file "
			"(default: epsilon.b).\n");
	bprintf("-quickup Load quickup.b after loading state."
			"    -nologo Skip initial msg.\n");
	bprintf("-noinit  Skip loading einit.ecm.\n\n");

	bprintf("-lFILE   Load FILE.b after loading files. ");
	bprintf("-rCMD    Run command or macro CMD.\n");
	bprintf("-teach   Run the tutorial.                ");
	bprintf("-dVAR!VAL Set variable to value.\n");
	bprintf("-add     Send files to prior instance.    ");
	bprintf("-noserver Don't accept sent files.\n");
	bprintf("-wait    Send files, then wait.           ");
	bprintf("-server SVR Set the server name for -add.\n");
	bprintf("-dir DIR  Interp. relative paths on cmd line.\n");
	bprintf("-p[VAL]  Set/toggle session-preserving.   ");
	bprintf("-pFILE    Load session from FILE.\n");
	bprintf("-wNUM    Set various directory control flags.\n");

	usage_new_page();
	bprintf("\n");
	bprintf("-fsDIRS  Use DIRS for temp space.         ");
	if (!(has_feature & FEAT_WINHELP))
		bprintf("-fdFILE  Documentation in file FILE.");
	bprintf("\n");

	bprintf("-vmono   Use monochrome color scheme.     ");
	bprintf("-vlNUM   Screen has NUM lines.\n");
	bprintf("-vcolor  Use colors.                      ");
	bprintf("-vcNUM   Screen has NUM columns.\n");
	bprintf("-vv      Put cmd line files side by side.\n\n");
	if (opsys == OS_UNIX) {
		bprintf("-vt      Don't use X11.                   ");
		bprintf("-bw PIX Window border width.\n");
		bprintf("-display DISP  Use this X11 display.            ");
		bprintf("-fn FONT Use this X11 font.\n");
		bprintf("-geometry GEOM  X11 window geometry spec.       ");
		bprintf("-name NAME Look up X11 resouces under NAME.\n");
		bprintf("-title TITLE  Set X11 window title.       ");
		bprintf("-xrm RES Specify an X11 resource name/val.\n");
	}
	bprintf("-kaNUM   Compatibility flags.\n");
	if (opsys == OS_DOS && !is_gui || opsys == OS_OS2) {
		bprintf("-kcNUM   Mouse graphic cursor/block/off.  ");
		bprintf("-kmNUM   Mouse relative/absolute/off.\n");
	}
	bprintf("-ksNUM   Speed (>100) vs. abort response.\n");
	if (opsys == OS_DOS && !is_gui || opsys == OS_OS2)
		usage_new_page();

	bprintf("\n");
	bprintf("-mNUM    Use NUM bytes of main memory.    ");
	bprintf("-m0      Use minimum memory.\n");
	bprintf("-mNUM    Use NUM kbytes (if NUM < 1000)   ");
	bprintf("-m       Use as needed.\n\n");
	if (opsys == OS_DOS && !is_gui && is_win32 != IS_WIN32_CONSOLE) {
		bprintf("-xNUM    Like -m; sets max XMS (extended) ");
		bprintf("-eNUM    Sets maximum EMS (expanded).\n");
		bprintf("-xiNUM   Initial allocation of XMS.       ");
		bprintf("-eiNUM   Initial allocation of EMS.\n");
		bprintf("-xfNUM   Allocation for EEL funcs in XMS. ");
		bprintf("-efNUM   Alloc. for EEL funcs in EMS.\n");
		bprintf("-xuNUM   Maximum use of upper memory blocks.\n\n");
	}
	bprintf("  NUMk means kilobytes, NUMm means megabytes.\n");
	bprintf("  Without suffix, NUM is in kilobytes if < 1000, "
			"otherwise bytes.\n");
	restore_vars();
	init_abort();
}

void set_color_handler(int activity, int buf, int from, int to);

set_color_xwin()
{
	int i, j, classes = number_of_color_classes();
	int *orig_fg, *orig_bg, *orig_schemes;
	int orig_scheme = selected_color_scheme, this_scheme = orig_scheme;
	char *s, cmd[FNAMELEN];
	int orig_class = get_character_syntax_color(point);

	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	on_exit free(orig_schemes), free(orig_fg), free(orig_bg);
	save_var bufnum;
	prepare_dialog_helper_cmdline(cmd);
	sprintf(cmd + strlen(cmd), " --color");

	for (;;) {
		bufnum = zap("-epsilon-dialog-data");
		bprintf("%d\n%s\n", font_styles, font_fixed);
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i))
				bprintf("%s\n", name_name(i));
		bprintf("\n");
		for (i = 0; i < classes; i++)
			bprintf("%s\n", name_color_class(i));
		bprintf("\n");
		bprintf("%s\n", (s = lookpath("colclass.txt")) ? s : "");
		bprintf("%s\n", name_name(this_scheme));

		selected_color_scheme = find_index("color-defaults");
		for (j = 0; j < classes; j++)
			bprintf("default %x %x\n", get_foreground_color(j, 1),
					get_background_color(j, 1));
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i)) {
				selected_color_scheme = i;
				for (j = 0; j < classes; j++)
					bprintf("%d %d %x %x\n", i, j, get_foreground_color(j, 1),
							get_background_color(j, 1));
			}
		make_color_class_list(1);
		point = size();
		s = name_color_class(orig_class);
		bprintf("\n%s\n", s ? s : "");
		selected_color_scheme = this_scheme;
		if (!pipe_text("-epsilon-dialog-data", "-epsilon-dialog-out", cmd,
					   NULL, PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH,
					   set_color_handler))
			file_error(errno, "Couldn't run dialog helper", "Error");
		this_scheme = selected_color_scheme;
		bufname = "-epsilon-dialog-out";
		i = process_exit_status;
		if (matches_at(0, 1, ".*(ibrary|shared|error)")) {
			dialog_helper_failed(0);
			return;
		}
		if (i < 0 || i > 2)
			i = !ask_yn("Keep changes", "The dialog process exited "
						"unexpectedly.  Keep these settings? ",
						"Yes", "No", 0);
		if (i == 0) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			full_redraw = 1;
			return;
		} else if (i == 1) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			full_redraw = 1;
			say("");
			return;
		} else {
			show_binding("Continue changing colors "
						 "by typing %s", "exit-level");
			full_redraw = 1;
			recursive_edit();
		}
	}
}

// If we find color class name, expand it and return 1.  Call this
// repeatedly to expand each level and reposition.
int set_color_search_expand(char *name)
{
	char pat[FNAMELEN];
	if (!name || !*name)
		return 0;
	sprintf(pat, "(^|\t)[-+ ]*%q($|\t)", name);
	point = 0;
	if (!re_search(1, pat))
		return 0;
	set_color_expand(1);
	return 1;
}	

comp_read(trystr, prmpt, finder, flags, def)
char *trystr, *prmpt;
char *(*finder)();
char *def;		/* default value */
{
	int oldbuf, editbuf, ocmd_len = cmd_len, olen_def_mac = _len_def_mac;
	int skipped_redisp = char_avail();
	char *s;

	save_var postpone_asynchronous_notify = 1;
	error_if_input(1);
	*trystr = 0;
	if (run_by_mouse && ungot_key == -1 && finder
		&& ((flags & MUST_MATCH) || show_mouse_choices))
		ungot_key = '?';
	else if (insert_default_response)
		strcpy(trystr, def);
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var return_raw_buttons = 0;
	save_var permit_window_keys |= PERMIT_WHEEL_KEY;
	if ((flags & COMP_FILE) && !completion_literal_char)
		save_var completion_literal_char = literal_if_quoted;
	_grab_word_loc = point;
	_grab_word_start = 0;
	oldbuf = bufnum;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	editbuf = bufnum = tmp_buf();
	if (flags & COMP_FILE)
		double_quotes(trystr);
	stuff(trystr);
	full_comp_read(trystr, prmpt, finder, flags, editbuf, def);
	bufnum = oldbuf;
	buf_delete(editbuf);
	if (_len_def_mac >= 0)	// now treat as one big cmd
		cmd_len = ocmd_len + _len_def_mac - olen_def_mac;
	check_abort();
	hide_auto_show_delimiter();
	if (!skipped_redisp)		// skipped before, so skip now
		maybe_refresh();
	if (completion_column_marker) {
		s = strchr(trystr, completion_column_marker);
		if (s)
			*s = 0;
	}
	if (!(flags & PASSWORD_PROMPT))
		save_completion(trystr, finder);
	if (flags & COMP_FILE)
		drop_quotes(trystr);
	say("");
}

char *general_matcher(s, flags)
char *s;
{
	char tmp[TAGLEN];

	save_var bufnum;
	save_var abort_searching = 0;
	bufname = _MATCH_BUF;
	save_var case_fold = (flags & COMP_FOLD) != 0;
	if (flags & STARTMATCH)
		point = 0;
	if (*s) {
		sprintf(tmp, "^%q", s);
		if (!re_search(1, tmp))
			return 0;
		to_begin_line();
	} else {
		if (curchar() == '\n')
			point++;
		if (point == size())
			return 0;
	}
	parse_string(1, ".+", _this_match);
	to_end_line();
	if ((flags & EXACTONLY) && strfcmp(_this_match, s))
		return 0;
	return _this_match;
}

drop_quotes(char *s)
{
	was_quoted = 0;
	while (s = index(s, '"')) {
		was_quoted = 1;
		strcpy(s, s + 1);
		if (*s)
			s++;
	}
}

// Replace each " character with two of them.
double_quotes(char *s)
{
	char tmp[FNAMELEN * 2], *t;
	if (strlen(s) + 1 >= ptrlen(tmp))
		return;
	strcpy(tmp, s);
	for (t = tmp; *t; s++, t++) {
		*s = *t;
		if (*t == '"')
			*++s = '"';
	}
	*s = 0;
}

command abort()
{
	highlight_off();
	error("Cancelled.");
}

get_binding(int k, short *tab)
{
	return ((k & CMD_INDEX_KEY) ? (k & ~CMD_INDEX_KEY)
								: ((tab && tab[k]) ? tab[k] : root_keys[k]));
}

check_running_processes()
{
	char *buf, *open = 0, *s, msg[FNAMELEN];

	buf = buffer_list(1);
	save_var bufnum;
	do {
		if (process_warn_on_exit && preserve_buffer_type(buf))
			if (process_warn_on_exit == 2 || !exit_this_process(buf)) {
				open = buf;
				break;
			}
		bufname = buf;
		if (is_process_buffer(bufnum) && discardable_buffer)
			exit_this_process(buf);
	} while (buf = buffer_list(0));
	restore_vars();
	if (open && process_warn_on_exit) {
		tiled_only();
		locate_window(open, "");
		refresh();
		s = preserve_buffer_type(open);
		if (windows_state() & WINSTATE_MINIMIZED)
			windows_restore();
		if (process_warn_on_exit != 2) {
			if (s && !strcmp(s, "Process"))
				error("Couldn't exit from concurrent process.");
			else
				error("Couldn't exit %s job.", s);
		} else {
			if (s && !strcmp(s, "Process"))
				sprintf(msg, "A process is running, confirm exit? ");
			else
				sprintf(msg, "%s job is running, confirm exit? ", s);
			if (!ask_yn("Confirm exit", msg, "Yes", "No", 2))
				error("Exiting cancelled.");
		}
	}
}

do_mode_default_settings(int retain_colors)
{
	indenter = indenter.default;
	if (!indenter)
		indenter = prev_indenter;
	auto_indent = auto_indent.default;
	buffer_maybe_break_line = 0;
	strcpy(comment_start, comment_start.default);
	strcpy(comment_pattern, comment_pattern.default);
	strcpy(comment_begin, comment_begin.default);
	strcpy(comment_end, comment_end.default);
	mode_keys = buf_tab;		/* use these keys */
	mode_move_level = 0;
	major_mode = strkeep("Fundamental");
	auto_show_matching_characters = fundamental_auto_show_delim_chars;
	mode_auto_show_delimiter = 0;
	when_displaying = default_when_displaying;
	recolor_range = recolor_from_here = 0;
	search_continuation = 0;
	if (!retain_colors) {
		drop_coloring(bufnum);
		coloring_flags &= ~(COLOR_MINIMAL | COLOR_INVALIDATE_FORWARD
							| COLOR_INVALIDATE_BACKWARD | COLOR_INVALIDATE_RESETS
							| COLOR_RETAIN_NARROWING);
	}
	make_mode();
}

do_diff(char *buf, int style)
{
	int start, other, osize, opt, other_pt;

	strcpy(_diff_buffer, buf);
	zap(_diff_buffer);
	drop_coloring(name_to_bufnum(_diff_buffer));
	if (style == VISUAL_DIFF_STYLE)
		buffer_printf(_diff_buffer, "%pComparing %p%s%p\n",
					  -1, color_class diff_removed,
					  (*filename && !force_save_as) ? filename : bufname, -1);
	highlight_off();
	opt = point;
	next_user_window(1);
	other = bufnum;
	osize = size();
	other_pt = point;
	if (style == VISUAL_DIFF_STYLE)
		buffer_printf(_diff_buffer, "with %p%s%p\n", color_class diff_added,
					  (*filename && !force_save_as) ? filename : bufname, -1);
	next_user_window(-1);
	if ((has_feature & FEAT_LCS_PRIMITIVES)
		  && size() <= diff_precise_limit && osize <= diff_precise_limit) {
		diff_via_lcs(bufnum, other, style);
		return;
	}

	diff_check_malloc();	/* see if we need to make room for data */
	_diff_line1 = _diff_line2 = 1;
	set_points(0, 0);
	do {
		start = point;
		diff_same();	// Found identical lines.
		note_same(bufnum, start, point, style);
	} while (resync_lines(style));
	set_points(opt, other_pt);
	say("Done.");
}

int dired_when_idle(int cnt);

dired_live_link() on dired_tab['l']
{
	iter = 0;
	add_buffer_when_idle(bufnum, dired_when_idle);	// Ensure it's set up.
	if (is_buffer_in_window(LIVE_LINK_BUF) == -1) {
		save_var window_handle;
		if (number_of_user_windows() == 1)
			window_split(has_arg ? VERTICAL : HORIZONTAL);
		else
			to_next_user_window(1);
		to_buffer(LIVE_LINK_BUF);
	}
}

// Insert size in each line, return length.
dired_sort_by_size_unix()
{
	char line[FNAMELEN * 2], *p;

	point = 0;
	while (re_search(1, UNIX_DATE)) {
		point = matchstart;
		if (!parse_string(-1, "[0-9]+[MKG]? *", line)) {
			note("Unrecognized format, can't sort.");
			strcpy(line, "0");
		}
		if (p = strchr(line, 'G'))
			strcpy(p, "000000000");
		if (p = strchr(line, 'M'))
			strcpy(p, "000000");
		if (p = strchr(line, 'K'))
			strcpy(p, "000");
		move_to_column(goal_column);
		bprintf("%20s", line);
		to_end_line();
	}
	return 20;
}

file_convert_write(char *fname, int trans, struct file_info *finfo)
{
	int tmp, err, mod, code;
	char *(*func)();

	code = file_write_newfile ? FILE_IO_NEWFILE : FILE_IO_ATSTART;
	if (is_remote_file(fname))
		return save_remote_file(fname, trans, finfo);
	if (!file_io_converter)
		return new_file_write(fname, trans, finfo, code, -1);
	if (!file_io_converter(FILE_CONVERT_WRITE | FILE_CONVERT_ASK))
		error("Cancelled.");
	func = file_io_converter;
	tmp = tmp_buf();
	buf_xfer(tmp, 0, size());
	copy_buffer_variables(tmp, bufnum);
	save_var bufnum = tmp;
	translation_type = trans;	// func might modify this
	func(FILE_CONVERT_WRITE);
	err = new_file_write(fname, translation_type, finfo, code, -1);
	mod = modified;
	restore_vars();
	modified = mod;
	return err;
}

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1,
			".* (not a directory|No such|not found)"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & (SCP_CHECK_IS_DIR
											  | SCP_CHECK_MATCHES)))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is| is now) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connect(ing|ed) to "
								 "|Connection closed|Listing directory "
								 "|Using username \"|sftp%>"
								 "|(.* |^)pass(word|phrase).*:)")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg && !scp_echoed_cmdline(msg, p->cmdlist))
				say("%s", msg);		// Any other line is an error msg.
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system"
								  "|Connection timed out|Could not)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

// Our winpty helper comes in both 32-bit and 64-bit versions.
// Determine which type of Cygwin install is on the path, if any, and
// what %w should expand to.
get_winpty(char *winpty, int turn_off_echo)
{
	char exename[FNAMELEN];

	*winpty = 0;
	if (opsys != OS_WINDOWS)
		return;
	if (!cygwin_type) {
		char *buf = temp_buf();
		shell("", "uname -m", buf, SHELL_HIDE);		// Get text with Cygwin install type.
		save_var bufname = buf;
		on_exit delete_buffer(buf);
		point = 0;
		cygwin_type = search(1, "i.86") ? 1 : search(-1, "_64") ? 2 : 3;
		show_text(-2, 0, "Cygwin type detected: %d", cygwin_type);
	}
	if (cygwin_type == 3)
		return;
	get_executable_directory(exename);
	convert_to_8_3_filename(exename);
	sprintf(winpty, "\"%swinpty%s\" -a%s",
			exename,
			(cygwin_type == 1) ? "32" : "64",
			turn_off_echo ? " -n" : "");
}

char *ssh_command_variable(char *dest, char *def, char *pat, struct url_parts *parts, int turn_off_echo)
{
	char s[FNAMELEN], winpty[FNAMELEN], *vals[1];
	int i;
	get_winpty(winpty, turn_off_echo);
	vals[0] = winpty;
	strcpy(dest, def);
	sprintf(s, pat, opsys == OS_UNIX ? "unix" : "windows");
	i = find_variable_by_host(s, parts->host);
	if (i)
		expand_string_template(dest, get_str_var(i), "w", vals);
}

do_scp_xfer(struct url_parts *parts, struct scp_job *p, char *template,
			int mode, int use_tmp)
{
	char tmpfile[FNAMELEN], cmd[FNAMELEN], *args[5], *s, cmdname[FNAMELEN];

	if (use_tmp) {
		new_file_write("", translation_type, NULL, FILE_IO_TEMPFILE, mode);
		strcpy(p->file, get_tempfile_name());
		p->flags |= SCP_TMP_FILE;
	}
	if (!(p->flags & (SCP_SCP_XFER | SCP_CLIENT_MODE))
		  && !strchr(p->file, '%')) {
		scp_run_helper(parts, p);
		p->flags |= SCP_IS_XFER;
		sprintf(cmd, (p->op == FTP_SEND) ? "put \"%s\" \"%%s\""
										 : "get \"%%s\" \"%s\"", p->file);
		if (is_win32)
			for (s = cmd; s = strchr(s, '\\'); )
				*s = '/';	// Windows sftp gets confused by \'s sometimes.
		scp_send_some_command(parts, p, cmd);
		return;
	}
	strcpy(tmpfile, get_tail(p->file, 1));
	args[0] = parts->usr;
	args[1] = parts->host;
	args[2] = parts->fname;
	args[3] = tmpfile;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command", parts, 1);
	args[4] = cmdname;
	if (expand_string_template(cmd, template, "uhflc", args))
		error("Invalid scp template.");
	scp_start_process(p, cmd);
}

char *scp_variable(struct url_parts *parts, char *varname);

scp_run_helper(struct url_parts *parts, struct scp_job *p)
{
	char cmd[FNAMELEN], *args[3], cmdname[FNAMELEN];

	args[0] = parts->usr;
	args[1] = parts->host;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command", parts, 1);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(parts, "run-helper"),
							   "uhc", args))
		error("Invalid run helper template.");
	if (!is_process_buffer(p->buf)) {
		if (is_win32) {
			char *oldval = getenv("CYGWIN");
			if (!oldval)
				putenv("CYGWIN=nodosfilewarning");
			else if (!strstr(oldval, "nodosfilewarning")) {
				char *newval = malloc(strlen(oldval) + 50);
				sprintf(newval, "CYGWIN=%s nodosfilewarning", oldval);
				putenv(strkeep(newval));
				free(newval);
			}
		}
		scp_start_process(p, cmd);
		if (p->homedir)
			free(p->homedir);
		p->homedir = 0;
		p->busy = 2;
	}
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			point = type_point;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 0x100)
				perform_unicode_conversion(bufnum, from, size(), CONV_TO_16, "UTF-8-No-BOM");
			if (telnet_interpret_output & 0x20) {
				point = from;
				simple_re_replace(0, "<#0-#1>", "");
			}
			if (telnet_interpret_output & 0x40)
				remove_xterm_sequences(from);
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			if (telnet_interpret_output & 0x80)
				process_backspace_sequences(from - 100);
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

void ssh_activity(int activity, int buf, int from, int to);
char *scp_variable(struct url_parts *parts, char *varname);

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = is_unix ? 6 : 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s", &parts, 0);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
	char url[FNAMELEN];
	sprintf(url, "ssh://%s", host);
	if (parts.have_usr)
		sprintf(url, "ssh://%s@%s", username, host);
	buffer_url = strkeep(url);
}

show_date_warning(temp, orig, then, now, save)
struct file_info *now, *then;
{
	char msg[FNAMELEN], *file;
	int mod, warn, auto;

	save_var bufnum = orig;
	file = filename;
	mod = modified;
	warn = want_warn;
	auto = auto_read_changed_file;
	bufnum = temp;
	buf_zap(temp);
	bprintf("%s\nhas changed since it was read from disk", file);
	bprintf(mod ? "!\n" : ".\nFortunately, you haven't edited it.\n");

	stuff_file_date("\nThe version being edited was dated %s on disk.\n",
		then);
	stuff_file_date("         The copy on disk is dated %s.\n\n", now);
	current_time_and_date(msg);
	bprintf("                         It is now %s.\n\n", msg);

	bprintf("Options:\n");
	bprintf("D) Display a diff listing comparing the two versions.\n");
	bprintf("C) Compare the two versions by reading the disk version ");
	bprintf("in its own buffer.\n");
	bprintf("   You may then use compare-windows or diff to ");
	bprintf("see the changes.\n");
	bprintf("R) Read in the new version now, replacing the ");
	bprintf("version already read in.\n");
	if (save)
		bprintf("S) Save your changes, replacing ");
	else
		bprintf("I) Ignore ");
	bprintf("the version on disk.\n\n");
	bprintf("!) %s this warning for this buffer.  ",
		warn ? "Disable" : "Re-enable");
	bprintf("(You may then choose\n");
	bprintf("   another option.)\n");

	bprintf("A) %s auto-read.  Read changed files without prompting for this buffer,\n"
			"   if it's safe. ", auto ? "Disable" : "Enable");
	bprintf("(You may then choose another option.)\n");

	bprintf("\nYour choice? ");
	point = 0;
}

ask_read_changed_file()
{
	char ques[200];
	int res;

	sprintf(ques, "%s for this buffer. %s auto-reading for new buffers too? ",
			auto_read_changed_file ? "Disabling" : "Enabling",
			auto_read_changed_file ? "Disable" : "Enable");
	if (!(is_gui && want_gui_prompts))
		res = ask_yn("Set default", ques, "Yes", "No", 2) ? 1 : 2;
	else
		res = button_dialog("Set default", ques, "Yes", "No", "Cancel", 2);
	if (!user_abort) {
		if (res == 1 || res == 2)
			auto_read_changed_file = !auto_read_changed_file;
		if (res == 1)
			auto_read_changed_file.default = auto_read_changed_file;
	}
	user_abort = 0;
}

/* Ask for a file name and save buffer to that file if one is given. */
command write_file() on cx_tab[CTRL('W')]
{
	char tmp[FNAMELEN], tmp2[FNAMELEN];
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;

	iter = 0;
	if (use_common_file_dialog()) {
		strcpy(tmp, force_save_as == FSA_NEWFILE ? "" : filename);
		if (!use_common_open_file_dlg(tmp, "Save file as", &flags, 1))
			return 2;
	} else
		get_file_dir(tmp, "Write file: ");
	if (*tmp && is_directory(tmp) && *filename) {
		make_copy_name(filename, tmp, tmp2);
		strcpy(tmp, tmp2);
	}
	if (*tmp) {
		if (get_file_read_only(tmp) == 1)
			error("File %s is read-only.", get_tail(tmp, 0));
		switch (warn_existing_file(tmp)) {
			case 1: return 0;
			case -1: save_var file_write_newfile = 1;
		}
		if (has_arg) {
			translation_type = give_line_translate(tmp);
			make_mode();
		}
		set_buffer_filename(tmp);
		already_made_backup = 0;
		modified = 1;
		return do_save_file(1, 0, 1);
	}
	say("Buffer %s not saved, no file name.", bufname);
	return 2;
}

do_print_region(b, all, silent)
{
	int buf, err, skipped_lines, cnt, cols, width = 80;
	char to[FNAMELEN], title[FNAMELEN], *dest;

	iter = 0;
	dest = is_unix ? print_destination_unix : print_destination;
	if ((has_feature & FEAT_OS_PRINTING) && want_gui_printing) {
		print_options.flags = all ? 0 : PD_SELECTION;
		if (silent)
			print_options.flags |= PD_RETURNDEFAULT;
		if (!select_printer(&print_options))
			return -1;		// User canceled.
		all = !(print_options.flags & PD_SELECTION);
		width = print_options.width;
	} else if (!silent) {
		get_strdef(to, "Print to device (or !command)", dest);
		strcpy(dest, to);
	}
	if (!*filename)
		strcpy(title, bufname);
	else if (print_heading & PRHEAD_PATH) {
		strcpy(title, filename);
		abbreviate_file_name(title, width);
	} else
		strcpy(title, get_tail(filename, 1));
	save_var bufnum = b;
	buf = tmp_buf();
	save_spot point, mark;
	fix_region();		// handle other types of regions
	recolor_buffer_range(0, size());
	skipped_lines = all ? 0 : lines_between(0, point, 0);
	if (all)
		buf_xfer_colors(buf, narrow_position(0), narrow_position(size()));
	else if (region_type() == REGRECT)
		extract_rectangle_colors(buf, 0, 1);
	else
		buf_xfer_colors(buf, point, mark);
	bufnum = buf;
	safe_copy_buffer_variables(buf, b);

	point = 0;
	if (!print_tabs || print_line_numbers)
		tab_convert(0, size(), 0);
	if (print_line_numbers) {
		cnt = lines_between(0, size(), 0) + 1 + skipped_lines;
		for (cols = 0; cnt > 0; cols++)
			cnt /= 10;		// Compute maximum width of line numbers.
		cnt = 1;
		do {
			if (point >= size())
				break;
			bprintf("%*d  ", cols, skipped_lines + cnt++);
			set_character_color(give_begin_line(), point, -1);
		} while (nl_forward());
		point = 0;
	}
	do_print_adjustments();
	if ((has_feature & FEAT_OS_PRINTING) && want_gui_printing) {
		gui_print(title, title);
		err = 0;
	} else
		err = send_to_file_or_cmd(buf, dest, title);
	bufnum = b;
	buf_delete(buf);
	if (!err && !all)
		highlight_off();
	if (!err && !((has_feature & FEAT_OS_PRINTING) && want_gui_printing))
		say("%s printed.", all ? "Buffer" : "Region");
	return err;
}

int ask_confirm_save(char *fmt)
{
	char othername[FNAMELEN], msg[FNAMELEN];

	relative(filename, othername);
	if (!(is_gui && want_gui_prompts))
		abbreviate_file_name(othername, screen_cols - strlen(fmt) - 2);
	sprintf(msg, fmt, othername);
	return ask_yn("Confirm save", msg, "Save", "Cancel", 0);
}

/* Returns 0=ok, 1=err, 2=user canceled. Prints status in echo area (except with
 * backup==2 and save-without-prompt). */
int do_save_file(int backup, int checkdate, int getdate)
{
	struct file_info *finfo;
	int err, old_abort = user_abort;

	iter = 0;
	if (!*filename || force_save_as)
		return write_file();
	if (checkdate && check_dates(1)) {
		user_abort = old_abort;	// Ignore aborting from modified-on-disk prompt.
		return 2;
	}
	if ((buffer_flags(0) & BF_UNICODE) && !GET_ENCODING(translation_type)
		  && (has_feature & FEAT_UNICODE)) {
		if (!buffer_has_pattern("<#256-#65535>"))
			save_var translation_type = SET_ENCODING(translation_type, 1);
		else		// Force to raw temporarily.
			pick_write_encoding();	// Have 16-bit chars, get an encoding for them.
		if (!GET_ENCODING(translation_type))
			error("Invalid encoding.");
	}
	if (_read_aborted)
		if (!ask_confirm_save((_read_aborted == 1)
							  ? "Only part of %s was read from disk.  "
							  "Save anyway? "
							  : "Error when reading %s.  Save anyway? "))
			return 2;
	if (is_retrieving_remote_file(bufnum)
		  && !ask_confirm_save("%s is still being downloaded.  "
							   "Save partial file now? "))
			return 2;
	if (backup)
		do_write_backup();
	prepare_save_file();
	finfo = getdate ? &file_info : 0;
	if (err = file_convert_write(filename, translation_type, finfo)) {
		auto_read_changed_file = 0;
		file_info.check_type = 0;	// Prevent checking date.
		file_error(err, filename, "write error");
		maybe_ding(bell_on_write_error);
	} else {
		if (!save_without_prompt || backup != 2)
			note_written(filename);
		undo_set_flag(UNDO_WRITE_FILE);
		if (finfo)
			update_readonly_warning(finfo);
		maybe_delete_autosave();
	}
	return (err != 0);
}

limit_buffer_size(int max)
{
	if (max > 0) {
		save_var narrow_start = 0, narrow_end = 0;
		save_var point = size() - max;
		to_begin_line();
		delete(0, point);
	}
}

buffer_to_end(char *buf, int limit)
{
	if (exist(buf)) {
		save_var bufname = buf;
		limit_buffer_size(limit);
		point = size();
	}
}

do_save_all_buffers(int newfile_bufs, int *count, int *queued, int *failed)
{
	int i;

	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		if (!*filename || !modified || buffer_not_saveable
			  || is_retrieving_remote_file(i))
			continue;
		if (newfile_bufs ? (force_save_as != FSA_NEWFILE)
						 : (force_save_as == FSA_NEWFILE))
			continue;	// Only do one group.
		if (newfile_bufs) {
			if (!size())		// Nothing to save, so don't bother to prompt.
				continue;
			if (save_all_without_asking || _doing_input) {
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not"
							  " save buffer %s, no file name\n", bufname);
				(*failed)++;
				continue;
			}
			to_buffer_num(i);	// Show the buffer we'll be prompting for.
		}
		switch (do_save_file(1, 1, 1)) {
			case 2:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Cancelled saving buffer %s\n",
							  bufname);
				(*failed)++;
				break;
			case 1:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not save %s\n",
							  filename);
				(*failed)++;
				break;
			case 0:
				show_text(-2, 1, "save-all-buffers: %s %s", 
						  is_remote_file(filename) ? "Queued" : "Saved",
						  filename);
				if (!save_without_prompt)
					if (is_remote_file(filename))
						(*queued)++;
					else
						(*count)++;
				break;
		}
	} while (i = buf_list(1, 1));
}

make_newfile_buf()
{
	char bname[FNAMELEN], fname[FNAMELEN];
	int unique = 2;

	strcpy(bname, "New File");
	for (;;) {
		sprintf(fname, "New%d%s", unique - 1, new_file_ext);
		absolute(fname);
		if (!exist(bname) && !do_file_match(fname, STARTMATCH | EXACTONLY)
			|| unique >= 99)
			break;
		sprintf(bname, "New File %d", unique++);
	}
	to_buffer(bname);
	force_save_as = FSA_NEWFILE;
	filename = fname;
	fundamental_mode();
	try_calling(new_file_mode);
	check_filevars();
}

// Run try_calling() on a subr name derived from a file name, first
// converting any characters invalid in a function name to underscores.
int safe_try_calling(char *func)
{
	char safe[FNAMELEN], *s, *t;
	for (s = func, t = safe; *s; s++, t++)
		*t = (*s < 128 && isident(*s)) ? *s : '_';
	*t = 0;
	return try_calling(safe);
}

call_by_suffix(file, pattern)	// construct a subr name from extension
char *file, *pattern;		// of file name, try to call it
{				// return 1 if we found something to call
	char *s, subr[FNAMELEN], def[FNAMELEN];

	s = get_extension(file);
	sprintf(subr, pattern, *s ? (s + 1) : "none");
	sprintf(def, pattern, "default");
	return safe_try_calling(subr) || safe_try_calling(def);
}

#define C_COMMENT_FILLER		"[^0-9a-zA-Z{}()\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

do_fill_comment(int generic, int margin)
{
	int pbuf, width = 1000, i, old_lines, new_lines, orig_lines;
	int col, col2, orig = bufnum, type;
	char last_prefix[FNAMELEN], orig_prefix[FNAMELEN], pat[FNAMELEN], comm[30];

	save_var margin_right = margin;
	save_spot point;
	type = locate_c_comment();	// 1 if block comment, 2 if one-line
	if (!type)					// comment, or 0 if not a comment.
		error("Can't fill comment here, not in a comment.");
	point += parse_string(1, C_COMMENT_FILLER);	// Past 1st line's prefix.
	sprintf(pat, "[ \t]*%s[^0-9a-z_\n]*", generic ? get_comment_start() : "/[*/]");
	grab(point, point - parse_string(-1, pat), orig_prefix);
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (!generic)
		point -= parse_string(-1, "<*>/");
	save_var narrow_end = size() - point + 1;

	if (type == 1 && !generic)		// Combine adjacent block comments.
		simple_re_replace(1, "[ \t]*<*>/[ \t]*\n?[ \t]*/<*>[ \t]*", " ");
	point = 0;
	while (nl_forward()) {	// Find width of narrowest prefix.
		i = parse_string(1, generic ? get_comment_start() : C_COMMENT_PREFIX);
		if (i > 0 && point > 0 && isalnum(character(point + i))
			  && strchr("<>", character(point + i - 1)))
			i--;		// Don't mess up e.g. <varname> at start of line.
		col = get_column(point + i);
		width = MIN(width, col);
	}
	pbuf = tmp_buf();
	point = 0;
	get_comment_marker(comm, generic);
	sprintf(pat, "[ \t]*%q", comm);
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		if (col > width)	// Make sure it's not too wide.
			col = width;
		if (type == 2) {	// But if it's made of one-line comments,
			i = parse_string(1, pat);
			col2 = get_column(point + i);
			if (col < col2)		// each prefix must include a starting delim like "//".
				col = col2;
		}
		i = column_to_pos(col);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	if (width == 1000)	// No prefix, so use default.
		width = 3;
	orig_lines = old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 20),
						   MAX(margin_right - width, 20));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;
	*last_prefix = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		grab(point, give_end_line(), last_prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	if (orig_lines < 2) {		// Default for new lines if 1 line originally.
		strcpy(last_prefix, orig_prefix);
		if (type == 2 && !strstr(last_prefix, comm))
			sprintf(last_prefix, "%s%s", comm,
					(*comm && comm[strlen(comm) - 1] == ' ') ? "" : " ");
		else if (type == 1 && !generic)
			if (!strstr(last_prefix, "/*"))
				strcpy(last_prefix, (auto_fill_comment_rules & 4) ? " " : " * ");
			else {
				replace_in_string(last_prefix, '/', ' ');
				if (fill_c_comment_plain)
					replace_in_string(last_prefix, '*', ' ');
			}
	}
	bufnum = orig;
	narrow_start = 0;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(last_prefix);			// Add some prefixes.
		if (generic) {
			save_var indenter_action = IACT_FILL_COMMENT;
			to_begin_line();
			if (indenter)
				(*indenter)();
		}
		nl_forward();
	}
	buf_delete(pbuf);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

int locate_c_comment()
{
	int com, type = 1, generic = strcmp(major_mode, "C");

	save_spot narrow_start, narrow_end;
	com = (mode_keys == c_tab) ? in_c_comment(c_look_back) : in_comment();
	if (com == IN_OLD_COMMENT || !com && !generic && parse_string(1, "[ \t]*/<*>"))
		narrow_to_block_comment();
	else if (narrow_to_new_comment())
		type = 2;
	else
		return 0;
	to_begin_line();
	if (parse_string(1, C_COMMENT_PREFIX "$"))
		point--;		// If on line right after a parag, go back.
	to_begin_line();
	if (parse_string(1, C_COMMENT_PREFIX "$")) // If no text on line,
		if (!re_search(-1, "[^ \t\n]"))	// Find a line with text.
			re_search(1, "[^ \t\n]");

	re_search(1, C_COMMENT_PREFIX "$");	// Find "blank" line w/only filler.
	if (parse_string(-1, C_COMMENT_PREFIX "$"))
		to_begin_line();	// And move to its start.
	re_search(-1, "[^ \t\n]");	// Find text of some parag before it.
	nl_forward();
	mark = point;
	point--;
	re_search(-1, C_COMMENT_PREFIX "$");	// Find "blank" line before.
	if (parse_string(1, C_COMMENT_PREFIX "\n"))
		nl_forward();
	return type;
}

int do_maybe_break_line(int c_like)
{
	int comm, res = 1, starts_with_comm, starts_in_comm, after_prefix;
	int comm_start = -1, break_block_comment = 0;
	int generic = strcmp(major_mode, "C");
	char fill[FNAMELEN], pat[FNAMELEN], raw_comm[FNAMELEN];
	spot after, next_ends = 0;

	delete_if_highlighted();

	// Various reasons we might want not want to fill right now.
	if (give_end_line() - give_begin_line() > FNAMELEN - 10)
		return 1;			// The line's too long for us.
	if (IS_ENTER_KEY(key)
		  && (c_like ? in_c_comment(c_look_back) : in_comment()) == IN_OLD_COMMENT
		  && (auto_fill_comment_rules & 1))
		break_block_comment = 1, res = 0;
	else if (current_column() + 1 <= margin_right)
		return 1;		// We're not past the margin.
	comm = c_like ? in_c_comment(c_look_back) : in_comment();
	if (!(comm & IN_COMMENT))
		return 1;			// We're not in a comment.

	if (c_like && !in_perl_buffer)
		generic = 0;
	// Collect info on the situation.
	if (key == ' ') {		// Insert <Space> now, then remember not
		insert(key);		// to later, so we can break here.
		res = 0;
	}
	after = alloc_spot(1);	// Remember where we should end up.
	to_begin_line();	// See if a comment's the first thing on this line.
	get_comment_marker(raw_comm, generic);
	sprintf(pat, "[ \t]*%s", generic ? raw_comm : "/[*/]");
	starts_with_comm = parse_string(1, pat);	// Yes, comment's first on line.
	starts_in_comm = c_like ? in_c_comment(c_look_back) : in_comment();
	// Does this line start with a comment?
	sprintf(pat, "[ \t%s]*", generic ? raw_comm : "*/");
	parse_string(1, pat, fill);	// Repeat this prefix on new lines.
	after_prefix = point;
	if (!starts_in_comm) {		// Find where this line's comment starts.
		to_begin_line();
		sprintf(pat, "<min>.*%s",
				generic ? raw_comm : comm == IN_NEW_COMMENT ? "//" : "/%*");
		point += parse_string(1, pat);
		comm_start = point;
	}
	if (nl_forward()) {		// Remember the end of the next line, unless
		to_end_line();		// there's a comment there.
		if (c_like && !in_c_comment(c_look_back))
			next_ends = alloc_spot(0);
	}

	// Look for a breaking point.
	point = *after;
	if ((c_like && c_auto_fill_mode != 1 || starts_with_comm || starts_in_comm)
		  && (auto_fill_comment_rules & 2) && !break_block_comment
		  && pick_break(margin_right) && point > after_prefix
		  && point > comm_start)
		break_block_comment = 1;

	if (break_block_comment)
		comm_break_line_here(c_like, comm, starts_with_comm, starts_in_comm,
							 comm_start, next_ends, after, fill);

	point = *after;
	free_spot(after);
	if (next_ends)
		free_spot(next_ends);
	return res;
}

// A quoted string starts at position p.  Color it and move past.
gams_parse_and_color_string(int p)
{
	char pat[50];
	int c = character(p);

	point = p + 1;
	sprintf(pat, "[%c\n]", c);
	re_search(1, pat);
	set_character_color(p, point, color_class gams_string);
}

// Return the comment character in use for the given buffer position.
// Inline may be nonzero for inline comment chars (else one-line
// comments).  Return 0 if not set, 2 if set to {{ }}, else char code.
// See gams_have_comment() for supported delimiters.
int gams_get_comment_char(int p, int inline)
{
	char *regname = inline ? "gams-inlinecom" : "gams-eolcom";
	int pos, i = get_tagged_region(regname, p, &pos);
	if (i >= 0)
		return i;
	save_var point = pos;	// Start at unknown section.
	to_begin_line();		// Assume it's the same as prev section, then scan it.
	int prev = (point > 0) ? get_tagged_region(regname, point - 1) : 0;
	set_tagged_region(regname, point, size(), prev);
	while (re_search(1, inline ? "^[$].*inlinecom +[</{]" : "^[$].*eolcom +[!/#?_]")) {
		if (curchar() == '{' && character(point - 1) == '{')	// Use special code for {{ }}.
			set_tagged_region(regname, point, size(), 2);
		else
			set_tagged_region(regname, point, size(), character(point - 1));
	}
	return get_tagged_region(regname, p);
}

int gams_have_comment(int p)
{
	switch (character(p)) {
		case '!':
			if (gams_get_comment_char(point, 0) == '!')
				return 1;
			break;
		case '<':
			if (gams_get_comment_char(point, 1) == '<' && character(p + 1) == '<')
				return 2;
			break;
		case '/':
			if (character(p + 1) == '*' && gams_get_comment_char(point, 1) == '/')
				return 2;
			if (character(p + 1) == '/' && gams_get_comment_char(point, 0) == '/')
				return 1;
			break;
		case '{':
			if (character(p + 1) == '{' && gams_get_comment_char(point, 1) == 2)
				return 2;			// Check for special code 2 for {{ }}.
			if (gams_get_comment_char(point, 1) == '{')
				return 2;
			break;
		case '#':
			if (gams_get_comment_char(point, 0) == '#')
				return 1;
			break;
		case '?':
			if (gams_get_comment_char(point, 0) == '?')
				return 1;
			break;
		case '*':
			if (matches_at(p, 1, "^%*"))
				return 1;
			break;
		case '_':			// Found _$, possible comment.
			if (gams_get_comment_char(point, 0) == '_' && character(p + 1) == '$')
				return 1;
			break;
		case '$':
			if (matches_at(p, 1, "^%$ontext"))
				return 1;
	}
	return 0;
}

int gams_color_common(int c, int p)
{
	switch (c) {
		case '"':		// found a string
		case '\'':
			gams_parse_and_color_string(p);
			return 1;
		case '*':	// Found possible comment.
		case '#':
		case '!':
		case '{':
		case '?':		// Found ??.
		case '<':		// Found <<.
		case '/':		// Found /* or //
		case '_':		// Found _$
			if (gams_have_comment(p)) {
				gams_move_past_comment(p);
				set_character_color(p, point, color_class gams_comment);
				return 1;
			}
			break;
		case '$':			// found control statement
			if (parse_string(1, "ontext")) {
				search(1, "\n$offtext");
				gams_color_comment(p, point);
				return 1;
			}
			nl_forward();
			color_gams_control(p, point);
			return 1;
	}
	return 0;
}

// Recolor some data in the middle of a data block.  Return 1 if data
// ended before the block did (must recolor more if so), else 0.
int gams_recolor_data_block(int sdata, int edata, int sblock, int eblock)
{
	save_var narrow_end = size() - eblock;
	save_var point = edata;
	nl_forward();
	narrow_end = size() - point;
	point = sdata;
	save_var narrow_start = sblock;
	nl_reverse();
	to_begin_line();
	narrow_start = 0;
	gams_color_data_list();
	return point < size() - narrow_end;
}

// These keywords are followed by identifiers and stuff we parse specially.
#define GAMS_DATA_KEYWORDS "set|sets|scalar|scalars|parameter|parameters|table" \
						   "|variable|variables|equation|equations|model|models" \
						   "|file|files|acronym|acronyms"

// These keywords (plus the ones above) start units.  Semicolon is
// optional before them.
#define GAMS_UNIT_STARTERS "option|options|loop|while|alias|abort|solve|display"

// Other keywords.
#define GAMS_KEYWORDS "card|ord|inf|na|eps|yes|no|smax|smin|not|and|or|sum|prod" \
					  "|exp|all|assign|diag|else|eq|for|ge|gt|if|le|lt|maximizing" \
					  "|minimizing|ne|putpage|puttl|repeat|sameas|system|then" \
					  "|until|using|xor"

// These can come before a VARIABLE declaration.
#define GAMS_TYPE_KEYWORDS "binary|negative|positive|integer|free|semicont" \
						   "|semiint|sos1|sos2|nonnegative|nonpositive"

// Identifiers that aren't really GAMS keywords but should be
// highlighted anyway.
#define GAMS_SPECIALS	"abs|arctan|ceil|cos|errorf|exp|floor|include|log" \
						"|log10|mapval|max|min|mod|normal|power|put|putclose" \
						"|round|sign|sin|sqr|sqrt|trunc|undf|uniform"

// A GAMS statement ends with a semicolon, or with a line break
// followed by a keyword.
#define GAMS_STATEMENT_END	"(;|[ \t]*\n([ \t\n]|[*$].*\n)*(" \
							GAMS_DATA_KEYWORDS "|" GAMS_UNIT_STARTERS "|" \
							GAMS_TYPE_KEYWORDS ")</word|lparen>)"

// A list of possible suffixes.  For speed we color any word
// using the syntax of a suffix, even if it's not in this list.
#define GAMS_SUFFIXES	"date|ifile|ofile|page|rdate|rfile|rtime|sfile|time" \
						"|title|ts|tl|te|l|lo|m|prior|scale|up|lj|nj|sj|tj" \
						"|lw|nw|sw|tw|nd|nr|nz|cc|hdcc|tlcc|cr|hdcr|tlcr|ll" \
						"|hdll|tlll|lp|ws|pc|ps|pw|tm|bm|case|bratio|domlim" \
						"|domusd|holdfixed|iterlim|iterusd|limcol|limrow" \
						"|modelstat|numequ|numinfes|numnopt|numnz|numunbnd" \
						"|numvar|optca|optcr|optfile|reslim|resusd|scaleopt" \
						"|solprint|solveopt|solvestat|sysout|workspace"

color_gams_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk = 0, c, talk_at = from + 40000, p, q, sdata, edata, data = 0;

	if (from >= to)
		return to;
	if (modified_buffer_region(&p, &q, "gams-regions")) {
		// Buffer's changed, so forget saved comment-char info after
		// this point.
		set_tagged_region("gams-inlinecom", p, size(), -1);
		set_tagged_region("gams-eolcom", p, size(), -1);
		if (get_tagged_region("color-gams-data", from, &sdata, &edata) > 0
			  && sdata < p && q < edata) {	// Modification within data block.
			if (!gams_recolor_data_block(p, q, sdata, edata)) {
				reset_modified_buffer_region("gams-regions");
				return to;
			}
		}
		set_tagged_region("color-gams-data", p, size(), -1);
		set_tagged_region("needs-color", to, size(), -1);
	}
	reset_modified_buffer_region("gams-regions");

	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	c = get_tagged_region("color-gams-as-unit", give_begin_line() - 1, &p);
	if (c && p < point)	// Jump back over data areas.
		point = p;
	if (data < 0)
		while (re_search(-1, GAMS_STATEMENT_END))	// Move back to start of a statement.
			if (get_character_color(point) != color_class gams_comment
				  && get_tagged_region("color-gams-as-unit", matchstart) == -1)
				break;
	from = point;
	set_character_color(point, to, -1);
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.]+|^[*$]|[!#?\"'{]|/![*/]|%<%<|_!%$")) {
			t = size();
			break;
		}
		t = matchstart;
		c = character(point - 1);
		if (!gams_color_common(c, t))
			if (isalnum(c) || strchr("_.", c))	// found identifier, kywd, or number
				gams_keyword_color(t, point);
		if (point > talk_at) {
			note("Coloring GAMS file: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point + 2000;
			talk = 1;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk)
		note("");
	return point;
}

// Set color of "identifier" in this buffer range (actually anything
// with alpha or digits).
gams_keyword_color(from, to)
{
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(*buf) - 10)
		save_var point = from + sizeof(buf) / sizeof(*buf) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-", buf[1]) || buf[1] == '.' && isdigit(buf[2])) {
		set_character_color(from, to, -1);
		return;
	}
	strcpy(buf + point - from + 1, "|");
	if (is_gams_data(buf)) {
		set_character_color(from, to, color_class gams_keyword);
		gams_color_descriptions();
		set_tagged_region("color-gams-as-unit", from, point - 1, 1);
		set_tagged_region("color-gams-as-unit", point - 1, point, -1);
	} else if (is_gams_keyword(buf))
		set_character_color(from, to, color_class gams_keyword);
	else {
		set_character_color(from, to, color_class gams_variable);
		if (parse_string(-1, "%.[a-z]+"))
			set_character_color(matchend, point, color_class gams_suffix);
	}
}

gams_color_descriptions()
{
	int descr, done, table, c, p, q;

	set_tagged_region("color-gams-data", point, size(), -1);
	if (parse_string(-1, GAMS_TYPE_KEYWORDS)) {
		gams_move_past_spaces_and_comments();
		if (parse_string(1, GAMS_DATA_KEYWORDS)) {
			set_character_color(matchstart, matchend, color_class gams_keyword);
			point = matchend;
		}
	}
	table = parse_string(-1, "table");	// Special rules for this keyword.
	gams_move_past_spaces_and_comments();
	move_past_gams_identifier();
	do {
		if (parse_string(1, GAMS_STATEMENT_END))
			return;
		point += parse_string(1, "[ \t]*");
		descr = point;
		done = 0;
		c = curchar();
		if (strchr("\"'", c)) {
			gams_parse_and_color_string(p = point);
		} else {
			if (re_search(1, "[\n,/]|" GAMS_STATEMENT_END)) {
				c = character(p = point - 1);
				if (isalpha(c)) {		// relaxed syntax rule: \n plus keyword means ;
					c = ';';
					p = matchstart;
					point = p + 1;
				}
			} else {
				done = 1;
				p = point;
				c = ';';
			}
		}
		set_character_color(descr, p, color_class gams_string);
		switch (c) {
			case ';':
				return;
			case '/':
				q = give_end_line();
				gams_color_data_list();
				set_tagged_region("color-gams-data", q, give_begin_line(), 1);
				if (parse_string(1, GAMS_STATEMENT_END)) {
					point = matchstart;
					return;
				}
				point += parse_string(1, "[ \t\n,]*");
				gams_move_past_spaces_and_comments();
				move_past_gams_identifier();
				break;
			case '\n':
			case ',':
			case '"':
			case '\'':
				if (parse_string(1, GAMS_STATEMENT_END)) {
					point = matchstart;
					return;
				}
				point += parse_string(1, "[ \t\n]*");
				if (table) {
					if (re_search(1, GAMS_STATEMENT_END))
						point = matchstart;
					return;
				}
				gams_move_past_spaces_and_comments();
				move_past_gams_identifier();
				break;
		}
	} while (!done);
}

void gams_read_list_file(int res, int buf);

command gams_mode()
{
	mode_default_settings();
	major_mode = _gams_mode_name;
	compile_buffer_cmd = compile_gams_cmd;	// can compile this
	post_compile_hook = gams_read_list_file;
	if (auto_show_gams_delimiters)
		auto_show_matching_characters = gams_auto_show_delim_chars;
	strcpy(comment_start, "^*[ \t]*");
	strcpy(comment_pattern, "^*.*$");
	strcpy(comment_begin, "* ");
	strcpy(comment_end, "");
	recolor_range = color_gams_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	coloring_flags |= COLOR_INVALIDATE_RESETS;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("gams-mode-hook");
	make_mode();
}

copy_matches(int tobuf, int flags, char *pat, char *rel, char *extra)
{
	int sbuf = bufnum, cnt = 0, i, more, srcherr, lines;
	int before, file, do_color;

	do_color = want_code_coloring && recolor_range;
	save_var point = 0;
	while ((srcherr = do_color_searching(flags, pat)) == 1) {
		i = give_position_at_column(matchstart, 0);
		more = nl_forward();
		lines = lines_between(i, point, 0);
		before = get_buf_point(tobuf);
		if (lines > 1)
			buf_printf(tobuf, "%s%s [%d]: ", extra, rel, lines);
		else
			buf_printf(tobuf, "%s%s: ", extra, rel);
		file = get_buf_point(tobuf);
		if (do_color) {			// Color in place, then copy.
			if (do_color == 1 && idle_coloring_size > 1000
				  && (recolor_from_here != recolor_by_lines)) {
				recolor_buffer_range(0, size());
				do_color = 2;		// Color all at once if coloring pieces is
			}						// slow.
			if (do_color == 1)
				recolor_buffer_range(i, point - more);
		}
		buf_xfer_colors(tobuf, i, point - more);
		bufnum = tobuf;
		set_character_color(before, file, color_class grep_filename);
		bprintf("\n");
		set_character_color(point - 1, point, color_class grep_filename);
		point = file;	// Now color matches within line.
		while (do_searching(flags, pat) == 1 && point > file) {
			set_character_color(matchstart, matchend, color_class grep_match);
			file = point;
		}
		point = size();
		bufnum = sbuf;
		cnt++;
		if (!more)
			return cnt;
	}
	return srcherr ? -srcherr : cnt;
}

copy_man_page_matches(int flags, char *pat)
{
	int before = buffer_size("grep");

	xfer("grep", give_begin_line(), give_end_line());
	save_var bufname = "grep";
	set_character_color(before, size(), 0);
	bprintf("%p\n", color_class grep_filename);
	save_var point = before;
	while (do_color_searching(flags, pat) == 1)
		set_character_color(matchstart, matchend, color_class grep_match);
}

prompt_file_replace(flags, inbufs)
{		// prompt for replacing in files/buffers, then do it
	char *str, with[10000], filepat[FNAMELEN];
	char old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	flags = get_replace_text(flags | (inbufs ? BUFGREP : FILEGREP),
				 with);
	if (flags == -1 || !get_grep_file(inbufs, filepat))
		return;
	str = default_search_string(flags);
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	getcd(buf_dir);
	if (!setjmp(top_level))
		do_file_replace(str, with, flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

command epsilon_manual()
{
	switch (get_epsilon_help_format()) {
		case HELP_FORMAT_WINHELP:
			win_help_string("", "Epsilon Manual Contents");
			return;
		case HELP_FORMAT_HTML:
			epsilon_manual_html();
			return;
	}
	epsilon_manual_info();
}

// Display topic in a browser and return 1.  Return 0 if running a web
// browser is unsupported, or -1 if we couldn't for some other reason.
int try_html_help_string(char *topic)
{
	char url[FNAMELEN], prog[FNAMELEN], *p, *q, cmdline[100], dir[FNAMELEN];

	if (epsilon_manual_port > 0)
		sprintf(url, "http://localhost:%d/man/", epsilon_manual_port);
	else
		strcpy(url, "http://www.lugaru.com/cgi-bin/lhelp/man/");
	for (p = topic, q = url + strlen(url); *p; p++, q++)
		if (*p == ' ') {
			strcpy(q, "%20");
			q += 2;
		} else
			*q = *p;
	*q = 0;
	if (!*topic)
		strcat(url, "0");
	p = lookpath("lhelp/epsilon.lhl");
	if (p) {		// Find name of directory containing our lhelp files.
		strcpy(dir, p);
		*get_tail(dir, 0) = 0;
	} else
		get_executable_directory(dir);
	convert_to_8_3_filename(dir);
	if (opsys == OS_UNIX) {
		sayput("Running a browser...");
		get_executable_file(prog, "lhelpd", 1);
		sprintf(prog + strlen(prog), " -p %d -d \"%s\" &",
				epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			shell("", prog, "");
	} else if (is_gui || is_win32) {
		sayput("Starting a browser, please wait...");
		expire_message = 1;
		get_executable_directory(prog);
		strcat(prog, "lhelp.exe");
		sprintf(cmdline, "lhelp -p %d -d %s", epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			winexec(prog, cmdline, SW_SHOWMINNOACTIVE, 0);
	} else
		return 0;
	if (!try_show_url(url))
		return -1;
	note("");
	return 1;
}

// Return matches for man pages that start with s.
char *man_match(s, flags)
char *s;
{
	char pat[FNAMELEN], match[FNAMELEN], *path, *t, *f, *q;
	char man_path[FNAMELEN * 2];

	if (strchr(s, '/') || strchr(s, '\\'))
		return do_file_match(s, flags);
	save_var cygwin_filenames |= 32;
	if ((flags & STARTMATCH) && *s && (!isalpha(*s) || s[1] == ' ')) {
		zap(_MATCH_BUF);	// Something like "-k foo" or "2 bar".
		buffer_printf(_MATCH_BUF, "%s%s\n", s, (flags & LISTMATCH) ? "" : " ");
		// Just return text itself as the match.
	} else if (flags & STARTMATCH) {
		show_text(-1, 20, "Listing matches...");	// Show for at least .2 secs.
		zap(_MATCH_BUF);
		build_man_path(path = man_path);
		while (!user_abort) {
			t = strchr(path, ':');	// Assume MANPATH uses Unix sep, not DOS.
			if (t)
				strncpy(pat, path, t - path);
			else
				strcpy(pat, path);
			convert_cygwin_filename(pat, 0);
			if (*pat && pat[strlen(pat) - 1] != '/')
				strcat(pat, "/");
			sprintf(pat + strlen(pat), "man*/%s*", s);
			f = do_file_match(pat, STARTMATCH);
			for (; f && !user_abort; f = do_file_match(pat, 0)) {
				strcpy(match, get_tail(f, 0));
				q = get_extension(match);
				if (q && !strcmp(q, ".gz"))
					*q = 0;
				q = get_extension(match);
				if (!q || !*q)
					continue;
				*q = 0;
				buffer_printf(_MATCH_BUF, "%s\n", match);
			}
			if (!t)
				break;
			path = t + 1;
		}
		save_var bufname = _MATCH_BUF;
		point = 0;
		save_var case_fold = 0;
		do_sort_region(0, size(), 0, 0);
		do_uniq(1, 1, 0);
		say("");
		restore_vars();
	}
	return general_matcher(s, flags);
}

command wall_chart()
{
	short *mkeys = mode_keys;

	iter = 0;
	tiled_only();
	noteput("Building wall chart...");
	to_buffer_num(zap("wall"));
	chart_table(mkeys, root_keys, "");
	point = 0;
	modified = 0;
	say("");
}

#define DEFN(c)    ((mtab[c] == 0) ? rtab[c] : mtab[c])

/* take name table index and make a string showing bindings */
find_bindings(msg, ind, mtab, rtab, prefix, count)
char *msg, *prefix;
short *mtab, *rtab;
{
	char tmp[300];
	short *newtab, *newalt;
	int i, maxwidth = screen_cols - 1;

	i = list_bindings(-1, mtab, rtab, ind);
	while (i != -1) {
		if (DEFN(i) == ind) {
			if (strlen(msg) > maxwidth - 24) {
				if (strlen(msg) < maxwidth - 18)
					strcat(msg, ", ...");
				return count;
			}
			if (count++)
				strcat(msg, ", ");
			strcat(msg, prefix);
			show_char(msg, i, 1);
		} else if (DEFN(i) > 0 && name_type(DEFN(i)) == NT_TABLE) {
			strcpy(tmp, prefix);
			show_char(tmp, i, 1);
			strcat(tmp, " ");
			newtab = index_table(mtab[i] ? mtab[i] : rtab[i]);
			newalt = index_table(rtab[i] ? rtab[i] : mtab[i]);
			newtab = newtab ? newtab : newalt;
			newalt = newalt ? newalt : newtab;
			count += find_bindings(msg, ind, newtab, newalt,
						tmp, count);
		}
		i = list_bindings(i, mtab, rtab, ind);
	}
	return count;
}

void hex_old_on_modify();
void hex_on_modify();
int color_hex_range(int from, int to);
char *hex_prefix();

command hex_mode()
{
	char bname[FNAMELEN];
	int pos = point, trans = GET_LINE_TRANSLATE(translation_type);
	struct file_info saved = file_info;

	if (is_hex_buffer(bufnum)) {	// "Turn off mode" -- save back to real buf.
		hex_save_key();
		return;
	}
	if (!size())
		error("Buffer is empty.");
	sprintf(bname, "%s%s", HEX_BUFNAME_PREFIX, bufname);
	if (exist(bname)) {
		say("This buffer is in hex mode.  Press %s? for help or %sq to quit.",
			  hex_prefix(), hex_prefix());
		to_buffer(bname);
		return;
	}
	if (trans == FILETYPE_MSDOS || trans == FILETYPE_MAC)
		hex_reload_file();
	buffer_to_hex(bname);
	if (user_abort) {
		delete_buffer(bname);
		error("Cancelled.");
	}
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_old_on_modify;	// special funcs can modify it.
	file_info.check_type = 0;	// Disable file date checking during hex mode.
	to_buffer(bname);
	file_info = saved;		// Remember old file date checking info.
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_on_modify;	// special funcs can modify it.
	mode_keys = hex_tab;		/* Use these keys. */
	major_mode = hex_overtype_mode ? _hex_overtype_mode_name : _hex_mode_name;
	recolor_range = color_hex_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_delay = -1; // No idle coloring, it slows down apply-changes.
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("hex-mode-hook");
	drop_all_colored_regions();
	make_mode();
	say("Entering hex mode, press %s? for help or %sq to quit.",
		  hex_prefix(), hex_prefix());
	hex_goto_place(pos, hex_overtype_mode);
}

color_css_range(from, to)
{
	int t = -1, talk, s, talked = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	while (from > narrow_start
		   && get_tagged_region("color-as-unit", from, &s) != -1)
		from = s - 1;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	talk = point + 10000;	// show status after coloring this many chars
	while (point < to) {
		if (!re_search(1, "[-0-9A-Za-z_.@%#]+|[\"']|/%*|%<%!--|-->")) {
			t = size();
			break;
		}
		t = matchstart;
		if (character(t) == '<') {
			set_character_color(t, point, color_class html_embedded);
			continue;
		}
		switch (character(point - 1)) {		// check last char
			case '>':	
				set_character_color(t, point, color_class html_embedded);
				break;
			case '*':			// found comment
				search(1, "*/");
				set_character_color(t, point, color_class css_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class css_string);
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class css_string);
				break;
			default:		// found identifier.
				if (parse_string(1, "([ \t\n]*[,:.]?[ \t\n]*"
								 "[-0-9A-Za-z_.:@%#]+)*[ \t\n]*{"))
						set_character_color(t, point, color_class css_selector);
				else if (parse_string(1, "[ \t\n]*:"))
					set_character_color(t, point, color_class css_class);
				else
					set_character_color(t, point, -1);
				break;
		}
		if (point > talk) {
			note("Coloring CSS: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk = point + 5000;
			talked = 1;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talked)
		note("");
	return point;
}

html_narrow_to_paragraph()
{
	char pat[1200];
	char *elems = (sublanguage_code == SUBLANG_XML) ? xml_paragraph_elements
		: html_paragraph_elements;
	save_var point;
	int pos = point, save, inside;
	inside = html_in_tag();
	if (inside >= 0) {		// If we're inside a tag like <p, go past it
		sprintf(pat, "[<](%s)</word>", elems); 	// so we find it in next step.
		if (matches_at(inside, 1, pat))
			point = matchend;
	} else if (parse_string(1, "<hspace>*[<]<alphanum>+"))
		point = matchend;	// Likewise if we're just before one.
	sprintf(pat, "\n<hspace>*\n|[<](%s)</word>", elems);
	if (re_search(-1, pat)) {
		save = point = matchstart;
		if (character(matchend) == '<') {
			point = matchend;
			int end = default_move_level(1, "<", ">");
			point = (end > 0) ? end : save;
			point += parse_string(1, "<hspace>*");
		}
		to_begin_line();
		narrow_start = point;
	}
	point = pos;
	sprintf(pat, "\n<hspace>*\n|[<]/(%s)</word>", elems);
	if (re_search(1, pat)) {
		point = matchstart;
		point -= parse_string(-1, "<wspace>*");
		narrow_end = size() - point;
	}
}

int html_restrict_break(int pos);

command html_fill_paragraph() on html_tab[ALT('q')]
{
	int region, start, end;

	recolor_buffer_range(point, point + 1);
	region = html_region_type(point);
	switch (region) {
		case HT_JSCRIPT:
			save_var major_mode = _c_mode_name;
			fill_comment();
			return;
		case HT_VBSCRIPT:
		case HT_PHP:
		case HT_PYTHON:
		case HT_CSS:
			fill_comment();
			return;
	}
	if (!is_highlight_on()) {
		save_var narrow_start, narrow_end;
		if (get_character_color(point, &start, &end) == color_class html_comment) {
			narrow_start = start;
			narrow_end = size() - end;
		} else
			html_narrow_to_paragraph();
	}
	save_var mode_restrict_break = html_restrict_break;
	fill_indented_paragraph();
}

// If bit is on in align_region_rules, look through region for all lines
// matching the regex pattern pat.  For each, find the boundary between
// #1 and #2 (less any whitespace).  Align the following text as close
// to the longest #1 as possible, plus the specified extra room, in all
// matching lines. Return # of lines.
int align_by_pattern(int bit, int start, int end, char *pat, int extra)
{
	int max = 0, i, cnt = 0;
	if (bit && !(align_region_rules & bit))
		return 0;
	save_var point, narrow_end = size() - end;
	point = start;
	to_begin_line();
	start = point;
	while (do_color_searching(REGEX, pat) == 1) {
		point = find_group(1, 0);
		if (align_region_rules & 1024)
			show_text(-2, 0, "Rule %d triggered on line %d at column %d "
					  "(adjusted to %d).",
					  bit, lines_between(0, point) + 1, current_column(),
					  get_column(point - parse_string(-1, "[ \t]*")));
		point -= parse_string(-1, "[ \t]*");
		i = current_column();
		to_end_line();
		cnt++;
		if (i > max)
			max = i;
		check_abort();
	}
	if (cnt < 2)
		return cnt;
	point = start;
	while (do_color_searching(REGEX, pat) == 1) {
		point = find_group(1, 0);
		indent_to_column(max + extra + align_region_extra_space);
		to_end_line();
		check_abort();
	}
	return cnt;
}

// Alignment rules for C mode.
c_mode_aligner()
{
	// Align variable names being defined.
	ALPAT(4, "((" C_IDENT "[ \t]+)+)([*(]*" C_IDENT "[ \t]*%)?[,;=].*)", 2);
	ALPAT(1, "((" C_IDENT "|[][ \t])+)(=.*)", 1); // Align = in assignments.
	ALPAT(8, "(#define[ \t]+" C_IDENT "(%((" C_IDENT
		  "|[, \t])*%))?[ \t]+)([^ \t\n]+)", 2);	// Align definitions.
	// Align comments.
	ALPAT(2, "(<^c:c-comment>.*)(<c:c-comment>/[/*].*)", 2);
	ALPAT(16, "(.)(\\)$", 2);
}

command align_by_tab()
{
	int spacing = 2;
	char num[FNAMELEN];
	iter = 0;
	if (has_arg) {
		get_strdef(num, "Extra spacing", "2");
		spacing = numtoi(num);
	}
	save_spot mark, point;
	fix_region();
	save_var indent_with_tabs = 0;
	while (ALPAT(0, "([^\t]*)*\t([^\t]*)", spacing) > 1)
		check_abort();
	restore_vars();
	if (indent_with_tabs)
		tab_convert(point, mark, 1);
}

char *info_get_std_dirs()
{
	if (is_win32)
		return "{/cygwin*,}/usr/*/info/dir";
	else
		return "/usr/*/info/dir";
}

int load_info_file(char *file, int flags)
{
	char modfile[FNAMELEN];
	int err;

	save_var no_popup_errors = 1;
	drop_quotes(file);
	set_read_only(0);		// Let us modify the info buffer.
	err = info_read_fix_case(file);
	if (err && err != EREADABORT) {		// Try adding suffixes.
		sprintf(modfile, "%s-info", file);
		err = info_read_fix_case(modfile);
		if (err && err != EREADABORT) {
			sprintf(modfile, "%s.info", file);
			err = info_read_fix_case(modfile);
		}
		if (err && err != EREADABORT) {
			sprintf(modfile, "%s.inf", file);
			err = info_read_fix_case(modfile);
		}
		if (!err)
			file = modfile;
	}
	if (err && err != EREADABORT)
		if (flags & 1)
			return 0;
		else
			file_error(err, file, "read error");
	if (err) {
		info_return(0);
		quick_abort();
	}
	info_mode();
	point = size();
	if (re_search(-1, "\n" INFO_CSTR "\f?\nTag table:[\t ]*\n!")
		  && parse_string(1, "%(Indirect%)"))
		info_load_indirect_file(file);
	perform_unicode_conversion(bufnum, 0, size(), CONV_TO_16, "UTF-8-No-BOM");
	modified = 0;
	translation_type = FILETYPE_AUTO;
	set_read_only(3);
	return 1;
}

command info_search() on info_tab['s']
{
	char *s, node[FNAMELEN];
	int flags, ret, found;

	flags = get_search_string("Search Info: ", UNKNOWN);
	flags &= ~(INCR | FAILING | BAD);
	s = default_search_string(flags);
	if (flags & UNKNOWN)
		flags = known_flags(flags);
	info_record();
	narrow_start = narrow_end = 0;
	while ((ret = do_searching(flags, s)) == 1) {
		found = point;		// Now see if it's in a node.
		if (re_search(-1, "\n" INFO_CSTR "\f?\n(.*[\t, ])?"
					  "Node:[\t ]*([^\t\n,]+)[\t\n,]")) {
			grab(find_group(2, 1), find_group(2, 0), node);
			info_go_node(node, 0);	// Narrow to this node.
			point = found;		// Narrowing might modify point here.
			if (point == found) {	// Yes, match is in this node.
				save_var point = 0;
				color_for_narrowed_area(narrow_start, size() - narrow_end);
				while ((ret = do_searching(flags, s)) == 1)
					set_character_color(matchstart, matchend,
										color_class info_match);
				return;
			}
			narrow_start = narrow_end = 0;
			point = found;		// No, so continue searching.
		} else
			point = found;
	}
	info_return(0);
	if (ret == DSABORT)
		say("Search cancelled.");
	else
		say("%s: not found.", s);
}

command copy_to_scratch() on cx_tab['x']
{
	char buf[80];

	mention("Save region in scratch buffer [A-Z, 0-9]: ");
	get_alnum();
	save_spot point, mark;
	fix_region();
	sprintf(buf, "-temp-buf-%c", key);
	zap(buf);
	if (region_type() == REGRECT)
		extract_rectangle(name_to_bufnum(buf), 0);
	else
		xfer(buf, point, mark);
	say("");
	highlight_off();
}

run_by_pattern(p)
char *p;
{
	char text[FNAMELEN], cmdline[FNAMELEN], *s;

	s = index(p, '!');
	save_spot point, mark;
	fix_region();
	if (is_highlight_on() && s && mark - point < sizeof(text) / sizeof(*text)) {
		grab(point, mark, text);
		highlight_off();
	} else
		*text = 0;
	if (s && *text) {
		strcpy(cmdline, s + 1);
		p = index(cmdline, '!');
		if (p) {
			strcpy(p, text);
			strcat(p, index(s + 1, '!') + 1);
		}
	} else if (s) {
		strcpy(cmdline, p);
		if (s = index(cmdline, '!'))
			*s = 0;
	}
	if (is_gui || is_win32) {
		char prog[FNAMELEN];
		say("%s", cmdline);
		if (*cmdline == '"') {
			strcpy(prog, cmdline + 1);
			s = index(prog, '"');
			if (s)
				*s = 0;
		} else {
			strcpy(prog, cmdline);
			s = index(prog, ' ');
			if (s)
				*s = 0;
		}
		winexec(prog, cmdline, SW_SHOWNORMAL, 0);
	} else
		shell("", cmdline, "");
}

menu_invoked() on reg_tab[WIN_MENU_SELECT], reg_tab[WIN_HELP_REQUEST]
{
	int i;

	if (_doing_input == DI_LINEINPUT && is_gui) {
		ungot_key = full_key;
		error_if_input(1);
	}
	if (*menu_command == '$') {		// help file
		char old[FNAMELEN], dir[FNAMELEN];
// MSVC 2 wants the current directory to be \msvc20\bin, when using the help
// file \msvc20\help\xxx.hlp.  The other help files we know about don't care.
// So temporarily set the current directory based on the help file name.
		getcd(old);
		strcpy(dir, menu_command + 1);
		*get_tail(dir, 0) = 0;
		sprintf(get_tail(dir, 0), "bin");
		chdir(dir);
		i = windows_help_from(menu_command + 1, winhelp_display_contents);
		chdir(old);
		if (!i)
			error("Could not display help from %s", menu_command + 1);
		return;
	} else if (*menu_command == '!') {
		run_by_pattern(menu_command + 1);
		return;
	} else if (*menu_command == '%') {
		run_on_file_template(menu_command + 1);
		return;
	}
	save_var run_by_mouse = 1;
	i = find_index(menu_command);
	switch (name_type(i)) {
		case NT_MACRO:
			run_by_mouse = 0;
		case NT_COMMAND:
		case NT_SUBR:
		case NT_AUTOLOAD:
		case NT_AUTOSUBR:
			add_key_to_macro(CMD_INDEX_KEY + i);
			if (shift_pressed() & (KB_CTRL_DN | KB_SHIFT_DN))
				has_arg = iter = 1;
			do_command(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR:
		case NT_VAR:
		case NT_BUILTVAR:
			do_set_variable(i);
			break;
		default:
			error("Couldn't call `%s' from menu.", menu_command);
	}
}

prepare_dialog_helper_cmdline(char *cmdline)
{
	char exe[FNAMELEN];

	get_executable_file(exe, "epsilon-dialogs", 1);
	sprintf(cmdline, "GTK_MODULES= %s --parent %x ", exe, screen_to_window_id(0));
}

// Return 0 if we can't because program's missing.
int set_font_for_unix()
{
	char buf[FNAMELEN];

	prepare_dialog_helper_cmdline(buf);
	sprintf(buf + strlen(buf), " --font \"%s\" 2>&1", font_fixed);
	if (!pipe_text("-epsilon-dialog-out", "-epsilon-dialog-out", buf, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF, 0))
		file_error(errno, "Couldn't run dialog helper", "Error");
	save_var bufname = "-epsilon-dialog-out";
	point = 0;
	if (parse_string(1, ".*(ibrary|egmentation|shared|X Error)")) {
		dialog_helper_failed(1);
		return simple_set_font_for_unix();
	} else if (re_search(1, "^=") && parse_string(1, ".*", buf))
		font_fixed = buf;
	return 1;			// User picked one, even if it was bad.
}

// Make a list of fonts by running xlsfonts; have the user pick one.
// Return 0 if we can't get the list.
int simple_set_font_for_unix()
{
	char buf[200];

	pipe_text(NULL, _MATCH_BUF, "xlsfonts", NULL,
			  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
	save_var bufname = _MATCH_BUF;
	if (lines_between(0, size(), 0) < 5) {
		say("Couldn't run xlsfonts to list the fonts.");
		return 0;
	}
	point = 0;
	if (!getenv("EPSMOREFONTS"))
		do_drop_matching_lines(REGEX, "-p-[^-\n]*-[^-\n]*-[^-\n]*$", 1);
	do_uniq(1, 1, 0);
	restore_vars();			// Discard proportional fonts.
	save_var resize_menu_list = 1;
	ungot_key = '?';		// Show font list immediately.
	prompt_comp_read(buf, "Select a font", general_matcher,
					 COMP_FOLD | MUST_MATCH | NONE_OK, "");
	if (!user_abort)
		font_fixed = buf;
	if (strcmp(font_fixed, buf))
		say("Sorry, can't use that font.");
	return 1;
}

command change_font_size()
{
	char *s, name[50], buf[50], old[50];
	int h, w, dir;

	if (is_unix)
		error("Please use the set-font command to select a font.");
	if (has_arg) {
		font_size_increment(iter);
		iter = 0;
		return;
	}
	strcpy(old, font_fixed);
	strcpy(name, font_fixed);
	s = index(name, ',');
	*s = 0;
	h = numtoi(s + 1);
	s = index(s + 1, ',');
	w = numtoi(s + 1);
	s = index(s + 1, ',') + 1;		// rest of spec
	for (;;) {
		note("Using font %s %dx%d, screen is %dx%d. "
			 "(Press an arrow key to change.)",
			 name, h, w, screen_cols, screen_lines);
		getkey();
		if (user_abort || key == BELL || key == ESC) {
			font_fixed = old;
			break;
		}
		dir = get_direction();
		switch (dir) {
			case BTOP:		if (h > 0) h--; break;
			case BBOTTOM:	h++; break;
			case BLEFT:		if (w > 0) w--; break;
			case BRIGHT:	w++; break;
		}
		if (dir != -1) {
			sprintf(buf, "%s,%d,%d,%s", name, h, w, s);
			font_fixed = buf;
			refresh();
		} else if (IS_ENTER_KEY(key) || key == ' ')
			break;
		else if (IS_TRUE_KEY(key))
			say("Press <Esc> to abort, <Enter> to keep these settings.");
	}
	say("");
}

// Increment and decrement the font size value in the current font name under Windows.
font_size_increment(int incr)
{
	char newfont[FNAMELEN], *p, *q;

	p = strchr(font_fixed, ',');
	if (p) {
		int size = strtoi(p + 1, 10);
		q = strchr(p + 1, ',');
		if (q) {
			strncpy(newfont, font_fixed, p - font_fixed);
			sprintf(newfont + (p - font_fixed), ",%d%s", size + incr, q);
			font_fixed = newfont;
			return;
		}
	}
}

// Color simple variable references within this string-like thing (already
// colored itself).
color_perl_interpolated(int from, int to, int comments)
{
	int p;

	save_var point = from;
	save_var narrow_end = size() - to;
	while (re_search(1, (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					 ? "[\\$]" : "[\\$@#]")) {
		p = point;
		if (character(point - 1) == '\\')
			point++;
		else if (character(point - 1) == '#') {
			if (!comments)
				continue;
			from = point - 1;
			to_end_line();
			if (point > to)
				point = to;
			set_character_color(from, point, color_class perl_comment);
		} else if (sublanguage_code == SUBLANG_PHP) {
			if (parse_string(-1, "[^\\]{%$")) {
				point -= 2;
				to_default_move_level(1, "{", "}", 1);
				set_character_color(p - 2, point, color_class perl_variable);
			} else if (parse_string(1, "<word>+(-%><word>+)*|{<word>+}")) {
				point = matchend;
				if (curchar() == '[')
					to_default_move_level(1, "[", "]", 0);
				set_character_color(p - 1, point, color_class perl_variable);
			}
		} else if (in_shell_buffer) {
			if (curchar() == '(') {
				shell_parse_substitution(point - 1);
			} else if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									"|[^ \t\n`]"))
				set_character_color(matchstart - 1, point = matchend,
									color_class perl_variable);
		} else if (curchar() == '{') {
			to_default_move_level(1, "{", "}", 1);
			set_character_color(p - 1, point, color_class perl_variable);
		} else if (parse_string(1, "#?([0-9]+|[A-Za-z_][0-9A-Za-z_]*"
								"[:0-9A-Za-z_]*)|%^[A-Z]"
								"|#?[]-[&`'+*|%=~^_./,\\\";:#?!@$<>()]") && point < to) {
			point = matchend;
			for (;;) {
				if (curchar() == '-' && character(point + 1) == '>'
					  && (isalnum(character(point + 2))
						  || strchr("[{", character(point + 2))))
					point += 2;
				else if (curchar() == '[' || curchar() == '{') {
					if (to_default_move_level(1, "[{", "]}", 1) < 0)
						break;
				} else
					break;
			}
			set_character_color(p - 1, point, color_class perl_variable);
		}
	}
}

// Match #1 at point has a terminating string.  See if it's a likely
// here document and color if so.
color_perl_here_doc(int start, int interpolated)
{
	char line[FNAMELEN], pat[FNAMELEN], *s;
	int p1, p2, end, type = 0;

	grab(find_group(1, 1), find_group(1, 0), line);
	p1 = give_end_line() + 1;
	if (sublanguage_code != SUBLANG_PHP
		  && get_tagged_region("color-perl-as-unit", p1) != -1) {
		get_tagged_region("color-perl-as-unit", p1, NULL, &p1);
		start = p1;
	}
	if (in_shell_buffer && parse_string(1, "[ \t]*-"))
		sprintf(pat, "^\t*%q</word>", line);
	else
		sprintf(pat, "^%q%s$", line, sublanguage_code == SUBLANG_PHP ? ";?" : "");
	if (re_search(1, pat))		// Find end of document (or eof).
		p2 = matchstart;
	else
		p2 = size();
	end = point;
	s = get_extension(line);
	if (s)
		if (!strcmp(s, ".tex"))
			type = 1;
		else if (!strcmp(s, ".html"))
			type = 2;
	set_character_color(start, point, color_class perl_constant);
	set_tagged_region("color-perl-as-unit", start, point, type);
	set_tagged_region("color-perl-as-unit", point, size(), -1);
	save_var point;
	save_var narrow_start = p1;
	save_var narrow_end = size() - p2;
	switch (type) {
		case 1: do_color_tex_range(p1, p2, interpolated); break;
		case 2: color_html_range(p1, p2); break;
	}
	restore_vars();
	if (interpolated)
		color_perl_interpolated(start, point, 0);
	point = end;
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
			   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				if (in_shell_buffer && !parse_string(-1, "([ \n\t|&;()<>]|^)#"))
					break;
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (curchar() == '(') {
						shell_parse_substitution(point - 1);
						break;
					}
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "[#$@]*(((<word>+('|::))*|'|::)<word>+|[-+])")) {
					set_character_color(t, point = matchend,
										color_class perl_variable);
				} else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer) {
					if (!(matches_at_length(t, -1, "\\+") & 1)) {
						search(1, "'");	// No \-quoting in '' for shell mode.
						set_character_color(t, point, color_class perl_string);
					}
				} else if (!strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && (matches_at_length(t, -1, "\\+") & 1))
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				if (curchar() == '/') {  // Found defined-or operator //
					point++;
					break;
				}
				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_><word>*)"
									 : "(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

int perl_smart_move_level(int dir, int comm, int stop_on_key, int ignore_parens)
{
	char delims[100], *pat = (dir > 0) ? "])}[({" : "[({])}";
	int c, i, depth = 0, t, cnt = 0, this_comm;

	while (re_search(dir, "[])}[({]")) {
		t = point - (dir > 0);
		c = character(t);
		i = strchr(pat, c) - pat;
		if (ignore_parens && strchr("()", c))
			continue;
		if (character(t - 1) == '$' && !strchr("{}", c)
			  && get_character_syntax_color(t - 1) == color_class perl_variable)
			continue;		// Actually a $[ or similar.
		if (stop_on_key && (cnt++ % 20 == 0)
			  && (char_avail() || is_key_repeating()))
			return -2;		// Give up search now?
		this_comm = perl_delim_type(t);
		if (this_comm != comm)
			if (!comm)		// We didn't start in comment/string, so
				continue;	// skip over delims in comments/strings.
			else if ((this_comm & 0xf00) == (comm & 0xf00))
				continue;	// Ignore quoted delims.
			else if ((this_comm & IN_COMMENT) && (comm & IN_STRING))
				continue;	// Ignore comment inside string.
			else			// If we started in comment/string,
				return -1;	// stop if we ever search outside it.
		if (i < 3) {		// Closing delim.
			if (depth < 1 || c != delims[--depth])
				return -1;	// This didn't match what we expected.
			if (!depth)		// We found a match, success.
				return point;
		} else if (depth < sizeof(delims) / sizeof(*delims) - 1)
			delims[depth++] = pat[i - 3];	// Opening delim, record it.
		else
			return -1;		// Too deep, give up.
	}
	return -1;
}

tag_mode_perl()
{
	char func[TAGLEN], pat[FNAMELEN], class_name[FNAMELEN];
	int start, i, have_class, class_end = -1, have_const, have_php_const;

	*class_name = 0;
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	sprintf(pat, "<^word>(%s)([a-zA-Z_]<word>*"
			"((::|')[a-zA-Z_]<word>*)?)",
			(sublanguage_code == SUBLANG_PHP)
			? "(function|class)[ \t]+|define[ \t\n]*%([ \t\n]*\""
			: "(sub|use[ \t\n]+constant)[ \t]+");
	while (re_search(1, pat)) {
		have_const = character(find_group(1, 1))
					 == ((sublanguage_code == SUBLANG_PHP) ? 'd' : 'u');
		have_php_const = have_const && (sublanguage_code == SUBLANG_PHP);
		if (get_character_color(find_group(1, 1))
			  == (have_php_const ? color_class perl_function : color_class perl_keyword)
			  && get_character_color(find_group(3, 1))
			  == (have_php_const ? color_class perl_string : color_class perl_function)) {
			grab(start = find_group(3, 1), find_group(3, 0), func);
			have_class = (sublanguage_code == SUBLANG_PHP)
						 && character(find_group(1, 1)) == 'c';
			if (*class_name && point > class_end)
				*class_name = 0;
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (curchar() != '{')
				have_class = 0;
			if (have_class && (i = perl_move_level(1, 0)) > 0) {
				if (tag_which_items & 16)
					add_tag(func, start);
				strcpy(class_name, func);
				class_end = i;
				continue;
			}
			if (!have_const && curchar() == '(') {
				i = perl_move_level(1, 0);
				if (i > 0 && i < size() - narrow_end)
					point = i;
			}
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (have_const) {
				if (tag_which_items & 4)
					add_tag(func, start);
			} else if (curchar() == '{'
					   || (curchar() == ':' || curchar() == '?') && sublanguage_code == SUBLANG_PHP
					   || (tag_declarations & 1))
				add_class_tag(class_name, "::", func, start);
		}
	}
}

do_push(cmdline, cap, show)
char *cmdline;	// cap nonzero means capture output, show means show it to user
{		// but show -1 means don't wait for keystroke
	char dir[FNAMELEN], old[FNAMELEN];
	int result, tbuf;

	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;
	if (opsys == OS_DOS && !is_gui && !is_win32)
		try_exit_concurrent();
	if (is_unix == IS_UNIX_XWIN && !*cmdline) {
		cmdline = push_cmd_unix_interactive;
		if (is_unix_flavor == IS_UNIX_MACOS)
			cmdline = push_cmd_unix_macos_interactive;
	}
	iter = 1;
	before_push();
	if (start_process_in_buffer_directory) {
		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
	}
	result = shell("", cmdline, cap ? bufnum_to_name(tbuf = tmp_buf())
					: "");
	if (should_restore_process_directory())
		chdir(old);
	if (cap) {
		process_captured(tbuf, cmdline);
		buf_delete(tbuf);
	}
	after_push();
	if (cmdline[0] && cap && show) {
		locate_window(PROCBUF, "");
		point = size();
	}
	if (result == -1 && errno) {
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	maybe_restart_concurrent();
	if (cmdline[0] && result)
		say("Process returned %d", result);
	else
		check_dates(0);
	return result;
}

command compile_buffer() on reg_tab[FALT(3)]
{		// compile the current buffer (mode-specific)
	int res, buf = bufnum;
	char cmd[FNAMELEN * 4], oscmd[FNAMELEN * 4], full[FNAMELEN * 4], *ext;

	save_var clear_process_buffer = (clear_process_buffer & 9) != 0;
	tiled_only();
	if (use_compile_command_file_variable && look_for_filevar_pattern(
		"compile-command", "([^-\" \n\t;]([^-;\n]|-[^*])*|\"[^\n\"]+\")", NULL)) {
		grab(find_group(1, 1), find_group(1, 0), oscmd);
		if (*oscmd == '"')
			drop_quotes(oscmd);
		if (use_compile_command_file_variable != 1)
			get_strdef(cmd, "Command to compile this buffer", oscmd);
		else
			strcpy(cmd, oscmd);
		if (*cmd)
			save_var compile_buffer_cmd = cmd;
	}
	if (!compile_buffer_cmd) {
		ext = get_extension(filename);
		sprintf(cmd, "compile-%s-cmd", *ext ? (ext + 1) : "none");
		sprintf(oscmd, "%s-%s", cmd, is_unix ? "unix" :
				is_win32 ? "win32" : "other");
		if ((ext = get_str_var(find_index(oscmd)))
			  || (ext = get_str_var(find_index(cmd))))
			save_var compile_buffer_cmd = ext;
	}
	if (!compile_buffer_cmd || !*compile_buffer_cmd) {
		get_string(cmd, "Command to compile this buffer: ");
		if (!*cmd)
			error("No command specified.");
		compile_buffer_cmd = strsave(cmd);
	}
	save_var concurrent_make;
	if (pre_compile_hook)	// hook can change above vars
		if ((*pre_compile_hook)())
			return;		// hook says don't do compile
	if (concurrent_compile < 3)
		concurrent_make = concurrent_compile;
	if (concurrent_compile == 0 && (is_gui || is_win32))	// Run in the
		concurrent_make = 4;		// concur proc if any, but keep control.
	if (start_process_in_buffer_directory && !is_remote_file(filename)) {
		char dir[FNAMELEN], old[FNAMELEN];

		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
		build_filename(full, compile_buffer_cmd, filename);
		chdir(old);
		if (start_process_in_buffer_directory == 2)	// For error msgs.
			save_var start_process_in_buffer_directory = 1;
	} else
		build_filename(full, compile_buffer_cmd, filename);
	res = do_compile(full);
	save_var bufnum = buf;
	if (post_compile_hook)
		(*post_compile_hook)(res, buf);
}

run_on_file_template(char *template)
{
	char fname[FNAMELEN];
	save_var start_process_in_buffer_directory = 1;
	build_filename(fname, template, filename);
	do_compile(fname);
}

remove_xterm_sequences(int start)
{
	save_var point = start - 25, narrow_start = point;
	save_var case_fold = 0;
	simple_re_replace(1, "\x1B%][0-9];.*\7", "");
	simple_re_replace(1, "\x1B%[K", "");
}

process_unicode_conversion(int start)
{
	save_var point;
	perform_unicode_conversion(bufnum, start, size(), CONV_TO_16, "UTF-8-No-BOM");
}

do_when_process_output(int buf, int start, int end)
{
	save_var bufnum = buf;
	save_var abort_searching = 0;
	set_character_color(start, end, color_class process_output);
	if (process_coloring_rules & 64)
		process_unicode_conversion(start);
	if (process_coloring_rules & 32)
		remove_xterm_sequences(start);
	if (process_coloring_rules & 2)
		color_ansi_output(start);
	if (process_coloring_rules & 8)
		process_backspace_sequences(start - 100);
	if (process_coloring_rules & 16)
		process_remove_controls(start);
	notice_password_prompt(start);
	limit_buffer_size(process_max_size);
	save_var position_window_on_screen_line = 100;
	if (buf_in_window(buf) >= 0)
		refresh();
}

command filter_region() on reg_tab[ALT('|')]
{
	int in_place = has_arg, b, width, orig = bufnum;
	char cmd[FNAMELEN];

	iter = 0;
	get_str_auto_def(cmd, in_place ? "Send region to command" :
			   "Filter region through command");
	b = zap("filter-output");
	if (region_type() == REGRECT) {
		width = extract_rectangle(b, 0);
	} else {
		save_spot point, mark;
		fix_region();		// handle other types of regions
		buf_xfer(b, point, mark);
	}
	if (!buf_pipe_text(b, b, cmd, "", PIPE_CLEAR_BUF
					   | ((in_place | (process_echo & 8)) ? PIPE_SYNCH : 0))) {
		if (bufnum != b)
			buf_delete(b);
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	if (in_place) {
		if (region_type() == REGRECT) {
			save_var bufnum = b;
			xfer_rectangle(orig, width, 1);
		} else {
			delete(point, mark);
			restore_vars();
			grab_buffer(b);
			fix_region();
		}
	} else {
		tiled_only();
		to_buffer_num(b);
		point = 0;
	}
	highlight_off();
}

// If this is a continuation line, move back one line and return 1.
python_back_over_continuation_line()
{
	if (!parse_string(-1, "\\\n[ \t]*"))
		return 0;
	point = matchend;
	return 1;
}

do_python_indent()
{
	int ind, orig = give_begin_line(), close, last;

	close = matches_at(orig, 1, "[ \t]*[])}]");
	save_var point;
	do {		// Find prev non-comment or non-blank line.
		if (!nl_reverse())		// First line in the file, no indent.
			return 0;
		to_begin_line();
	} while (parse_string(1, python_indent_to_comment ? "[ \t]*\n"
						  : "[ \t]*[#\n]"));
	last = point;
	while (python_back_over_continuation_line())
		;
	to_indentation();
	ind = current_column();
	recolor_buffer_range(point, orig);
	// Does this (previous) line end with a : character (plus maybe a
	// comment)?  Or does the previous line end in \, or is there an
	// open delim on prev line?  If so, indent one level more.
	if (matches_at(last, 1, ".*<^c:python-*>:[ \t\n]*(#.*)?$")
		  || matches_at(orig, -1, "\\\n")
		  || python_open_delimiter(point, orig))
		ind += python_indent;
	else if (close)
		ind -= python_indent;
	restore_vars();
	indent_to_column(ind);
}

exit_incr()	/* does this key exit an incremental search? */
{
	int b = get_binding(key, mode_keys);
	if (name_type(b) == NT_MACRO)
		return 0;
	return key != '\t' && key != CTRL('Q') && (key < ' ' || !IS_ASCII_KEY(key))
			&& ((binding_of(full_key, comp_tab) != (short) inp_grab_word
				 && binding_of(full_key, comp_tab) != (short) inp_grab_line)
				|| !IS_ALT_KEY(key)) && !norm_binding();
}

repeat_search(cnt, flags)	 // search & report errors
{
	int orig, ret = 0, wrap = (search_wraps & 2) ? 1 : 0;
	char *s;

	s = default_search_string(flags);
	while (cnt-- > 0) {
        if (wrap == 2)      // Just wrapped, no more wrapping.
            wrap = 0;
        else
            orig = point;
		ret = do_color_searching(flags, s);
		if (ret == 1)
			continue;
		if (ret == DSABORT)
			say("Search cancelled.");
        else if (wrap == 1 && ask_wrap(flags)) {
            wrap = 2;
            point = (flags & REVERSE) ? size() : 0;
            say("Search wrapped.");
            cnt++;
            continue;
        } else {
			say("%s: not found.", s);
			point = orig;
			if (in_macro(1))
				quick_abort();
		}
		break;
	}
	if (search_positions_at_start && ret == 1)
		point = matchstart;
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end)) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command(help_on_search(flags));
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		} else if (IS_MOUSE_SINGLE(key)) {
			ret = 1;
			ungot_key = full_key;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
						show = 1;
						goto ask;
					}
					if (!IS_ENTER_KEY(key))
						ungot_key = full_key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(str, with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
			case ',': case '^':	skip = 1; goto ask;
			case '.':		ret = 1; goto done;
			case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement cancelled.");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
			    replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

rest_sess_wins(b, wins, skip, def_buf, lines, cols)
{
	int nums[10], i, err = 0, bnum, must_default;
	char buf[FNAMELEN];
	struct window_info w;

	for (i = 0; i < wins; i++) {
		if (grab_numbers(b, nums) < 4)
			err = 6;
		w.left = nums[0];
		w.top = nums[1];
		w.right = nums[2];
		w.bottom = nums[3];
		if (grab_numbers(b, nums) < 4 || w.left < 0 || w.top < 0 || w.right < 0 || w.bottom < 0)
			err = 7;
		w.textcolor = nums[0];
		w.hbordcolor = nums[1];
		w.vbordcolor = nums[2];
		w.titlecolor = nums[3];
		if (grab_numbers(b, nums) < 2)
			err = 8;
		w.borders = nums[0];
		w.other = nums[1];
		grab_line(b, buf);
		w.point = -1;
		w.dpoint = -1;
		bnum = name_to_bufnum(buf);
		must_default = !bnum;
		grab_line(b, buf);
		w.prevbuf = name_to_bufnum(buf);
		w.bufnum = bnum ? bnum : w.prevbuf ? w.prevbuf :
			(bnum = give_prev_buf()) ? bnum : def_buf;
		if (grab_numbers(b, nums) < 3)
			err = 9;
		w.dcolumn = nums[1];
		if (!must_default) {
			w.point = nums[0];
			w.dpoint = nums[2];
		}
		if (err)
			break;
		if (!skip) {
			window_create(!i, &w);
			fix_window_start();
		}
	}
	nums[1] = screen_lines - 1;	/* echo line */
	nums[2] = 0;			/* want perm menu? */
	nums[3] = 0;			/* top avoid lines */
	nums[4] = 1;			/* bottom avoid lines */
	if (!err && !grab_numbers(b, nums))
		err = 10;
	if (err)
		one_window();
	else if (!skip) {
		echo_line = nums[1];
		select_low_window(nums[0], nums[3], nums[4], lines, cols);
		i = permanent_menu;
		restore_menu(nums[2]);
		set_menu_bar(i);	// undo setting in session file
	}
	return err;
}

load_binding()		/* new-style binding declaration */
{
	char name[80], junk[80];
	int *onekey, start;
	short *tab;
	int i, now_type, cnt = 1;

	point++;			/* skip initial ~ */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	now_type = (i > 0) ? name_type(i) : -1;
	if (now_type != NT_TABLE)
		init_error("%s is not a table.", name);
	tab = index_table(i);
	point++;			/* skip space before key */
	onekey = get_keycode();
	if (!onekey || onekey[0] != 2)
		init_error("Bad key name.", "");
	start = onekey[1];
	if (parse_string(1, " \"")) {
		point++;
		onekey = get_keycode();
		if (!onekey || onekey[0] != 2)
			init_error("Bad final key name.", "");
		cnt = onekey[1] - start + 1;
	}
	if (!parse_string(1, ": ", junk) || !parse_string(1, ".*", name))
		init_error("Bad key binding.", "");
	i = find_index(name);
	if (i <= 0) {		/* allow name table index too */
		i = numtoi(name);
		if (!i && *name != '0') {	/* not a name, not a number */
			init_warning("No such command %s.", name);
			return;
		}
	}
	set_range(tab, start, i, cnt);
	rebuild_menu();
}

dequote_str(char *s)
{
	for (; *s; s++)
		if (*s == '\\')
			strcpy(s, s + 1);
}

// Get variable name, value, possible default option, and set.
cf_set_variable()
{
	char varname[FNAMELEN], val[FNAMELEN], def[FNAMELEN];
	int i;

	parse_var_name(varname);
	point += parse_string(1, "[ \t\n]*");
	if (!parse_string(1, "-?[0-9][0-9a-zA-Z]*", val)) {
		get_str(val);
		dequote_str(val);
	}
	point += parse_string(1, "[ \t\n]*");
	*def = 0;
	parse_string(1, "-?[0-9][0-9a-zA-Z]*", def);
	i = find_index(varname);
	if (!vartype(i))
		init_error("%s: not a variable", varname);
	if (!var_settable(varname))
		init_error("%s: can't set variable of this type", varname);
	safe_set_var_val(i, val, *def ? numtoi(def) : 3);
}

#include "tags.e"

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class") || !strcmp(func, "__interface"))
		return tag_c_struct(1);		// or similar class or managed c++ def
	else if (!strcmp(func, "interface") || !strcmp(func, "package"))
		return tag_c_struct(1);		// for java or managed c++
	else if ((c_extra_keywords & OBJC_KEYWORDS) && character(start - 1) == '@'
			 && word_in_list(func, "|interface|implementation|protocol|", 0))
		return tag_objc_block();
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if ((c_extra_keywords & CPP_KEYWORDS) && !strcmp(func, "using")) {
		tag_modifier_kwd |= 4;	// just a declaration 
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (!strcmp(func, "__attribute__") && parse_string(1, "<wspace>*%(%(")
			   && !skip_c_delims('(', ')', 1)) {
		return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	if (parse_string(1, "(_ANSI_ARGS_|__P|PROTO)</word>")) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template();
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
												"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (tag_c_variable()
			   && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
				   || (tag_modifier_kwd & 2))
			   && (!(c_extra_keywords & JS_KEYWORDS))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}


tag_c_skip_before_body(int show)
{
	if (parse_string(1, "#.*(\\\n.*)*")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(try|const|volatile|NOTHROW)![^_A-Za-z0-9]")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(AVAILABLE|DEPRECATED_IN)_MAC_OS_X_VERSION_<word>+")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "throws[ \t]")) {	// for java
		point = matchend;
		skip_identifier_list();
		return 1;
	}
	if (parse_string(1, "(throw|__attribute__|__DARWIN_ALIAS)[ \t]*!%(")) {
		point = matchend;	// throw for C++, attribute for gcc
		if (skip_c_delims('(', ')', 1)) {
			if (show)
				buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
							  filename, lines_between(0, point, 0) + 1);
			return 0;
		}
		return 1;
	}
	if ((c_extra_keywords & JS_KEYWORDS) && parse_string(1, ":<wspace>*<alnum>+")) {
		point = matchend;
		return 1;
	}
	return 0;
}

tag_c_func(show)
{
	int ret = 2;

	if (skip_c_delims('(', ')', 1)) {
		if (show && !in_c_comment(0))
			buffer_printf("-tag-errors", "Warning: couldn't parse %s at line %d\n",
						  filename, lines_between(0, point, 0) + 1);
		return 0;	/* after identifier, skip over arg list */
	}
	c_fwd_space();
	while (tag_c_skip_before_body(show))
		c_fwd_space();
	if (parse_string(1, "on[ \t]+<word>+[ \t]*<[>")) {	// eel command
		search(1, "\n{");
		point--;
		ret = 2;
	} else if (parse_string(1, "[A-Za-z{]")) // func definition
		ret = skip_param_decls();
	else if (curchar() == ':')		// C++ initializer
		ret = skip_cpp_init();
	else if (curchar() == '=')		// C++ function using =0.
		return 3;
	else if (!parse_string(1, "[,;]"))	// not even a declaration
		ret = (point >= size());	// Skip parens if it looks like a macro.
	else if (!(tag_declarations & 1))	// a declaration, don't tag it?
		ret = 1;
	if (ret)
		tag_modifier_kwd = 0;
	return ret;
}

char *tag_c_ident_pattern()
{
	return (c_extra_keywords & (CSHARP_KEYWORDS | JS_KEYWORDS | JAVA_KEYWORDS)) ?
			"<word|.>+" : "<word>+(::<word>+)?";
}

tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start, i;

	if (flavor == 1 && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(-1, "@class"))
		return tag_objc_class_list();
	c_fwd_space();
	start = i = point;
	if (parse_string(1, "__declspec</word>")) {
		point = matchend;
		c_fwd_space();
		if (curchar() == '(')
			c_skip_level('(', ')');
		c_fwd_space();
	}

	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	if (flavor == 0 && (c_extra_keywords & CPP_KEYWORDS))
		flavor = 1;	// In C++, treat structs just like classes.
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer() && !(c_extra_keywords & JS_KEYWORDS)
		  && parse_string(1, tag_c_ident_pattern())) {
		i = point;
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (strchr(";[,", curchar()) && i > start) {
		point = i;	// A plain definition.
		return 1;
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		tag_c_add_class(func);
		i = ++point;	// skip over { so we do stuff within
		c_fwd_space();
		if (parse_string(1, "#(else|elif)"))
			if (locate_by_ifdef(1) != 2)
				point = i;
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

parse_other_class()
{
	c_fwd_space();
	if (is_java_buffer() && curchar() == '<')
		c_skip_level('<', '>');
	c_fwd_space();
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (parse_string(1, tag_c_ident_pattern())
				  || parse_string(1, "[,:]"))
				point = matchend;
			else if (!find_template_end())
				break;
		}
		c_fwd_space();
	} else if (c_extra_keywords & (JAVA_KEYWORDS)) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}

int color_tcl_range(from, to);
void tcl_indenter();

command tcl_mode()
{
	mode_default_settings();
	mode_keys = tcl_tab;		/* Use these keys. */
	major_mode = _tcl_mode_name;
	compile_buffer_cmd = compile_tcl_cmd;	// can compile this
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 128) != 0;
	recolor_range = color_tcl_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	indenter = tcl_indenter;
	auto_indent = 1;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_tcl_delimiters)
		auto_show_matching_characters = tcl_auto_show_delim_chars;
	try_calling("tcl-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

make_ctrl(k)
{
	return CTRL(k);
}

select_menu_item(resbuf, menuwin, owin, dir)
{	// let user select line from menu window, maybe copy line to resbuf
	int res; // dir tells how to srch on norm keys, or 0 if they're cmds
	char str[FNAMELEN], pat[FNAMELEN * 2];

	save_var mode_keys;
	cursor_to_column = -1;
	do {
		mode_keys = menu_tab;
		res = do_recursion();
		if (!is_window(menuwin))
			res = INP_ABORT;
		else if (res == INP_ADD && search_in_menu == 2) {
			sprintf(pat, "^%c", key);
			window_handle = menuwin;
			to_end_line();
			if (!re_search(1, pat)) {
				point = 0;
				re_search(1, pat);
			}
			to_begin_line();
			res = 0;
		} else if (res == INP_ADD && search_in_menu && dir) {
			window_handle = owin;
			run_topkey();
			window_handle = owin;
			grab(0, FNAMELEN - 1, str);
			sprintf(pat, "^%q", str);
			window_handle = menuwin;
			point = (dir == 1) ? 0 : size();
			re_search(dir, pat);
			to_begin_line();
			res = 0;
		} else if (res == INP_ADD && !dir)
			switch (isalpha(key) ? toupper(key) : 0) {
				case 'N':
				case 'P':	ungot_key = CTRL(key);
				default:	res = 0;	// fall through
				case 'Q':	break;
			}
	} while (!res);
	switch (res) {		// try all other chars
		case INP_ADD:	ungot_key = full_key;
						break;
		case INP_GET_EXIT:
		case INP_GET:	if (resbuf)
							get_menu_line(resbuf);
						break;
	}
	if (res == INP_ABORT && !user_abort)	// user deleted edit window?
		user_abort = 1;
	remove_line_highlight();
	window_handle = owin;
	remove_window(menuwin);
	return res;
}

tag_mode_jscript()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (re_search(1, "(^|;)[ \t]*(var|function)[ \t]+(<word>+)")) {
		int var = tolower(character(find_group(2, 1))) == 'v';
		grab(start = find_group(3, 1), find_group(3, 0), func);
		if (!isdigit(*func))
			add_tag(func, start);
		if (var)
			for (;;) {
				point += parse_string(1, "([ \t\n]|//.*)*");
				if (curchar() == '=')
					re_search(1, "![\n,;]");
				if (!parse_string(1, ",[ \t\n]*(<word>+)"))
					break;
				grab(start = find_group(1, 1), find_group(1, 0), func);
				point = matchend;
				if (!isdigit(*func))
					add_tag(func, start);
			}
		else
			c_skip_level('{', '}');
	}
}

tag_mode_latex()
{
	char tag[TAGLEN], pat[FNAMELEN * 2];
	int start, end;

	recolor_buffer_range(0, size());
	save_var case_fold = 0, point = 0;
	sprintf(pat, "\\(%s){([^{}]+)}", latex_tag_keywords);
	while (do_color_searching(REGEX, pat) == 1) {
		start = find_group(2, 1);
		end = find_group(2, 0);
		if (end - start > sizeof(tag) - 2)
			continue;
		grab(start, end, tag);
		add_tag(tag, start);
	}
}

tag_mode_python()
{
	char func[TAGLEN];
	int start, color;

	python_mode();
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	while (re_search(1, "^[ \t]*(def|class)[ \t]+([a-zA-Z_]<word>*)[(:]")) {
		color = get_character_color(matchstart);
		if (color == color_class python_comment
			  || color == color_class python_string)
			continue;
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

tag_mode_tcl()
{
	char func[TAGLEN];
	int start;

	save_var case_fold = 1;
	while (re_search(1, "^[ \t]*(proc)[ \t]*(" TCL_IDENT ")")) {
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

tag_mode_vbasic()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (re_search(1, "(^|%>)[ \t]*((public|private|static|default|shared"
					 "|widening|protected|friend|shadows|readonly|writeonly"
					 "|overloads|overrides|overridable|notoverridable"
					 "|mustoverride|narrowing)[ \t]+)*"
					 "(((sub|function|type|property([ \t]+(let|get|set))?"
					 ")[ \t]+([a-z_0-9]+))"
					 "|(operator((<wspace>*([-+*/\\&^=<>]|[<>][<>=]))|"
					 "(<wspace>+(IsFalse|IsTrue|Not|Ctype|And|Like|Mod|Or"
					 "|Xor)))))")) {
		if (tolower(character(find_group(4, 1))) == 'o')
			grab(start = find_group(10, 1), find_group(10, 0), func);
		else
			grab(start = find_group(9, 1), find_group(9, 0), func);
		add_tag(func, start);
	}
}

tag_mode_vhdl()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (do_color_searching(REGEX, "^[ \t]*<^c:vhdl-comment"
							  "|vhdl-string>(procedure|function|entity)"
							  "[ \t]+([a-zA-Z_]<word>*)")) {
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

when_loading()
{
	drop_name("have_gams_directive");  // Deleted subroutine.
	process_coloring_rules |= 64;	   // Adjust for new default values.
	reindent_c_comments |= 16;
	ssh_interpret_output &= ~0x10;
	telnet_interpret_output &= ~0x10;
	strcpy(scp_windows_sftp_command, "%w sftp");
	strcpy(ssh_command_windows, "%w ssh");
}
