/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2017 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13.14 to 13.15.

#include "eel.h"
#include "colcode.h"
#include "c.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "info.h"
#include "kill.h"
#include "search.h"
#include "session.h"
#include "proc.h"
#include "diff.h"
#include "tags.h"
#include "gams.h"
#include "hex.h"
#include "indent.h"
#include "perl.h"
#include "python.h"
#include "tcl.h"
#include "tex.h"

#define C_COMMENT_FILLER		"[^0-9a-zA-Z{}()\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

do_fill_comment(int generic, int margin)
{
	int pbuf, width = 1000, i, old_lines, new_lines, orig_lines;
	int col, col2, orig = bufnum, type;
	char last_prefix[FNAMELEN], orig_prefix[FNAMELEN], pat[FNAMELEN], comm[30];

	save_var margin_right = margin;
	save_spot point;
	type = locate_c_comment();	// 1 if block comment, 2 if one-line
	if (!type)					// comment, or 0 if not a comment.
		error("Can't fill comment here, not in a comment.");
	point += parse_string(1, C_COMMENT_FILLER);	// Past 1st line's prefix.
	sprintf(pat, "[ \t]*%s[^0-9a-z_\n]*", generic ? get_comment_start() : "/[*/]");
	grab(point, point - parse_string(-1, pat), orig_prefix);
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (!generic)
		point -= parse_string(-1, "<*>/");
	save_var narrow_end = size() - point + 1;

	if (type == 1 && !generic)		// Combine adjacent block comments.
		simple_re_replace(1, "[ \t]*<*>/[ \t]*\n?[ \t]*/<*>[ \t]*", " ");
	point = 0;
	while (nl_forward()) {	// Find width of narrowest prefix.
		i = parse_string(1, generic ? get_comment_start() : C_COMMENT_PREFIX);
		if (i > 0 && point > 0 && matches_at(point + i - 1, 1, "<^word|wspace><alnum>"))
			i--;		// Don't mess up e.g. <varname> at start of line.
		col = get_column(point + i);
		width = MIN(width, col);
	}
	pbuf = tmp_buf();
	point = 0;
	get_comment_marker(comm, generic);
	sprintf(pat, "[ \t]*%q", comm);
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		if (col > width)	// Make sure it's not too wide.
			col = width;
		if (type == 2) {	// But if it's made of one-line comments,
			i = parse_string(1, pat);
			col2 = get_column(point + i);
			if (col < col2)		// each prefix must include a starting delim like "//".
				col = col2;
		}
		i = column_to_pos(col);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	if (width == 1000)	// No prefix, so use default.
		width = 3;
	orig_lines = old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 20),
						   MAX(margin_right - width, 20));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;
	*last_prefix = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		grab(point, give_end_line(), last_prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	if (orig_lines < 2) {		// Default for new lines if 1 line originally.
		strcpy(last_prefix, orig_prefix);
		if (type == 2 && !strstr(last_prefix, comm))
			sprintf(last_prefix, "%s%s", comm,
					(*comm && comm[strlen(comm) - 1] == ' ') ? "" : " ");
		else if (type == 1 && !generic)
			if (!strstr(last_prefix, "/*"))
				strcpy(last_prefix, (auto_fill_comment_rules & 4) ? " " : " * ");
			else {
				replace_in_string(last_prefix, '/', ' ');
				if (fill_c_comment_plain)
					replace_in_string(last_prefix, '*', ' ');
			}
	}
	bufnum = orig;
	narrow_start = 0;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(last_prefix);			// Add some prefixes.
		if (generic) {
			save_var indenter_action = IACT_FILL_COMMENT;
			to_begin_line();
			if (indenter)
				(*indenter)();
		}
		nl_forward();
	}
	buf_delete(pbuf);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}


load_tags(file)
char *file;
{
	int i;

	save_var synchronous_file_io = 1;
	tags_from_bsc = 0;
	*_last_tag = *_last_tag_file = 0;
	if (!strfcmp(get_extension(file), ".bsc")) {
		tags_from_bsc = 1;
		strcpy(bsc_file, file);
		absolute(bsc_file);
		return;
	}
	save_var bufnum = _tags_buffer = zap("-tags");
	i = read_file(file, FILETYPE_AUTO);
	set_read_only(0);
	if (i == 2)			/* new file */
		mark_tags(0);		/* make new-fmt tags file */
	else if (i) {			/* error reading file */
		restore_vars();
		buf_delete(_tags_buffer);
		quick_abort();
	} else if (!tags_header()) {
		if (!ask_yn("Convert", "Convert old-format tags file? ",
					"Yes", "No", 1))
			quick_abort();
		sayput("Converting old-format tags file...");
		if (!string_replace(";", "\t", 0)) {
			mark_tags(0);
			do_save_tags();
		}
	}
}
