/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B9 to 13 B10, assuming further updaters.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "session.h"
#include "dired.h"
#include "grep.h"
#include "html.h"
#include "indent.h"
#include "vspell.h"
#include "proc.h"

dired_delete_hierarchy(char *dir)
{
	char pat[FNAMELEN], tail[FNAMELEN], *s;
	int cnt = -1, last;

	strcpy(pat, dir);
	add_final_slash(pat);
	save_var file_pattern_rules |= FPAT_SKIP_DIR_SYMLINKS
								   | FPAT_SKIP_RECUR_SYMLINKS;
	save_var file_pattern_wildcards = 0;
	strcat(pat, "**/*");
	do {	// Delete everything that matches our pattern.
		s = do_file_match(pat, STARTMATCH | EXACTONLY);
		last = cnt;
		for (cnt = 0; s; s = do_file_match(pat, EXACTONLY), cnt++) {
			check_abort();
			strcpy(tail, get_tail(s, 0));
			drop_final_slash(tail);
			if (!strcmp(tail, ".") || !strcmp(tail, ".."))
				continue;
			if (get_file_read_only(s))
				set_file_read_only(s, 0);
			if (cnt % 10 == 0)
				noteput("Deleting %f ...", s);
			if (is_directory(s))
				rmdir(s);
			else
				delete_file(s);
		}	// Repeat until the number of matches stops decreasing.
	} while (cnt < last || last == -1);
	set_file_read_only(dir, 0);
	return rmdir(dir);
}

do_shell_execute(char *fname, char *op, char *descr)
{
	char dirname[FNAMELEN * 3], file[FNAMELEN * 3];

	strcpy(file, fname);
	absolute(file);
	strcpy(dirname, file);
	*get_tail(dirname, 1) = 0;
	if (is_remote_file(file) && !*op)
		op = "Open";
	if ( do_run_viewer(file, op, dirname) == 0)
		say("Could not %s for \"%f\"", descr, file);
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];
	int show_rel = 1;		// Show the relative path of the default.

	if (!inbufs) {
		if (is_dired_buf() && !inp_get_other_name)
			save_var inp_get_other_name = get_dired_buffer_pattern();
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory
			  || grep_default_directory == 3)
			show_rel = 0;
		build_prompt(pr, "In file(s)", _grep_files, 0, show_rel);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory)
			strcpy(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else {
		sprintf(pr, "In buffers matching pattern (default: this buffer): ");
		comp_read(filepat, pr, b_match,
				  ((preserve_filename_case & 4) ? FM_FOLD : 0), bufname);
		if (!*filepat)
			strcpy(filepat, bufname);
	}
	return 1;
}

do_html_indent(int manual)
{
	int region, start, i;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 1);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1)))
		return;
	save_var narrow_start = start;
	i = (1 << (region - 1 + (IS_ENTER_KEY(key) ? 8 : 0)));
	if (region > 0 && region <= HT_CSS && !(html_indenting_rules & i))
		region = 0;
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
		case HT_PHP:
			html_exclude_script_head();
			do_php_indent();
			return;
		case HT_PYTHON:
			html_exclude_script_head();
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			html_exclude_script_head();
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

int do_html_auto_show_delimiter()
{
	int res, dir, i;
	char tag[FNAMELEN];

	save_var point, case_fold = (sublanguage_code != SUBLANG_XML);
	save_var matchstart, matchend, abort_searching = 0;
	save_var narrow_start = 0, narrow_end = 0;
	init_auto_show_delimiter();   // Must do this first.
	if (html_in_tag() < 0)
		return (auto_show_html_delimiters & 1) != 0;
	if (!(auto_show_html_delimiters & 2))
		return (auto_show_html_delimiters & 1) != 0;
	point -= parse_string(-1, HTML_NAME);
	*highlight_area_start[0] = point;
	dir = character(point - 1) == '<' ? 1 : -1;
	if ((i = parse_string(-1, "%</?")) && parse_string(1, HTML_NAME)) {
		*highlight_area_end[0] = matchend;
		res = do_html_find_matching_tag(point - i, tag);
		if (res == -1)
			switch (html_is_start_tag(point - i, 1)) {
				case 1:
				case -1:
					html_show_mismatch(point, *highlight_area_end[0]);
					return 2;
				case 0:
					return (auto_show_html_delimiters & 1) != 0;
			}

		if (res < 0)
			return 2;
		point = res;
		if (!parse_string(-dir, HTML_NAME))
			return 2;
	} else
		return (auto_show_html_delimiters & 1) != 0;
	*highlight_area_start[1] = matchstart;   // Mark the far end.
	*highlight_area_end[1] = matchend;
	modify_region(SHOW_MATCHING_REGION, MRTYPE, REGNORM);
			 // Make the highlighting visible.
	return 3;
}

command html_close_last_tag() on html_tab[ALT('E')]
{
	if (html_in_tag() >= 0)
		error("Can't insert an end tag inside another tag.");
	int tmp = tmp_buf();
	do_list_element_nesting(tmp, 1);
	save_var bufnum = tmp;
	point = size();
	char msg[FNAMELEN];
	if (parse_string(-1, "[<]" HTML_NAME "[>]\n", msg)) {
		msg[strlen(msg) - 2] = 0;
		restore_vars();
		bprintf("</%s>", msg + 1);
		save_spot point;
		search(-1, "<");
		if (current_column() <= get_indentation(point))
			do_html_indent(1);
	} else if (parse_string(-1, ".*\n", msg)) {
		msg[strlen(msg) - 1] = 0;
		error("%s", msg);
	} else
		error("Not in any element.");
}

command html_list_element_nesting() on html_tab[ALT('i')]
{
	int b = zap("-list-element-nesting"), sz;

	iter = 0;
	do_list_element_nesting(b, 1);
	set_buf_point(b, 0);
	sz = buf_size(b);
	char *name = *filename ? filename : bufname;
	if (!want_gui_prompts || !is_gui)
		buf_set_character_color(b, 0, size(), 0);
	buf_printf(b, "Element nesting in %s%s:\n",
			   strlen(name) > 20 ? "\n" : "", name);
	save_var _view_title = "Element Nesting";
	if (sz)
		view_buf(b, 1);
	else
		say("Not in any element.");
}

align_by_pattern(int bit, int start, int end, char *pat, int extra)
{
	int max = 0, i, cnt = 0;
	if (bit && !(align_region_rules & bit))
		return;
	save_var point, narrow_end = size() - end;
	point = start;
	to_begin_line();
	start = point;
	while (do_color_searching(REGEX, pat) == 1) {
		point = find_group(1, 0);
		if (align_region_rules & 1024)
			show_text(-2, 0, "Rule %d triggered on line %d at column %d "
					  "(adjusted to %d).",
					  bit, lines_between(0, point) + 1, current_column(),
					  get_column(point - parse_string(-1, "[ \t]*")));
		point -= parse_string(-1, "[ \t]*");
		i = current_column();
		to_end_line();
		cnt++;
		if (i > max)
			max = i;
		check_abort();
	}
	if (cnt < 2)
		return;
	point = start;
	while (do_color_searching(REGEX, pat) == 1) {
		point = find_group(1, 0);
		indent_to_column(max + extra + align_region_extra_space);
		to_end_line();
		check_abort();
	}
}

c_mode_aligner()
{
	// Align variable names being defined.
	ALPAT(4, "((" C_IDENT "[ \t]+)+)([*(]*" C_IDENT "[ \t]*[,;=].*)", 2);
	ALPAT(1, "((" C_IDENT "|[][ \t])+)(=.*)", 1); // Align = in assignments.
	ALPAT(8, "(#define[ \t]+" C_IDENT "(%((" C_IDENT
		  "|[, \t])*%))?[ \t]+)([^ \t\n]+)", 2);	// Align definitions.
	// Align comments.
	ALPAT(2, "(<^c:c-comment>.*)(<c:c-comment>/[/*].*)", 2);
}

MOUSE_CURSOR pan_neutral, pan_up, pan_down;

command mouse_pan()
{
	TIMER ts;
	int rate, per_elem, dir, lines, t, repaint_speed = 1;
	int before, we_moved = 0;

//	clip_mouse();
//	to_mouse_window(mouse_x, mouse_y, mouse_screen);

	save_var mouse_cursor;
	save_var mouse_panning = 1;
	save_var mouse_mask |= MASK_UP | MASK_MOVE;
	save_var permit_window_keys |= PERMIT_WHEEL_KEY;

	time_begin(&ts, 0);
	for (;;) {
		if (char_avail()) {		// If a key or mouse event arrives, stop.
			getkey();			// But if we've clicked the mouse button
			if (key != MOUSE_MOVE && (!IS_MOUSE_UP(key) || we_moved))
				break;			// without moving, continue (auto-scroll).
		}		// Otherwise, exit (unless it's just mouse movement).

		rate = mouse_panning_rate(40, 2000, 1); // Screen range, max, min.
						// Get scroll speed (lines/millisec).
		if (!rate) {					// Don't scroll right now.
			mouse_cursor = &pan_neutral;
			delay(1, COND_KEY);
			continue;
		} else if (rate > 0)
			mouse_cursor = &pan_down;
		else
			mouse_cursor = &pan_up;
		we_moved = 1;		// Distinguish panning from auto-scroll.

		per_elem = (rate > 0) ? rate : -rate;
		dir = (rate > 0) ? 1 : -1;
		if (per_elem > repaint_speed) {		// Scroll slower than we could,
			lines = 1;						// even going line by line.
			t = per_elem - repaint_speed;
		} else {		// Faster scroll--do multiple lines at once.
			lines = (repaint_speed + per_elem - 1) / per_elem;
			t = per_elem * lines - repaint_speed;
		}
		if (time_remaining(&ts) > t / 10)	// Moved to a faster rate,
			time_begin(&ts, t / 10);		// so skip some delaying.

		if (!time_done(&ts)) {		// Time for the next scroll?
			delay(1, COND_KEY);
			continue;
		}
		before = time_ms();
		window_scroll(dir * lines);	// Now scroll, then prepare to delay.
		set_display_func_name();
		redisplay();
		time_begin(&ts, t / 10);
		repaint_speed = time_ms() - before;	// Time taken for one refresh.
	}
}

command process_enter() on process_tab['\r']
{
	int set_err = 1;

	if ((process_enter_whole_line & 3) == 2 && point < type_point) {
		char line[FNAMELEN];

		to_begin_line();
		point += parse_string(1, "[^>$#%]*[>$#%] *");
		parse_string(1, ".*", line);
		point = size();	// grab current line excluding prompt
		stuff(line);	// and copy to the end
		return;
	}
	if ((process_enter_whole_line & 3) && point >= type_point)
		to_end_line();
	if (parse_string(-1, "Terminate batch job.*"))
		set_err = 0;
	enter_key();

// Pressing Enter should usually set the error spot.  When shouldn't it?
// 1. When you've pre-typed the next compilation command in the middle
// of a previous compile?
// 2. When you've just hit enter without typing anything?
//	if (point > type_point + 1 && (spot_to_buffer(error_spot) != bufnum
//								   || point < *error_spot))
// 3. When you're confirming aborting a batch file in Windows.

	if (set_err)
		set_error_spot(), (*error_spot)--;
	_last_was_grep = 0;
	if (point >= window_end && process_output_to_window_bottom)
		line_up_to_bottom();
}

process_get_previous_cmd(int down)
{
	int buf = bufnum;
	char cmd[16384], hist_buf[FNAMELEN];

	sprintf(hist_buf, "-process-history-%s", bufname);
	point = size();
	if (!exist(hist_buf))
		return;
	*cmd = 0;
	if (has_arg) {
		iter = 0;
		save_var bufname = hist_buf;
		if (point == size())
			point--;
		to_begin_line();
		restore_vars();
		if (!get_choice(name_to_bufnum(hist_buf), cmd,
						"Previous commands", "Select a previous command",
						"OK", "Cancel", ""))
			return;
	}
	if (spot_to_buffer(process_input_starts_at) <= 0)
		process_input_starts_at = alloc_spot();
	if (lines_between(*process_input_starts_at, point))
		*process_input_starts_at = type_point;
	if (!lines_between(*process_input_starts_at, point))
		delete(*process_input_starts_at, point);
	else
		*process_input_starts_at = point;
	if (*cmd)
		stuff(cmd);
	else {
		save_var bufname = hist_buf;
		down_or_up_lines(down);
		buf_xfer(buf, give_begin_line(), give_end_line());
	}
	restore_vars();
	if (process_enter_whole_line & 4)
		point = *process_input_starts_at;
}

session_save_buffer(int b, char *sdir)
{
	char file[FNAMELEN];
	int transl = GET_LINE_TRANSLATE(translation_type);
	int uni_compat = 0;

	if (translation_type & FORCED_TRANS)
		transl = GET_LINE_TRANSLATE(translation_type);
	if (has_feature & FEAT_UNICODE) {
		char *s = encoding_to_name(GET_ENCODING(translation_type));
		if (s)
			uni_compat = strstr(s, "UTF-16") != 0;
		if (uni_compat)
			transl = FILETYPE_BINARY;
	}
	buf_printf(b, "%s\n", bufname);
	if (is_dired_buf()) {
		insert_char_in_buf(b, 0);	// Hide file pattern from old versions.
		buf_printf(b, "%s\n", get_dired_buffer_pattern());
	} else {
		relative(filename, file, sdir);
		buf_printf(b, "%s\n", file);
	}
	buf_printf(b, "%d %d %d %d %d %d %d",
			   point, mark, transl,
			   is_dired_buf(), dired_buffer_pattern_wildcards,
			   dired_sorts_files, force_save_as);
	if ((has_feature & FEAT_UNICODE) && (translation_type & FORCED_TRANS))
		buf_printf(b, " %d @%s", GET_LINE_TRANSLATE(translation_type),
				   encoding_to_name(GET_ENCODING(translation_type)));
	buf_printf(b, "\n");
	buf_printf(b, "%d %d %d %d\n", over_mode, fill_mode,
			   indent_with_tabs, margin_right);
	transl = FILETYPE_AUTO;
	if (uni_compat || (translation_type & FORCED_TRANS))
		transl = GET_LINE_TRANSLATE(translation_type);
	// This flavor of the translation type is used by non-Unicode/
	// pre-v12 versions only.
	buf_printf(b, "%d %d %d %d %d %d %d %d %d %d\n", tab_size,
			   case_fold, 0, want_warn,
			   _buf_readonly, buf_accessed,
			   (file_io_converter == oem_file_converter),
			   uni_compat,
			   transl, soft_tab_size);
}

int session_restore_one_file(int b, char *file, char *sdir, int *readerrs)
{
	int i, trans = FILETYPE_AUTO;
	int num1[10], num1cnt, num2[10], num2cnt, num3[10], num3cnt;
	char firstline[150];
	jmp_buf this_level;

	grab_line(b, firstline);
	if ((num1cnt = break_into_numbers(firstline, num1)) < 3)
		return 2;
	if ((num2cnt = grab_numbers(b, num2)) < 4)
		return 3;
	if ((num3cnt = grab_numbers(b, num3)) < 5)
		return 4;
	if (session_restore_files)
		filename = file;
	save_var top_level = &this_level;
	if (setjmp(top_level))
		return -1;
	if (num1cnt >= 4 && num1[3] && session_restore_files) {	// restore dired buf
		filename = "";
		if (num1cnt >= 5)
			save_var file_pattern_wildcards = num1[4];
		save_var show_status = 0;
		if (num1cnt >= 6)
			dired_sorts_files = num1[5];
		create_dired_listing(file, 0);
		if (is_remote_file(get_dired_buffer_pattern()))
			dired_orig_position = -1;
	} else if (*file && session_restore_files) {
		if (!is_remote_file(file))
			absolute(file, sdir);
		if (has_feature & FEAT_UNICODE) {
			if (num1cnt >= 7) {		 // If explicit trans, force.
				char *p;
				if (p = strchr(firstline, '@')) {
					i = encoding_from_name(p + 1);
					if (i)
						trans = MAKE_TRANSLATE(i, num1[6], 1);
				}
			}
		} else {
			if (num3cnt > 8 && num3[7])
				trans = num1[2];	// If explicit trans, or common unicode, force.
		}
		i = do_file_read(file, trans);
		force_save_as = 0;
		if (num1cnt >= 7)
			force_save_as = num1[6];
		if (user_abort || i == EREADABORT)
			return 0;
		if (!i || i == 2) {
			if (!i)
				set_readonly_from_file();
			window_bufnum = bufnum;		// make suffix- kbd macros work
			call_mode(filename);
			make_mode();
		} else if (!*readerrs++) {
			file_error(i, file, "read error");
			maybe_ding(bell_on_read_error);
		} else
			show_text(0, -1, "%d files could not be read", *readerrs);
	}
	if (session_restore_files) {
		if (bufnum != b)	/* same tmp name as session buffer */
			point = num1[0];
		mark = num1[1];
		if (is_dired_buf())
			move_to_column(goal_column);
	}
	if (session_restore_files) {
		session_restore_one_file_1(num2, num2cnt);
		session_restore_one_file_2(num3, num3cnt);
	}
	return -1;
}

command spell_configure()
{
	char dir[FNAMELEN], file[FNAMELEN], *arch, archpath[FNAMELEN];
	char url[FNAMELEN], other[FNAMELEN], cmdline[FNAMELEN], *s;
	if (exist(WORD_LIST_BUFFER))
		reload_spell_word_list();
	save_var want_gui_prompts = 0;
	get_customization_directory(dir);
	absolute(dir);
	strcat(dir, "/spell");
	mkdir(dir);
	if (check_file(dir) != CHECK_DIR)
		error("Couldn't create working directory %f", dir);
	sprintf(file, "%s/final/english-words.10", dir);
	if (check_file(file) != CHECK_FILE) {		// Download and extract.
		arch = is_win32 ? "spell.exe" : "spell.tgz";
		int sz = is_win32 ? 2_398_133 : 2_258_676;
		struct file_info fi;
		sprintf(archpath, "%s/%s", dir, arch);
		sprintf(url, "http://www.lugaru.com/%s", arch);
		s = lookpath(arch);
		if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9) {
			if (s && check_file(s, &fi) == CHECK_FILE && fi.fsize * 10 >= sz * 9)
				strcpy(archpath, s);
			else if (!ask_yn("Download and install dictionary file?", 
							 "Download and install a dictionary file "
							 "(about 2.3 MB)? ", "Yes", "No", 1))
					return;
			else {
				say("Please wait...");
				if (download_file_to_disk(url, archpath, sz))
					error("Couldn't download dictionary file from %s", url);
			}
		}
		if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9)
			error("Bad dictionary archive %f, please retrieve manually.",
				  archpath);
		convert_to_8_3_filename(archpath);
		sprintf(cmdline, is_win32 ? "%s" : "tar xzf %s", archpath);
		pipe_text(NULL, "-spell-config", cmdline,
				  dir, PIPE_SYNCH | PIPE_NOREFRESH, 0);
	}
	if (check_file(file) != CHECK_FILE)
		error("Couldn't extract files from dictionary archive.");
	
	int region = get_key_response("Use American, British, British with -ize, "
								  "or Canadian spellings [ABZC]? ",
								  "ABZC", 0, "spell-configure");

	int help = tmp_buf(), w = add_popup(10, 10, 60, 10, _view_border, help);
	on_exit buf_delete(help), remove_window(w);
	window_title(w, BTOP, TITLECENTER,  " Spell Configuration Help ");

	show_popup_help(w, "Select a dictionary size.  Bigger dictionaries "
					" recognize rarer words, but won't catch misspellings "
					"that result in rare words.");
	int dictsz = get_key_response("Dictionary size (small, medium, large) "
								  "[SML, default L]? ", "SML", 'L',
								  "spell-configure");

	save_var insert_default_response = 1;
	show_popup_help(w, "Should the dictionary include abbreviations, "
					"contractions, uppercase words, or proper names?  Enter "
					"A, C, U, and P, in any combination, to select these.\n\nThe "
					"Abbreviations category includes abbreviations and acronyms "
					"like \"TV\", \"VIP\", and \"rpm\".\n\nThe Contractions "
					"category includes \"doesn't\" and \"I'm\".\n\nThe Uppercase "
					"category includes words that must start with an uppercase "
					"letter, like \"Doctor\", \"December\", or \"Europe\".\n\n"
					"The Proper Names category includes less common names like "
					"\"Brian\" or \"Altman\" that are not included in the "
					"Uppercase category.\n\nSome words appear in more than one "
					"category.");
	get_strdef(other, "Include Abbrevs, Contractions, "
			   "Uppercase, Proper names in dictionary", "ACUP");

	show_popup_help(w, "Should the speller permit variant spellings, instead "
					"of permitting only a single preselected spelling of each "
					"word?\n\nThis permits words like \"fiascos\" in addition to "
					"\"fiascoes\" or \"Hanukah\" in addition to \"Hanukkah\".");
	sprintf(other + strlen(other), "%c",
			get_key_response("Permit variant spellings (None, Some, Many, Lots) "
							 "[Some]? ", "NSML", 'S', "spell-configure"));
	spell_configure_helper(w);
	save_var bufnum = zap("-spell-config");
	switch (dictsz) {
		case 'L':
			grab_spell_dict_size("70", region, other);
			grab_spell_dict_size("60", region, other);
			grab_spell_dict_size("55", region, other);
		case 'M':
			grab_spell_dict_size("50", region, other);
			grab_spell_dict_size("40", region, other);
		case 'S':
			grab_spell_dict_size("35", region, other);
			grab_spell_dict_size("20", region, other);
			grab_spell_dict_size("10", region, other);
	}
	sprintf(file, "%s/../%s", dir, WORD_ORDER_FILE);
	int err = file_write(file, FILETYPE_AUTO);
	if (err)
		file_error(err, file, "write error");
	case_fold = 0;
	do_sort_region(0, size(), 0, 0);
	sprintf(file, "%s/../%s", dir, WORD_FILE_NAME);
	err = file_write(file, FILETYPE_AUTO);
	if (err)
		file_error(err, file, "write error");
	restore_vars();
	reload_spell_word_list();
	say("Dictionary configured, run spell-mode or spell-buffer.");
}

int run_spell_helper_program()
{
	save_var bufnum = create("-spell-helper");
	point = size();
	if (is_process_buffer(bufnum))
		return bufnum;
	char prog[FNAMELEN];
	expand_string_template(prog, spell_helper_program, "", NULL);
	convert_to_8_3_filename(prog);
	concur_shell("", prog, NULL, bufname);
	spell_wait_for_response("^@");
	return bufnum;
}

when_loading()
{
	xml_auto_indent = html_auto_indent;
}
