/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B10 to 13 B11, assuming further updaters.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "session.h"
#include "dired.h"
#include "grep.h"
#include "html.h"
#include "indent.h"
#include "vspell.h"
#include "proc.h"
#include "epsnet.h"
#include "vbasic.h"
#include "perl.h"
#include "python.h"
#include "vhdl.h"
#include "tags.h"
#include "tex.h"
#include "chelp.h"

color_asm_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	while (from > 0 && get_tagged_region("color-as-unit", from, &s) != -1
		   && s <= from)
		from = s - 1;
	c_init_color(from, to);
	point = from;
	to_begin_line();
	if (talk = (to - from > 50000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.@#%?$]+|[\"';]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case ';':			// found comment
				nl_forward();
				set_character_color(t, point, color_class c_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, asm_keyword_color(t));
				if (asm_title_line()) {
					t = point;
					point++;
					nl_forward();
					set_character_color(t, point, color_class c_string);
				}
				asm_comment_line();
				break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring ASM program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

do_reindent_previous_line()
{
	int i = get_mode_variable("%s-reindent-previous-line");

	if (got_bad_number)
		i = default_reindent_previous_line;
	if (!i)
		return;
	save_spot point;
	nl_reverse();
	to_indentation();
	save_var indenter_action = IACT_REINDENT_PREV;
	if (curchar() != '\n')
		(*indenter)();
}

/* Delete trailing whitespace, then break line here and indent it. */
break_and_indent_line()
{
	delete(point, point - parse_string(-1, "[ \t]*"));
	insert('\n');
	do_reindent_previous_line();
	save_var indenter_action = IACT_AUTO_INDENT;
	(*indenter)();
}

command goto_line()
{
	char res[FNAMELEN], *p;
	int num = iter, percent = 0;

	got_bad_number = iter = 0;
	if (has_arg) {
		go_line(num);
		return;
	}
	get_string(res, "Go to line: ");
	if (tolower(*res) == 'p') {
		num = exptoi(res + 1);
		if (got_bad_number)
			error("Bad position %s.", res + 1);
		point = num;
		return;
	}
	if (res[0] && *(p = res + strlen(res) - 1) == '%') {
		percent = 1;
		*p = 0;
	}
	num = exptoi(res);
	if (!got_bad_number) {
		if (percent) {
			point = muldiv(size(), num, 100);	// handles intermediates > 32 bits
			if (num)
				nl_forward();
		} else
			go_line(num);
	}
	p = strchr(res, ':');		// a number after : or . is a column
	if (!p)
		p = strchr(res, '.');
	if (p) {
		num = exptoi(p + 1);
		if (!got_bad_number)
			to_column_and_center(num);
	}
}

command show_point()
{
	char msg[300], cname[300], regmsg[300];
	int sz = MAX(point - mark, mark - point), c;

	*regmsg = 0;
	if (is_highlight_on())
		sprintf(regmsg, ", %'d char%s in region", sz, (sz != 1) ? "s" : "");
	if (point < size()) {
		sprintf(msg, "Column %'d, char %'d of %'d is ", virtual_column(),
				point, size());
		if (c = curchar())
			sprintf(msg + strlen(msg),
				"'%c'=%d decimal=%02x hex", c, c, c);
		else
			sprintf(msg + strlen(msg), "'\\0'");
		if (*regmsg)
			strcat(msg, regmsg);
		else if (c > 127) {
			*cname = 0;
			show_char(cname, c, 1);	// Show full name first, if it all won't fit.
			if (strlen(msg) + strlen(cname) + 2 >= screen_cols)
				show_text(0, MAX(see_delay, 200), "'%c'=%s", c, cname);
			sprintf(msg + strlen(msg), ", %s", cname);
		}
	} else
		sprintf(msg, "Column %'d, at end of %'d chars%s",
				virtual_column(), size(), regmsg);
	say("%s", msg);
	expire_message = 1;
	iter = 0;
}

color_c_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	char pat[200];
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
		       "|-?%.?[0-9]([A-Za-z0-9._]|[Ee]-)*");
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
		case '/':			// found // one-line comment
			while (nl_forward() && character(point - 2) == '\\')
				;		// Continued // comment.
			set_character_color(t, point, color_class c_comment);
			break;
		case '*':			// found /* starting comment
			search(1, "*/");
			set_character_color(t, point, color_class c_comment);
			if (get_character_color(point, (int *) 0, &s)
				  == color_class c_comment)
				to = MAX(s, to);	// Comment extended further before; rescan it.
			break;
		case '#':			// found preproc line
			c_preproc_color();	// move past correct part
			set_character_color(t, point,
					    color_class c_preprocessor);
			break;
		case '"':		// found a string literal
			point = t;
			if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
				t--;
			if ((c_extra_keywords & CSHARP_KEYWORDS)
				  && character(t - 1) == '@')
				t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
			else
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
			set_character_color(t, point, color_class c_string);
			if (get_character_color(point, (int *) 0, &s) == 
			    color_class c_string && s > to)  // fix up after
				c_init_color(point, to = s); // quoted "'s
			break;
		case '\'':		// found a char const
			point = t;
			re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
			set_character_color(t, point, color_class c_charconst);
			break;
		default:		// found identifier, kywd, or number
			if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
				t--;
			set_character_color(t, point, c_keyword_color(t));
			if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
				&& (point - t == 4)			// uuid's in IDL files.
				&& parse_string(-1, "uuid")
				&& parse_string(1, "%([0-9A-Fa-f]"))
				maybe_color_uuid();
			break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

telnet_send_interrupt()
{
	do_send_invisible("\03", (mode_keys == ssh_tab));
}

command visit_file()
{
	char tmp[FNAMELEN];

	iter = 0;
	save_var show_mouse_choices = 1;
	get_file(tmp, "Visit file", force_save_as == FSA_NEWFILE ? "" : filename);
	if (!*tmp)
		return;
	if (is_process_buffer(bufnum)
		  && !ask_yn("Confirm replacing buffer contents", "Are you sure "
					 "you want to begin executing the lines of this file? ",
					 "Yes", "No", 2))
		return;
	if (!has_arg && modified) {	/* buffer need to be saved? */
		if (ask_save_buffer())
			do_save_file(1, 1, 1);
	}
	read_file(tmp, reading_translation_type());
	do_file_read_kibitz();
}

// Is this buffer unsaved?
int is_unsaved_buffer()
{
	if (discardable_buffer || buffer_not_saveable)
		return 0;
	if (modified && *filename && !is_dired_buf())
		if (size() || check_file(filename) == CHECK_FILE)
			return 1;
	if (ftp_job && ftp_job->buf == bufnum && ftp_job->operation == FTP_SEND)
		return 1;
	return 0;
}

// Find a file using default translation.
std_find_it(char *fname)
{
	int ret = find_it(fname, get_default_translation_type(fname));
	if (!ret)
		do_file_read_kibitz();
	return ret;
}

#define C_COMMENT_FILLER		"[^0-9a-zA-Z\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

do_fill_comment(int generic, int margin)
{
	int pbuf, width = 1000, i, old_lines, new_lines, orig_lines;
	int col, col2, orig = bufnum, type;
	char last_prefix[FNAMELEN], orig_prefix[FNAMELEN], pat[FNAMELEN], comm[30];

	save_var margin_right = margin;
	save_spot point;
	type = locate_c_comment();	// 1 if block comment, 2 if one-line
	if (!type)					// comment, or 0 if not a comment.
		error("Can't fill comment here, not in a comment.");
	point += parse_string(1, C_COMMENT_FILLER);	// Past 1st line's prefix.
	sprintf(pat, "[ \t]*%s[^0-9a-z_\n]*", generic ? get_comment_start() : "/[*/]");
	grab(point, point - parse_string(-1, pat), orig_prefix);
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (!generic)
		point -= parse_string(-1, "<*>/");
	save_var narrow_end = size() - point + 1;

	if (type == 1 && !generic)		// Combine adjacent block comments.
		simple_re_replace(1, "[ \t]*<*>/[ \t]*\n?[ \t]*/<*>[ \t]*", " ");
	point = 0;
	while (nl_forward()) {	// Find width of narrowest prefix.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		width = MIN(width, col);
	}
	pbuf = tmp_buf();
	point = 0;
	get_comment_marker(comm, generic);
	sprintf(pat, "[ \t]*%s", comm);
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		if (col > width)	// Make sure it's not too wide.
			col = width;
		if (type == 2) {	// But if it's made of one-line comments,
			i = parse_string(1, pat);
			col2 = get_column(point + i);
			if (col < col2)		// each prefix must include a "//".
				col = col2;
		}
		i = column_to_pos(col);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	if (width == 1000)	// No prefix, so use default.
		width = 3;
	orig_lines = old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 20),
						   MAX(margin_right - width, 20));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;
	*last_prefix = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		grab(point, give_end_line(), last_prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	if (orig_lines < 2) {		// Default for new lines if 1 line originally.
		strcpy(last_prefix, orig_prefix);
		if (type == 2 && !strstr(last_prefix, comm))
			sprintf(last_prefix, "%s ", comm);
		else if (type == 1 && !generic)
			if (!strstr(last_prefix, "/*"))
				strcpy(last_prefix, (auto_fill_comment_rules & 4) ? " " : " * ");
			else {
				replace_in_string(last_prefix, '/', ' ');
				if (fill_c_comment_plain)
					replace_in_string(last_prefix, '*', ' ');
			}
	}
	bufnum = orig;
	narrow_start = 0;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(last_prefix);			// Add some prefixes.
		if (generic) {
			save_var indenter_action = IACT_FILL_COMMENT;
			to_begin_line();
			if (indenter)
				(*indenter)();
		}
		nl_forward();
	}
	buf_delete(pbuf);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

comm_break_line_here(int c_like, int comm, int starts_with_comm, int starts_in_comm,
					 int comm_start, spot next_ends, spot after, char *fill)
{
	int offset, generic = strcmp(major_mode, "C");
	char rest[FNAMELEN], *new_comm;

	delete_horizontal_space();
	insert('\n');
	if (display_column > 0)	// Horizontally scroll back to left margin.
		display_column = 0;
	if (key == '\n')	// User might be trying to break line elsewhere as
		key = ' ';		// we're breaking it here.  Prevent a mess.

	if (comm == IN_NEW_COMMENT || !c_like) {		// A //-style C++ comment.
		new_comm = generic ? get_comment_begin() : "// ";
		if (starts_with_comm)	// This line began with a comment, so just
			stuff(fill);		// repeat its prefix (indentation and such).
		else if (next_ends && get_column(comm_start) > get_column(*next_ends)) {
			grab(point, give_end_line(), rest);	// Continue it on the next line
			offset = give_end_line() - *after;	// since there's no comment
			delete(point - 1, give_end_line());	// there.
			point = *next_ends;
			to_column(get_column(comm_start) - 2);
			bprintf("%s%s%s", isspace(character(point - 1)) ? "" : " ",
					new_comm, rest);
			*after = point - offset;	// Point might be in the middle of moved text.
		} else {			// Make a new line after the commented code line we
			stuff(new_comm);	// broke.
			point -= strlen(new_comm);
			save_var indenter_action = IACT_AUTO_FILL_COMMENT;
			if (c_like)
				indent_to_column(c_compute_indent());
			else if (indenter)
				(*indenter)();
		}
	} else {			/* Old-style K&R comment. */
		if (strstr(fill, "*/"))
			*fill = 0;
		if (starts_in_comm)		// Just repeat the *'s or whatever at start of
			stuff(fill);		// old line.
		else if (starts_with_comm) {
			if (!strcmp(fill, "/**"))	// A Javadoc comment.
				fill = " * ";
			replace_in_string(fill, '/', ' ');
			if (fill_c_comment_plain)	// Like above, but convert that first /*.
				replace_in_string(fill, '*', ' ');
			stuff(fill);
		} else {
			delete(point - 1, point);	// Delete the \n we just inserted.
			stuff(" */\n/* ");			// And put this instead.
			to_begin_line();
			indent_to_column(c_compute_indent());
		}
	}
}

break_line_here()	/* replace whitespace with \n and indent */
{
	delete_horizontal_space();
	insert('\n');
	save_var indenter_action = IACT_AUTO_FILL;
	if (auto_fill_indents && indenter)
		(*indenter)();
	if (display_column > 0)
		display_column = 0;
}

do_fill_indented_paragraph()
{
	int pbuf, width, i, old_lines, new_lines, orig = bufnum;
	char prefix[FNAMELEN];

	save_spot point;
	point--;
	forward_paragraph();
	mark = point - (point < size() - narrow_end);
	backward_paragraph(); 
	point += parse_string(1, "([ \t]*\n)+");
	parse_string(1, "[ \t]*", prefix);
	width = current_column();
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (character(point - 1) == '\n')
		point--;
	save_var narrow_end = size() - point;

	point = 0;
	while (nl_forward())	// Find width of narrowest indentation.
		width = MIN(width, get_indentation(point));
	pbuf = tmp_buf();
	point = 0;
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = column_to_pos(width);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 2),
						   MAX(margin_right - width, 2));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		if (point < give_end_line())
			grab(point, give_end_line(), prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	bufnum = orig;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(prefix);			// Add some prefixes.
		nl_forward();
	}
	buf_delete(pbuf);
}

command fill_indented_paragraph()
{
	iter = 0;
	if (has_arg)
		save_var margin_right = current_column() + 1;
	if (is_highlight_on()) {
		save_spot point, mark, narrow_start, narrow_end;
		fix_region();
		narrow_start = point;
		narrow_end = size() - mark;
		do {
			do_fill_indented_paragraph();
			forward_paragraph();
		} while (point < size() - narrow_end);
	} else {
		do_fill_indented_paragraph();
	}
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

int html_recolor_from_here(int safe)
{
	int start = point;

	if (sublanguage_code == SUBLANG_PHP)	// Can have <?'s in strings.
		return safe;		// Can't parse it backwards, so start from top.
	// Look for start or end of XML's CDATA block, but ignore the
	// marker if it might be part of a string or comment.
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>!")
		  && !parse_string(-1, "(['\"#]|//||\\\n).*"))
		return point;
	point = start;
	while (re_search(-1, sublanguage_code == SUBLANG_XML
				  ? "%<%!--|--[ \t\n]*>|%<%?php"
				  : "%<%!--|-[ \t\n]*>|%<[%?]|[%?]%>") && curchar() == '<')
		if (html_region_type(point) == -1)
			return point;
	return (sublanguage_code == SUBLANG_PHP) ? safe : start;
}

html_color_value(int from, int to)
{
	set_character_color(from, to, color_class html_value);
	save_var narrow_start = point = from, narrow_end = size() - to;
	int cnt = 0;
	while (re_search(1, "[<][%?]")) {
		if (!cnt++)
			set_tagged_region("color-as-unit", from, to, 99);
		// Force recoloring of rest of attribute.
		set_tagged_region("needs-color", point, to, -1);
		color_html_include(-1);
	}
}

color_html_include(int limit)
{
	int s, t, c, lang, orig = point, stop;
	char *end, *name;

	search(-1, "<");	// Find start of include.
	c = character(point + 1);
	s = point;
	switch (c) {
		case '%':	if (parse_string(1, "%<%%--")) {
						search(1, "--%>");	// Found a JSP "hidden comment".
						set_character_color(s, point, color_class html_comment);
						return 1;
					}
					point += parse_string(1, "%<%%[a-z0-9]*");
					end = "%>";
					lang = get_html_asp_language(sublanguage_code
						== SUBLANG_XML ? xml_asp_coloring : html_asp_coloring);
					break;
		case '?':	if (sublanguage_code == SUBLANG_XML)
						lang = xml_php_coloring;
					else if (parse_string(1, "%<%?php"))
						lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
					else if (parse_string(1, "%<%?python"))
						lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
					else
						lang = html_php_coloring;
					point += parse_string(1, "%<%?[a-z0-9]*");
					end = "?>";
					break;
		default:	point++;
					end = ">";
					lang = get_html_asp_language(html_asp_coloring);
					break;
	}
	if (lang == HT_IGNORE) {
		point = orig;
		return 0;
	}
	t = point;
	// First mark whole section embedded.
	set_character_color(s, point, color_class html_embedded);
	set_tagged_region("color-as-unit", s, point, 99);
	for (;;) {
		stop = search(1, end) ? matchstart : point;	// Find the end marker.
		color_embedded_code(t, point, lang, limit);	// Color it too for now.
		if (c == '%' && lang == HT_VBSCRIPT)
			break;	// Don't do this for comment in ASP vbscript.
		// If it wound up being colored as a comment or string, that's
		// not really the end, so find the next marker.
		name = name_color_class(get_character_color(stop, &t));
		if (point <= stop || !name || !strstr(name, "comment")
			  && !strstr(name, "string"))
			break;
		// Force recoloring of that entire comment/string.
		set_tagged_region("needs-color", t, s = point, -1);
	}
	// Color the marker ending the embedded code.
	set_character_color(stop, point, color_class html_embedded);
	set_tagged_region("color-as-unit", stop, point, 99);
	return 1;
}

int color_html_range(int from, int to) // recolor just this section
{			// last colored region may go past to
	int i, j;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to) {
		set_character_color(from, to, color_class html_text);
		set_tagged_region("needs-color", from, to, -1);	// Say it needs coloring.
	}
	point = from;
	save_var case_fold = 1;
	color_html_move_back();
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &from) != -1)
		point = from;		// Move before script areas.
	if (to > size() - narrow_end)
		to = size() - narrow_end;
	set_tagged_region("color-as-unit", from, to, -1);
	from = point;
	for (;;) {
		set_tagged_region("html-coloring-postponed", 0, size(), -1); // Clear rgn.
		color_html_loop(to);
		to = 0;
		while (to < size() - narrow_end) { // Colored partial script block...
			if (get_tagged_region("html-coloring-postponed", to, &i, &j) == 1) {
				if (i >= to) {
					point = i;	// ...so invalidate rest of buffer.
					set_tagged_region("needs-color", point, size(), -1);
					set_tagged_region("color-as-unit", point, size(), 98);
					return point;
				}
			}
			if (j <= to)
 				break;
			to = j;
		}

		set_tagged_region("needs-color", from, point, 0);	// Say it's colored.
		i = get_tagged_region("color-as-unit", point, NULL, &to);
		if (i == -1 || i == 99 || point >= size() - narrow_end || to <= point)
			break;
		// The next text was colored as scripting, but it shouldn't be
		// any more, since we never stop coloring inside scripting.
		// Continue coloring it as HTML, and mark the rest of the buffer
		// to be recolored later.
		set_tagged_region("needs-color", point, size(), -1);
	}
	return point;
}

html_is_self_delimiting(int pos, char *color)
{
	save_var point = pos;
	char pat[FNAMELEN];
	sprintf(pat, "<c:%s>[>]", color);
	return do_color_searching(REGEX, pat) && parse_string(-1, "/[ \t\n]*[>]");
}

// Return indent of this HTML line by examining start and end tags.
int html_indent_level(int manual)
{
	char tag[FNAMELEN];
	save_var case_fold = (sublanguage_code != SUBLANG_XML);
	int ind = get_indentation(point), i, level = 0;
	i = html_in_tag();
	if (i >= 0)
		return html_indent_in_tag(i, ind);
	// Indent end tag if due to pressing tab, or to reindent previous
	// line option, but not when auto-fill moves an end tag down.
	if ((manual || indenter_action != IACT_AUTO_FILL)
		  && parse_string(1, "[ \t]*[<]/"))	// Indenting an end tag.
		return html_indent_of_matching_tag(matchend - 2, ind);
	if (sublanguage_code != SUBLANG_XML	// Special rules for some tags.
		  && parse_string(1, "[ \t]*%<(" HTML_NAME ")</" HTML_NAME_CHAR ">")) {
		grab(find_group(1, 1), find_group(1, 0), tag);
		if (!strfcmp(tag, "option") && (i = html_open_option(point)) >= 0)
			return get_indentation(i);
		if (!strfcmp(tag, "li") && (i = html_open_li(point)) >= 0)
			return get_indentation(i);
		if ((!strfcmp(tag, "dd") || !strfcmp(tag, "dt"))
			  && (i = html_open_dd_dt(point)) >= 0)
			return get_indentation(i);
		if (!strfcmp(tag, "colgroup"))
			level += html_colgroup_adjust();
	}
	save_var point;
	if (do_color_searching(REGEX | REVERSE, "<^c:html-comment>[^ \t\n]") != 1)
		return 0;
	save_var narrow_end = size() - give_end_line();
	to_indentation();
	while ((i = html_in_tag()) >= 0 && i < give_begin_line()) {
		point = i;	// Go to start of tag.
		to_indentation();
	}

	ind = current_column();
	level += html_tag_level();
	return ind + level * give_html_indent();
}

do_html_find_matching_tag(int pos, char *tag)
{
	char pat[FNAMELEN], *endcolor = "html-punctuation";
	int level, dir = 0, i;

	*tag = 0;
	save_var point = pos, case_fold = (sublanguage_code != SUBLANG_XML);
	i = parse_string(1, "%<[ \t\n]*");
	if (!i)
		return -2;
	point += i;
	if (curchar() == '/') {
		dir = REVERSE;
		level = 0;
		point++;
	} else
		level = 1;
	point += parse_string(1, "[ \t\n]");
	int color = get_character_syntax_color(point);
	if (color == color_class html_value)
		endcolor = name_color_class(color);
	if (!parse_string(1, "[-_./A-Za-z0-9:]+", tag))
		return -2;
	if (html_is_self_delimiting(point, endcolor))
		return -3;
	sprintf(pat, "%%<(/?<wspace>*<c:%s>%q)</word>",
			(color >= 0) ? name_color_class(color) : "-1", tag);
	while (do_color_searching(REGEX | dir, pat) == 1) {
		i = (character(find_group(1, 1)) == '/') ? -1 : 1;
		if (!html_is_self_delimiting(point, endcolor))
			if (!(level += i))
				break;
	}
	if (dir == REVERSE)
		point++;
	if (level)
		return -1;
	return point;
}

inner_php_func_name_finder()
{
	int from, to, orig = point;

	save_var point, case_fold = 1, narrow_end;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:*comment><^c:*string>"
							  "(function)</word>") == 1) {
		if (!must_find_func_name && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		if (!parse_string(1, "^[ \t]*(function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		narrow_end = size() - (orig + 1000);
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		move_level(1, "{", "}", 0, 0);
		if (give_end_line() < orig)
			*display_func_name = 0;
		narrow_end = 0;
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

to_mouse_position(x, y)	// Move to position for these screen coords,
{			// assuming we're in right window, & return nonzero
	int pt;		// if we could.

	x -= window_text_left(window_handle);
	y -= window_text_top(window_handle);
	pt = window_line_to_position(y);
	if (pt == -1) {			// try again, in case window's
		build_window();		// never been seen
		pt = window_line_to_position(y);
		if (pt == -1)
			return 0;		// must be on the border
	}

	point = pt;
	if (draw_line_numbers & 1)
		x -= line_number_width + 1;
	if (display_column < 0)
		save_var narrow_start = point;
	else
		x += display_column;
	if (virtual_space)
		to_virtual_column(x);
	else
		move_to_column(x);
	return 1;
}

select_loop(by_words, display_region)
{	/* select a buffer pos while pressing mouse button */
	int bgn, pt, mk, curcol, otype = -1;
	TIMER ts;

	bgn = mark;
	time_begin(&ts, 0);
	pt = point; mk = mark; curcol = cursor_to_column;
	save_var mouse_mask |= MASK_MOVE | MASK_UP;
	do {
		if (char_avail()) {
			getkey();
			if (key == GETFOCUS || key == LOSEFOCUS) {
				do_topkey();
				continue;
			}
		} else
			delay(1, COND_KEY);

		if (mouse_x < 0 && display_column <= 0)
			mouse_x = 0;
		if (in_text_area(mouse_x, mouse_y))
			point_to_pos(mouse_x, mouse_y, mouse_screen);
		else {
			if (!has_elapsed(&ts))
				continue;
			handle_scroll(mouse_x, mouse_y);
		}
		if (by_words)
			region_by_words(bgn);
		set_region_type((shift_pressed() & KB_ALT_DN)
				? REGRECT : REGNORM);
		if (point != pt || mark != mk || curcol != cursor_to_column
		    || region_type() != otype) {
			if (point != mark || curcol != cursor_to_column)
				if (display_region)
					highlight_on();
			refresh();
			mouse_display = 1;
			pt = point;
			mk = mark;
		}
		otype = region_type();
	} while (!IS_MOUSE_UP(key) && mouse_pressed());
	mouse_display = 0;
}

do_next_error(int count, char *cmd)	// return 0 if no errors, else 1
{
	char file[FNAMELEN], number[30], line[FNAMELEN], *p;
	int lineno, origpt, dir, from, to, is_bufname, file_paren, line_paren;
	int orig = bufnum, procbuf, start;

	tiled_only();
	_last_was_grep = 0;
	if (spot_to_buffer(error_spot) <= 0 && !exist(PROCBUF))
		error("No process buffer to read errors from.");
	tiled_only();
	if (spot_to_buffer(error_spot) > 0)
		bufnum = spot_to_buffer(error_spot);
	else
		bufname = PROCBUF;
	telnet_error_set_to_telnet();
	procbuf = bufnum;
	origpt = point;
	case_fold = 1;
	if (spot_to_buffer(error_spot) > 0)
		point = *error_spot;
	else {
		error_spot = alloc_spot();
		point = 0;
	}
	if (!count && procbuf == orig)
		point = origpt;

	noteput("Searching...");
	if (!count)
		find_error(-1, &file_paren, &line_paren), count = 1;
	dir = (count > 0) ? 1 : -1;
	for (; count; count -= dir) {
		if (!find_error(dir, &file_paren, &line_paren)) {
			point = origpt;
			bufnum = orig;
			if (_had_errors == 2)
				return 0;
			else if (_had_errors)
				say("No more errors.");
			else if (cmd)
				say("No errors running \"%s\".", cmd);
			else
				say("No errors.");
			return 0;
		}
	}
	_had_errors = 1;
	grab(find_group(file_paren, 1), find_group(file_paren, 0), file);
	grab(find_group(line_paren, 1), find_group(line_paren, 0), number);
	start = find_group(0, 1);
	from = find_group(line_paren, 0);
	to = find_group(0, 0);
	if (to - from < 20) {	// error message at end is too short
		from = start;	// so try including whole line
		if (to - from > screen_cols - 5) // Only rightmost if it's too big.
			from = to - screen_cols + 5;
	}
	if (to > from + sizeof(line) / sizeof(*line) - 1)
		to = from + sizeof(line) / sizeof(*line) - 1;
	grab(from, to, line);
	is_bufname = (parse_string(1, "Buffer ") != 0);
	if (dir > 0)
		to_end_line();
	*error_spot = point;
	point = origpt;
	lineno = strtoi(number, 10);
	bufnum = orig;
	if (!*file || !*number) {
		locate_window(bufnum_to_name(procbuf), "");
		point = *error_spot;
		to_begin_line();
		say("=> %s", line);
		return 1;
	}
	if (is_bufname)
		locate_window(file, "");
	else {
		process_find_directory(file, procbuf, from);
		if (!check_file(file) && !strcmp(get_extension(file), ".java"))
			proc_find_java_source(file, procbuf, start);
		locate_window("", file);
		if (process_view_error_lines
			  && window_number == number_of_windows() - 1)
			window_number = 0;
		std_find_it(file);
	}
	go_line(lineno);
	if (is_remote_file(file))
		ftp_go_to_line = lineno;
	for (p = line; *p && index("() :;,\t0123456789", *p); p++)
		;			// skip over these chars at start of msg
	say("=> %s", p);
	restore_vars();
	proc_view_error_lines(procbuf, start);
	return 1;
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
			   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "#?((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_variable);
				else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer ? character(t - 1) != '\\'
									: !strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && character(t - 1) == '\\')
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_>+)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_>+)"
									 : "(<alpha|_>+)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

to_comment_column()	/* replace whitespace here with comment indentation */
{
	int this;

	delete_horizontal_space();
	this = current_column();
	save_var indenter_action = IACT_COMMENT;
	if (!this && !comment_column)
		return;
	else if (this >= comment_column)
		stuff(" ");	/* code extends past comment col */
	else if (!this && indenter && indent_comment_as_code)
		(*indenter)();
	else
		insert_to_column(this, comment_column);
}

color_python_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		if (!re_search(1, "[A-Za-z_][A-Za-z0-9_]*"
					   "|-?%.?[0-9]([A-Za-z0-9._]|[Ee]-)*j?|[\"'#]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '#':			// found comment
				nl_forward();
				set_character_color(t, point, color_class python_comment);
				break;
			case '"':		// found a string literal
				python_string_color('"');
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class python_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\'':
				python_string_color('\'');
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class python_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, python_keyword_color(t));
				break;
		}
		if (talk && point > talk_at + 2000 && time_done(&talk_now)) {
			note("Coloring Python program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

remember_this_buffer()	/* should we record this buffer in session file? */
{
	int limit;

	if (*bufname == '-')
		return 0;
	if (is_process_buffer(bufnum) || buffer_not_saveable || _read_aborted)
		return 0;
	if (file_info.year && file_info.check_type != CHECK_FILE)
		return 0;
	if (is_dired_buf()) {
		if (!(session_restore_directory_buffers
			  & (is_remote_file(get_dired_buffer_pattern()) ? 2
				 : strstr(get_dired_buffer_pattern(), "**") ? 4 : 1)))
			return 0;
	} else {
		if (!*filename)
			return 0;
		limit = is_remote_file(filename) ? session_restore_biggest_remote_file
										 : session_restore_biggest_file;
		if (limit >= 0 && size() > limit)
			return 0;		// too big to restore
		if (!size() && check_file(filename) != CHECK_FILE)
			return 0;
	}
	return 1;
}

tag_mode_c()		/* tag all c functions in this file */
{
	int end;

	save_var case_fold = 0;	// don't confuse STRUCT with struct
	tag_modifier_kwd = 0;
	*c_tagging_class = 0;	// initially, not inside a class block
	if (old_tagging_class) {
		free(old_tagging_class);
		old_tagging_class = 0;
	}
	while (re_search(1, "([{};<]|/<*|/>|[a-zA-Z0-9_]+)")) {
		end = point;		/* find braces, ;, <, comment open, or ident */
		point = matchstart;
		switch (curchar()) {
			case '{':	skip_c_braces();
						break;
			case '<':	skip_cpp_template();
						break;
			case '}':	tag_remove_class();
						point++;
						break;
			case ';':	tag_modifier_kwd = 0;	// assume next var
						point++;		// is being defined
						break;
			case '/':	if (character(point + 1) == '*') {
							point += 2;
							search(1, "*/");
						} else
							nl_forward();	/* c++ comment */
						break;
			default:	if (!good_c_tag())
							point = end;
						break;
		}
	}
}

char *tag_c_ident_pattern();

tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start, i;

	if (flavor == 1 && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(-1, "@class"))
		return tag_objc_class_list();
	c_fwd_space();
	start = i = point;
	if (parse_string(1, "__declspec</word>")) {
		point = matchend;
		c_fwd_space();
		if (curchar() == '(')
			c_skip_level('(', ')');
		c_fwd_space();
	}

	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	if (flavor == 0 && (c_extra_keywords & CPP_KEYWORDS))
		flavor = 1;	// In C++, treat structs just like classes.
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer()
		  && parse_string(1, tag_c_ident_pattern())) {
		i = point;
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (strchr(";[,", curchar()) && i > start) {
		point = i;	// A plain definition.
		return 1;
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		tag_c_add_class(func);
		i = ++point;	// skip over { so we do stuff within
		c_fwd_space();
		if (parse_string(1, "#(else|elif)"))
			if (locate_by_ifdef(1) != 2)
				point = i;
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

int tex_start_construct()
{
	int s;

	if (get_tagged_region("color-tex-as-unit", point, &s) == 1)
		return s;
	save_var point, narrow_end = size() - point;
	re_search(-1, "\n([ \t]*\n|\\def[^a-z0-9])");
	for (; re_search(1, "[%$\\]"); ) {	// Look for %, $, or \\ characters
		s = matchstart;
		switch (character(point - 1)) {
			case '\\':
				if (curchar() == '[') {
					if (!search(1, "\\]"))
						return s;
				} else
					point++;
				break;
			case '$':
				if (curchar() == '$') {
					if (!search(1, "$$"))
						return s;
				} else
					if (!search(1, "$"))
						return s;
				break;
			case '%':
				to_end_line();
				break;
		}
	}
	return point;
}

color_tex_math_env(char *envs, int start, int color)
{
	char envname[FNAMELEN], pat[FNAMELEN];

	if (!envs)
		return;
	sprintf(pat, "{(%s)}", envs);
	if (!parse_string(-1, pat))
		return;
	grab(matchend, matchstart, envname);
	sprintf(pat, "\\end%q", envname);
	int orig = point;
	if (re_search(1, pat)) {
		color_tex_math(orig, point = matchstart, color);
		set_tagged_region("color-tex-as-unit", start, give_end_line(), 1);
	} else
		point = orig;
}

color_tex_cmd(int s)
{
	char cmd[FNAMELEN];
	int color = color_class tex_literal, orig;

	strcpy(cmd, "|");
	if (parse_string(1, "[a-z@]*", cmd + 1)) {
		strcat(cmd, "|");
		if (strstr("|begin|end|", cmd))
			color = color_class tex_begin_end;
		else if (strstr("|chapter|section|subsection|subsubsection"
						"|appendix|bibliography|title|tableofcontents"
						"|input|include|listoffigures|listoftables|part"
						"|usepackage|requirepackage|paragraph|subparagraph|",
						cmd))
			color = color_class tex_section;
		else
			color = color_class tex_command;
	} else if (curchar() == '[') {
		orig = point;
		if (re_search(1, "\n[ \t]*\n|\\par![^a-z]|\\]")
			&& character(point - 1) == ']') {
			color_tex_math(orig - 1, point, color_class tex_display_math);
			return;
		}
		point = orig + 1;
	} else if (curchar() == '(') {
		orig = point;
		if (re_search(1, "\n[ \t]*\n|\\par![^a-z]|\\%)")
			  && character(point - 1) == ')') {
			color_tex_math(orig - 1, point, color_class tex_math);
			return;
		}
		point = orig + 1;
	} else
		point++;
	set_character_color(s, point, color);
	if (cmd[1] && strstr(latex_non_text_argument, cmd)
		  && parse_string(1, "{[^ \t\n\\#$%&~^_{}]+}")) {
		set_character_color(point, point + 1, color_class tex_curly_brace);
		set_character_color(point + 1, matchend - 1, color_class tex_keyword);
		set_character_color(matchend - 1, matchend, color_class tex_curly_brace);
		point = matchend;
		if (!strcmp("|begin|", cmd)) {
			color_tex_math_env(latex_math_env_pat, s, color_class tex_math);
			color_tex_math_env(latex_display_math_env_pat, s,
							   color_class tex_display_math);
		}
	}
}

color_tex_math(int from, int to, int color)
{
	int s;

	set_character_color(from, to, color);
	set_tagged_region("tex-math-mode", from, to, 1);
	save_var narrow_start = from, narrow_end = size() - to, point = from;
	while (re_search(1, "[\\%&]")) {
		s = point - 1;		// Color % and & specially in math mode unless
		switch (character(s)) {	// protected by \ character.
			case '\\':
				if (parse_string(1, "[a-z@]+"))
					color_tex_cmd(s);
				else
					point++;
				break;
			case '%':
				tex_color_comment(s);
				break;
			case '&':
				set_character_color(s, point, color_class tex_section);
				break;
		}
	}
}

color_vbasic_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, old, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = from;
	to_begin_line();
	set_character_color(point, to, -1);
	if (talk = (to - from > 20000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		old = point;
		if (!re_search(1, "[#0-9A-Za-z_]+[$@#%&!]?|[\"']")) {
			set_character_color(old, point, -1);
			t = size();
			break;
		}
		set_character_color(old, point, -1);
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '\'':			// found comment
				nl_forward();
				set_character_color(t, point, color_class vbasic_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"[^\"\n]*[\"\n]");
				set_character_color(t, point, color_class vbasic_string);
				break;
			case 'm':
				if (parse_string(-1, "</word>rem")) {
					nl_forward();
					set_character_color(t, point, color_class vbasic_comment);
					break;
				}					// fall through
			default:		// found identifier, kywd, or number
				set_character_color(t, point, vbasic_keyword_color(t));
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring VBasic program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

color_vhdl_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	save_var case_fold = 1;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		if (!re_search(1, "[a-z_][a-z0-9_]*('[a-z0-9_]+)*"
					   "|[0-9.]([a-z0-9._]|e[-+][0-9])*l?|--|[\"'\\]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '-':			// found comment
				nl_forward();
				set_character_color(t, point, color_class vhdl_comment);
				break;
			case '"':		// found a string literal
				t = --point;
				if (strchr("BOX", toupper(character(point - 1)))
					  && parse_string(-1, "</word>[box]"))
					t = --point;
				if (re_search(1, "[box]?\"([^\"\\\n]|\\.)*[\"\n]"))
					set_character_color(t, point, color_class vhdl_string);
				if (get_character_color(point, (int *) 0, &s) ==
					  color_class vhdl_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\'':
//				if (re_search(1, "([^'\\\n]|\\.)*['\n]"))
				if (re_search(1, "([^'\n])*['\n]"))
					set_character_color(t, point, color_class vhdl_string);
				if (get_character_color(point, (int *) 0, &s) ==
					  color_class vhdl_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\\':		// \arbitrary name\ is an identifier.
				if (re_search(1, "[\\\n]"))
					set_character_color(t, point, color_class vhdl_identifier);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, vhdl_keyword_color(t));
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring VHDL program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

int spell_get_and_check_word(int opt, char *word)
{
	int start = matchstart, end = matchend;

	if (matchend - matchstart > FNAMELEN - 1)
		return 1;
	grab(matchstart, matchend, word);
	if (spell_ignore_word(opt, matchstart, matchend, word))
		return 2;
	if (good_word_spelling(word))
		return 3;
	char *s = get_mode_string_variable("%s-spell-ignore-pattern-prefix");
	if (s && *s && matches_at(start, -1, s))
		return 4;
	s = get_mode_string_variable("%s-spell-ignore-pattern-suffix");
	if (s && *s && matches_at(end, 1, s))
		return 4;
	return 0;
}

char *get_spell_word_pattern();

command spell_buffer_or_region()
{
	if (!load_spell_word_list(0))
		return;
	save_spot point, mark;
	spot end = alloc_spot(1);
	on_exit if (spot_to_buffer(end) != -1) free_spot(end);
	on_exit highlight_off();
	if (is_highlight_on()) {
		fix_region();
		*end = mark;
	} else {
		point = 0;
		*end = size();
	}
	int opt = get_num_var(vspell_get_mode_variable());
	char *pat = get_spell_word_pattern();
	while (point < *end && re_search(1, pat) == 1) {
		char word[FNAMELEN];
		mark = matchstart;
		if (spell_get_and_check_word(opt, word))
			continue;
		highlight_on();
		if (do_spell_correct(1) == 1) {
			highlight_on();
			show_binding("Continue spell checking by typing %s", "exit-level");
			if (!recursive_edit_preserve())
				break;
		}
	}
	say("Spell checking completed.");
}

command spell_grep()
{
	if (!load_spell_word_list(0))
		return;
	char fname[FNAMELEN], dir[FNAMELEN];
	save_var point, mark;
	int pos = point;
	if (is_highlight_on()) {
		fix_region();
		save_var narrow_start = point, narrow_end = size() - mark;
	}
	int buf = tmp_buf(), cnt = 0, eline = 0, lstart = 0;
	on_exit buf_delete(buf);
	recolor_buffer_range(0, size());
	init_copy_for_grep(fname, dir);
	save_var abort_searching = 0;
	point = 0;
	int opt = get_num_var(vspell_get_mode_variable());
	char *pat = get_spell_word_pattern();
	while (point < size() && re_search(1, pat) == 1
		   && !user_abort) {
		char word[FNAMELEN];
		mark = matchstart;
		if (spell_get_and_check_word(opt, word))
			continue;
		if (point < eline) {
			buf_set_character_color(buf, lstart + mark - give_begin_line(),
									lstart + point - give_begin_line(),
									color_class spell_misspelled);
		} else
			lstart = copy_line_for_grep(fname, buf, mark, point,
										color_class spell_misspelled);
		eline = give_end_line();
		cnt++;
		if (size() > 20000)
			note("%d misspelled words, %d%% complete.",
				 cnt, muldiv(point, 100, size()));
	}
	user_abort = 0;
	if (size() > 20000)
		say("");
	point = pos;
	if (buf_size(buf))
		finish_copy_for_grep(buf, "misspelled words", fname, dir);
	else
		say("No misspelled words in %s.",
			  is_highlight_on() ? "region" : "buffer");
	highlight_off();
}

// Does the specified buffer appear in any window?  If so, return its
// handle, else -1.
is_buf_in_window(int buf)
{
	int orig = window_number;

	if (buf < 0)
		return -1;
	save_var narrow_start = 0, narrow_end = 0, window_bufnum, point;
	save_var window_handle;
	bufnum = window_bufnum;
	do {
		if (!system_window && buf == bufnum)
			return window_handle;
	} while (++window_number != orig);
	return -1;
}

when_loading()
{
	strcpy(context_help_rule_python,
		   "=http://www.google.com/search?q=*+site:docs.python.org&l=en&btnI=I");
}
