/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B11 to 13 B12, assuming later updaters.

#include "eel.h"
#include "epsnet.h"
#include "html.h"
#include "tags.h"
#include "proc.h"

int telnet_activity(int activity, int buf, int from, int to);
int ssh_activity(int activity, int buf, int from, int to);
char *scp_variable(struct url_parts *parts, char *varname);

void do_telnet(char *host, int port, char *bname)
{
	to_buffer(bname);
	type_point = 0;
	telnet_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_net_activity = telnet_activity;
	telnet_id = telnet_host(host, port, bname);
}

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s");
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
}

html_find_matching_quote(int c)
{
	char pat[FNAMELEN];
	int level = 0, j, orig = point;
	strcpy(pat, "X|%<%%|%%%>|%<%?|%?%>|%<(bean):");
	pat[0] = c;
	while (re_search(1, pat)) {
		switch (character(point - 1)) {
			case '?':
			case '%':			// Found <% or <?.
				level++;
				break;
			case '>':			// Found %> or ?>.
				level--;
				break;
			case ':':
				j = point;
				if (!search(1, ">") || character(point - 2) != '/')
					point = j;
				break;
			default:
				if (level <= 0)
					return;
		}
		if (level > 3 || level < 0) {	// Mismatched, so use simple rules.
			point = orig;
			pat[1] = 0;
			search(1, pat);
			return;
		}
	}
}

int tag_find_file()
{
	save_var synchronous_file_io = 1;
	switch (tag_to_window_key) {
		case '1':
		case '2':
		case '5':
		case 'z':
			has_arg = iter = 0;
			do_command(cx_tab[key]);
			break;
		case 'n':
		case 'p':
		case 'o':
			if (number_of_windows() < 2)
				split_window();
			window_number += tag_to_window_key == 'p' ? -1 : 1;
			break;
		case '.':		// Stay in this window.
			break;
		default:
			int dir = get_direction_from_key(tag_to_window_key);
			if (dir != -1)
				window_handle = window_from_direction(dir, window_handle);
			else
				locate_window("", _last_tag_file);
			break;
	}
	return std_find_it(_last_tag_file);
}

when_loading()
{
	clear_process_buffer |= 0x20;
}
