/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B12 to 13 B13, assuming further updaters.

#include "eel.h"
#include "epsnet.h"
#include "html.h"
#include "tags.h"
#include "proc.h"
#include "colcode.h"
#include "c.h"
#include "vbasic.h"
#include "python.h"
#include "perl.h"

char *find_framework_header(char *hdr)
{
	char *p, *result, srchpath[FNAMELEN], fname[FNAMELEN], bufdir[FNAMELEN];

	p = index(hdr, path_sep);	// Must be in Fmwk/Header.h format.
	if (!p || !*mac_framework_dirs)
		return 0;
	sprintf(fname, "%.*s.framework/{Headers,PrivateHeaders}%s",
			p - hdr, hdr, p);
	// Look for a header in a root framework.
	if (result = look_on_path(fname, PATH_PERMIT_WILDCARDS, mac_framework_dirs))
		return result;
	// If that didn't work, look for sub-frameworks if current file is
	// inside a framework.
	get_buffer_directory(bufdir);
	if (*bufdir && (p = strstr(bufdir, ".framework"))) {
		*p = 0;		// Make bufdir into path to framework folder.
		sprintf(srchpath, "%s.framework/Frameworks", bufdir);
		result = look_on_path(fname, PATH_PERMIT_WILDCARDS, srchpath);
	}
	return result;
}

char *c_search_include_path(char *rel, int flags);

c_find_linked_file(char *file)
{
	char rel[FNAMELEN], curdir[FNAMELEN], *p;
	int flags = PATH_ADD_CUR_DIR, is_abs = 0;

	to_begin_line();
	if (parse_string(1, "[ \t]*#[ \t]*(include|import)[ \t]*[<\"](.*)([>\"])"
					 "[ \t]*(/[*/].*)?$")) {
		grab(find_group(2, 1), find_group(2, 0), rel);
		is_abs = character(find_group(3, 1)) == '>'
				 && !find_linked_file_ignores_angles;
		// Using #include "" or <> syntax?
	} else if (is_java_buffer() &&  parse_string(1, "[ \t]*(package|import)"
		"[ \t]*(<word>+(%.<word>+)*(%.%*)?)[ \t]*;")) {
		grab(find_group(2, 1), find_group(2, 0), rel);
		while (p = strchr(rel, '.'))	// Transform java.lang to
			*p = path_sep;				// java/lang filename.
		p = rel + strlen(rel) - 1;
		if (*p == '*') {
			strcpy(p, ".");
			flags |= PATH_PERMIT_DIRS;
		} else
			strcat(p, ".java");
	} else {
		default_find_linked_file(file);
		return;
	}
	strcpy(file, rel);
	get_buffer_directory(curdir);
	if (is_abs)				// Don't look for relative names at all.
		flags = 0;
	else {
		absolute(file, curdir);
		if (check_file(file))
			return;
	}
	p = c_search_include_path(rel, flags);
	if (!p && is_abs && is_unix_flavor == IS_UNIX_MACOS)
		p = find_framework_header(rel);
	if (!p)
		error("%s not found in include-directories path", rel);
	strcpy(file, p);
}

color_embedded_code(int from, int to, int lang, int limit)
{
	save_var point;
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	set_tagged_region("color-as-unit", from, to, lang);
	if (limit > 0 && to > limit + 1000 && lang == HT_PHP
		  && count_windows_with_buf(bufnum, 0) <= 1) {
		set_tagged_region("html-coloring-postponed", limit, to, 1);
		to = limit;		// Note that we only colored part of this.
	}
	if (lang == HT_JSCRIPT) {
		save_var c_extra_keywords |= JS_KEYWORDS;
		save_var recolor_range = color_c_range;
		save_var recolor_from_here = color_c_from_here;
	} else if (lang == HT_VBSCRIPT) {
		save_var recolor_range = color_vbasic_range;
		save_var recolor_from_here = recolor_by_lines;
		save_var sublanguage_code = SUBLANG_VBSCRIPT;
	} else if (lang == HT_PHP) {
		save_var recolor_range = color_full_perl_range;
		save_var recolor_from_here = recolor_from_top;
		save_var sublanguage_code = SUBLANG_PHP;
		// Any buffer with embedded PHP must invalidate coloring after
		// edits, as in Perl mode.
		coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_RESETS;
	} else if (lang == HT_PYTHON) {
		save_var recolor_range = color_python_range;
		save_var recolor_from_here = recolor_from_top;
	} else if (lang == HT_CSS) {
		save_var recolor_range = color_css_range;
		save_var recolor_from_here = recolor_from_top;
	} else {
		set_character_color(from, to, color_class html_attribute);
		return;
	}
	color_for_narrowed_area(from, to);
}

html_find_linked_file(char *file)
{
	char buf[FNAMELEN], *p;
	int orig;

	if (!re_search(1, "(href|src)[ \t\n]*=[ \t\n]*")) {
		to_begin_line();
		if (!re_search(1, "(href|src)[ \t\n]*=[ \t\n]*"))
			error("Can't find any link here.");
	}
	orig = point;
	if (curchar() == '"') {
		point++, orig++;
		search(1, "\"");
	} else if (curchar() == '\'') {
		point++, orig++;
		search(1, "'");
	} else	// No quotes, assume it's whitespace-delimited.
		re_search(1, "[^ \t\n<>]*"), point++;
	grab(orig, point - 1, buf);		// Get link from buffer.
	p = strchr(buf, '#');		// Remove #fragment if any.
	if (p)
		*p = 0;
	if (!*buf)
		error("Can't follow fragment links.");
	point = orig;
	if (parse_string(1, "[A-Za-z]+:"))
		strcpy(file, buf);					// An absolute URL.
	else if (parse_string(1, "/")) {		// An absolute URL path.
		strcpy(file, filename);
		strcpy(get_url_file_part(file, 1), buf);
	} else {								// A relative URL path.
		strcpy(file, filename);
		strcpy(get_tail(file, 1), buf);
	}
}

tag_c_skip_before_body(int show)
{
	if (parse_string(1, "#.*(\\\n.*)*")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(try|const|volatile|NOTHROW)![^_A-Za-z0-9]")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(AVAILABLE|DEPRECATED_IN)_MAC_OS_X_VERSION_<word>+")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "throws[ \t]")) {	// for java
		point = matchend;
		skip_identifier_list();
		return 1;
	}
	if (parse_string(1, "(throw|__attribute__|__DARWIN_ALIAS)[ \t]*!%(")) {
		point = matchend;	// throw for C++, attribute for gcc
		if (skip_c_delims('(', ')', 1)) {
			if (show)
				buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
							  filename, lines_between(0, point, 0) + 1);
			return 0;
		}
		return 1;
	}
	return 0;
}

// Move past a series of comma-separated identifiers with
// whitespace/comments between.  Return how many.
int skip_identifier_list()
{
	int cnt = 0;
	for (;; point++) {
		c_fwd_space();
		if (!parse_string(1, "[a-zA-Z0-9_.]+"))
			break;
		point = matchend;
		if (curchar() == '<')
			c_skip_level('<', '>');
		cnt++;
		c_fwd_space();
		if (curchar() != ',')
			break;
	}
	return cnt;
}

char *tag_c_ident_pattern();

parse_other_class()
{
	c_fwd_space();
	if (is_java_buffer() && curchar() == '<')
		c_skip_level('<', '>');
	c_fwd_space();
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (parse_string(1, tag_c_ident_pattern())
				  || parse_string(1, "[,:]"))
				point = matchend;
			else if (!find_template_end())
				break;
		}
		c_fwd_space();
	} else if (is_java_buffer()) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}

do_unicode_convert(int flags)
{
	char msg[FNAMELEN], *enc;
	int err = 0, from = 0, to = size(), high, scol, ecol;

	high = is_highlight_on();
	check_modify(bufnum);
	sprintf(msg, "Convert %s %s", high ? "region" : "buffer",
			(flags & CONV_TO_16) ? "to Unicode from encoding"
								 : "from Unicode to encoding");
	if (!(enc = encoding_to_name(ask_encoding(msg, 1))))
		return;
	if (high) {
		save_spot point, mark;
		fix_region();
		from = point;
		to = mark;
	} else
		save_spot point;
	if (high == REGRECT) {
		scol = get_column(point);
		ecol = get_column(mark);
		for (; point < mark && !err; nl_forward())
			err = perform_unicode_conversion(bufnum, column_to_pos(scol),
											 column_to_pos(ecol), flags, enc);
	} else
		err = perform_unicode_conversion(bufnum, from, to, flags, enc);
	if (err == ETOOBIG)
		error("This is not an 8-bit encoding; a character is already wide.");
	else if (err)
		file_error(err, filename ? filename : bufname, "conversion error");
	if (!high && (flags & CONV_TO_16)) {
		int trans = GET_LINE_TRANSLATE(translation_type);
		int conv = encoding_from_name(enc);
		translation_type = MAKE_TRANSLATE(conv, trans,
										  (trans != FILETYPE_AUTO || conv > 0));
		make_mode();
	}
}

when_loading()
{
	drop_name("count_windows_with_buf");	// Now a prim.
}
