/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B13 to 13 B14, assuming further updaters.

#include "eel.h"
#include "session.h"
#include "html.h"
#include "tags.h"
#include "proc.h"
#include "colcode.h"
#include "c.h"
#include "grep.h"
#include "python.h"
#include "diff.h"
#include "epsnet.h"
#include "dired.h"

color_class html_altcomment = ocher | EFONT_ITALIC;
color_scheme "standard-mono" {
	color_class html_altcomment = color_class text;
};

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	point--;	// See if "class foo" is before this.
	strcpy(last, "");
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (c_ignore_declspec())
			continue;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords & (CSHARP_KEYWORDS
				| JAVA_KEYWORDS | JS_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum") || !strcmp(this, "__interface")
				  || !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, (c_extra_keywords == CSHARP_KEYWORDS) ? "." : "::");
						class += strlen(class);
					}
					strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

is_switch(p)	/* is there a switch-like statement at position p? */
{
	if (!index("scuni_", character(p)))
		return 0;
	save_var point = p;
	return parse_string(1, "(switch|struct|union|class|namespace"
						"|(__)?interface)[^a-zA-Z0-9_]");
}

indent_after_block(orig)
{
	int ind;

	c_rev_skip_level('{', '}');	/* find matching { */
	to_begin_line();
	ind = get_indentation(point);
	if (Topindent && (ind == 0 || parse_string(1, "[ \t]*{")
					  && (ind <= c_brace_offset || ind <= c_top_braces)))
		return 0;	// found a func def block
	if (!Topindent && c_top_braces && !Closeback && ind <= c_top_braces)
		return 0;
	return indent_after_statement(orig);
}

c_default_indent(ind, orig, defmult)
{
	if ((c_extra_keywords & CSHARP_KEYWORDS) && parse_string(1, "[ \t\f]*%[")
		  && have_csharp_attribute(point))
		return get_indentation(point);
	if (matches_at(orig, 1, "[ \t\f]*//"))
		return ind;
	if ((c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS | CSHARP_KEYWORDS))
		  && matches_at(orig, 1, "[ \t\f]*(class|(__)?interface)</word>"))
		return ind;		// Put class keyword in continuation back on left.
	if (ind && indent_preprocessor_contin == 255)
		return ind;
	if (ind)
		return ind + defmult * get_c_indent() + c_contin_offset;
	if (parse_string(1, "[ \t\f]*(struct|union|class|namespace)[ \t\f\n]+"
			 "[a-zA-Z0-9_]*[ \t\f\n]*{[ \t\f\n]*[\n/]"))
		return ind + c_top_struct;
	if (parse_string(1, "[ \t\f]*(template)[ \t\f\n]*%<"))
		return ind + c_top_template;
	if (just_open(orig))
		return c_top_braces;
	point = orig;
	cind_space();
	return is_contin_statement(point, 0, orig)
			? c_top_contin : c_param_decl;
}

// Flavor: 1 means do pre-einit -d flags, 2 means do all other -d flags
// plus -p and -l flags, 3 means do -r flags.
parse_command_line_flags(int flavor)
{
	int i, j;
	char *s, *flag, *arg, buf[1024];

	for (i = 1; i < argc; i++) {
		flag = argv[i];
		if (*flag != '-')
			continue;
		if (!strcmp(flag, "--"))
			return;
		if (!strcmp(flag, "-dir")) {
			i++;
			continue;
		}
		arg = (flag[1] && flag[2] || i + 1 >= argc) ? (flag + 2) : argv[++i];
		if (!strchr((flavor == 3 ? "r" : flavor == 2 ? "lpd"  : "d"), flag[1]))
			continue;
		switch (flag[1]) {
			case 'p':
				if (!flag[2])
					preserve_session_once = preserve_session ? 1 : 2;
				else if (isdigit(flag[2]))
					preserve_session_once = numtoi(flag + 2) + 1;
				else
					cmd_line_session_file = flag + 2;
				break;
			case 'l':
				load_commands(arg); break;
			case 'd':
			case 'r': 
				strcpy(buf, arg);
				s = index(buf, '=');
				if (!s)		// accept either VAR=VAL or VAR!VAL
					s = index(buf, '!');
				if (s)
					*s = 0;
				if (j = find_index(buf)) {
					if (flag[1] == 'd') {
						if (!s)
							bad_flag(flag);
						if ((flavor == 1) != !strcmp(buf, "einit-file-name"))
							continue;	// Do this var only if flavor is 1.
						now_parsing_cmds = 1;	// Don't record this.
						safe_set_var_val(j, s + 1, 1);
						now_parsing_cmds = 0;
					} else {
						if (s)
							(*(int (*)())j)(s + 1);
						else
							(*(int (*)())j)();
					}
					break;
				}
				if (flag[1] == 'd') {
					bprintf("No such variable %s.\n", buf);
					init_help_abort();
				} else
					say("No command %s, %s ignored.", buf, flag);
				break;
		}
	}
}

when_restoring()
{
	int tflag = 0, i, splitmode = HORIZONTAL, have_files = 0;
	char *flag, *arg, init_dir[FNAMELEN], *dir = init_dir;

	getcd(init_dir);
	early_restoring();
	early_init();
	set_toolbar();

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			have_files = 1;
			continue;
		}
		if (!strcmp(argv[i], "--"))
			break;
		if (!stricmp(argv[i], "-noinit")) {
			noinit_flag = 1;
			continue;
		} else if (!strcmp(argv[i], "-dir")) {
			if (++i < argc)
				dir = argv[i];
			continue;
		}
		flag = argv[i];
		switch (flag[1]) {
		case 'f':
			arg = (flag[2] && flag[3] || i + 1 >= argc) ? (flag + 3) : argv[++i];
			switch (flag[2]) {
			case 's':
				set_swapname(arg);
				break;
			case 'd':
				strcpy(_docname, arg);
				break;
			default:
				bad_flag(flag);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'd':
		case 'r':
		case 'l': if (!flag[2] && i + 1 < argc)
					  i++;
				  break;
		case 'p': break;
		case 'h': case '?': usage(); break;
		case 'v':
			switch (flag[2]) {
				case 'v':
					splitmode = VERTICAL;
					break;
				default:
					bad_flag(flag);
			}
			break;
		case '-':
			if (!strcmp(flag + 2, "help"))
				usage();
			else
				bad_flag(flag);
			break;
		default:
			bad_flag(flag);
		}			/* i is 1st file argument */
	}
	if (i < argc)
		have_files = 1;
	term_init();
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	delete_buffer("-customizations");	// Don't record any settings before now.
	parse_command_line_flags(1);
	if (!noinit_flag && load_from_state)
		init_from_file();
	parse_command_line_flags(2);
	if (tflag) {
		preserve_session_once = 1;
		tutorial();
	} else {
		if (load_from_state && (!have_files || session_always_restore)) {
			if (session_restore_files == 2 && have_files)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (have_files)
			load_cmd_line_files(1, splitmode, dir);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	restore_check_version();
	show_text(0, 1, "Epsilon Programmer's Editor %s, EEL version %s",
			  version, eel_version);
	try_calling("start-up");
	parse_command_line_flags(3);
}

/* Start at argv[i].  If dirname, interpret filenames relative to it. */
load_cmd_line_files(int i, int splitmode, char *dirname)
{
	int wins = 0, tot_size, line, column, pos, j, all_files = 0;
	char fname[FNAMELEN], fixed[FNAMELEN], orig_dir[FNAMELEN], *s, first = 1;

	if (dirname) {
		getcd(orig_dir);
		on_exit chdir(orig_dir);
		chdir(dirname);
	}
	one_window();			// Don't use windows specified by session
	tot_size = window_height;
	if (is_gui && i + 1 < argc && argv[i][0] != '+') {	// multiple args
		for (j = i, *fixed = 0; j < argc; j++) {
			if (strlen(fixed) + strlen(argv[j]) >= FNAMELEN - 1) {
				*fixed = 0;
				break;		// See if perhaps a stray association is
			}				// passing us a file name w/o the quotes.
			if (*fixed)
				strcat(fixed, " ");
			strcat(fixed, argv[j]);
		}
		if (check_file(fixed) == 1 && !check_file(argv[i])) {
			argv[i] = strsave(fixed);
			argc = i + 1;
		}
	}
	line = column = pos = -1;
	for (; i < argc; i++) {
		if (argv[i][0] == '+') {
			if (argv[i][1] != 'p') {
				line = strtoi(argv[i] + 1, 10);
				s = strchr(argv[i], ':');
				if (s)
					column = strtoi(s + 1, 10);
			} else
				pos = strtoi(argv[i] + 2, 10);
		} else if (argv[i][0] == '-' && !all_files) {
			if (argv[i][1] == '-')
				all_files = 1;
			else if (!strcmp(argv[i], "-dir")
					 || strchr("ldr", argv[i][1]) && !argv[i][2]
					 || strchr("f", argv[i][1]) && !argv[i][3])
				i++;
		} else {
			strcpy(fname, argv[i]);
			convert_cygwin_filename(fname, 0);
			absolute(fname);
			if ((directory_flags & DIR_CWD_FROM_FILE) && first) {
				char dir[FNAMELEN];
				strcpy(dir, fname);
				*get_tail(dir, 1) = 0;
				chdir(dir);
				first = 0;
			}
			s = expand_wildcards ? do_file_match(fname, STARTMATCH | EXACTONLY)
				: fname;
			if (!s)		/* new file */
				s = fname;
			do {
				if (wins < max_initial_windows && wins++ > 0) {
					window_split(splitmode);
					window_height = 1000;
				}
				if (!(is_pattern(s) & 2) && !is_directory(s)
					&& !is_remote_file(s))	// Has quotable wildcards.
					find_it(s, get_default_translation_type(s));
				else
					do_find(s, get_default_translation_type(s));
				if (line > 0)
					go_line(line);
				if (pos >= 0)
					point = pos;
				if (column >= 0)
					move_to_column(column);
			} while (expand_wildcards
				&& (s = do_file_match(fname, EXACTONLY)));
			line = column = pos = -1;
		}
	}
	for (; window_number > 0; window_number--)
		window_height = tot_size / number_of_user_windows();
}

int set_color_handler(int activity, int buf, int from, int to);

set_color_xwin()
{
	int i, j, classes = number_of_color_classes();
	int *orig_fg, *orig_bg, *orig_schemes;
	int orig_scheme = selected_color_scheme, this_scheme = orig_scheme;
	char *s, cmd[FNAMELEN];
	int orig_class = get_character_syntax_color(point);

	if (no_mac_dialog_helper()) {
		mac_gtk_message(0);
		return;
	}
	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	on_exit free(orig_schemes), free(orig_fg), free(orig_bg);
	save_var bufnum;
	get_executable_file(cmd, "epsilon-dialogs", 1);
	sprintf(cmd + strlen(cmd), " --parent %x --color", screen_to_window_id(0));

	for (;;) {
		bufnum = zap("-epsilon-dialog-data");
		bprintf("%d\n%s\n", font_styles, font_fixed);
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i))
				bprintf("%s\n", name_name(i));
		bprintf("\n");
		for (i = 0; i < classes; i++)
			bprintf("%s\n", name_color_class(i));
		bprintf("\n");
		bprintf("%s\n", (s = lookpath("colclass.txt")) ? s : "");
		bprintf("%s\n", name_name(this_scheme));

		selected_color_scheme = find_index("color-defaults");
		for (j = 0; j < classes; j++)
			bprintf("default %x %x\n", get_foreground_color(j, 1),
					get_background_color(j, 1));
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i)) {
				selected_color_scheme = i;
				for (j = 0; j < classes; j++)
					bprintf("%d %d %x %x\n", i, j, get_foreground_color(j, 1),
							get_background_color(j, 1));
			}
		make_color_class_list(1);
		point = size();
		s = name_color_class(orig_class);
		bprintf("\n%s\n", s ? s : "");
		selected_color_scheme = this_scheme;
		if (!pipe_text("-epsilon-dialog-data", "-epsilon-dialog-out", cmd,
					   NULL, PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH,
					   set_color_handler))
			file_error(errno, "Couldn't run dialog helper", "Error");
		this_scheme = selected_color_scheme;
		bufname = "-epsilon-dialog-out";
		i = process_exit_status;
		if (matches_at(0, 1, ".*(Library not loaded|can't open library)")) {
			mac_gtk_message(0);
			return;
		}
		if (i < 0 || i > 2)
			i = !ask_yn("Keep changes", "The dialog process exited "
						"unexpectedly.  Keep these settings? ",
						"Yes", "No", 0);
		if (i == 0) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			return;
		} else if (i == 1) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			return;
		} else {
			show_binding("Continue changing colors "
						 "by typing %s", "exit-level");
			recursive_edit();
		}
	}
}

int set_color_search_expand(char *name)
{
	char pat[FNAMELEN];
	sprintf(pat, "(^|\t)[-+ ]*%q($|\t)", name);
	point = 0;
	if (!re_search(1, pat))
		return 0;
	set_color_expand(1);
	return 1;
}	

#define FG(c)		((c) & 15)	/* extract foreground & background */
#define BG(c)		(((c) >> 4) & 7)
#define COLOR_COL	57
#define COLOR_LINE	2
#define SCRPLACE(c)	(COLOR_COL + FG(c)), (COLOR_LINE + BG(c))

command set_color()		/* query user to set colors */
{
	int class, color, i, j, *orig, classes, docbuf, wantwin, cmd;
	int swin, ccwin, cwin, infowin, newcolor, newscheme, orig_scheme;
	int *orig_fg, *orig_bg, *orig_schemes;
	int fg_color, bg_color;
	char buf[100], class_line[100], *s;
	int sbuf, ccbuf, cbuf, infobuf, orig_win = window_handle, win, page;
	char *new, *old;
	jmp_buf this_level;

	int orig_class = get_character_syntax_color(point);
	save_var abort_searching = 0, shift_selects = 0;
	if (is_unix == IS_UNIX_XWIN) {
		set_color_xwin();
		return;
	}
	save_var top_level = &this_level;
	classes = number_of_color_classes();

	sbuf = tmp_buf();	/* list of color schemes */
	bufnum = sbuf;	/* now highlight current line in buffers */
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	make_color_scheme_list(sbuf);

	orig_scheme = selected_color_scheme;
	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	selected_color_scheme = orig_scheme;
	orig = (int *) malloc(classes * sizeof(int));
	for (i = 0; i < classes; i++)
		orig[i] = alter_color(i, -1);

	class = iter = page = 0;

	bufnum = docbuf = tmp_buf();	// read help text for classes into buf
	on_exit buf_delete(docbuf);
	if (new = lookpath("colclass.txt"))
		file_read(new, FILETYPE_AUTO);
	ccbuf = tmp_buf();	/* list of color classes */
	cbuf = tmp_buf();	/* window with colors */ 
	infobuf = tmp_buf();	/* help buffer with options */ 

	bufnum = ccbuf;
	make_color_class_list(0);

winagain:
	swin = add_popup(4, 1, 20, 10, ALL_BORD(BSINGLE), sbuf);
	window_title(swin, BTOP, TITLECENTER, " Color Schemes ");
	set_menu_look(swin);

	ccwin = add_popup(28, 1, 25, 10, ALL_BORD(BSINGLE), ccbuf);
	display_column = 0;		/* prevent continuation lines */
	window_title(ccwin, BTOP, TITLECENTER, " Color Classes ");
	set_menu_look(ccwin);
	point = 0;
	set_color_expand(1);	// Expand first category.
	if (orig_class >= 0 && set_color_search_expand(name_color_class(orig_class))
		  && set_color_search_expand(name_color_class(orig_class))
		  && set_color_search_expand(name_color_class(orig_class)))
		class = this_color_class_to_index();

	if (!(has_feature & FEAT_ANYCOLOR)) {		// Matrix of color choices.
		cwin = add_popup(56, 1, 18, 10, ALL_BORD(BSINGLE), cbuf);
		window_title(cwin, BTOP, TITLECENTER, " Colors ");
		invisible_window = 1;
		insert('\n');
		if (selectable_colors > 128)
			window_title(cwin, BBOTTOM, TITLERIGHT(1), " More ");
	}

	infowin = add_popup(2, 14, 75, 8, _view_border, infobuf);
	display_column = 0;		/* prevent continuation lines */
	mouse_cursor = &left_pointer;
	text_color = is_gui ? color_class text : color_class menu;
	display_scroll_bar = 0;
	newscheme = 2;
	if (has_feature & FEAT_COLOR_DIALOG) {
		display_dialog_box("SetColorBox", "Set colors",
						   swin, ccwin, infowin, "OK", "Cancel", "Help");
		save_var permit_window_keys = PERMIT_SCROLL_KEY;
		save_var return_raw_buttons = 1;
	}
	wantwin = ccwin;

	if (setjmp(top_level))	/* come here if subcommand's aborted */
		;
	for (;;) {
		if (class >= 0) {
			color = alter_color(class, -1);
			fg_color = get_foreground_color(class, 0);
			bg_color = get_background_color(class, 0);
			if ((color >= 128) != page && !newscheme)
				newscheme = 1;
			page = (color >= 128);
		}
		if (newscheme) {
			if (!(has_feature & FEAT_ANYCOLOR)) {	// Sample window has 2 pages.
				window_handle = cwin;
				point = page;
			}
			window_handle = swin;
			if (newscheme == 2) {
				make_color_scheme_list(bufnum);
				delete(size() - 1, size());	/* no final \n */
			}
			point = 0;
			col_search(name_name(selected_color_scheme), 0);
		}
		if (!(has_feature & FEAT_ANYCOLOR)) {
			j = page ? 128 : 0;
			for (i = 0; i < 128; i++, j++)
				term_write_attr(SCRPLACE(i), '+',
								(class >= 0) ? j : 0);
		}
		window_handle = ccwin;
		if (class >= 0) {
			point = 0;
			sprintf(buf, "^[-+ ]*%q$", name_color_class(class));
			re_search(1, buf);
		}
		to_begin_line();
		parse_string(1, "[^\n\t]*", class_line);
		to_begin_line();
		buf_zap(infobuf);
		window_handle = infowin;
		if (!(has_feature & FEAT_COLOR_DIALOG))
			bprintf("Type F or B to change the color scheme, D to "
					"define a new color scheme.\n");
		if (class >= 0) {
			if (!(has_feature & FEAT_COLOR_DIALOG)) {
				bprintf("Or use the arrow keys to change the "
						"%s color class.\n", name_color_class(class));
			} else
				bprintf("Sample text using the %s color class:\n",
						name_color_class(class));
			i = point;
		}
		bufnum = docbuf;
		point = 0;
		s = class_line;
		while (strchr(" -+", *s))
			s++;
		sprintf(buf, "^%q\t+", s);
		if (re_search(1, buf))
			buf_xfer(infobuf, point, give_end_line());
		else if (class >= 0)
			buf_printf(infobuf, "This is sample text.");
		bufnum = infobuf;
		j = point + 1;
		if (!(has_feature & FEAT_COLOR_DIALOG)) {
			set_color_explain_keys(class_line);
			if (class >= 0)
				bprintf("Color was initially 0x%02x, this color is 0x%02x.",
						orig[class], color);
		} else if (string_matches_pattern(class_line, " *<+>"))
			bprintf("\nPress + to expand or - to collapse a category.");
		else if (string_matches_pattern(class_line, " *-"))
			bprintf("\nPress - to collapse this category.");
		if (class >= 0)
			set_character_color(i, j, class);
		point = 0;

		// Disable 2 color-setting buttons and 3 font style checkboxes only.
		disable_dialog_controls = (class < 0) ? 0x1807 : 0;
		dialog_checkboxes
				= (class < 0) ? 0 : (((fg_color & EFONT_BOLD) ? 1 : 0)
									 | ((fg_color & EFONT_ITALIC) ? 2 : 0)
									 | ((fg_color & EFONT_UNDERLINED) ? 4 : 0));

		if (has_feature & FEAT_ANYCOLOR) {
			window_handle = wantwin;
			maybe_refresh();
			getkey();
			wantwin = window_handle;
		} else {
			window_handle = cwin;
			do {
				maybe_refresh();
				if (class >= 0)
					term_write_attr(SCRPLACE(color), 'C', color);
				full_getkey("Your choice: ", 0);
				if (class >= 0)
					term_write_attr(SCRPLACE(color), '+', color);
			} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		}
		key = toupper(key);
		newcolor = -100;	// Assume it's an invalid key.
		newscheme = 0;
		cmd = binding_of(key, reg_tab);
		if (cmd == (short) next_page || cmd == (short) previous_page
			  || cmd == (short) goto_beginning || cmd == (short) goto_end) {
			do_command(cmd);
			newcolor = -99;		// Mark as a valid key.
		}
		if (has_feature & FEAT_COLOR_DIALOG) {
			window_handle = ccwin;	// Scrolling might have changed these.
			class = this_color_class_to_index();

			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		}
		if (key_is_button == 4 && class >= 0) {
			i = comm_dlg_color(fg_color, "Foreground color");
			if (i != -1)
				set_color_pair(class, i | (fg_color & EFONT_STYLE), bg_color);
		} else if (key_is_button == 5 && class >= 0) {
			i = comm_dlg_color(bg_color, "Background color");
			if (i != -1)
				set_color_pair(class, fg_color, i);
		} else if (key_is_button == 8 && class >= 0) {
			set_color_pair(class, fg_color ^= EFONT_BOLD, bg_color);
		} else if (key_is_button == 9 && class >= 0) {
			set_color_pair(class, fg_color ^= EFONT_ITALIC, bg_color);
		} else if (key_is_button == 10 && class >= 0) {
			set_color_pair(class, fg_color ^= EFONT_UNDERLINED, bg_color);
		} else if (key == 'C' || key == 'Q' || key == ESC || user_abort
				   || key_is_button == 2) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			break;
		} else if (key == 'S' || key == ' ' || IS_ENTER_KEY(key)
				   || key_is_button == 1) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			break;
		} else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win))
				continue;
			window_handle = win;
			get_text_area(win);
			if (win == cwin) {
				if (bar_on_border()) {
					i = scroll_bar_code();
					if ((i ? (i > 0) : !page) != point)
						newcolor = color ^ 128;
				} else {
					newcolor = (mouse_y - COLOR_LINE) * 16;
					newcolor += (mouse_x - COLOR_COL);
					if (page)
						newcolor += 128;
				}
			} else if (in_text_area(mouse_x, mouse_y))
				point_to_pos(mouse_x, mouse_y, mouse_screen);
			else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
			if (win == ccwin) {
				class = this_color_class_to_index();
				if (class < 0)
					class = set_color_expand(2);
			} else if (win == swin)
				newscheme = 1;
			else if (win == infowin && isupper(curchar())
					 && character(point - 1) == ' '
					 && character(point + 1) == ' ')
				ungot_key = curchar();
			if (win != infowin)
				wantwin = win;
		} else if ((has_feature & FEAT_COLOR_DIALOG)
				   && full_key == NUMSHIFT(GREYTAB)) {
			prev_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && generic_key(key) == '\t') {
			next_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && key == GETFOCUS) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win) || win == infowin)
				continue;
			wantwin = win;
		}
		if (!(has_feature & FEAT_ANYCOLOR))
			switch (get_direction()) {
				case BTOP:		newcolor = color + 256 - 16; break;
				case BBOTTOM:	newcolor = color + 16; break;
				case BLEFT:		newcolor = color + 256 - 1; break;
				case BRIGHT:	newcolor = color + 1; break;
			}
		else
			switch (get_direction()) {
				case BLEFT:
					if (wantwin == ccwin && class == -2) {
						key = '-';
						break;
					}
				case BTOP: 
					key = (wantwin == ccwin) ? 'P' : 'B'; break;
				case BRIGHT:
					if (wantwin == ccwin && class == -3) {
						key = '+';
						break;
					}						
				case BBOTTOM:
					key = (wantwin == ccwin) ? 'N' : 'F'; break;
			}
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'N' && win == swin)
			key = 'F';
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'P' && win == swin)
			key = 'B';
		switch (key) {
			case 'N': 
				window_handle = ccwin;
				to_end_line();
				if (!nl_forward())
					point = 0;
				class = this_color_class_to_index();
				break;
			case 'P':
				window_handle = ccwin;
				if (!re_search(-1, "^.+"))
					point = size();
				class = this_color_class_to_index();
				break;
			case 'T':
				show_binding("Continue changing colors "
					"by typing %s", "exit-level");
				remove_window(swin);
				remove_window(ccwin);
				remove_window(cwin);
				remove_window(infowin);
				abort_searching = 1, shift_selects = 1;
				recursive_edit();
				abort_searching = 0, shift_selects = 0;
				goto winagain;
				break;
			case 'F':
				wantwin = window_handle = swin;
				if (!nl_forward())
					point = 0;
				newscheme = 1;
				break;
			case 'B':
				wantwin = window_handle = swin;
				if (!nl_reverse())
					point = size();
				newscheme = 1;
				break;
			case 'D':
				wantwin = swin;
				maybe_refresh();
				get_strpopup(buf, "New scheme name: ", "", "set-color");
				if (*buf) {
					new_variable(buf, NT_COLSCHEME, 0, 0);
					new = varptr(find_index(buf));
					old = varptr(selected_color_scheme);
					for (i = 0; i < ptrlen(old, 1); i++)
						*new++ = *old++;
					newscheme = 2;
				}
				break;
			default:
				if (cmd == (short) help || key == 'h' || key == '?'
					  || key_is_button == 3) {
					if (!(has_feature & FEAT_COLOR_DIALOG))
						say("Press <Space> to continue.");
					help_on_command("set-color");
					if (!(has_feature & FEAT_COLOR_DIALOG))
						newscheme = 1;
				} else if (is_color_search_key()) {
					window_handle = (has_feature & FEAT_COLOR_DIALOG)
									? wantwin : ccwin;
					run_color_search_key(window_handle == ccwin);
					if (window_handle == ccwin)
						class = this_color_class_to_index();
					else if (window_handle == swin)
						newscheme = 1;
				} else if ((key == '+' || key == '-') && (wantwin == ccwin
					|| !(has_feature & FEAT_COLOR_DIALOG))) {
					window_handle = ccwin;
					class = set_color_expand(key == '+');
				} else if (IS_TRUE_KEY(key) && newcolor == -100) {
					maybe_ding(bell_on_bad_key);
					newscheme = 1;
				}
				break;
		}
		if (newcolor > -99 && class >= 0)	/* did it change? */
			alter_color(class, newcolor % selectable_colors);
		else if (newscheme) {
			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		}
	}
	remove_window(swin);
	remove_window(ccwin);
	if (!(has_feature & FEAT_ANYCOLOR))
		remove_window(cwin);
	remove_window(infowin);
	buf_delete(ccbuf);
	buf_delete(cbuf);
	buf_delete(infobuf);
	if (is_window(orig_win))
		window_handle = orig_win;
	buf_delete(sbuf);
	full_redraw = 1;
	*get_color_scheme_variable() = selected_color_scheme;
	free(orig);
	free(orig_fg);
	free(orig_bg);
}

int buffer_range_has_file_name(int start, int end)
{
	char fname[FNAMELEN];
	if (end - start >= FNAMELEN - 1 || end - start <= 0)
		return 0;
	grab(start, end, fname);
	switch (check_file(fname)) {
		case CHECK_FILE:
		case CHECK_DIR:
			return 1;
	}
	return 0;
}

// Remove the filename before any ~ following a /.  But given a name
// like /Volumes//Volumes/~vol, only delete the initial /Volumes/ part.

unix_delete_file_path_before_tilde()
{
	if (!parse_string(1, "/.*/!~"))
		return;
	int p = matchend;
	save_spot point;
	int end = point + parse_string(1, "/.*/~[^/]*");
	while (point < end) {	// From tilde, go back to each /.
		if (buffer_range_has_file_name(point, end)) {
			delete(0, point);
			return;
		} else if (!search(1, "/"))
			break;
	}
	delete(0, p);
}

restart_file_paths()
{
	char pat[FNAMELEN];
	int res, res2;

	if (insert_default_response != 3
		  || prompt_with_buffer_directory == PBD_JUSTINS)
		return;
	save_var point = 0, abort_searching = 0, case_fold = 1;
	sprintf(pat, "(%s):[\\/][\\/][^\\/]+", url_services);
	point += parse_string(1, pat);
		// Don't consider ftp://host.com part, which may have \\ after.
	if (opsys == OS_UNIX)		// Delete /dir/name/ before a ~.
		unix_delete_file_path_before_tilde();
	sprintf(pat, "((%s%s):)?"	// optional a: or ftp:
		"[\\/].*[\\/]!"			// slash, stuff, slash
			"([\\/]|(%s|[a-z]):)",		// slash or a: or ftp:
		url_services, (opsys == OS_UNIX) ? "" : "|[a-z]", url_services);
	res = parse_string(1, pat);
	if (res) {	// Delete abc:\dirname\dir2\ or \dir\ or k:\dir\, etc.
		point += res;	// But exclude a trailing URL part.
		sprintf(pat, "(%s):[\\/]+", url_services);
		res2 = parse_string(-1, pat);
		if (res2 && res2 < point)
			point -= res2;
		sprintf(pat, "[/\\]!(%s):[\\/][\\/].*", url_services);
		res2 = parse_string(-1, pat);
		if (res2 && res2 < point)
			point -= res2;
		if (character(point) == '\\' && parse_string(-1, "\\\\"))
			point--;	// Preserve a UNC name following a path.
		delete(0, point);
	}
	if (opsys != OS_UNIX) {
		sprintf(pat, "[a-z]:[\\/]!([\\/]|(%s|[a-z]):)", url_services);
		res = parse_string(1, pat);
		if (res)	// Delete c:\ or similar before an absolute path.
			delete(0, point + res);
	}
}

do_diff(char *buf, int style)
{
	int start, other, osize, opt, other_pt;

	strcpy(_diff_buffer, buf);
	zap(_diff_buffer);
	drop_coloring(name_to_bufnum(_diff_buffer));
	if (style == VISUAL_DIFF_STYLE)
		buffer_printf(_diff_buffer, "%pComparing %p%s%p\n",
					  -1, color_class diff_removed,
					  *filename ? filename : bufname, -1);
	highlight_off();
	opt = point;
	next_user_window(1);
	other = bufnum;
	osize = size();
	other_pt = point;
	if (style == VISUAL_DIFF_STYLE)
		buffer_printf(_diff_buffer, "with %p%s%p\n", color_class diff_added,
					  *filename ? filename : bufname, -1);
	next_user_window(-1);
	if ((has_feature & FEAT_LCS_PRIMITIVES)
		  && size() <= diff_precise_limit && osize <= diff_precise_limit) {
		diff_via_lcs(bufnum, other, style);
		return;
	}

	diff_check_malloc();	/* see if we need to make room for data */
	_diff_line1 = _diff_line2 = 1;
	set_points(0, 0);
	do {
		start = point;
		diff_same();	// Found identical lines.
		note_same(bufnum, start, point, style);
	} while (resync_lines(style));
	set_points(opt, other_pt);
	say("Done.");
}

char *dired_skip_line_pat()
{
	if (dired_format == DF_VMS)
		return "^(     Directory of |  total [0-9]+|  .*:(%[.*%])?$|\n"
				"|  Directory .*:%[|  Total of .* (files|blocks)"
				"|                    |  USR%$DISK:"
				"|" OMITLINEPAT ")";
	else
		return "^(     Directory of |  total [0-9]+|\n"
				"|  Directory .*:%[|  Total of .* (files|blocks)"
				"|                    "
				"|" OMITLINEPAT ")";
}

dired_repl(newchar)
{
	to_begin_line();
	if (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	if (index("DCMULRG ", curchar()) && character(point + 1) == ' ')
		replace(point, newchar);
	save_var iter = 1;		// call us to do each repetition
	dired_down_line();
}

dired_move_by_marking(int dir)
{
	if (dir < 0)
		nl_reverse();
	to_begin_line();
	char pat[FNAMELEN];
	sprintf(pat, "^(%s|[^%c])", dired_skip_line_pat(), character(give_begin_line()));
	re_search(dir, pat);
	to_begin_line();
	if (dir < 0 || parse_string(1, dired_skip_line_pat()))
		nl_forward();
	move_to_column(goal_column);
}

dired_down_line() on dired_tab['n']
{
	down_line();
	to_begin_line();
	while (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	move_to_column(goal_column);
}

dired_up_line() on dired_tab['p']
{
	up_line();
	to_begin_line();
	while (parse_string(1, dired_skip_line_pat())) {
		if (!nl_reverse())
			break;
		to_begin_line();
	}
	move_to_column(goal_column);
}

int dired_getname(char *fullname, int flags)
{
	char *p, *q;

	dired_try_getdir(fullname, 1);
	p = fullname + strlen(fullname);
	dired_getindex(p, goal_column);		/* the file name */
	to_begin_line();
	if (parse_string(1, dired_skip_line_pat()))  /* on dir line, make name illegal */
		*p = 0;
	move_to_column(goal_column);
	if (!*p || isspace(*p))
		if (flags & 1)
			error("Can't read file name");
		else
			return 0;
	q = strstr(p, " -> ");			// replace file name with link dest
	if (q)
		if (flags & 2)
			*q = 0;
		else
			strcpy(p, q + 4);
	if (is_path_separator(*p) == 1)	// If link dest starts with /,
		strcpy(get_url_file_part(fullname, 0), p);	// just use that.
	if (dired_format == DF_SIMPLE) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
	} else if (dired_format == DF_VMS) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
		q = strstr(fullname, ".DIR;");	// Refer to directories without
		if (q)							// .DIR extension to see contents.
			*q = 0;
		q = strstr(fullname, ";");		// Remove version number.
		if (q && isdigit(q[1]))
			*q = 0;
	}
	p += strlen(p) - 1;		// Unix file name might use "ls -F" format
	if (*p == '/' || *p == '*')
		*p = 0;
	return 1;
}

dired_strip_dotfiles()
{
	char name[FNAMELEN];
	int filecnt = 0, dircnt = 0, specialcnt = 0;

	if (dired_force_show_dotfiles || dired_show_dotfiles == 15
		  || dired_format && dired_format != DF_UNIX)
		return;
	if (string_matches_pattern(get_dired_buffer_pattern(),
							   ".*(^|[,;{/\\])%.%.?[^./\\]+"))
		return;		// Pattern explicitly mentions dotfile.
	save_var modified, point = 0;
	for (; search(1, " ."); to_end_line()) { // Quickly get to possible dotfile.
		if (give_end_line() >= size())	// Don't do partial lines at eof.
			break;
		to_begin_line();
		if (parse_string(1, dired_skip_line_pat()))
			continue;		// Ignore directory separator lines.
		dired_getindex(name, goal_column);
		if (*name == '.') {
			if (strlen(name) > 1 && is_path_separator(name[strlen(name) - 1]))
				name[strlen(name) - 1] = 0;
			if (!strcmp(name, ".")) {
				if (dired_show_dotfiles & 1)
					continue;
				specialcnt++;
			} else if (!strcmp(name, "..")) {
				if (dired_show_dotfiles & 2)
					continue;
				specialcnt++;
			} else if (!dired_isdir()) {
				if (dired_show_dotfiles & 8)
					continue;
				filecnt++;
			} else {
				if (dired_show_dotfiles & 4)
					continue;
				dircnt++;
			}
			delete(give_begin_line() - 1, give_end_line());
		}
		check_abort();
	}
	point = size();
	if (character(point - 1) == '\n' && (filecnt || dircnt)
		  && !(dired_show_dotfiles & 16))
		bprintf("  %d file%s, %d director%s omitted due to "
				"dired-show-dotfiles\n",
				filecnt, filecnt == 1 ? "" : "s",
				dircnt + specialcnt, dircnt + specialcnt == 1 ? "y" : "ies");
}

color_dired_range(from, to)
{
	int color, namecolor, old, namestart, nameend, isdir;
	char buf[FNAMELEN];

	if (from >= to)
		return to;
	detect_dired_format();
	save_var point, matchstart, matchend;
	point = from;
	save_var narrow_start = give_begin_line();
	point = to;
	nl_forward();
	save_var narrow_end = size() - point;
	point = 0;
	while (point < size() - narrow_end) {
		switch (curchar()) {
			case 'D': color = color_class dired_deleted; break;
			case 'U': case 'L': case 'R': case 'G':
			case 'M': color = color_class dired_moved; break;
			case 'C': color = color_class dired_copied; break;
			default: color = color_class dired_normal; break;
		}
		old = point;
		if (parse_string(1, dired_skip_line_pat()) != 0) {
			color = color_class dired_header;
			namestart = matchend;
			nameend = give_end_line();
			namecolor = color_class dired_dir_name;
		} else {
			dired_getindex(buf, goal_column);
			namestart = matchstart;
			nameend = matchend;
			namecolor = color_class dired_file_name;
			if (dired_format == DF_VMS)
				nameend = point + parse_string(1, "[^\n;]*;?[0-9]*");
			isdir = dired_isdir();
			if (isdir) {
				namecolor = color_class dired_dir_name;
				if (!dired_format)
					namestart = column_to_pos(ISDIR_START);
				if (isdir == 3 && parse_string(1, ".* -> "))
					nameend = matchend;
			}
			if (!is_remote_file(get_dired_buffer_pattern())
				  && character(namestart - 1) == ' ')
				namestart--;
		}
		nl_forward();
		set_character_color(old, point - 1, color);
		set_character_color(namestart, nameend, namecolor);
	}
	return point;
}

detect_dired_format()
{
	int old = dired_format, i;

	goal_column = NAME_START;
	save_var point = 0, case_fold = 1;
	dired_format = DF_VMS;		// Skip VMS lines too for now.
	while (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	dired_format = 0;
	if (point >= size())
		return;
	if (parse_string(1, ".*" UNIX_DATE)) {
		dired_format = DF_UNIX;
		goal_column = get_column(matchend);
	} else if (parse_string(1, ".*" OLDNT_DATE)) {
		dired_format = DF_OLDNT;
		goal_column = 41;
	} else if (parse_string(1, ".*" VMS_DATE)) {
		dired_format = DF_VMS;
		goal_column = 2;
	} else if (!parse_string(1, ". [ 0-9][0-9]-[a-z?][a-z?][a-z?]-[0-9][0-9]")
			   && !parse_string(1, ". [0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]")) {
		dired_format = DF_SIMPLE;
		goal_column = 2;
	} else {
		for (i = 1; i <= 8; i <<= 1)
			if (dired_layout & i)
				goal_column++;
	}
	if (dired_format != old)
		drop_all_colored_regions();
}

sort_for_dired()
{
	int start, done = 0, chat = size() - narrow_end - narrow_start > 100000;

	if (dired_format == DF_VMS)
		return;		// Sorting not supported.
	if (chat)
		noteput("Sorting...");
	save_var modified;
	window_start = point = 0;
	save_var abort_searching = 0;
	while (!done) {
		nl_forward();	/* skip header line */
		to_begin_line();
		start = point;
		if (re_search(1, dired_skip_line_pat()))
			to_begin_line();
		else
			done = 1;
		if (start < point - 1)
			dired_sorter(start, point - (parse_string(-1, "\n\n") ? 1 : 0));
	}
	point = 0;
	dired_down_line();		/* position on 2nd line */
	window_start = 0;
	if (chat)
		say("");
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "(sh|csh|bash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], *(*finder)();
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	char extra[FNAMELEN];
	spot start;

	flags |= (inbufs ? BUFGREP : FILEGREP);
	save_var synchronous_file_io = 1;
	noteput("0 matches, searching...");
	before_head = prepare_grep_buffer(buf, flags, pat, filepat, NULL);
	start = alloc_spot(0);
	save_var modified;
	save_var abort_file_matching = ABORT_ERROR;
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning, no_popup_errors = 1, abort_file_io = ABORT_ERROR;
	if (grep_on_changed_file & 1)
		readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	save_var use_file_variables &= ~2;	// Turn off .epsilon_vars unless
	if (use_file_variables & 8)			// we have additional flag.
		use_file_variables |= 2;
	if (file_read_kibitz & 2)
		save_var file_read_kibitz = 0;
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (user_abort) {
			aborted = 1;
			break;
		}
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("%d matches, searching %f...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = std_find_it(abs);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		*extra = 0;
		if (grep_include_timestamp && !inbufs && file_info.check_type == CHECK_FILE)
			if (grep_include_timestamp == 2)
				sprintf(extra, "[%04d-%02d-%02d %02d:%02d:%02d] ",
						file_info.year, file_info.month, file_info.day,
						file_info.hour, file_info.minute, file_info.second);
			else
				sprintf(extra, "[%02d-%.3s-%04d %02d:%02d:%02d] ",
						file_info.day,
						(file_info.month >= 1 && file_info.month <= 12)
						? months + (file_info.month - 1) * 4 : "???",
						file_info.year,
						file_info.hour, file_info.minute, file_info.second);
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel, extra);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			update_grep_title(before_head, cnt);
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		point = i;
		if (skipped) {
			say("Skipped %d file%s due to grep-ignore-file- variables.",
				  skipped, (skipped == 1) ? "" : "s");
			bprintf("%pSkipped %d file%s due to grep-ignore-file- variables.\n",
					color_class grep_header,
					skipped, (skipped == 1) ? "" : "s");
		}
	}
}

do_file_replace(str, with, flags, filepat, inbufs)
char *str, *with, *filepat;	// (query-) replace across files/buffers
{
	int res, oldbuf = 1, prevbuf = 0, total = 0, cnt = 0;
	char rel[FNAMELEN], abs[FNAMELEN], *s, *(*finder)(), msg[FNAMELEN];
	int total_found = 0, total_changed = 0, total_readonly = 0, old_readonly;

	save_var synchronous_file_io = 1;
	old_readonly = readonly_warning;
	save_var readonly_warning;			// Don't show msg or ring bell
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL); // unless matches.
	save_var abort_file_matching = ABORT_ERROR;
	save_var bufnum, window_bufnum;
	if (file_read_kibitz & 4)
		save_var file_read_kibitz = 0;
	finder = give_grep_finder(inbufs, filepat, &s);	// Get 1st match
	for (; s; s = (*finder)(filepat, 0)) {
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs))
			continue;
		noteput("Searching %s...", rel);
		if (!inbufs) {
			oldbuf = look_file(abs);
			if (std_find_it(abs) == EREADABORT)
				user_abort = 1;
		} else
			to_buffer(s);
		undo_mainloop();
		if (!user_abort) {
			res = one_file_replace(str, with, flags, oldbuf, old_readonly);
			total_found += replace_num_found;
			total_changed += replace_num_changed;
			total++;
			if (res == -2)
				total_readonly++;
			else if (res >= 0)
				cnt++;
		} else {
			res = DSABORT;
			user_abort = 0;
		}
		if (prevbuf && buf_in_window(prevbuf) == -1) {
			buf_delete(prevbuf);	// delete untouched new bufs
			prev_forget(_prev_num - 1); // fix default
		}
		prevbuf = (oldbuf || modified) ? 0 : bufnum;
		if (res == 0 && (flags & QUERY) && key == '!'
			&& ask_continue_replace("Replace without asking"))
			flags &= ~QUERY;	// no more querying
		else if (res == DSABORT || res == 1 && ungot_key != -1) {
			if (!ask_continue_replace("Continue replacing"))
				break;
		} else if (res > 0) {	// User said to stop, or bad pattern.
			if (res == DSBAD)
				show_replace(str, with, flags | BAD);
			user_abort = 0;
			return;
		}
		ungot_key = -1;
		if (user_abort)
			break;
	}
	ungot_key = -1;
	user_abort = 0;
	sprintf(msg, "Replaced %d of %d matches in %d of %d %s",
	    total_changed, total_found, cnt, total,
	    inbufs ? "buffers" : "files");
	if (total_readonly)
		sprintf(msg + strlen(msg), ", %d read-only, see %s buffer",
				total_readonly, MSG_BUFFER);
	say("%s.", msg);
	restore_vars();
	if (prevbuf && buf_in_window(prevbuf) == -1) {
		buf_delete(prevbuf);	// delete untouched new bufs
		prev_forget(_prev_num - 1); // fix default
	}
}

mshelp2_check_std_name(char *name, char *col)
{
	point = 0;
	if (search(1, name)) {
		grab_full_line(bufnum, col);
		return 1;
	}
	return 0;
}

// If there's a collection with a date, use it.  Prefer later dates to
// earlier ones.
mshelp2_choose_collection(char *col)
{
	if (*mshelp2_collection) {		// Override default selection.
		strcpy(col, mshelp2_collection);
		return;
	}
	*col = 0;
	if (!mshelp2_wait_until_pat("Using default collection.*\n"))
		return;
	point = matchend;
	bprintf("listcoll\n");
	int before = size();
	if (!mshelp2_wait_until_pat("^--End\n"))
		return;
	int tmp = tmp_buf();
	buf_xfer(tmp, before, matchstart);
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	case_fold = 1;
	if (mshelp2_check_std_name("MS.VSCC.v90", col)
		  || mshelp2_check_std_name("MS.VSCC.v80", col)
		  || mshelp2_check_std_name("MS.VSCC.2003", col)
		  || mshelp2_check_std_name("MS.VSCC", col))
		return;
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ", pat[100];
	int year, month;
	for (year = 2020; year > 2000; year--) {
		for (month = 12; month > 0; ) {
			sprintf(pat, "qtr.%d%.3s", year, months + --month * 4);
			point = 0;
			if (search(1, pat)) {
				grab_full_line(bufnum, col);
				return;
			}
		}
		sprintf(pat, "%d", year);
		point = 0;
		if (search(1, pat)) {
			grab_full_line(bufnum, col);
			return;
		}
	}
}

color_html_attributes(int script, int orig, int limit)
{
	int len, start, lang;

	if (script == 2)		// A <jsp:> tag.
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x10, HT_JSCRIPT));
	else if (script == 3)
		lang = HT_CSS;
	else if (script == 4)
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x20, HT_JSCRIPT));
	else if (script)
		lang = get_html_asp_language(html_other_coloring);
	for (;;) {
		if (curchar() == '<' && parse_string(1, "%<%!---")) {
			point = matchend;
			color_html_comment();
			re_search(1, "[ \t\n]*");	// Parse whitespace.
		}
		if (!(len = parse_string(1, "[-_./a-z0-9]*")))	// An attribute?
			break;
		if (script == 1 && parse_string(1, "(language|type)[ \t\n]*=[ \t\n]*"
										"[\"']*(text/)?")) {
			start = matchend;
			if (matches_at(start, 1, "(javascript|jscript|ecmascript)"))
				lang = HTML_SPECIFIC_COLORING(1, HT_JSCRIPT);
			else if (matches_at(start, 1, "(vbscript)"))
				lang = HTML_SPECIFIC_COLORING(2, HT_VBSCRIPT);
			else if (matches_at(start, 1, "(php)"))
				lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
			else if (matches_at(start, 1, "(python)"))
				lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
		}
		if (character(point + len) == ':'
			  && matches_at(point + len, 1, ":[-_./a-z0-9]*")) {
			set_character_color(point, point + len + 1, color_class html_namespace);
			point += len + 1;
			len = matchend - point;
		}
		set_character_color(point, point + len, color_class html_attribute);
		point += len;			// Followed by a value?
		if (len = parse_string(1, "[ \t\n]*=[ \t\n]*")) {
			set_character_color(point, point + len, color_class html_punctuation);
							// Color the = and whitespace.
			point += len;		// Value might be quoted.
			start = point;
			if (curchar() == '"') {
				point++;
				html_find_matching_quote('"');
			} else if (curchar() == '\'') {
				point++;
				html_find_matching_quote('\'');
			} else	// No quotes, assume it's whitespace-delimited.
				re_search(1, "[^ \t\n<>]*");
			html_color_value(start, point);
		}
		re_search(1, "[ \t\n]*");	// Parse whitespace.
		set_character_color(matchstart, matchend, color_class html_punctuation);
	}
	if (curchar() == '>') {
		set_character_color(point, point + 1, color_class html_punctuation);
		point++;
	}
	start = point;
	if (script && lang != HT_IGNORE && !parse_string(-1, "/%>")) {
		if (re_search(1, script == 3 ? "%</style%>"
					  : script == 2 ? "%</jsp:.*%>" : "%</(cf)?script(text)?%>")) {
			set_tagged_region("color-as-unit", orig, point, 99);
			point = matchstart;
		} else
			set_tagged_region("color-as-unit", orig, point, 99);
		color_embedded_code(start, point, lang, limit);
	}
}

color_html_move_back()
{
	int start = point, p;

	save_var narrow_start = point - color_html_look_back;
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>")
		  && !parse_string(-1, "%<script[^<]*%><wspace>*"))
		if (curchar() == '<')
			return 1;
		else {
			point += parse_string(1, "%]%]%>");
			return -1;
		}
	point = start;
	if (re_search(-1, "<Langle>[ \t\n]*/?((cf)?script|style)")
		&& parse_string(1, "<Langle>[ \t\n]*((cf)?script|style)"))
		return 2;				// Found a <script>, not a </script>.
	point = start;
	if (re_search(-1, "<Langle>[%?]|[%?]<Rangle>") && curchar() == '<'
		  && get_tagged_region("color-as-unit", point) == -1)
		return 3;				// Found a <% or <?, not a %> or ?>.
	point = start;
	do {			// start coloring from the previous tag
		if (!search(-1, "<"))
			break;	// but ignore <'s before " or ' characters
		if (get_character_color(point, &p, NULL) == color_class html_comment) {
			point = p + 2;	// If we're in a comment,
			search(-1, "<!");	// go back to its start.
		}
	} while (html_in_quoted_attribute());
	return 0;
}

color_html_loop(int end)
{
	int s, t, i, script;

	for (s = point; point < end; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							 ? "([#a-z0-9]+);?"
							 : "([a-z]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()) && color_html_include(end))
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = (s == 6 && parse_string(1, "script")
					  || s == 10 && sublanguage_code == SUBLANG_XML
					  && parse_string(1, "scripttext"));
			if (s == 3 && parse_string(1, "jsp:(declaration|expression"
									   "|scriptlet)</word>"))
				script = 2;
			if (s == 5 && parse_string(1, "style"))
				script = 3;
			if (s == 8 && parse_string(1, "cfscript"))
				script = 4;
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			for (i = point; strchr(" \t\n", curchar()); )
				point++;	// Parse whitespace.
			set_character_color(i, point, color_class html_punctuation);
			if (parse_string(1, "/[ \t\n]*>")) {
				set_character_color(point, point + 1, color_class html_tag);
				set_character_color(point + 1, matchend,
									color_class html_punctuation);
				point = matchend;
			} else
				color_html_attributes(script, t, end);
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
}

// Narrowing may not have excluded the <script line itself, so do it
// now.
html_exclude_script_head()
{
	save_spot point = narrow_start;
	if (parse_string(1, "<langle>[%?][A-Za-z0-9]*"))
		point = matchend;
	else if (parse_string(1, "<Langle>[ \t\n]*((cf)?script|style)"))
		search(1, ">");
	if (parse_string(1, "%<%!--"))
		point = matchend;
	if (parse_string(1, "[ \t\n]*%<%!%[CDATA%["))
		point = matchend;
	narrow_start = point;
}

// Flags: 1 to treat elements in html_no_indent_elements as
// self-terminating.
int html_is_start_tag(int pos, int flags)
{
	char tag[FNAMELEN];
	int i, is_end = 0;

	save_var point = pos, case_fold = (sublanguage_code != SUBLANG_XML);
	save_var narrow_end = 0;	// Use only new limit.
	narrow_end = size() - narrow_position(point + html_tag_match_look_back);
	i = parse_string(1, "%<[ \t\n]*");
	if (!i)
		return -2;
	point += i;
	if (curchar() == '/')			// An end tag.
		is_end = 1, point++;
	point += parse_string(1, "[ \t\n]");	// Get the tag name.
	if (!parse_string(1, "[-_./A-Za-z0-9:]+", tag))
		return -2;
	if (is_end)
		return (flags & 1) && word_in_list(tag, html_no_indent_elements, 1)
				? 0 : -1;
	if (do_color_searching(REGEX, "<c:html-punctuation>%>") != 1)
		return -2;		// Find the closing >.
	if (parse_string(-1, "/%>"))
		return 0;		// It's marked as self-terminating.
	if (sublanguage_code == SUBLANG_XML || (html_style_rules & 1))
		return 1;		// In XML, all self-terminating tags must be marked.
	// In HTML, we must check more.
	if (word_in_list(tag, html_empty_elements, 1))
		return 0;
	if (tolower(tag[0]) == 'c' && tolower(tag[1]) == 'f'
		  && word_in_list(tag + 2, coldfusion_empty_elements, 1))
		return 0;
	if ((flags & 1) && word_in_list(tag, html_no_indent_elements, 1))
		return 0;	// Assume start tags are optional for these.
	if (!strfcmp(tag, "p") && !html_check_paragraph_is_container())
		return 0;
	if (!strfcmp(tag, "option") && html_open_option(pos) >= 0)
		return 0;
	if (!strfcmp(tag, "li")) {
		// These rarely have end tags, so be lenient.
		if (html_open_li(pos) >= 0 || html_find_open_tag(pos + 1,
			"li|ol|ul|dir|menu",
			"li|ol|ul|dir|menu|/ol|/ul|/dir|/menu", 0) >= 0)
			return 0;
	}
	if ((!strfcmp(tag, "dd") || !strfcmp(tag, "dt"))
		  && html_open_dd_dt(pos) >= 0)
		return 0;
	return 1;
}

suffix_cfc()
{
	html_mode();
}

suffix_rss()
{
	xml_mode();
}

suffix_xsdconfig()
{
	xml_mode();
}

drag_drop_handler() on reg_tab[WIN_DRAG_DROP]
{
	char buf[FNAMELEN], *s, *t, dir[FNAMELEN], orig_dir[FNAMELEN];
	int line = -1, column = -1, j, dir_flags = directory_flags, first = 1;
	int had_action = 0, next_is_dir = 0, pos = -1, all_files = 0;

	tiled_only();
	if (mouse_screen >= 0)
		to_mouse_window(mouse_x, mouse_y, mouse_screen);
	while (j = drag_drop_result(buf)){
		if (!strcmp(buf, "-dir"))
			next_is_dir = 1;
		else if (*buf == '+') {			// a +linenum parameter
			if (buf[1] != 'p') {
				line = numtoi(buf + 1);
				s = strchr(buf, ':');
				if (s)
					column = numtoi(s + 1);
			} else
				pos = numtoi(buf + 2);
		} else if (*buf == '-' && !all_files) {
			switch (buf[1]) {
				case 'd':
					if (buf[2])
						t = buf + 2;
					else if (!drag_drop_result(t = buf)) {
						say("-d flag missing variable name.");
						break;
					}
					s = index(t, '=');
					if (!s)		// accept either VAR=VAL or VAR!VAL
						s = index(t, '!');
					if (!s)
						break;
					*s = 0;
					if (j = find_index(t)) {
						set_var_val(j, s + 1, 1);
						continue;
					} else
						say("No such variable %s.", t);
					break;
				case 'l':
					if (buf[2])
						load_commands(buf + 2);
					else if (drag_drop_result(buf))
						load_commands(buf);
					else
						say("-l flag missing file name.");
					break;
				case 'r':
					had_action = 1;
					if (buf[2])
						t = buf + 2;
					else if (!drag_drop_result(t = buf)) {
						say("-r flag missing function name.");
						break;
					}
					s = index(t, '=');
					if (!s)		// accept either CMD=ARG or CMD!ARG
						s = index(t, '!');
					if (s)
						*s = 0;
					if (j = find_index(t)) {
						if (s)
							(*(int (*)())j)(s + 1);
						else
							(*(int (*)())j)();
					} else
						say("No such function %s.", t);
					break;
				case 'w':
					dir_flags = numtoi(buf + 2);
					break;
				case '-':
					all_files = 1;
					break;
			}
			continue;
		} else if (*buf) {
			if (next_is_dir) {
				getcd(orig_dir);
				on_exit chdir(orig_dir);
				chdir(buf);
				getcd(dir);
				next_is_dir = 0;
				continue;
			}
			windows_foreground();
			had_action = 1;
			convert_cygwin_filename(buf, 0);
			absolute(buf);
			if (is_directory(buf) || is_remote_file(buf) || (is_pattern(buf) & 2))
				do_find(buf, get_default_translation_type(buf));
			else
				find_it(buf, get_default_translation_type(buf));
			if (line >= 0)
				go_line(line);
			if (pos >= 0)
				point = pos;
			if (column >= 0)
				move_to_column(column);
			line = column = pos = -1;
			if ((dir_flags & DIR_CWD_FROM_FILE) && first) {
				strcpy(dir, buf);
				*get_tail(dir, 1) = 0;
				chdir(dir);
			}
			first = 0;
			if (j == 2)
				show_binding("Use %s to release the Epsilon client when you finish.",
							 "resume-client");
		}
	}
	if (!had_action)			// Ensure we're not minimized, unless we got
		windows_foreground();	// a -r command but no file.
}

// Does this region have unmatched ), ] or }?
int python_close_delimiter(int start, int end)
{
	save_var narrow_start = start;
	save_var point = end;
	return re_search(-1, "[])}]!") && get_character_syntax_color(point - 1) == -1
			&& default_move_level(-1, "])}", "[({") == -1;
}

// If this is a continuation line, move back one line and return 1.
python_back_over_continuation_line(int orig)
{
	if (parse_string(-1, "\\\n[ \t]*"))
		point = matchend;
	else if (!python_close_delimiter(give_begin_line(), orig) || !nl_reverse())
		return 0;
	return 1;
}

do_python_indent()
{
	int ind, orig = give_begin_line(), close, last;

	close = matches_at(orig, 1, "[ \t]*[])}]");
	save_var point;
	do {		// Find prev non-comment or non-blank line.
		if (!nl_reverse())		// First line in the file, no indent.
			return 0;
		to_begin_line();
	} while (parse_string(1, python_indent_to_comment ? "[ \t]*\n"
		: "[ \t]*[#\n]"));
	last = point;
	while (python_back_over_continuation_line(orig))
		;
	to_indentation();
	ind = current_column();
	recolor_buffer_range(point, orig);
	// Does this (previous) line end with a : character (plus maybe a
	// comment)?  Or does the previous line end in \, or is there an
	// open delim on prev line?  If so, indent one level more.
	if (matches_at(last, 1, ".*<^c:python-*>:[ \t\n]*(#.*)?$")
		  || matches_at(orig, -1, "\\\n")
		  || python_open_delimiter(point, orig))
		ind += python_indent;
	else if (close)
		ind -= python_indent;
	restore_vars();
	indent_to_column(ind);
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class") || !strcmp(func, "__interface"))
		return tag_c_struct(1);		// or similar class or managed c++ def
	else if (!strcmp(func, "interface"))
		return tag_c_struct(1);		// for java or managed c++
	else if ((c_extra_keywords & OBJC_KEYWORDS) && character(start - 1) == '@'
			 && word_in_list(func, "|interface|implementation|protocol|", 0))
		return tag_objc_block();
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if ((c_extra_keywords & CPP_KEYWORDS) && !strcmp(func, "using")) {
		tag_modifier_kwd |= 4;	// just a declaration 
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (!strcmp(func, "__attribute__") && parse_string(1, "<wspace>*%(%(")
			   && !skip_c_delims('(', ')', 1)) {
			return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	if (curchar() == '_' && parse_string(1, "(_ANSI_ARGS_|__P)</word>")) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template();
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
			"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (tag_c_variable()
			   && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
				   || (tag_modifier_kwd & 2))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}
