/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B14 to 13 B15.

#include "eel.h"
#include "vspell.h"
#include "html.h"
#include "colcode.h"
#include "c.h"
#include "grep.h"
#include "kill.h"

c_indenter()		/* replace surrounding indentation with new */
{
	if (c_extra_keywords & IDL_KEYWORDS)
		return;
	if (in_shell_buffer)
		prev_indenter();
	else if (!(in_c_comment(c_look_back) & IN_COMMENT))
		indent_to_column(c_compute_indent());
	else if ((reindent_c_comments
			  & (indenter_action == IACT_REINDENT_PREV ? 4 : 1))
			 && !in_perl_buffer)
		indent_c_comment();
}

suffix_h()
{
	c_mode();
	c_extra_keywords = CPP_KEYWORDS | EEL_KEYWORDS;
	// recognize all keywords; can't tell if it's C/C++/EEL from extension
	save_var point = 0, narrow_start = 0, narrow_end, case_fold = 0;
	narrow_end = size() - c_look_back;
	if (re_search(1, "^@(" OBJC_TOPBLOCKS ")"))
		c_extra_keywords |= OBJC_KEYWORDS;
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */
	int in_initializer;

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!(reindent_c_comments & 1) && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if (!reindent_one_line_c_comments && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' && reindent_c_preprocessor_lines == 1
		  || first_char == '\f'
		  || first_char == '@' && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(1, "@(" OBJC_TOPBLOCKS ")[ \t]+<word>|@end"))
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (parse_string(-1, PREPROC_ELSE ".*\n")) {
		point = matchend;
		locate_by_ifdef(-1);
	}
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig) || c_align_break_with_case
		  && matches_at(orig, 1, "[ \t]*break;")) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig) && !((c_extra_keywords & OBJC_KEYWORDS)
									 && matches_at(pstart, 1, "^[-+@]"))) {
		char type[FNAMELEN];		// Could it be a bitfield, not a label?
		find_c_func_info(type, NULL, NULL, 0);
		if (!strcmp(type, "struct") || !strcmp(type, "union"))	// Yes, a bitfield.
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_initializer = in_c_initializer(orig)) {
		return ind;
	} else if ((c_extra_keywords & OBJC_KEYWORDS)
			   && objc_next_no_indent(orig, pstart)) {
		return 0;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, in_initializer);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

int objc_next_no_indent(int orig, int pstart)
{
	int first_char = character(orig);
	if (strchr("-+", character(pstart)) && strchr("-+", first_char)
		  && find_objc_class(NULL, NULL))
		return 1;
	save_var point = orig;
	cind_space();
	to_begin_line();
	if (parse_string(1, "@(" OBJC_TOPBLOCKS ")|@end"))
		return 1;
	return 0;
}

is_contin_statement(p, fd, orig) // would line after p be a continuation line?
{			// assumes p is just after last interesting char
	// if fd, pretend func param decls are a "continuation" of func decl
	// orig is later position in buffer, for unclosed "for" scan
	switch (character(p - 1)) {
		case ';':		// check for multi-line "for" statemnt
			if (unmatched_rparen(p, orig))
				return 1;
		case '{':
		case '}':
			return 0;
		case ')':
			save_var point = p;
			save_var narrow_start = p - 500;
			if (!c_rev_skip_level('(', ')'))
				return 1;
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && parse_string(-1, "@(synchronized|catch)"))
				return 0;
			if (sublanguage_code == SUBLANG_PHP)
				return !parse_string(-1, "[^a-z0-9_](if|while|for"
									 "|switch|foreach)[ \t\f]*")
						&& (fd || !Topindent || get_indentation(point));

			return !parse_string(-1, "[^a-z0-9_](if|while|for"
								 "|switch)[ \t\f]*")
					&& (fd || !Topindent || get_indentation(point))
					|| in_perl_buffer;
		case 'e':
			return !is_else(p - 4);
		case 'o':
			save_var point = p;
			return !parse_string(-1, "[^a-z0-9_]do");
		case 'y':
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && matches_at(p, -1, "@(try|finally)"))
				return 0;
			return 1;
		case ':':
			save_var point = p;
			to_indentation();
			return !is_label(point);
		default:
			return 1;
	}
}

is_c_keyword(p)		// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|if|else|for|do|while|return|break|continue|switch|"
		   "case|default|goto|char|short|int|static|volatile|const|"
		   "float|double|long|void|", p))
		return 1;
	if (p[1] == '_' &&
	    strstr("|_asm|_based|_cdecl|_emit|_except|_export|_far|_fastcall|"
		   "_fortran|_huge|_interrupt|_loadds|_near|_pascal|_saveregs|"
		   "_segment|_segname|_stdcall|_syscall|_try|__int64|", p))
		return 1;
	if ((c_extra_keywords & EEL_KEYWORDS) &&
	    strstr("|keytable|buffer|window|command|on|color_class|"
		   "spot|save_spot|save_var|on_exit|user|zeroed|byte|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
	    strstr("|asm|catch|class|delete|friend|inline|new|operator|"
			   "private|protected|public|template|try|this|virtual|"
			   "throw|_self|and|and_eq|bitand|bitor|bool|compl|const_cast|"
			   "dynamic_cast|explicit|false|mutable|namespace|not|not_eq|"
			   "or|or_eq|reinterpret_cast|static_cast|true|typeid|typename|"
			   "using|wchar_t|xor|xor_eq|", p))
		return 1;
	if ((c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|abstract|boolean|byte|catch|class|extends|false|final|"
			   "finally|function|implements|import|in|instanceof|"
			   "interface|native|new|null|package|private|protected|"
			   "public|super|synchronized|this|throw|throws|transient|"
			   "true|try|var|with|", p))
		return 1;
	if (!(c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|auto|enum|extern|register|unsigned|signed|sizeof|"
			   "struct|union|typedef|", p))
		return 1;		// C keywords not in Java.
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|appobject|auto_handle|bindable|boolean|broadcast|byte|"
			   "callback|coclass|code|comm_status|context_handle|"
			   "control|decode|defaultbind|dispinterface|displaybind|"
			   "dual|enable_allocate|encode|error_status_t|"
			   "explicit_handle|fault_status|handle|handle_t|hidden|"
			   "hyper|idempotent|ignore|immediatebind|import|in|"
			   "include|interface|library|licensed|local|maybe|module|", p))
		return 1;
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|ms_union|nocode|nonextensible|notify|object|odl|"
			   "oleautomation|optional|out|pipe|propget|propput|"
			   "propputref|public|ptr|readonly|ref|requestedit|"
			   "restricted|retval|small|source|string|unique|"
			   "usesgetlasterror|v1_enum|vararg|", p))
		return 1;
	if ((c_extra_keywords & CSHARP_KEYWORDS) &&
		  strstr("|abstract|as|base|bool|byte|catch|checked|decimal|delegate|"
				 "event|explicit|false|finally|fixed|foreach|implicit|in|"
				 "interface|internal|is|lock|namespace|new|null|object|"
				 "operator|out|override|params|private|protected|public|"
				 "readonly|ref|sbyte|sealed|sizeof|stackalloc|string|this|"
				 "throw|true|try|typeof|uint|ulong|unchecked|unsafe|ushort|"
				 "using|virtual|", p))
		return 1;
	if ((c_extra_keywords & OBJC_KEYWORDS) &&
		  strstr("|self|super|BOOL|id|Class|nil|YES|NO|NS_DURING|NS_HANDLER|"
				 "NS_ENDHANDLER|in|out|inout|bycopy|byref|oneway|@interface|"
				 "@implementation|@end|@class|@private|@protected|@public|"
				 "@encode|@protocol|@selector|@try|@catch|@finally|@throw|"
				 "@synchronized|Nil|IBAction|IBOutlet|", p))
			return 1;
	return 0;
}

full_comp_read(trystr, prmpt, finder, flags, edit, def)
char *trystr, *prmpt, *def;
char *(*finder)();
{
	int i, j, num_match = 0, maxlen = ptrlen(trystr) - 1, pbuf, res;
	int pwin, editwin, owin = window_handle, recall_cnt = 0;
	int olen = 0, in_menu = 0, menu = -1, mbuf = 0;
	jmp_buf this_level;
	int *(*lister)();
	char password_display_class[256];

	save_var last_index;
	save_var _help_on_cmd = last_index;
	save_var _doing_input = DI_LINEINPUT;
	save_var _doing_completion = (finder != 0);
	save_var top_level = &this_level;

	while (!in_macro() && !char_avail() && unseen_msgs())
		check_abort();	// Wait for unseen msgs to display so we don't cover them.
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
		pbuf = tmp_buf());	/* make system window with prompt */
	stuff(prmpt);
	i = current_column();
	system_window = 1;	/* don't let user switch to it */
	point = display_column = 0;
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;

	editwin = add_popup(i, echo_line, screen_cols - i, 1,
							0, edit);
	display_column = 0;	/* horizontally scroll this window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	if (flags & PASSWORD_PROMPT) {
		set_chars(password_display_class, '*', 256);
		_display_class = password_display_class;
		build_first = 1;
		save_var disable_recording_last_keys = 1;
	}
	mouse_handler = allow_mouse_switching;
	if (insert_default_response == 1) {
		highlight_on();
		shift_selecting = 2;
	}
	if ((flags & POP_UP_PROMPT) && (has_feature & FEAT_GUI_DIALOGS)) {
		remove_window(pwin);
		do_display_dialog_box("OneLineBox", prmpt, editwin, 0, 0,
							  "OK", "Cancel", "Help", NULL, NULL);
	}
	maybe_refresh();
	if (setjmp(top_level)) {
		if (in_menu == 2) {	/* menu is half-created */
			remove_window(menu);
			buf_delete(mbuf);
			in_menu = mbuf = 0;
			if (is_gui && want_gui_prompts) {
				res = INP_ABORT;
				goto done;
			}
		}
	}
	for (;;) {
		if (mbuf) {
			window_handle = menu;
			if (in_menu && line_gets_highlight())
				make_line_highlight();
			else
				remove_line_highlight();
		}
		window_handle = in_menu ? menu : editwin;
		mode_keys = in_menu ? menu_tab : comp_tab;
		_doing_input = (response_tracks_list && mbuf)
					   ? DI_LISTBOX : DI_LINEINPUT;
		if (_doing_input == DI_LISTBOX) {
			maybe_refresh();
			getkey();
			cmd_result = 0;
			run_topkey();
			res = cmd_result;
		} else
			res = do_recursion();	/* let user edit input */
		if (!is_window(editwin) || in_menu && !is_window(menu)) {
			res = INP_ABORT;
			break;
		} else if (window_handle == (in_menu ? editwin : menu))
			in_menu = !in_menu;
		window_handle = editwin;
		if (res == INP_ADD) {
			window_handle = editwin;
			run_topkey();
			if (search_in_menu && mbuf) {
				window_handle = editwin;
				grab(0, maxlen, trystr);
				if (strlen(trystr) >= olen) {
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			} else
				in_menu = 0;
		} else if (res == INP_SEEN)
			in_menu = 0;
		else if (res == INP_TO_LIST && mbuf) {
			window_handle = menu;
			in_menu = 1;
		} else if ((res == INP_GET || res == INP_GET_EXIT) && mbuf) {
			window_handle = menu;
			get_menu_line(edit);
			window_handle = editwin;
			in_menu = 0;
		} else if (!res) {
			if (_doing_input == DI_LISTBOX)
				if (in_menu) {
					window_handle = menu;
					get_menu_line(edit);
				} else {
					window_handle = editwin;
					grab(0, maxlen, trystr);
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			continue;
		}
		restart_file_paths();
		if (res == INP_ABORT || res == INP_GET_EXIT || res == INP_EXIT)
			break;
		else if (res == INP_GRAB_WORD) {
			grab_buffer_word(owin, 0);
			continue;
		} else if (res == INP_GRAB_LINE) {
			grab_buffer_word(owin, 1);
			continue;
		} else if (res == INP_DEFAULT) {
			if (!size())	/* type in default text */
				stuff(def);
			in_menu = 0;
			continue;
		} else if (res == INP_LAST_RECALL || res == INP_NEXT_RECALL
				   || res == INP_PREV_RECALL) {
			if (res == INP_LAST_RECALL)
				get_last_recall(finder, trystr);
			else
				recall_cnt = get_last_recall_from_end(finder, trystr,
					recall_cnt + (res == INP_NEXT_RECALL ? -1 : 1));
			if ((*trystr || res != INP_LAST_RECALL) && !mbuf) {
				delete(0, size());
				stuff(trystr);
				if (recall_response_selected && *trystr) {
					highlight_on();
					shift_selecting = 2;
				}
			}
			in_menu = 0;
		} else if (res == INP_GET_OTHER) {
			if (inp_get_other_name && *inp_get_other_name) {
				delete(0, size());
				stuff(inp_get_other_name);
			}
			in_menu = 0;
		} else if (res == INP_SHOW_MATCHES || res == INP_RECALL) {
show_matches:		grab(0, maxlen, trystr);	/* typed '?' */
			if (is_highlight_on() || res == INP_RECALL) {
				if (flags & COMP_FILE)
					*get_tail(trystr, 0) = 0;
				else
					*trystr = 0;
				highlight_off();
			}
			olen = strlen(trystr);
			if (!mbuf) {
				menu = new_menu(mbuf = tmp_buf());
				if (is_gui && want_gui_prompts) {
					remove_window(pwin);
					say("");
					make_prompt_box(res == INP_RECALL, prmpt, editwin, menu, finder);
				}
			}
			in_menu = 2;		/* initializing */
			if (res == INP_RECALL)
				lister = list_recalls;
			else if (completion_lister)
				lister = completion_lister;
			else
				lister = list_matches;
			make_menu(menu, trystr, finder, flags, editwin, mbuf, lister);
			in_menu = 1;
			mode_keys = menu_tab;
			cursor_to_column = -1;
		} else if (res == INP_TRY || res == INP_COMPLETE) {
			if (res == INP_TRY && cursor_to_column
						== current_column() + 1)
				break;	/* exactly 1 match last time */
			grab(0, maxlen, trystr);
			i = strlen(trystr);
			num_match = complete(trystr, finder, flags);
			j = strlen(trystr);
			if ((flags & SPACE_VALID) && key == ' ' && trystr[i] == ' '
				  && could_add_space(trystr, finder, flags) == 1)
				i++, num_match++;
			else if (num_match && i != j)
				i = j;
			else if (num_match != 1) {
				if ((flags & SPACE_VALID) && key == ' ' && is_window(menu)
//					  && trystr[i] == ' '
// Given dirs c:\program files and c:\programs, this commented-out code
// made selecting the former fail.
					  && could_add_space(trystr, finder, flags)) {
					trystr[i++] = ' ';
				} else {
					maybe_ding(bell_on_completion);
					if (completion_pops_up)
						ungot_key = '?';
				}
			}
			if ((flags & COMP_FILE) && num_match == 1 && i > 0
			    && is_path_separator(trystr[i - 1]))
				num_match++;	/* might be directory */
			trystr[i] = 0;
			buf_zap(edit);
			stuff(trystr);
			if (num_match == 1) {
				if (res == INP_COMPLETE)
					break;	/* <Esc> worked */
				maybe_refresh();
				cursor_to_column = current_column() + 1;
			}
		}
	}
done:
	if (res != INP_ABORT) {
		grab(0, maxlen, trystr);
		if (size() > maxlen)
			say("Response exceeds internal limit of %d characters; truncated.",
				  maxlen);
	} else if (!user_abort)		/* user deleted edit window */
		user_abort = 1;
	if (res == INP_EXIT && must_match(trystr, finder, flags))
		goto show_matches;
	if (is_window(owin))
		window_handle = owin;
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(pbuf);
	if (mbuf) {
		remove_window(menu);
		buf_delete(mbuf);
	}
}

error_if_remote(char *file)
{
	if (is_remote_file(file))
		error("Remote files not supported for this command.");
}

command insert_file()
{
	char file[FNAMELEN];
	int flags = OFN_HIDEREADONLY;

	save_var show_mouse_choices = 1, file_pattern_rules |= FPAT_PERMIT_NO_URLS;
	if (use_common_file_dialog()) {
		strcpy(file, _region_file);
		if (!use_common_open_file_dlg(file, "Insert file", &flags, 0))
			return;
		save_var file_pattern_wildcards = 0; // []'s etc in file name are literal.
	} else {
		if (!insert_file_remembers_file || !*_region_file)
			get_buffer_directory(_region_file);
		get_file(file, "Insert file", _region_file);
	}
	if (!*file)
		return;
	strcpy(_region_file, file);
	mark = point;			/* put mark before it */
	set_region_type(REGNORM);
	build_first = 1;
	do_insert_file(file, give_line_translate(file));
	iter = 0;
}

do_insert_file(file, strip)	/* insert file into current buffer at pt */
char *file;
{
	int orig = bufnum, buf, err;

	error_if_remote(file);
	bufnum = buf = tmp_buf();
	err = err_file_read(file, strip);
	if (!err)
		buf_xfer(orig, 0, size());
	bufnum = orig;
	buf_delete(buf);
	return err;
}

char *low_file_match(char *s, int flags)
{
	char *remote_file_match();

	if (is_remote_file(s))
		return (file_pattern_rules & FPAT_PERMIT_NO_URLS) ? 0
				: remote_file_match(s, flags);
	return file_match(s, flags);
}

grep_prompt(flags, inbufs)
{
	char filepat[FNAMELEN], old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	_last_was_grep = 1;
	tiled_only();
	iter = 0;
	error_if_input(1);	/* error if we're in the middle of input */
	flags = get_search_string("List lines with: ",
				  flags | (inbufs ? BUFGREP : FILEGREP));
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	if (!inbufs)
		save_var full_path_on_mode_line = 1;
	all_must_build_mode = 1;
	if (!get_grep_file(inbufs, filepat))
		return;
	restore_vars();
	all_must_build_mode = 1;
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	if (inbufs && (search_in_region & 2) && !strcmp(filepat, bufname)
		  && is_highlight_on()) {
		save_var point, mark;
		fix_region();
		save_var narrow_start = point, narrow_end = size() - mark;
	}
	if (!setjmp(top_level))
		do_grep("grep", default_search_string(flags), flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

char *hex_prefix();

hex_on_modify()	// Tried to modify buffer with hex display.
{
	call_on_modify = 1;
	if (this_cmd != UNDO_CMD && this_cmd != UNDO_MOVEMENTS)
		error("Edit this buffer using hex-mode commands.  "
			  "Press %s? for help.", hex_prefix());
}

color_html_comment()
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (parse_string(1, "%<%!---")) {	// A legal ColdFusion comment.
		int level = 0;
		while (re_search(1, "%<%!---|---%>"))
			if (character(matchstart) == '<')
				level++;
			else if (--level <= 0)
				break;
		set_character_color(s, point, color_class html_altcomment);
		set_tagged_region("color-as-unit", s, point, 99);
		return;
	} else if (parse_string(1, "%<[ \t\n]*%![ \t\n]*--")) {
		point = matchend;		// A legal comment.
		re_search(1, "--[ \t\n]*>");	// Search for end of legal comment.
	} else if (parse_string(1, "%<%!DOCTYPE[^][<>]*%[")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		return;
	} else if (parse_string(1, "%<%!(ENTITY|ELEMENT|ATTLIST)[ \t\n]*")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		s = point;
		search(1, ">");
		set_character_color(s, point, color_class html_value);
		if (character(point - 1) == '>')
			set_character_color(point - 1, point, color_class html_punctuation);
		return;
	} else if (parse_string(1, "%<%!DOCTYPE")) {
		if (re_search(1, "[[>]") && character(matchstart) == '[') {
			search(1, "]");
			search(1, ">");
		}
		set_character_color(s, point, color_class html_value);
		return;
	} else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

do_reindent_after_yank()
{
	int tmp, orig = bufnum, from, to, stuff_before, max;
	jmp_buf this_level;

	if (_doing_input)	// Don't indent at prompts.
		return;
	max = get_mode_variable("reindent_after_%s_yank");
	if (got_bad_number)
		max = reindent_after_yank;
	if (!max || max > 0 && point - mark > max)
		return;	// Text is too large.
	stuff_before = matches_at(mark, -1, "^.*[^ \n\t].*") > 0;
	if (lines_between(mark, point, 0) < 1) {
		if (parse_string(1, "[ \t]*[^ \t\n]"))
			return;
		if (stuff_before)	// Other stuff on the line, don't reindent.
			return;
	}
	save_var narrow_start = 0, narrow_end = 0;
	save_var reindent_c_comments &= ~1;
	if (reindent_c_comments & 2)
		reindent_c_comments |= 1;
	if (reindent_one_line_c_comments == 1)
		save_var reindent_one_line_c_comments = 0;
	prev_cmd = 0;
	from = mark;	// Make a copy of the buffer and do indenting
	to = point;		// there, so undo records just one operation.
	tmp = tmp_buf();	// (Yank-pop requires undo.)
	buf_xfer(tmp, 0, size());
	bufnum = tmp;
	mark = from;
	point = to;
	safe_copy_buffer_variables(tmp, orig);	// Duplicate bufvars into tmp.
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		spot end = alloc_spot(1);
		save_var iter = 1, has_arg = 0;
		if (stuff_before) {		// Don't indent the first line;
			exchange_point_and_mark();	// there was already an indent.
			nl_forward();
			exchange_point_and_mark();
		}
		indent_region();
		if (!buffers_identical(orig, tmp)) {
			fix_region();
			if (stuff_before)	// Indent-region can expand the region.
				point = from;
			else
				from = point;
			to = size() - *end;
			delete(0, point);	// Discard all but the region.
			delete(*end, size());
			bufnum = orig;
			delete(from, size() - to);
			grab_buffer(tmp);
			_oldfrom = mark;
			_oldto = point;		/* end of yank */
		}
	}					// User aborted, just clean up.
	bufnum = orig;
	buf_delete(tmp);
}

undo_it(is_undo, style, talk)
{
	int ret, before_count;
	char msg[200], *name = is_undo ? "un" : "re", *spec = 0;
	int orig_nar_start = narrow_start;
	int orig_nar_end = narrow_end;

	this_cmd = UNDO_CMD;
	if (undo_count(is_undo) < 1)
		if (talk & 4)
			error("Nothing to %sdo.", name);
		else
			return -1;
	before_count = undo_count(0);
	if (!before_count)
		_buffer_saved_here = -1;
	if (!undo_keeps_narrowing)
		narrow_start = narrow_end = 0;
	*name = tolower(*name);
	if (style)
		ret = undo_group_movements(is_undo, style);
	else					/* plain undo */
		ret = undo_op(is_undo);
	narrow_start = orig_nar_start;
	narrow_end = orig_nar_end;
	if ((point < orig_nar_start || point > size() - orig_nar_end)
	   && !undo_keeps_narrowing) {
		if (talk & 2)
			say("Undoing outside narrowed region, removing narrowing.");
		narrow_start = narrow_end = 0;
	}
	*name = toupper(*name);
	if (ret & UNDO_FLAG)
		switch (undo_flag) {
			case UNDO_READ_FILE:	spec = "Buffer read from disk";
									break;
			case UNDO_WRITE_FILE:	spec = "Buffer written to disk";
									break;
			case UNDO_UNMODIFIED:	spec = "Buffer was marked unmodified";
									break;
		}
	if (spec) {
		sprintf(msg, "%s, ", spec);
		if (_buffer_saved_here == -1)
			_buffer_saved_here = undo_count(0);
		if (undo_count(0) == _buffer_saved_here || before_count == _buffer_saved_here)
			modified = 0;
	} else if (ret & (UNDO_CHANGES | UNDO_MOVE | UNDO_REDISP | UNDO_MAINLOOP))
		sprintf(msg, "%sdid %s%s%s%s%s%s%s",
			name,
			(ret & UNDO_END ? "Partial " : ""),
			(ret & UNDO_MAINLOOP ? "Command, " : ""),
			(ret & UNDO_REDISP ? "Display, " : ""),
			(ret & UNDO_MOVE ? "Move, " : ""),
			(ret & UNDO_INSERT ? "Insert, " : ""),
			(ret & UNDO_DELETE ? "Delete, " : ""),
			(ret & UNDO_REPLACE ? "Replace, " : ""));
	else
		sprintf(msg, "%sdid, ", name);
	if (talk & 1)
		show_text(0, 1, "%s%d bytes remain", msg, undo_count(is_undo));
	highlight_off();
	return ret;
}

spell_configure_helper(int w)
{
	char file[FNAMELEN], ok[FNAMELEN], pr[FNAMELEN];

	sprintf(ok, "NAI%sU", (is_unix_flavor == IS_UNIX_MACOS || is_win32) ? "M" : "");
	sprintf(pr, "External spell helper? (None, Aspell, Ispell%s, Unchanged) [U] ",
			is_win32 ? ", MicroSpell" :
			(is_unix_flavor == IS_UNIX_MACOS) ? ", Mac OS X" : "");
	show_popup_help(w, "The speller can optionally use an external helper "
					"program to provide better, faster suggestions when "
					"correcting misspellings.  If you have one of the supported "
					"programs installed, select it now.\n\nThe aspell and "
					"earlier ispell programs are commonly installed on Unix "
					"systems, and may also be used on Windows systems via the "
					"Cygwin system.\n\nMicroSpell is a commercial spelling "
					"checker that provides hotkey spell checking under "
					"Windows.  See www.microspell.com.  You must have version "
					"9.1f or later.\n\nThe OS X speller is available under Mac "
					"OS X systems (but not over ssh).\n\nThis option sets "
					"the spell-helper-program variable.  If your aspell/ispell "
					"program is not on your path, you'll have to modify this "
					"variable to include its full path.");
	switch (get_key_response(pr, ok, 'U', "spell-configure")) {
		case 'A':	strcpy(spell_helper_program, "aspell -a");
					break;
		case 'I':	strcpy(spell_helper_program, "ispell -a");
					break;
		case 'M':	if (is_unix_flavor == IS_UNIX_MACOS) {
						strcpy(spell_helper_program, "%xmacspell");
						break;
					}
					get_executable_file(file, "mspellcmd.exe", 0);
					delete_file(MICROSPELL_CMDLINE);
					if (copyfile(file, MICROSPELL_CMDLINE))
						say("Error installing %s, check.", MICROSPELL_CMDLINE);
					else
						strcpy(spell_helper_program, MICROSPELL_CMDLINE);
					break;
		case 'N':	*spell_helper_program = 0;
					break;
	}
}

// Load the two dictionary files.  Sort the user one.  Return 1 if ok.
int load_spell_word_list(int force_reload)
{
	if (!spell_load_file(SPELLB_ORDERED, force_reload)) {
		if (!vspell_data.told_err) {
			say("Can't find speller word list %f, run spell-configure.",
				  WORD_ORDER_FILE);
			vspell_data.told_err = 1;
		}
		return 0;
	}
	spell_load_file(SPELLB_SORTED, force_reload);
	spell_load_file(SPELLB_GLOBAL, force_reload);
	spell_load_file(SPELLB_PERDIR, force_reload);
	spell_load_file(SPELLB_PERFILE, force_reload);
	spell_load_file(SPELLB_PERMODE, force_reload);
	return 1;
}

int vspell_buffer_when_displaying();

do_when_make_mode_vspell()
{
	if (use_default)
		return;
	vspell_data.want = get_num_var(vspell_get_mode_variable()) & 1;
	if (vspell_data.buf_want)
		vspell_data.want = (vspell_data.buf_want == 1);
	if (vspell_data.last_want == vspell_data.want)
		return;
	if (!spell_mode_ok(bufnum))
		vspell_data.want = 0;
	vspell_data.last_want = vspell_data.want;
	if (modify_region(SPELL_REGION_HANDLE, MRTYPE, -1) == REGNORM)
		remove_region(SPELL_REGION_HANDLE);	// Remove if it exists.
	if (vspell_data.want) {
		if (load_spell_word_list(0)) {
			add_buffer_when_displaying(bufnum, vspell_buffer_when_displaying);
			vspell_data.window_start = -1;
		} else
			vspell_data.want = vspell_data.last_want = vspell_data.buf_want = 0;
	} else
		delete_buffer_when_displaying(bufnum, vspell_buffer_when_displaying);
}

command spell_mode()
{
	set_spell_mode(0);
}

command buffer_spell_mode()
{
	set_spell_mode(1);
}
