/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B1-B6 to 13 B7, assuming further updaters.

#include "eel.h"
#include "colcode.h"
#include "c.h"
#include "session.h"
#include "perl.h"
#include "diff.h"
#include "epsnet.h"
#include "proc.h"
#include "grep.h"
#include "html.h"
#include "tags.h"
#include "dired.h"
#include "bufed.h"
#include "lowlevel.h"
#include "indent.h"
#include "info.h"
#include "inimode.h"
#include "misclang.h"
#include "python.h"
#include "tex.h"
#include "vbasic.h"

when_loading()
{
	int dorest = 0;
	if (!strcmp(eel_version, "13b1") || dorest) {
			// Apply new defaults if orig setting is unchanged.
		if (dired_layout == 0)
			dired_layout = 1;
		if (!*print_color_scheme)
			strcpy(print_color_scheme, "standard-gui");
		if (!strcmp(mode_format, " %b%E%M %d%p %s%f"))
			strcpy(mode_format, " %b%E%M %l,%c %d%p %s%f");
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b2") || dorest) {
		if (!*ignore_file_basename)
			strcpy(ignore_file_basename, "%.");
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b3") || dorest) {
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b4") || dorest) {
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b5") || dorest) {
		if (diff_match_characters_limit < 50_000)
			diff_match_characters_limit = 50_000;
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b6") || dorest) {
		if (!strcmp(auto_save_name, "%p%b.asv"))
			strcpy(auto_save_name, AUTOSAVENAME);
		dorest = 1;
	}
}

color_class spell_misspelled = black on light_pink;

#include "vspell.e"

command normal_character()
{
	int i;

	check_abort();
	delete_if_highlighted();
	if (!call_maybe_break_line(3))
		return;
	if (IS_ALT_KEY(key) && KEY_PLAIN(key) < 128)
		key |= 128;			// turn on meta bit
	if (!IS_ASCII_KEY(full_key))
		if (IS_ASCII_KEY(key))
			full_key = key;
		else if (IS_ASCII_KEY(KEY_PLAIN(key)))
			full_key = KEY_PLAIN(key);
	if (full_key == '\n') {
		if (over_mode) {
			if (!nl_forward())
				insert('\n');
		} else if (auto_indent && indenter)
			break_and_indent_line();
		else
			insert('\n');
	} else if (over_mode && point < size() && curchar() != '\n') {
		if (curchar() == '\t')
			i = display_width('\t', current_column());
		else
			i = 0;
		replace(point++, full_key);
		while (--i > 0)		/* insert i-1 spaces after pt */
			insert(' '), point--;
	} else
		insert(full_key);
}

down_or_up_lines(down_lines)
{
	int hit_end = 0, wantcol;

	do_shift_selects();
	if (iter < 0)
		iter = -iter, down_lines = !down_lines;
	if (prev_cmd != SAMECOL)
		_preserve_column = virtual_space ? virtual_column()
					: current_column();
	while (iter-- > 0)
		if (!down_lines)
			nl_reverse();
		else if (!nl_forward()) {
			hit_end = 1;
			break;
		}
	if (is_highlight_on())
		save_var goal_column = -1;
	wantcol = (goal_column >= 0) ? goal_column : _preserve_column;
	if (virtual_space)
		to_virtual_column(wantcol);
	else if (!hit_end || goal_column >= 0)	/* stay at end if we tried */
		move_to_column(wantcol);	/* to move past it */
	this_cmd = SAMECOL;
	iter = 0;
}

charname_lister(char *s, char *(*finder)(), int flags, int mbuf);
char *charname_matcher(char *s, int flags);

command insert_ascii()
{
	int res;
	char buf[FNAMELEN];

	if (run_by_mouse)
		has_arg = 0;
	save_var completion_lister = charname_lister;
	save_var menu_width = MAX(menu_width, 70);
	comp_read(buf, "Insert character with ASCII code: ",
			  charname_matcher, COMP_FOLD | SPACE_VALID, "");
	res = exptoi(buf);
	if (got_bad_number || res < 0 || res > MAX_CHAR)
		error("Invalid ASCII code %s.", buf);
	insert(res);
}

int numtoi(char *s)	/* like strtoi(), but get base from string */
{
	int neg = 0, i;
	char *from = "nbtrf", *to = "\n\b\t\r\f", *p;

	got_bad_number = 0;
	while (isspace(*s))
		s++;
	if (*s == '\'') {		/* allow 'x' as a number (give ascii code) */
		if (s[1] != '\\' && s[2] == '\'')
			return s[1];
		if (s[1] == '\\' && s[2] && s[3] == '\'') {
			p = strchr(from, s[2]);
			if (p)
				return to[p - from];
			return s[2];
		}
		got_bad_number = 1;
		return s[1];
	}
	if (*s == '<') {
		i = key_value(s);
		if (i >= 0)
			return i;
		got_bad_number = 1;
		return 0;
	}
	while (*s == '-')
		s++, neg = !neg;
	if (s[0] == '0' && tolower(s[1]) == 'x')
		i = strtoi(s + 2, 16);
	else if (s[0] == '0' && tolower(s[1]) == 'o')
		i = strtoi(s + 2, 8);
	else if (s[0] == '0' && tolower(s[1]) == 'b')
		i = strtoi(s + 2, 2);
	else
		i = strtoi(s, 10);
	return neg ? -i : i;
}

int exptoi(char *s)
{
	save_var use_default = 0;	// Don't run compile with use_default 1.
	while (isspace(*s))
		s++;
	if (!strchr("<", *s) && (strpbrk(s, "+/*^|&<>%=")
							  || *s && strpbrk(s + 1, "-"))
		  && (has_feature & FEAT_EEL_COMPILE))
		return evaluate_numeric_expression(s);
	return numtoi(s);
}

quote_bufname(bname)	/* add quotes around bname if necessary */
char *bname;
{
	char tmp[FNAMELEN * 2];

	strcpy(tmp, bname);
	if (index(bname, ' ') || strlen(bname) >= bufed_column_width - 2) {
		sprintf(bname, "\"%s\"", tmp);
	} else
		sprintf(bname, " %s", tmp);
}

command count_lines()
{
	int count, here, disksize, reg_lines, trans;
	char regmsg[FNAMELEN], msg[FNAMELEN];

	noteput("Counting...");
	here = lines_between(0, point, 1);
	check_abort();
	noteput("Counting... point on line %'d...", here + 1);
	count = here + lines_between(point, size(), 1);
	check_abort();
	disksize = size();
	trans = GET_LINE_TRANSLATE(translation_type);
	if (trans == FILETYPE_AUTO)
		trans = get_fallback_translation_type();
	if (trans == FILETYPE_MSDOS)
		disksize += count;
	if (size() && character(size() - 1) != '\n')
		count++;
	save_var point, mark;
	fix_region();
	*regmsg = 0;
	if (is_highlight_on()) {
		reg_lines = lines_between(point, mark, 1);
		if (mark > point && character(mark - 1) != '\n')
			reg_lines++;
		sprintf(regmsg, ", %'d in region", reg_lines);
	}
	sprintf(msg, "%'d line%s%s, point on line %'d.",
		  count, count == 1 ? "" : "s", regmsg, here + 1);
	sprintf(msg + strlen(msg), "  %'d %s on disk.",
			disksize, (buffer_flags() & BF_UNICODE) ? "chars" : "bytes");
	say("%s", msg);
	expire_message = 1;
	iter = 0;
}

buf_set_character_color(int buf, int from, int to, int color)
{
	save_var bufnum = buf;
	set_character_color(from, to, color);
}

int find_objc_method(char *func)
{
	char funcname[FNAMELEN];
	int i, j;

	*funcname = 0;
	i = point;
	point += parse_string(1, "[-+][ \t]*");
	save_var point, narrow_start = point;
	re_search(1, "[{@]");
	save_var narrow_end = size() - point;
	point = 0;
	if (curchar() == '(')
		if (!c_skip_level('(', ')'))
			return 0;
	point += parse_string(1, "[ \t\n]*");
	if (parse_string(1, "(<word>*)[ \t]*:[ \t\n]*")) {	// Parse keyword selector:
		for (;;) {			// A series of (optional) words with :'s after each.
			grab(find_group(1, 1), find_group(1, 0), funcname + strlen(funcname));
			point = matchend;
			strcat(funcname, ":");
			if (curchar() == '(')
				c_skip_level('(', ')');
			j = parse_string(1, "[ \t\n]*<word>+[ \t\n]*");
			if (!j)
				break;
			point += j;
			if (!parse_string(1, "[ \t\n]*(<word>*)[ \t]*:[ \t\n]*"))
				break;
		}
	} else if (!parse_string(1, "<word>+", funcname))
		return 0;
	start_of_function = i;
	if (func)
		strcpy(func, funcname);
	if (search(1, ";"))	// A mere declaration.  (We're already narrowed.)
		return 2;
	return 1;
}

// Fill in type and class if inside scope of @implementation or
// similar, returning CF_ code.
int find_objc_class(char *type, char *class)
{
	save_var point;
	to_end_line();
	if (re_search(-1, "^@(end|((implementation|interface)[ \t]+(<word>+)))")
		  && character(point + 1) != 'e') {
		if (type)
			grab(find_group(3, 1), find_group(3, 0), type);
		if (class)
			grab(find_group(4, 1), find_group(4, 0), class);
		return CF_INFO_TYPE | CF_INFO_CLASS;
	}
	return 0;
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN], classname[FNAMELEN], pat[FNAMELEN];
	int start, funcstart, j, got_class = 0;

	start_of_function = -1;
	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	if (c_extra_keywords & OBJC_KEYWORDS)
		got_class |= find_objc_class(type, class);
	to_begin_line();
	start = point;
	point--;		// Begin searching at the last } char on the left margin...
	while (search(-1, "\n}")) {
		point++;		// ...that's not in a comment.
		if (!in_c_comment(c_look_back))
			break;
		point--;
	}

	strcpy(pat, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(");
	if (c_extra_keywords & OBJC_KEYWORDS)
		strcat(pat, "|\n[-+]");
	while (re_search(1, pat)) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*(define|if).*(\\\n.*)*")) {
			nl_forward();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				j = point;
				if (c_inside_block(point, start)) {
					got_class |= find_c_class_info(start, type, class, func);
					point = j + 1;
					if (got_class)
						continue;
				}
				point = j;
				c_skip_level('{', '}');
				break;
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					nl_forward();	/* c++ comment */
				break;
			case '\n':		// Objective-C method
				point++;
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				if (find_objc_method(func) == 1)
					return CF_INFO_FUNC | got_class;
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return got_class;
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + 1000);
				if (tag_c_func(0) == 2 && !find_c_func_oneliner(start)) {
					point = funcstart;	// Checks out as a function def.
					start_of_function = give_begin_line();
					*classname = 0;
					if (parse_string(-1, (c_extra_keywords == CSHARP_KEYWORDS)
									 ? "[ \t\n]*%.[ \t\n]*"
									 : "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						parse_string(-1, "[a-zA-Z0-9_]+", classname);
					}
					if (func)
						strcpy(func, funcname);
					if (class && *classname)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0)
							| got_class;
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return got_class;
}

suffix_ii()		// Preprocessed C++.
{
	suffix_cpp();
}

suffix_mi()		// Preprocessed Objective-C.
{
	suffix_m();
}

suffix_mmi()		// Preprocessed Objective-C++.
{
	suffix_mm();
}

indent_c_contin(ind, orig, lastl, in_initializer)
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0, slevel = 0;
	int objc_sel = 0, objc_indent = -1;

 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	if (!c_align_contin_lines)	// never line up
		if (!in_initializer)
			return c_default_indent(ind, orig, 1);
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	if ((c_extra_keywords & OBJC_KEYWORDS) && c_align_selectors
		  && matches_at(give_begin_line(), 1, "^[ \t]*(<word>+):"))
		objc_sel = find_group(1, 0) - find_group(1, 1);
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[][()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0 || objc_indent >= 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (objc_indent >= 0)
					break;
				if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				if (objc_sel && !plevel && !slevel
					  && parse_string(-1, "(^|<word>|[]])[ \t]*<word>+")
					  && current_column() <= c_align_selectors)
					objc_indent = current_column() - objc_sel;
				else
					qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				if (objc_indent >= 0)
					return objc_indent;
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '[':
				if (slevel-- > 0 || objc_indent >= 0)
					break;
				if (findc != ']')
					point++;
				save_var c_align_extra_space = 0;
				return limit_c_contin_indent(ind);
			case ']':
				slevel++;
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)) || objc_indent >= 0)
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (objc_indent >= 0)
		return objc_indent;
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		if (curchar() == '#')
			past_initial_preprocessor();
		return limit_c_contin_indent(ind);
	}
	return c_default_indent(ind, orig, 2);
}

int c_rev_skip_level(left_d, right_d)
char left_d, right_d;
{
	char pattern[40], ch;
	int level = 0;

	if (in_perl_buffer)
		return perl_skip_level(-1, left_d, right_d);
	sprintf(pattern, "<*>/|[%c%c'\"]|//.*\n", right_d, left_d);

bgn:
	do {
		if (!re_search(RE_REVERSE, pattern))
			return 0;
		switch (ch = curchar()) {
			case '/':
				if (at_c_comment(point + 2, 100) & IN_STRING)
					to_end_line();
				goto bgn;
			case '\'': case '"': case '*':
				point = matchstart;
				if (ch != '*' && (in_c_comment(100) & IN_COMMENT)) {
					  point--;
					  goto bgn;
				}
				c_skip_item(character(point - 1), RE_REVERSE);
				goto bgn;
			default:
				if (ch == right_d)
					level++;
				else if (ch == left_d)
					level--;
		}
	} while (level > 0);
	return 1;
}

int c_skip_level(left_d, right_d)
char left_d, right_d;
{
	char pattern[40], ch;
	int level = 0;

	if (in_perl_buffer)
		return perl_skip_level(1, left_d, right_d);
	sprintf(pattern, "/[/*]|[%c%c'\"]", right_d, left_d);

bgn:	do {
		if (!re_search(RE_FORWARD, pattern))
			return 0;
		switch (ch = character(matchstart)) {
			case '/':
				search(1, character(matchstart + 1) == '/'
				       ? "\n" : "*/");
				break;
			case '\'': case '"':
				point = matchstart;
				c_skip_item(ch, RE_FORWARD);
				goto bgn;
			default:
				if (ch == right_d)
					level++;
				else if (ch == left_d)
					level--;
		}
	} while (level < 0);
	return 1;
}

// Move forward past spaces, tabs, optionally newlines, and C/C++
// comments.
c_skip_fwd_space(int skip_newlines)
{
	for (;;)
		switch (curchar()) {
			case '\n':
				if (!skip_newlines)
					return;
			case ' ': case '\t':
				point++;
				break;
			case '/':
				switch (character(point + 1)) {
					case '/':
						nl_forward();
						break;
					case '*':
						point + 2;
						search(1, "*/");
						break;
					default:
						return;
				}
			default:
				return;
		}
}

char *c_search_include_path(char *rel, int flags);

// Search on line for link, set file to its name.  Abort if none.
c_find_linked_file(char *file)
{
	char rel[FNAMELEN], curdir[FNAMELEN], *p;
	int flags = PATH_ADD_CUR_DIR, is_abs = 0;

	to_begin_line();
	if (parse_string(1, "[ \t]*#[ \t]*include[ \t]*[<\"](.*)([>\"])"
					 "[ \t]*(/[*/].*)?$")) {
		grab(find_group(1, 1), find_group(1, 0), rel);
		is_abs = character(find_group(2, 1)) == '>'
				 && !find_linked_file_ignores_angles;
		// Using #include "" or <> syntax?
	} else if (is_java_buffer() &&  parse_string(1, "[ \t]*(package|import)"
		"[ \t]*(<word>+(%.<word>+)*(%.%*)?)[ \t]*;")) {
		grab(find_group(2, 1), find_group(2, 0), rel);
		while (p = strchr(rel, '.'))	// Transform java.lang to
			*p = path_sep;				// java/lang filename.
		p = rel + strlen(rel) - 1;
		if (*p == '*') {
			strcpy(p, ".");
			flags |= PATH_PERMIT_DIRS;
		} else
			strcat(p, ".java");
	} else {
		default_find_linked_file(file);
		return;
	}
	strcpy(file, rel);
	get_buffer_directory(curdir);
	if (is_abs)				// Don't look for relative names at all.
		flags = 0;
	else {
		absolute(file, curdir);
		if (check_file(file))
			return;
	}
	p = c_search_include_path(rel, flags);
	if (!p)
		error("%s not found in include-directories path", rel);
	strcpy(file, p);
}

// Run this program on the current word.
do_context_help_run(char *prog)
{
	char item[FNAMELEN], cmdline[FNAMELEN], buf[FNAMELEN];
	char winprog[FNAMELEN * 2], fixprog[FNAMELEN * 2], *s;
	int sync = 1;

	context_help_get_item(item, word_pattern);
	strcpy(fixprog, prog);
	if (*fixprog == '"' && (s = strchr(fixprog + 1, '"'))) {
		strncpy(winprog, fixprog + 1, s - fixprog - 1);
		convert_to_8_3_filename(winprog);
		strcat(winprog, s + 1);
		strcpy(fixprog, winprog);
	}
	s = fixprog + strlen(fixprog) - 1;
	if (strlen(fixprog) > 1 && *s == '&') {
		*s = 0;
		sync = 0;
	}
	if (s = strchr(fixprog, '*')) {
		strncpy(cmdline, fixprog, s - fixprog);
		quote_for_shell(cmdline + strlen(cmdline), item);
		strcat(cmdline, s + 1);
	} else {
		sprintf(cmdline, "%s ", fixprog);
		quote_for_shell(cmdline + strlen(cmdline), item);
	}
	sprintf(buf, "context-help-%s", item);
	if (opsys == OS_UNIX || !sync) {
		pipe_text(NULL, buf, cmdline, NULL, (sync ? PIPE_SYNCH : 0) |
				  PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			point = 0;
		}
	} else {
		strcpy(winprog, cmdline);
		if (index(winprog, ' '))
			*index(winprog, ' ') = 0;
		if (winexec(winprog, cmdline, SW_SHOWNORMAL, 0) == -1)
			say("Could not start %s.", winprog);
	}
}

// Change \n to \r\n when copying to clipboard, \r\n to \n when retrieving?
clipboard_convert_newlines()
{
	return (GET_LINE_TRANSLATE(translation_type) != FILETYPE_BINARY
			&& opsys != OS_UNIX) ? CLIP_CONVERT_NEWLINES : 0;
}

char *clipboard_buf_name();

command copy_to_clipboard()	// copy region to clipboard
{
	int err = 0, tmp = zap(clipboard_buf_name());
	save_spot point, mark;
	fix_region();
	if (region_type() == REGRECT)
		extract_rectangle(tmp, 0);
	else
		buf_xfer(tmp, point, mark);
	if (!buffer_to_clipboard(tmp, clipboard_convert_newlines(), clipboard_format))
		err = errno;
	zap(clipboard_buf_name());	// Load it next time.
	if (err)
		file_error(err, "", "Couldn't copy to clipboard.");
	else {
		say("Selection copied to clipboard.");
		highlight_off();
	}
}

int html_clipboard_format()
{
	return call_dll("USER32.DLL", "RegisterClipboardFormatW",
					"p", "S", "HTML Format");
}

// Epsilon's RGB encoding isn't the same as HTML's; convert.
int color_to_html(int color)
{
	int r = GETRED(COLOR_STRIP_ATTR(color));
	int g = GETGREEN(COLOR_STRIP_ATTR(color));
	int b = GETBLUE(COLOR_STRIP_ATTR(color));
	return (r << 16) | (g << 8) | b;
}

// Encode the 8-bit characters of buffer buf into the specified Unicode
// encoding, preserving colors.
int color_perform_unicode_conversion(int buf, char *enc)
{
	int pos, sz, color;
	save_var bufnum = buf;
	save_var point = 0;			// Convert any non-ascii to UTF8.
	while (re_search(1, "<^ascii>")) {
		pos = point;
		sz = size();	// Do it one character at a time to preserve colors.
		color = get_character_color(point - 1);
		if (perform_unicode_conversion(bufnum, point - 1, point, 0, enc))
			error("Failed to convert to %s at %d in %s.", enc, pos, bufname);
		point = pos + size() - sz;
		set_character_color(pos - 1, point, color);
		check_abort();
	}
}

// Look for each use of open following close.  If it has the same value
// (in match #1) as the previous one, delete this open and close.
html_clean_repeated(char *open, char *close)
{
	char last[FNAMELEN], this[FNAMELEN];

	*last = 0;
	save_var point = 0;
	while (re_search(1, open)) {
		grab(find_group(1, 1), find_group(1, 0), this);
		int end = matchend;
		if (!strcmp(last, this) && matches_at(matchstart, -1, close))
			delete(matchend, end);
		strcpy(last, this);
	}
}

// Put the current buffer's file name on the clipboard.
command copy_file_name() on cc_tab[ALT('n')]
{
	char buf[FNAMELEN], *s = filename;

	if (is_dired_buf())
		dired_getname(s = buf, 1); 
	else if (!*s)
		s = bufname;
	copy_line_to_clipboard(s, 0);
} 

int quickup_would_work()
{
	char def[FNAMELEN];
	if (!*state_file)
		return 0;
	strcpy(def, state_file);
	set_default_state_file_name(def);
	absolute(def);
	return get_file_read_only(state_file) != 1 && !strcmp(state_file, def);
}

offer_quickup()
{
	char cmdline[16384];
	int i;

	get_string(cmdline, "Run quickup now, then restart this instance [Y]: ");
	if (tolower(*cmdline) == 'n')
		return;
	if (shell_with_args(argv[0], "-quickup") != 42)
		error("Auto-quickup failed, please update manually.");
	*cmdline = 0;
	for (i = 1; original_argv(i); i++) {
		quote_for_shell(cmdline + strlen(cmdline), original_argv(i));
		if (original_argv(i + 1))
			strcat(cmdline, " ");
	}
	int err = shell_with_args_flags(original_argv(0), cmdline,
									SHELL_NO_SYNCH | SHELL_KEEP_ENV);
	if (err)
		file_error(err, original_argv(0),
				   "Couldn't restart Epsilon, please restart manually");
	else
		leave(0);
}

// To deal with change in version string format, process a version
// string into a standard format: strip colon and make part after b
// two-digit.
standardize_version_string(char *dest, char *src)
{
	char *s;
	strcpy(dest, src);
	while (s = strpbrk(dest, ": "))	// Strip all these chars.
		strcpy(s, s + 1);
	s = dest + strlen(dest) - 1;	// Last char.
	if (strchr("abcd.", s[-1])) {	// Second-to-last char.
		s[2] = 0;		// One digit, make two: Convert 13b4 to 13b04
		s[1] = s[0];
		s[0] = '0';
	}
	for (; *dest; dest++)
		*dest = tolower(*dest);
}

restore_check_version()
{
	char clean_exe_version[30], clean_eel_version[30];

	// To deal with change in version string format, process
	// versioned_file_string into a standard format: strip colon and
	// make part after b two-digit.
	standardize_version_string(clean_exe_version, versioned_file_string);
	standardize_version_string(clean_eel_version, eel_version);
	if (!load_from_state || strpbrk(clean_exe_version, "ad"))
		return;
	if (strcmp(clean_eel_version, clean_exe_version) >= 0)
		if (!strchr(clean_eel_version, 'b') || strchr(clean_exe_version, 'b'))
			return;
	// Each user has a private state file, so we can't update them
	// all automatically.
	save_var bufnum = zap(HELPBUF);
	bprintf("Your version %s state file requires updating.\n"
			"You need to start Epsilon with its -quickup flag.\n", 
			eel_version);
	bprintf("Until you do this, Epsilon will not work correctly.\n");
	view_help(" Update Needed! ", 0);
	if (quickup_would_work())
		offer_quickup();
}

when_restoring()
{
	int tflag = 0, i, splitmode = HORIZONTAL;
	char *flag, *arg, init_dir[FNAMELEN], *dir = init_dir;

	getcd(init_dir);
	early_restoring();
	early_init();
	set_toolbar();

	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		if (!stricmp(argv[i], "-noinit")) {
			noinit_flag = 1;
			continue;
		} else if (!strcmp(argv[i], "-dir")) {
			if (++i < argc)
				dir = argv[i];
			continue;
		}
		flag = argv[i];
		switch (flag[1]) {
		case 'f':
			arg = (flag[2] && flag[3] || i + 1 >= argc) ? (flag + 3) : argv[++i];
			switch (flag[2]) {
			case 's':
				set_swapname(arg);
				break;
			case 'd':
				strcpy(_docname, arg);
				break;
			default:
				bad_flag(flag);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'd':
		case 'r':
		case 'l': if (!flag[2] && i + 1 < argc)
					  i++;
				  break;
		case 'p': break;
		case 'h': case '?': usage(); break;
		case 'v':
			switch (flag[2]) {
				case 'v':
					splitmode = VERTICAL;
					break;
				default:
					bad_flag(flag);
			}
			break;
		case '-':
			if (!flag[2])
				break;
			if (!strcmp(flag + 2, "help"))
				usage();
			else
				bad_flag(flag);
			break;
		default:
			bad_flag(flag);
		}			/* i is 1st file argument */
	}
	term_init();
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	delete_buffer("-customizations");	// Don't record any settings before now.
	parse_command_line_flags(1);
	if (!noinit_flag && load_from_state)
		init_from_file();
	parse_command_line_flags(2);
	if (tflag) {
		preserve_session_once = 1;
		tutorial();
	} else {
		if (load_from_state && (i >= argc || session_always_restore)) {
			if (session_restore_files == 2 && i < argc)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (i < argc)
			load_cmd_line_files(i, splitmode, dir);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	restore_check_version();
	show_text(0, 1, "Epsilon Programmer's Editor %s, EEL version %s",
			  version, eel_version);
	try_calling("start-up");
	parse_command_line_flags(3);
}

#define MAX_BUFFER_WHEN_DISPLAYING	5
buffer int (*buffer_when_displaying[MAX_BUFFER_WHEN_DISPLAYING])();

// Just call the buffer-specific functions for this buffer.
default_when_displaying(int from, int to)
{
	UNREFERENCED_PARAMETER(from);
	for (to = 0; to < MAX_BUFFER_WHEN_DISPLAYING; to++)
		if (buffer_when_displaying[to])
			(*buffer_when_displaying[to])();
}

add_buffer_when_displaying(int buf, int (*func)())
{
	if (!add_to_buffer_func_list(buffer_when_displaying,
								 MAX_BUFFER_WHEN_DISPLAYING, buf, func))
		error("Too many when-displaying functions for this buffer.");
	if (!when_displaying)
		when_displaying = default_when_displaying;
}

delete_buffer_when_displaying(int buf, int (*func)())
{
	delete_from_buffer_func_list(buffer_when_displaying,
								 MAX_BUFFER_WHEN_DISPLAYING, buf, func);
}

// On each refresh, recolor_partial_code() will be called with the
// buffer range that was modified since the last redisplay.  Its params
// give the range of the buffer that was modified, but we don't use
// this.
recolor_partial_code(from, to)
{
	init_color_code();
	if (!(coloring_flags & COLOR_IN_PROGRESS)) {
		coloring_flags |= COLOR_IN_PROGRESS; // ensure errors don't restart us
		if (!(coloring_flags & COLOR_MINIMAL) != !minimal_coloring) {
			drop_all_colored_regions();		// If setting's changed.
			coloring_flags &= ~COLOR_MINIMAL;
			if (minimal_coloring)
				coloring_flags |= COLOR_MINIMAL;
		}
		save_var point, matchstart, matchend, abort_searching = 0;
		if (!(coloring_flags & COLOR_RETAIN_NARROWING))
			save_var narrow_start = 0, narrow_end = 0;
		shrink_colored_area();		// Buffer modifications shrink list of regions.
		if (size() < color_whole_buffer) // Now expand list of regions.
			color_for_area(0, size());
		else
			color_for_area(window_start, MIN(window_end + 1, size()));
		coloring_flags &= ~COLOR_IN_PROGRESS;	// no errors aborted us,
	}					// let us run next time
	default_when_displaying(from, to);
}

c_keyword_color(from)	// return color for "identifier" from here to point
{			// (something with alpha or digits)
	char buf[500];

	save_var point;
	if (point - from > sizeof(buf) / sizeof(char) - 10)
		point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-.", buf[1]))
		return c_number_color(buf + 1);
	strcpy(buf + point - from + 1, "|");
	if (is_c_keyword(buf) && (!(c_extra_keywords & CSHARP_KEYWORDS) ||
							  character(from - 1) != '@'))
		return color_class c_keyword;
	if (color_class c_function != color_class c_identifier
		  && paren_follows())
		return color_class c_function;
	return color_class c_identifier;
}

c_preproc_color()	// color from t to end of preprocessor text
{
	int orig = point;

	re_search(1, "[ \t]*");
	if (parse_string(1, "include|import|tryinclude")) {
		while (nl_forward() && character(point - 2) == '\\')
			;			// color contin lines
		save_var narrow_end = size() - point;
		point = orig;
		while (re_search(1, "/[*/]"))	// stop before comment
			if (!(in_c_comment(c_look_back) & IN_STRING)) {
				point -= 2;
				break;
			}
	} else if (parse_string(1, "(define|undef)[ \t]+[A-Za-z0-9_]+")) {
		point = matchend;
		point += parse_string(1, "%([A-Za-z0-9_, \t]*%)");
	} else
		re_search(1, "[A-Za-z_0-9]*");
}

// always call the following after setting want_code_coloring variable
// set-variable calls this automatically
when_setting_want_code_coloring()
{
	when_displaying = default_when_displaying;
	if (want_code_coloring)
		want_code_coloring = 1;
	if (!want_code_coloring)
		drop_coloring(bufnum);
	if (want_code_coloring && recolor_range && recolor_from_here)
		when_displaying = recolor_partial_code;
	coloring_flags &= ~(COLOR_DO_COLORING | COLOR_IN_PROGRESS);
}

#include "color.e"

/* put matches for s in menu buf */
list_matches(char *s, char *(*finder)(), int flags, int mbuf)
{
	save_var bufnum = mbuf;
	buf_zap(mbuf);
	load_matches(s, finder, flags);
	buf_sort_and_uniq(mbuf);
	point = 0;
	delete_final_newlines();
}

ok_file_match(s)	/* is filename s ok as a match, or excluded? */
char *s;	/* see if extension occurs in ignore_file_extensions, */
{		/* but if only_file_extensions, use that instead */
	char pat[FNAMELEN];

	save_var case_fold = (FOLDFILE != 0);
	sprintf(pat, "|%s|", get_extension(s));
	if (*ignore_file_pattern
		  && string_matches_pattern(s, ignore_file_pattern))
		return 0;
	if (*ignore_file_basename
		  && string_matches_pattern(get_tail(s, 0), ignore_file_basename))
		return 0;
	if (only_file_extensions)
		return strstr(only_file_extensions, pat) != NULL;
	return strstr(ignore_file_extensions, pat) == NULL;
}

get_number(pr)	/* return the numeric argument, or prompt if none */
	char *pr;
{
	char res[FNAMELEN];
	int v = iter;

	got_bad_number = iter = 0;
	if (has_arg)
		return v;
	get_string(res, pr);
	return exptoi(res);
}

// Return nonzero if start of word appears in list of |-delimited words.
int starts_with_in_list(char *word, char *list, int fold)
{
	char *s, *t;
	save_var case_fold = fold;
	for (s = list; t = strchr(s + 1, '|'); s = t)
		if (!strnfcmp(s + 1, word, t - s - 1))
			return 1;
	return 0;
}

// Return nonzero if string matches regex pattern.
int string_matches_regex(char *str, char *pat, int fold)
{
	int tmp = tmp_buf();
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	stuff(str);
	point = 0;
	case_fold = fold;
	if (re_compile(1, pat))
		return 0;
	return (re_match() != -1);
}

// Return length of initial string that matches regex pattern.
int string_matches_pattern(char *str, char *pat)
{
	int tmp = tmp_buf(), res;
	save_var bufnum = tmp;
	stuff(str);
	point = 0;
	res = parse_string(1, pat);
	restore_vars();
	buf_delete(tmp);
	return res;
}

command exit()
{
	if (!has_arg) {
		error_if_input(0);
		check_running_processes();
	}
	if (opsys == OS_DOS && !is_gui && is_win32 != IS_WIN32_CONSOLE
		&& another_process()) {
		if (!ask_yn("Kill process", "A subprocess is running.  Kill it? ",
					"Yes", "No", 1)) {
			check_abort();
			tiled_only();
			locate_window(PROCBUF, "");
			point = size();
			say("You must kill this subprocess before exiting.");
			iter = 0;
			return;
		}
		exit_process();
	}
	if (!has_arg) {
		if (unsaved_buffers() && !bufed_ask_exit())
			return;
		if (load_from_state)
			save_session();
	}
	sayput("");
	wait_for_unseen_msgs();
	delete_all_autosaves();
	finish_up();
	if (!is_gui) {
		term_position(0, screen_lines - 1);
		term_mode(0);
	}
	int i = name_match("do_when_exiting_", 1);
	for (; i; i = name_match("do_when_exiting_", 0))
		if (name_type(i) == NT_SUBR)
			(*(int (*)())i)(has_arg);
	when_exiting();
	leave(has_arg ? iter : 0);
}

do_mode_default_settings(int retain_colors)
{
	indenter = indenter.default;
	if (!indenter)
		indenter = prev_indenter;
	auto_indent = auto_indent.default;
	buffer_maybe_break_line = 0;
	strcpy(comment_start, comment_start.default);
	strcpy(comment_pattern, comment_pattern.default);
	strcpy(comment_begin, comment_begin.default);
	strcpy(comment_end, comment_end.default);
	mode_keys = buf_tab;		/* use these keys */
	mode_move_level = 0;
	major_mode = strkeep("Fundamental");
	auto_show_matching_characters = fundamental_auto_show_delim_chars;
	when_displaying = default_when_displaying;
	recolor_range = recolor_from_here = 0;
	search_continuation = 0;
	if (!retain_colors) {
		drop_coloring(bufnum);
		coloring_flags &= ~(COLOR_MINIMAL | COLOR_INVALIDATE_FORWARD
							| COLOR_INVALIDATE_BACKWARD | COLOR_INVALIDATE_RESETS
							| COLOR_RETAIN_NARROWING);
	}
	make_mode();
}

char *strkeep(char *s)
{
	if (!strkeep_data)
		strkeep_data = (char **) malloc(0);
	int sz = ptrlen(strkeep_data, 1) / sizeof(*strkeep_data), i;
	for (i = 0; i < sz; i++)
		if (!strcmp(s, strkeep_data[i]))	// Already here.
			return strkeep_data[i];
	strkeep_data = (char **) realloc((char *) strkeep_data, (sz + 1)
									 * sizeof(*strkeep_data) / sizeof(char));
	return strkeep_data[sz] = strsave(s);
}

char *strpbrk(char *s, char *charset)
{
	for (; *s; s++)
		if (strchr(charset, *s))
			return s;
	return 0;
}

// Return a pointer to the first character in s that's also in the list
// of characters charset, ignoring the first skip such characters.
// Return null if there aren't that many.
char *strpbrk_cnt(char *s, char *charset, int skip)
{
	char *p;

	for (;;) {
		p = strpbrk(s, charset);
		if (!p || --skip < 0)
			return p;
		s = p + 1;
	}
}

// Add func to cnt-element list of functions for the buffer buf.
int add_to_buffer_func_list(int (**list)(), int cnt, int buf, int (*func)())
{
	int i;
	save_var bufnum = buf;
	for (i = 0; i < cnt; i++)
		if (list[i] == func)	// Already there.
			return 1;
	for (i = 0; i < cnt; i++)
		if (!list[i]) {
			list[i] = func;
			return 1;
		}
	return 0;
}

// Remove func from cnt-element list of functions for the buffer buf.
delete_from_buffer_func_list(int (**list)(), int cnt, int buf, int (*func)())
{
	int i;
	save_var bufnum = buf;
	for (i = 0; i < cnt; i++)
		if (list[i] == func)
			list[i] = 0;
}

#define MAX_BUFFER_WHEN_IDLE	5
buffer int (*buffer_when_idle[MAX_BUFFER_WHEN_IDLE])();

add_buffer_when_idle(int buf, int (*func)())
{
	if (!add_to_buffer_func_list(buffer_when_idle, MAX_BUFFER_WHEN_IDLE,
								 buf, func))
		error("Too many idle-time functions for this buffer.");
}

delete_buffer_when_idle(int buf, int (*func)())
{
	delete_from_buffer_func_list(buffer_when_idle, MAX_BUFFER_WHEN_IDLE,
								 buf, func);
}

diff_style(int style)
{
	char newbuf[FNAMELEN];

	tiled_only();
	if (number_of_user_windows() < 2)
		error("Only one window");
	iter = 0;
	if (has_arg)
		get_buf(newbuf, "List differences to buffer", "#diff#");
	else
		strcpy(newbuf, "#diff#");
	do_diff(newbuf, style);
	locate_window(newbuf, "");
	discardable_buffer = 1;
	point = 0;
}

// Copy src to dest, putting \ before shell meta chars.
quote_for_shell(char *dest, char *src)
{
	if (opsys != OS_UNIX) {
		sprintf(dest, "\"%s\"", src);
		return;
	}
	while (*src) {
		if (!isalnum(*src) && !strchr("-~/", *src))
			*dest++ = '\\';
		*dest++ = *src++;
	}
	*dest = 0;
}

do_shell_command(char *file, char *dir)
{
	char cmd[FNAMELEN], cmdline[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN];
	int sync = 1;

	sprintf(pr, "! on %s: ", get_tail(file, 0));
	get_string(cmd, pr);
	if (strlen(cmd) > 1 && cmd[strlen(cmd) - 1] == '&') {
		cmd[strlen(cmd) - 1] = 0;
		sync = 0;
	}
	sprintf(buf, "shell-output-%s-%s", cmd, get_tail(file, 0));
	if (!strchr(cmd, '*'))
		strcat(cmd, " *");
	strcpy(cmdline, cmd);
	strcpy(strchr(cmdline, '*'), " ");
	quote_for_shell(cmdline + strlen(cmdline), file);
	strcat(cmdline, strchr(cmd, '*') + 1);
	if (opsys == OS_UNIX || !sync) {
		pipe_text(NULL, buf, cmdline, dir,
				  (sync ? PIPE_SYNCH : 0) | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			point = 0;
		}
	} else
		do_push(cmdline, 1, 1);
}

// Return 1 if ok, 0 if error.
int dired_do_set_attributes(char *fname, char *dir)
{
	char flags[FNAMELEN], cmd[FNAMELEN], buf[FNAMELEN], msg[FNAMELEN];
	char abbrev[FNAMELEN];

	save_var _show_help_on = "dired-mode";
	strcpy(abbrev, fname);
	abbreviate_file_name(abbrev, screen_cols - 50);
	if (is_unix) {
		sprintf(msg, "Flags for chmod command on %s", abbrev);
		get_str_auto_def(flags, msg);
		if (!*flags)
			return 0;
		sprintf(cmd, "chmod %s ", flags);
		quote_for_shell(cmd + strlen(cmd), fname);
		strcpy(buf, "-dired-set-attributes-result");
		pipe_text(NULL, buf, cmd, dir,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			discardable_buffer = 1;
			point = 0;
			return 0;
		}
	} else {
		struct file_info fi;
		int attrs;

		if (!check_file(fname, &fi))
			error("Can't get attributes for %f", fname);
		attrs = fi.opsysattr;
		sprintf(cmd, "%s%s%s%s",
				(attrs & DOS_ATTR_RDONLY) ? "R" : "",
				(attrs & DOS_ATTR_SYSTEM) ? "S" : "",
				(attrs & DOS_ATTR_HIDDEN) ? "H" : "",
				(attrs & DOS_ATTR_ARCHIVE) ? "A" : "");
		sprintf(msg, "Set attributes (RO, Sys, Hid, Arch) for %s to: ", abbrev);
		get_strnone(flags, msg, cmd);
		save_var case_fold = 1;
		attrs = (strstr(flags, "R") ?  DOS_ATTR_RDONLY : 0)
				| (strstr(flags, "S") ? DOS_ATTR_SYSTEM : 0)
				| (strstr(flags, "H") ? DOS_ATTR_HIDDEN : 0)
				| (strstr(flags, "A") ? DOS_ATTR_ARCHIVE : 0);
		set_file_opsys_attribute(fname, attrs);
	}
	return 1;
}

dired_isdir_unix()
{
	save_var point;
	to_begin_line();
	save_var case_fold = 1;
	if (!parse_string(1, dired_dirsearch_unix))
		return 0;
	switch (character(matchend - 1)) {
		case 'a':
		case 'l':	return 3;
		case 'A':
		case 'L':	return 2;
	}
	return 1;
}

dired_sorter(start, end)
{
	int tmp;
	char *pat = DIRSEARCH;
	int i = get_dired_item("dired-dirsearch-", 0);

	if (i)
		pat = get_str_var(i);
	save_var case_fold = 1, sort_case_fold = (dired_format != DF_UNIX);
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	if (dired_groups_dirs) {
		save_var point = 0;
		tmp = tmp_buf();
		int dirs = tmp_buf();	// Put ., then .., in here, don't sort.
		while (column_color_searching(REGEX, "<.>[/\\]?$", goal_column, -1)
			   && !user_abort)
			move_line_to_buffer(dirs);
		point = 0;
		while (column_color_searching(REGEX, "<.><.>[/\\]?$", goal_column, -1)
			   && !user_abort)
			move_line_to_buffer(dirs);
		point = 0;
		while (re_search(1, pat) && !user_abort)
			move_line_to_buffer(tmp);
		dired_group_sorter(start, point);
		point = 0;
		grab_buffer(dirs);
		buf_delete(dirs);
		grab_buffer(tmp);
		buf_delete(tmp);
		dired_group_sorter(mark, point);
	} else
		dired_group_sorter(start, end);
}

command set_tab_size()
{
	char res[FNAMELEN];

	tab_size = get_number("Tab size: ");
	if (tab_size < 2 || tab_size > 256)
		tab_size = 8;
	sprintf(res, "%d", tab_size);
	set_var_val(find_index("tab-size"), res, 0);
	say("Tab size is %d in this buffer.", tab_size);
	build_first = 1;
}

make_mode()	/* tell eps to rebuild mode line for each window showing */
{		/* this buf */
	must_build_mode = 1;
	int i = name_match("do_when_make_mode_", 1);
	for (; i; i = name_match("do_when_make_mode_", 0))
		if (name_type(i) == NT_SUBR)
			(*(int (*)())i)();
}

command show_version()
{
	say("Epsilon Programmer's Editor %s, EEL version %s.", version, eel_version);
}

#define HTTP_HEADERS	"HTTP Headers"

http_activity(int activity, int buf, int from, int to)
{
	int headers, pos, orig;
	char err[FNAMELEN];

	UNREFERENCED_PARAMETER(to);
	if (!buf_exist(buf))
		return;
	save_var bufnum = buf;
	if (activity == NET_DONE) {
		set_mode_message(NULL);
		refresh();
		return;
	}
	if (!from)
		position_in_all_windows(buf, 0);
	bufnum = buf;
	if (mode_extra && !strstr(mode_extra, "Connecting")) {
		if (from < 1000 && mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
		return;
	}
	save_spot point = 0;
	if (re_search(1, "\r?\n\r?\n")) {
		set_mode_message("[HTTP: Retrieving] ");
		pos = point;
		headers = create(HTTP_HEADERS);	// Got end of headers yet?
		set_buf_point(headers, orig = buf_size(headers));
		buffer_printf(HTTP_HEADERS, "Headers for %s\n", filename);
		xfer(HTTP_HEADERS, 0, pos);
		point = 0;
		if (parse_string(1, "HTTP/[.0-9]+ [1345]")) {
			int m = matchend - 1;
			grab(m, MIN(parse_string(1, "[^\n\r]*"), 200), err);
			say("Response: %s, see buffer \"HTTP Headers\" for details.", err);
		}
		save_var modified = 0;	// Ignore following change to buffer.
		delete(0, pos);
		if (mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
		bufname = HTTP_HEADERS;
		discardable_buffer = 1;	// Deleting the headers is ok.
		save_var narrow_start = orig;
		simple_re_replace(1, "\r\n", "\n");
		translation_type = FILETYPE_MSDOS;
	}
}

scp_handle_done(struct scp_job *p)
{
	char fname[FNAMELEN];

	p->busy = 0;
	if (!buf_exist(p->destbuf))
		return;
	if (p->cmd) {
		if (p->cmd == strchr(p->cmdlist, '\n') + 1) {
			say("Connection to site was lost; reconnecting.");
			buffer_printf("-scp-trace", "Restart connection: |%s| of |%s|\n",
				  p->cmd, p->cmdlist ? p->cmdlist : "<none>");
			p->cmd = p->cmdlist;
			p->flags |= SCP_RESTART;
			return;
		}
	}
	set_buf_mode_message(p->destbuf, NULL);
	if (p->op == FTP_RECV && (p->flags & SCP_TMP_FILE)) {
		save_var bufnum = p->destbuf;		// The cmd running was supposed to
		strcpy(fname, filename);			// retrieve a file.  Now load it.
		do_file_read(p->file, translation_type);
		filename = fname;
		file_info.check_type = 0;	// Prevent checking its date.
		file_info.year = 0;			// Let session save it, though.
		call_mode(filename);
		check_filevars();
	}
	if (p->op == FTP_SEND && (p->flags & SCP_TMP_FILE)
		  && !(p->flags & SCP_CMD_ERROR)) {
		save_var bufnum = p->destbuf;
		say("%f: written.", filename);
	}
	if (p->flags & SCP_TMP_FILE)
		delete_file(p->file);
}

// Is buffer b currently used for scp?  Return op type, or -1 if none.
is_busy_doing_scp(int b)
{
	struct scp_job *p;

	save_var bufnum = b;
	for (p = scp_job; p; p = p->next)
		if (p->busy && (p->destbuf == b || p->buf == b))
			return p->op;
	return -1;
}

is_retrieving_remote_file(int b)
{
	save_var bufnum = b;
	return ftp_job && (ftp_job->operation & FTP_OP_MASK) == FTP_RECV
			|| (is_busy_doing_scp(b) & FTP_OP_MASK) == FTP_RECV;
}

check_dates(save)
{
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum, pos;

	if (num_of_popups() > 0)	// Don't pop up over bufed or similar.
		return 0;
	if (file_date_skip_drives && check_dates_skip_drive())
		return 0;
	ignoring_file_change = 0;	// Resume checking on getting focus.
	if (!want_warn || !filename || then->check_type != CHECK_FILE
			|| check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if ((readonly_warning & ROWARN_REFRESH) && !save
		  && (file_info.attr ^ now.attr) & ATTR_READONLY)
		return 2;

	if (!compare_dates(&now, then)) {
		maybe_add_readonly_once(&now);
		return 0;
	}
	switch_to_buffer(bufnum);
	maybe_add_readonly_once(&now);

	if ((auto_read_changed_file || _force_auto_read_changed_file)
		&& !save && maybe_auto_read(&now))
		return 1;
	if ((readonly_warning & (ROWARN_GREP | ROWARN_REFRESH)) && !save)
		return 2;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 30;
	save_var return_raw_buttons = 1;
	while (!user_abort) {
		show_date_warning(temp, orig, then, &now, save);
		bufnum = orig;
		pos = point;
		if (is_gui && want_gui_prompts) {
			int win, owin = window_handle;

			remove_final_view();
			save_var _doing_input = DI_VIEW;
			win = add_popup(0, 0, 75, 15, _view_border, temp);
			go_to_link = 0;
			display_dialog_box("FileDateBox2", "File has changed on disk",
							   win, 0, 0,
							   "Compare", "Read", save ? "Save" : "Ignore");
			text_color = (is_gui && want_gui_prompts) ? color_class text
						 : color_class viewed_text;
			point = 0;
			view_loop(win);
			switch (key_is_button) {
				case 1:	key = 'C'; break;
				case 2:	key = 'R'; break;
				case 3:	key = save ? 'S' : 'I'; break;
			}
			window_handle = owin;
			remove_window(win);
			if (user_abort)
				break;
		} else {
			view_buf(temp, 0);
			if (key == abort_key)
				user_abort = 1;
			if (user_abort)
				break;
			getkey();
		}
		char junk[FNAMELEN];
		if (key == WIN_DRAG_DROP)	// Ignore dropped files here.
			while(drag_drop_result(junk))
				;
		key = toupper(key);
		if (key == 'C') {
			ignoring_file_change = 2;	// Don't check for a while.
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, get_default_translation_type(filename));
			point = pos;
			to_begin_line();
			set_buffer_filename("");
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, reading_translation_type()))
				say("%s: Read from disk.", filename);
			point = pos;
			to_begin_line();	// Don't check immediately after dialog ends.
			ignoring_file_change = is_gui ? 2 : 0;
			break;
		} else if (key == 'D') {
			save_var compare_to_prior_version_style = 1;
			switch (do_compare_to_prior_version(0)) {
				case 1:
					say("The version on disk is identical.");
					delay(200, 0);
					break;
				case 0:
					say("Displaying changes from disk version to edited version.");
					window_one();
					buf_delete(temp);
					quick_abort();		/* abort out of calling func */
					break;
			}
			break;
		} else if (key == (save ? 'S' : 'I')) {
			if (key == 'I')
				ignoring_file_change = 100;	// Don't check on getting focus.
			else
				ignoring_file_change = 2;	// Don't check immediately after dialog ends.
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		else if (key == 'A')
			ask_read_changed_file();
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

do_save_file(backup, checkdate, getdate) /* 0=ok, 1=err, 2=user canceled */
{				/* always prints status in echo area */
	struct file_info *finfo;
	int err, old_abort = user_abort;

	iter = 0;
	if (!*filename || force_save_as)
		return write_file();
	if (checkdate && check_dates(1)) {
		user_abort = old_abort;	// Ignore aborting from modified-on-disk prompt.
		return 2;
	}
	if ((buffer_flags(0) & BF_UNICODE) && !GET_ENCODING(translation_type)
		  && (has_feature & FEAT_UNICODE)) {
		if (!buffer_has_pattern("<#256-#65535>"))
			save_var translation_type = SET_ENCODING(translation_type, 1);
		else		// Force to raw temporarily.
			pick_write_encoding();	// Have 16-bit chars, get an encoding for them.
		if (!GET_ENCODING(translation_type))
			error("Invalid encoding.");
	}
	if (_read_aborted)
		if (!ask_confirm_save((_read_aborted == 1)
							  ? "Only part of %s was read from disk.  "
							  "Save anyway? "
							  : "Error when reading %s.  Save anyway? "))
			return 2;
	if (is_retrieving_remote_file(bufnum)
		  && !ask_confirm_save("%s is still being downloaded.  "
							   "Save partial file now? "))
			return 2;
	if (backup)
		do_write_backup();
	prepare_save_file();
	finfo = getdate ? &file_info : 0;
	if (err = file_convert_write(filename, translation_type, finfo)) {
		auto_read_changed_file = 0;
		file_info.check_type = 0;	// Prevent checking date.
		file_error(err, filename, "write error");
		maybe_ding(bell_on_write_error);
	} else {
		if (!save_without_prompt && backup != 2)
			note_written(filename);
		undo_set_flag(UNDO_WRITE_FILE);
		if (finfo)
			update_readonly_warning(finfo);
		maybe_delete_autosave();
	}
	return (err != 0);
}

autosave_ok(int buf)
{
	save_var bufnum = buf;
	return want_auto_save && *filename && !buffer_not_saveable
			&& !discardable_buffer && !is_remote_file(filename)
			&& check_file(filename) == CHECK_FILE;
}

delete_all_autosaves()
{
	save_var bufnum;
	if (!want_auto_save || (want_auto_save & 8))
		return;		// No auto-save, or no auto-cleanup, do nothing.
	int i = buf_list(0, 0);
	do {
		bufnum = i;
		maybe_delete_autosave();
	} while (i = buf_list(1, 1));
}

maybe_delete_autosave()
{
	if (!(want_auto_save & 4) && strcmp(auto_save_name, "%f")
		  && autosave_ok(bufnum)) {
		char name[FNAMELEN];
		build_filename(name, auto_save_name, filename);
		delete_file(name);
		if (check_file(filename) != CHECK_FILE)
			error("Deleted auto-saved file and now original file is gone: %f",
				  filename);
	}
}

do_when_idle_autosave(int cnt)
{
	if (auto_save_idle_seconds < 1 || !want_auto_save)
		return -1;
	if (!cnt)
		time_begin(&auto_save_idle_timer, auto_save_idle_seconds * 100);
	int left = time_remaining(&auto_save_idle_timer);
	if (left > 0)
		return left;
	_auto_save_counter = 0;
	auto_save_buffers();
	return -1;
}

auto_save_buffers()	/* maybe save buffers */
{
	int i;

	save_var bufnum;
	i = buf_list(0, 0);	/* start with oldest buf */
	do {
		bufnum = i;
		if (modified && autosave_ok(bufnum) && size() <= auto_save_biggest_file)
			auto_save_one();
	} while (i = buf_list(1, 1));
}

auto_save_one()
{
	char name[FNAMELEN], altname[FNAMELEN];
	int err, origfile = !strcmp(auto_save_name, "%f");

	build_filename(name, auto_save_name, filename);
	if (_read_aborted || !autosave_ok(bufnum))		// If we didn't read
		return;						// whole file in, don't save it.
	else if (!origfile) {			// auto-save to a different file
		sprintf(altname, "%s.epstmp", name);
		delete_file(altname);
		rename_file(name, altname);
		err = file_convert_write(name, translation_type, NULL);
		save_var case_fold = (get_foldfile() != 0);
		if (strfcmp(altname, name))
			delete_file(altname);
	} else if (check_dates(1))	// make sure other prog hasn't changed
		return;
	else				// get new time info
		err = file_convert_write(filename, translation_type, &file_info);
	if (err) {
		save_var no_popup_errors = 1; // Don't pop up dialog when autosave fails.
		file_error(err, name, "couldn't autosave");
		maybe_ding(bell_on_autosave_error);
	} else if (want_auto_save & 2)
		show_text(0, 1, "%f autosaved.", name);
	if (!origfile)
		modified = 1;
}

#define URL_PATTERN	"|((ftp|http|scp|telnet|ssh):[/\\][/\\][^\n\t >]+)"

default_find_linked_file(char *file)
{
	char *pat;

	save_var case_fold = 1;
	if (opsys == OS_UNIX)
		pat = "[+-9A-z~]+[/.][+-9A-z]*[--9+A-z]" URL_PATTERN;
	else
		pat = "([a-zA-Z]:)?[+-9A-z~]+<.>[+-9A-z]*[--9+A-z]" URL_PATTERN;
	point -= parse_string(-1, "[^ \t\n!-*]*");
	if (!re_search(1, pat)) {
		to_begin_line();
		if (!re_search(1, pat))
			error("Can't find any file name on this line.");
	}
	if (strchr("\"'`<", character(matchstart)))
		matchstart++;
	if (strchr(")", character(matchend - 1))
		  && strchr("\"'`.,>", character(matchend - 2)))
		matchend--;
	if (strchr("\"'`.,>", character(matchend - 1)))
		matchend--;
	grab(matchstart, matchend, file);
}

// Look for an .epsilon_vars file with settings for many files and apply
// its settings to the current buffer.
apply_vars_file()
{
	char varsfile[FNAMELEN], bname[FNAMELEN], pat[FNAMELEN];
	char *ext = get_extension(filename), *mode = major_mode;
	int orig = bufnum;

	if (!stricmp(get_tail(filename, 1), ".epsilon_vars")
		  || !get_buffer_directory(varsfile) || is_remote_file(filename))
		return;
	strcat(varsfile, ".epsilon_vars");
	if (check_file(varsfile) != CHECK_FILE)
		return;
	save_var bufnum;
	if (!look_file(varsfile)) {	// Load the .epsilon_vars file we found.
		make_bname(varsfile, bname);
		bufnum = create(bname);
		find_in_this_buf(varsfile, get_default_translation_type(varsfile));
	}
	int tmp = tmp_buf();	// Make a buffer with just the matching settings.
	buf_printf(tmp, "Local Variables:\n");
	on_exit buf_delete(tmp);
	save_var bufnum;
	save_var point = 0, case_fold = 0, abort_searching = ABORT_ERROR;
	while (re_search(1, "^(Extensions|Modes): *([^ \n]*)") == 1) {
		grab(find_group(2, 1), find_group(2, 0), pat);
		if (fpatmatch(toupper(character(matchstart)) == 'E'
								   ? ext : mode, pat, 0, 0)) {
			int start = give_end_line();
			if (re_search(1, "^(Extensions|Modes): *"))
				to_begin_line();
			buf_xfer(tmp, start, point);
			buf_printf(tmp, "\n");
		}
	}
	bufnum = tmp;	// Now apply settings from tmp to orig buffer.
	point = 0;
	simple_re_replace(1, "^#.*", "");
	check_buffer_filevars(orig);
	if (!retain_forced_mode)
		check_filevar_mode(orig);
}

look_up_tree(res, file, dir, stop)
char *res, *file, *dir, *stop;
{
	char subdir[FNAMELEN], stopdir[FNAMELEN];
	int len, last = 1000;

	strcpy(stopdir, stop ? stop : "");
	if (*stopdir == '~')		// Expand home directory for Unix.
		absolute(stopdir);
	if (dir)
		strcpy(subdir, dir);
	else
		getcd(subdir);
	for (;;) {
		if (stopdir && *stopdir && fnamecmp(subdir, stopdir)
			&& !is_in_tree(subdir, stopdir))
			return 0;
		strcpy(res, file);
		absolute(res, subdir);
		if (check_file(res, NULL) == CHECK_FILE)
			return 1;
		if (strchr(file, '\\') || strchr(file, '/') || strchr(file, ':'))
			return 0;
		*get_tail(subdir, 0) = 0;
		len = strlen(subdir);
		if (is_path_separator(subdir[0]) && is_path_separator(subdir[1])
			  && !strpbrk_cnt(subdir, "/\\", 3))
			return 0;				// A UNC path; stop when we hit share name.
		if (is_remote_file(subdir) && !strpbrk_cnt(subdir, "/\\", 2))
			return 0;
		if (len > 3 && is_path_separator(subdir[len - 1]))
			subdir[--len] = 0;
		if (len >= last)
			break;
		last = len;
	}
	return 0;
}

// Get user to select an encoding for writing this file.
pick_write_encoding()
{
	int tmp = tmp_buf();
	int t = GET_LINE_TRANSLATE(translation_type);
	int e = GET_ENCODING(translation_type);
	int def = encoding_from_name(default_write_encoding);

	save_var window_bufnum;
	to_buffer_num(bufnum);
	save_var _view_title = "Choose an encoding";
	buf_printf(tmp, "%s\ncontains characters that cannot be represented\n",
			   bufname);
	buf_printf(tmp, "in a disk file without using an encoding.  Please choose an\n");
	buf_printf(tmp, "encoding for this file.  The default encoding shown or any\n");
	buf_printf(tmp, "of the UTF- encodings will preserve all the characters.\n");
	buf_printf(tmp, "See \"Using National Characters\" in the online manual.\n\n");
	buf_printf(tmp, "Here are some of the characters that can't be represented\n");
	buf_printf(tmp, "without an encoding:\n\n");
	grab_regex_match_lines(tmp, "<#256-#65535>", " ", " ", 10000);
	view_buf(tmp, 0);
	if (def < 0)
		def = 2;		// Default to UTF-8-No-BOM
	save_var translation_type = MAKE_TRANSLATE(def, t, 1);
	e = ask_encoding("Set buffer encoding to", 1);
	iter = 1;
	buf_delete(tmp);
	restore_vars();
	if (e < 1)
		e = def;
	translation_type = MAKE_TRANSLATE(e, t, (e || t != FILETYPE_AUTO));
	make_mode();
}

get_foldfile()
{
	if (opsys == OS_UNIX && is_unix_flavor != IS_UNIX_MACOS)
		return 0;
	return COMP_FOLD;
}

#define TEX_PARAG_SEP	"^\\(begin|end)|^[ \t]*(%%|%$%$|\\[][])"

command backward_paragraph()
{
	do_shift_selects();
	if (tex_paragraphs) {
		if (character(point - 1) == '\n')
			re_search(-1, "^(([@.]|" TEX_PARAG_SEP ").*\n)*");
		re_search(-1, "[^ \t\n]");
		if (matches_at(give_begin_line(), 1, TEX_PARAG_SEP))
			to_begin_line();
		if (indents_separate_paragraphs) {
			if (re_search(-1, "\n[ \t\n]|\f|" TEX_PARAG_SEP))
				re_search(1, "\n*");
		} else
			re_search(-1, "(\n[ \t]*\n|\f|" TEX_PARAG_SEP ")!");
		if (curchar() != '\f')
			re_search(1, "^(([@.\f]|" TEX_PARAG_SEP ").*\n)*");
	} else {
		re_search(-1, "[^ \t\n]");
		if (indents_separate_paragraphs) {
			if (re_search(-1, "\n[ \t\n]|\f|\n(" TEX_PARAG_SEP ")"))
				re_search(1, "\n*");
		} else
			re_search(-1, "((\n|<bob>)[ \t]*\n|\f)!");
	}
}

/* determine where to break current line given right margin */
pick_break(col)
int col;
{
	int orig = point, start, colpos;

	to_begin_line();
	re_search(1, "[ \t]*");	/* skip indentation */
	start = point;
	move_to_column(col);	/* find first space before col */
	colpos = point;
	while (re_search(-1, "[ \t]")) {
		if (point <= start)
			break;
		if (!mode_restrict_break || mode_restrict_break(point)) {
			point++;
			return 1;
		}
	}
	point = colpos;		/* else find first space after col */
	while (re_search(1, "[ \t]")) {
		if (point >= orig)
			break;
		if (!mode_restrict_break || mode_restrict_break(point - 1)) {
			point--;
			return 1;
		}
	}
	point = orig;
	return 0;
}

int maybe_break_this_line()
{
	int h, found = 1;
	spot old;

	h = current_column();
	if (h >= margin_right) {
		old = alloc_spot(1);
		if (h == margin_right && (key == ' ' || key == '\n')) {
			break_line_here();	/* just change the ' ' to a '\n' */
			return 0;
		}
		found = pick_break(margin_right);
		if (found) {
			break_line_here();
			if (IS_ENTER_KEY(key))
				full_key = key = ' ';
			point = *old;
		} else if (key == ' ')
			break_line_here();
		free_spot(old);
		if (!found && key == ' ')
			return 0;
	}
	return 1;
}

command buffer_grep()
{
	grep_prompt(REGEX | UNKNOWN, 1);
}

prompt_for_help()
{
	for (;;) {
		mention("Help on (? for choices): ");
		do {
			getkey();
		} while (!IS_TRUE_KEY(key));
		check_abort();
		say("");
		if (binding_of(key, root_keys) == (short) help || key == '?')
			dodesc("help", "help", 2);
		else if (key < 128)
			switch (tolower(key)) {
				case 'k': describe_key(); return;
				case 'c': describe_command(); return;
				case 'a': apropos(); return;
				case 'b': show_bindings(); return;
				case 'q': what_is(); return;
				case 'l': show_last_keys(); return;
				case 'v': about_epsilon(); return;
				case 'r': describe_variable(); return;
				case 'm': help_on_mode(); return;
				case 'i': info(); return;
				case 'f': epsilon_info_look_up(); return;
				case 'w': windows_only_help();
						  win_help_contents(""); return;
				case 'h': epsilon_html_look_up(); return;
				case 't': context_help(); return;
				case CTRL('C'): info_goto_epsilon_command(); return;
				case CTRL('K'): info_goto_epsilon_key(); return;
				case CTRL('V'):
				case CTRL('R'): info_goto_epsilon_variable(); return;
			}
	}
}

int mshelp2_wait_until_pat(char *pat)
{
	int i;
	for (i = 0; i < 60; i++) {
		point = 0;
		if (re_search(1, pat))
			return 1;
		delay(100, COND_PROC, bufnum);
		if (i > 5)
			noteput("Starting MS-Help2...  Giving up in %d seconds", 60 - i);
	}
	return 0;
}

// If there's a collection with a date, use it.  Prefer later dates to
// earlier ones.
mshelp2_choose_collection(char *col)
{
	if (*mshelp2_collection) {		// Override default selection.
		strcpy(col, mshelp2_collection);
		return;
	}
	*col = 0;
	if (!mshelp2_wait_until_pat("Using default collection.*\n"))
		return;
	point = matchend;
	bprintf("listcoll\n");
	int before = size();
	if (!mshelp2_wait_until_pat("^--End\n"))
		return;
	int tmp = tmp_buf();
	buf_xfer(tmp, before, matchstart);
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	case_fold = 1;
	point = 0;
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ", pat[100];
	int year, month;
	for (year = 2020; year > 2000; year--) {
		for (month = 12; month > 0; ) {
			sprintf(pat, "qtr.%d%.3s", year, months + --month * 4);
			point = 0;
			if (search(1, pat)) {
				grab_full_line(bufnum, col);
				return;
			}
		}
		sprintf(pat, "%d", year);
		point = 0;
		if (search(1, pat)) {
			grab_full_line(bufnum, col);
			return;
		}
	}
}

show_windows_help2(char *topic)
{
	char prog[FNAMELEN], cmdline[FNAMELEN], *s, col[FNAMELEN];
	int talk = 0;

	if (!is_win32)
		error("This type of help is only available under Windows.");
	save_var bufnum = create("-windows-help2-helper");
	if (!is_process_buffer(bufnum)) {
		if (size())
			say("Possible problem running mshelp2.vbs helper detected.");
		if (RegOpenKey(HKEY_CLASSES_ROOT, "AppID\\Dexplore.exe", &key))
			return 1;
		RegCloseKey(key);
		sayput("Starting MS-Help2...");
		talk = 1;
		s = lookpath("mshelp2.vbs");
		if (!s) {
			get_executable_directory(prog);
			strcat(prog, "mshelp2.vbs");
		} else
			strcpy(prog, s);
		convert_to_8_3_filename(prog);
		sprintf(cmdline, "cscript %s", prog);
		if (concur_shell("", cmdline, NULL, bufname))
			return 2;
		mshelp2_choose_collection(col);
		point = size();
		if (*col)
			bprintf("collection %s\n", col);
	}
	point = size();
	bprintf("keyword %s\n", topic);
	delay(10, 0);
	if (talk)
		say("");
	if (!is_process_buffer(bufnum))
		return 3;
	return 0;
}

// When Epsilon exits, shut down the helper script.  Otherwise the
// external help program will never exit, even if the user closes its
// window.  (Unfortunately this will force the external help program to
// close when Epsilon closes.)
do_when_exiting_mshelp2(int forced)
{
	int cnt = 1;
	UNREFERENCED_PARAMETER(forced);
	if (!exist("-windows-help2-helper"))
		return;
	save_var bufnum = create("-windows-help2-helper");
	if (!is_process_buffer(bufnum))
		return;
	while (cnt <= 5) {
		point = size();
		stuff("exit\r\n");
		delay(500, COND_PROC_EXIT | COND_TRUE_KEY, bufnum);
		if (!is_process_buffer(bufnum))
			return;
		check_abort();
		note("Attempt %d/5: Can't shut down helper in -windows-help2-helper.",
			 ++cnt);
	}
}

// Look for MSDev help in typical places.  Add logic for other C++
// compilers.
context_help_windows_compilers()
{
	char item[FNAMELEN];

	context_help_get_item(item, word_pattern);
	if (!show_windows_help2(item))
		return;
	do_context_help_from_msdn(item);
}

context_help_from_msdn()
{
	char item[FNAMELEN];

	context_help_get_item(item, word_pattern);
	do_context_help_from_msdn(item);
}

do_context_help_from_msdn(char *item)
{
	char dir[FNAMELEN * 2], *s;
	int sz = sizeof(dir) / sizeof(*dir), key;

	strcpy(dir, "C:\\Program Files");
	if (!RegOpenKey(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\"
					"Windows\\CurrentVersion", &key)) {
		RegQueryValueEx(key, "ProgramFilesDir", NULL, dir, &sz);
		RegCloseKey(key);
	}
	strcat(dir, "\\*\\**\\msdn*.col");
	if (s = do_file_match(dir, STARTMATCH | EXACTONLY))
		win_help_string(s, item);
	else
		error("Can't find help files.  Set the "
			  "context-help-rule-... variable for this mode.");
}

dodesc(char *str, char *also, int flags)
{
	char *old = bufname, msg[80];
	int start, i, browser_ok = !(flags & 2);

	if (use_winhelp()) {
		show_help_for(str);
		return;
	}
	if (!get_doc(0))
		return;
	i = find_index(str);
	if (!i || name_help(i) < 0) {
		gripe("No information on %s", str, flags & 1);
		bufname = old;
		return;
	}
	point = name_help(i);
	while (isspace(curchar()))
		point++;
	start = point;
	re_search(1, "^!~");
	xfer(HELPBUF, start, point);
	bufname = HELPBUF;
	if (also && *also)
		point = 0, stuff(also);
	insert('\n');
	point = size();
	if (is_unix != IS_UNIX_XWIN && !is_gui && !is_win32)
		browser_ok = 0;
	if (!(flags & 1) && browser_ok)
		bprintf("Press H to see this topic in a web browser, with\n"
				"links to the rest of Epsilon's online manual.\n");
	point = 0;
	bufname = old;
	sprintf(msg, " Help on %s ", str);
	view_help(msg, flags & 1);
	if (browser_ok && tolower(key) == 'h') {
		ungot_key = -1;
		html_help_string(str);
		drop_pending_says();
	}
}

css_indenter()
{
	int orig, ind, pos;

	to_begin_line();
	orig = point;
	if (Closeback && parse_string(1, "[ \t]*}")) {
		point = matchend;
		if ((pos = default_move_level(-1, "}", "{")) >= 0)
			ind = get_indentation(pos);
	} else if (!re_search(-1, "[^ \t\n]"))	// Find previous non-blank line.
		ind = 0;
	else if ((pos = css_unclosed(give_begin_line(), orig)) >= 0) {
		// If there's an unmatched ( between start of this line and
		// orig, align to text after (.
		point = pos;
		point += parse_string(1, "{[ \t]*");
		ind = current_column();
	} else  {
		ind = get_indentation(point);
		to_end_line();
		if (parse_string(-1, "{[ \t]*"))
			ind += css_indent;
		else if (parse_string(-1, "}[ \t]*")) {
			if ((pos = default_move_level(-1, "}", "{")) >= 0) {
				// If prev line ended in }, don't use its indent, but
				// indent of word before matching { line.
				point = pos;
				if (parse_string(-1, "<word><wspace>+"))
					point = matchend;
				ind = get_indentation(point);
			} else if (!Closeback)
				ind -= css_indent;
		}
	}
	point = orig;
	to_indentation();
	indent_to_column(ind);
}

css_indent_and_show_matching_delimiter()
{
	normal_character();
	save_spot point;
	css_indenter();
	save_var inside_show_matching_delimiter = 1;
	restore_vars();
	if (Matchdelim && strchr(")]}", key))
		find_delimiter();
}

html_find_open_tag(int pos, char *main, char *ok, int flags)
{
	char pat[FNAMELEN], tag[FNAMELEN];
	save_var case_fold = (sublanguage_code != SUBLANG_XML), point = pos;
	sprintf(pat, "<c:html-punctuation>[<](/?(%s))</" HTML_NAME_CHAR ">", main);
	if (do_color_searching(REGEX | ((flags & 1) ? REVERSE : 0), pat) != 1)
		return -2;
	strcpy(tag, "|");
	grab(find_group(1, 1), find_group(1, 0), tag + 1);
	strcat(tag, "|");
	sprintf(pat, "|%s|", ok);
	if (strstr(pat, tag))
		return point;
	return -1;
}

int html_tag_level()
{
	int level = 0;

	if (parse_string(1, "[ \t]*[<]/") && html_is_start_tag(matchend - 2, 1))
		level++;
	while (do_color_searching(REGEX, "<c:html-punctuation>[<]") == 1)
		switch (html_is_start_tag(point - 1, 1)) {
			case -1:	level--; break;
			case 1:		level++; break;
		}
	if (html_indenting_rules & 128)
		level = (level > 0) ? 1 : (level < 0) ? -1 : 0;
	return level;
}

command html_fill_paragraph()
{
	int region, start, end;

	recolor_buffer_range(point, point + 1);
	region = html_region_type(point);
	switch (region) {
		case HT_JSCRIPT:
			save_var major_mode = _c_mode_name;
			fill_comment();
			return;
		case HT_VBSCRIPT:
		case HT_PHP:
		case HT_PYTHON:
		case HT_CSS:
			fill_comment();
			return;
	}
	if (get_character_color(point, &start, &end) == color_class html_comment)
		save_var narrow_start = start, narrow_end = size() - end;
	fill_paragraph();
}

command html_forward_tag() on html_tab[ALT('F')]
{
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]"
						   "|<c:html-text><word>+") != 1
		  || get_character_color(matchstart) == color_class html_text)
		return;
	char tag[FNAMELEN];
	int pos = point, match;
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") == 1
		  && !parse_string(1, "[<][ \t]*/")
		  && (match = do_html_find_matching_tag(point, tag)) >= 0) {
		point = match;
		do_color_searching(REGEX, "<c:html-punctuation>[>]");
		return;
	}
	point = pos;
}

// Inside an end tag, move before its matching start tag.  Inside some
// other kind of tag, move to its start.  Outside a tag, move by words.

command html_backward_tag() on html_tab[ALT('B')]
{
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]"
						   "|<c:html-text><word>+") != 1
		  || get_character_color(matchstart) == color_class html_text)
		return;
	char tag[FNAMELEN];
	int pos = point, match;
	if (parse_string(1, "[<][ \t]*/")
		  && (match = do_html_find_matching_tag(point, tag)) >= 0) {
		point = match;
		do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]");
		return;
	}
	point = pos;
}

// Delete current tag or next one.  It if has a matching start/end tag,
// delete that one too.
command html_delete_tag() on html_tab[ALT('D')]
{
	save_spot point;
	int pos = html_in_tag(), orig = point;
	char tag[FNAMELEN];
	if (pos >= 0)
		point = orig = pos;
	else if (parse_string(1, "<wspace>*[<]"))
		point = matchend - 1;
	else
		error("Not in or before a tag.");
	spot match = alloc_spot();
	on_exit free_spot(match);
	*match = do_html_find_matching_tag(point, tag);
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]") == 1)
		delete(orig, point);
	if (*match >= 0) {
		point = *match;
		if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") == 1) {
			orig = point;
			if (do_color_searching(REGEX, "<c:html-punctuation>[>]") == 1)
				delete(orig, point);
		}
	}
}

xml_sort_attrs_for_one_tag(int pos)
{
	save_var point = pos;
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") != 1)
		return;
	point += parse_string(1, "[<]<wspace>*" HTML_NAME "<wspace>+");
	save_var narrow_start = point;
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]") != 1)
		return;
	point -= parse_string(-1, "<wspace>*/?[>]");
	save_var narrow_end = size() - point, abort_searching = 0;
	simple_re_replace(1, "\n", XML_SORT_NL_MARKER);
	point = 0;
	stuff(XML_SORT_MARKER);
	while (point < size() - narrow_end && !user_abort) {
		point += parse_string(1, HTML_NAME XML_SORT_WSPACE "*");
		int len;
		if (curchar() == '=') {
			len = parse_string(1, "=" XML_SORT_WSPACE "*(<^wspace|\"|'>+"
							   "|\"[^\"]*\"|'[^']*')");
			if (!len)
				break;
			point += len;
		}
		len = parse_string(1, XML_SORT_WSPACE "+");
		if (!len)
			break;
		delete(point, point + len);
		stuff("\n" XML_SORT_MARKER);
	}
	point = size();
	stuff("\n");
	do_sort_region(narrow_start, size() - narrow_end, 0, 0);
	point = 0;
	delete(point, point + parse_string(1, XML_SORT_MARKER));
	simple_re_replace(1, "\n?^" XML_SORT_MARKER, " ");
	point = size();
	delete(point - parse_string(-1, "\n?"), point);
	simple_re_replace(1, XML_SORT_NL_MARKER, "\n");
}

// Fill in name with the name of the n'th attribute (from 0) in the
// current tag and return 1.  If not that many, return 0.
int xml_find_nth_attr(int n, char *name)
{
	*name = 0;
	point = 0;
	if (do_color_searching(REGEX, "<c:html-punctuation>[<]") != 1)
		return 0;
	point += parse_string(1, "<wspace>*" HTML_NAME "<wspace>+");
	save_var narrow_start = point;
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]") != 1)
		return 0;
	point -= parse_string(-1, "<wspace>*/?[>]");
	save_var narrow_end = size() - point, abort_searching = 0;
	point = 0;
	while (point < size() - narrow_end && !user_abort) {
		point += parse_string(1, HTML_NAME, name);
		if (strlen(name) <= 0)
			return 0;
		if (n-- <= 0)
			return 1;
		point += parse_string(1, "<wspace>*");
		if (curchar() == '=')
			point += parse_string(1, "=<wspace>*(<^wspace|\"|'>+"
								  "|\"[^\"]*\"|'[^']*')");
		int len = parse_string(1, "<wspace>+");
		if (!len)
			break;
		point += len;
	}
	return 0;
}

// Find cnt'th attribute of the first tag in the buffer.  Align all
// attributes with that name in the region.  Return 0 if there's no
// n'th attr.
int xml_align_one_attr(int cnt)
{
	char name[FNAMELEN], pat[FNAMELEN];
	if (!xml_find_nth_attr(cnt, name))
		return 0;
	save_spot point;
	sprintf(pat, "( )(<c:html-attribute|html-namespace>%q)", name);
	align_by_pattern(16, narrow_start, size() - narrow_end, pat, 1);
	return 1;
}

// Go to each tag and reorder its attrs, then align them.
command xml_sort_by_attribute_name() on html_tab[ALT('R')]
{
	save_spot point, mark;
	if (is_highlight_on()) {
		fix_region();
		save_var narrow_start = point;
		save_var narrow_end = size() - mark;
		point = 0;
		while (do_color_searching(REGEX, "<c:html-punctuation>[<]")) {
			xml_sort_attrs_for_one_tag(point);
			check_abort();
		}
		int cnt = 0;
		while (xml_align_one_attr(cnt++))
			check_abort();
   } else {
		point++;
		if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") == 1)
			xml_sort_attrs_for_one_tag(point + 1);
	}
}

html_trim_element_nesting(int b)
{
	save_var bufnum = b;
	int space = window_width
				- strlen(get_tail(*filename ? filename : bufname, 0))
				- html_display_nesting_width;
	if (space < 15) {		// If we'd have too little room, skip it.
		*display_func_name = 0;
		return;
	}
	point = size() - space;
	if (point > 0) {
		point += 4;		// Make room for "..."
		if (!search(1, " "))
			if (search(-1, " "))
				point++;
		delete(0, point);
		stuff("... ");
	}
	point = size() - ptrlen(display_func_name) - 10;
	if (point) {
		delete(0, point);
		stuff("... ");
	}
	grab(0, size(), display_func_name);
}

int html_restrict_break(int pos)
{
	if (get_character_syntax_color(pos) != color_class html_value)
		return 1;
	int start;
	get_character_color(pos, &start);
	return !strchr("\"'", character(start));
}

int html_maybe_break_line(int mode)
{
	char *var = (sublanguage_code == SUBLANG_XML) ? &xml_auto_fill_mode
				: &html_auto_fill_mode;
	switch (mode) {
		case 0:
		case 1:
			*var = (mode != 0);
			return 0;
		case 2:
			return *var;
	}
	if (!*var)		// We're not in auto-fill mode.
		return 1;
	recolor_buffer_range(point, point + 1);
	if (((html_region_type(point) == -1) ? 2 : 3) == *var)
		return 1;	// Only do one sort of filling?
	switch (html_region_type(point)) {
		case HT_PHP:			 // Break only if in a comment.
		case HT_VBSCRIPT:
		case HT_PYTHON:
		case HT_CSS:
			return do_maybe_break_line(0);
		case HT_JSCRIPT:
			return do_maybe_break_line(1);
	}
	save_var mode_restrict_break = html_restrict_break;
	return maybe_break_this_line();	// Break always as text.
}

int html_indenter();
int html_move_level(int dir, int stop_on_key);
int html_auto_show_delimiter();
int color_html_range(int from, int to);
int html_recolor_from_here(int safe);

html_like_mode(int force_xml)
{
	mode_default_settings();
	sublanguage_code = 0;
	mode_keys = html_tab;		/* Use these keys. */
	if (force_xml || matches_at(0, 1, "<wspace>*%<%?xml"))
		sublanguage_code = SUBLANG_XML;
	if (sublanguage_code == SUBLANG_XML) {
		major_mode = strkeep("XML");
		compile_buffer_cmd = compile_xml_cmd;
	} else {
		major_mode = strkeep("Html");
		compile_buffer_cmd = compile_html_cmd;
	}

	strcpy(comment_start, "%<%!-- *");
	strcpy(comment_pattern, "%<%!--(.|<newline>)*--%><FirstEnd>");
	strcpy(comment_begin, "<!-- ");
	strcpy(comment_end, " -->");
	buffer_maybe_break_line = html_maybe_break_line;
	indenter = html_indenter;
	auto_indent = 1;
	mode_move_level = html_move_level;
	if (auto_show_html_delimiters)
		auto_show_matching_characters = html_auto_show_delim_chars;
	mode_auto_show_delimiter = html_auto_show_delimiter;
	recolor_range = color_html_range;	// set up coloring rules
	recolor_from_here = html_recolor_from_here;
	idle_coloring_size = 5000;
	coloring_flags |= COLOR_IGNORE_INDENT;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("html-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

command php_mode()
{
	html_like_mode(0);
	compile_buffer_cmd = compile_php_cmd;
	major_mode = _php_mode_name;
	sublanguage_code = SUBLANG_PHP;
	auto_show_matching_characters = auto_show_php_delimiters
									? php_auto_show_delim_chars : 0;
	if (php_tab_override > 0)
		tab_size = php_tab_override;
	try_calling("php-mode-hook");
}

command align_region()
{
	int manual = has_arg, flags, ind;
	char num[FNAMELEN];

	iter = has_arg = 0;
	if (manual) {
		highlight_off();
		flags = get_search_string("Pattern (aligned between #1 and #2): ",
								  REGEX);
		if (flags & SEARCH_EMPTY)
			manual = 0;
		get_strdef(num, "Extra spacing", "2");
		save_var align_region_extra_space = numtoi(num);
	}
	save_spot mark, point;
	fix_region();
	if (!manual) {
		ind = get_mode_based_index("%s-mode-aligner");
		if (is_runnable(name_name(ind)))
			(*(int (*)())ind)();
		else
			error("No valid aligner for %s mode.", major_mode);
	} else
		align_by_pattern(0, point, mark, default_search_string(flags), 0);
	highlight_off();
}

int info_node_by_start(int p)
{
	char node[FNAMELEN];
	int res = 1;

	narrow_start = narrow_end = 0;
	point = p;
	if (!re_search(-1, "\n" INFO_CSTR "\f?\n")) {
		point = 0;
		res = 0;
	}
	if (re_search(1, "\n" INFO_CSTR "\f?\n(.*[\t, ])?"
				  "Node:[\t ]*([^\t\n,]+)[\t\n,]")) {
		grab(find_group(2, 1), find_group(2, 0), node);
		info_go_node(node, 0);	// Narrow to this node.
		if (p < narrow_start || p > size() - narrow_end)
			res = 0;
	} else
		res = 0;
	return res;
}

int color_ini_range(int from, int to);

command ini_mode()
{
	mode_default_settings();
	major_mode = _ini_mode_name;
	mode_keys = inifile_tab;		/* Use these keys. */
	strcpy(comment_start, "[;#][ \t]*");
	strcpy(comment_pattern, "[;#].*$");
	strcpy(comment_begin, "; ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_ini_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 256) != 0;
	try_calling("ini-mode-hook");
	make_mode();
}

when_loading()
{
	inifile_tab[ALT('q')] = (short) fill_comment;
}

command backward_delete_character()
{
	if (delete_if_highlighted())
		return;
	if (virtual_space && cursor_to_column > 0)
		virtual_move_back();
	else {
		int tab = (soft_tab_size > 0) ? soft_tab_size : tab_size;
		int hack = get_mode_variable("%s-delete-hacking-tabs");
		if (got_bad_number)
			hack = delete_hacking_tabs;
		int many = (hack & 2) && character(point - 1) == ' ' && !has_arg;
		if (many && (hack & 4) && current_column() - 1 > get_indentation(point))
			many = 0;
		if (many && (hack & 8) && current_column() % tab)
			many = 0;
		if (many && (hack & 16) && (curchar() == ' ' || curchar() == '\t'))
			many = 0;
		if (hack & 1)
			hack_tabs(-iter);
		do_delete_characters(-1);
		if (many)
			while (current_column() % tab && parse_string(-1, "[ \t]"))
				delete(point - 1, point);
	}
}

// Scan the rest of the current buffer for EEL-created things that
// couldn't be defined.  If there are any, remove them from the buffer,
// show them to the user, and return nonzero.  If none, return zero.
list_undefined_customizations()
{
	int list = 0;
	char funcname[FNAMELEN];

	while (re_search(1, "^([^ \n\t:]+): (command|subroutine)\n")) {
		grab(find_group(1, 1), find_group(1, 0), funcname);
		if (!is_runnable(funcname)) {
			if (!list) {
				list = zap("-unknown-customizations");
				save_var bufnum = list;
				bprintf("Epsilon couldn't find definitions for the functions "
						"listed below.\nIf your previous setup included any EEL "
						"extension files, you\nmay want to rerun this command "
						"and include their names.\n\n");
				restore_vars();
			}
			buf_xfer(list, matchstart, matchend);
		}
		delete(matchstart, matchend);
	}
	if (list) {
		set_buf_point(list, 0);
		view_buf(list, 1);
	}
	return list != 0;
}

int shell_with_args(char *cmd, char *args)
{
	return shell_with_args_flags(cmd, args, SHELL_KEEP_ENV);
}

shell_with_args_flags(char *cmd, char *args, int flags)
{
	char full_cmd[1000];
	int err;

	if (is_unix) {	// Run via the shell, so it splits the args at whitespace.
		quote_for_shell(full_cmd, cmd);
		strcat(full_cmd, " ");
		strcat(full_cmd, args);
		err = shell("", full_cmd, "", flags);
	} else		// Run without the shell, so there's no window.
		err = shell(cmd, args, "", flags);
	build_first = 1;
	screen_messed();
	return err;
}

// Return a buffer listing the differences in settings between orig and
// cur, two buffers with lists of settings.  Put the original settings
// corresponding to the modified settings into a buffer
// "original-settings".
int make_difference_list(int orig, int cur)
{
	noteput("Comparing...");
	save_var show_status = 0;
	scan_binding_ranges(orig, cur);
	scan_binding_ranges(cur, orig);
	save_var bufnum = orig;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	save_var bufnum = cur;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	int out = tmp_buf();
	do_compare_sorted(orig, cur, "original-settings",
					  bufnum_to_name(out), NULL);
	bufnum = out;
	point = 0;
	save_var case_fold = 1;
	custom_apply_immediately();
	// Drop vars starting with -, but not color scheme ones.
	do_drop_matching_lines(REGEX, "^%+-[^o][^u][^r][^-]", 1);
	drop_boring_change_lines();
	custom_fix_mode_vars();
	insert_key_unbindings(name_to_bufnum("original-settings"));
	do_drop_matching_lines(REGEX, "^%+(push-cmd|echo-line|permanent-menu"
						   "|avoid-(top|bottom)-lines|bufed-width"
						   "|mode-start|mode-end) ", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for after-exiting: .*$", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for .*: "
						   "%[0x4000000 on 0x0%]$", 1);
	do_drop_matching_lines(REGEX, "^~.*: case-indirect$", 1);
	// We might have original.sta from 9.0 with epsilon.sta from 9.02,
	// due to 9.02 updater failing to update original.sta.  Ignore
	// resulting differences.  We ignore the latex setting only if it
	// looks like this is a 9.0/9.02 update.
	if (do_drop_matching_lines(REGEX, "^(fix-buf-accessed-clock|fix-tex-math-coloring"
						   "|http-activity|tex-start-construct): subroutine$", 1))
		do_drop_matching_lines(REGEX, "^%+latex-2e-or-3 char variable: 0$", 1);
	do_drop_matching_lines(REGEX, "^%+ftp-passive-transfers char variable: 1$", 1);

	return out;		// Don't count vars saved in session file, others.
}

int set_font_for_unix()
{
	char buf[FNAMELEN];

	if (no_mac_dialog_helper()) {
		mac_gtk_message(1);
		return simple_set_font_for_unix();
	}
	get_executable_file(buf, "epsilon-dialogs", 1);
	sprintf(buf + strlen(buf), " --parent %x --font \"%s\"",
			screen_to_window_id(0), font_fixed);
	if (!pipe_text("-epsilon-dialog-out", "-epsilon-dialog-out", buf, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF, 0)) {
		file_error(errno, "Couldn't run dialog helper", "Error");
		return 0;
	}
	save_var bufname = "-epsilon-dialog-out";
	point = 0;
	if (parse_string(1, ".*Library not loaded")) {
		mac_gtk_message(1);
		return simple_set_font_for_unix();
	} else if (re_search(1, "^=") && parse_string(1, ".*", buf))
		font_fixed = buf;
	return 1;			// User picked one, even if it was bad.
}

// Return nonzero to indicate we should show the Mac gtk message
// instead of trying to run the dialog helper.
int no_mac_dialog_helper()
{
	if (is_unix_flavor != IS_UNIX_MACOS)
		return 0;
	return (check_file("/sw/lib/libgtk-1.2.0.dylib") != CHECK_FILE);
}

// Display a message explaining how to install needed library, trying
// to set font or colors.
mac_gtk_message(int is_font)
{
	int b = zap("-epsilon-dialog-error");
	buf_printf(b, "This command requires GTK+ 1.4 shared libraries, which ");
	buf_printf(b, "are not\ninstalled.  Install the fink application from ");
	buf_printf(b, "fink.sourceforge.net,\nthen run\n\n");
	buf_printf(b, "   fink install gtk+-shlibs\n\n");
	if (is_font)
		buf_printf(b, "Epsilon will revert to a simpler font setting method.\n");
	else {
		buf_printf(b, "For now, you can set colors by running the Alt-x ");
		buf_printf(b, "export-colors\ncommand, editing the source file it ");
		buf_printf(b, "generates, and loading it\nusing the compile-buffer ");
		buf_printf(b, "command on Alt-F3.\n");
	}
	save_var _view_title = " Can't run dialog helper ";
	view_buf(b, 0);
}

color_conf_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "([^ \t\n#:;,]+[ \t]*)="
					   "([ \t]*([^ \t\n#:;,]+|'[^\n']*'|\"[^\n\"]*\"))|^;|[#]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '#':
			case ';':
				nl_forward();
				set_character_color(t, point, color_class conf_comment);
				break;
			default:
				set_character_color(t, find_group(1, 0),
									color_class conf_attribute);
				set_character_color(find_group(2, 1), point,
									color_class conf_value);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

proc_view_error_lines(int buf, int pos)
{
	if (!process_view_error_lines)
		return;
	if (process_view_error_lines < 0
		  || process_view_error_lines > screen_lines - 2)
		process_view_error_lines = screen_lines / 3;
	int owin = window_handle;
	int win = buf_in_window(buf);
	if (win >= 0)
		window_handle = win;
	else {
		window_number = 0;
		window_number--;		// Get to last window on screen.
		while (window_height < process_view_error_lines) {
			int oh = window_height++;
			if (oh == window_height)
				if (give_window_space(BTOP))
					break;
		}
		if (window_height > process_view_error_lines) {
			window_split(HORIZONTAL);
			window_height = process_view_error_lines;
		}
		to_buffer_num(buf);
	}
	point = window_start = pos;
	if (is_window(owin))
		window_handle = owin;
	else
		window_number--;
}

proc_view_start_compile(int buf)
{
	if (!process_view_error_lines)
		return;
	if (buf_in_window(buf) >= 0)
		proc_view_error_lines(buf, buf_size(buf));
}

#define FNAME_NO_SPACE	"(([a-z]:)?[+-9A-z~]+<.>[+-9A-z]+[0-9<>]*)"
#define FNAME_OK_SPACE	"(([a-z]:)?[+-9A-z~][+-9A-z~ ]*<.>[+-9A-z]+[0-9<>]*)"

// Some things we want to ignore in an error message.

// 3> or E " or W " or ***, for example.
#define START_STUFF1	"[1-9]>|e \"|w \"|<*>*"
// One of these three keywords, or [someother].
#define START_STUFF2	"error|warning|fatal| *%[[a-z]+%]"
// Optionally, a code like E123 with a space before it, then perhaps
// some punc chars * : after.  Or just punc chars " or (.
#define START_STUFF3	"( [a-z][0-9]+)?[*:]*|[\"(]?"

// Between a file name and line number, we must have space, open paren,
// comma or colon, then perhaps L or the word "line", then perhaps
// space.  Also skip over any " at the start, which may have surrounded
// file name.
#define BEFORE_NUM		"\"?([ (:,]( *line|L)?) *"
// The line number.
#define LINE_NUM		"([0-9]+)"
// Stuff to require after line number.  Either ) or colon, or else some
// stuff with colon or semicolon after it.
#define END_STUFF		"(<rparen|:>|[^0-9\\\n].*[:;])"

/* If you change ERROR_PATTERN, you may have to change the next 3 macros
 * as well. */
#define ERROR_PATTERN	"^(" START_STUFF1 \
						"(" START_STUFF2 "|[a-z0-9]+ )" \
						START_STUFF3 ") *" \
						FNAME_NO_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN	4	/* filename is in 4th () pair above */
#define LINE_PAREN	8	/* line# is in 8th pair of parens above */

#define SIMP_PATTERN	"[:;)]| line "	/* search with simple pattern first */

// This version permits spaces in file names, but with more
// restrictions on ancillary text before it.
#define ERROR_PATTERN2	"^(" START_STUFF1 \
						"(" START_STUFF2 ")" \
						"( [a-z][0-9]+)?[*:]*|[\"(]?) *" \
						FNAME_OK_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN2	4	/* filename is in 4th () pair above */
#define LINE_PAREN2	8	/* line# is in 8th pair of parens above */

#define ERROR_PATTERN3	".* at (.+) line ([0-9]+)[.,].*"
#define FILE_PAREN3	1	/* filename in () pair above */
#define LINE_PAREN3	2	/* line# in pair of parens above */

// An error pattern that permits file names with no . character, but
// with more restrictions on the format.
#define ERROR_PATTERN4	"^([+-9A-z~:]+[0-9<>]*)" BEFORE_NUM \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN4	1	/* filename is in 1st () pair above */
#define LINE_PAREN4	2	/* line# is in 2nd pair of parens above */

// An error pattern for Java runtime errors.

#define ERROR_PATTERN5	"^.*%(([+-9A-z~:]+):([0-9]+)%)"
#define FILE_PAREN5	1	/* filename is in 1st () pair above */
#define LINE_PAREN5	2	/* line# is in 2nd pair of parens above */

// An error pattern that permits file names with no . character,
// allowing spaces, but with even more restrictions on the rest of the
// message.
#define ERROR_PATTERN6	"^([a-z]:\\[ +-9A-z~]+)[(]" \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN6	1	/* filename is in 1st () pair above */
#define LINE_PAREN6	2	/* line# is in 2nd pair of parens above */

find_error(int dir, int *file_paren, int *line_paren)
{
	int old = point;

	if (point > 0 && dir > 0)
		to_end_line();
	else
		to_begin_line();
	while (re_search(dir, SIMP_PATTERN)) {
		to_begin_line();
		if ((!*ignore_error || !parse_string(1, ignore_error, NULL))
			  && !parse_string(1, "In file included from ", NULL)) {
			if (parse_string(1, ERROR_PATTERN, (char *)0)) {
				*file_paren = FILE_PAREN;
				*line_paren = LINE_PAREN;
				return 1;
			}
#ifdef ERROR_PATTERN2
			if (parse_string(1, ERROR_PATTERN2, (char *)0)) {
				*file_paren = FILE_PAREN2;
				*line_paren = LINE_PAREN2;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN3
			if (parse_string(1, ERROR_PATTERN3, (char *)0)) {
				*file_paren = FILE_PAREN3;
				*line_paren = LINE_PAREN3;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN4
			if (parse_string(1, ERROR_PATTERN4, (char *)0)) {
				*file_paren = FILE_PAREN4;
				*line_paren = LINE_PAREN4;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN5
			if (parse_string(1, ERROR_PATTERN5, (char *)0)) {
				*file_paren = FILE_PAREN5;
				*line_paren = LINE_PAREN5;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN6
			if (parse_string(1, ERROR_PATTERN6, (char *)0)) {
				*file_paren = FILE_PAREN6;
				*line_paren = LINE_PAREN6;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN7
			if (parse_string(1, ERROR_PATTERN7, (char *)0)) {
				*file_paren = FILE_PAREN7;
				*line_paren = LINE_PAREN7;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN8
			if (parse_string(1, ERROR_PATTERN8, (char *)0)) {
				*file_paren = FILE_PAREN8;
				*line_paren = LINE_PAREN8;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN9
			if (parse_string(1, ERROR_PATTERN9, (char *)0)) {
				*file_paren = FILE_PAREN9;
				*line_paren = LINE_PAREN9;
				return 1;
			}
#endif
		}
		if (dir > 0)
			nl_forward();
	}
	*error_spot = old;
	return 0;
}

color_perl_here_doc(int start, int interpolated)
{
	char line[FNAMELEN], pat[FNAMELEN], *s;
	int p1, p2, end, type = 0;

	grab(find_group(1, 1), find_group(1, 0), line);
	p1 = give_end_line() + 1;
	if (sublanguage_code != SUBLANG_PHP
		  && get_tagged_region("color-perl-as-unit", p1) != -1) {
		get_tagged_region("color-perl-as-unit", p1, NULL, &p1);
		start = p1;
	}
	if (in_shell_buffer && parse_string(1, "[ \t]*-"))
		sprintf(pat, "^\t*%q</word>", line);
	else
		sprintf(pat, "^%q%s$", line, sublanguage_code == SUBLANG_PHP ? ";?" : "");
	if (re_search(1, pat))		// Find end of document (or eof).
		p2 = matchstart;
	else
		p2 = size();
	end = point;
	s = get_extension(line);
	if (s)
		if (!strcmp(s, ".tex"))
			type = 1;
		else if (!strcmp(s, ".html"))
			type = 2;
	set_character_color(start, point, color_class perl_constant);
	set_tagged_region("color-perl-as-unit", start, point, type);
	save_var point;
	save_var narrow_start = p1;
	save_var narrow_end = size() - p2;
	switch (type) {
		case 1: do_color_tex_range(p1, p2, interpolated); break;
		case 2: color_html_range(p1, p2); break;
	}
	restore_vars();
	if (interpolated)
		color_perl_interpolated(start, point, 0);
	point = end;
}

// Found a word; see if it's decorated like a variable or subr.
color_perl_word(int start)
{
	int true_perl = !in_shell_buffer && sublanguage_code != SUBLANG_PHP;

	if (isdigit(character(start))) {		// a number
		if (matches_at(start, -1, "-?%.?"))	// number like -.25
			start = matchend;
		point += parse_string(1, "([A-Za-z0-9._]|[Ee]-)*");		// -3.0e-7
		set_character_color(start, point, color_class perl_constant);
		return;
	}
	if (sublanguage_code == SUBLANG_PHP)
		point += parse_string(1, "(::<word>+)+");	// pkg::var
	else if (!in_shell_buffer)
		if (strchr("$*@%&<", character(start - 1))
			  || !parse_string(-1, "</word>(q|qq|qx|qw|m|qr|s|tr)"))
			point += parse_string(1, "(('|::)<word>+)+");	// pkg::var
	if (strchr(!true_perl ? "$" : "$*@%",
			   character(start - 1))) {	// $var, @var etc.
		set_character_color(start - 1, point, color_class perl_variable);
	} else if (true_perl && character(start - 1) == '&') {	// &funcname
		set_character_color(start - 1, point, color_class perl_function);
	} else if (true_perl && character(start - 1) == '<'
			   && character(point) == '>') {
		set_character_color(start - 1, point, color_class perl_constant);
	} else if (sublanguage_code == SUBLANG_PHP) {
		color_perl_keyword(start);
	} else if (true_perl)
		if (parse_string(1, "[ \t\n]*=>"))
			set_character_color(start, point, color_class perl_constant);
		else if (character(start - 1) == '{' && curchar() == '}')
			set_character_color(start, point, color_class perl_constant);
		else
			color_perl_keyword(start);
}

prefix_fill_one_paragraph(char *pat, char *skip, int force)
{
	char prefix[FNAMELEN], quoted[FNAMELEN];

	save_spot point;
	re_search(1, "[^\n\t ]");
	if (force || matches_at(give_begin_line(), 1, pat)) {
		save_spot mark, narrow_start, narrow_end, margin_right;
		to_begin_line();
		*prefix = 0;
		parse_string(1, skip, prefix);
		if (!*prefix) {
			fill_paragraph();
			return;
		}
		margin_right -= current_column();
		if (margin_right < 1)
			margin_right = 1;
		sprintf(quoted, "^%q", prefix);
		prefix_find_fill_region(1, skip, pat);
		// We've now narrowed to one paragraph, and know the quoting
		// prefix shared by all its lines.  So strip the prefix, fill,
		// and put it back.
		if (*quoted)
			simple_re_replace(1, quoted, "");
		fill_paragraph();
		if (*prefix)
			simple_re_replace(1, "^", prefix);
	} else
		fill_paragraph();
}

before_push()
{
	if (is_gui || opsys == OS_UNIX)
		return;
	save_screen(&push_info);
	sayput("");
	wait_for_unseen_msgs();
	term_position(0, screen_lines - 1);
	term_mode(0);
}

command compile_buffer()
{		// compile the current buffer (mode-specific)
	int res, buf = bufnum;
	char cmd[FNAMELEN * 4], oscmd[FNAMELEN * 4], full[FNAMELEN * 4], *ext;

	save_var clear_process_buffer = (clear_process_buffer & 9) != 0;
	tiled_only();
	if (use_compile_command_file_variable && look_for_filevar_pattern(
		"compile-command", "([^-\" \n\t;]([^-;\n]|-[^*])*|\"[^\n\"]+\")", NULL)) {
		grab(find_group(1, 1), find_group(1, 0), oscmd);
		if (*oscmd == '"')
			drop_quotes(oscmd);
		if (use_compile_command_file_variable != 1)
			get_strdef(cmd, "Command to compile this buffer", oscmd);
		else
			strcpy(cmd, oscmd);
		if (*cmd)
			save_var compile_buffer_cmd = cmd;
	}
	if (!compile_buffer_cmd) {
		ext = get_extension(filename);
		sprintf(cmd, "compile-%s-cmd", *ext ? (ext + 1) : "none");
		sprintf(oscmd, "%s-%s", cmd, is_unix ? "unix" :
				is_win32 ? "win32" : "other");
		if ((ext = get_str_var(find_index(oscmd)))
			  || (ext = get_str_var(find_index(cmd))))
			save_var compile_buffer_cmd = ext;
	}
	if (!compile_buffer_cmd || !*compile_buffer_cmd) {
		get_string(cmd, "Command to compile this buffer: ");
		compile_buffer_cmd = strsave(cmd);
	}
	save_var concurrent_make;
	if (pre_compile_hook)	// hook can change above vars
		if ((*pre_compile_hook)())
			return;		// hook says don't do compile
	if (concurrent_compile < 3)
		concurrent_make = concurrent_compile;
	if (concurrent_compile == 0 && (is_gui || is_win32))	// Run in the
		concurrent_make = 4;		// concur proc if any, but keep control.
	if (start_process_in_buffer_directory) {
		char dir[FNAMELEN], old[FNAMELEN];

		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
		build_filename(full, compile_buffer_cmd, filename);
		chdir(old);
		if (start_process_in_buffer_directory == 2)	// For error msgs.
			save_var start_process_in_buffer_directory = 1;
	} else
		build_filename(full, compile_buffer_cmd, filename);
	res = do_compile(full);
	save_var bufnum = buf;
	if (post_compile_hook)
		(*post_compile_hook)(res, buf);
}

int run_helper(char *prog, char *file, char *moreargs)
{
	char cmd[FNAMELEN * 2], quoted[FNAMELEN];

	quote_for_shell(quoted, file);
	sprintf(cmd, "~/.epsilon/%s %s %s >/dev/null 2>&1",
			prog, quoted, moreargs);
	if (!shell("", cmd, ""))	// Run user's version; fail silently.
		return 1;		// If we failed, look on path.
	sprintf(cmd, "%s %s %s >/dev/null 2>&1", prog, quoted, moreargs);
	if (!shell("", cmd, ""))	// Search along the path; fail silently.
		return 1;		// If we failed, try again with std copy we provide.
	get_executable_file(cmd, prog, 1);
	sprintf(cmd + strlen(cmd), " %s %s", quoted, moreargs);
	if (!shell("", cmd, ""))
		return 1;
	return 0;
}

// Write to dest the full pathname of the file prog in the same
// directory as Epsilon's executable, optionally inside quotes.
get_executable_file(char *dest, char *prog, int quoted)
{
	char path[FNAMELEN];

	get_executable_directory(path);
	convert_to_8_3_filename(path);
	strcat(path, prog);
	if (quoted)
		quote_for_shell(dest, path);
	else
		strcpy(dest, path);
}	

to_line_comment()	/* go to start of comment on this line, or eol */
{			/* return 1 if there was comment, else 0 */
	save_var narrow_end = size() - give_end_line();
	to_begin_line();
	if (!re_search(1, get_comment_start()))
		return 0;
	point = matchstart;
	return 1;
}

python_string_color(int c)
{
	char pat[100];
	int start = point - 1;

	if (character(point) == c && character(point + 1) == c) {
		// a triple-quoted string
		point += 2;
		sprintf(pat, "\\.|%c%c%c", c, c, c);
	} else
		sprintf(pat, "\\.|%c|\n", c);
	while (re_search(1, pat))
		if (character(matchstart) != '\\')
			break;
	set_character_color(start, point, color_class python_string);
}

int python_open_delimiter(int start, int end)
{
	save_var point = start;
	save_var narrow_end = size() - end;
	return re_search(1, "![[({]") && get_character_syntax_color(point) == -1
			&& default_move_level(1, "[({", "])}") == -1;
}

command python_fill_comment() on python_tab[ALT('q')]
{
	int from, to;

	iter = 0;
	if (has_arg)
		save_var margin_right = current_column() + 1;
	if (get_character_syntax_color(point) == color_class python_string
		  && get_character_color(point, &from, &to) == color_class python_string
		  && matches_at(from, 1, "'''|\"\"\"")) {
		save_var narrow_start = from, narrow_end = size() - to;
		fill_paragraph();
	} else
		fill_comment();
}

int grab_search_string(int flags, char **str)
{
	save_var abort_searching = 0;
	grab_expanding(0, size(), str, 256);
	if (size() != strlen(*str)) {		// Has nulls.
		if ((flags & SEARCH_REGEX) && !(flags & SEARCH_REGEX_HIDDEN)) {
			// If it's a regex pattern, replace literal nulls with
			// regex coding that finds them.
			spot sp = alloc_spot(1);	// Preserve point w/left-inserting spot.
			on_exit point = *sp, free_spot(sp);
		} else {
			int tmp = tmp_buf();		// Convert literal text to regex.
			on_exit buf_delete(tmp);
			buf_xfer(tmp, 0, size());	// Make a copy of buffer with nulls.
			save_var bufnum = tmp;
			point = size();
			while (re_search(-1, "[][|()+*?.$%<>^]")) {
				insert('%');		// Quote these literal characters for regex.
				point--;
			}
			flags |= SEARCH_REGEX | SEARCH_REGEX_HIDDEN;
		}
		simple_re_replace(1, "<#0>", "<#0>");	// Nulls => 4 chars "<#0>".
		grab_expanding(0, size(), str, 256);	// Get revised pattern.
	}
	return flags;
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	if (search_in_region && is_highlight_on()) {
		save_var _highlight_control;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	} else if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var _highlight_control;
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = full_key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))));
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		flags = grab_search_string(flags, &trystr);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = editwin;
	if (size()) {
		flags = grab_search_string(flags, &trystr);
		save_search_recall(flags, trystr);
	}
	if (key == abort_key || user_abort) {
		window_handle = origwin;
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			if (_last_was_regex)
				grab_search_string(flags, &_default_regex_search);
			else
				grab_search_string(flags, &_default_search);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

set_flag_modes(fp)	/* maybe set *fp from key, return 1 if we could */
int *fp;
{
	if (key == CTRL('C')) {
		if ((*fp & REPLACE) && replace_by_case == 2) {
			if (*fp & CASE_REPLACE)
				*fp &= ~(CASE_REPLACE | FOLD);
			else if (*fp & FOLD)
				*fp |= CASE_REPLACE;
			else
				*fp |= FOLD;
		} else
			*fp ^= FOLD;
		*fp |= MODFOLD;
		return 1;
	} else if (key == CTRL('W'))
		*fp ^= WORD;
	else if (key == CTRL('T')) {
		if (*fp & SEARCH_REGEX_HIDDEN)
			*fp &= ~SEARCH_REGEX_HIDDEN;
		else
			*fp ^= REGEX;
		if (*fp | REGEX) /* if entering regex, not incremental */
			*fp &= ~INCR;
	} else
		return 0;
	if (*fp & UNKNOWN)	/* setting regex or word makes type known */
		*fp &= ~(INCR | UNKNOWN);
	return 1;
}

make_search_prompt(s, flags)
char *s;
{
	sprintf(s, "%s%s%s%s%s%s%s",
		(flags & SEARCH_CONTIN) ? "Continued " :
		(flags & SEARCH_OVERWRAPPED) ? "Overwrapped " :
			(flags & SEARCH_WRAPPED) ? "Wrapped " : "",
		(flags & BAD) ? "Bad " : (flags & FAILING) ? "Failing " : "",
		(flags & REVERSE) ? "Reverse " : "",
		!(flags & MODFOLD) ? "" : (flags & CASE_REPLACE) ? "CaseRepl "
			: (flags & FOLD) ? "Folding " : "Exact ",
		(flags & REGEX) && !(flags & SEARCH_REGEX_HIDDEN) ? "R-E " : "",
		(flags & WORD) ? "Word " : "",
		(flags & INCR) ? "I-" : "");
}

save_search_recall(int flags, char *str)
{
	char *srch;
	int buf = tmp_buf();

	save_var bufnum = buf, abort_searching = 0;
	bprintf("%s%s%s",
		!(flags & MODFOLD) ? "" : (flags & FOLD) ? "Fold " : "Nofold ",
		(flags & WORD) ? "Word " : "",
		(flags & REGEX) ? "Regex " : "");
	if (character(point - 1) == ' ')
		delete(point - 1, point);
	bprintf(": %s", str);
	simple_re_replace(1, "%<NL>", "%<NL>");	// Put a % before actual <NL> strings.
	simple_re_replace(1, "\n", "<NL>");	// Convert newlines to <NL>.
	srch = malloc(size() + 1);
	grab(0, size(), srch);
	restore_vars();
	buf_delete(buf);
	save_recall(srch, "search");
	free(srch);
}

int column_color_searching(int flags, char *pat, int startcol, int endcol)
{
	while (do_color_searching(flags, pat) == 1)
		if (startcol < 0 || get_column(matchstart) == startcol)
			if (endcol < 0 || get_column(matchend) == endcol)
				return 1;
	return 0;
}

// A quick and simple function for replacing one regex string with
// another.  Returns number of replacements done.  Use string_replace
// for replacement text with #1, etc.
simple_re_replace(int flags, char *str, char *repl)
{
	int cnt = 0;
	save_spot point = (flags & RE_REVERSE) ? size() : 0;
	while (re_search(flags, str)) {
		delete(matchstart, matchend);
		stuff(repl);
		cnt++;
	}
	return cnt;
}

save_session()
{
	char file[FNAMELEN];

	if (preserve_session_once)
		save_var preserve_session = preserve_session_once - 1;
	if (!(preserve_session & (SAVE_RESTORE_SESSION | SAVE_SESSION)))
		return 0;
	if (!preserve_session_flags && preserve_session != SAVE_SESSION)
		return 0;	// Never read a session, don't write (unless flags say to).
	give_session_file(file);
	int err = do_save_session(file);
	if (err) {
		save_var no_popup_errors = 1;
		file_error(err, file, "session file write error");
		if (ask_yn("Can't save session, exit anyway?",
				   "Error writing session file, exit anyway? ",
				   "Yes", "No", 1) != 1)
			quick_abort();
	}
	return err;
}

do_setting(int i, char *tmp, int def)
{
	save_var use_default = def;
	switch (vartype(i)) {
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT:
			if (*tmp)
				set_num_var(i, exptoi(tmp));
			break;
		case TYPE_CARRAY:
			strncpy(varptr(i), tmp, ptrlen(varptr(i)) - 1);
			int j = ptrlen(varptr(i)) - strlen(tmp);
			if (j > 0)
				set_chars(varptr(i) + strlen(tmp), 0, j);
			else
				say("%s truncated: %d characters, room for %d.",
					  name_name(i), strlen(tmp), ptrlen(varptr(i)) - 1);
			break;
		case TYPE_CPTR:
			set_str_var(i, strsave(tmp)); break;
	}
}

set_default_state_file_name(char *state_file)
{
	char fname[FNAMELEN], newname[FNAMELEN];

	build_filename(fname, "%x%f", default_state_file_name);
	strcpy(newname, default_state_file_name);
	absolute(newname);
	// Force a different name if the original is read-only or the
	// suggested name differs from the usual one in the executable's
	// directory.
	if (*state_file && (strcmp(newname, fname)
						|| get_file_read_only(state_file) == 1)) {
		get_customization_directory(state_file);
		absolute(state_file);		// Strip off any trailing /.
		sprintf(state_file + strlen(state_file), "%c%s",
				path_sep, default_state_file_name);
	} else if (!*state_file)
		strcpy(state_file, default_state_file_name);
}

do_save_state(fname)		/* save state in file, aborting on error */
char *fname;
{
	int err;
	char file[FNAMELEN], backupname[FNAMELEN];

	strcpy(file, fname);
	copy_bindings();	/* fix up changed bindings */
	sayput("Writing...");

	strcpy(get_extension(file), state_extension);
	if (want_state_file_backups) {
		build_filename(backupname, state_file_backup_name, file);
		make_backup(file, backupname);
	}
	if (err = save_state(file)) {
		file_error(err, file, "write error");
		maybe_ding(bell_on_write_error);
		quick_abort();
	} else
		say("State saved on %f.", file);
}

void eval_template(char *expr, char *src, int in_buf)
{
	char quoted[300], *p, *q, *format = 0, copy[300];

	strcpy(copy, expr);
	expr = copy;
	for (p = expr; p = strchr(p, ';'); p++) {
		if (strchr(p, '"'))
			continue;
		format = p + 1;
		*p = 0;
		break;
	}
	save_var bufnum = zap(src);
	for (p = expr, q = quoted; *q = *p; p++, q++)
		if (*p == '"') {
			*q++ = '\\';
			*q = '"';
		}
	ev_header();
	bprintf("ev_template(){\n");
	bprintf("char *strval;\n");
	bprintf("  if (sizeof(%s) <= sizeof(int))\n", expr);
	char prefix[20];
	*prefix = 0;
	if (format)
		if (p = strpbrk(format, "xo"))
			sprintf(prefix, "0%c", *p);
	if (in_buf)
		bprintf("    bprintf(\"%s%%%s\", (%s));\n",
				prefix, format ? format : "d", expr);
	else
		bprintf("    say(\"%%s == %s%%%s\", \"%s\", (%s));\n",
				prefix, format ? format : "'d", quoted, expr);
	bprintf("  else {\n");
	bprintf("    strval = (char *)(%s);\n", expr);
	bprintf("    if (!strval) strval = \"<null>\";\n");
	if (in_buf)
		bprintf("    bprintf(\"%%%s\", strval);\n", format ? format : "s");
	else
		bprintf("    say(\"%%s == %%%s\", \"%s\", strval);\n",
				format ? format : "s", quoted);
	bprintf("}\n");
	bprintf("}\n");
}

int ev_numexpr()		// Replaced by functions below.
{
	return 0;
}

// Take an expression like 1+2 and evaluate it.  If it's invalid,
// return 0.  Set got_bad_number to indicate if it's valid.
int evaluate_numeric_expression(char *expr)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	save_var bufnum = zap(src);
	got_bad_number = 0;
	ev_header();
	bprintf("ev_numexpr(){ return %s; }\n", expr);
	if (!eel_compile(src, 0, "-s -b", "eval-comp"))
		return ev_numexpr();
	got_bad_number = 1;
	return 0;
}

load_binding()		/* new-style binding declaration */
{
	char name[80], junk[80];
	int *onekey, start;
	short *tab;
	int i, now_type, cnt = 1;

	point++;			/* skip initial ~ */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	now_type = (i > 0) ? name_type(i) : -1;
	if (now_type != NT_TABLE)
		init_error("%s is not a table.", name);
	tab = index_table(i);
	point++;			/* skip space before key */
	onekey = get_keycode();
	if (!onekey || onekey[0] != 2)
		init_error("Bad key.");
	start = onekey[1];
	if (parse_string(1, " \"")) {
		point++;
		onekey = get_keycode();
		if (!onekey || onekey[0] != 2)
			init_error("Bad final key.");
		cnt = onekey[1] - start + 1;
	}
	if (!parse_string(1, ": ", junk) || !parse_string(1, ".*", name))
		init_error("Bad key binding.");
	i = find_index(name);
	if (i <= 0) {		/* allow name table index too */
		i = numtoi(name);
		if (!i && *name != '0') {	/* not a name, not a number */
			init_warning("No such command %s.", name);
			return;
		}
	}
	set_range(tab, start, i, cnt);
	rebuild_menu();
}

jmp_buf jbuf;
int *keyseq;		/* sequence of key codes, in format used by macros */

load_other()
{
	char name[80], junk[20];
	int i;

	parse_string(1, "[^ :\n]+", name);
	i = find_index(name);
	if (parse_string(1, ": macro ", junk)) {
		get_keyseq();
		if (i > 0)
			if (name_type(i) == NT_MACRO || !strcmp(name, "last-kbd-macro"))
				drop_name(name);
			else {
				init_warning("%s is already defined.  Use Alt-x delete-name "
							 "to overwrite.", name);
				return;
			}
		name_macro(name, keyseq);
	} else if (parse_string(1, ": table", junk)) {
		if (i > 0) {
			if (name_type(i) == NT_TABLE)
				return;
			drop_name(name);
		}
		new_table(name);
	} else if (parse_string(1, ": command", junk)) {
		if (i <= 0 || name_type(i) != NT_COMMAND
			  && name_type(i) != NT_AUTOLOAD)
			init_warning("%s is not a currently defined command.", name);
	} else if (parse_string(1, ": subroutine", junk)) {
		if (i <= 0 || name_type(i) != NT_SUBR
			  && name_type(i) != NT_AUTOSUBR)
			init_warning("%s is not a currently defined subroutine.", name);
	} else
		init_error("%s: unrecognized syntax.", name);
}

int get_tag_window_direction(char *tag)
{
	while (get_direction() == -1 && (!IS_ASCII_KEY(key)
									 || !strchr("125znpo.", key))) {
		noteput("Show %s in window (press an arrow key, "
				"or one of .np125z) ", tag && *tag ? tag : "tag");
		getkey();
		check_abort();
	}
	say("");
	return key;
}

command pluck_tag()
{		 /* read a function name at point & go there via tags */
	char tag[TAGLEN];

	tiled_only();
	init_tags();
	iter = 0;
	grab_tag_name(tag);
	if (has_arg)
		save_var tag_to_window_key = get_tag_window_direction(tag);
	modal_goto_tag(tag);
}

command goto_tag()
{
	char tag[TAGLEN], pr[FNAMELEN], look_for[FNAMELEN];
	int to_win = has_arg && !iter;

	tiled_only();
	init_tags();
	iter = 0;
	*tag = 0;
	tags_look_for(look_for);
	if (!has_arg || to_win) {
		if (!*_last_tag)
			sprintf(pr, "Find tag: ");
		else if (tags_from_bsc)
			sprintf(pr, "Find tag [next %s]: ", look_for);
		else
			sprintf(pr, "Find tag [next %s of %d]: ",
					look_for, count_tags(look_for));
		if (*_last_tag)
			save_completion(_last_tag, tag_match);
		get_tag(tag, pr);
	}
	if (to_win)
		save_var tag_to_window_key = get_tag_window_direction(tag);
	modal_goto_tag(tag);
}

get_to_tag_match_line(char *fname, int linenum)
{
	char tdir[FNAMELEN];

	strcpy(_last_tag_file, fname);
	strcpy(tdir, bsc_file);
	*get_tail(tdir, 0) = 0;
	absolute(_last_tag_file, tdir);
	if (tag_find_file())
		return;	// Error reading the file, already reported.
	go_line(linenum);
	window_start = prev_screen_line(show_tag_line);
}

do_go_tag(s, n_fwd)
char *s;
{
	char pat[2048], tdir[FNAMELEN], *file, *p;
	int pos, oldbuf = bufnum, i;

	i = get_mode_based_index("tag_to_next_%s");
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	if (*s) {
		if (p = index(s, '\t')) {
			strcpy(_last_tag_file, p + 1);
			*p = 0;		/* remove file name from s */
			if (p = index(_last_tag_file, '\t'))
				*p = 0;
		} else
			*_last_tag_file = 0;
		strcpy(_last_tag, s);
		if (!*_last_tag)	/* no last tag */
			return 0;
		if (*_last_tag_file && p && !strchr(p + 1, '\t'))
			sprintf(pat, "^%q\t%q\t%q", _last_tag, _last_tag_file, p + 1);
		else if (*_last_tag_file)
			sprintf(pat, "^%q\t%q\t",
				_last_tag, _last_tag_file);
		else
			sprintf(pat, "^%q\t", _last_tag);
		point = 0;
		if (!re_search(1, pat))		/* look for 1st */
			return 1;
	} else {		/* find next matching tag */
		pos = point;
		sprintf(pat, "^%q\t", _last_tag);
		switch (name_type(i)) {		// Call mode-specific version.
			case NT_SUBR: case NT_COMMAND:
			case NT_AUTOSUBR: case NT_AUTOLOAD:
				if ((*(int (*)())i)(pat, n_fwd))
					return 1;
				break;
			default:
				if (!tag_search_by_offset(pat, n_fwd))
					return 1;
		}
	}
	grab(give_begin_line(), give_end_line(), pat);	/* tag, file, pos */
	file = index(pat, '\t');
	*file++ = 0;

	p = index(file, '\t');		/* find start of pos */
	pos = strtoi(p + 1, 10);
	*p = 0;
	p = index(p + 1, '\t');		/* if p, p+1 has line's text */
	strcpy(_last_tag_file, file);
	put_directory(tdir);
	bufnum = oldbuf;
	absolute(_last_tag_file, tdir);
	tag_find_file();
	if (p)
		sprintf(tdir, "%q[^a-zA-Z0-9_]", _last_tag);
	else			// Might not include class info.
		tag_make_loose_pattern(_last_tag, tdir);
	point = pos;
	if (!parse_string(1, tdir, NULL))
		if (!p || !prox_line_search(p + 1))	/* search by text */
			return 2;
	to_begin_line();
	window_start = prev_screen_line(show_tag_line);
	return 0;
}

skip_c_braces()		/* skip over c function definition */
{
	if (parse_string(-1, "<dquote>[A-Za-z0-9_]*<dquote>[ \t\n]*")
		  && !at_c_comment(point, 0) && !parse_string(-1, "color_scheme.*")) {
		point++;	// if C++'s extern "C" { or similar, don't skip
		return;
	}
	skip_c_delims('{', '}', 0);
}

tag_c_skip_before_body(int show)
{
	if (parse_string(1, "#.*(\\\n.*)*")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(try|const|volatile|NOTHROW)![^_A-Za-z0-9]")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(AVAILABLE|DEPRECATED_IN)_MAC_OS_X_VERSION_<word>+")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "throws[ \t]")) {	// for java
		point = matchend;
		skip_identifier_list();
		return 1;
	}
	if (parse_string(1, "(throw|__attribute__)[ \t]*!%(")) {
		point = matchend;	// throw for C++, attribute for gcc
		if (skip_c_delims('(', ')', 1)) {
			if (show)
				buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
							  filename, lines_between(0, point, 0) + 1);
			return 0;
		}
		return 1;
	}
	return 0;
}

// Remember we're now in this class (or namespace or similar).
tag_add_class(char *class, char *sep)
{
	if (!old_tagging_class)
		old_tagging_class = (char **) malloc(sizeof(*old_tagging_class) * 20);
	char **p = old_tagging_class;
	while (*p)	// Add ptr to end of class list to end of this list.
		p++;
	*p = c_tagging_class + strlen(c_tagging_class);
	if (*c_tagging_class)
		strcat(c_tagging_class, sep);
	strcat(c_tagging_class, class);
}

// Remember we're now in this class (or namespace or similar).
tag_c_add_class(char *class)
{
	tag_add_class(class, "::");
}

// Leave a class's scope.
tag_remove_class()
{
	char **p;
	if (old_tagging_class && *old_tagging_class) {
		for (p = old_tagging_class; *p; p++)
			;
		**--p = 0;		// Chop off last class.
		*p = 0;			// Forget chop-off point.
	} else
		*c_tagging_class = 0;
}

add_class_tag(char *class, char *sep, char *func, int pos)
{
	char fulltag[FNAMELEN * 2];
	if (class && *class && *sep)
		sprintf(fulltag, "%s%s%s", class, sep, func);
	else
		strcpy(fulltag, func);
	add_tag(fulltag, pos);
}

// Return a regex pattern that matches an identifier, including :: or .
// as apppropriate.
char *tag_c_ident_pattern()
{
	return (c_extra_keywords == CSHARP_KEYWORDS) ?
			"<word|.>+" : "<word>+(::<word>+)?";
}

tag_c_define(keyw)	// found a preproc line, maybe enter #define
char *keyw;		// either way, skip past line
{
	char func[TAGLEN], pat[FNAMELEN];
	int start = point;

	sprintf(pat, "[ \t]+%s</word>[ \t]*($|/[/*])", tag_c_preprocessor_skip_pat);
	if (!strcmp(keyw, "if") && parse_string(1, pat))
		if (!locate_by_ifdef(1))
			point = start;
	if (!strcmp(keyw, "asm"))	// Found #asm, look for matching #endasm.
		return re_search(1, "^[ \t]*#[^#\n]*endasm");
	if (!strcmp(keyw, "define")) {
		point += parse_string(1, "[ \t]*");
		start = point;
		if (parse_string(1, "[a-zA-Z0-9_]+", func) && !isdigit(*func)
			  && (tag_which_items & 4))
			add_tag(func, start);
	}
	if (parse_string(1, ".*/<*>"))
		search(1, "*/");
	do {
		if (!nl_forward())
			break;
	} while (character(point - 2) == '\\');
	return 1;
}

// Found start of an Objective-C block; parse it.
int tag_objc_block()
{
	char class[FNAMELEN], func[FNAMELEN];
	int i, instance_vars = 0, end;

	point += parse_string(1, "[ \t]*");
	if (!parse_string(1, "<word>+", class))
		return 1;
	if (tag_which_items & 16)
		add_tag(class, matchstart);
	tag_c_add_class(class);
	c_fwd_space();
	if (curchar() == '(')	// Skip category.
		c_skip_level('(', ')');
	c_fwd_space();
	point += parse_string(1, ":[ \t]*<word>+");		// Skip superclass.
	c_fwd_space();
	if (curchar() == '<')	// Skip protocol list.
		c_skip_level('<', '>');
	c_fwd_space();
	if (curchar() == '{') {	// Include instance variables.
		instance_vars++;	// Any vars before next } are instance vars.
		point++;
	}
	tag_modifier_kwd = 0;
	while (re_search(1, "(^@end|^[-+][ \t]*|[{};<]|/<*|/>|[a-zA-Z0-9_]+)")
		   && character(matchstart) != '@') {
		end = point;
		point = matchstart;
		switch (curchar()) {
			case '-':
			case '+':
				if (i = find_objc_method(func))
					if ((i == 1 || (tag_declarations & 1) || instance_vars)
						  && (tag_which_items & 1))
						add_class_tag(c_tagging_class, "::", func,
									  start_of_function);
				re_search(1, "![@;{]");
				break;
			case '{':
				skip_c_braces();
				break;
			case '<':
				skip_cpp_template();
				break;
			case '}':
				if (instance_vars)
					instance_vars = 0;
				else
					tag_remove_class();
				point++;
				break;
			case ';':
				tag_modifier_kwd = 0;	// assume next var
				point++;		// is being defined
				break;
			case '/':
				if (character(point + 1) == '*') {
					point += 2;
					search(1, "*/");
				} else
					nl_forward();	/* c++ comment */
				break;
			default:
				if (!good_c_tag())
					point = end;
				break;
		}
	}
	tag_remove_class();
	return 1;
}

// Found @class, parse class names that follow.
tag_objc_class_list()
{
	char class[FNAMELEN];

	for (;;) {
		c_fwd_space();
		if (!parse_string(1, "<word>+", class))
			break;
		if (tag_which_items & 16)
			add_tag(class, give_begin_line());
		c_fwd_space();
		point += parse_string(1, ",");
	}
	return 1;
}

command tex_fill_paragraph()
{
	if (get_character_syntax_color(point) == color_class tex_comment
		  || get_character_syntax_color(point - 1) == color_class tex_comment)
		fill_comment();
	else
		fill_paragraph();
}

is_vbasic_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	save_var case_fold = 1;

	if (sublanguage_code == SUBLANG_VBNET) {
		if (strstr("|#const|#else|#elseif|#externalsource|#externalsource|#if"
				   "|#region|#region|#then|addhandler|addressof|alias|and|andalso"
				   "|ansi|as|assembly|auto|boolean|byref|byte|byval|call|case"
				   "|catch|cbool|cbyte|cchar|cdate|cdbl|cdec|char|cint|class|clng"
				   "|cobj|const|cshort|csng|cstr|ctype|date|decimal|declare"
				   "|default|delegate|dim|directcast|do|double|each|else|elseif"
				   "|end|enum|erase|error|event|exit|false|finally|for|friend"
				   "|function|get|gettype|goto|handles|if|implements|imports|in"
				   "|inherits|integer|interface|is|let|lib|like|long|loop|me|mod"
				   "|module|mustinherit|mustoverride|mybase|myclass|namespace|new"
				   "|next|not|nothing|notinheritable|notoverridable|object|on"
				   "|option|optional|or|orelse|overloads|overridable|overrides"
				   "|paramarray|preserve|private|property|protected|public"
				   "|raiseevent|readonly|redim|rem|removehandler|resume|return"
				   "|select|set|shadows|short|single|static|step|stop"
				   "|string|structure|sub|synclock|then|throw|to|true|try|typeof"
				   "|unicode|until|variant|when|while|with|withevents|writeonly"
				   "|xor|widening|narrowing|operator|datetime|int16|int32|int64"
				   "|sbyte|uinteger|uint32|ulong|uint64|ushort|uint16", p))
			return 1;
		return 0;
	}
	if (strstr("|abs|array|asc|ascb|ascw|atn|cbool|cbyte|ccur|cdate|cdbl"
			   "|cdec|choose|chr|chr$|chrb|chrb$|chrw|cint|clng|command"
			   "|command$|cos|createobject|csng|cstr|curdir|curdir$|cvar"
			   "|cvdate|cverr|date|date$|dateadd|datediff|datepart"
			   "|dateserial|datevalue|day|ddb|dir|dir$|doevents|environ"
			   "|environ$|eof|error|error$|exp|fileattr|filedatetime|filelen"
			   "|fix|format|format$|freefile|fv|getallsettings|getattr"
			   "|getobject|getsetting|hex|hex$|hour|iif|imestatus|input"
			   "|input$|inputb|inputb$|inputbox|instr|instrb|int|ipmt|irr"
			   "|isarray|isdate|isempty|iserror|ismissing|isnull|isnumeric"
			   "|isobject|lbound|lcase|lcase$|left|left$|leftb|leftb$|len"
			   "|lenb|loc|lof|log|long|ltrim|ltrim$|mid|mid$|midb|midb$|minute"
			   "|mirr|month|msgbox|now|nper|npv|oct|oct$|partition|pmt|ppmt"
			   "|pv|qbcolor|rate|rgb|right|right$|rightb|rightb$|rnd|rtrim"
			   "|rtrim$|second|seek|sgn|shell|sin|sln|space|space$|spc|sqr"
			   "|str|str$|strcomp|strconv|string|string$|switch|syd|tab|tan"
			   "|time|time$|timer|timeserial|timevalue|trim|trim$|typename"
			   "|ubound|ucase|ucase$|val|vartype|weekday|year|", p))
			return 2;		// Function
	if (strstr("|#const|#else|#elseif|#end|if|#if|access|alias|any"
			   "|appactivate|append|as|base|beep|binary|byref|byval|call"
			   "|case|cdecl|chdir|chdrive|close|compare|const|date|declare"
			   "|defbool|defbyte|defcur|defdate|defdbl|defdec|defint|deflng"
			   "|defobj|defsng|defstr|defvar|deletesetting|dim|do|each|else"
			   "|elseif|empty|end|endif|erase|error|exit|explicit|false"
			   "|filecopy|for|function|get|global|gosub|goto|if|in|input|is"
			   "|kill|len|let|lib|line|load|local|lock|loop|lset|me|mid|midb"
			   "|mkdir|module|name|new|next|nothing|null|on|open|option"
			   "|optional|output|paramarray|preserve|print|private|property"
			   "|public|put|random|randomize|read|redim|rem|reset|resume"
			   "|return|rmdir|rset|savesetting|seek|select|sendkeys|set"
			   "|setattr|shared|static|step|stop|string|sub|text|then|time"
			   "|to|true|type|typeof|unload|unlock|until|wend|while|width"
			   "|with|withevents|write|", p))
			return 1;	// Statement or keyword.
	if (sublanguage_code == SUBLANG_VBSCRIPT &&
		  strstr("|class|eval|execute|scriptengine|", p))
		return 3;		// VBScript-specific
	return 0;
}

int vbasic_indent_more()
{
	char kwd[FNAMELEN], pat[FNAMELEN];

	if (vbasic_indent_subroutines
		  && parse_string(1, "[ \t]*(%<[^\n>]+%>)?[ \t]*"
						  "(</word>(Private|Public|Default|Friend"
						  "|Static|Protected|Shadows|MustInherit"
						  "|NotInheritable|Overloads|Overrides|Overridable"
						  "|NotOverridable|MustOverride|Shared|ReadOnly"
						  "|WriteOnly|Widening|Narrowing)</word>[ \t]*)*"
						  "</word>(Sub|Function|Property|Class|Operator"
						  "|Module|Interface|Namespace|Structure)</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(If)</word>.*"
					 "</word>Then</word>[ \t]*[^ \t\n':]"))
		return 0;		// Old one-line form of If statement.
	if (parse_string(1, "[ \t]*</word>((private|public)[ \t]+)?"
					 "(Type|Enum|Else|With|Select|Case|Try|Catch|Finally"
					 "|SyncLock|If|Elseif)</word>")) {
		grab(find_group(3, 1), find_group(3, 0), kwd);
		sprintf(pat, ".*:[ \t]*</word>End[ \t]+%s</word>", kwd);
		if (!parse_string(1, pat))	// Multiple statements: ignore block.
			return 1;
	}
	if (parse_string(1, "[ \t]*</word>(For)</word>")
		  && !parse_string(1, ".*:[ \t]*</word>Next</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(Do)</word>")
		  && !parse_string(1, ".*:[ \t]*</word>Loop</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(While)</word>")
		  && !parse_string(1, ".*:[ \t]*</word>Wend</word>"))
		return 1;
	if (parse_string(1, "[ \t]*(Get|Set)[ \t]*(%([ \t]*%))?[ \t]*[\n(':]"))
		return 1;	// Get/set keyword with nothing after it, or (, starts block.
	return 0;
}

vbasic_func_name_finder()
{
	int from, to, kwdstart;

	save_var case_fold = 1, point;
	to_end_line();
	while (re_search(-1, "(^|%>)[ \t]*((end|public|private|static|default|shared"
					 "|widening|protected|friend|shadows|readonly|writeonly"
					 "|overloads|overrides|overridable|notoverridable"
					 "|mustoverride|narrowing)[ \t]+)*"
					 "(sub|function|type|property([ \t]+(let|get|set))?"
					 "|operator)</word>")) {
		if (tolower(character(find_group(2, 1))) == 'e')
			break;
		kwdstart = find_group(4, 1);
		if (!must_find_func_name && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		if (tolower(character(kwdstart)) == 'o') {
			if (!matches_at(kwdstart, 1, "(operator((<wspace>*"
							"([-+*/\\&^=<>]|[<>][<>=]))|"
							"(<wspace>+(IsFalse|IsTrue|Not|Ctype|And|Like|Mod|Or"
							"|Xor))))"))
				continue;
		} else if (!matches_at(matchstart, 1, "[ \t]+([a-z_0-9]+)"))
			continue;
		from = find_group(1, 1);
		to = find_group(1, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

tag_mode_vbasic()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (re_search(1, "(^|%>)[ \t]*((public|private|static|default|shared"
					 "|widening|protected|friend|shadows|readonly|writeonly"
					 "|overloads|overrides|overridable|notoverridable"
					 "|mustoverride|narrowing)[ \t]+)*"
					 "(((sub|function|type|property([ \t]+(let|get|set))?"
					 ")[ \t]+([a-z_0-9]+))"
					 "|(operator((<wspace>*([-+*/\\&^=<>]|[<>][<>=]))|"
					 "(<wspace>+(IsFalse|IsTrue|Not|Ctype|And|Like|Mod|Or"
					 "|Xor)))))")) {
		if (tolower(character(find_group(4, 1))) == 'o')
			grab(start = find_group(10, 1), find_group(10, 0), func);
		else
			grab(start = find_group(9, 1), find_group(9, 0), func);
		add_tag(func, start);
	}
}

move_by_pages(dir)	// page up or down by one page
{
	int col, wide = window_width - 1, n;

	if (prev_cmd != SAMECOL)
		_preserve_column = virtual_space ? virtual_column()
					: current_column();
	if (window_overlap < 0)
		n = ((100 - (-window_overlap)) * text_height()) / 100;
	else
		n = text_height() - window_overlap;
	if (n < 1)
		n = 1;
	n *= dir;		// move by n lines
	window_start = move_by_screen_lines(window_start, n);
	if (paging_centers_window)	// go halfway down window
		point = move_by_screen_lines(window_start, text_height() / 2);
	else				// go one screenful from orig place
		point = move_by_screen_lines(point, n);
	if (is_highlight_on())
		save_var goal_column = -1;
	col = (goal_column >= 0) ? goal_column : _preserve_column;
	if (virtual_space)
		to_virtual_column(col);
	else {
		if (display_column < 0)	/* move to same screen column */
			col = (col % wide) + (current_column() / wide) * wide;
		move_to_column(col);
	}
}

/* Convert current key to direction code or -1. */
int get_direction()
{
	return get_direction_from_key(key);
}

/* Convert key to direction code or -1. */
int get_direction_from_key(int k)
{
	if (IS_TRUE_KEY(k))
		k = NUMPLAIN(generic_key(k));
	switch (k) {
		case KEYUP: case CTRL('P'): case NUMDIGIT(8): case '8':
			return BTOP;
		case KEYDOWN: case CTRL('N'): case NUMDIGIT(2): case '2':
			return BBOTTOM;
		case KEYLEFT: case CTRL('B'): case NUMDIGIT(4): case '4':
			return BLEFT;
		case KEYRIGHT: case CTRL('F'): case NUMDIGIT(6): case '6':
			return BRIGHT;
	}
	return -1;
}

// Return handle of window based on direction code from this window.
int window_from_direction(int dir, int win)
{
	int edge, row, col, max_cols = screen_cols;

	save_var window_handle = win;
	edge = get_wattrib(window_number, dir);
	get_window_pos(point, &row, &col);	/* get pt's window coords */
	row += window_top;		/* convert to screen coords */
	col += window_left;
	switch (dir) {
		case BTOP:	row = edge - 1; break;
		case BBOTTOM:	row = edge + 1; break;
		case BLEFT:	col = edge - 1; break;
		case BRIGHT:	col = edge + 1; break;
	}				/* now adjust to screen bounds */
	if (is_gui && col == screen_cols && (dir == BTOP || dir == BBOTTOM))
		col--;			// Cursor can be after all characters.
	row = (row + screen_lines - 1) % (screen_lines - 1);
	col = (col + max_cols) % max_cols;
	return window_at_coords(row, col);
}

command move_to_window()
{
	int dir;

	tiled_only();
	while ((dir = get_direction()) == -1) {
		say("Move to window (press an arrow key)");
		getkey();
		check_abort();
	}
	window_handle = window_from_direction(dir, window_handle);
	say("");
	switching_to_buffer();
	check_dates(0);
}

command cd()
{
	char resp[FNAMELEN], dir[FNAMELEN];

	save_var recall_id = "changedir";
	if (get_buffer_directory(dir))
		save_completion(dir, 0);
	getcd(dir);
	if (*dir && is_path_separator(dir[strlen(dir) - 1]) != 1)
		sprintf(dir + strlen(dir), "%c", path_sep);
	save_var show_mouse_choices = 1, insert_default_response = 3;
	save_var completion_lister = list_file_matches;
	save_var abort_file_matching = ABORT_JUMP;
	prompt_comp_read(resp, "Change to directory", f_match,
					 COMP_FILE | FOLDFILE | FM_ONLY_DIRS, dir);
	if (*resp && chdir(resp)) {
		if (errno == 2)
			error("%s: no such directory", resp);
		else
			file_error(errno, resp, "unknown error");
	} else {
		getcd(dir);
		say("Current directory: %f", dir);
	}
}

command indent_region()
{
	int temp;
	int (*func)() = (int (*)()) (mode_keys['\t'] ? mode_keys['\t']
		: root_keys['\t']);
	int progress, cnt = 0;

	fix_region();
	highlight_off();
	if (nl_reverse())
		point++;
	temp = point;
	progress = (mark - point > 4000);
	this_cmd = CMD_INDENT_REG;
	while (point < mark) {
		(*func)();
		if (curchar() == '\n')
			delete_horizontal_space();
		nl_forward();
		if (!(cnt % 25) && progress && temp - mark != 0)
			note("Indenting region: %d%%...",
				 100 * (point - temp ) / (mark - temp));
		cnt++;
		check_abort();
	}
	if (progress)
		say("");
	mark = temp;
	this_cmd = iter = 0;
}

command transpose_characters()
{
	int p = point;
	char c, d;

	if (p == size() || curchar() == '\n')
		p--;		/* switch last two at end of line */
	else if (p == 0 || character(point - 1) == '\n')
		p++;
	c = character(p);	/* store away current */
	d = character(p - 1);
	if (p > 0 && p < size() && c != '\n' && d != '\n') {
		replace(p, d);
		replace(p - 1, c);
	}
	iter = 0;
}

command unbind_key()
{
	char keyname[300];
	short *binding, val = -1;

	binding = ask_key("Unbind key: ", keyname, 1);
	sayput("Make \"%s\" Unbound, or make it Self-insert [U, S]? ", keyname);
	while (!user_abort && val == -1) {
		getkey();
		switch (toupper(key)) {
			case 'U':	val = 0; break;
			case 'S':	val = normal_character; break;
		}
	}
	say("");
	check_abort();
	if (val != -1) {
		*binding = val;
		save_binding(binding);
	}
	rebuild_menu();
}

dired_try_getdir(fullname, showerr)	// grab directory, add / or \ to end
char *fullname;		// if we can't find, set fullname "" if !showerr
{			// return nonzero if ok
	int i;
	char *p, path_char;

	*fullname = 0;
	save_var point;
	to_end_line();
	if (!re_search(-1, DIRLINEPAT))	// find prev "Directory of "
		if (showerr)		// bad dired buf
			error("Can't determine directory of files");
		else
			return 0;
	point = matchstart;
	re_search(1, " *");		// skip initial spaces
	parse_string(1, FILECHARS, fullname);
	path_char = is_remote_file(fullname) ? '/' : path_sep;
	i = strlen(fullname);
	if (is_remote_file(fullname))
		while (i > 0 && !is_path_separator(fullname[i - 1]))
			fullname[--i] = 0;
	p = fullname + i;
	if (i > 0 && !is_path_separator(fullname[i - 1])
		  && (strcmp(fullname, "~") || !is_unix))
		p[0] = path_char, p[1] = 0; // add final / or \ if none
	return 1;
}

char *(*give_grep_finder)()(inbufs, filepat, ps)
char *filepat, **ps;	// put 1st match in ps & return finder for next matches
{
	char *(*finder)();

	finder = inbufs ? b_match : do_file_match;
	*ps = (*finder)(filepat, STARTMATCH | EXACTONLY
					| (inbufs ? ((preserve_filename_case & 4) ? FM_FOLD : 0)
							  : FM_NO_DIRS));
	return finder;
}

int extract_rectangle_colors(copybuf, remove, copy_colors)
{
	int oldpoint, t, tmp = 0;
	short start_col, stop_col;	/* columns to operate on */

	if (remove)
		check_modify(bufnum);
	start_col = virtual_column();
	t = point;
	point = mark;
	if (mark_to_column >= 0)
		stop_col = mark_to_column;
	else
		stop_col = current_column();
	point = t;
	if (point > mark)
		t = point, point = mark, mark = t;
	if (start_col == stop_col)
		return 0;
	else if (start_col > stop_col)
		t = start_col, start_col = stop_col, stop_col = t;
	oldpoint = point;
	if (!remove && _buf_readonly) {
		tmp = tmp_buf();
		t = give_begin_line();
		point = mark;
		buf_xfer_maybe_colors(tmp, t, give_end_line(), copy_colors);
		save_var bufnum = tmp;
		exchange_point_and_mark();
	}
	for (;;) {			/* for each line */
		force_to_column(start_col);
		t = point;
		force_to_column(stop_col);
		if (copybuf)
			buf_xfer_maybe_colors(copybuf, t, point, copy_colors);
		if (remove)
			delete(t, point);
		if (remove > 1)
			insert_to_column(start_col, stop_col);
		if (point >= mark || !nl_forward())
			break;
		if (copybuf)
			buf_xfer_maybe_colors(copybuf, point - 1, point, copy_colors);
	}
	mark = point;		/* lower right of rectangle */
	point = oldpoint;	/* upper left */
	force_to_column(start_col);
	if (tmp) {
		restore_vars();
		buf_delete(tmp);
	}
	return stop_col - start_col;
}

file_convert_write(char *fname, int trans, struct file_info *finfo)
{
	int tmp, err, mod, code;
	char *(*func)();

	code = file_write_newfile ? FILE_IO_NEWFILE : FILE_IO_ATSTART;
	if (is_remote_file(fname))
		return save_remote_file(fname, trans, finfo);
	if (!file_io_converter)
		return new_file_write(fname, trans, finfo, code, -1);
	if (!file_io_converter(FILE_CONVERT_WRITE | FILE_CONVERT_ASK))
		error("Canceled.");
	func = file_io_converter;
	tmp = tmp_buf();
	buf_xfer(tmp, 0, size());
	copy_buffer_variables(tmp, bufnum);
	save_var bufnum = tmp;
	translation_type = trans;	// func might modify this
	func(FILE_CONVERT_WRITE);
	err = new_file_write(fname, translation_type, finfo, code, -1);
	mod = modified;
	restore_vars();
	modified = mod;
	return err;
}

format_date(msg, year, month, day, hour, minute, second)
char *msg;
{
	char ampm = 'a', args[24][30], *vals[24], *s, *t;
	char *months = "January February March April May June July August "
				   "September October November December ";
	char *weekdays = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday ";
	int i, dow;

	sprintf(args[0], "%d", hour);
	sprintf(args[1], "%02d", hour);
	if (hour >= 12)
		ampm = 'p', hour -= 12;
	sprintf(args[2], "%d", hour ? hour : 12);
	sprintf(args[3], "%02d", hour ? hour : 12);
	sprintf(args[4], "%d", minute);
	sprintf(args[5], "%02d", minute);
	sprintf(args[6], "%d", second);
	sprintf(args[7], "%02d", second);
	sprintf(args[8], "%c", ampm);
	sprintf(args[9], "%c", toupper(ampm));
	sprintf(args[10], "%d", day);
	sprintf(args[11], "%02d", day);
	sprintf(args[12], "%d", month);
	sprintf(args[13], "%02d", month);
	sprintf(args[14], "%d", year);
	sprintf(args[15], "%02d", year % 100);
	for (i = month - 1, s = months; i > 0 && s; i--)
		s = strchr(s, ' ') + 1;
	if (month >= 1 && month <= 12 && s && (t = strchr(s, ' ')))
		sprintf(args[16], "%.*s", t - s, s);
	else
		sprintf(args[16], "???");
	strncpy(args[18], args[16], 3);
	for (i = 0; i < strlen(args[16]) + 1; i++)
		args[17][i] = toupper(args[16][i]);
	for (i = 0; i < strlen(args[18]) + 1; i++)
		args[19][i] = toupper(args[18][i]);
	dow = day_of_week(day, month, year);
	for (i = dow, s = weekdays; i > 0 && s; i--)
		s = strchr(s, ' ') + 1;
	if (dow >= 0 && dow <= 6 && s && (t = strchr(s, ' ')))
		sprintf(args[20], "%.*s", t - s, s);
	else
		sprintf(args[20], "???");
	strncpy(args[22], args[20], 3);
	for (i = 0; i < strlen(args[20]) + 1; i++)
		args[21][i] = toupper(args[20][i]);
	for (i = 0; i < strlen(args[22]) + 1; i++)
		args[23][i] = toupper(args[22][i]);
	for (i = 0; i < 24; i++)
		vals[i] = args[i];
	if (expand_string_template(msg, date_format, "iIhHnNeEaAdDmMyYfFtTwWsS", vals))
		error("Invalid date-format variable.");
}

format_date(msg, year, month, day, hour, minute, second)
char *msg;
{
	char ampm = 'a', args[24][30], *vals[24], *s, *t;
	char *months = "January February March April May June July August "
				   "September October November December ";
	char *weekdays = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday ";
	int i, dow;

	sprintf(args[0], "%d", hour);
	sprintf(args[1], "%02d", hour);
	if (hour >= 12)
		ampm = 'p', hour -= 12;
	sprintf(args[2], "%d", hour ? hour : 12);
	sprintf(args[3], "%02d", hour ? hour : 12);
	sprintf(args[4], "%d", minute);
	sprintf(args[5], "%02d", minute);
	sprintf(args[6], "%d", second);
	sprintf(args[7], "%02d", second);
	sprintf(args[8], "%c", ampm);
	sprintf(args[9], "%c", toupper(ampm));
	sprintf(args[10], "%d", day);
	sprintf(args[11], "%02d", day);
	sprintf(args[12], "%d", month);
	sprintf(args[13], "%02d", month);
	sprintf(args[14], "%d", year);
	sprintf(args[15], "%02d", year % 100);
	for (i = month - 1, s = months; i > 0 && s; i--)
		s = strchr(s, ' ') + 1;
	if (month >= 1 && month <= 12 && s && (t = strchr(s, ' ')))
		sprintf(args[16], "%.*s", t - s, s);
	else
		sprintf(args[16], "???");
	strncpy(args[18], args[16], 3);
	for (i = 0; i < strlen(args[16]) + 1; i++)
		args[17][i] = toupper(args[16][i]);
	for (i = 0; i < strlen(args[18]) + 1; i++)
		args[19][i] = toupper(args[18][i]);
	dow = day_of_week(day, month, year);
	for (i = dow, s = weekdays; i > 0 && s; i--)
		s = strchr(s, ' ') + 1;
	if (dow >= 0 && dow <= 6 && s && (t = strchr(s, ' ')))
		sprintf(args[20], "%.*s", t - s, s);
	else
		sprintf(args[20], "???");
	strncpy(args[22], args[20], 3);
	for (i = 0; i < strlen(args[20]) + 1; i++)
		args[21][i] = toupper(args[20][i]);
	for (i = 0; i < strlen(args[22]) + 1; i++)
		args[23][i] = toupper(args[22][i]);
	for (i = 0; i < 24; i++)
		vals[i] = args[i];
	if (expand_string_template(msg, date_format, "iIhHnNeEaAdDmMyYfFtTwWsS", vals))
		error("Invalid date-format variable.");
}

char *f_only_match(char *s, int flags);

get_file(res, pr, def)
char *res, *pr, *def;
{
	char rel[FNAMELEN], other[FNAMELEN];

//	relative(def, rel);
	get_buffer_filename(other);
	if (!inp_get_other_name || !*inp_get_other_name)
		save_var inp_get_other_name = other;
	strcpy(rel, def);
	if (!*rel) {
		get_buffer_directory(rel);
		save_var insert_default_response = 3;
	}
	if (*rel && is_directory(rel))
		save_var insert_default_response = 3;
	save_var completion_lister = list_file_matches;
	save_var abort_file_matching = ABORT_JUMP;
	save_var resize_menu_list = 1;
	if (only_file_extensions)  {
		prompt_comp_read(res, pr, f_only_match, COMP_FILE | FOLDFILE, rel);
	} else {
		save_var list_finder = do_file_match;
		prompt_comp_read(res, pr, f_match, COMP_FILE | FOLDFILE, rel);
	}
	if (*res && !wild_pattern(res))
		absolute(res);
}

go_tag_helper(s, n_fwd)
char *s;
{
	char msg[FNAMELEN];
	short err;

	if (!*s && !*_last_tag)
		error("No current tag.");
	if (tags_from_bsc)
		return go_bsc_tag(s);
	err = do_go_tag(s, n_fwd);
	if (err != 2)
		return err;
	if (narrow_start || narrow_end) {
		say("Tag not found in narrowed buffer, removing narrowing.");
		narrow_start = narrow_end = 0;
		make_mode();
		err = do_go_tag(_last_tag, n_fwd);
		if (err != 2)
			return err;
	}
	if (find_tagger(0)) {
		sprintf(msg, "Tag %s has moved, retag file %s? ",
				_last_tag, _last_tag_file);
		if (tag_ask_before_retagging)
			if (!ask_yn("Retag file", msg, "Yes", "No", 1))
				return 0;
		do_retag_file(_last_tag_file);		// Tag has moved, so retag.
		do_save_tags();
	}
	return do_go_tag(_last_tag, n_fwd);
}

html_in_quoted_attribute()
{
	save_var point;
	if (!re_search(-1, "['\"]"))
		return 0;
	if (!parse_string(-1, "%<" HTML_NAME "([ \t\n]+" HTML_ATTRIBUTE ")*"
					  "[ \t\n]+" HTML_NAME HTML_EQUALS))
		return 0;
	restore_vars();
	point = matchend + 1;
	return 1;
}

#define MAX_HELP_DESCR	40

is_help_exe()
{
	char fname[FNAMELEN], find_cmd;

	save_var point;
	save_var narrow_end = size() - give_end_line();
	if (!search(1, ".exe"))
		return 0;
	move_to_column(MAX_HELP_DESCR);
	grab(point, size(), fname);
	convert_to_8_3_filename(fname);
	delete(point, size());		// Is it Find or Keyword Search?
	find_cmd = character(give_begin_line()) == 'F';
	bprintf("!%s!%s /%c !", fname, fname, find_cmd ? 's' : 'k');
	return 1;
}

is_python_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	save_var case_fold = 0;
	if (strstr("|and|del|for|is|raise|assert|elif|from|lambda|return|"
			   "break|else|global|not|try|class|except|if|or|while|"
			   "continue|exec|import|pass|def|finally|in|print|yield|", p))
		return 1;
	return 0;
}

mouse_move_scroll_box()		// grabbed box in center of scroll bar
{
	int i, pix, max, offset, last = -1;
	char *p = _display_characters + SCROLLCHARSTART + 2, old = *p;

	save_var mouse_display = 0, mouse_auto_on = 0;
	*p = p[1];
	pix = y_pixels_per_char();
	max = (MAX(text_height(), 5) - 4) * pix;
	i = mouse_pixel_y - (w_top + 2) * pix;
	offset = point - size() * i / max;
	while (mouse_pressed() && !user_abort) {
		if (char_avail())
			getkey();
		i = mouse_pixel_y - (w_top + 2) * pix;
		point = size() * i / max + offset;
		if (point > 0 && goal_column >= 0)
			move_to_column(goal_column);
		window_start = prev_screen_line(text_height() / 2);
		if (point != last)
			refresh();
		last = point;
	}
	*p = old;
}

ok_grep_file_match(char *s, char *abs)
{
	char pat[FNAMELEN];
	struct file_info info;

	if (!use_grep_ignore_file_variables)
		return 1;
	save_var case_fold = (FOLDFILE != 0);
	sprintf(pat, "|%s|", get_extension(s));
	if (strstr(grep_ignore_file_extensions, pat))
		return 0;
	if (*grep_ignore_file_pattern
		  && string_matches_pattern(abs, grep_ignore_file_pattern))
		return 0;
	if (*grep_ignore_file_basename
		  && string_matches_pattern(get_tail(abs, 1), grep_ignore_file_basename))
		return 0;
	if (grep_ignore_file_types && check_file(abs, &info)) {
		if ((grep_ignore_file_types & 1) && info.check_type == CHECK_DEVICE)
			return 0;
		if ((grep_ignore_file_types & 2) && info.check_type == CHECK_PIPE)
			return 0;
		if ((grep_ignore_file_types & 4) && info.check_type == CHECK_OTHER)
			return 0;
	}
	return 1;
}

int perl_skip_level(int dir, char left_d, char right_d)
{
	char pattern[40], ch;
	int level = 0, pos;

	sprintf(pattern, "[%c%c]", right_d, left_d);
	do {
		if (!re_search(dir, pattern))
			return 0;
		pos = point - (dir > 0);
		if (in_perl_comment(pos))
			continue;
		if (character(pos - 1) == '$' && get_character_syntax_color(pos - 1)
			  == color_class perl_variable)
			continue;		// Actually a $[ or similar.
		ch = character(pos);
		if (ch == right_d)
			level++;
		else if (ch == left_d)
			level--;
	} while ((dir > 0) ? (level < 0) : (level > 0));
	return 1;
}

int proc_complete_list_matches(char *pat, int initial,
							   int skip_dots, int flags)
{
	int proc_tab_size = tab_size;
	char *pat2 = 0, file[FNAMELEN], *s, *t, cmdnames = 0;

	if (initial && (s = getenv("PATH")) && !strchr(pat, '/') &&
		  (opsys == OS_UNIX || !strchr(pat, '\\') && !strchr(pat, ':'))) {
		pat2 = malloc(strlen(s) + strlen(pat) + 100);
		sprintf(pat2, (opsys == OS_UNIX) ? "{%s}/%s*" :
				"{%s,.}/%s*.{exe,bat,cmd,com}", s, pat);
		while (s = strchr(pat2, path_list_char))
			*s = ',';
		flags |= EXACTONLY;
		pat = pat2;
		cmdnames = 1;
	}
	zap(_MATCH_BUF);
	if (s = do_file_match(pat, STARTMATCH | flags))
		do {
			if (opsys == OS_UNIX && initial && lowaccess(s, LOWACC_X))
				continue;
			if (cmdnames) {
				strcpy(file, s);
				s = get_tail(file, 0);
				if (opsys != OS_UNIX)
					*get_extension(s) = 0;
			}
			if (skip_dots && *get_tail(s, 0) == '.'
				  && string_matches_pattern(get_tail(s, 0), "%.%.?[/\\]?<eof>"))
				continue;
			if (cygwin_filenames & 64)
				while (t = strchr(s, '\\'))
					*t = '/';
			buffer_printf(_MATCH_BUF, "%s\n", s);
		} while ((s = do_file_match(pat, 0)) && !user_abort);
	free(pat2);	// Free if we allocated it.
	save_var bufname = _MATCH_BUF;
	save_var abort_searching = 0;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	tab_size = proc_tab_size;
	return lines_between(0, size(), 0);
}

window_mouse_handler(win)	// try running window-specific handler
{				// return 1 if handler says to stop now
	int owin = window_handle;
	int (*f)();

	if (!is_window(win))
		return 0;
	window_handle = win;
	f = mouse_handler;
	window_handle = owin;
	return f ? (*f)(win) : 0;
}

when_loading()
{
	int i;

	int dorest = 0;
	if (!strcmp(eel_version, "13b1") || dorest) {
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b2") || dorest) {
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b3") || dorest) {
		save_var bufnum;
		i = buf_list(0, 0);
		do {
			bufnum = i;
			if (strstr("Fundamental HTML XML", major_mode))
				major_mode = strkeep(major_mode);
		} while (i = buf_list(1, 1));
		if (strstr("Fundamental HTML XML", major_mode.default))
			major_mode.default = strkeep(major_mode.default);
		drop_name("-fundamental-mode-name");
		drop_name("-html-mode-name");
		drop_name("-xml-mode-name");
		drop_name("tag-c-remove-class");
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b4") || dorest) {
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b5") || dorest) {
		dorest = 1;
	}
	if (!strcmp(eel_version, "13b6") || dorest) {
		when_displaying.default = default_when_displaying;
		if (!strcmp(grep_ignore_file_extensions,
					"|.obj|.exe|.o|.b|.dll|.lib|.zip|.tgz|"))
			strcpy(grep_ignore_file_extensions,
				   "|.obj|.exe|.o|.b|.dll|.lib|.zip|.tgz|.lo|.a|.so|");
		dorest = 1;
	}
}
