/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B7 to 13 B8, assuming further updaters.

#include "eel.h"
#include "colcode.h"
#include "c.h"
#include "session.h"
#include "perl.h"
#include "diff.h"
#include "dired.h"
#include "epsnet.h"
#include "proc.h"
#include "grep.h"
#include "html.h"
#include "tags.h"

//#include "vspell.e"

command down_line()
{
	int true;

	down_or_up_lines(1);
	true = point;
	while (scroll_at_end > 0 && true > window_end) {
		window_scroll(scroll_at_end);
		if (window_end + 1 >= size() - narrow_end || in_macro())
			break;		/* make sure we stop at end */
		point = window_end + 1;	/* compute window_end explicitely */
		window_end = next_screen_line(scroll_at_end) - 1;
		check_abort();
	}
	point = true;
}

int get_key_response(char *pr, char *valid, int def, char *helpcmd)
{
	do {
		noteput("%s", pr);
		getkey();
		check_abort();
		if (IS_ENTER_KEY(key) && def > 0)
			key = def;
		if (IS_HELP_KEY(key) || key == '?')
			if (helpcmd)
				help_on_command(helpcmd);
	} while (!IS_ASCII_KEY(key) || !strchr(valid, toupper(key)));
	say("");
	return toupper(key);
}

delete_user_buffer(bname)	/* delete a buffer that might have coloring */
char *bname;
{
	if (!exist(bname))
		return;
	save_var bufname = bname;
	maybe_delete_autosave();
	if (telnet_id)
		telnet_send(telnet_id, NULL);
	restore_vars();
	drop_coloring(name_to_bufnum(bname));
	highlight_off();
	delete_buffer(bname);
}

char *preserve_buffer_type(char *buf)
{
	struct scp_job *p;
	save_var bufname = buf;

	for (p = scp_job; p; p = p->next)
		if (p->buf == bufnum)
			return p->busy ? "Scp" : 0; // "Scp (Idle)";
	if (is_process_buffer(bufnum) || can_send_invisible())
		if (*buf != '-')
			return strcmp(major_mode, "Fundamental") ? major_mode : "Process";
	return 0;
}

// Return full path to include file rel by searching paths, using
// current mode's rules.
char *c_search_include_path(char *rel, int flags)
{
	char tmp[FNAMELEN], *p, *classpath, *include;

	flags |= PATH_PERMIT_WILDCARDS;
	include = getenv("INCLUDE");
	classpath = getenv("CLASSPATH");
	p = 0;
	if (!p && !fnamecmp(get_extension(filename), ".e")) {
		sprintf(tmp, "include%c%s", path_sep, rel);
		p = lookpath(tmp, 0);
	}
	if (!p && is_java_buffer() && classpath)
		p = look_on_path(rel, flags, classpath);
	if (!p && include)
		p = look_on_path(rel, flags, include);
	if (!p && *include_directories)
		p = look_on_path(rel, flags, include_directories);
	if (!p)
		p = look_on_path(rel, flags, (opsys == OS_UNIX)
						 ? INCL_PATH_UNIX : INCL_PATH_WIN);
	return p;
}

// Return the shortest suffix of the file name f that can be found via
// the include path.
char *c_find_on_include_path(char *f, int flags)
{
	char *s = f + strlen(f) - 1, *t, full[FNAMELEN];
	for (; s >= f; s--) {
		if (s == f) {
			break;
		} else if (is_path_separator(s[-1])
				   && (t = c_search_include_path(s, flags))) {
			strcpy(full, t);
			absolute(full);
			if (!fnamecmp(full, f))
				break;
		}
	}
	return s;
}

copy_include_file_name_c()
{
	char line[FNAMELEN], *s = c_find_on_include_path(filename, 0);

	if (copy_include_file_name_options & 1)	// Always use ""?
		sprintf(line, "#include \"%s\"", s);
	else
		sprintf(line, "#include <%s>", s);
	if (is_java_buffer()) {
		if (is_path_separator(*s))
			s++;
		sprintf(line, "import %s", s);
		if (s = strstr(line, ".java"))
			*s = 0;
		strcat(line, ";");
	} else if (is_path_separator(*s) || opsys == OS_DOS && s[1] == ':') {
		s = c_find_on_include_path(filename, PATH_ADD_CUR_DIR);
		sprintf(line, "#include \"%s\"", s);
	}
	while (s = strchr(line, '\\'))
		*s = '/';
	if (is_java_buffer())
		while (s = strchr(line, '/'))
			*s = '.';
	copy_line_to_clipboard(line, 1);
}

command copy_include_file_name() on cc_tab[ALT('i')]
{
	char *s, line[FNAMELEN], mode[FNAMELEN], name[FNAMELEN];
	char orig_dir[FNAMELEN], dir[FNAMELEN];
	int i;

	if (!*filename)
		error("This buffer has no file name.");
	clean_mode(mode);
	sprintf(name, "copy-include-file-name-%s", mode);
	i = find_index(name);
	switch (name_type(i)) {		// i has name table index of subr/var
		case NT_SUBR: case NT_COMMAND:
		case NT_AUTOSUBR: case NT_AUTOLOAD:
			(*(void (*)())i)();
			return;
		case NT_VAR: case NT_BUFVAR:
		case NT_WINVAR: case NT_BUILTVAR:
			strcpy(dir, filename);
			*get_tail(dir, 1) = 0;
			getcd(orig_dir);
			on_exit chdir(orig_dir);
			chdir(dir);
			s = get_str_var(i);
			if (s && *s) {		// Got a template back.
				build_filename(line, s, filename);
				copy_line_to_clipboard(line, 1);
				return;
			}
			break;
	}
	sprintf(line, "import %r", get_tail(filename, 0));	// Default.
	copy_line_to_clipboard(line, 1);
}

early_restoring()
{
	char fname[FNAMELEN];
	int i;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_DOS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
	if (load_from_state)
		show_text(-2, 0, "Loaded from state file %s", state_file);
	else if (*state_file)
		show_text(-2, 0, "Loaded from state file %s (quickup)", state_file);
	else
		show_text(-2, 0, "Loaded from bytecode files");
	for (i = 0; original_argv(i); i++)
		show_text(-2, 0, "Argv[%d]=%s", i, original_argv(i));
}

init_from_file()
{
	char *p, file[FNAMELEN];
	int res;
	jmp_buf this_level;

	save_var top_level = &this_level;	// Don't let errors abort,
	if (setjmp(top_level)) {		// so session file can still load.
		show_text(0, -1, "Error loading %f, see %s buffer.",
				  einit_file_name, MSG_BUFFER);
		return;
	}
	save_var _abort_load_on_error = 2;
	if ((load_customizations & 1) && (p = lookpath(einit_file_name))) {
		strcpy(file, p);
		show_text(-2, 0, "Loaded init file %s", file);
		save_var bufnum = zap("-customizations");
		do_file_read(file, default_translation_type);
		ini_mode();
		res = parse_cmds();
		tab_size = tab_size.default;
		restore_vars();
		if (res)
			show_text(0, -1, "Error loading %f, see %s buffer.", file, MSG_BUFFER);
	}
}

get_file_dir(res, pr)
	char *res, *pr;
{
	jmp_buf *old_level = top_level, this_level;
	char def[FNAMELEN], old[FNAMELEN], other[FNAMELEN];
	int ret = 0, ok;

	get_buffer_filename(other);
	if (!inp_get_other_name || !*inp_get_other_name)
		save_var inp_get_other_name = other;
	ok = get_buffer_directory(def);
	save_var completion_lister = list_file_matches;
	save_var list_finder = do_file_match;
	save_var abort_file_matching = ABORT_JUMP;
	save_var resize_menu_list = 1;
	save_var insert_default_response = 0;

	if (prompt_with_buffer_directory == PBD_INSERT
		  || prompt_with_buffer_directory == PBD_JUSTINS)
		insert_default_response = 3;
	if (prompt_with_buffer_directory == PBD_TEMPCD) {
		top_level = &this_level;
		getcd(old);
		if (ok)
			chdir(def);
		ret = setjmp(top_level);
	}
	if (!ret)
		comp_read(res, pr, f_match, COMP_FILE | FOLDFILE, def);
	if (prompt_with_buffer_directory == PBD_TEMPCD)
		chdir(old);
	if (prompt_with_buffer_directory == PBD_INSERT && !strcmp(res, def)
		  && !is_remote_file(res))
		*res = 0;
	top_level = old_level;
	if (*res && !is_remote_file(res) && ok && !wild_pattern(res)
		&& prompt_with_buffer_directory == PBD_TEMPCD)
		absolute(res, def);
	else if (*res && !wild_pattern(res))
		absolute(res);
	if (ret)
		longjmp(top_level, ret);
}

dired_examine_deleting() on dired_tab['E']
{
	int buf = bufnum;
	dired_examine();
	if (buf && buf != bufnum)
		buf_delete(buf);
}

make_name_for_mode_line(char *bname, int room)
{
	char fname[FNAMELEN];
	int in_parens = 0;

	if (*filename && force_save_as != FSA_NEWFILE) {
		if (full_path_on_mode_line == 2)
			strcpy(fname, get_tail(filename, 0));
		else if (full_path_on_mode_line)
			strcpy(fname, filename);
		else
			relative(filename, fname);		/* if file, put name */
	} else {
		strcpy(fname, (full_path_on_mode_line == 2) ? get_tail(bufname, 0)
			   : bufname);
		if (!is_dired_buf())
			in_parens = 1;
	}
	abbreviate_file_name(fname, room);
	sprintf(bname, in_parens ? "(%s) " : "%s ", fname);
}

assemble_mode_line(char *line)
{
	int want_func = 0, want_mode = 0, want_name = 0, size = 0, maxsize;
	int want_bigname, want_lastname;
	char bname[FNAMELEN * 2], funcname[FNAMELEN], mode[FNAMELEN], *s, *t, *q;
	char extra[FNAMELEN], bigname[FNAMELEN * 2], lastname[FNAMELEN * 2];

	strcpy(extra, mode_more_extra);
	if (mode_extra)
		strcat(extra, mode_extra);
	for (s = mode_format; *s; s++)
		if (*s == '%' && s[1])
			switch (*++s) {
				case 'B':	want_bigname = 1; break;
				case 'b':	want_name = 1; break;
				case 'L':	want_lastname = 1; break;
				case 'M':	want_mode = 1; break;
				case 'f':	want_func = 1; break;
				case 'E':	size += strlen(extra);
							break;
				default:	// Approximate the size of various options.
					if (strchr("dDmsSanehH", *s))
						size += 2;
					else if (strchr("cClPp", *s))
						size += 4;
					else
						size++;
					break;
			}
		else
			size++;
	*lastname = *bigname = *bname = *funcname = *mode = 0;
	if ((want_func) && display_func_name_method() == 1) {
		strncpy(funcname, display_func_name, ptrlen(funcname) - 2);
		if (*funcname)
			strcat(funcname, " ");
	}
	size += strlen(funcname);
	if (want_mode)
		make_mode_for_mode_line(mode);
	size += strlen(mode);
	maxsize = MIN(window_width, ptrlen(line));
	if (want_name)
		make_name_for_mode_line(bname, maxsize - size);
	size += strlen(bname);
	save_var full_path_on_mode_line = 2;
	if (want_lastname)
		make_name_for_mode_line(lastname, maxsize - size);
	size += strlen(lastname);
	full_path_on_mode_line = 1;
	if (want_bigname)
		make_name_for_mode_line(bigname, maxsize - size);
	for (s = mode_format, t = line; (*t++ = *s) && ptrlen(t) > 2; s++)
		if (*s == '%' && s[1]) {
			q = 0;
			switch (*++s) {
				case 'B':	q = bigname; break;
				case 'b':	q = bname; break;
				case 'L':	q = lastname; break;
				case 'M':	q = mode; break;
				case 'f':	q = funcname; break;
				case 'E':	q = extra; break;
				default:	*t++ = *s; break;
			}
			if (q) {
				strncpy_double_percents(--t, q, ptrlen(t) - 1);
				t += strlen(t);
			}
		}
	*t = 0;
}

prepare_url_operation(char *file, int op, struct url_parts *parts)
{
	char *p;

	save_var case_fold = 1;
	if (!parse_url(file, parts))
		error("Bad URL %s", file);
	if (!strstr(" ftp scp ", parts->service) && (op & FTP_OP_MASK) != FTP_RECV)
		error("Can't do that with %s: buffers, only ftp: or scp: buffers.",
			  parts->service);
	if (!strstr(" ftp scp http telnet ssh ", parts->service))
		error("Unknown service %s: in URL.", parts->service);

	if (!strcmp(parts->service, "ftp")) {	// An FTP URL?
		if (!parts->have_usr)		// No user specified.
			strcpy(parts->usr, "anonymous");
		if (!parts->have_password)	// If no password yet,
			if (!strfcmp(parts->usr, "anonymous")	// Use anon pwd?
				|| !strfcmp(parts->usr, "guest"))
				strcpy(parts->pwd, anon_ftp_password);
			else				// Or prompt for one/retrieve old.
				get_password(parts->pwd, parts->host, parts->usr);
	} else if (!strcmp(parts->service, "http")) {
		if (parts->have_usr && !parts->have_password)
			get_password(parts->pwd, parts->host, parts->usr);
	} else if (!strcmp(parts->service, "scp") || !strcmp(parts->service, "ssh")) {
	} else if (parts->have_usr)
		error("User name not supported in %s URL's", parts->service);

	while (p = index(parts->fname, '\\'))
		*p = '/';			// Convert all \'s to /'s.
}

// Try to retrieve a file via http and store it on disk.  Sz is the
// expected size, for progress messages.  Returns 0 if OK.
int download_file_to_disk(char *url, char *fname, int sz)
{
	int orig = bufnum, b, got_sz;
	do_find(url, FILETYPE_AUTO);
	b = bufnum;
	if (strcmp(filename, url))
		return 1;
	to_buffer_num(orig);
	bufnum = b;
	while (mode_extra) {
		got_sz = size();
		bufnum = orig;
		if (sz)
			note("Downloading file... %'d of %'d", got_sz, sz);
		delay(1, COND_KEY);
		check_abort();
		bufnum = b;
	}
	bufnum = orig;
	say("Done, writing %f to disk...", fname);
	bufnum = b;
	int err = file_write(fname, FILETYPE_AUTO);
	bufnum = orig;
	drop_buffer(bufnum_to_name(b));
	if (!err)
		say("");
	return err;
}

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1,
			".* (not a directory|No such|not found)"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & (SCP_CHECK_IS_DIR
											  | SCP_CHECK_MATCHES)))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is| is now) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connecting to "
								 "|Connection closed|Listing directory "
								 "|Using username \"|sftp%>"
								 "|(.* |^)pass(word|phrase).*:)")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

// Handle activity in scp buffer.
scp_handler(int activity, int buf, int from, int to)
{
	struct scp_job *p;

	save_var bufnum = buf;
	for (p = scp_job; p; p = p->next)	// Find the data for this buffer.
		if (bufnum == p->buf)
			break;
	if (!p || !p->destbuf)
		return;
	if (activity == NET_DONE)
		scp_handle_done(p);
	else if (activity == NET_RECV) {
		if (p->op == FTP_LIST || p->op == FTP_MISC || (p->flags & SCP_IS_XFER))
			scp_handle_list(p, from, to);
		else
			scp_handle_xfer(p, from, to);
		if (matches_at(to, -1, "(.* |^)pass(word|phrase).*: ?")) {
			ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
			menu_command = "password-prompt";
		}
	}
	refresh();
}

char *ssh_command_variable(char *dest, char *def, char *pat)
{
	char s[FNAMELEN];
	int i;

	strcpy(dest, def);
	sprintf(s, pat, opsys == OS_UNIX ? "unix" : "windows");
	i = find_index(s);
	if (i)
		expand_string_template(dest, get_str_var(i), "", NULL);
}

do_scp_xfer(struct url_parts *parts, struct scp_job *p, char *template,
			int mode, int use_tmp)
{
	char tmpfile[FNAMELEN], cmd[FNAMELEN], *args[5], *s, cmdname[FNAMELEN];

	if (use_tmp) {
		new_file_write("", translation_type, NULL, FILE_IO_TEMPFILE, mode);
		strcpy(p->file, get_tempfile_name());
		p->flags |= SCP_TMP_FILE;
	}
	if (!(p->flags & (SCP_SCP_XFER | SCP_CLIENT_MODE))
		  && !strchr(p->file, '%')) {
		scp_run_helper(parts, p);
		p->flags |= SCP_IS_XFER;
		sprintf(cmd, (p->op == FTP_SEND) ? "put \"%s\" \"%%s\""
										 : "get \"%%s\" \"%s\"", p->file);
		if (is_win32)
			for (s = cmd; s = strchr(s, '\\'); )
				*s = '/';	// Windows sftp gets confused by \'s sometimes.
		scp_send_some_command(parts, p, cmd);
		return;
	}
	strcpy(tmpfile, get_tail(p->file, 1));
	args[0] = parts->usr;
	args[1] = parts->host;
	args[2] = parts->fname;
	args[3] = tmpfile;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command");
	args[4] = cmdname;
	if (expand_string_template(cmd, template, "uhflc", args))
		error("Invalid scp template.");
	scp_start_process(p, cmd);
}

char *scp_variable(struct url_parts *parts, char *varname);

scp_run_helper(struct url_parts *parts, struct scp_job *p)
{
	char cmd[FNAMELEN], *args[3], cmdname[FNAMELEN];

	args[0] = parts->usr;
	args[1] = parts->host;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command");
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(parts, "run-helper"),
							   "uhc", args))
		error("Invalid run helper template.");
	if (!is_process_buffer(p->buf)) {
		scp_start_process(p, cmd);
		if (p->homedir)
			free(p->homedir);
		p->homedir = 0;
		p->busy = 2;
	}
}

struct scp_job *do_scp(struct url_parts *parts, int op);

int scp_is_url_a_dir(char *s, struct url_parts *parts)
{
	if (scp_flags(parts) & (SCP_CLIENT_MODE | SCP_NO_DIR_CHECK))
		return 0;
	if (!*parts->fname)
		return 1;
	int tmp = tmp_buf();
	scp_get_response(2, s, parts, tmp);
	int isdir = !buf_size(tmp);
	buf_delete(tmp);
	return isdir;
}

int url_operation(char *file, int op)
{
	struct url_parts parts;
	int port, i;

	prepare_url_operation(file, op, &parts);
	port = numtoi(parts.port);
	if (*parts.port && got_bad_number)
		say("Ignoring invalid port number %s", parts.port);
	if (i = find_variable_by_host("fallback_remote_translation_type", parts.host))
		fallback_translation_type = get_num_var(i);
	if (!strfcmp(parts.service, "http")) {
		if (got_bad_number)
			port = 80;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTP files.");
		do_http(&parts, port);
	} else if (!strfcmp(parts.service, "telnet")) {
		if (got_bad_number)
			port = 23;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with telnet buffers.");
		do_telnet(parts.host, port, bufname);
	} else if (!strfcmp(parts.service, "ssh")) {
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with ssh buffers.");
		do_ssh(parts.host, bufname, parts.usr);
	} else if (!strfcmp(parts.service, "scp")) {
		do_scp(&parts, op);
	} else
		do_ftp_op(bufname, parts.host, parts.port, parts.usr, parts.pwd,
				  parts.fname, op);
	return 0;		// No error possible yet (asynchronous).
}

void ssh_activity(int activity, int buf, int from, int to);

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	zap(bname);
	to_buffer(bname);
	type_point = 0;
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s");
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
}

command ssh()
{
	char hostname[FNAMELEN], buf[FNAMELEN], dir[FNAMELEN], *p;
	int i = 1;
	struct url_parts parts;

	inet_only();
	if (!(has_feature & FEAT_MULT_CONCUR))
		error("SCP functions require Unix or Windows NT/2K/XP");
	tiled_only();
	*hostname = 0;
	if (get_buffer_directory(dir) && is_remote_file(dir)
		  && parse_url(dir, &parts) && strstr(" ssh scp ", parts.service)) {
		if (parts.have_usr)
			sprintf(dir, "%s@%s", parts.usr, parts.host);
		else
			sprintf(dir, "%s", parts.host);
		get_strdef(hostname, "Ssh to (hostname or user@hostname)", dir);
	} else
		get_str_auto_def(hostname, "Ssh to (hostname or user@hostname)");
	if (!*hostname)
		return;
	sprintf(buf, "ssh-%s", hostname);
	p = buf + strlen(buf);
	for (;;) {		// Find an unused buffer.
		if (!exist(buf))
			break;
		bufname = buf;
		if (!is_process_buffer(bufnum))
			break;
		sprintf(p, " %d", i++);
	}
	do_ssh(hostname, buf, "");
}

int is_ftp_url_a_dir(char *s)
{
	char cmd[FNAMELEN];
	struct url_parts parts;
	int port, res;

	prepare_url_operation(s, FTP_MISC | FTP_WAIT, &parts);
	if (!strcmp(parts.service, "scp"))
		return scp_is_url_a_dir(s, &parts);
	if (ftp_compatible_dirs & 4)
		return 0;
	if (*parts.fname)
		sprintf(cmd, "CWD %s", parts.fname);
	else
		return 1;
	port = numtoi(parts.port);
	if (got_bad_number)
		port = 21;
	char *t1 = temp_buf(), *t2 = temp_buf();
	ftp_op(t1, t2, parts.host, port, parts.usr, parts.pwd, cmd,
		   FTP_MISC | FTP_WAIT);
	save_var bufname = t2;
	point = 0;
	res = !re_search(1, "^(550|530) ");
	restore_vars();
	delete_buffer(t1);
	delete_buffer(t2);
	return res;
}

// Should we use the GUI dialog?
use_common_dialog(int var)
{
	if (!(has_feature & FEAT_FILE_DIALOG))
		return 0;
	switch (var) {
		case 0:	return 0;	// Never use.
		case 2:	return 1;	// Always use.
		case 3:	return !run_by_mouse;	// Unless via menu or toolbar.
	}
	return run_by_mouse;	// Use if command was invoked by menu or toolbar.
}

command force_common_file_dialog()
{
	save_var want_common_file_dialog;
	switch (want_common_file_dialog) {
		case 0:		want_common_file_dialog = 2; break;
		case 2:		want_common_file_dialog = 0; break;
		default:	want_common_file_dialog = 3; break;
	}
	do {
		mention("Common dialog: %s ", want_common_file_dialog ? "On" : "Off");
		getkey();
	} while (key == MOUSE_MOVE);
	say("");
	do_topkey();
}

// Maybe display why the file was read as binary.
do_file_read_kibitz()
{
	if (file_read_kibitz && *get_file_read_kibitz())
		say("%f: %s.", filename, get_file_read_kibitz());
}

command find_file()
{
	char fname[FNAMELEN];
	int want_readonly = 0, flags = 0;

	tiled_only();
	was_quoted = 0;
	save_var show_mouse_choices = 1;
	if (use_common_file_dialog()) {
		*fname = 0;
		if (!use_common_open_file_dlg(fname, "Open file", &flags, 0))
			return;
		if (flags & OFN_READONLY)
			want_readonly = 1;
		save_var file_pattern_wildcards = 0; // []'s etc in file name are literal.
	} else
		get_file_dir(fname, "Find file: ");
	if (!*fname)
		dired_one(fname);
	else if (was_quoted && !is_directory(fname) && !is_remote_file(fname))
		find_it(fname, give_line_translate(fname));
	else
		do_find(fname, give_line_translate(fname));
	do_file_read_kibitz();
	if (want_readonly)
		force_save_as = FSA_READONLY;
	iter = 0;
}

// Search for a URL at or after point, return nonzero.
int find_url_here()
{
	int start = point;
	to_begin_line();
	while (re_search(1, "(www%.|(https?|ftp):[/\\][/\\])[^\n\t <>\"]+"
					 "[^\n\t <>()\"',.]|mailto:<word|@|.>+"))
		if (matchend >= start)
			return 1;
	return 0;
}

// Find a file.  If has_arg, ask for its translation.
int ask_find_it(char *fname)
{
	int ret = find_it(fname, give_line_translate(fname));
	do_file_read_kibitz();
	return ret;
}

// This runs after we compile a GAMS file using compile-buffer.  It
// finds the matching .lst file, then translates any error messages
// there into a traditional error message format, inserting the error
// lines into compilation-output where next-error and friends will look
// for them.
gams_read_list_file(int res, int buf)
{
	char list_file[FNAMELEN], orig_file[FNAMELEN], num[100], inc_file[FNAMELEN];
	char gams_err[FNAMELEN];
	int out = create("compilation-output"), i;

	if (res == 2)
		say("Interrupted, skipping error message conversion.");
	if (res == 2 || res == 3 || res == 5)	// Do_compile() didn't complete
		return;								// successfully.
	save_var bufnum = buf;
	strcpy(orig_file, filename);
	strcpy(list_file, filename);
	strcpy(get_extension(list_file), ".lst");
	int tmp = tmp_buf();
	save_var bufnum = tmp;
	if (!err_file_read(list_file, FILETYPE_AUTO)) {
		save_var point = 0;
		while (re_search(1, "^%*%*%*%* *%$")) {
			i = give_end_line();
			grab(point, i, gams_err);	// Get the error code.
			nl_reverse();		// Now copy the source line.
			buf_xfer(out, give_begin_line(), i + 1);
			point = i + 1;
			if (parse_string(1, " *([0-9]+)")) {	// Error in main file.
				grab(find_group(1, 1), find_group(1, 0), num);
				buf_printf(out, "%s %d: GAMS error %s\n",
						   orig_file, numtoi(num) - 1, gams_err);
			} else if (parse_string(1, "<*>+ *([0-9]+.+)\n *([0-9]+)")) {
				grab(find_group(1, 1), find_group(1, 0), gams_err);
				grab(find_group(2, 1), find_group(2, 0), num);
				buf_printf(out, "%s %d: GAMS error %s\n",
						   orig_file, numtoi(num) - 1, gams_err);
			} else if (parse_string(1, "%*%*%*%* *LINE +([0-9]+) +"
									"INPUT +([^ \n].*)")) {	// In another file.
				grab(find_group(1, 1), find_group(1, 0), num);
				grab(find_group(2, 1), find_group(2, 0), inc_file);
				buf_printf(out, "%s %d: GAMS error %s\n",
						   inc_file, numtoi(num), gams_err);
			}
		}
	}
	restore_vars();
	buf_delete(tmp);
	next_error();
}

prepare_grep_buffer(char *buf, int flags, char *pat, char *filepat,
					char *forcedir)
{
	char search_type[FNAMELEN], dir[FNAMELEN];
	int before_head;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (grep_empties_buffer)
		modified = 0;
	save_var modified;
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	if (forcedir && *forcedir)
		strcpy(dir, forcedir);
	else
		getcd(dir);
	bprintf("Found 0 lines with %s\"%s\" in %s%s (cur dir \"%s\")\n",
			search_type, pat, (flags & BUFGREP) ? "(buffers) " : "",
			filepat, dir);
	set_character_color(before_head, point, color_class grep_header);
	if (display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	refresh();
	return before_head;
}

init_copy_for_grep(char *fname, char *dir)
{
	*dir = 0;
	if (*filename) {
		strcpy(fname, get_tail(filename, 0));
		strcpy(dir, filename);
		*get_tail(dir, 0) = 0;
	} else
		sprintf(fname, BUFPREFIX "%s", bufname);
}

// Copy current line to grep-like buffer buf as a match in fname,
// highlighting range from...to in color.  Return offset of copied line.
int copy_line_for_grep(char *fname, int buf, int from, int to, int color)
{
	buf_printf(buf, "%p%s: ", color_class grep_filename, fname);
	int pos = get_buf_point(buf), s;
	buf_xfer_colors(buf, s = give_begin_line(), give_end_line());
	buf_set_character_color(buf, pos + from - s, pos + to - s, color);
	buf_printf(buf, "%p\n", color_class grep_filename);
	return pos;
}

// Copy temp buffer to grep buffer, adding header line.
finish_copy_for_grep(int buf, char *type, char *fname, char *dir)
{
	bufnum = buf;
	int lines = lines_between(0, size(), 0);
	update_grep_title(prepare_grep_buffer("grep", 0, type,
										  fname, dir), lines);
	bufnum = buf;
	buf_xfer_colors(name_to_bufnum("grep"), 0, size());
	bufnum = window_bufnum;
	backward_paragraph();
	if (parse_string(1, "Found [0-9]+ lines with \""))
		delete(matchend - 1, matchend);
	if (parse_string(1, "Found [0-9]+ lines with [^\n\"]*\""))
		delete(matchend - 1, matchend);
	nl_forward();
}

command search_man_pages()
{
	char pagepat[FNAMELEN], section[FNAMELEN], pat[FNAMELEN], *searchpat;
	int flags, before_head, cnt = 0, this_cnt;
	char man_path[FNAMELEN * 2], match[FNAMELEN], *path, *t, *q, *f;

	flags = get_search_string("Search man pages for text: ", UNKNOWN | FILEGREP);
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	searchpat = default_search_string(flags);

	save_var _help_on_cmd = last_index;
	get_last_recall((char *(*)()) 0, pat);
	if (!*pat)
		strcpy(pat, "*");
	get_strnone(section, "Only search these man page sections, "
			   "such as 1 2 3, or * for all: ", pat);
	for (t = section, q = match; *t; t++)
		if (isdigit(*t))
			*q++ = *t;
	*q = 0;
	if (*match)
		sprintf(section, "[%s]", match);
	else
		*section = 0;

	save_var recall_id = "search-man-pages pat";
	get_last_recall((char *(*)()) 0, pat);
	if (!*pat)
		strcpy(pat, "*");
	get_strnone(pagepat, "Restrict to man pages matching pattern: ", pat);
	if (!*pagepat)
		strcpy(pagepat, "*");

	save_var cygwin_filenames |= 32;
	save_var show_status = 0;
	build_man_path(path = man_path);
	before_head = prepare_grep_buffer("grep", flags, searchpat, "man pages", NULL);

	while (!user_abort) {
		t = strchr(path, ':');
		if (t)
			strncpy(pat, path, t - path);
		else
			strcpy(pat, path);
		if (*pat && pat[strlen(pat) - 1] != '/')
			strcat(pat, "/");
		convert_cygwin_filename(pat, 0);
		sprintf(pat + strlen(pat), "man%s*/%s", section, pagepat);
		f = do_file_match(pat, STARTMATCH);
		for (; f && !user_abort; f = do_file_match(pat, 0)) {
			q = get_extension(f);
			if (!q || !*q)
				continue;
			strcpy(match, get_tail(f, 0));
			noteput("Searching %s...", match);
			q = get_extension(match);
			if (q && !strcmp(q, ".gz"))
				*q = 0;
			q = get_extension(match);
			if (!q || !*q)
				continue;
			this_cnt = search_one_man_page(f, flags, searchpat, match);
			if (this_cnt) {
				update_grep_title(before_head, cnt += this_cnt);
				build_first = 1;
				refresh();
			}
		}
		if (!t)
			break;
		path = t + 1;
	}
	point = before_head;
	nl_forward();
	if (!user_abort)
		modified = 0;
	say("");
}

#define MANCONF_PATH	"^(MANDATORY_|OPTIONAL_)?MANPATH" \
						"(_MAP[ \t]+[^ \t\n]+)?[ \t]+([^ \t\n]+)$"

build_man_path(char *dest)
{
	char *path = getenv("MANPATH"), out[FNAMELEN * 4];
	int tmp;

	if (path && *path) {
		strcpy(dest, path);
		return;
	}
	save_var bufnum = tmp = tmp_buf();
	if (!file_read("/etc/manpath.config", FILETYPE_AUTO)
		  || !file_read("/etc/man.config", FILETYPE_AUTO)) {
		do_drop_matching_lines(REGEX, MANCONF_PATH, 0);
		string_replace(MANCONF_PATH, "#3", REGEX);
	}
	stuff("/usr/man\n/usr/share/man\n");
	*out = 0;
	replace_file_with_output(out, is_win32 ? "man --path %s" : "manpath %s");
	while (path = strchr(out, ':'))
		*path = '\n';
	stuff(out);
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	simple_re_replace(0, "\n", ":");
	simple_re_replace(0, ":<eof>", "");
	grab(0, size(), dest);
	restore_vars();
	buf_delete(tmp);
}

html_invalid_tag(int pos)
{
	save_var point = pos;
	return do_color_searching(REGEX, "<c:html-punctuation>[<>]")
			&& character(point - 1) == '<';
}

// Return indent for attrs inside a tag that starts at pos, with current
// indentation cur.
html_indent_in_tag(int pos, int cur)
{
	int attrlen = 0, tagind;
	point += parse_string(1, "[ \t]*");
	// Get width of "attname=" on current line.
	if ((html_indenting_rules & 0x8000) && parse_string(1, HTML_NAME "[ \t]*="))
		attrlen = get_column(matchend) - current_column();
	save_var point = pos;
	if (parse_string(-1, "(^|\n)[ \t]*")) {
		// Get indentation of "<tagname".
		tagind = get_indentation(point);
		point += parse_string(1, "[<]" HTML_NAME "[ \t]*");
		if (attrlen && parse_string(1, "[ \t\n]*" HTML_NAME "[ \t]*=")) {
			int i = get_column(matchend) - attrlen;	// Indent to align ='s.
			if (i >= 0)		// Always indent at least to level of tag name.
				return MAX(i, tagind + 1);
		}		// Otherwise, don't align ='s.
		return current_column();
	}
	restore_vars();
	save_var point;
	if (!re_search(-1, "[^ \t\n]"))
		return cur;
	return get_indentation(point);
}

command html_list_mismatched_tags() on html_tab[ALT('L')]
{
	char tag[FNAMELEN], fname[FNAMELEN], dir[FNAMELEN];
	int buf = tmp_buf(), cnt = 0, pos = point;
	save_var point = 0;
	save_var case_fold = (sublanguage_code != SUBLANG_XML);
	save_var narrow_start = 0, narrow_end = 0;
	on_exit buf_delete(buf);
	recolor_buffer_range(0, size());
	init_copy_for_grep(fname, dir);
	save_var abort_searching = 0;
	while (do_color_searching(REGEX, "<c:html-punctuation>[<]") == 1
		   && !user_abort) {
		if (do_html_find_matching_tag(point - 1, tag) != -1)
			continue;
		switch (html_is_start_tag(point - 1, 1)) {
			case 1:
			case -1:
				point += parse_string(1, "<wspace|/>*");
				copy_line_for_grep(fname, buf, point, point + strlen(tag),
								   color_class html_mismatch);
				cnt++;
				break;
		}
		if (size() > 20000)
			note("%d mismatched tags, %d%% complete.",
				 cnt, muldiv(point, 100, size()));
	}
	user_abort = 0;
	if (size() > 20000)
		say("");
	point = pos;
	if (buf_size(buf))
		finish_copy_for_grep(buf, "mismatched tags", fname, dir);
	else
		say("No mismatched tags.");
}

command html_find_matching_tag()
{
	char tag[FNAMELEN];

	save_var point;
	if (curchar() != '<' && !search(-1, "<"))
		error("No tag found.");		// Get to opening <.
	int res = do_html_find_matching_tag(point, tag);
	restore_vars();
	if (res == -2)
		error("No tag found.");
	else if (res == -1 || res == -3) {
		say("No match for %s tag.", tag);
		expire_message = 1;
	} else
		point = res;
}

tag_mode_jscript()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (re_search(1, "(^|;)[ \t]*(var|function)[ \t]+(<word>+)")) {
		int var = tolower(character(find_group(2, 1))) == 'v';
		grab(start = find_group(3, 1), find_group(3, 0), func);
		if (!isdigit(*func))
			add_tag(func, start);
		if (var)
			for (;;) {
				point += parse_string(1, "([ \t\n]|//.*)*");
				if (curchar() == '=')
					re_search(1, "![\n,;]");
				if (!parse_string(1, ",[ \t\n]*(<word>+)"))
					break;
				grab(start = find_group(1, 1), find_group(1, 0), func);
				point = matchend;
				if (!isdigit(*func))
					add_tag(func, start);
			}
	}
}

// Narrow and tag this region if it's scripting.
html_tag_region(int type, int from, int to)
{
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	point = 0;
	switch (type) {
		case HT_JSCRIPT:
			tag_mode_jscript();
			break;
		case HT_VBSCRIPT:
			tag_mode_vbasic();
			break;
		case HT_PHP:
			save_var sublanguage_code = SUBLANG_PHP;
			tag_mode_perl();
			break;
		case HT_PYTHON:
			tag_suffix_python();
			break;
	}
}

// For each scripting block, tag its contents appropriately.
tag_mode_html()
{
	int from, to, type;
	recolor_buffer_range(0, size());	// Divide into scripting blocks.
	save_var point = 0;
	do {
		type = get_tagged_region("color-as-unit", point, &from, &to);
		html_tag_region(type, from, to);
		check_abort();
		point = to + 1;
	} while (point < size() - narrow_end);
}

tag_mode_xml()
{
	tag_mode_html();
}

tag_mode_php()
{
	tag_mode_html();
}

command list_customizations()
{
	char cmd[FNAMELEN], *args, time[FNAMELEN];
	char ext_file[FNAMELEN], settings_file[FNAMELEN], *s;
	int tmp, cur, ext, i, err;
	int cust = prep_customizations(), clist;

	iter = 1;
	if (!has_arg)
		clean_customizations();
	tiled_only();
	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	delete_file(settings_file);
	i = name_match("-loaded-eel-file-", STARTMATCH);
	save_var bufnum = ext = zap("-list-customizations");
	give_binding(cmd, "exit-level");
	bprintf("Epsilon will include the following EEL extension files as part\n"
			"of your customizations.  Edit this list if needed, then\n"
			"press %s to begin.\n\n%p\n", cmd, color_class viewed_text);
	point--;
	for (; i; i = name_match("", 0))	// List new extensions by searching
		if (name_type(i) == NT_VAR) {	// for id variables.
			s = get_str_var(i);
			bprintf("%s%s\n", s, (check_file(s) == CHECK_FILE)
					? "" : " (not found)");
		}
	to_buffer_num(ext);
	point = 0;
	do {
		recursive_edit();
		if (!buf_exist(ext))
			return;
		point = 0;
		point += parse_string(1, "(.|\n.)*\n*");
		i = test_valid_files(point, size());
		if (i == 2) {
			buf_zap(ext);
			return;
		}
	} while (i);
	point = 0;		// Drop the explanatory text.
	delete(0, parse_string(1, "(.|\n.)*\n*"));
	simple_re_replace(1, " *%(not found%)", "");
	if (err = file_write(ext_file, FILETYPE_BINARY)) {
		file_error(err, ext_file, "write error");
		return;
	}
	save_var reject_client_connections = 1;
	strcpy(cmd, argv[0]);
	convert_to_8_3_filename(cmd);
	sayput("Running %s to create original settings list...", cmd);
	args = " -s original -p0 -noinit -r build-settings-file";
	err = shell_with_args(cmd, args);
	if (err)
		error("Error running %s %s", cmd, args);
	sayput("Reading original settings list...");
	bufnum = tmp = tmp_buf();
	if (err_file_read(settings_file, FILETYPE_BINARY))
		return;
	bufnum = cur = tmp_buf();
	list_all_to_buffer(cur, 3);
	clist = make_difference_list(tmp, cur);
	restore_vars();
	buf_delete(tmp);
	buf_delete(cur);
	to_buffer_num(cust);
	current_time_and_date(time);
	save_spot point;
	i = point;
	bprintf("\n# List-customizations created these settings %s.\n", time);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	buf_delete(clist);
	delete_file(settings_file);
	delete_file(ext_file);
	bprintf("\n");
	point = i;
	if (!has_arg)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("");
}

command clean_customizations()
{
	int dup;

	iter = 0;
	tiled_only();
	to_buffer_num(prep_customizations());
	point = 0;		// First delete previous batch of cleanings.
	simple_re_replace(0, "###.*\n", "");
	point = 0;
	while (point < size() - narrow_end) {
		dup = 0;
		if (parse_string(1, "^(%+<word|dash>+ )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%&<word|dash>+ color class "
							  "for <word|dash>+: )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^%(bind-to-key \"[^\"\n]+\" (\"[^\"\n]+\")%)"))
			dup = check_dup_customization(1, "^%(bind-to-key \"[^\"\n]+\" <>%)");
		else if (parse_string(1, "^(<word|dash|#128-#255>+: macro )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%(load-eel-from-path \"[^\"\n]+\" 2%)$)"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(~<^wspace>+ \".*\": )"))
			dup = check_dup_customization(1, "^<>");
		if (dup) {
			stuff("### ");
			nl_forward();
			continue;
		}
		switch (curchar()) {
			case '(':	skip_paren_cmd(); break;
			case ';':
			case '#':	nl_forward(); break;
			default:
				nl_forward();
				break;
		}
	}
}

command import_customizations()
{
	int tmp = zap("import-customizations"), i, ver = 0, err, clean = !has_arg;
	int debug = (iter == 99);
	char buf[2000], maindir[FNAMELEN], bindir[FNAMELEN];
	char origsta[FNAMELEN], cursta[FNAMELEN], *s, *exe_ext = "", pat[FNAMELEN];
	char answer[6][2000], ranswer[6][2000], oldname[FNAMELEN];
	char origdir[FNAMELEN], *epsout = lookpath("epsout.e");

	iter = 1;
	tiled_only();
	getcd(origdir);
	on_exit chdir(origdir);
	putenv("EPSRUNS");		// Prevent error msg in old Epsilons.
	save_var bufnum;
	if (!epsout)
		error("The file epsout.e is missing.");
	for (i = EELVERSION / 10 - 1; i >= 8 ; i--) {
		if (is_win32) {
			sprintf(buf, "EPSPATH%02d", (i < 10) ? (i * 10) : i);
			if (!verenv(buf))
				continue;
		} else {
			sprintf(buf, "~/.epsilon/epsilon-v%d.sta", i);
			if (!check_file(buf))
				continue;
		}
		buf_printf(tmp, "%d\n", ver = i);
	}
	set_buf_point(tmp, 0);
	switch (count_lines_in_buf(tmp, 0)) {
		case 0:		ver = 0; break;
		case 1:		break;
		default:
			*buf = 0;
			buf_delete_final_newlines(tmp);
			if (!get_choice(tmp, buf, " Update from version ",
							"Update from which version? ",
							"OK", "Cancel", ""))
				return;
			ver = numtoi(buf);
			say("");
	}
	*maindir = *bindir = *origsta = *cursta = 0;
	if (is_win32 && ver) {
		exe_ext = ".exe";
		sprintf(buf, "EPSPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (s = verenv(buf)) {
			strcpy(maindir, s);
			while (s = strchr(maindir, ';')) {
				*s = 0;
				sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
				if (check_file(origsta) == CHECK_FILE)
					break;
				strcpy(maindir, s + 1);
			}
			absolute(maindir);
			sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
		}
		sprintf(buf, "BINPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (*maindir)
			sprintf(bindir, "%s\\bin", maindir);
		s = verenv(buf);
		if (s)
			strcpy(bindir, s);
		absolute(bindir);
		if (*maindir)
			sprintf(cursta, "%s\\epsilon.sta", maindir);
		if (ver >= 13) {
			get_customization_directory(cursta);
			sprintf(cursta + strlen(cursta), "\\epsilon-v%d.sta", ver);
		}
	} else if (ver) {		// Set vars for Unix.
		sprintf(buf, "/usr/local/epsilon%d*/bin/epsilon", ver);
		int max = 0;		// Find newest match of pattern.
		struct file_info fi;
		s = file_match(buf, STARTMATCH);
		*buf = 0;
		while (s && check_file(s, &fi)) {
			if (!max || fi.raw_file_date_low > max) {
				strcpy(buf, s);
				max = fi.raw_file_date_low;
			}
			s = file_match("", 0);
		}
		if (*buf) {
			get_tail(buf, 0)[-1] = 0;
			strcpy(bindir, buf);
			strcpy(origsta, buf);
			strcpy(get_tail(origsta, 0), "original.sta");
		}
		sprintf(cursta, "~/.epsilon/epsilon-v%d.sta", ver);
	}
	save_var window_bufnum;
	to_buffer_num(tmp);
	buf_zap(bufnum);
	sprintf(oldname, "Epsilon %d", ver);
	if (!ver)
		strcpy(oldname, "the old version");
	bprintf("Epsilon will try to import any customizations you saved in your\n");
	bprintf("state file in %s.  Please correct any of\n", oldname);
	give_binding(buf, "exit-level");
	bprintf("these settings as necessary, then press %s\n", buf);
	bprintf("to begin the import process.\n\n");

	if (!ver) {
		bprintf("Epsilon couldn't find any %s,\n", is_win32
				? "Epsilon registry entries"
				: "customized state file in ~/.epsilon");
		bprintf("so you'll have to fill in this information.\n\n");
	}
	bprintf("1. Your customized state file for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text,
			*cursta ? cursta : "full path for your .sta file", -1);
	if (*cursta && check_file(cursta) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	bprintf("2. The original pre-customization state file for %s\n", oldname);
	bprintf("is:\n%p%s\n%p\n",
			color_class viewed_text,
			*origsta ? origsta : "full path for original.sta", -1);
	if (*origsta)
		if (check_file(origsta) != CHECK_FILE)
			bprintf("(This file doesn't exist.)\n\n");
	bprintf("3. If you customized %s by loading EEL files,\n", oldname);
	bprintf("please list each .e file's pathname here, one per line:\n%p\n%p\n",
			color_class viewed_text, -1);
	sprintf(buf, "full path for epsilon%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%c%s%s", bindir, path_sep,
				(ver < 10) ? "e32" : "epsilon", exe_ext);
	bprintf("4. The Epsilon%s executable file for %s is:\n%p%s\n%p\n",
			is_win32 ? " (GUI, not console version)" : "", oldname,
			color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	sprintf(buf, "full path for eel%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%ceel%s", bindir, path_sep, exe_ext);
	bprintf("5. The EEL compiler executable for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	get_customization_directory(buf);
	strcat(buf, "upd");
	mkdir(buf);
	bprintf("6. Use this working directory for intermediate files:\n%p%s\n%p\n",
			color_class viewed_text, buf, -1);
	if (buf && check_file(buf) != CHECK_DIR)
		bprintf("(Warning: Couldn't create this directory.)\n\n");
	for (;;) {
		err = point = 0;
		recursive_edit();
		if (!buf_exist(tmp))
			return;
		bufnum = tmp;
		point = 0;		// Now retrieve the answers.
		for (i = 0; i < 6; i++) {
			sprintf(pat, "^%d%%. [^:]+:\n((.|\n.)*)\n\n", i + 1);
			if (!re_search(1, pat)) {
				say("Can't interpret response %d", i + 1);
				err = 1;
				break;
			}
			grab(find_group(1, 1), find_group(1, 0), answer[i]);	// 8.3 version.
			grab(find_group(1, 1), find_group(1, 0), ranswer[i]);	// For display.
			if (i == 2) {
				err = test_valid_files(find_group(1, 1), find_group(1, 0));
				if (err == 2)
					quick_abort();
				err = err ? 3 : 0;	// Don't prompt again.
			} else if (check_file(answer[i]) != ((i == 5)
				? CHECK_DIR : CHECK_FILE)) {
				say("Response %d: Can't read %s \"%f\".",
					  i + 1, (i == 5) ? "directory" : "file", answer[i]);
				err = 1;
				break;
			} else
				convert_to_8_3_filename(answer[i]);
		}
		if (!err)
			break;
		if (err != 3
			  && !ask_yn("Try again", "Would you like to correct your response? ",
						 "Yes", "No", 0))
			quick_abort();
	}
	if (chdir(answer[5]))
		error("Couldn't set directory to %f.", answer[5]);
	buf_zap(tmp);	// Compile epsout & extensions, then save list of them.
	compile_for_import_customizations(epsout, answer[4], answer[5], 0);
	compile_for_import_customizations(answer[2], answer[4], answer[5], tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}

	// Run the old version with orig state file.
	save_var reject_client_connections = 1;
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[1], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	buf_zap(tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	// Run the old version with new state file.
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[0], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	int cur = tmp_buf();
	on_exit buf_delete(cur);
	bufnum = cur;
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	int clist = make_difference_list(tmp, cur);
	on_exit buf_delete(clist);	// Now format the list of differences.
	if (clean)
		clean_customizations();
	int cust = prep_customizations();
	bufnum = cust;
	current_time_and_date(buf);
	save_var point;
	i = point;
	bprintf("\n# Import-customizations created these settings %s\n", buf);
	bprintf("# from %s.\n", ranswer[0]);
	int ext = tmp_buf();
	on_exit buf_delete(ext);
	buf_printf(ext, "%s\n", ranswer[2]);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	bprintf("\n");
	restore_vars();
	to_buffer_num(cust);
	point = i;
	if (clean)
		comment_old_customizations(point);
	list_undefined_customizations();
	say("Done.  Save file, then restart Epsilon, to use these new settings.");
}

command prefix_fill_paragraph()
{
	prefix_fill_region("^<^alphanum|(|)|wspace>",
					   "^<^alphanum|(|)|newline>+", 1);
}

compile_handler(int activity, int buf, int from, int to)
{
	UNREFERENCED_PARAMETER(to);
	if (activity == NET_DONE) {
		save_var bufnum = buf;
		point = size();
		if (size() && character(point - 1) != '\n')
			bprintf("\n");
		bprintf("Process exited with status %d\n", from);
		update_mode_more_extra(-1);
		if (post_compile_hook_buffer) {
			save_var bufnum = post_compile_hook_buffer;
			if (post_compile_hook)		// Track this compilation back to orig
				(*post_compile_hook)(6, bufnum); // buffer, then run its hook.
		}
	}
}

// Run the given cmd, then search for errors.
// Return 0 if the compile was ok,
// 1 if the process returned an error code,
// 2 if we ran it concurrently and the user started typing,
// 3 if the user didn't ok saving,
// 4 if the process returned 0, but next-error found some errors.
// 5 if we can't tell when the process has finished.
// 6 is reserved; we don't return it, but a post_compile_hook function
// can receive when the compilation finishes it if the user interrupted
// us and we returned 2.
do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
        process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			post_compile_hook_buffer = obuf;
			proc_view_start_compile(bufnum);
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE
					  && !IS_WIN_PASSIVE_KEY(key)) {
					ungot_key = full_key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
			bufnum = obuf;
			check_dates(0);
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		proc_view_start_compile(bufnum);
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

do_execute_eel(char *expr)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	execute_eel_template(expr, src);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		error("Could not compile EEL statement.");
	}
	ev_template();
}

command execute_eel()
{
	char expr[FNAMELEN];

	get_string(expr, "Execute EEL statement: ");
	if (*expr)
		do_execute_eel(expr);
}

load_var()	/* variable definition */
{
	int i, new_type, vtype, hasnum;
	char name[80], junk[80], num[20], *p, *tmp;

	point++;			/* skip initial + */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	new_type = parse_string(1, " buffer", junk)
			? NT_BUFVAR : parse_string(1, " window", junk)
			? NT_WINVAR : NT_VAR;
	if (parse_string(1, " (char ptr )?variable: %<complex>", junk)) {
		if (i <= 0 || name_type(i) != new_type)
			init_warning("Variable %s too complex to define.", name);
		nl_forward();
		return;
	} else if (parse_string(1, " char array ", junk))
		vtype = TYPE_CARRAY;
	else if (parse_string(1, " int ", junk))
		vtype = TYPE_INT;
	else if (parse_string(1, " short ", junk))
		vtype = TYPE_SHORT;
	else if (parse_string(1, " char ", junk))
		vtype = TYPE_CHAR;
	else
bad:		init_error("Bad definition for variable %s.", name);
	if (!parse_string(1, "variable: ", junk))
		goto bad;
	hasnum = parse_string(1, "[-0-9][a-zA-Z0-9]*", num);
	if (vtype == TYPE_CARRAY) {
		if (i <= 0 && hasnum) {
			i = new_variable(name, new_type, TYPE_CARRAY, numtoi(num));
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!parse_string(1, " *<dquote>", junk)
				|| name_type(i) != new_type
				|| vartype(i) != TYPE_CARRAY)
			goto bad;
		use_default = 1;
		if (hasnum && new_type == NT_VAR && numtoi(num) > ptrlen(varptr(i))) {
			drop_name(name);		// Make it bigger by recreating it.
			i = new_variable(name, new_type, TYPE_CARRAY, numtoi(num));
			set_name_user(i, 1);		// Make it a user variable.
			tmp = p = varptr(i);
		}
		tmp = p = varptr(i);
		use_default = 0;
		for (; curchar() != '"'; point++) {
			if (ptrlen(p) < 1)
				init_error("Definition too long for variable %s.", name);
			if (curchar() != '\\') {
				*p++ = curchar();
				continue;
			}
			point++;
			switch (curchar()) {
				case 't': *p++ = '\t'; break;
				case 'r': *p++ = '\r'; break;
				case 'n': *p++ = '\n'; break;
				case 'z': *p++ = 0; break;
				default: *p++ = curchar();
			}
		}
		while (p - tmp < ptrlen(tmp))
			*p++ = 0;
		if (!parse_string(1, "<dquote>$", junk))
			goto bad;
	} else if (vtype) {
		if (i <= 0) {
			i = new_variable(name, new_type, vtype, 1);
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!hasnum && parse_string(1, ".*", junk))
			sprintf(num, "%d", find_index(junk));
		set_var_val(i, num, 2);
	}
}

command untag_files()
{
	int t, p, count;
	char this[FNAMELEN], resp[FNAMELEN], msg[FNAMELEN];

	init_tags();
	if (tags_from_bsc)
		error("Can't modify a .BSC file.");
	t = zap(TAGGED_FILES_BUF);
	get_tagged_file_names(t);

	save_var bufnum, window_bufnum;
	to_buffer_num(t);
	p = tmp_buf();
	buf_xfer(p, 0, size());
	point = 0;
	stuff("# Press Ctrl-X Ctrl-Z to untag all files not named below.\n");
	stuff("# Only files remaining in this list will keep their tags.\n\n");
	recursive_edit();
	if (!buf_exist(t) || !buf_exist(p)) {
		say("No files untagged.");
		return;
	}
	to_buffer_num(t);
	simple_re_replace(1, "^(# .*)?\n", "");
	case_fold = sort_case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	bufnum = p;
	case_fold = sort_case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	restore_vars();
	do_compare_sorted(t, p, NULL, "-untag-these-files", NULL);
	buf_delete(p);

	save_var bufname = "-untag-these-files";
	count = lines_between(0, size(), 0);
	restore_vars();
	sprintf(msg, (count == 1) ? " Untag this %d file? "
							  : " Untag these %d files? ", count);
	if (!(is_gui && want_gui_prompts))
		say("%s", msg + 1);
	else
		say("");
	if (count) {
		save_var _view_title = msg;
		save_var paging_retains_view = 1;
		view_buffer("-untag-these-files", 0);
		if (is_gui && want_gui_prompts && key == abort_key)
			count = 0;		// User clicked Cancel.
		else if (is_gui && want_gui_prompts) {	// Already prompted so don't
			if (toupper(key) != 'Y' && !IS_ENTER_KEY(key) // ask again.
				&& toupper(key) != 'O')
				count = 0;
			ungot_key = -1;
		} else {
			get_string(resp, msg + 1);
			if (toupper(*resp) != 'Y')
				count = 0;
		}
		user_abort = 0;
	}
	if (count) {
		save_var bufname = "-untag-these-files";
		for (point = 0; point < size(); nl_forward()) {
			grab(give_begin_line(), give_end_line(), this);
			if (size() > 1000)
				noteput("Deleting tags: %2d%%...", muldiv(point, 100, size()));
			do_delete_tags(this);
		}
		do_save_tags();
		say("%d file%s untagged.", count, count == 1 ? "" : "s");
	} else
		say("No files untagged.");

	restore_vars();
	delete_buffer("-untag-these-files");
	buf_delete(t);
}

when_loading()
{
	strcpy(ssh_no_user_template, "%c %h");
	strcpy(ssh_template, "%c -l %u %h");
	strcpy(recognize_password_pattern, 
		   "^((Enter |type the |.*@.* )?pass(word|phrase)( for .*)?"
		   "|Password: *\nResponse):[ \t]*");
	drop_name("load-one-spell-word-list");
	drop_name("scp-command-variable");
}
