/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2006 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 13 B8 to 13 B9, assuming further updaters.

#include "eel.h"
#include "colcode.h"
#include "c.h"
#include "session.h"
#include "perl.h"
#include "diff.h"
#include "dired.h"
#include "epsnet.h"
#include "proc.h"
#include "grep.h"
#include "html.h"
#include "tags.h"
#include "tex.h"

#include "vspell.e"

color_class tex_keyword = color_class tex_text;

command do_c_indent()
{
	int orig = point, ind = get_c_indent();
	int orig_column = current_column();

	if (maybe_indent_rigidly(0))
		return;
	to_indentation();
	if (!c_tab_always_indents && orig_column > current_column() || in_shell_buffer) {
					/* if not in indentation */
		point = orig;
		if (this_cmd != CMD_INDENT_REG)
			to_column(orig_column + tab_size - orig_column % tab_size);
	} else if (prev_cmd == C_INDENT)	/* repeated, make bigger */
		to_column(orig_column + ind);
	else if ((c_extra_keywords & IDL_KEYWORDS))
		prev_indenter();
	else {
		c_indenter();
		if (in_perl_buffer && !get_indentation(point) && !orig_column
			  && this_cmd != CMD_INDENT_REG)
			// Nothing to indent from, but indent anyway.
			if (sublanguage_code != SUBLANG_PHP)
				to_column(ind);
			else
				to_column(php_top_level_indent);
	}
	if (this_cmd != CMD_INDENT_REG)
		this_cmd = C_INDENT;
}

in_c_comment(lookback)		// is point inside a c comment, string, etc?
{
	int i = -1, orig = point;

	if (in_perl_buffer)
		return in_perl_comment(point);
	if (lookback) {			// If !lookback, check only this line.
		save_var narrow_start = point - lookback;
		save_var point;
		for (; re_search(RE_REVERSE | RE_FASTEST, "/<*>|<*>/"); point++) {
			i = scan_for_string();	// Parse just this line.  A string?
			if (curchar() == '/')
				switch (character(point - 1)) {
					case '/':		// Found //*, so no block comment.
						continue;
					case '*':		// Found */*, end of comment, not start.
						point--;
						break;
				}
			if (!i || i == IN_OLD_COMMENT)
				break;		// The delimiter wasn't quoted.
			if (curchar() == '*' && c_in_old_comment()) {
				save_var narrow_start = point;
				break;	// Found */ sequence, ends a comment?
			}	// If it looks like it was quoted, double-check.
		}
		if (i != -1 && curchar() == '/' && character(point + 1) == '*')
			return IN_OLD_COMMENT;	// Found /* sequence.
	}
	point = orig;	// Not in a block comment, so use a local check.
	return scan_for_string();
}

command copy_formatting_as_html()
{
	int buf = zap("#html-clipboard#");
	int color, colstart, colend, tab = tab_size;
	save_var point, mark;
	fix_region();
	buf_xfer_colors(buf, point, mark);		// Make a copy of the region.
	save_var bufnum = buf;
	tab_size = tab;
	tab_convert(0, size(), 0);				// Untabify it.
	color_perform_unicode_conversion(bufnum, "utf8");	// To UTF8.
	point = 0;						// Convert special chars to HTML.
	simple_re_replace(1, "%&", "&amp;");
	simple_re_replace(1, "%<", "&lt;");
	simple_re_replace(1, "%>", "&gt;");
	spot before = alloc_spot(), after = alloc_spot(1);
	on_exit free_spot(before), free_spot(after);
	while (point < size()) {	// Insert HTML for the colors.
		color = get_character_color(point, &colstart, &colend);
		if (color == -1)
			color = color_class text;
		*before = *after = colend;
		bprintf("<span style=\"background: #%06x\"><span style=\"color: #%06x\">",
				color_to_html(get_background_color(color)),
				color_to_html(get_foreground_color(color)));
		if (get_foreground_color(color) & EFONT_ITALIC)
			bprintf("<i>");
		if (get_foreground_color(color) & EFONT_BOLD)
			bprintf("<b>");
		if (get_foreground_color(color) & EFONT_UNDERLINED)
			bprintf("<u>");
		point = *before;
		if (get_foreground_color(color) & EFONT_UNDERLINED)
			bprintf("</u>");
		if (get_foreground_color(color) & EFONT_BOLD)
			bprintf("</b>");
		if (get_foreground_color(color) & EFONT_ITALIC)
			bprintf("</i>");
		bprintf("</span></span>");
		point = *after;
		check_abort();
	}				// Now simplify the HTML coloring just inserted.
	html_clean_repeated("%<span style=\"background: #(<word>+)\"%>", "%</span%>");
	html_clean_repeated("%<span style=\"color: #(<word>+)\"%>", "%</span%>");
	html_mode();
	point = size();
	bprintf("</pre>");
	point = 0;
	bprintf("<pre>");
	point = 0;
	mark = size();
	copy_to_clipboard();
	restore_vars();
	highlight_off();
	say("Selection copied to clipboard.");
	if (!is_win32)
		return;
	int frag = tmp_buf();	// For Win32 only, make an HTML Fragment version.
	save_var bufnum = frag;
	grab_buffer(buf);
	simple_re_replace(1, "\n", "\r\n");		// Data will be raw binary.
	translation_type = FILETYPE_BINARY;
	point = size() - 6;		// Before </pre>.
	bprintf("<!--EndFragment-->");
	point = size();
	bprintf("</body></html>");
	point = 0;
	bprintf("Version:0.9\r\nStartHTML:00000097\r\nEndHTML:%08d\r\n",
			size() + 135);		// Create char offsets.
	bprintf("StartFragment:00000135\r\nEndFragment:%08d\r\n",
			size() + 46);
	bprintf("<html><body>");
	point += 5;			// After <pre>.
	bprintf("<!--StartFragment -->");
	if (!buffer_to_clipboard(bufnum, CLIP_ADD_FORMAT, html_clipboard_format()))
		say("Error %d copying HTML fragment to clipboard.", errno);
	restore_vars();
	buf_delete(frag);
}

// Copy some text to the clipboard.  Flags: 1=add a newline after it,
// 2=don't display msg.
copy_line_to_clipboard(char *line, int flags)
{
	int tmp = tmp_buf();
	buf_printf(tmp, "%s%s", line, (flags & 1) ? "\n" : "");
	buffer_to_clipboard(tmp, CLIP_CONVERT_NEWLINES, 0);
	buf_delete(tmp);
	if (!(flags & 2))
		say("Copied to clipboard: %s", line);
}

// Create class list in the current buffer.
make_color_class_list(int expanded)
{
	int i, classes = number_of_color_classes();

	// First, set up a good top-level order.
	stuff("-General\n-Language Modes\n-Special-Purpose Modes\n-Dialogs\n");
	stuff("-Non-GUI Dialogs\n-Non-GUI Menu Bar\n");
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	tab_size = 100;		// Hide rest of tree.
	save_var narrow_start, narrow_end;
	for (i = 0; i < classes; i++) {
		char cat[FNAMELEN], pat[FNAMELEN], *s, firstcat[FNAMELEN];
		set_color_give_category(i, cat);
		point = 0;
		strcpy(firstcat, cat);
		if (s = strchr(firstcat, '|'))
			*s = 0;
		sprintf(pat, "^-%q\n", firstcat);
		if (!re_search(1, pat))
			bprintf("-%s\n", firstcat);
		narrow_start = point;
		if (re_search(1, "^-"))
			to_begin_line();
		narrow_end = size() - point;
		if (s = strchr(cat, '|')) {
			sprintf(pat, "^  -%q\n", s + 1);
			point = 0;
			if (!re_search(1, pat))
				bprintf("  -%s\n", s + 1);
			else
				re_search(1, "^!([^ \n]|  [^ \n])");
			bprintf("  ");
		}
		narrow_start = narrow_end = 0;
		bprintf("  %s\n", name_color_class(i));
	}
	if (!expanded) {	// Now collapse all categories.
		delete(size() - 1, size());
		simple_re_replace(1, "^  -", "  +");
		simple_re_replace(1, "^-", "+");
		simple_re_replace(1, "\n ", "\t ");
	}
}

check_running_processes()
{
	char *buf, *open = 0, *s, msg[FNAMELEN];

	buf = buffer_list(1);
	save_var bufnum;
	do {
		if (process_warn_on_exit && preserve_buffer_type(buf))
			if (process_warn_on_exit == 2 || !exit_this_process(buf)) {
				open = buf;
				break;
			}
		bufname = buf;
		if (is_process_buffer(bufnum) && discardable_buffer)
			exit_this_process(buf);
	} while (buf = buffer_list(0));
	restore_vars();
	if (open && process_warn_on_exit) {
		tiled_only();
		locate_window(open, "");
		refresh();
		s = preserve_buffer_type(open);
		if (windows_state() & WINSTATE_MINIMIZED)
			windows_restore();
		if (process_warn_on_exit != 2) {
			if (s && !strcmp(s, "Process"))
				error("Couldn't exit from concurrent process.");
			else
				error("Couldn't exit %s job.", s);
		} else {
			if (s && !strcmp(s, "Process"))
				sprintf(msg, "A process is running, confirm exit? ");
			else
				sprintf(msg, "%s job is running, confirm exit? ", s);
			if (!ask_yn("Confirm exit", msg, "Yes", "No", 2))
				error("Exiting canceled.");
		}
	}
}

// Wait until all timed say() messages have expired.  But if there's a
// message with an infinite timeout, accept user keypresses that would
// go to the next message.  In that case, eventually return even if the
// user never presses a key.
wait_for_unseen_msgs()
{
	TIMER tm;
	int res;
	time_begin(&tm, 500);
	while (!in_macro() && (res = unseen_msgs_time())) {
		if (res == -1) {
			if (time_done(&tm))
				return;
			else if (char_avail())
				getkey();
		} else
			time_begin(&tm, 500);
		check_abort();
		delay(1, COND_KEY);
	}
}

int recursive_edit_preserve()
{
	int owin = window_handle, obuf = window_bufnum;
	recursive_edit();
	if (!is_window(owin) || !buf_exist(obuf))
		return 0;
	window_handle = owin;
	if (window_bufnum != obuf)
		to_buffer_num(obuf);
	return 1;
}

int scp_handler(int activity, int buf, int from, int to);


// Find or allocate an scp_job structure associated with this buffer,
// and make a work buffer for it too.
struct scp_job *find_scp_buffer(char *host, char *uname, int op)
{
	char buf[FNAMELEN];
	struct scp_job *p;
	int i, b, separate_xfers;

	for (p = scp_job; p; p = p->next)
		if (!strfcmp(host, p->host) && !strcmp(uname, p->uname)
			  && !p->busy && buf_exist(p->buf)) {
			separate_xfers = p->flags & (SCP_CLIENT_MODE | SCP_SCP_XFER);
			if (!separate_xfers || (op & 2) == (p->op & 2))
				return p;
		}
	for (i = 1; ; i++) {
		sprintf(buf, "scp-helper-%s-%s-%d", host, uname, i);
		if (!exist(buf))
			break;
	}
	b = create(buf);
	p = (struct scp_job *) malloc(sizeof(struct scp_job));
	p->host = strsave(host);
	p->uname = strsave(uname);
	p->buf = b;
	p->next = scp_job;
	scp_job = p;
	save_var bufnum = b;
	when_activity = scp_handler;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	return p;
}

net_during_wait(int cond)
{
	delay(1, cond);
	if (ungot_key == WIN_HELP_REQUEST && menu_command
		  && !strcmp(menu_command, "password-prompt")) {
		ungot_key = -1;
		password_prompt();
	}
}	

struct scp_job *do_scp(struct url_parts *parts, int op);

scp_get_response(int code, char *url, struct url_parts *parts, int b)
{
	char cmd[FNAMELEN], filter[FNAMELEN], *s, *val;
	struct scp_job *p;

	save_var bufnum = b;
	buf_zap(bufnum);
	p = do_scp(parts, FTP_MISC);
	strcpy(p->file, url);
	*get_url_file_part(p->file, 0) = 0;
	if (!(p->flags & SCP_CLIENT_MODE)) {
		switch (code) {
			case 1:
				s = "";
				break;
			case 2:
				p->flags |= SCP_CHECK_IS_DIR;
				s = "cd \"%s\"";
				break;
			default:		// Cases 0 and 3: list matches.
				p->flags |= SCP_CHECK_MATCHES;
				val = get_str_var(find_variable_by_host("scp-list-flags",
					parts->host));
				sprintf(s = cmd, "ls %s", (p->flags & SCP_OLD_SFTP) ? ""
					: val ? val : "-l");
				if (s[strlen(s) - 1] != ' ')
					strcat(s, " ");
				strcat(s, "%s");
				strcpy(filter, get_tail(parts->fname, 1));
				*get_tail(parts->fname, 1) = 0;	// Must list all files.
				break;
		}
		scp_send_some_command(parts, p, s);
	} else {
		sprintf(cmd, (code == 1 || code == 2) ? "prdir %s\n" : "comp %s\n",
				parts->fname);
		p->cmd = p->cmdlist = strsave(cmd);
		if (p->busy != 2)
			scp_next_cmd(p);
	}
	delayed_say(0, 100, 30, "Working...");
	while (p->busy && !user_abort)
		net_during_wait(COND_KEY);
	set_buf_mode_message(bufnum, NULL);
	user_abort = 0;
	delayed_say(0, 0, 0, "");
	say("");
	point = 0;
	if (!(scp_flags(parts) & SCP_CLIENT_MODE)) {
		switch (code) {
			case 0:
			case 3:		// Discard the extra info and filter nonmatches.
				scp_filter_sftp_matches(filter, code != 3, url);
				break;
			case 1:
				buf_printf(b, "%s%s", p->file, p->lastdir);
				set_buf_point(b, 0);
				break;
		}
	}
}

ftp_file_match(char *pat, int flags)
{
	char mypat[FNAMELEN * 2], name[FNAMELEN * 2], *p;

	save_var bufnum = zap("-ftp-file-match-listing");
	strcpy(mypat, pat);
	if (!(flags & EXACTONLY))
		strcat(mypat, "*");
	strcpy(name, mypat);
	p = get_tail(name, 1);	// If name contains wildcards, strip them.
	if (index(p, '*') || index(p, '?')) {
		*p = 0;								// Make FTP listing
		save_var ftp_compatible_dirs |= 1;	// show just the basename.
	}
	url_operation(name, FTP_LIST | FTP_WAIT);
	point = 0;
	bprintf("     Directory of %s\n", name);
	while (re_search(1, "^.")) {
		point--;
		stuff("  ");
		point++;
	}
	modified = 0;
	dired_orig_position = point = 0;
	dired_ftp_activity(NET_DONE, bufnum, 0, 0);
	color_dired_range(0, size());
	int b = zap(_MATCH_BUF);
	point = 0;
	for (point = 0; point < size() && nl_forward(); )
		if (dired_getname(name, 2) && (dired_isdir() ? !(flags & FM_NO_DIRS)
			: !(flags & FM_ONLY_DIRS)))
			if (fpatmatch(name, mypat, 0, 0))
				buf_printf(b, "%s%s\n", name, dired_isdir() == 1 ? "/" : "");
	set_buf_point(b, 0);
}

wait_for_remote_file(int b)
{
	save_var bufnum = b;
	for (;;) {
		if (!ftp_job && is_busy_doing_scp(b) == -1)
			return;
		net_during_wait(0);
		check_abort();
	}
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

// Check for file vars in the current buffer, then apply to target
// buffer.
check_buffer_filevars(int target)
{
	char buf[FNAMELEN], var[FNAMELEN], pat[2048], *list, *s;
	int i, cur = bufnum;

	save_var bufnum;
	if (look_for_filevar_pattern("tab-(size|width)", "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			tab_size = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(margin-right|fill-column)",
								 "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			margin_right = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(auto-)?fill-mode",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			call_maybe_break_line(i);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(indent-with-tabs|indent-tabs-mode)",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		indent_with_tabs = numtoi(buf);
		bufnum = cur;
	}
	list = "c-indent|soft-tab-size|over-mode|auto-indent"
			"|goal-column|auto-fill-indents|virtual-space"
			"|want-backups"
			"|case-fold|sort-case-fold|delete-hacking-tabs"
			"|undo-size|indents-separate-paragraphs"
			"|tex-paragraphs|comment-column|want-warn"
			"|auto-read-changed-file|perl-indent"
			"|html-paragraph-is-container|tex-force-latex"
			"|concurrent-compile|vbasic-indent";
	sprintf(pat, "(%s)", list);
	if (look_for_filevar_pattern(pat, "([0-9]+|nil)", NULL)) {
		for (;;) {
			strncpy(var, list, (sizeof(var) / sizeof(*var)) - 1);
			if (s = strchr(var, '|'))
				*s = 0;
			if (look_for_filevar_pattern(var, "([0-9]+|nil)", NULL)) {
				grab(find_group(1, 1), find_group(1, 0), buf);
				bufnum = target;
				if (i = find_index(var))
					do_setting(i, buf, 0);
				else
					say("Unknown mode variable %s.", var);
				bufnum = cur;
			}
			if (s = strchr(list, '|'))
				list = s + 1;
			else
				break;
		}
	}
}

grep_get_curdir(char *dir)
{
	save_var point;
	search(-1, "\nFound ");
	point += parse_string(1, "\nFound .* lines with ");
	if (curchar() == '"') {
		point++;
		search(1, "\"");
	}
	if (parse_string(1, ".*%(cur dir \"(.*)\"%)\n"))
		grab(find_group(1, 1), find_group(1, 0), dir);
	else
		getcd(dir);
}

// Set the text in a popup window used for help.  Fill the text using
// the window's current width, then adjust its height to match the
// text, and position it just above the echo area.
show_popup_help(int win, char *text)
{
	save_var window_handle = win;
	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	buf_zap(bufnum);
	stuff(text);
	margin_right = window_width - 4;
	fill_region();
	int lines = lines_between(0, size(), 0) + 1;
	window_height = lines + 2;
	window_top = screen_lines - lines - 5;
	display_scroll_bar = 0;
	restore_vars();
	refresh();
}

do_php_indent()
{
	save_var in_perl_buffer = 1;
	save_var sublanguage_code = SUBLANG_PHP;
	save_var perl_indent = php_indent;
	save_var perl_closeback = php_closeback;
	save_var perl_topindent = php_topindent;
	save_var perl_label_indent = php_label_indent;
	save_var perl_contin_offset = php_contin_offset;
	save_var perl_align_contin_lines = php_align_contin_lines;
	save_var perl_brace_offset = php_brace_offset;
	save_var perl_top_braces = php_top_braces;
	save_var perl_top_contin = php_top_contin;
	save_var perl_top_struct = php_top_struct;
	do_c_indent();
}

int do_list_element_nesting(int b, int flags)
{
	char tag[FNAMELEN], depth[FNAMELEN];
	save_var point, case_fold = (sublanguage_code != SUBLANG_XML);
	int open = html_in_tag() != -1;
	while (do_color_searching(REVERSE | REGEX, "<c:html-punctuation>[<]") == 1) {
		if ((flags & 2) && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		set_buf_point(b, 0);
		int i = html_is_start_tag(point, 0);
		if (i == -1) {
			i = do_html_find_matching_tag(point, tag);
			if (i >= 0) {
				point = i;
				search(-1, "<");	// Get from element name to before its <.
			} else if (!(flags & 1))
				buf_printf(b, "%s? ", tag);
			else if (i == -1)
				buf_printf(b, "</%s> unmatched at line %d, position %d.\n",
						   tag, lines_between(0, point) + 1, point);
			else
				buf_printf(b, "Can't recognize tag at line %d, position %d.\n",
						   lines_between(0, point) + 1, point);
		} else if ((i == 1 || !i && open) && html_get_tag_name(point, tag) == 1) {
			char *bad = html_invalid_tag(point + 1) ? "?" : "";
			if ((flags & 1) && *bad)
				buf_printf(b, "Unclosed <%s at line %d, position %d.\n",
						   tag, lines_between(0, point) + 1, point);
			else
				buf_printf(b, !(flags & 1) ? (open ? "<%s%s " : "%s%s ")
										   : (open ? "<%s%s\n" : "<%s%s>\n"),
						   tag, bad);
		}
		open = 0;
	}
	if (flags & 1) {
		save_var bufnum = b, point = 0;
		*depth = 0;
		while (re_search(1, "^[<]" HTML_NAME)) {
			set_character_color(matchstart + 1, matchend, color_class html_tag);
			point = matchstart;
			bprintf("%s", depth);
			strcat(depth, "  ");
			to_end_line();
		}
	}
	return 1;
}

get_soft_tab_size()
{
	if (soft_tab_size > 0)
		return soft_tab_size;
	int tab = get_mode_variable("%s-soft-tab-size");
	if (!got_bad_number && tab > 0)
		return tab;
	return tab_size;
}

/* Delete the previous character */
command backward_delete_character()
{
	if (delete_if_highlighted())
		return;
	if (virtual_space && cursor_to_column > 0)
		virtual_move_back();
	else {
		int tab = get_soft_tab_size();
		int hack = get_mode_variable("%s-delete-hacking-tabs");
		if (got_bad_number)
			hack = delete_hacking_tabs;
		int many = (hack & 2) && character(point - 1) == ' ' && !has_arg;
		if (many && (hack & 4) && current_column() - 1 > get_indentation(point))
			many = 0;
		if (many && (hack & 8) && current_column() % tab)
			many = 0;
		if (many && (hack & 16) && (curchar() == ' ' || curchar() == '\t'))
			many = 0;
		if (hack & 1)
			hack_tabs(-iter);
		do_delete_characters(-1);
		if (many)
			while (current_column() % tab && parse_string(-1, "[ \t]"))
				delete(point - 1, point);
	}
}

process_find_directory(char *file, int buf, int pos)
{
	char dir[FNAMELEN], url[FNAMELEN];
	int level = 0, got = 0, c;

	save_var bufnum = buf, point = pos;
	*dir = 0;
	if (is_process_buffer(bufnum) && *process_current_directory)
		strcpy(dir, process_current_directory);
	while (!got && re_search(-1, "(Entering|Leaving|Running `.*' in)"
							 " directory `(.*)'$")) {
		switch (c = character(find_group(1, 1))) {
			case 'L':	level++; break;
			case 'E':	if (--level >= 0)
							break;
			case 'R':	got = c;
						grab(find_group(2, 1), find_group(2, 0), dir);
						break;
		}
	}
	if (telnet_error_converter(file, *dir ? dir : NULL, url)) {
		strcpy(file, url);
		return;
	}
	if (got != 'R')	// "Running" message is from us, so don't translate it.
		convert_cygwin_filename(dir, 0);
	if (opsys == OS_DOS && strchr(file, '/'))
		convert_cygwin_filename(file, 0);
	if (*dir)
		absolute(file, dir);
	else
		absolute(file);
}

tag_mode_perl()
{
	char func[TAGLEN], pat[FNAMELEN];
	int start, i;

	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	sprintf(pat, "<^word>(%s)[ \t]+([a-zA-Z_]<word>*"
			"((::|')[a-zA-Z_]<word>*)?)",
			(sublanguage_code == SUBLANG_PHP) ? "function" : "sub");
	while (re_search(1, pat)) {
		if (get_character_color(find_group(1, 1)) == color_class perl_keyword
			&& get_character_color(find_group(2, 1)) == color_class perl_function) {
			grab(start = find_group(2, 1), find_group(2, 0), func);
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (curchar() == '(') {
				i = perl_move_level(1, 0);
				if (i > 0 && i < size() - narrow_end)
					point = i;
			}
			point += parse_string(1, "([ \t\n]|(#.*))*");
			if (curchar() == '{' || (tag_declarations & 1))
				add_tag(func, start);
		}
	}
}

paren_cmds()		/* handle parenthesized commands */
{
	char funcname[80], numarg[80];
	int i;

	must_char('(');
	re_search(1, "[ \t\n]*");
	*numarg = 0;
	parse_string(1, "-?[0-9][0-9a-z]*[ \t\n]+", numarg);
	parse_string(1, "[^ \t\n()\"']*", funcname);
	if (!strcmp(funcname,"define-macro"))
		cf_macro();
	else if (!strcmp(funcname, "bind-to-key"))
		cf_bind();
	else if (!strcmp(funcname, "create-prefix-command"))
		cf_prefix();
	else if (!strcmp(funcname, "set-variable"))
		cf_set_variable();
	else if (!strcmp(funcname, "change-name"))
		cf_change_name();
	else if (i = is_runnable(funcname))
		cf_any(i, numarg);
	else
		init_error("Bad command %s in command file. ", funcname);
	re_search(1, "[ \t\n]*");
	must_char(')');
}

int *keyseq;		/* sequence of key codes, in format used by macros */

cf_macro()		/* handle define-macro */
{
	char name[80];

	parse_var_name(name);
	get_keyseq();
	name_macro(name, keyseq);
}

cf_bind()		/* handle bind-to-key */
{
	int index;
	char name[80];

	parse_var_name(name);
	index = find_index(name);
	if (index)
		put_on_key(index);
	else
		init_error("No such command %s.", name);
	rebuild_menu();
}

// Parse a variable name, optionally quoted, and put in s.
parse_var_name(char *s)
{
	re_search(1, "[ \t\n]*");
	if (!parse_string(1, "[-a-zA-Z][-a-zA-Z0-9]*", s))
		get_str(s);
}

// Get variable name, value, possible default option, and set.
cf_set_variable()
{
	char varname[FNAMELEN], val[FNAMELEN], def[FNAMELEN];
	int i;

	parse_var_name(varname);
	point += parse_string(1, "[ \t\n]*");
	if (!parse_string(1, "-?[0-9][0-9a-zA-Z]*", val))
		get_str(val);
	point += parse_string(1, "[ \t\n]*");
	*def = 0;
	parse_string(1, "-?[0-9][0-9a-zA-Z]*", def);
	i = find_index(varname);
	if (!vartype(i))
		init_error("%s: not a variable", varname);
	if (!var_settable(varname))
		init_error("%s: can't set variable of this type", varname);
	safe_set_var_val(i, val, *def ? numtoi(def) : 3);
}

cf_change_name()
{
	char old[FNAMELEN], new[FNAMELEN];

	parse_var_name(old);
	point += parse_string(1, "[ \t\n]*");
	parse_var_name(new);
	switch (replace_name(old, new)) {
		case 1: init_error("%s: not found!", old); break;
		case 2: init_error("%s exists, delete first", new); break;
	}
}

load_tags(file)
char *file;
{
	int i;

	save_var synchronous_file_io = 1;
	tags_from_bsc = 0;
	*_last_tag = *_last_tag_file = 0;
	if (!strfcmp(get_extension(file), ".bsc")) {
		tags_from_bsc = 1;
		strcpy(bsc_file, file);
		absolute(bsc_file);
		return;
	}
	save_var bufnum = _tags_buffer = zap("-tags");
	i = read_file(file, FILETYPE_AUTO);
	if (i == 2)			/* new file */
		mark_tags(0);		/* make new-fmt tags file */
	else if (i) {			/* error reading file */
		restore_vars();
		buf_delete(_tags_buffer);
		quick_abort();
	} else if (!tags_header()) {
		if (!ask_yn("Convert", "Convert old-format tags file? ",
					"Yes", "No", 1))
			quick_abort();
		sayput("Converting old-format tags file...");
		if (!string_replace(";", "\t", 0)) {
			mark_tags(0);
			do_save_tags();
		}
	}
}

command retag_files()
{
	char file[FNAMELEN], tdir[FNAMELEN];
	int t, aborted = 0;

	init_tags();
	save_var synchronous_file_io = 1;
	if (tags_from_bsc)
		error("Can't modify a .BSC file.");
	say("");
	if (!ask_yn("Retag files", "Retag all files? ", "Yes", "No", 1))
		return;
	t = zap(TAGGED_FILES_BUF);
	get_tagged_file_names(t);
	to_another_buffer("-tag-errors");
	delete_buffer("-tag-errors");
	save_var tag_batch_mode = 1;
	put_tag_directory(tdir);
	while (grab_line(t, file) > 0 && !user_abort) {
		absolute(file, tdir);
		if (do_retag_file(file))
			user_abort = aborted = 1;
	}
	if (!aborted)
		do_save_tags();
	if (exist("-tag-errors")) {
		to_buffer("-tag-errors");
		bprintf("\nOld tags for the above files were removed.\n");
		point = 0;
		discardable_buffer = 1;
	}
	if (!aborted)
		say("All files retagged.");
}

command tag_files()
{
	char pat[FNAMELEN], *s, file[FNAMELEN];
	int aborted = 0;

	init_tags();
	save_var synchronous_file_io = 1;
	if (tags_from_bsc)
		error("Must tag files using the BSCMAKE program.");
	if (has_arg)
		save_var tag_declarations = 3;
	get_file(pat, "Add/update tags for files matching", filename);
	iter = 0;
	if (!(s = do_file_match(pat, STARTMATCH | EXACTONLY | FM_NO_DIRS)))
		error("No matches");
	to_another_buffer("-tag-errors");
	delete_buffer("-tag-errors");
	save_var tag_batch_mode = 1;
	for (; s; s = do_file_match(pat, EXACTONLY | FM_NO_DIRS)) {
		strcpy(file, s);
		absolute(file);
		if (aborted = do_retag_file(file))
			break;
	}
	if (!aborted) {
		do_save_tags();
		say("%s tagged.", pat);
	}
	if (exist("-tag-errors")) {
		to_buffer("-tag-errors");
		discardable_buffer = 1;
	}
}

tag_mode_asm()	/* tag all labels or procs in the file */
{
	char func[TAGLEN];
	int start;

	save_var case_fold = 1, point = 0;
	while (re_search(1, "^[ \t]*([a-z0-9@$_]+)[ \t]*:")) {
		grab(start = find_group(1, 1), find_group(1, 0), func);
		add_tag(func, start);
	}
	point = 0;
	while (re_search(1, "^[ \t]*([a-z0-9@$_]+)[ \t]+proc[ \t\n;]+")) {
		grab(start = find_group(1, 1), find_group(1, 0), func);
		add_tag(func, start);
	}
}

tag_mode_latex()
{
	char tag[TAGLEN], pat[FNAMELEN * 2];
	int start, end;

	recolor_buffer_range(0, size());
	save_var case_fold = 0, point = 0;
	sprintf(pat, "\\(%s){([^{}]+)}", latex_tag_keywords);
	while (do_color_searching(REGEX, pat) == 1) {
		start = find_group(2, 1);
		end = find_group(2, 0);
		if (end - start > sizeof(tag) - 2)
			continue;
		grab(start, end, tag);
		add_tag(tag, start);
	}
}

when_loading()
{
	if (name_type(find_index("php-top-indent")) == NT_VAR)
		php_top_level_indent = get_num_var(find_index("php-top-indent"));
	drop_name("php-top-indent");
	drop_name("spell-buffer");
	server_raises_window = 1;		// New defaults.
	if (!strstr(grep_ignore_file_extensions, ".pyd"))
		strcat(grep_ignore_file_extensions, ".asv#|.class|.pll|.pyd|");
	if (strstr(scp_windows_sftp_command, "winpty"))
		strcpy(scp_windows_sftp_command, "%X\winpty -a -n sftp");
	if (strstr(ssh_command_windows, "winpty"))
		strcpy(ssh_command_windows, "%X\winpty -a ssh");
	c_spell_options &= 1;			// Clear the reassigned bits.
	default_spell_options &= 1;
	fundamental_spell_options &= 1;
	html_spell_options &= 1;
	latex_spell_options &= 1;
	tex_spell_options &= 1;
	xml_spell_options &= 1;
}
